// LI2/ENIB - AReVi-Tools - BM font builder
//----------------------------------------------------------------------------
// file        : Tools/BMFontBuilder/bmFontBuilder.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This tool loads a X11 font and generates a bitmap font file
//  suitable for AReVi.
//
//----------------------------------------------------------------------------

#include "AReVi/arSystem.h"
#include "AReVi/Utils/fileStream.h"
#include "AReVi/Lib3D/util3D_priv.h"

using namespace AReVi;

#include <X11/Xlib.h>
#include <X11/Xutil.h>

//----------------------------------------------------------------------------

class BMFont : public ArObject
{
public:

  AR_CLASS(BMFont)

  AR_CONSTRUCTOR_1(BMFont,
                   const StlString &, name)

  virtual
  void
  writeToStream(ArRef<AbstractOStream> stream) const;

protected:
  Display * _dpy;
  XFontStruct * _font;
  Pixmap _pix;
};

//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(BMFont,ArObject)

BMFont::BMFont(ArCW & arCW,
               const StlString & name)
: ArObject(arCW),
  _dpy((Display *)0),
  _font((XFontStruct *)0),
  _pix((Pixmap)0)
{
_dpy=XOpenDisplay("");
if(!_dpy)
  {
  setErrorMessage("Cannot open display");
  return;
  }
_font=XLoadQueryFont(_dpy,name.c_str());
if(!_font)
  {
  setErrorMessage("Cannot load font "+name);
  return;
  }
if(_font->min_byte1||_font->max_byte1)
  {
  setErrorMessage("16 bit fonts are not supported");
  return;
  }
unsigned int height=_font->ascent+_font->descent;
unsigned int maxWidth=_font->max_bounds.rbearing;
_pix=XCreatePixmap(_dpy,DefaultRootWindow(_dpy),
                   maxWidth,height,24);
if(!_pix)
  {
  setErrorMessage("Cannot create pixmap");
  return;
  }
}

BMFont::~BMFont(void)
{
if(_pix)
  {
  XFreePixmap(_dpy,_pix);
  }
if(_font)
  {
  XFreeFont(_dpy,_font);
  }
if(_dpy)
  {
  XCloseDisplay(_dpy);
  }
}

void
BMFont::writeToStream(ArRef<AbstractOStream> stream) const
{
GC textGC=XCreateGC(_dpy,DefaultRootWindow(_dpy),0,(XGCValues *)0);
XSetForeground(_dpy,textGC,0x00000000);
XSetFont(_dpy,textGC,_font->fid);
GC clearGC=XCreateGC(_dpy,DefaultRootWindow(_dpy),0,(XGCValues *)0);
XSetForeground(_dpy,clearGC,0x00ffffff);
unsigned char data[0x100];
unsigned int ascent=_font->ascent;
unsigned int descent=_font->descent;
unsigned int height=ascent+descent;
unsigned int maxWidth=_font->max_bounds.rbearing;
if(!stream->writeUChar('A')||
   !stream->writeUChar('r')||
   !stream->writeUChar('B')||
   !stream->writeUChar('M')||
   !stream->writeUChar('F')||
   !stream->writeUChar(ascent)||
   !stream->writeUChar(descent))
  {
  return;
  }
for(unsigned int code=0;code<256;code++)
  {
  char str[2];
  str[0]=(char)code;
  str[1]='\0';
  unsigned int width=XTextWidth(_font,str,1);
  if(!width)
    {
    str[0]=' ';
    width=XTextWidth(_font,str,1);
    }
  unsigned int lineSize=(width+7)>>3;
  unsigned int dataSize=lineSize*height;
  if(dataSize>255)
    {
    const_cast<BMFont *>(this)->setErrorMessage(
                                "Data size is greater than 255 for symbol "+
                                uintToStr(code));
    cerr << getErrorMessage() << endl;
    return;
    }
  XFillRectangle(_dpy,_pix,clearGC,0,0,maxWidth,height);
  XDrawString(_dpy,_pix,textGC,0,ascent,str,1);
  XImage * image=XGetImage(_dpy,_pix,0,0,maxWidth,height,AllPlanes,ZPixmap);
  memset(data,0,0x100);
  for(unsigned int y=0;y<height;y++)
    {
    for(unsigned int x=0;x<width;x++)
      {
      unsigned int pixel=XGetPixel(image,x,height-(y+1));
      if(!pixel)
        {
        data[(y*lineSize)+(x>>3)]|=(1<<(7-(x&7)));
        }
      }
    }
  if(!stream->writeUChar(width)||
     !stream->writeUChar(dataSize)||
     !stream->writeData(data,dataSize))
    {
    return;
    }
  XDestroyImage(image);
  }
XFreeGC(_dpy,textGC);
XFreeGC(_dpy,clearGC);
}

//----------------------------------------------------------------------------

void registerBMFontClasses(void);

int
main(int argc,
     char ** argv)
{
ArSystem arevi(argc,argv,false);
registerBMFontClasses();
if(argc<3)
  {
  cerr << endl;
  cerr << "usage: " << argv[0] << " X11_font_name output_file.arbmf" << endl;
  cerr << endl;
  return(-1);
  }
ArRef<BMFont> font=BMFont::NEW(argv[1]);
if(font->fail())
  {
  cerr << font->getErrorMessage() << endl;
  }
else
  {
  ArRef<FileOStream> stream=FileOStream::NEW(argv[2],false);
  if(stream->fail())
    {
    cerr << stream->getErrorMessage() << endl;
    }
  else
    {
    font->writeToStream(stream);
    if(stream->fail())
      {
      cerr << "Problem while writing to "
           << stream->getFileName() << " !" << endl;
      cerr << stream->getErrorMessage() << endl;
      }
    }
  }
return(0);
}

