// LI2/ENIB - AReVi-Tools - GL font builder
//----------------------------------------------------------------------------
// file        : Tools/GLFontBuilder/glFontBuilder.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This tool uses the FreeType2 API (http://www.freetype.org).
//
//  It reads a true-type font file and generates a 3D font file
//  suitable for AReVi.
//
//----------------------------------------------------------------------------

#include "AReVi/arSystem.h"
#include "AReVi/Utils/fileStream.h"
#include "AReVi/Lib3D/util3D_priv.h"

using namespace AReVi;

#include <GL/glu.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#define SIMPLIFY_PRECISION 1e-2

//----------------------------------------------------------------------------

class GLFontSymbol : public ArObject
{
public:

  AR_CLASS(GLFontSymbol)

  AR_CONSTRUCTOR_3(GLFontSymbol,
                   unsigned int, code,
                   FT_GlyphSlot, glyph,
                   double, scale)
  virtual
  void
  writeToStream(ArRef<AbstractOStream> stream) const;

protected:

#ifdef __CYGWIN__
  typedef void (__attribute__ ((__stdcall__)) * GLUTessCB)();
#else // not __CYGWIN__
  typedef void (*GLUTessCB)();
#endif // not __CYGWIN__

  virtual
  void
  _simplify(double precision);

  static
  void
  _edgeFlagCB(GLboolean /* flag */,
           GLFontSymbol * /* symbol */);

  static
  void
  _errorCB(GLenum /* errorCode */,
           GLFontSymbol * symbol);

  static
  void
  _beginCB(GLenum primitive,
           GLFontSymbol * symbol);

  static
  void
  _vertexCB(unsigned int index,
            GLFontSymbol * symbol);

  static
  void
  _combineCB(GLdouble coords[3],
             unsigned int /* vertexData */ [4],
             GLfloat /* weight */ [4],
             unsigned int * dataOut,
             GLFontSymbol * symbol);

  static
  void
  _endCB(GLFontSymbol * /* symbol */);

protected:
  unsigned int _code;
  vector<Util3D::Dbl3> _vertices;
  vector<Util3D::Face> _triangles;
  vector<unsigned int> _lines;
  unsigned int _vertexCount;
  Util3D::Face _currentFace;
};

//----------------------------------------------------------------------------

class GLFont : public ArObject
{
public:

  AR_CLASS(GLFont)

  AR_CONSTRUCTOR_1(GLFont,
                   const StlString &, name)

  virtual
  void
  writeToStream(ArRef<AbstractOStream> stream) const;

protected:
  StlVector<ArRef<GLFontSymbol> > _symbols;
};

//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(GLFontSymbol,ArObject)

GLFontSymbol::GLFontSymbol(ArCW & arCW,
                           unsigned int code,
                           FT_GlyphSlot glyph,
                           double scale)
: ArObject(arCW),
  _code(code),
  _vertices(),
  _triangles(),
  _lines(),
  _vertexCount(0),
  _currentFace()
{
unsigned int nbC=glyph->outline.n_contours;
_lines.reserve(nbC);
unsigned int nbV=(nbC ? glyph->outline.contours[nbC-1]+1 : 0);
_vertices.reserve(2*nbV);
if(glyph->outline.flags&FT_OUTLINE_REVERSE_FILL)
  {
  unsigned int i;
  for(i=0;i<nbV;i++)
    {
    _vertices.push_back(Util3D::Dbl3(glyph->outline.points[i].x/scale,
                                     glyph->outline.points[i].y/scale,
                                     0.0));
    }
  for(i=0;i<nbC;i++)
    {
    _lines.push_back(glyph->outline.contours[i]);
    }
  }
else // reverse for a CCW GLU tesselation
  {
  unsigned int i;
  for(i=nbV;i--;)
    {
    _vertices.push_back(Util3D::Dbl3(glyph->outline.points[i].x/scale,
                                     glyph->outline.points[i].y/scale,
                                     0.0));
    }
  for(i=nbC;i--;)
    {
    unsigned int last=(i ? glyph->outline.contours[i-1]+1 : 0);
    _lines.push_back(nbV-(last+1));
    }
  }
_simplify(SIMPLIFY_PRECISION);
GLUtesselator * tess=gluNewTess();
if(glyph->outline.flags&FT_OUTLINE_EVEN_ODD_FILL)
  {
  gluTessProperty(tess,GLU_TESS_WINDING_RULE,GLU_TESS_WINDING_ODD);
  }
else
  {
  gluTessProperty(tess,GLU_TESS_WINDING_RULE,GLU_TESS_WINDING_NONZERO);
  }
gluTessCallback(tess,(GLenum)GLU_TESS_BEGIN_DATA,
                reinterpret_cast<GLUTessCB>(&GLFontSymbol::_beginCB));
gluTessCallback(tess,(GLenum)GLU_TESS_VERTEX_DATA,
                reinterpret_cast<GLUTessCB>(&GLFontSymbol::_vertexCB));
gluTessCallback(tess,(GLenum)GLU_TESS_COMBINE_DATA,
                reinterpret_cast<GLUTessCB>(&GLFontSymbol::_combineCB));
gluTessCallback(tess,(GLenum)GLU_TESS_END_DATA,
                reinterpret_cast<GLUTessCB>(&GLFontSymbol::_endCB));
gluTessCallback(tess,(GLenum)GLU_TESS_EDGE_FLAG_DATA,
                reinterpret_cast<GLUTessCB>(&GLFontSymbol::_edgeFlagCB));
gluTessCallback(tess,(GLenum)GLU_TESS_ERROR_DATA,
                reinterpret_cast<GLUTessCB>(&GLFontSymbol::_errorCB));
gluTessBeginPolygon(tess,this);
unsigned char index=0;
for(unsigned int i=0;i<nbC;i++)
  {
  gluTessBeginContour(tess);
  for(;index<=_lines[i];index++)
    {
    gluTessVertex(tess,&_vertices[index].x,reinterpret_cast<void *>(index));
    }
  gluTessEndContour(tess);
  }
gluTessEndPolygon(tess);
gluDeleteTess(tess);
if(fail())
  {
  return;
  }
_simplify(SIMPLIFY_PRECISION);
if(_vertices.size()>255)
  {
  setErrorMessage("Too many points for char "+uintToStr(code));
  return;
  }
if(_triangles.size()>255)
  {
  setErrorMessage("Too many triangles for char "+uintToStr(code));
  return;
  }
}

GLFontSymbol::~GLFontSymbol(void)
{
}

void
GLFontSymbol::writeToStream(ArRef<AbstractOStream> stream) const
{
if(!stream->writeUChar(_code)||
   !stream->writeUChar(_vertices.size())||
   !stream->writeUChar(_triangles.size())||
   !stream->writeUChar(_lines.size()))
  {
  return;
  }
unsigned int i;
for(i=0;i<_vertices.size();i++)
  {
  if(!stream->writeFloat(_vertices[i].x)||
     !stream->writeFloat(_vertices[i].y))
    {
    return;
    }
  }
for(i=0;i<_triangles.size();i++)
  {
  if(!stream->writeUChar(_triangles[i].p1)||
     !stream->writeUChar(_triangles[i].p2)||
     !stream->writeUChar(_triangles[i].p3))
    {
    return;
    }
  }
for(i=0;i<_lines.size();i++)
  {
  if(!stream->writeUChar(_lines[i]))
    {
    return;
    }
  }
}

void
GLFontSymbol::_simplify(double precision)
{
unsigned int i;
double sqrPrecision=precision*precision;
for(i=1;i<_vertices.size();i++)
  {
  double dx=_vertices[i].x-_vertices[i-1].x;
  double dy=_vertices[i].y-_vertices[i-1].y;
  if(dx*dx+dy*dy<sqrPrecision)
    {
    _vertices[i-1].x+=0.5*dx;
    _vertices[i-1].y+=0.5*dy;
    _vertices.erase(_vertices.begin()+i);
    unsigned int j;
    for(j=_lines.size();j--;)
      {
      if(_lines[j]>=i)
        {
        --_lines[j];
        }
      }
    for(j=_triangles.size();j--;)
      {
      if(_triangles[j].p1>=i)
        {
        --_triangles[j].p1;
        }
      if(_triangles[j].p2>=i)
        {
        --_triangles[j].p2;
        }
      if(_triangles[j].p3>=i)
        {
        --_triangles[j].p3;
        }
      }
    }
  }
for(i=_triangles.size();i--;)
  {
  double x1=_vertices[_triangles[i].p1].x;
  double y1=_vertices[_triangles[i].p1].y;
  double x2=_vertices[_triangles[i].p2].x-x1;
  double y2=_vertices[_triangles[i].p2].y-y1;
  double x3=_vertices[_triangles[i].p3].x-x1;
  double y3=_vertices[_triangles[i].p3].y-y1;
  if(x2*y3-y2*x3<sqrPrecision)
    {
    _triangles.erase(_triangles.begin()+i);
    }
  }
}

void
GLFontSymbol::_edgeFlagCB(GLboolean /* flag */,
                          GLFontSymbol * /* symbol */)
{
  // Nothing to be done
}

void
GLFontSymbol::_errorCB(GLenum /* errorCode */,
                       GLFontSymbol * symbol)
{
symbol->setErrorMessage("Error during tesselation");
}

void
GLFontSymbol::_beginCB(GLenum primitive,
                       GLFontSymbol * symbol)
{
if(primitive!=GL_TRIANGLES)
  {
  symbol->setErrorMessage("Bad primitive during tesselation");
  }
}

void
GLFontSymbol::_vertexCB(unsigned int index,
                        GLFontSymbol * symbol)
{
switch((symbol->_vertexCount++)%3)
  {
  case 0:
    {
    symbol->_currentFace.p1=index;
    break;
    }
  case 1:
    {
    symbol->_currentFace.p2=index;
    break;
    }
  case 2:
    {
    symbol->_currentFace.p3=index;
    symbol->_triangles.push_back(symbol->_currentFace);
    break;
    }
  }
}

void
GLFontSymbol::_combineCB(GLdouble coords[3],
                         unsigned int /* vertexData */ [4],
                         GLfloat /* weight */ [4],
                         unsigned int * dataOut,
                         GLFontSymbol * symbol)
{
(*dataOut)=symbol->_vertices.size();
symbol->_vertices.push_back(Util3D::Dbl3(coords[0],coords[1],coords[2]));
}

void
GLFontSymbol::_endCB(GLFontSymbol * /* symbol */)
{
  // Nothing to be done
}

//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(GLFont,ArObject)

GLFont::GLFont(ArCW & arCW,
               const StlString & name)
: ArObject(arCW),
  _symbols()
{
FT_Error error;
FT_Library library;
error=FT_Init_FreeType(&library);
if(error)
  {
  setErrorMessage("PB in FT_Init_FreeType()");
  return;
  }
FT_Face face;
error=FT_New_Face(library,name.c_str(),0,&face);
if(error)
  {
  setErrorMessage("PB in FT_New_Face() with font "+name);
  return;
  }
if(!FT_IS_SCALABLE(face))
  {
  setErrorMessage("Font "+name+" is not scalable");
  return;
  }
if(!FT_HAS_HORIZONTAL(face))
  {
  setErrorMessage("Font "+name+" is not horizontal");
  return;
  }
double scale=face->units_per_EM;
if(!scale)
  {
  setErrorMessage("Bad scale for font "+name);
  return;
  }
for(unsigned int code=0;code<256;code++)
  {
  FT_UInt glyph_index=FT_Get_Char_Index(face,code);
  if(glyph_index)
    {
    error=FT_Load_Glyph(face,glyph_index,
                        FT_LOAD_NO_SCALE|
                        FT_LOAD_NO_BITMAP|
                        FT_LOAD_NO_HINTING|
                        FT_LOAD_LINEAR_DESIGN);
    if(error)
      {
      setErrorMessage("Pb in FT_Load_Glyph() for char "+uintToStr(code));
      cerr << getErrorMessage() << endl;
      setErrorMessage("");
      }
    else if(face->glyph->format!=FT_GLYPH_FORMAT_OUTLINE)
      {
      setErrorMessage("No outline for char "+uintToStr(code));
      cerr << getErrorMessage() << endl;
      setErrorMessage("");
      }
    else
      {
      ArRef<GLFontSymbol> symbol=GLFontSymbol::NEW(code,face->glyph,scale);
      if(symbol->fail())
        {
        cerr << symbol->getErrorMessage() << endl;
        }
      else
        {
        _symbols.push_back(symbol);
        }
      }
    }
  }
if(_symbols.empty())
  {
  setErrorMessage("Font "+name+" is empty");
  }
}

GLFont::~GLFont(void)
{
}

void
GLFont::writeToStream(ArRef<AbstractOStream> stream) const
{
unsigned int i;
if(!stream->writeUChar('A')||
   !stream->writeUChar('r')||
   !stream->writeUChar('G')||
   !stream->writeUChar('L')||
   !stream->writeUChar('F')||
   !stream->writeUChar(_symbols.size()))
  {
  return;
  }
for(i=0;i<_symbols.size();i++)
  {
  _symbols[i]->writeToStream(stream);
  if(stream->fail())
    {
    return;
    }
  }
}

//----------------------------------------------------------------------------

void registerGLFontClasses(void);

int
main(int argc,
     char ** argv)
{
ArSystem arevi(argc,argv,false);
registerGLFontClasses();
if(argc<3)
  {
  cerr << endl;
  cerr << "usage: " << argv[0] << " font_file.ttf output_file.arglf" << endl;
  cerr << endl;
  return(-1);
  }
ArRef<GLFont> font=GLFont::NEW(argv[1]);
if(font->fail())
  {
  cerr << font->getErrorMessage() << endl;
  }
else
  {
  ArRef<FileOStream> stream=FileOStream::NEW(argv[2],false);
  if(stream->fail())
    {
    cerr << stream->getErrorMessage() << endl;
    }
  else
    {
    font->writeToStream(stream);
    if(stream->fail())
      {
      cerr << "Problem while writing to "
           << stream->getFileName() << " !" << endl;
      cerr << stream->getErrorMessage() << endl;
      }
    }
  }
return(0);
}

