#include "AReVi/Lib3D/base3D.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/material3D.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/transform3D.h"

#include "path3D.h"

AR_CLASS_DEF(Path3D, ArObject)

Path3D::Path3D(ArCW& arCW)
: Path(arCW),
  _updating(false),
  _precision(100),
  _line(Object3D::NEW()),
  _wayPointSh(VrmlShape3D::NEW()),
  _movingBaseSh(VrmlShape3D::NEW()),
  _lineSh(Shape3D::NEW()),
  _lineShp(LineSet3D::NEW()),
  _duration(1.0) {

  _lineShp->setThickness(2);
  StlVector<StlVector<Util3D::Dbl3> >& c = _lineShp->accessColors();
  c.push_back(StlVector<Util3D::Dbl3>());
  c.back().push_back(Util3D::Dbl3(1.0, 1.0, 1.0));
  _lineSh->addRootPart(_lineShp);
  _line->setShape(_lineSh);

  ArRef<Transform3D> tr = Transform3D::NEW();
  tr->postScale(0.5, 0.5, 0.5);

  _wayPointSh->parseURL("wayPoint.wrl");
  _wayPointSh->writeTransformation(tr);

  _movingBaseSh->parseURL("movingBase.wrl");
  _movingBaseSh->writeTransformation(tr);

  _derivate[0] = Arrow::NEW();
  _derivate[0]->addMotionCB(thisRef(), &Path3D::_onArrowMotion);
  _derivate[1] = Arrow::NEW();
  _derivate[1]->addMotionCB(thisRef(), &Path3D::_onArrowMotion);

  _scale[0] = 1.0;
  _scale[1] = 1.0;

  _movingBase = Object3D::NEW();
  ar_down_cast<Object3D>(_movingBase)->setShape(_movingBaseSh);
}

Path3D::~Path3D(void) {
}

void
Path3D::applyChanges(void) {
  _updating = true;

  Path::applyChanges();

  if(_wayPoints.empty()) {
    _derivate[0]->linkTo(Base3D::nullRef());
    _derivate[1]->linkTo(Base3D::nullRef());
  } else {
    _derivate[0]->linkTo(_wayPoints.front());
    _derivate[1]->linkTo(_wayPoints.back());
  }

  _updateLine();

  _updating = false;
}

void
Path3D::addWayPoint(int p) {
  ArRef<Material3D> m = Material3D::NEW();
  m->setDiffuseColor(1.0, 0.407, 0.117);
  m->setTransparency(0.3);

  ArRef<MyObject3D> o = MyObject3D::NEW();
  o->setShape(_wayPointSh);
  o->writeSelectedMaterial(m);
  o->addTranslationCB(thisRef(), &Path3D::_onWayPointTranslation);

  if(_wayPoints.empty() ||
     p >= (int)_wayPoints.size()) {
    _wayPoints.push_back(o);
    _autoOrientations.push_back(true);
  } else if(p == 0) {
    o->setLocation(_wayPoints.front());
    _wayPoints.insert(_wayPoints.begin(), o);
    _autoOrientations.insert(_autoOrientations.begin(), true);
  } else {
    double u = (_parameters[p] - _parameters[p - 1]) * 0.5 + _parameters[p - 1];
    o->setLocation(evaluate(u * _duration));
    _wayPoints.insert(_wayPoints.begin() + p, o);
    _autoOrientations.insert(_autoOrientations.begin() + p, true);
  }

  applyChanges();
}

void
Path3D::removeWayPoint(int position) {
  /*
  StlVectorFastErase(_wayPoints, position);
  StlVectorFastErase(_autoOrientations, position);
  */

  _wayPoints.erase(_wayPoints.begin() + position);
  _autoOrientations.erase(_autoOrientations.begin() + position);

  applyChanges();
}

ArConstRef<Object3D>
Path3D::getWayPoint(int position) const {
  return ar_down_cast<Object3D>(_wayPoints[position]);
}

ArRef<Object3D>
Path3D::accessWayPoint(int position) {
  return ar_down_cast<Object3D>(_wayPoints[position]);
}

void
Path3D::setStartDerivate(const Vector3d& d) {
  Path::setStartDerivate(d);

  _scale[0] = d.length();
  ArRef<Transform3D> tr = Transform3D::NEW();
  tr->postScale(_scale[0], 1.0, 1.0);

  ArRef<Shape3D> sh = _derivate[0]->accessShape();
  sh->writeTransformation(tr);

  Vector3d u = d.unit();
  Vector3d axis = Vector3d(1.0, 0.0, 0.0) ^ u;
  double cosAngle = Vector3d(1.0, 0.0, 0.0) * u;
  double sinAngle = axis.length();
  double angle = atan2(sinAngle, cosAngle);
  axis.normalize();

  _derivate[0]->setRotation(axis.x(), axis.y(), axis.z(), angle);
}

void
Path3D::setEndDerivate(const Vector3d& d) {
  Path::setEndDerivate(d);

  _scale[1] = d.length();
  ArRef<Transform3D> tr = Transform3D::NEW();
  tr->postScale(_scale[1], 1.0, 1.0);

  ArRef<Shape3D> sh = _derivate[1]->accessShape();
  sh->writeTransformation(tr);

  Vector3d u = d.unit();
  Vector3d axis = Vector3d(1.0, 0.0, 0.0) ^ u;
  double cosAngle = Vector3d(1.0, 0.0, 0.0) * u;
  double sinAngle = axis.length();
  double angle = atan2(sinAngle, cosAngle);
  axis.normalize();

  _derivate[1]->setRotation(axis.x(), axis.y(), axis.z(), angle);
}

void
Path3D::setPrecision(int p) {
  _precision = p;
  _updateLine();
}

double
Path3D::getPrecision(void) const {
  return _precision;
}

void
Path3D::setDuration(double d) {
  _duration = d;
}

double
Path3D::getDuration(void) const {
  return _duration;
}

ArConstRef<Base3D> // base to attach stuffs
Path3D::evaluate(double p) {
  return Path::evaluate(p / _duration);
}

void
Path3D::serialize(SerializationDependencies& dependenciesOut,
		  ArRef<AbstractOStream> output) const {
  // Serialize it as a simple path
  ArRef<Path> p = Path::NEW();
  p->setName(getName());
  StlVector<ArRef<Base3D> >& wp = p->accessWayPoints();
  StlVector<bool>& ao = p->accessAutoOrientations();
  for(size_t i = 0; i < _wayPoints.size(); i++) {
    ArRef<Base3D> b = Base3D::NEW();
    b->setLocation(_wayPoints[i]);
    wp.push_back(b);
    ao.push_back(_autoOrientations[i]);
  }
  p->setStartDerivate(getStartDerivate());
  p->setEndDerivate(getEndDerivate());

  p->serialize(dependenciesOut, output);
}

bool
Path3D::unserialize(const SerializationDependencies& dependencies,
		    ArRef<AbstractIStream> input) {
  ArRef<Path> p = Path::NEW();
  if(!p->unserialize(dependencies, input)) return false;

  while(!_wayPoints.empty()) {
    removeWayPoint(0);
  }

  setName(p->getName());
  setStartDerivate(p->getStartDerivate());
  setEndDerivate(p->getEndDerivate());

  StlVector<ArRef<Base3D> >& wp = p->accessWayPoints();
  StlVector<bool>& ao = p->accessAutoOrientations();

  ArRef<Material3D> m = Material3D::NEW();
  m->setDiffuseColor(1.0, 0.407, 0.117);
  m->setTransparency(0.3);

  for(size_t i = 0; i < wp.size(); i++) {
    ArRef<MyObject3D> o = MyObject3D::NEW();
    o->setLocation(wp[i]);
    o->setShape(_wayPointSh);
    o->writeSelectedMaterial(m);
    o->addTranslationCB(thisRef(), &Path3D::_onWayPointTranslation);
    _wayPoints.push_back(o);
    _autoOrientations.push_back(ao[i]);
  }
  if(!_wayPoints.empty()) {
    _movingBase->setLocation(_wayPoints.front());
  }
  applyChanges();

  return true;
}

void
Path3D::_onWayPointTranslation(const MyObject3D::TranslationEvent&) {
  if(!_updating) {
    applyChanges();
  }
}

void
Path3D::_onArrowMotion(const Base3D::MotionEvent& evt) {
  if(!_updating) {
    ArRef<Arrow> a = ar_down_cast<Arrow>(evt.source);

    double x, y, z;
    evt.source->getLocalX(x, y, z);

    if(a == _derivate[0]) {
      _sd = Vector3d(x, y, z) * _scale[0];
    } else {
      _ed = Vector3d(x, y, z) * _scale[1];
    }

    applyChanges();
  }
}

void
Path3D::_updateLine(void) {
  StlVector<StlVector<Util3D::Dbl3> >& v = _lineShp->accessVertices();
  v.clear();
  v.push_back(StlVector<Util3D::Dbl3>());

  StlVector<StlVector<Util3D::Dbl3> >& c = _lineShp->accessColors();
  c.clear();
  c.push_back(StlVector<Util3D::Dbl3>());
  c.back().push_back(Util3D::Dbl3(1.0, 1.0, 1.0));

  if(_wayPoints.size() > 1) {
    for(int i = 0; i <= _precision; i++) {
      double u = (double)i / (double)_precision;
      ArConstRef<Base3D> b = evaluate(u * _duration);
      double x, y, z;
      b->getPosition(x, y, z);
      v.back().push_back(Util3D::Dbl3(x, y, z));
    }
    /*
    for(int i = 0; i <= _precision; i++) {
      double u = (double)i / (double)_precision;
      ArConstRef<Base3D> b = evaluate(u * _duration);
      double x, y, z;
      b->getPosition(x, y, z);

      Vector3d d;
      _curve.computeDerive(u, d);

      v.push_back(StlVector<Util3D::Dbl3>());
      v.back().push_back(Util3D::Dbl3(x, y, z));
      v.back().push_back(Util3D::Dbl3(x + d.x(), y + d.y(), z + d.z()));

      c.push_back(StlVector<Util3D::Dbl3>());
      c.back().push_back(Util3D::Dbl3(0.0, 0.0, 1.0));
    }
    */
  }

  _lineShp->applyChanges(true);
}
