#ifndef _AR_MATH_MATRIX4X4_H_
#define _AR_MATH_MATRIX4X4_H_

namespace AReVi {

class Matrix4x4d {
public :

  enum Matrix4x4Index {
    R00 = 0, R01 = 4, R02 =  8,  TX = 12,
    R10 = 1, R11 = 5, R12 =  9,  TY = 13,
    R20 = 2, R21 = 6, R22 = 10,  TZ = 14,
     Z0 = 3,  Z1 = 7,  Z2 = 11, ONE = 15
  };

  /** Construction / Destruction
   */
  inline Matrix4x4d(void);
  inline Matrix4x4d(double m00, double m01, double m02, double tx,
                   double m10, double m11, double m12, double ty,
                   double m20, double m21, double m22, double tz,
                   double z0 , double z1 , double z2 , double one);
  inline Matrix4x4d(double a);
  inline Matrix4x4d(const Matrix3x3d& m);
  inline Matrix4x4d(const Vector3d& translation);
  inline Matrix4x4d(const Vector3d& axe, double angle);
  inline Matrix4x4d(const Quaterniond& q);
  inline Matrix4x4d(const Quaterniond& q, const Vector3d& t);
  inline Matrix4x4d(const double* m);
  inline Matrix4x4d(const Matrix4x4d& m);

  /** Accessor
   */
  inline double at(Matrix4x4Index i) const;
  inline double at(size_t row, size_t col) const;
  inline double operator[](Matrix4x4Index i) const;

  /** Affectation
   */
  inline Matrix4x4d& operator=(const Matrix4x4d& m);

  /** Add / sum / mul / div
   */
  inline Matrix4x4d& operator+=(const Matrix4x4d& m);
  inline Matrix4x4d  operator+ (const Matrix4x4d& m) const;
  inline Matrix4x4d& operator-=(const Matrix4x4d& m);
  inline Matrix4x4d  operator- (const Matrix4x4d& m) const;
  inline Matrix4x4d& operator*=(const double& a);
  inline Matrix4x4d  operator* (const double& a) const;
  inline Matrix4x4d& operator*=(const Matrix4x4d& m);
  inline Matrix4x4d  operator* (const Matrix4x4d& m) const;
  inline Matrix4x4d& operator/=(const double& a);
  inline Matrix4x4d  operator/ (const double& a);

  inline Point3d  operator*(const Point3d& p) const;
  inline Vector3d operator*(const Vector3d& v) const;

  /** Inverse / Transpose
   */
  inline bool invert(void);
  inline void transpose(void);

  /** Matrix4x4 operations
   */
  inline Matrix4x4d& set(const double* m);
  inline Matrix4x4d& set(double m00, double m01, double m02, double tx,
                           double m10, double m11, double m12, double ty,
                           double m20, double m21, double m22, double tz,
                           double z0 , double z1 , double z2 , double one);
  inline Matrix4x4d& setIdentity(void);
  inline Matrix4x4d& setTranslation(const Vector3d& v);
  inline Matrix4x4d& setScale(double x, double y, double z);
  inline Matrix4x4d& setXRotation(double angle);
  inline Matrix4x4d& setYRotation(double angle);
  inline Matrix4x4d& setZRotation(double angle);
  inline Matrix4x4d& setRotation(const Matrix3x3d& m);
  inline Matrix4x4d& setRotation(const Vector3d& axe, double angle);
  inline Matrix4x4d& setRotation(const Quaterniond& q);

  /** Determinant
   */
  inline double determinant(void) const;

  /** Cast
   */
  inline operator double* (void);
  inline operator const double* (void) const;
  inline       double* accessMatrix(void);
  inline const double* getMatrix(void) const;

private :
  inline double _det3x3(double m00, double m01, double m02,
                   double m10, double m11, double m12,
                   double m20, double m21, double m22);

  double _v[16];
};

inline Matrix4x4d operator*(double a, const Matrix4x4d& m);

inline Matrix4x4d operator/(double a, const Matrix4x4d& m);

inline ostream& operator<<(ostream& o, const Matrix4x4d& m);

} // namespace AReVi

#endif // _AR_MATH_MATRIX4X4_H_
