#ifndef _AR_QUATERNION_H_
#define _AR_QUATERNION_H_

/** A quaternion has the following form :
    q = q0 + q1.i + q2.j + q3.k
 */

namespace AReVi {

class Quaterniond {
  friend Quaterniond pow(const Quaterniond& q, double p);
public :
  /** Construction / Destruction
   */
  inline Quaterniond(void);
  inline Quaterniond(double a, double b, double c, double d);
  inline Quaterniond(const Vector3d& axis, double angle);
  inline Quaterniond(const Matrix3x3d& m);
  inline Quaterniond(double rx, double ry, double rz);
  inline Quaterniond(const Quaterniond& q);

  /** Accessors
   */
  inline double q0(void) const;
  inline double q1(void) const;
  inline double q2(void) const;
  inline double q3(void) const;

  /** Affectation
   */
  inline Quaterniond& operator=(const Quaterniond& q);

  /** Comparison
   */
  inline bool operator==(const Quaterniond& q) const;
  inline bool operator!=(const Quaterniond& q) const;

  /** Get the norme of the quaternion
   */
  inline double length(void) const;

  /** Modifiers
   */
  inline void conjugate(void);
  inline bool invert(void);
  inline bool normalize(void);

  inline void setAxisAngle(const Vector3d axis, double angle);
  inline void setRotation(const Matrix3x3d m);

  /** Get modified quaternions
   */
  inline void getConjugate(Quaterniond& result) const;
  inline bool getInvert(Quaterniond& result) const;
  inline bool getNormalized(Quaterniond& result) const;

  /** Extract angle & axis or matrix rotation
   */
  inline bool getAxisAngle(Vector3d& axis, double& angle) const;
  inline bool getRotationMatrix(Matrix3x3d& m) const;
  inline bool getEulerAngles(double& rx, double& ry, double& rz) const;

  /** Operations
   */
  inline Quaterniond& operator+=(const Quaterniond& q);
  inline Quaterniond  operator+ (const Quaterniond& q) const;
  inline Quaterniond& operator-=(const Quaterniond& q);
  inline Quaterniond  operator- (const Quaterniond& q) const;
  inline Quaterniond& operator*=(double a);
  inline Quaterniond  operator* (double a) const;
  inline Quaterniond& operator*=(const Quaterniond& q);
  inline Quaterniond  operator* (const Quaterniond& q) const;
  inline Quaterniond& operator/=(double a);
  inline Quaterniond  operator/ (double a) const;

  inline Point3d  operator*(const Point3d& p) const;
  inline Vector3d operator*(const Vector3d& v) const;

  /** Dot product
   */
  inline double dot(const Quaterniond& q) const;

private :
  double _q0, _q1, _q2, _q3;
};

inline Quaterniond operator*(double a, const Quaterniond& q);

inline Quaterniond pow(const Quaterniond& q, double p);

inline
Vector3d log(const Quaterniond& q);

inline
Quaterniond exp(const Vector3d v);

/** Liner interpolation */
inline
Quaterniond
lerp(const Quaterniond& q1,const Quaterniond& q2, double p);

/** Spherical linear interpolation */
inline
Quaterniond
slerp(const Quaterniond& q1, const Quaterniond& q2, double p);

/** Spherical cubic interpolation */
inline
Quaterniond
squad(const Quaterniond& p,
      const Quaterniond& a,
      const Quaterniond& b,
      const Quaterniond& q,
      double t);

inline
ostream& operator<<(ostream& o, const Quaterniond& q);

} // namespace AReVi

#endif // _AR_QUATERNION_H_
