// LI2/ENIB - AReVi - Contrib
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/arNurbs.h
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_CONTRIB_ARNURBS_H
#define AREVI_CONTRIB_ARNURBS_H 1

#include "AReVi/arObject.h"
#include "AReVi/Lib3D/util3D_priv.h"
#include "AReVi/Contrib/arMath.h"

namespace AReVi {

class AbstractOStream;
class AbstractIStream;

//----------------------------------------------------------------------------
// Class ArNurbs (abstract)
//----------------------------------------------------------------------------

class ArNurbs
{
public:
//----------------------------------------------------------------------------
// Class ArNurbs::Curve3d
//----------------------------------------------------------------------------

  class Curve3d
  {
  public:
    inline Curve3d(void);

    inline Curve3d(const StlVector<Vector3d > &ctrlPoints, int degree);

  //-------- Curve dimensions---------------------------------------------------
    inline
    void // setup knots and ctrlPoints
    resize(unsigned int nbCtrlPoints,
           unsigned int degree);

    inline
    unsigned int
    getDegree(void) const;

    inline
    bool // check curve validity
    valid(void) const;

  //-------- Control Points ----------------------------------------------------

    inline
    const StlVector<Vector3d > &
    getCtrlPoints(void) const;

    inline
    StlVector<Vector3d > &
    accessCtrlPoints(void);

 //-------- Weights Points ----------------------------------------------------

    inline
    const StlVector<double> &
    getWeights(void) const ;

    inline
    StlVector<double> &
    accessWeights(void);

  //-------- Knot vector -------------------------------------------------------

    inline
    const StlVector<double> &
    getKnotVector(void) const;

    inline
    StlVector<double> &
    accessKnotVector(void);

    inline
    double
    begin(void) const;

    inline
    double
    end(void) const;

    inline
    int
    findMult(int r) const;

  //-------- Compute points ----------------------------------------------------

    inline
    void
    computePoint(double u, Vector3d &pointOut);

    inline
    void
    computePoint(double u, int span, Vector3d &pointOut);

    inline
    void
    computeDerive(double u, Vector3d &deriv);

    inline
    void
    computeDerive(double u, int span, Vector3d &deriv);

  //-------- Curve Serialization ----------------------------------------------
    // Specialized for Curve3f and Curve3d
    void
    writeToStream(ArRef<AbstractOStream> output) const;

    bool // success
    readFromStream(ArRef<AbstractIStream> input);

  //-------- Member attributes -------------------------------------------------
  protected:
    StlVector<double> _knots;

    unsigned int _degree;
    StlVector<Vector3d > _ctrlPoints;
    StlVector<double> _weights;
  };

//----------------------------------------------------------------------------
// Class ArNurbs::Tab
//----------------------------------------------------------------------------
  template <class T>
  class Tab
  {
  public:
    inline Tab(void);
    inline Tab(unsigned int x, unsigned int y);

    inline T &operator()(unsigned int x,unsigned int y);
    inline const T &operator()(unsigned int x,unsigned int y) const;

    inline void resize(unsigned int x, unsigned int y);

    inline unsigned int sizeX(void) const; //rows
    inline unsigned int sizeY(void) const; //cols
    inline unsigned int size(void) const;

    inline const StlVector<T> &getData(void) const;
    inline StlVector<T> &accessData(void);

    // Decompose matrix using LH method. Pivots values must be used to inverseIn method.
    inline bool decompose(const Tab<T>& a, StlVector<int> &pivotOut);
    // Inverse Matrix.
    inline bool inverseIn(const StlVector<int> &pivot, Tab<T>& inc);

    // Multipli axb and store the result un this
    inline bool mult(const Tab<T> &a,const Tab<T> &b);
  //-------- Member attributes -------------------------------------------------
  protected:
    unsigned int _sizeX,_sizeY;
    StlVector<T> _tab;
  };

//----------------------------------------------------------------------------
// Class ArNurbs::Surface3d
//----------------------------------------------------------------------------
  class Surface3d
  {
  public:
    inline
    Surface3d(void);

    inline
    Surface3d(unsigned int uSize, unsigned int vSize,
            const StlVector<Vector3d > &ctrlPoints,
            unsigned int degree);

  //-------- Surface Size ------------------------------------------------------

    inline
    void // setup knots and ctrlPoints
    resize(unsigned int sizeU,
           unsigned int sizeV,
           unsigned int degreeU,
           unsigned int degreeV);

    inline
    bool // check surface validity
    valid(void) const;

  //-------- Control Points ----------------------------------------------------

    inline
    const StlVector<Vector3d > &
    getCtrlPoints(void) const;

    inline
    StlVector<Vector3d > &
    accessCtrlPoints(void);

    inline
    Vector3d &
    accessCtrlPoint(unsigned int u,unsigned int v);

    inline
    unsigned int
    getSizeU(void) const;

    inline
    unsigned int
    getSizeV(void) const;

  //-------- Weights Points ----------------------------------------------------

    inline
    const StlVector<double> &
    getWeights(void) const ;

    inline
    StlVector<double> &
    accessWeights(void);

    inline
    double &
    accessWeight(unsigned int u,unsigned int v);

  //-------- Knot vector ------------------------------------------------------

    inline const StlVector<double> &
    getUKnotVector(void) const;

    inline StlVector<double> &
    accessUKnotVector(void);

    inline
    double
    beginU(void) const;

    inline
    double
    endU(void) const;

    inline
    const StlVector<double> &
    getVKnotVector(void) const;

    inline
    StlVector<double> &
    accessVKnotVector(void);

    inline double
    beginV(void) const;

    inline double
    endV(void) const;

  //-------- Compute points ----------------------------------------------------

    inline
    void
    computePoint(double u, double v, Vector3d &pointOut) const;

    inline
    void
    computePoint(double u, int spanU,
                 double v, int spanV,
                 Vector3d  &pointOut) const;

    inline
    void
    computeDerives(double u, double v,
                   unsigned int d, Tab<Vector3d > &deriv) const;

    inline
    void
    computeDerives(double u, int spanU,
                   double v, int spanV,
                   unsigned int d, Tab<Vector3d > &deriv) const;

    inline
    void
    computeFirstDerive(double u, int spanU,
                       double v, int spanV,
                       Tab<Vector3d > &deriv) const;

    inline
    void
    computeNormal(double u, double v, Vector3d &normal) const;

    inline
    void
    computeVertice(double u, double v, Vector3d &vertex, Vector3d  &normal) const;

    inline
    void
    computeVertice(double u, int uspan, double v, int vspan, Vector3d &vertex, Vector3d  &normal) const;

    inline
    int // Span index (u must be valid!)
    findSpanU(double u) const;

    inline
    int // Span index (v must be valid!)
    findSpanV(double v) const;

  //-------- Surface Serialization ---------------------------------------------
    // Specialized for Surface3d
    void
    writeToStream(ArRef<AbstractOStream> output) const;

    bool // success
    readFromStream(ArRef<AbstractIStream> input);

  protected:
    // Compute derive without weights...
    inline
    void
    _computeDerives(double u, int spanU,
                    double v, int spanV,
                    unsigned int d,
                    Tab<Vector3d > &dersP,
                    Tab<double> &dersW) const;


    // Compute first derive without weights...
    inline
    void
    _computeFirstDerive(double u, int spanU,
                         double v, int spanV,
                         Tab<Vector3d > &dersP,
                         Tab<double> &dersW) const;


    inline static
    void
    _basisFunsDers(int spanIndex,                   // span index
                   double u,                        // u => Point to find Span from (Knot vector index)
                   int deg,                         // deg => Curve degree
                   const StlVector<double> &knots,  // knot vector
                   int d,                           // derivation degree
                   Tab<double> &ders);// Basis funct values N[k] => kth derivation

    inline static
    void
    _basisFunsFirstDers(int spanIndex,                   // span index
                   double u,                        // u => Point to find Span from (Knot vector index)
                   int deg,                         // deg => Curve degree
                   const StlVector<double> &knots,  // knot vector
                   Tab<double> &ders);// Basis funct values N[k] => kth derivation

  //-------- Member attributes -------------------------------------------------
  protected:
    StlVector<double> _uKnots,_vKnots;
    unsigned int _uDegree,_vDegree;

    Tab<Vector3d > _ctrlPoints; // [U,V]
    Tab<double> _weights; // [U,V]
  };

//----------------------------------------------------------------------------
// Class ArNurbs::CurveN
//----------------------------------------------------------------------------

  template <class T, int N>
  class CurveN
  {
  public:
    inline CurveN(void);

    inline CurveN(const StlVector<VectorN<T,N> > &ctrlPoints, int degree);

  //-------- Curve dimensions---------------------------------------------------
    inline
    void // setup knots and ctrlPoints
    resize(unsigned int nbCtrlPoints,
           unsigned int degree);

    inline
    unsigned int
    getDegree(void) const;

    inline
    bool // check curve validity
    valid(void) const;

  //-------- Control Points ----------------------------------------------------

    inline
    const StlVector<VectorN<T,N> > &
    getCtrlPoints(void) const;

    inline
    StlVector<VectorN<T,N> > &
    accessCtrlPoints(void);

 //-------- Weights Points ----------------------------------------------------

    inline
    const StlVector<T> &
    getWeights(void) const ;

    inline
    StlVector<T> &
    accessWeights(void);

  //-------- Knot vector -------------------------------------------------------

    inline
    const StlVector<T> &
    getKnotVector(void) const;

    inline
    StlVector<T> &
    accessKnotVector(void);

    inline
    T
    begin(void) const;

    inline
    T
    end(void) const;

    inline
    int
    findMult(int r) const;

  //-------- Compute points ----------------------------------------------------

    inline
    void
    computePoint(T u, VectorN<T,N> &pointOut);

    inline
    void
    computePoint(T u, int span, VectorN<T,N> &pointOut);

    inline
    void
    computeDerive(T u, VectorN<T,N> &deriv);

    inline
    void
    computeDerive(T u, int span, VectorN<T,N> &deriv);

  //-------- Member attributes -------------------------------------------------
  protected:
    StlVector<T> _knots;

    unsigned int _degree;
    StlVector<VectorN<T,N> > _ctrlPoints;
    StlVector<T> _weights;
  };


//----------------------------------------------------------------------------
// Class ArNurbs::Utils
//----------------------------------------------------------------------------

  template <class T>
  inline static
  int // Span index (u must be valid!)
  findSpan(int n,                     // n => Nb Ctrls Points-1
           int p,                     // p => Curve degree
           T u,                  // u => Point to find Span from (Knot vector index)
           const StlVector<T> &knots); // knot vector

  template <class T>
  inline static
  void
  basisFuns( int spanIndex,                   // n => Nb Ctrls Points-1
             T u,                        // u => Point to find Span from (Knot vector index)
             int p,                           // p => Curve degree
             const StlVector<T> &knots,  // knot vector
             StlVector<T> &Nout);        // Basis funct values

  template <class T>
  inline static
  void
  basisFunsFirstDers(int spanIndex,                   // span index
                   T u,                        // u => Point to find Span from (Knot vector index)
                   int deg,                         // deg => Curve degree
                   const StlVector<T> &knots,  // knot vector
                   Tab<T> &ders);// Basis funct values N[k] => kth derivation


  template <class T>
  inline static
  void
  fillKnotVector(unsigned nbCtrlPoints,
                 unsigned degree,
                 bool closed,
                 StlVector<T> &knots); // knot vector

  template <class T>
  inline static
  void
  binomialCoef(Tab<T>& bin);

  inline static
  void
  projectToLine(const Vector3d S, const Vector3d& Trj, const Vector3d& pnt, Vector3d& p);

  inline static
  int
  intersectLine(const Vector3d& p1,
                const Vector3d& t1,
                const Vector3d& p2,
                const Vector3d& t2,
                Vector3d& p);

//-------- Curve construction facilities --------------------------------------

  inline static
  void
  makeLine(unsigned nbCtrlPoints,
           unsigned degree,
           const Vector3d &position,
           const Vector3d &direction,
           Curve3d &curveOut); // curve

  inline static
  void
  makeCircle(const Point3d &position,
              double radius,
              Curve3d &curveOut); // circle

  inline static
  void
  make3DPlane(unsigned uRes, double uSize,
              unsigned vRes, double vSize,
              unsigned degree,
              Surface3d &surfaceOut); // plane

  inline static
  void
  extrude(const Curve3d &curve,
          const Vector3d &direction, double length,
          Surface3d &surfaceOut); // Extrusion

  inline static
  void
  revolve(const Curve3d &curve,
          const Vector3d &center,
          const Vector3d &direction,
          double angle, // in degree
          Surface3d &surfaceOut); // Make a revolution surface


// ------------------- GLOBAL INTERPOLATIONS ALGORITHMS ----------------------------------------------

  // Q is a list of way points of the curve, ubOut is the parametrization of the curve
  // Code de la Nurbs++ lib
  inline static
  double // total length
  chordLengthParam(const StlVector<Vector3d >& Q, StlVector<double> &ubOut);

  // Compute a curve passing throught "points". This function use chord length method to
  // make the parametrization
  // Code de la Nurbs++ lib
  inline static
  bool // sucess . Note that points.size() must be > degree
  globalInterpolation(const StlVector<Vector3d > &points, unsigned degree, Curve3d &curveOut);

  // ub allow to specify parametrization of the curve
  // Code de la Nurbs++ lib
  inline static
  bool // sucess . Note that points.size() must be > degree
  globalInterpolation(const StlVector<Vector3d > &points, const StlVector<double> &ub, unsigned degree, Curve3d &curveOut);


// --- Version with N dimensions -----
  // Q is a list of way points of the curve, ubOut is the parametrization of the curve
  // Code de la Nurbs++ lib
  template <class T, int N>
  inline static
  T // total length
  chordLengthParamN(const StlVector<VectorN<T,N> >& Q, StlVector<T> &ubOut);

  // Compute a curve passing throught "points". This function use chord length method to
  // make the parametrization
  // Code de la Nurbs++ lib
  template <class T, int N>
  inline static
  bool // sucess . Note that points.size() must be > degree
  globalInterpolationN(const StlVector<VectorN<T,N> > &points, unsigned degree, CurveN<T,N> &curveOut);

  // ub allow to specify parametrization of the curve
  // Code de la Nurbs++ lib
  template <class T, int N>
  inline static
  bool // success . Note that points.size() must be > degree
  globalInterpolationN(const StlVector<VectorN<T,N> > &points, const StlVector<T> &ub, unsigned degree, CurveN<T,N> &curveOut);

  // Compute a curve passing throught "points". This function use chord length method to
  // make the parametrization. This version take derivative for first and ending points
  template <class T, int N>
  inline static
  bool // success . Note that points.size() must be > degree
  globalInterpolationDerivN(const StlVector<VectorN<T,N> > &points, unsigned degree, const VectorN<T,N> &derivBegin,bool userDerivBegin, const VectorN<T,N> &derivEnd, bool userDerivEnd, CurveN<T,N> &curveOut);

  // Compute a curve passing throught "points".
  // ub allow to specify parametrization of the curve
  // make the parametrization. This version take derivative for first and ending points
  template <class T, int N>
  inline static
  bool // success . Note that points.size() must be > degree
  globalInterpolationDerivN(const StlVector<VectorN<T,N> > &points, const StlVector<T> &ub,unsigned degree, const VectorN<T,N> &derivBegin,bool userDerivBegin, const VectorN<T,N> &derivEnd, bool userDerivEnd, CurveN<T,N> &curveOut);

  // Compute a curve passing throught "points". This function use chord length method to
  // make the parametrization. This version take derivative for each point
  // derivs is the list of pair < deriv, useThisDeriv> for each point.
  template <class T, int N>
  inline static
  bool // success . Note that points.size() must be > degree and points.size()==derivs.size()
  globalInterpolationDerivN(const StlVector<VectorN<T,N> > &points, unsigned degree, const StlVector<pair< VectorN<T,N> , bool> > &derivs, CurveN<T,N> &curveOut);

  // Compute a curve passing throught "points". parametrization if gived by user.
  // This version take derivative for each point
  // Special implementation of The Nurbs book algorithm that allow specify wayPoint derivative we want
  template <class T, int N>
  inline static
  bool // success . Note that points.size() must be > degree and points.size()==derivs.size()
  globalInterpolationDerivN(const StlVector<VectorN<T,N> > &points, const StlVector<T> &ub,unsigned degree, const StlVector<pair< VectorN<T,N> , bool> > &derivs, CurveN<T,N> &curveOut);


  // Init knot vector, using average technique (p365)
  // Code de la Nurbs++ lib
  template <class T>
  inline static
  void
  knotAveraging(const StlVector<T>& uk, unsigned deg, StlVector<T>& U);

  // Solves the linear system [A][X]=[B] using LU decompositio
  // Code de la Nurbs++ lib
  template <class T>
  inline static
  bool // success
  solveLU(const Tab<T>& A, const Tab<T>& B, Tab<T>& Xout);
};

} // namespace AReVi

//-------- Inline source code ------------------------------------------------

#include "AReVi/Contrib/arNurbsCurve.Ci"

#include "AReVi/Contrib/arNurbsCurveN.Ci"

#include "AReVi/Contrib/arNurbsUtils.Ci"

#include "AReVi/Contrib/arNurbsSurface.Ci"


#endif // AREVI_CONTRIB_ARNURBS_H

//----------------------------------------------------------------------------
