// LI2/ENIB - AReVi-Contrib - Cg context
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/cgContext.h
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_CONTRIB_CGCONTEXT_H
#define AREVI_CONTRIB_CGCONTEXT_H 1

#include "AReVi/arObject.h"

#include <GL/gl.h>

namespace AReVi {

extern "C" bool CgWrapper_init(const StlVector<StlString> *);
extern "C" void CgWrapper_destroy(void);

class CgContext : public ArObject {
public :
  AR_UNSTUBBED AR_CLASS(CgContext)
  AR_CONSTRUCTOR(CgContext)

  /** Data types
   */

  typedef void* CgCTX;
  typedef void* Program;
  typedef void* Param;

  enum Profile {
    PROFILE_ARBVP1,
    PROFILE_ARBFP1,
    PROFILE_VP20,
    PROFILE_FP20,
    PROFILE_VP30,
    PROFILE_FP30,
    PROFILE_VP40,
    PROFILE_FP40,
    LATEST_VERTEX,
    LATEST_FRAGMENT
  };

  enum StateMatrixType {
    MODELVIEW_MATRIX,
    PROJECTION_MATRIX,
    TEXTURE_MATRIX,
    MODELVIEW_PROJECTION_MATRIX
  };

  enum TransformMatrixType {
    MATRIX_IDENTITY,
    MATRIX_TRANSPOSE,
    MATRIX_INVERSE,
    MATRIX_INVERSE_TRANSPOSE
  };

  /*******************************************************
      Programs
  ********************************************************/

  /** Creation / destruction
   */

  CgContext::Program
  createProgram(const StlString& code,
                CgContext::Profile profile,
                const StlString& entry,
                const StlVector<StlString> * args);

  CgContext::Program
  createProgramFromURL(const StlString& url,
                       CgContext::Profile profile,
                       const StlString& entry,
                       const StlVector<StlString> * args);

  void
  destroyProgram(CgContext::Program program);

  /** Profile
   */

  void
  enableProfile(CgContext::Profile profile);

  void
  disableProfile(CgContext::Profile profile);

  bool
  isProfileSupported(CgContext::Profile profile);

  /** Execution
   */

  void
  loadProgram(CgContext::Program program);

  void
  bindProgram(CgContext::Program program);

  /*******************************************************
      Errors
  ********************************************************/

  const char*
  getError(void);

  void
  setErrorCallback(void (*callback)(void));

  /*******************************************************
      Parameters
  ********************************************************/

  CgContext::Param
  getNamedParameter(CgContext::Program program,
                    const StlString& name);

  /** Setting uniform scalar and uniform vector parameters
   */

  void setParameter1f(CgContext::Param parameter, float x);
  void setParameter2f(CgContext::Param parameter, float x, float y);
  void setParameter3f(CgContext::Param parameter, float x, float y, float z);
  void setParameter4f(CgContext::Param parameter, float x, float y, float z, float w);

  void setParameter1fv(CgContext::Param parameter, const float* array);
  void setParameter2fv(CgContext::Param parameter, const float* array);
  void setParameter3fv(CgContext::Param parameter, const float* array);
  void setParameter4fv(CgContext::Param parameter, const float* array);

  void getParameter1f(CgContext::Param parameter, float* array);
  void getParameter2f(CgContext::Param parameter, float* array);
  void getParameter3f(CgContext::Param parameter, float* array);
  void getParameter4f(CgContext::Param parameter, float* array);

  void setParameter1d(CgContext::Param parameter, double x);
  void setParameter2d(CgContext::Param parameter, double x, double y);
  void setParameter3d(CgContext::Param parameter, double x, double y, double z);
  void setParameter4d(CgContext::Param parameter, double x, double y, double z, double w);

  void setParameter1dv(CgContext::Param parameter, const double* array);
  void setParameter2dv(CgContext::Param parameter, const double* array);
  void setParameter3dv(CgContext::Param parameter, const double* array);
  void setParameter4dv(CgContext::Param parameter, const double* array);

  void getParameter1d(CgContext::Param parameter, double* array);
  void getParameter2d(CgContext::Param parameter, double* array);
  void getParameter3d(CgContext::Param parameter, double* array);
  void getParameter4d(CgContext::Param parameter, double* array);

  /** Setting uniform matrix parameters
   */

  void setMatrixParameterfr(CgContext::Param parameter, const float* matrix);
  void setMatrixParameterfc(CgContext::Param parameter, const float* matrix);
  void setMatrixParameterdr(CgContext::Param parameter, const double* matrix);
  void setMatrixParameterdc(CgContext::Param parameter, const double* matrix);

  void getMatrixParameterfr(CgContext::Param parameter, float* matrix);
  void getMatrixParameterfc(CgContext::Param parameter, float* matrix);
  void getMatrixParameterdr(CgContext::Param parameter, double* matrix);
  void getMatrixParameterdc(CgContext::Param parameter, double* matrix);

  void setStateMatrixParameter(CgContext::Param parameter,
                               CgContext::StateMatrixType state,
			       CgContext::TransformMatrixType transform);

  /** Setting uniform arrays of scalar, vector and matrix parameters
   */

  void setParameterArray1f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const float* array);
  void setParameterArray2f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const float* array);
  void setParameterArray3f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const float* array);
  void setParameterArray4f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const float* array);

  void getParameterArray1f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, float* array);
  void getParameterArray2f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, float* array);
  void getParameterArray3f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, float* array);
  void getParameterArray4f(CgContext::Param parameter, int startIndex,
			   int numberOfElements, float* array);

  void setParameterArray1d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const double* array);
  void setParameterArray2d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const double* array);
  void setParameterArray3d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const double* array);
  void setParameterArray4d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, const double* array);

  void getParameterArray1d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, double* array);
  void getParameterArray2d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, double* array);
  void getParameterArray3d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, double* array);
  void getParameterArray4d(CgContext::Param parameter, int startIndex,
			   int numberOfElements, double* array);

  void setMatrixParameterArrayfr(CgContext::Param parameter, int startIndex,
				 int numberOfElements, const float* array);
  void setMatrixParameterArrayfc(CgContext::Param parameter, int startIndex,
				 int numberOfElements, const float* array);

  void getMatrixParameterArrayfr(CgContext::Param parameter, int startIndex,
				 int numberOfElements, float* array);
  void getMatrixParameterArrayfc(CgContext::Param parameter, int startIndex,
				 int numberOfElements, float* array);

  void setMatrixParameterArraydr(CgContext::Param parameter, int startIndex,
				 int numberOfElements, const double* array);
  void setMatrixParameterArraydc(CgContext::Param parameter, int startIndex,
				 int numberOfElements, const double* array);

  void getMatrixParameterArraydr(CgContext::Param parameter, int startIndex,
				 int numberOfElements, double* array);
  void getMatrixParameterArraydc(CgContext::Param parameter, int startIndex,
				 int numberOfElements, double* array);

  /** Setting varying parameters
   */

  void enableClientState(CgContext::Param parameter);
  void disableClientState(CgContext::Param parameter);
  void setParameterPointer(CgContext::Param parameter,
                           GLint size, GLenum type,
			   GLsizei stride, GLvoid* array);

  /** Setting sampler parameters
   */

  void enableTextureParameter(CgContext::Param parameter);
  void disableTextureParameter(CgContext::Param parameter);
  GLuint getTextureParameter(CgContext::Param parameter);
  void setTextureParameter(CgContext::Param parameter, GLuint textureId);

  /*******************************************************
      Private members
  ********************************************************/

private :
  friend bool CgWrapper_init(const StlVector<StlString> *);
  friend void CgWrapper_destroy(void);

  static bool _initSuccess;

  static CgContext::CgCTX  (*_createContext)(void);
  static void   (*_destroyContext)(CgContext::CgCTX context);
  static CgContext::Program (*_createProgram)(CgContext::CgCTX context, const StlString& code, CgContext::Profile profile, const StlString& entry, const StlVector<StlString> * args);
  static CgContext::Program (*_createProgramFromFile)(CgContext::CgCTX, const StlString& filename, CgContext::Profile profile, const StlString& entry, const StlVector<StlString> * args);
  static void   (*_destroyProgram)(CgContext::Program program);
  static void   (*_enableProfile)(CgContext::Profile profile);
  static void   (*_disableProfile)(CgContext::Profile profile);
  static bool   (*_isProfileSupported)(CgContext::Profile profile);
  static void   (*_loadProgram)(CgContext::Param program);
  static void   (*_bindProgram)(CgContext::Param program);
  static const char* (*_getError)(void);
  static void   (*_setErrorCallback)(void (*callback)(void));
  static CgContext::Param (*_getNamedParameter)(CgContext::Param program, const StlString& name);
  static void   (*_setParameter1f)(CgContext::Param parameter, float x);
  static void   (*_setParameter2f)(CgContext::Param parameter, float x, float y);
  static void   (*_setParameter3f)(CgContext::Param parameter, float x, float y, float z);
  static void   (*_setParameter4f)(CgContext::Param parameter, float x, float y, float z, float w);
  static void   (*_setParameter1fv)(CgContext::Param parameter, const float* array);
  static void   (*_setParameter2fv)(CgContext::Param parameter, const float* array);
  static void   (*_setParameter3fv)(CgContext::Param parameter, const float* array);
  static void   (*_setParameter4fv)(CgContext::Param parameter, const float* array);
  static void   (*_getParameter1f)(CgContext::Param parameter, float* array);
  static void   (*_getParameter2f)(CgContext::Param parameter, float* array);
  static void   (*_getParameter3f)(CgContext::Param parameter, float* array);
  static void   (*_getParameter4f)(CgContext::Param parameter, float* array);
  static void   (*_setParameter1d)(CgContext::Param parameter, double x);
  static void   (*_setParameter2d)(CgContext::Param parameter, double x, double y);
  static void   (*_setParameter3d)(CgContext::Param parameter, double x, double y, double z);
  static void   (*_setParameter4d)(CgContext::Param parameter, double x, double y, double z, double w);
  static void   (*_setParameter1dv)(CgContext::Param parameter, const double* array);
  static void   (*_setParameter2dv)(CgContext::Param parameter, const double* array);
  static void   (*_setParameter3dv)(CgContext::Param parameter, const double* array);
  static void   (*_setParameter4dv)(CgContext::Param parameter, const double* array);
  static void   (*_getParameter1d)(CgContext::Param parameter, double* array);
  static void   (*_getParameter2d)(CgContext::Param parameter, double* array);
  static void   (*_getParameter3d)(CgContext::Param parameter, double* array);
  static void   (*_getParameter4d)(CgContext::Param parameter, double* array);
  static void   (*_setMatrixParameterfr)(CgContext::Param parameter, const float* matrix);
  static void   (*_setMatrixParameterfc)(CgContext::Param parameter, const float* matrix);
  static void   (*_setMatrixParameterdr)(CgContext::Param parameter, const double* matrix);
  static void   (*_setMatrixParameterdc)(CgContext::Param parameter, const double* matrix);
  static void   (*_getMatrixParameterfr)(CgContext::Param parameter, float* matrix);
  static void   (*_getMatrixParameterfc)(CgContext::Param parameter, float* matrix);
  static void   (*_getMatrixParameterdr)(CgContext::Param parameter, double* matrix);
  static void   (*_getMatrixParameterdc)(CgContext::Param parameter, double* matrix);
  static void   (*_setStateMatrixParameter)(CgContext::Param parameter, CgContext::StateMatrixType state, CgContext::TransformMatrixType transform);
  static void   (*_setParameterArray1f)(CgContext::Param parameter, int startIndex, int numberOfElements, const float* array);
  static void   (*_setParameterArray2f)(CgContext::Param parameter, int startIndex, int numberOfElements, const float* array);
  static void   (*_setParameterArray3f)(CgContext::Param parameter, int startIndex, int numberOfElements, const float* array);
  static void   (*_setParameterArray4f)(CgContext::Param parameter, int startIndex, int numberOfElements, const float* array);
  static void   (*_getParameterArray1f)(CgContext::Param parameter, int startIndex, int numberOfElements, float* array);
  static void   (*_getParameterArray2f)(CgContext::Param parameter, int startIndex, int numberOfElements, float* array);
  static void   (*_getParameterArray3f)(CgContext::Param parameter, int startIndex, int numberOfElements, float* array);
  static void   (*_getParameterArray4f)(CgContext::Param parameter, int startIndex, int numberOfElements, float* array);
  static void   (*_setParameterArray1d)(CgContext::Param parameter, int startIndex, int numberOfElements, const double* array);
  static void   (*_setParameterArray2d)(CgContext::Param parameter, int startIndex, int numberOfElements, const double* array);
  static void   (*_setParameterArray3d)(CgContext::Param parameter, int startIndex, int numberOfElements, const double* array);
  static void   (*_setParameterArray4d)(CgContext::Param parameter, int startIndex, int numberOfElements, const double* array);
  static void   (*_getParameterArray1d)(CgContext::Param parameter, int startIndex, int numberOfElements, double* array);
  static void   (*_getParameterArray2d)(CgContext::Param parameter, int startIndex, int numberOfElements, double* array);
  static void   (*_getParameterArray3d)(CgContext::Param parameter, int startIndex, int numberOfElements, double* array);
  static void   (*_getParameterArray4d)(CgContext::Param parameter, int startIndex, int numberOfElements, double* array);
  static void   (*_setMatrixParameterArrayfr)(CgContext::Param parameter, int startIndex, int numberOfElements, const float* array);
  static void   (*_setMatrixParameterArrayfc)(CgContext::Param parameter, int startIndex, int numberOfElements, const float* array);
  static void   (*_getMatrixParameterArrayfr)(CgContext::Param parameter, int startIndex, int numberOfElements, float* array);
  static void   (*_getMatrixParameterArrayfc)(CgContext::Param parameter, int startIndex, int numberOfElements, float* array);
  static void   (*_setMatrixParameterArraydr)(CgContext::Param parameter, int startIndex, int numberOfElements, const double* array);
  static void   (*_setMatrixParameterArraydc)(CgContext::Param parameter, int startIndex, int numberOfElements, const double* array);
  static void   (*_getMatrixParameterArraydr)(CgContext::Param parameter, int startIndex, int numberOfElements, double* array);
  static void   (*_getMatrixParameterArraydc)(CgContext::Param parameter, int startIndex, int numberOfElements, double* array);
  static void   (*_enableClientState)(CgContext::Param parameter);
  static void   (*_disableClientState)(CgContext::Param parameter);
  static void   (*_setParameterPointer)(CgContext::Param parameter, GLint size, GLenum type, GLsizei stride, GLvoid* array);
  static void   (*_enableTextureParameter)(CgContext::Param parameter);
  static void   (*_disableTextureParameter)(CgContext::Param parameter);
  static GLuint (*_getTextureParameter)(CgContext::Param parameter);
  static void   (*_setTextureParameter)(CgContext::Param parameter, GLuint textureId);

  CgContext::CgCTX _context;
};

} // namespace AReVi

#endif // AREVI_CONTRIB_CGCONTEXT_H

//----------------------------------------------------------------------------
