// LI2/ENIB - AReVi-Contrib - Sound player
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/soundPlayer.h
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_CONTRIB_SOUNDPLAYER_H
#define AREVI_CONTRIB_SOUNDPLAYER_H 1

#include "AReVi/arObject.h"
#include "AReVi/scheduler.h"
#include "AReVi/Utils/utilSound.h"

namespace AReVi {

class MemoryBlock;
class SoundPlayer ;

extern "C" bool SoundPlayer_init(const StlVector<StlString> *);
extern "C" void SoundPlayer_destroy(void);

//----------------------------------------------------------------------------
// Class SoundPlayer (abstract)
//----------------------------------------------------------------------------

class SoundPlayer : public ArObject
{
//-------- Type information --------------------------------------------------

public:

  AR_CLASS(SoundPlayer)

  AR_CONSTRUCTOR_2(SoundPlayer,
                   UtilSound::SoundFormat, format,
                   unsigned int, rate)

//-------- Sound properties --------------------------------------------------

  virtual
  UtilSound::SoundFormat // sound format
  getFormat(void) const;

  virtual
  unsigned int // sound rate (commons rate are 11025, 22050, 44100)
  getRate(void) const;

  virtual
  void
  setAutoPop(bool autoPop);

  virtual
  bool // buffers are auto erased from queue when played
  getAutoPop(void) const;

  virtual
  bool // success (volume [0,1])
  setVolume(double leftVolume, double rightVolume);

  virtual
  void
  getVolume(double &leftVolumeOut, double &rightVolumeOut) const;

//-------- Playback control --------------------------------------------------

  virtual
  bool
  play(void);

  virtual
  bool
  stop(void);

  virtual
  bool
  seek(int bufferID);

  virtual
  bool // sound is playing
  isPlaying(void) const;

  virtual
  bool // sound is in play mode (auto-run when buffers available)
  getPlay(void) const;

//-------- Buffers control --------------------------------------------------

  virtual
  int // bufferID, -1 if fail
  pushBuffer(ArConstRef<MemoryBlock> buffer);

 // if allBuffers = true, clear all buffers
 // if allBuffers = false, remove only buffers already played
  virtual
  unsigned int
  popBuffers(bool allBuffers);

  virtual
  int // buffer ID
  getCurrentBuffer(void) const;

  virtual
  unsigned int
  getNbBuffers(void) const;

//-------- Time control --------------------------------------------------

  virtual
  unsigned int // time in millisecond
  getTime(void) const;

  virtual
  bool // success
  setTime(unsigned int newTimePosition);

  virtual
  unsigned int // total buffers duration
  getDuration(void) const;

//-------- Event management --------------------------------------------------

  AR_CALLBACK(SoundPlayer,Seek,_seekCB,
              /* no field */)

  AR_CALLBACK(SoundPlayer,Setup,_setupCB,
              /* no field */)

  AR_CALLBACK(SoundPlayer,Control,_controlCB,
              /* no field */) // play, stop, pause

//-------- Forbidden operations ----------------------------------------------
protected:

  friend bool SoundPlayer_init(const StlVector<StlString> *);
  friend void SoundPlayer_destroy(void);

  virtual void
  _onPolling(const Scheduler::PollingEvent & evt);

  virtual
  void
  _onSetup(void);

  virtual
  void
  _onControl(void);

//-------- Member attributes -------------------------------------------------
protected:

  int _deviceID;
  void *_pluginData;

  bool _playing;

  UtilSound::SoundFormat _soundFormat;
  unsigned _sampleRate;

  bool _autoErase;

  double _leftVolume,_rightVolume;


  CallbackManager<SoundPlayer,SeekEvent> _seekCB;
  CallbackManager<SoundPlayer,SetupEvent> _setupCB;
  CallbackManager<SoundPlayer,ControlEvent> _controlCB;

//-------- Static attributes -------------------------------------------------
protected:
  static bool _initSuccess;

  static void *( *_openSoundOutputFnct)(SoundPlayer *, const StlString &, int &);
  static void ( *_closeSoundOutputFnct)(void *);
  static bool ( *_setPlayFnct)(void *, bool );
  static bool ( *_isPlayingFnct)(const void *);

  static bool ( *_setOutputVolumeFnct)(void *);

  static int (*_pushBufferFnct)(void *, ArConstRef<MemoryBlock> );
  static unsigned ( *_popBuffersFnct)(void *, bool );

  static unsigned ( *_getNbBuffersFnct)(const void *);
  static bool ( *_setCurrentBufferFnct)(void *, int );
  static int ( *_getCurrentBufferFnct)(const void *);

  static unsigned int  ( *_getDurationFnct)(const void *);
  static bool           ( *_setTimeFnct)(void *, unsigned int );
  static unsigned int  ( *_getTimeFnct)(const void *);
};

} // namespace AReVi

#endif // AREVI_CONTRIB_SOUNDPLAYER_H

//----------------------------------------------------------------------------
