// LI2/ENIB - AReVi-Contrib - Text To Speech service
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/text2Speech.h
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

// NOTE D'UTILISATION :
// Need Elans Speech (/usr/elan/nscube/bin/nscube V4.5 => only support ISO-8859-1 encoding)
//                or   /opt/Acapela/TelecomTTS/bin/nscube V5.2
// It is necessary to have nscube server launched.
//
// Default plugin use, is :
//        ArSystem::loadPlugin("Text2Speech")
//
// Speech port can be customized by (0 is the default port number):
//        StlVector<StlString> params;
//        params.push_back("5151");
//        ArSystem::loadPlugin("Text2Speech", &params);

// Server adress can be also specified
//        StlVector<StlString> params;
//        params.push_back("192.168.20.222");
//        params.push_back("5151");
//        ArSystem::loadPlugin("Text2Speech", &params);

#ifndef AREVI_CONTRIB_TEXT2SPEECH_H
#define AREVI_CONTRIB_TEXT2SPEECH_H 1

#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/arObject.h"
#include "AReVi/Utils/utilSound.h"

typedef   long unsigned int     TTSCHANELID;

// Language (windows-like) identifiers
#define TEXT2SPEECH_FRENCH    1036  // french
#define TEXT2SPEECH_BRAZILIAN 1046  // brazilian
#define TEXT2SPEECH_GERMAN    1031  // german
#define TEXT2SPEECH_ENGLISH   2057  // english
#define TEXT2SPEECH_ITALIAN   1040  // italian
#define TEXT2SPEECH_SPANISH   1034  // spanish
#define TEXT2SPEECH_AMERICAN  1033  // us english
#define TEXT2SPEECH_RUSSIAN   1049  // russian
#define TEXT2SPEECH_POLISH    1045  // polish
#define TEXT2SPEECH_DUTCH     1043  // dutch
#define TEXT2SPEECH_LAS       8202  // latin american spanish
#define TEXT2SPEECH_ARABIC    5121  // arabic (Algeria)
#define TEXT2SPEECH_ARABIC_SA 1025  // arabic (Saudi Arabia)


namespace AReVi {

class Activity;
class DataSound3D;
class SoundPlayer;

extern "C" bool Text2Speech_init(const StlVector<StlString> *);
extern "C" void Text2Speech_destroy(void);

//----------------------------------------------------------------------------
// Class Text2Speech
//----------------------------------------------------------------------------

class Text2Speech : public ArObject
{
public:
  typedef bool (ArObject::*TTSSayTextToMemoryMethod)(ArRef<MemoryBlock>,     // NewBuffer
                                                     unsigned int ,          // Frequency
                                                     UtilSound::SoundFormat ,  // SoundFormat
                                                     bool ); // finished
  typedef ArPtr<ArObject>  TTSSayTextToMemoryPerformer;

//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Text2Speech)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR(Text2Speech)

  AR_CONSTRUCTOR_2(Text2Speech, bool , useSoundPlayer,
                                bool , useSound3D)

//-------- TTS Commands  -----------------------------------------------------
//! Speaker say text. If speaker is already speaking, text will be pushed back
//! This function is non blocking (return before end of text generation)
  virtual
  bool
  sayText(const StlString & str);

//! Make a MemoryBlock from text (lock execution!)
  virtual
  ArRef<MemoryBlock>
  sayTextToMemory(const StlString & str, unsigned int & freqOut , UtilSound::SoundFormat &formatOut);

//! Return if TTS is Speaking (generating and/or speaking using SoundPlayer or Sound3D)
  virtual
  bool
  isSpeaking(void) const;

//! Abort Speaking
  virtual
  void
  abortSpeaking(void);

//-------- TTS Configuration -------------------------------------------------
  //! Pitch [0 (1) 4]
  virtual double  getVoicePitch(void) const;
  virtual bool    setVoicePitch(double pitch);

  //! Volume [0 1]
  virtual double  getVoiceVolume(void) const;
  virtual bool    setVoiceVolume(double volume);

  //! Speed [0 (1) 4?]
  virtual double   getVoiceSpeed(void) const;
  virtual bool     setVoiceSpeed(double speed);

  //! Get list of all voices
  //! If specificLanguage!=-1 => getVoices only return voices with this
  //     id (eg : 1036 for french)
  //! If specificFrequency!=-1 => getVoices only return voices with this
  //      frequency(8000, 16000, 22050)
  static
  bool
  getVoices(StlVector<StlString> & voicesOut,
            const int &specificLanguage=-1,
            const int &specificFrequency=-1); // GetOnly Voice with specificFrequency

  //! Return informations about voiceID
  //!   - frequency  = 8000,16000, 22050
  //!   - languageid = 1036 (TEXT2SPEECH_FRENCH) => See above...
  static
  bool
  getVoiceSettings(const StlString &voiceID,
                    unsigned int &frequencyOut,
                    UtilSound::SoundFormat &formatOut,
                    unsigned int &langageIdOut);


  //! Select a voice
  virtual
  bool
  selectVoice(const StlString &voiceName);

  //! Get current voice name
  virtual
  bool
  getCurrentVoice(StlString &voiceNameOut) const;

  //! Select input text to say encoding
  //! default is ISO-8859-1
  virtual
  bool
  setTextEncoding(const StlString &encoding);

  //! get Selected text encoding
  virtual
  StlString
  getTextEncoding(void) const;

//-------- Sound Speaker -----------------------------------------------------

  //! All generated sounds will be played by this dataSound3D
  virtual
  ArConstRef<DataSound3D> // Current 3D player
  getSound3D(void) const;

  virtual
  ArRef<DataSound3D> // Current 3D player
  accessSound3D(void) const;

  //! All generated sounds will be played by this soundPlayer
  virtual
  ArConstRef<SoundPlayer>
  getSoundPlayer(void) const;

  virtual
  ArRef<SoundPlayer>
  accessSoundPlayer(void) const;

//-------- Forbidden operations ----------------------------------------------
protected:
  friend bool Text2Speech_init(const StlVector<StlString> *);
  friend void Text2Speech_destroy(void);

  //! CB pour signaler la fin d'un traitement asynchrone
  virtual bool
  _onSayTextToMemory(ArRef<MemoryBlock> mem,
                      unsigned int freq,
                      UtilSound::SoundFormat format,
                      bool finished);

  //! CB update of text processing => asynchronous mode only.
  virtual bool
  _onSpeaking(ArRef<Activity> act, double dt);

  //! Close Channels
  virtual bool
  _closeAll(void);

// ----------------- DATA -------------------------------------
protected:
  TTSCHANELID _channelID;
  ArRef<Activity> _speakingAct;
  bool _onCancel;

  bool _useSoundPlayer, _useSound3D;
  ArRef<DataSound3D> _dataSound;
  ArRef<SoundPlayer> _soundPlayer;
  StlString _encoding;

  StlString _currentVoice;
  double _voicePitch, _voiceVolume, _voiceSpeed;
//----------------- Plugin datas --------------------------------
protected:
  static bool   _initSuccess;
  static void * _serverHandle;
  static void * _dispatchHandle;

  void * _pInstanceData; // data used in async mode

//-------- Static attributes -------------------------------------------------
  static ArRef<MemoryBlock>  (*_sayTextToMemory)(TTSCHANELID, const StlString &, const StlString &, unsigned int & , UtilSound::SoundFormat &);
  static bool                (*_sayTextToMemoryAsync)(void *, const StlString &, const StlString &);
  static void *              (*_sayTextToMemoryAsyncInit)(TTSCHANELID, TTSSayTextToMemoryPerformer performer, TTSSayTextToMemoryMethod method, void *oldData);

  static bool                (*_getVoices)(StlVector<StlString> &,const int &,const int &);
  static TTSCHANELID         (*_openVoice)(void *, void *, const StlString &);
  static bool                (*_closeVoice)(void *, TTSCHANELID);
  static bool                (*_getVoiceSettings)(const StlString &, unsigned int &, UtilSound::SoundFormat &, unsigned int &);

  static bool                (*_setVoiceParams)(TTSCHANELID, double ,double  ,double );
  static bool                (*_getVoiceParams)(TTSCHANELID, double &,double &,double &);

  static bool                (*_continueProcessing)(void *, int);
  static bool                (*_abortProcessing)(void *);

  static void                (*_getEncodingSupportList)(StlVector<StlString> &encodingListOut);
};


} // namespace AReVi

#endif // AREVI_CONTRIB_TEXT2SPEECH_H

//----------------------------------------------------------------------------
