// LI2/ENIB - AReVi-Contrib - XML management
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/xmlParser.h
// author(s)   : E. Cazeaux, E. Maffre
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_CONTRIB_XMLPARSER_H
#define AREVI_CONTRIB_XMLPARSER_H 1

#include "AReVi/arObject.h"

#define XMLPARSER_DEFAULT_ENCODING    "ISO-8859-1"

namespace AReVi {

extern "C" bool XmlParser_init(const StlVector<StlString> *);
extern "C" void XmlParser_destroy(void);

//----------------------------------------------------------------------------
// Class XmlNode
//----------------------------------------------------------------------------
// IMPORTANT : Les noeuds XmlNode, ne peuvent etre instancie que par le
//               XmlParser. La reference sur un noeud Xml est unique et est
//               valide (!=nullRef()) tant que le document (XmlParser) n'est
//               pas detruit ou qu'un parent du noeud n'a pas ete detruit par
//               'removeChild()'.
//             Dans certains cas, le parseur ajouter des balises 'speciales' qui
//               apparaissent dans la structure (ex: comment, include, text).
//             Tous les textes manipules par XmlParser sont supposes etre codes
//               en UTF8 (16bits). Deux fonctions sont mis a disposition dans
//               AReVi pour effectuer la conversion strToUTF8/strFromUTF8.
//               Les fonctions UTF8ToISO et ISOToUTF8, sont pour le moment
//               conservees, mais sont emmenees a disparaitre.
class XmlParser;

class XmlNode : public ArObject
{
public:
  AR_CLASS(XmlNode);

//----------------------------------------------------------
//! Inspect Node
  //! Nom du noeud XML (ex : "Nom" dans <Nom></Nom>)
  virtual StlString getName(void);

  //! Renvoi le texte dans une balise.
  //! <A>Mon texte</A> => il faut faire nodeA->getTextString();
  //! Retourne false si la balise est vide eg: <A/> ou <A></A>
  //! If useISOencoding=true => use UTF8ToISO to convert text encoding
  virtual bool getTextString(StlString &textOut, const bool &useISOencoding=false);

  //! Renvoi le texte dans une balise. Attention, pour recuperer le texte de :
  //! <A>Mon texte</A> => il faut faire nodeA->getChild("text")->getString();
  virtual StlString getString(void);

  virtual
  StlString // concatenation of text nodes under this one
  flattenText(void);

  //! Lecture des property ex : <A prop1="1" prop2="eee"/>
  virtual	StlString	getProperty(const StlString &name);
  virtual	bool	    getPropertyString(const StlString &name, StlString &value);
  virtual	bool	    getPropertyInteger(const StlString &name, int &value);
  virtual	bool	    getPropertyReal(const StlString &name, double &value);
  virtual	bool	    getPropertyBoolean(const StlString &name, bool &value);
  virtual	bool	    hasProperty(const StlString &name);

  //! Liste des attributs definis pour le noeud
  virtual bool      getPropertyList(StlVector<StlString> &namesOut);

//----------------------------------------------------------
//! Navigate in nodes
  virtual ArRef<XmlNode> getNext(void);
  virtual ArRef<XmlNode> getPrevious(void);
  virtual ArRef<XmlNode> getFirstChild(void);
  virtual ArRef<XmlNode> getLastChild(void);
  virtual	ArRef<XmlNode> getParent(void);

  //! Find the occurence ieme of the child with name "childName"
  virtual ArRef<XmlNode> getChild(const StlString &childName, unsigned int occurence=1);

  //! Test nodes types
  virtual bool isText(void) const;
  virtual bool isComment(void) const;

//----------------------------------------------------------
//! Access on nodes tree
  virtual ArRef<XmlNode> addChild(const StlString &childName);
  virtual ArRef<XmlNode> addText(const StlString &text); // Add a text node (name is "text")
  virtual ArRef<XmlNode> addComment(const StlString &comment); // Add a comment node (name is "comment")
  virtual bool           removeChild(ArRef<XmlNode> node);

//----------------------------------------------------------
//! Setup node
  virtual void setName(const StlString &name);
  virtual	void setProperty(const StlString &name, const StlString &value);
  virtual	void setPropertyString(const StlString &name, const StlString &value);
  virtual	void setPropertyInteger(const StlString &name,const  int &value);
  virtual	void setPropertyReal(const StlString &name, const double &value);
  virtual	void setPropertyBoolean(const StlString &name, const bool &value);
  virtual	void setString(const StlString &texte); // Attention : Remplace tout le "Content" du noeud (texte et enfants)

//-------- TOOLS ----------------------------------------------
  //! Conversion de codage de caractere
  //! UTF8ToISO => affichage sur les flux std (cerr, cout)
  //! ISOToUTF8 => eriture des accents dans le fichier XML
  static StlString UTF8ToISO(const StlString &utf8String);
  static StlString ISOToUTF8(const StlString &ISOString);

//-------- Forbidden operations ----------------------------------------------
 protected:
  //! Ajoute les enfants a un noeud XML
  virtual bool _buildXmlNodeChilds(void);

  //! Enleve un enfant de la liste
  virtual bool _removeChild(void *);

  ArRef<XmlNode> _getChild(void *child)
  {
    StlMap<void *, ArRef<XmlNode> >::iterator it=_childs.find(child);
    if (it!=_childs.end())
      return it->second;
    else return XmlNode::nullRef();
  }

 //-------- Construction / Destruction ----------------------------------------
 //-------- Type information --------------------------------------------------
  friend class XmlParser;
  friend class XmlFlowOut;

  //! Attach the Node to a XmlNode
  XmlNode(ArCW & arCW, ArRef<XmlNode> parent, void *xmlNode);

  static ArRef<XmlNode> _newXmlNode(ArRef<XmlNode> parent, void *node);

  // DATA(s)
  void * _xmlNode;
  ArRef<XmlNode> _parent;
  StlMap<void *, ArRef<XmlNode> > _childs;

  friend bool XmlParser_init(const StlVector<StlString> *);
  friend void XmlParser_destroy(void);
//-------- Static attributes -------------------------------------------------
  static void   (* _freeString)(void * strPtr);
  static char * (* _getName)(void * node);
  static char * (* _getString)(void * node);
  static char * (* _getTextString)(void * node);
  static char * (* _getProperty)(void * node,  StlString propName);
  static void * (* _getNext)(void * node);
  static void * (* _getPrevious)(void * node);
  static void * (* _getFirstChild)(void * node);
  static void * (* _getLastChild)(void * node);
  static void * (* _getParent)(void * node);
  static void * (* _getDoc)(void * node);

  static void   (* _setName)(void * node, const char * name);
  static void * (* _newNode)(void * ns, const char * name);
  static void * (* _copyNode)(const void * node, int recursive);
  static void   (* _freeNode)(void * node, bool unlinkFirst);
  static void * (* _addChild)(void * parent, void *node);
  static void * (* _addText)(void * parent, const char * name);
  static void * (* _addComment)(void * parent, const char * name);
  static bool   (* _isText)(void * node);
  static bool   (* _isComment)(void * node);
  static void   (* _setString)(void * node, const char * name);
  static void   (* _setProperty)(void * node, const StlString &propName, const StlString &value);
  static void   (* _addProperty)(void * node, const StlString &propName, const StlString &value);
  static bool   (* _hasProperty)(void * node, const StlString &propName);
  static bool   (* _getPropertyList)(void * node, StlVector<StlString> &namesOut);

  //! Construit une refence sur un XmlNode
  static ArRef<XmlNode> (* _newRef)(ArRef<XmlNode> parent, void *node);

  static StlString (* _UTF8ToISO)(const StlString &utf8String);
  static StlString (* _ISOToUTF8)(const StlString &ISOString);
};

//----------------------------------------------------------------------------
// Class XmlParser
//----------------------------------------------------------------------------
// NOTA : Les noeuds XmlNode crees via un parseFile/parseFlow,
//        dependent de l'instance du XmlParser qui les a instancie Il faut
//        donc conserver une reference du XmlParser

class XmlParser : public ArObject
{
//-------- Type information --------------------------------------------------
  public:

  AR_CLASS(XmlParser);
//-------- Construction / Destruction ----------------------------------------
  AR_CONSTRUCTOR(XmlParser);

//-------- Reading XML -------------------------------------------------------
//! Read xml File
  virtual	bool parseFile(StlString fileName, bool processInclude=true, bool verbose=true);

//! Read flowed xml File (no-root element!)
  virtual	bool parseFlow(StlString fileName, bool processInclude=true, bool verbose=true);

//! Read xml from stream
  virtual	bool parseStream(ArRef<AbstractIStream> stream, bool processInclude=true, bool verbose=true);

  //-------- XmlParseError -------------------------------------------------------------
  struct XmlParseError
  {
  StlString message; // human-readable informative error message
  StlString file; // In any...
  int line; // Line number of error
  int level; // how consequent is the error (0=>None, 1=>WARNING, 2=>ERROR, 3=>FATAL ERROR)
  };

  virtual const StlVector<XmlParseError> &getParseErrors(void) const;
  static void displayErrors(const StlVector<XmlParseError> &errors) ;

//-------- Writing XML -------------------------------------------------------
//! Save xml File. NB : fileName="-" => stdout
//! encoding ="" => no encoding.
//! Sinon, converti tous les textes en utilisant l'encodage specifie
//! NB : si les textes accentues ne sont pas en UTF8 encoding doit etre "".
  virtual	bool saveFile(StlString fileName,const StlString  &encoding=XMLPARSER_DEFAULT_ENCODING);

//! Save xml top stream
  virtual	bool saveStream(ArRef<AbstractOStream> stream,const StlString  &encoding=XMLPARSER_DEFAULT_ENCODING);


//! Free old document and create an empty one
  virtual	bool newDocument(void);

//! getRoot Node
  virtual	ArRef<XmlNode> getRoot(void);

//! Set the root node
  virtual bool setRoot(const StlString &rootname);

//-------- XPath -------------------------------------------------------------
  struct XmlContent
  {
  ArRef<XmlNode> node;
  StlString property; // "" means content represent directly the node
  };

  bool
  matchPath(const StlString &pathExpr, StlList<XmlContent> &contentsOut) const;

//-------- XSLT -------------------------------------------------------------
  static
  bool // success
  applyXSLT(ArRef<AbstractIStream> xmlInput,
            ArRef<AbstractIStream> styleSheet,
            ArRef<AbstractOStream> output,
            bool processInclude=true,
            StlVector<XmlParser::XmlParseError> *errorsOut=NULL);


//-------- Forbidden operations ----------------------------------------------
 protected:
  //! Construit l'arbre de noeud XML
  virtual bool _buildTreeNodes(void);

//-----------  DATA --------------------------------------------------------------
  void * _xmlDocPtr;
  ArRef<XmlNode> _rootNode;

  StlVector<XmlParseError> _errors;

  friend class XmlNode;
  friend class XmlFlowOut;

  friend bool XmlParser_init(const StlVector<StlString> *);
  friend void XmlParser_destroy(void);

//-------- Static attributes -------------------------------------------------
  static bool _initSuccess;
  static bool   (* _newDocument)(void ** docPtr);
  static bool   (* _parseFile)(StlString fileName,void ** docPtr, bool processInclude, StlVector<XmlParser::XmlParseError> &errors);
  static bool   (* _parseDoc)(char *buffer,void ** docPtr, bool processInclude, StlVector<XmlParser::XmlParseError> &errors);
  static bool   (* _saveFile)(StlString fileName,void * docPtr, const char *encoding);
  static bool   (* _saveDoc)(ArRef<AbstractOStream> stream,void * docPtr, const char *encoding);
  static void   (* _freeDoc)(void * docPtr);
  static void * (* _getRoot)(void * docPtr);
  static bool   (* _setRoot)(void * docPtr, void * node);
  static bool   (* _matchPath)(void * doc, const StlString &, StlList<XmlParser::XmlContent> &);
  static bool   (* _applyXSLT)(ArRef<AbstractIStream> , ArRef<AbstractIStream>, ArRef<AbstractOStream>, bool,StlVector<XmlParser::XmlParseError> &);
};

//----------------------------------------------------------------------------
// Class XmlFlowOut
//----------------------------------------------------------------------------
// Enregistrement directe d'elements XML dans un flux
class XmlFlowOut : public XmlParser
{
  friend class XmlParser;

//-------- Type information --------------------------------------------------
  public:
    AR_CLASS(XmlFlowOut);
//-------- Construction / Destruction ----------------------------------------
    AR_CONSTRUCTOR_1(XmlFlowOut, const StlString &, fileName);

//----------------------------------------------------------
//! Add XmlNode (and subnodes...)
//! NB : si les textes accentues sont en UTF8 encoding doit etre "".
    virtual	bool addXmlNode(ArRef<XmlNode> node, const int &level=0,
                            const StlString &encoding=XMLPARSER_DEFAULT_ENCODING);

//-------- Forbidden operations ----------------------------------------------
 protected:
    void * _xmlBufferPtr;
    FILE *  _file;

  friend bool XmlParser_init(const StlVector<StlString> *);
  friend void XmlParser_destroy(void);

//-------- Static attributes -------------------------------------------------
  static int    (*_xmlNodeDump)(void * bufferPtr, void *nodePtr,const int &level, const char *encoding);
  static int    (*_xmlSaveFileTo)(void * bufferPtr, void * docPtr, const char *encoding);
  static void * (*_xmlOutputBufferCreateFile)(FILE *file);
  static int    (*_xmlOutputBufferFlush)(void *bufferPtr);
  static int    (*_xmlOutputBufferClose)(void *bufferPtr);
};

} // namespace AReVi

#endif // AREVI_CONTRIB_XMLPARSER_H

//----------------------------------------------------------------------------


