// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/base3D.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_BASE3D_H
#define AREVI_LIB3D_BASE3D_H 1

#include "AReVi/arObject.h"

#include "AReVi/Lib3D/util3D_priv.h"

namespace AReVi {

class Index3D;
class PointIndex3D;
class VolumeIndex3D;

//----------------------------------------------------------------------------
// Class Base3D
//----------------------------------------------------------------------------

class Base3D : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Base3D)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR(Base3D)

//-------- Transformations ---------------------------------------------------

  virtual
  void
  globalToLocalLocation(ArRef<Base3D> baseInOut) const;

  virtual
  void
  localToGlobalLocation(ArRef<Base3D> baseInOut) const;

  virtual
  void
  globalToLocalPosition(double & xInOut,
                        double & yInOut,
                        double & zInOut) const;

  virtual
  void
  localToGlobalPosition(double & xInOut,
                        double & yInOut,
                        double & zInOut) const;

  virtual
  void
  globalToLocalVector(double & vxInOut,
                      double & vyInOut,
                      double & vzInOut) const;

  virtual
  void
  localToGlobalVector(double & vxInOut,
                      double & vyInOut,
                      double & vzInOut) const;

  virtual
  void
  globalToLocalOrientation(double & rollInOut,
                           double & pitchInOut,
                           double & yawInOut) const;

  virtual
  void
  localToGlobalOrientation(double & rollInOut,
                           double & pitchInOut,
                           double & yawInOut) const;

  virtual
  void
  getLocalX(double & xOut,
            double & yOut,
            double & zOut) const;

  virtual
  void
  getLocalY(double & xOut,
            double & yOut,
            double & zOut) const;

  virtual
  void
  getLocalZ(double & xOut,
            double & yOut,
            double & zOut) const;

//-------- Location ----------------------------------------------------------

  virtual
  void
  setLocation(ArConstRef<Base3D> base);

  virtual
  void
  move(ArConstRef<Base3D> base);

  virtual
  void
  invertLocation(void);

  virtual
  bool // success
  adjustOrthonormality(void);

//-------- Position ----------------------------------------------------------

  virtual
  void
  getPosition(double & pxOut,
              double & pyOut,
              double & pzOut) const;

  virtual
  void
  setPosition(double px,
              double py,
              double pz);

  virtual
  void
  translate(double dx,
            double dy,
            double dz);

//-------- Orientation -------------------------------------------------------

  virtual
  void
  extractOrientation(double & rollOut,
                     double & pitchOut,
                     double & yawOut) const;

  virtual
  void
  extractRotation(double & xOut,
                  double & yOut,
                  double & zOut,
                  double & angleOut) const;

  virtual
  void
  cancelRotation(void);

  virtual
  void
  setOrientation(double roll,
                 double pitch,
                 double yaw);

  virtual
  void
  setRotation(double x,
              double y,
              double z,
              double angle);

  virtual
  void
  rotate(double x,
         double y,
         double z,
         double dAngle);

  virtual
  void
  roll(double dRoll);

  virtual
  void
  pitch(double dPitch);

  virtual
  void
  yaw(double dYaw);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Base3D,Motion,_motionCB,
              /* no field */)

  virtual
  void
  maskMotion(void);

  virtual
  void
  unmaskMotion(void);

//-------- Attachment --------------------------------------------------------

  enum Attachment { ATTACH_NONE=0,
                    ATTACH_TX=1,
                    ATTACH_TY=2,
                    ATTACH_TZ=4,
                    ATTACH_TRANSLATION=7,
                    ATTACH_ROLL=8,
                    ATTACH_PITCH=16,
                    ATTACH_YAW=32,
                    ATTACH_ROTATION=56,
                    ATTACH_ALL=63 };

  virtual
  void
  attachTo(ArRef<Base3D> base,
           int attachment=Base3D::ATTACH_ALL);

  virtual
  int // attachement to base
  getAttachmentTo(ArConstRef<Base3D> base) const;

  virtual
  void
  getAttachments(StlVector<ArRef<Base3D> > & srcOut,
                 StlVector<ArRef<Base3D> > & dstOut) const;

//-------- Detection ---------------------------------------------------------

  virtual
  void
  viewFirst(double yawVision,
            double pitchVision,
            double range,
            const StlVector<ArRef<Base3D> > & candidates,
            ArRef<Base3D> & elemOut,
            double & yawAngleOut,
            double & pitchAngleOut,
            double & distanceOut) const;

  virtual
  void
  view(double yawVision,
       double pitchVision,
       double range,
       const StlVector<ArRef<Base3D> > & candidates,
       StlVector<ArRef<Base3D> > & elemsOut,
       StlVector<double> & yawAnglesOut,
       StlVector<double> & pitchAnglesOut,
       StlVector<double> & distancesOut) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Data access (internal) --------------------------------------------

  AR_UNSTUBBED inline virtual
  const Util3D::Base &
  getBase(void) const;

  AR_UNSTUBBED inline virtual
  Util3D::Base &
  accessBase(void);

//-------- Forbidden operations ----------------------------------------------
protected:
  friend class Index3D;
  friend class PointIndex3D;
  friend class VolumeIndex3D;

  virtual
  void
  _onMotion(void);

  virtual
  void
  _addIndex(ArRef<Index3D> index,
            void * gridData,
            double precision,
            double factor,
            int level);

  virtual
  void
  _removeIndex(ArRef<Index3D> index);

//-------- Member attributes -------------------------------------------------
protected:

  struct AttachInfo
    {
    ArPtr<Base3D> src;
    ArRef<Base3D> dst;
    int attachment;
    bool propagating;
    Util3D::Dbl3 absT;
    Util3D::Dbl3 absR;
    Util3D::Dbl3 relT;
    Util3D::Dbl3 relR;
    AttachInfo(ArPtr<Base3D> s,
               ArRef<Base3D> d,
               int a);
    };

  struct IndexInfo
    {
    ArRef<Index3D> index;
    void * gridData;
    double precision;
    double factor;
    int level;
    inline IndexInfo(void)
      : index(), gridData((void *)0), precision(0.0), factor(0.0), level(0) {}
    inline IndexInfo(ArRef<Index3D> i,void * g,double p,double f,int l)
      : index(i), gridData(g), precision(p), factor(f), level(l) {}
    };

  Util3D::Base _base;
  CallbackManager<Base3D,MotionEvent> _motionCB;
  unsigned int _motionMask;
  bool _moving;
  StlVector<Base3D::AttachInfo *> _attachInfos;
  Util3D::Dbl3 _oldPos;
  StlVector<Base3D::IndexInfo> _indexes;
};

//-------- Inline code -------------------------------------------------------

inline
const Util3D::Base &
Base3D::getBase(void) const
{
return(_base);
}

inline
Util3D::Base &
Base3D::accessBase(void)
{
return(_base);
}

} // namespace AReVi

#endif // AREVI_LIB3D_BASE3D_H

//----------------------------------------------------------------------------
