// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/index3D.h
// author(s)   : F. Harrouet & K. Manac'h
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_INDEX3D_H
#define AREVI_LIB3D_INDEX3D_H 1

#include "AReVi/arObject.h"

namespace AReVi {

class Base3D;

//----------------------------------------------------------------------------
// Class Index3D (abstract)
//----------------------------------------------------------------------------

class Index3D : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Index3D)

//-------- Detection ---------------------------------------------------------

  virtual
  double // reference cell size
  getUnitSize(void) const;

  virtual
  void
  getCellContent(ArConstRef<ArClass> elemClass,
                 double x,
                 double y,
                 double z,
                 double precision,
                 StlVector<ArRef<Base3D> > & elemsOut) const;

  virtual
  void
  getAll(ArConstRef<ArClass> elemClass,
         StlVector<ArRef<Base3D> > & elemsOut) const;

//-------- Content management ------------------------------------------------

  virtual
  bool // success
  addElement(ArRef<Base3D> elem,
             double precision)=0;

  virtual
  void
  removeElement(ArRef<Base3D> elem)=0;

  virtual
  double // precision with which element is contained or 0
  getElementPrecision(ArConstRef<Base3D> elem) const;

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Index3D,CellChange,_cellChangeCB,
              ArRef<Base3D> element;
              double oldX;
              double oldY;
              double oldZ;
              bool enterIndex;
              bool leaveIndex;)

//-------- Forbidden operations ----------------------------------------------
protected:
  friend class Base3D;

  Index3D(ArCW & arCW,
          double unitSize);

  virtual
  void
  _removeAllElements(void);

  virtual
  void
  _onMotion(Base3D * elem,
            void * gridData,
            double factor,
            int level)=0;

//-------- Member attributes -------------------------------------------------
protected:

  struct Grid
    {
    typedef StlVector<Base3D *> Cell;
    struct XAxis
      {
      int offset;
      StlDeque<Index3D::Grid::Cell *> x;
      inline XAxis(void) : offset(0), x() {}
      };
    struct YAxis
      {
      int offset;
      StlDeque<Index3D::Grid::XAxis *> y;
      inline YAxis(void) : offset(0), y() {}
      };
    struct ZAxis
      {
      int offset;
      StlDeque<Index3D::Grid::YAxis *> z;
      inline ZAxis(void) : offset(0), z() {}
      };
    struct GAxis
      {
      int offset;
      StlDeque<Index3D::Grid::ZAxis *> g;
      inline GAxis(void) : offset(0), g() {}
      };
    Index3D::Grid::GAxis grid;
    StlVector<ArPtr<Base3D> > elems;
    inline void removeCell(int x,int y,int z,int g);
    inline Index3D::Grid::Cell * getExistingCell(int x,int y,int z,int g);
    inline Index3D::Grid::Cell * getCellIfExists(int x,int y,int z,int g);
    inline Index3D::Grid::Cell * getOrCreateCell(int x,int y,int z,int g);
    };

  typedef StlMap<const ArClass *,Index3D::Grid *> GridMap;

  double _unitFactor;
  Index3D::GridMap _grids;
  CallbackManager<Index3D,CellChangeEvent> _cellChangeCB;
};


//----------------------------------------------------------------------------
// Class PointIndex3D
//----------------------------------------------------------------------------

class PointIndex3D : public Index3D
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(PointIndex3D)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(PointIndex3D,
                   double, unitSize)

//-------- Detection ---------------------------------------------------------

  virtual
  void
  getApproxNeighborhood(ArConstRef<ArClass> elemClass,
                        double x,
                        double y,
                        double z,
                        double range,
                        StlVector<ArRef<Base3D> > & elemsOut) const;

  virtual
  void
  viewFirst(ArConstRef<Base3D> origin,
            double yawVision,
            double pitchVision,
            double range,
            ArConstRef<ArClass> elemClass,
            ArRef<Base3D> & elemOut,
            double & yawAngleOut,
            double & pitchAngleOut,
            double & distanceOut) const;

  virtual
  void
  view(ArConstRef<Base3D> origin,
       double yawVision,
       double pitchVision,
       double range,
       ArConstRef<ArClass> elemClass,
       StlVector<ArRef<Base3D> > & elemsOut,
       StlVector<double> & yawAnglesOut,
       StlVector<double> & pitchAnglesOut,
       StlVector<double> & distancesOut) const;

//-------- Content management ------------------------------------------------

  virtual
  bool // success
  addElement(ArRef<Base3D> elem,
             double precision);

  virtual
  void
  removeElement(ArRef<Base3D> elem);

//-------- Forbidden operations ----------------------------------------------
protected:

  virtual
  void
  _onMotion(Base3D * elem,
            void * gridData,
            double factor,
            int level);

};


//----------------------------------------------------------------------------
// Class VolumeIndex3D
//----------------------------------------------------------------------------

class VolumeIndex3D : public Index3D
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(VolumeIndex3D)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(VolumeIndex3D,
                   double, unitSize)

//-------- Detection ---------------------------------------------------------

  virtual
  void
  getApproxCoverage(ArConstRef<ArClass> elemClass,
                    double x,
                    double y,
                    double z,
                    StlVector<ArRef<Base3D> > & elemsOut) const;

  virtual
  void
  getApproxNeighborhood(ArConstRef<ArClass> elemClass,
                        double x,
                        double y,
                        double z,
                        double range,
                        StlVector<ArRef<Base3D> > & elemsOut) const;

//-------- Content management ------------------------------------------------

  virtual
  bool // success
  addElement(ArRef<Base3D> elem,
             double precision);

  virtual
  void
  removeElement(ArRef<Base3D> elem);

//-------- Forbidden operations ----------------------------------------------
protected:

  virtual
  void
  _onMotion(Base3D * elem,
            void * gridData,
            double factor,
            int level);

};

} // namespace AReVi

//-------- Inline source code ------------------------------------------------

#include "AReVi/Lib3D/index3D.Ci"

#endif // AREVI_LIB3D_INDEX3D_H

//----------------------------------------------------------------------------
