// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/object3D.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_OBJECT3D_H
#define AREVI_LIB3D_OBJECT3D_H 1

#include "AReVi/Lib3D/base3D.h"

namespace AReVi {

class Shape3D;
class ShapePart3D;
class BoundingBox3D;
class AbstractInteractor;
class Material3D;
class Scene3D;

//----------------------------------------------------------------------------
// Class Object3D
//----------------------------------------------------------------------------

class Object3D : public Base3D
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Object3D)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR(Object3D)

//-------- Shape management --------------------------------------------------

  virtual
  void
  setShape(ArRef<Shape3D> shape);

  virtual
  void
  setLODShapes(const StlVector<double> & distances,
               const StlVector<ArRef<Shape3D> > & shapes);

  virtual
  ArConstRef<Shape3D> // first shape in LOD
  getShape(void) const;

  virtual
  ArRef<Shape3D> // first shape in LOD
  accessShape(void);

  virtual
  void
  queryShapes(StlVector<double> & distancesOut,
              StlVector<ArRef<Shape3D> > & shapesOut) const;

  virtual
  bool // currently displayed by a renderer
  isCurrentlyDisplayed(void) const;

//-------- Appearance --------------------------------------------------------

  virtual
  void
  setWireFrame(bool wireFrameOn);

  virtual
  bool // wire frame is on
  getWireFrame(void) const;

  virtual
  void
  setShapeBoundingBox(bool shapeBoundingBoxOn);

  virtual
  bool // shapes bounding boxes are drawn
  getShapeBoundingBox(void) const;

  virtual
  void
  setShapePartBoundingBox(bool shapePartBoundingBoxOn);

  virtual
  bool // shape parts bounding boxes are drawn
  getShapePartBoundingBox(void) const;

  virtual
  void
  writeSelectedMaterial(ArConstRef<Material3D> material);

  virtual
  void
  readSelectedMaterial(ArRef<Material3D> materialOut) const;

//-------- Detection ---------------------------------------------------------

  virtual
  void
  computeBoundingBox(ArRef<BoundingBox3D> bboxOut,
                     ArConstRef<Base3D> viewPoint) const;

  virtual
  void
  setIntersectionShape(ArRef<Shape3D> shape);

  virtual
  ArConstRef<Shape3D> // intersection shape
  getIntersectionShape(void) const;

  virtual
  ArRef<Shape3D> // intersection shape
  accessIntersectionShape(void);

  virtual
  void
  setIntersectionLevel(unsigned int level);

  virtual
  unsigned int // 0: nothing  1: shape bbox   2: shape part bbox  3: detail
  getIntersectionLevel(void) const;

  virtual
  double // intersection distance
  intersectRay(ArConstRef<Base3D> ray,
               unsigned int level,
               bool findNearest,
               ArConstRef<Base3D> viewPoint,
	       ArRef<ShapePart3D> & interPartOut,
               unsigned int & triIndexOut) const;

  virtual
  double // intersection distance
  simpleIntersectRay(double sourceX,
                     double sourceY,
                     double sourceZ,
                     double unitDirX,
                     double unitDirY,
                     double unitDirZ,
                     unsigned int level,
                     bool findNearest,
		     ArRef<ShapePart3D> & interPartOut,
                     unsigned int & triIndexOut) const;

  virtual
  bool // collision occured
  testCollision(ArConstRef<Object3D> target,
                unsigned int level,
                ArConstRef<Base3D> viewPoint) const;

  AR_UNSTUBBED virtual
  bool // collision occured
  intersect(ArConstRef<Object3D> target,
            StlVector<Util3D::TriInter> & itOut,
            ArConstRef<Base3D> viewPoint) const;

//-------- Interaction -------------------------------------------------------

  virtual
  void
  setMouseButtonInteraction(bool mouseButtonInteraction);

  virtual
  bool // object is sensible to mouse button interaction
  getMouseButtonInteraction(void) const;

  virtual
  void
  onMouseButtonInteraction(ArRef<AbstractInteractor> source,
		           ArRef<ShapePart3D> part,
		           unsigned int triIndex,
                           int button,
                           bool pressed);

  virtual
  void
  setKeyboardInteraction(bool keyboardInteraction);

  virtual
  bool // object is sensible to keyboard interaction
  getKeyboardInteraction(void) const;

  virtual
  void
  onKeyboardInteraction(ArRef<AbstractInteractor> source,
		        ArRef<ShapePart3D> part,
		        unsigned int triIndex,
                        const StlString & key,
                        bool pressed);

  virtual
  void
  setMotionInteraction(bool motionInteraction);

  virtual
  bool // object is sensible to motion interaction
  getMotionInteraction(void) const;

  virtual
  void
  onMotionInteraction(ArRef<AbstractInteractor> source,
		      ArRef<ShapePart3D> part,
		      unsigned int triIndex,
                      ArRef<Base3D> newLocation,
                      bool begin,
                      bool end);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> output);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Object3D,ShapeChange,_shapeChangeCB,
              /* no field */)

  AR_CALLBACK(Object3D,Appearance,_appearanceCB,
              /* no field */)

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class Scene3D;
  friend class Shape3D;
  friend class Renderer3D;

  virtual
  void
  _onMotion(void);

  virtual
  void
  _rebuildShapesIfNeeded(void);

  virtual
  void
  _shapeDestroyed(Shape3D * shape);

  virtual
  void
  _addScene(ArRef<Scene3D> scene);

  virtual
  void
  _removeScene(ArRef<Scene3D> scene);

  virtual
  void
  _postRedisplay(void);

  virtual
  const Shape3D * // shape to be displayed from point of view (if any)
  _findShape(const Base3D * viewPoint) const;

  virtual
  void
  _onShapeChange(void);

//-------- Member attributes -------------------------------------------------
protected:

  StlVector<ArPtr<Scene3D> > _scenes;
  StlVector<double> _distances;
  StlVector<ArRef<Shape3D> > _shapes;
  ArRef<Shape3D> _interShape;
  bool _wireFrame;
  bool _drawShapeBBox;
  bool _drawShapePartBBox;
  Util3D::Material _selectedMaterial;
  unsigned int _intersectionLevel;
  unsigned int _frameID;
  bool _mouseButtonInteraction;
  bool _keyboardInteraction;
  bool _motionInteraction;
  CallbackManager<Object3D,ShapeChangeEvent> _shapeChangeCB;
  CallbackManager<Object3D,AppearanceEvent> _appearanceCB;
};

} // namespace AReVi

#endif // AREVI_LIB3D_OBJECT3D_H

//----------------------------------------------------------------------------
