// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/osd.h
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_OSD_H
#define AREVI_LIB3D_OSD_H 1

#include "AReVi/arObject.h"
#include "AReVi/Lib3D/util3D_priv.h"

namespace AReVi {

class AbstractInteractor;
class BoundingBox3D;
class Renderer3D;

//----------------------------------------------------------------------------
// Class OSD (abstract)
//----------------------------------------------------------------------------

class OSD : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(OSD)

//-------- Position ----------------------------------------------------------

  enum Location { // Position relative to rendering area
                  LOC_TOP_LEFT,
                  LOC_TOP_CENTER,
                  LOC_TOP_RIGHT,
                  LOC_CENTER_LEFT,
                  LOC_CENTER,
                  LOC_CENTER_RIGHT,
                  LOC_BOTTOM_LEFT,
                  LOC_BOTTOM_CENTER,
                  LOC_BOTTOM_RIGHT,
                  // Absolute position in pixels
                  LOC_FREE };

  virtual
  void
  setLocation(OSD::Location loc);

  virtual
  OSD::Location
  getLocation(void) const;

  // New position in pixels if location type is OSD::LOC_FREE
  virtual
  void
  setPosition(int x,
              int y);

  // renderer not needed if location type is OSD::LOC_FREE
  virtual
  bool // success
  getPosition(ArConstRef<Renderer3D> renderer,
              int & xOut,
              int & yOut) const;

  virtual
  void
  setOrientation(double orientation);

  virtual
  double
  getOrientation(void) const;

  // Spacing between bounds of renderer and osd if location is not OSD::LOC_FREE
  virtual
  void
  setPadding(double x,
             double y);

  virtual
  void
  getPadding(double & xOut,
             double & yOut) const;

  virtual
  void
  setDepth(double depth);

  virtual
  double
  getDepth(void) const;

//-------- Transparency ------------------------------------------------------

  virtual
  void
  setAlpha(double alpha);

  virtual
  double
  getAlpha(void) const;

//-------- Size --------------------------------------------------------------

  virtual
  void
  readBoundingBox(ArConstRef<Renderer3D> renderer,
		  ArRef<BoundingBox3D> bboxOut) const;

//-------- Interaction -------------------------------------------------------

  virtual
  void
  setMouseButtonInteraction(bool mouseButtonInteraction);

  virtual
  bool // osd is sensible to mouse button interaction
  getMouseButtonInteraction(void) const;

  virtual
  void
  onMouseButtonInteraction(ArRef<AbstractInteractor> source,
                           int button,
                           bool pressed);

  virtual
  void
  setMouseMotionInteraction(bool mouseMotionInteraction);

  virtual
  bool // osd is sensible to mouse motion interaction
  getMouseMotionInteraction(void) const;

  virtual
  void
  onMouseMotionInteraction(ArRef<AbstractInteractor> source,
                           int xMouse,
                           int yMouse);

  virtual
  void
  setKeyboardInteraction(bool keyboardInteraction);

  virtual
  bool // osd is sensible to keyboard interaction
  getKeyboardInteraction(void) const;

  virtual
  void
  onKeyboardInteraction(ArRef<AbstractInteractor> source,
                        const StlString & key,
                        bool pressed);

  virtual
  void
  setMotionInteraction(bool motionInteraction);

  virtual
  bool // osd is sensible to motion interaction
  getMotionInteraction(void) const;

  virtual
  void
  onMotionInteraction(ArRef<AbstractInteractor> source,
		      int newX,
		      int newY,
		      double newOrientation,
                      bool begin,
                      bool end);

//-------- Attachment --------------------------------------------------------

  enum Attachment { ATTACH_NONE=0,
                    ATTACH_TX=1,
                    ATTACH_TY=2,
                    ATTACH_TRANSLATION=3,
                    ATTACH_ROTATION=4,
                    ATTACH_ALL=7 };

  virtual
  bool // success if location is OSD::LOC_FREE
  attachTo(ArRef<OSD> osd,
           int attachment=OSD::ATTACH_ALL);

  virtual
  int // attachement to osd
  getAttachmentTo(ArConstRef<OSD> osd) const;

  virtual
  void
  getAttachments(StlVector<ArRef<OSD> > & srcOut,
                 StlVector<ArRef<OSD> > & dstOut) const;

//-------- Event management --------------------------------------------------

  AR_CALLBACK(OSD,Setup,_setupCB,
              /* no field */)

  AR_CALLBACK(OSD, Motion, _motionCB,
	      /* no field */)

  virtual
  void
  maskMotion(void);

  virtual
  void
  unmaskMotion(void);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:
  friend class Renderer3D;
  friend class AbstractInteractor;

  OSD(ArCW& arCW);

  virtual
  void
  _addRenderer(ArPtr<Renderer3D> renderer);

  virtual
  void
  _removeRenderer(ArPtr<Renderer3D> renderer);

  virtual
  void
  _getSize(unsigned int rendererWidth,
	   unsigned int rendererHeight,
	   unsigned int & widthOut,
	   unsigned int & heightOut) const=0;

  virtual
  void
  _getPositionAndSize(unsigned int rendererWidth,
		      unsigned int rendererHeight,
		      int & xOut,
		      int & yOut,
		      unsigned int & widthOut,
		      unsigned int & heightOut) const;

  virtual
  void
  _postRedisplay(void);

  virtual
  void
  _render(unsigned int width,
          unsigned int height) const=0;

  virtual
  void
  _onSetup(void);

  virtual
  void
  _onMotion(void);

//-------- Member attributes -------------------------------------------------
protected:

  struct DepthPrecedence
    {
    inline
    bool
    operator()(ArRef<OSD> osd1,
               ArRef<OSD> osd2)
      {
      return(osd1->getDepth()>osd2->getDepth());
      }
    };

  struct AttachInfo
    {
    ArPtr<OSD> src;
    ArRef<OSD> dst;
    int attachment;
    bool propagating;
    int absTx,absTy;
    int relTx,relTy;
    double absR,relR;
    AttachInfo(ArPtr<OSD> s,
               ArRef<OSD> d,
               int a);
    };

  OSD::Location _location;
  int _x,_y;
  double _orientation;
  double _depth;
  double _alpha;
  Util3D::Dbl2 _padding;
  StlVector<ArPtr<Renderer3D> > _renderers;
  bool _mouseButtonInteraction;
  bool _mouseMotionInteraction;
  bool _keyboardInteraction;
  bool _motionInteraction;
  CallbackManager<OSD,SetupEvent> _setupCB;
  CallbackManager<OSD,MotionEvent> _motionCB;
  unsigned int _motionMask;
  bool _moving;
  StlVector<OSD::AttachInfo *> _attachInfos;
};

} // namespace AReVi

#endif // AREVI_LIB3D_OSD_H

//----------------------------------------------------------------------------
