// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/particleSystem.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_PARTICLESYSTEM_H
#define AREVI_LIB3D_PARTICLESYSTEM_H 1

#include "AReVi/Lib3D/base3D.h"

namespace AReVi {

class Scene3D;
class Renderer3D;
class BoundingBox3D;
class ParticleSystem;
class Texture;

//----------------------------------------------------------------------------
// Class ParticleCollider (abstract)
//----------------------------------------------------------------------------

class ParticleCollider : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(ParticleCollider)

//-------- Location ---------------------------------------------------------

  virtual
  ArConstRef<Base3D> // location
  getLocation(void) const;

  virtual
  ArRef<Base3D> // location
  accessLocation(void);

//-------- Collision management ----------------------------------------------

  virtual
  void
  setDestroyFlag(bool destroyFlag);

  virtual
  bool // destroy flag
  getDestroyFlag(void) const;

  AR_CALLBACK(ParticleCollider,ParticleCollision,_particleCollisionCB,
              unsigned int count;
              double normalSpeedX;
              double normalSpeedY;
              double normalSpeedZ;
              double tangentialSpeedX;
              double tangentialSpeedY;
              double tangentialSpeedZ;
              double dt;
              ArRef<ParticleSystem> particleSystem;)

//-------- Surface parameters ------------------------------------------------

  virtual
  bool // success
  setNormalFactor(double normalFactor);

  virtual
  double // normal factor
  getNormalFactor(void) const;

  virtual
  bool // success
  setTangentialFactor(double tangentialFactor);

  virtual
  double // tangential factor
  getTangentialFactor(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);


//-------- Event management --------------------------------------------------

  AR_CALLBACK(ParticleCollider,Setup,_setupCB,
              /* no field */)

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class ParticleSystem;

  ParticleCollider(ArCW & arCW); // only for serialisation

  ParticleCollider(ArCW & arCW,
                   ArRef<Base3D> location);

  virtual
  void
  _addParticleSystem(ArPtr<ParticleSystem> particleSystem);

  virtual
  void
  _removeParticleSystem(ArPtr<ParticleSystem> particleSystem);

  virtual
  void
  _preUpdate(void);

  struct Particle
    {
    double duration;
    Util3D::Dbl3 oldPos;
    Util3D::Dbl4 pos;
    Util3D::Dbl4 speed;
    Util3D::Dbl4 color;
    Util3D::Dbl2 size;
    ParticleCollider * lastCollider;
    double kmin;
    };

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut)=0;

  virtual
  void
  _onSetup(void);

//-------- Member attributes -------------------------------------------------
protected:

  ArRef<Base3D> _location;
  StlVector<ArPtr<ParticleSystem> > _particleSystems;
  bool _destroyFlag;
  double _nFactor;
  double _tFactor;
  unsigned int _collisionCount;
  Util3D::Dbl3 _nSpeed;
  Util3D::Dbl3 _tSpeed;
  Util3D::Dbl3 _nSpeedSum;
  Util3D::Dbl3 _tSpeedSum;
  CallbackManager<ParticleCollider,ParticleCollisionEvent>
                 _particleCollisionCB;
  CallbackManager<ParticleCollider,SetupEvent> _setupCB;
};


//----------------------------------------------------------------------------
// Class PlaneParticleCollider
//----------------------------------------------------------------------------

class PlaneParticleCollider : public ParticleCollider
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(PlaneParticleCollider)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(PlaneParticleCollider,
                   ArRef<Base3D>, location)

//-------- Forbidden operations ----------------------------------------------
protected:

  PlaneParticleCollider(ArCW & arCW); // only for serialisation

  virtual
  void
  _preUpdate(void);

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut);

//-------- Member attributes -------------------------------------------------
protected:

  double _a,_b,_c,_d;
};


//----------------------------------------------------------------------------
// Class RectangleParticleCollider
//----------------------------------------------------------------------------

class RectangleParticleCollider : public ParticleCollider
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(RectangleParticleCollider)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(RectangleParticleCollider,
                   ArRef<Base3D>, location)

//-------- Size management ---------------------------------------------------

  virtual
  bool // success
  setSize(double sizeX,
          double sizeY);

  virtual
  void
  getSize(double & sizeXOut,
          double & sizeYOut) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:

  RectangleParticleCollider(ArCW & arCW); // only for serialisation

  virtual
  void
  _preUpdate(void);

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut);

//-------- Member attributes -------------------------------------------------
protected:

  double _xSize,_halfXSize;
  double _ySize,_halfYSize;
  Util3D::Dbl3 _center;
  Util3D::Dbl3 _xDir;
  Util3D::Dbl3 _yDir;
  double _a,_b,_c,_d;
};


//----------------------------------------------------------------------------
// Class DiskParticleCollider
//----------------------------------------------------------------------------

class DiskParticleCollider : public ParticleCollider
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(DiskParticleCollider)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(DiskParticleCollider,
                   ArRef<Base3D>, location)

//-------- Radius management -------------------------------------------------

  virtual
  bool // success
  setRadius(double radius);

  virtual
  double // radius
  getRadius(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:

  DiskParticleCollider(ArCW & arCW); // only for serialisation

  virtual
  void
  _preUpdate(void);

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut);

//-------- Member attributes -------------------------------------------------
protected:

  double _radius,_r2;
  Util3D::Dbl3 _center;
  double _a,_b,_c,_d;
};


//----------------------------------------------------------------------------
// Class SphereParticleCollider
//----------------------------------------------------------------------------

class SphereParticleCollider : public ParticleCollider
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(SphereParticleCollider)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(SphereParticleCollider,
                   ArRef<Base3D>, location)

//-------- Radius management -------------------------------------------------

  virtual
  bool // success
  setRadius(double radius);

  virtual
  double // radius
  getRadius(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:

  SphereParticleCollider(ArCW & arCW); // only for serialisation

  virtual
  void
  _preUpdate(void);

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut);

//-------- Member attributes -------------------------------------------------
protected:

  double _radius,_r2,_rInv;
  Util3D::Dbl3 _center;
};


//----------------------------------------------------------------------------
// Class CylinderParticleCollider
//----------------------------------------------------------------------------

class CylinderParticleCollider : public ParticleCollider
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(CylinderParticleCollider)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(CylinderParticleCollider,
                   ArRef<Base3D>, location)

//-------- Size management ---------------------------------------------------

  virtual
  bool // success
  setRadius(double radius);

  virtual
  double // radius
  getRadius(void) const;

  virtual
  bool // success
  setHeight(double height);

  virtual
  double // height
  getHeight(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:

  CylinderParticleCollider(ArCW & arCW); // only for serialisation

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut);

//-------- Member attributes -------------------------------------------------
protected:

  double _radius,_r2,_rInv;
  double _height,_halfHeight;
};


//----------------------------------------------------------------------------
// Class ConeParticleCollider
//----------------------------------------------------------------------------

class ConeParticleCollider : public ParticleCollider
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(ConeParticleCollider)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(ConeParticleCollider,
                   ArRef<Base3D>, location)

//-------- Size management ---------------------------------------------------

  virtual
  bool // success
  setRadius(double radius);

  virtual
  double // radius
  getRadius(void) const;

  virtual
  bool // success
  setHeight(double height);

  virtual
  double // height
  getHeight(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:

  ConeParticleCollider(ArCW & arCW); // only for serialisation

  virtual
  void
  _collide(ParticleCollider::Particle & partInOut);

//-------- Member attributes -------------------------------------------------
protected:

  double _radius,_r2,_rInv;
  double _height,_hInv,_h2Inv,_hCoef;
};


//----------------------------------------------------------------------------
// Class ParticleSystem
//----------------------------------------------------------------------------

class ParticleSystem : public Base3D
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(ParticleSystem)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR(ParticleSystem)

//-------- Emission ----------------------------------------------------------

  virtual
  bool // success
  setSection(double width,
             double height,
             double widthAngle,
             double heightAngle,
             bool rounded);

  virtual
  void
  getSection(double & widthOut,
             double & heightOut,
             double & widthAngleOut,
             double & heightAngleOut,
             bool & roundedOut) const;

  virtual
  bool // success
  setParticleRate(double particleRate);

  virtual
  double // particle rate
  getParticleRate(void) const;

  virtual
  void
  setEmissionSpeed(double speed);

  virtual
  double // particle emission speed
  getEmissionSpeed(void) const;

  virtual
  void
  setEmissionSpeedDispersion(double dispersion);

  virtual
  double // dispersion around particle emission speed
  getEmissionSpeedDispersion(void) const;

//-------- Evolution ---------------------------------------------------------

  virtual
  bool // success
  setParticleDuration(double particleDuration);

  virtual
  double // particle duration
  getParticleDuration(void) const;

  virtual
  unsigned int // number of particles
  getNbParticles(void) const;

  virtual
  void
  setGravity(double gx,
             double gy,
             double gz);

  virtual
  void
  getGravity(double & gxOut,
             double & gyOut,
             double & gzOut) const;

  virtual
  void
  readBoundingBox(ArRef<BoundingBox3D> bboxOut) const;

  virtual
  void
  update(double dt);

//-------- Rendering ---------------------------------------------------------

  virtual
  void
  setLineAppearance(bool drawLines);

  virtual
  bool // draws lines
  getLineAppearance(void) const;

  virtual
  bool // success
  setThickness(unsigned int thickness);

  virtual
  unsigned int // line/point thickness
  getThickness(void) const;

  virtual
  void
  setTexture(ArRef<Texture> texture);

  virtual
  ArConstRef<Texture> // particle texture
  getTexture(void) const;

  virtual
  ArRef<Texture> // particle texture
  accessTexture(void);

  virtual
  void
  setInitialAngle(double angle);

  virtual
  double // initial particle angle
  getInitialAngle(void) const;

  virtual
  void
  setInitialAngleDispersion(double dispersion);

  virtual
  double // dispersion around initial particle angle
  getInitialAngleDispersion(void) const;

  virtual
  void
  setAngularSpeed(double speed);

  virtual
  double // angular particle speed
  getAngularSpeed(void) const;

  virtual
  void
  setAngularSpeedDispersion(double dispersion);

  virtual
  double // dispersion around angular particle speed
  getAngularSpeedDispersion(void) const;

  virtual
  void
  setSize(double width,
          double height);

  virtual
  void
  getSize(double & widthOut,
          double & heightOut) const;

  virtual
  void
  setSizeDispersion(double widthDispersion,
                    double heightDispersion);

  virtual
  void
  getSizeDispersion(double & widthDispersionOut,
                    double & heightDispersionOut) const;

  virtual
  bool // success
  setColor(double r,
           double g,
           double b,
           double a);

  virtual
  void
  getColor(double & rOut,
           double & gOut,
           double & bOut,
           double & aOut) const;

  virtual
  void
  setLighting(bool lighting);

  virtual
  bool // lighting is used for color
  getLighting(void) const;

  virtual
  bool // currently displayed by a renderer
  isCurrentlyDisplayed(void) const;

//-------- Collider management -----------------------------------------------

  virtual
  void
  addCollider(ArRef<ParticleCollider> collider);

  virtual
  void
  removeCollider(ArRef<ParticleCollider> collider);

  virtual
  bool // collider is contained
  hasCollider(ArConstRef<ParticleCollider> collider) const;

  virtual
  unsigned int // number of colliders
  getNbColliders(void) const;

  virtual
  ArConstRef<ParticleCollider> // index'th collider
  getCollider(unsigned int index) const;

  virtual
  ArRef<ParticleCollider> // index'th collider
  accessCollider(unsigned int index);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);


//-------- Event management --------------------------------------------------

  AR_CALLBACK(ParticleSystem,Update,_updateCB,
              double dt;)

  AR_CALLBACK(ParticleSystem,Emission,_emissionCB,
              /* no field */)

  AR_CALLBACK(ParticleSystem,Evolution,_evolutionCB,
              /* no field */)

  AR_CALLBACK(ParticleSystem,Rendering,_renderingCB,
              /* no field */)

  AR_CALLBACK(ParticleSystem,AddCollider,_addColliderCB,
              ArRef<ParticleCollider> collider;)

  AR_CALLBACK(ParticleSystem,RemoveCollider,_removeColliderCB,
              ArRef<ParticleCollider> collider;)

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class ParticleCollider;
  friend class Renderer3D;
  friend class Texture;
  friend class Scene3D;

  virtual
  void
  _addScene(ArRef<Scene3D> scene);

  virtual
  void
  _removeScene(ArRef<Scene3D> scene);

  virtual
  void
  _postRedisplay(void);

  virtual
  bool // blending is needed for rendering
  _needsBlending(void) const;

  virtual
  bool // alpha-test is needed for rendering
  _needsAlphaTest(void) const;

  virtual
  bool // keep particle
  _manageCollisions(ParticleCollider::Particle & partInOut);

  class Particle : public ParticleCollider::Particle
  {
  public:
    inline double getDuration(void) const
      { return(duration); }
    inline ArConstRef<ParticleCollider> getCollider(void) const
      { return(ArConstRef<ParticleCollider>(lastCollider)); }
    inline ArRef<ParticleCollider> accessCollider(void)
      { return(ArRef<ParticleCollider>(lastCollider)); }
    inline const Util3D::Dbl3 & getOldPosition(void) const
      { return(oldPos); }
    inline Util3D::Dbl3 & accessOldPosition(void)
      { return(oldPos); }
    inline const Util3D::Dbl4 & getPosition(void) const
      { return(pos); }
    inline Util3D::Dbl4 & accessPosition(void)
      { return(pos); }
    inline const Util3D::Dbl4 & getSpeed(void) const
      { return(speed); }
    inline Util3D::Dbl4 & accessSpeed(void)
      { return(speed); }
    inline const Util3D::Dbl4 & getColor(void) const
      { return(color); }
    inline Util3D::Dbl4 & accessColor(void)
      { return(color); }
    inline const Util3D::Dbl2 & getSize(void) const
      { return(size); }
    inline Util3D::Dbl2 & accessSize(void)
      { return(size); }
  private:
    friend class ParticleSystem;
    Particle(void); // no implementation
    Particle(const Particle & p); // no implementation
    Particle & operator=(const Particle & p); // no implementation
  };

  virtual
  bool // keep particle
  _initParticle(ParticleSystem::Particle & particleInOut);

  virtual
  bool // keep particle
  _updateParticle(double dt,
                  ParticleSystem::Particle & particleInOut);

  static
  double // randomize value
  _randomize(double value,
             double dispersion);

  virtual
  void
  _render(const Renderer3D * renderer) const;

  virtual
  void // Emission section
  _onEmission(void);

  virtual
  void // Evolution section except update(dt); => updateCB
  _onEvolution(void);

  virtual
  void // Rendering section
  _onRendering(void);

//-------- Member attributes -------------------------------------------------
protected:

  StlVector<ArPtr<Scene3D> > _scenes;
  double _width;
  double _height;
  double _widthAngle;
  double _heightAngle;
  bool _rounded;
  double _particleRate;
  double _emissionSpeed;
  double _emissionSpeedDisp;
  double _particleDuration;
  Util3D::Dbl3 _gravity;
  bool _lines;
  double _thickness;
  ArRef<Texture> _texture;
  double _initAngle;
  double _initAngleDisp;
  double _angularSpeed;
  double _angularSpeedDisp;
  Util3D::Dbl2 _size;
  Util3D::Dbl2 _sizeDisp;
  Util3D::Dbl4 _color;
  bool _lighting;
  Util3D::Dbl3 _avgDir;
  Util3D::BoundingBox _bbox;
  unsigned int _nbBlended;
  StlList<ParticleCollider::Particle> _particles;
  StlVector<ArRef<ParticleCollider> > _colliders;
  unsigned int _frameID;
  CallbackManager<ParticleSystem,UpdateEvent> _updateCB;
  CallbackManager<ParticleSystem,EmissionEvent> _emissionCB;
  CallbackManager<ParticleSystem,EvolutionEvent> _evolutionCB;
  CallbackManager<ParticleSystem,RenderingEvent> _renderingCB;
  CallbackManager<ParticleSystem,AddColliderEvent> _addColliderCB;
  CallbackManager<ParticleSystem,RemoveColliderEvent> _removeColliderCB;
};

} // namespace AReVi

#endif // AREVI_LIB3D_PARTICLESYSTEM_H

//----------------------------------------------------------------------------
