// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/renderer3D.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_RENDERER3D_H
#define AREVI_LIB3D_RENDERER3D_H 1

#include "AReVi/Lib3D/base3D.h"

#include "AReVi/Lib3D/llw_priv.h"

namespace AReVi {

class Scene3D;
class Shape3D;
class ShapePart3D;
class Object3D;
class Light3D;
class ParticleSystem;
class BoundingBox3D;
class OSD;
class AbstractInteractor;

struct StereoData;

//----------------------------------------------------------------------------
// Class Renderer3D (abstract)
//----------------------------------------------------------------------------

class Renderer3D : public Base3D
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Renderer3D)

//-------- View parameters ---------------------------------------------------

  AR_UNSTUBBED virtual
  const Util3D::Perspective &
  getPerspective(void) const;

  virtual
  void
  setFieldOfView(double fieldOfView);

  virtual
  double // field of view
  getFieldOfView(void) const;

  virtual
  double // current aspect ratio
  getAspectRatio(void) const;

  virtual
  void
  setFixedAspectRatio(double aspectRatio);

  virtual
  double // fixed aspect ratio (whatever width/height) or 0.0 if not fixed
  getFixedAspectRatio(void) const;

  virtual
  void
  setNearDistance(double nearDistance);

  virtual
  double // near clipping plane distance
  getNearDistance(void) const;

  virtual
  void
  setFarDistance(double farDistance);

  virtual
  double // far clipping plane distance
  getFarDistance(void) const;

  virtual
  void
  setNearYOffset(double nearYOffset);

  virtual
  double // near clipping plane horizontal offset
  getNearYOffset(void) const;

  virtual
  void
  setNearZOffset(double nearZOffset);

  virtual
  double // near clipping plane vertical offset
  getNearZOffset(void) const;

  virtual
  double // left clipping plane slope
  getLeftSlope(void) const;

  virtual
  double // right clipping plane slope
  getRightSlope(void) const;

  virtual
  double // up clipping plane slope
  getUpSlope(void) const;

  virtual
  double // down clipping plane slope
  getDownSlope(void) const;

  virtual
  void
  setWireFrame(bool wireFrameOn);

  virtual
  bool // wire frame is on
  getWireFrame(void) const;

  virtual
  void
  setHeadLight(bool headLightOn);

  virtual
  bool // head light is on
  getHeadLight(void) const;

  virtual
  void
  setOriginVisible(bool originVisible);

  virtual
  bool // origine is visible
  getOriginVisible(void) const;

  virtual
  bool // something is visible
  getCenterOfView(double & xOut,
                  double & yOut,
                  double & zOut) const;

  virtual
  void
  setShapeBoundingBox(bool shapeBoundingBoxOn);

  virtual
  bool // shapes bounding boxes are drawn
  getShapeBoundingBox(void) const;

  virtual
  void
  setShapePartBoundingBox(bool shapePartBoundingBoxOn);

  virtual
  bool // shape parts bounding boxes are drawn
  getShapePartBoundingBox(void) const;

  enum StereoMode
    {
    STEREO_NONE,
    STEREO_SPLIT,
    STEREO_COLOR
    };

  virtual
  void
  setStereoMode(Renderer3D::StereoMode stereoMode);

  virtual
  Renderer3D::StereoMode // current stereo mode
  getStereoMode(void) const;

  virtual
  void
  setStereoLeftColor(double red,
                     double green,
                     double blue);

  virtual
  void
  getStereoLeftColor(double & redOut,
                     double & greenOut,
                     double & blueOut) const;

  virtual
  void
  setStereoRightColor(double red,
                      double green,
                      double blue);

  virtual
  void
  getStereoRightColor(double & redOut,
                      double & greenOut,
                      double & blueOut) const;

  virtual
  void
  setColorStereoMaxResolution(unsigned int maxWidth,
                              unsigned int maxHeight);

  virtual
  void
  getColorStereoMaxResolution(unsigned int & maxWidthOut,
                              unsigned int & maxHeightOut) const;

  virtual
  void
  setStereoViewDistance(double viewDistance);

  virtual
  double // view distance for stereo
  getStereoViewDistance(void) const;

  virtual
  void
  setStereoEyeSeparation(double eyeSeparation);

  virtual
  double // eye separation for stereo
  getStereoEyeSeparation(void) const;

  virtual
  void
  setStereoGlance(bool stereoGlanceOn);

  virtual
  bool // stereo glance is drawn
  getStereoGlance(void) const;

//-------- On Screen Display -------------------------------------------------

  virtual
  void
  addOSD(ArRef<OSD> osd);

  virtual
  void
  removeOSD(ArRef<OSD> osd);

  virtual
  bool // osd is contained
  hasOSD(ArConstRef<OSD> osd) const;

  unsigned int // number of OSD contained
  getNbOSD(void) const;

  ArConstRef<OSD> // index'th OSD contained
  getOSD(unsigned int index) const;

  ArRef<OSD> // index'th OSD contained
  accessOSD(unsigned int index);

//-------- Rendering ---------------------------------------------------------

  virtual
  void
  selectScene(ArRef<Scene3D> scene);

  virtual
  ArConstRef<Scene3D> // main scene
  getScene(void) const;

  virtual
  ArRef<Scene3D> // main scene
  accessScene(void);

  virtual
  void
  setBackScene(ArRef<Scene3D> scene);

  virtual
  ArConstRef<Scene3D> // back scene
  getBackScene(void) const;

  virtual
  ArRef<Scene3D> // back scene
  accessBackScene(void);

  virtual
  void
  setFrontScene(ArRef<Scene3D> scene);

  virtual
  ArConstRef<Scene3D> // front scene
  getFrontScene(void) const;

  virtual
  ArRef<Scene3D> // front scene
  accessFrontScene(void);

  static
  double // frames / second
  getFrameRate(void);

  static
  void
  setFrameRateLimit(double limit);

  static
  double // max frames / second allowed
  getFrameRateLimit(void);

//-------- Appearance --------------------------------------------------------

  virtual
  void
  setBackgroundColor(double red,
                     double green,
                     double blue);

  virtual
  void
  getBackgroundColor(double & redOut,
                     double & greenOut,
                     double & blueOut) const;

  virtual
  void
  setHeadLightColor(double red,
                    double green,
                    double blue);

  virtual
  void
  getHeadLightColor(double & redOut,
                    double & greenOut,
                    double & blueOut) const;

  virtual
  void
  setBoundingBoxColor(double red,
                      double green,
                      double blue);

  virtual
  void
  getBoundingBoxColor(double & redOut,
                      double & greenOut,
                      double & blueOut) const;

//-------- Rendering area ----------------------------------------------------

  virtual
  void
  setRenderSize(unsigned int width,
                unsigned int height)=0;

  virtual
  void
  getRenderSize(unsigned int & widthOut,
                unsigned int & heightOut) const;

  static
  void
  getScreenSize(unsigned int & widthOut,
                unsigned int & heightOut);

//-------- Snapshot ----------------------------------------------------------

  virtual
  bool // success
  snapshotToFile(const StlString & fileName) const;

  virtual
  bool // success (dataOut filled from bottom to top)
  takeSnapshot(unsigned int & widthOut,
               unsigned int & heightOut,
               unsigned int & componentsOut,
               ArRef<MemoryBlock> dataOut) const;

  static
  bool // success
  saveSnapshot(const StlString & fileName,
               unsigned int width,
               unsigned int height,
               unsigned int components,
               ArConstRef<MemoryBlock> data);

//-------- Interaction -------------------------------------------------------

  static
  void
  moveSystemCursor(int dx,
                   int dy);

  virtual
  void
  computeCursorDirection(int xMouse,
                         int yMouse,
                         double & localYawOut,
                         double & localPitchOut) const;

  virtual
  void
  onMotionInteraction(ArRef<AbstractInteractor> source,
                      ArRef<Base3D> newLocation,
                      bool begin,
                      bool end);

  enum CloseAction
    {
    CLOSE_NOP,     // do nothing
    CLOSE_DESTROY, // destroy renderer
    CLOSE_LEAVE    // leave simulation
    };

  virtual
  void
  setCloseAction(Renderer3D::CloseAction closeAction);

  virtual
  Renderer3D::CloseAction // action performed on close event
  getCloseAction(void) const;

  enum Modifier
    {
    MDF_NONE=LLW::MDF_NONE,
    MDF_CONTROL=LLW::MDF_CONTROL,
    MDF_SHIFT=LLW::MDF_SHIFT,
    MDF_CAPSLOCK=LLW::MDF_CAPSLOCK,
    MDF_NUMLOCK=LLW::MDF_NUMLOCK,
    MDF_LEFT=LLW::MDF_LEFT,
    MDF_MIDDLE=LLW::MDF_MIDDLE,
    MDF_RIGHT=LLW::MDF_RIGHT
    };

  virtual
  Renderer3D::Modifier // key/button modifiers
  getModifiers(void) const;

  virtual
  bool // [Shift] key is pressed
  shiftPressed(void) const;

  virtual
  bool // [Control] key is pressed
  controlPressed(void) const;

  virtual
  bool // [CapsLock] key is locked
  capsLockPressed(void) const;

  virtual
  bool // [NumLock] key is locked
  numLockPressed(void) const;

  virtual
  bool // left button is pressed
  leftPressed(void) const;

  virtual
  bool // middle button is pressed
  middlePressed(void) const;

  virtual
  bool // right button is pressed
  rightPressed(void) const;

  virtual
  void
  getLastMousePosition(int & xOut,
                       int & yOut) const;

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Renderer3D,Close,_closeCB,
              /* no field */)

  AR_CALLBACK(Renderer3D,Keyboard,_keyboardCB,
              StlString key;
              bool pressed;
              int xMouse;
              int yMouse;)

  AR_CALLBACK(Renderer3D,MouseButton,_mouseButtonCB,
              int button;
              bool pressed;
              int xMouse;
              int yMouse;)

  AR_CALLBACK(Renderer3D,MouseMotion,_mouseMotionCB,
              int xMouse;
              int yMouse;)

  AR_CALLBACK(Renderer3D,TimeOut,_timeOutCB,
              unsigned int timeOut;)

  virtual
  void
  simulateClose(void);

  virtual
  void
  simulateKeyboard(Renderer3D::Modifier modifiers,
                   const StlString & key,
                   bool pressed,
                   int xMouse,
                   int yMouse);

  virtual
  void
  simulateMouseButton(Renderer3D::Modifier modifiers,
                      int button,
                      bool pressed,
                      int xMouse,
                      int yMouse);

  virtual
  void
  simulateMouseMotion(Renderer3D::Modifier modifiers,
                      int xMouse,
                      int yMouse);

  virtual
  unsigned int // expiration time
  postTimeOut(unsigned int delay);

  virtual
  void
  cancelTimeOut(unsigned int timeOut);

  AR_CALLBACK(Renderer3D,ViewParameters,_viewParametersCB,
              /* no field */)

  AR_CALLBACK(Renderer3D,RenderingSetup,_renderingSetupCB,
              /* no field */) // and rendering size

  AR_CALLBACK(Renderer3D,AppearanceSetup,_appearanceSetupCB,
              /* no field */)

  AR_CALLBACK(Renderer3D,AddOSD,_addOSDCB,
              ArPtr<OSD> osd;)

  AR_CALLBACK(Renderer3D,RemoveOSD,_removeOSDCB,
              ArPtr<OSD> osd;)

  AR_CALLBACK(Renderer3D,AboutToRender,_aboutToRenderCB,
              /* no field */)

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class ArSystem;
  friend class Scene3D;
  friend class Object3D;
  friend class OSD;
  friend class Shape3D;
  friend class ShapePartRenderer;
  friend class ParticleSystem;
  friend class AbstractInteractor;

  static
  void
  _init(void);

  static
  void
  _destroy(void);

  Renderer3D(ArCW & arCW);

  virtual
  void
  _afterRenderAreaSetup(void);

  static
  void
  _checkEvents(bool blocking);

  virtual
  bool // event processed
  _processEvent(const LLW::Event * evt);

  virtual
  void
  _manageClose(void);

  virtual
  void
  _onMotion(void);

  virtual
  void
  _rebuildShapesIfNeeded(void);

  virtual
  void
  _postRedisplay(void);

  static
  bool // redisplay was needed
  _redisplay(void);

  virtual
  void
  _updateLookAt(double eyeX,
                double eyeY,
                double eyeZ);

  virtual
  void
  _preRender(void);

  virtual
  void
  _renderNoSwap(void);

  virtual
  void
  _postRender(void);

  virtual
  void
  _render(void);

  virtual
  void
  _beginScene(const Scene3D * scene);

  virtual
  void
  _renderScene(const Scene3D * scene);

  virtual
  void
  _endScene(const Scene3D * scene);

  struct RenderInfo
    {
    const ParticleSystem * particleSystem;
    Util3D::Base location;
    const Object3D * object;
    const Shape3D * shape;
    const ShapePart3D * part;
    double squareDistance;
    bool  selected;
    inline RenderInfo(void)
      : location(false) {} // other members are not initialized !
    };

  virtual
  void
  _drawVisibleInfos(StlVector<Renderer3D::RenderInfo *> & info,
                    const Scene3D * scene);

  static
  Renderer3D::RenderInfo * // (re)allocated render info
  _makeParticleSystemInfo(const ParticleSystem * particleSystem);

  static
  Renderer3D::RenderInfo * // (re)allocated render info
  _makeObject3DInfo(const Util3D::Base & location,
                    const Object3D * object,
                    const Shape3D * shape,
                    const ShapePart3D * part,
                    bool selected);

  virtual
  void
  _insertTransparentInfo(Renderer3D::RenderInfo * info);

  virtual
  void
  _drawBoundingBox(const Util3D::BoundingBox & bbox,
                   const Scene3D * scene);

  virtual
  void
  _drawGlance(const Scene3D * scene);

  virtual
  void
  _onViewParameters(void);

  virtual
  void
  _onRenderingSetup(void);

  virtual
  void
  _onAppearanceSetup(void);

//-------- Member attributes -------------------------------------------------
protected:

  Util3D::Perspective _perspective;
  double _fixedAspectRatio;
  Util3D::Dbl3 _backgroundColor;
  Util3D::Dbl3 _headLightColor;
  Util3D::Dbl3 _bboxColor;
  bool _wireFrame;
  bool _headLight;
  bool _originVisible;
  bool _drawShapeBBox;
  bool _drawShapePartBBox;
  Renderer3D::StereoMode _stereoMode;
  Util3D::Dbl3 _leftColor;
  Util3D::Dbl3 _rightColor;
  unsigned int _stereoMaxWidth;
  unsigned int _stereoMaxHeight;
  bool _drawStereoGlance;
  double _stereoViewDistance;
  double _stereoEyeSeparation;
  Util3D::Base _lookAt;
  ArRef<Scene3D> _scenes[3];
  StlVector<Renderer3D::RenderInfo *> _opaqueInfos;
  StlVector<Renderer3D::RenderInfo *> _alphaTestInfos;
  StlVector<Renderer3D::RenderInfo *> _transparentInfos;
  StlVector<const Light3D *> _lights;
  StlVector<ArRef<OSD> > _osd;
  unsigned int _nbSeen;
  double _xSeen,_ySeen,_zSeen;
  LLW::AreaID _renderArea;
  LLW::GLBufID _stereoBuffer;
  Renderer3D::CloseAction _closeAction;
  Renderer3D::Modifier _modifiers;
  int _lastMouseX,_lastMouseY;
  CallbackManager<Renderer3D,CloseEvent> _closeCB;
  CallbackManager<Renderer3D,KeyboardEvent> _keyboardCB;
  CallbackManager<Renderer3D,MouseButtonEvent> _mouseButtonCB;
  CallbackManager<Renderer3D,MouseMotionEvent> _mouseMotionCB;
  CallbackManager<Renderer3D,TimeOutEvent> _timeOutCB;
  CallbackManager<Renderer3D,ViewParametersEvent> _viewParametersCB;
  CallbackManager<Renderer3D,RenderingSetupEvent> _renderingSetupCB;
  CallbackManager<Renderer3D,AppearanceSetupEvent> _appearanceSetupCB;
  CallbackManager<Renderer3D,AddOSDEvent> _addOSDCB;
  CallbackManager<Renderer3D,RemoveOSDEvent> _removeOSDCB;
  CallbackManager<Renderer3D,AboutToRenderEvent> _aboutToRenderCB;

//-------- Static attributes -------------------------------------------------
protected:

  static bool _needsRedisplay;
  static unsigned int _lastFrameRate;
  static unsigned int _currentFrameID;
  static unsigned int _nbFrames;
  static double _frameRate;
  static unsigned int _lastRedraw;
  static unsigned int _redrawInterval;
  static StlVector<ArPtr<Renderer3D> > * _renderers;
  static StlVector<Renderer3D::RenderInfo *> * _renderInfoPool;
};

} // namespace AReVi

#endif // AREVI_LIB3D_RENDERER3D_H

//----------------------------------------------------------------------------
