// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/sound3D.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_SOUND3D_H
#define AREVI_LIB3D_SOUND3D_H 1

#include "AReVi/Lib3D/base3D.h"
#include "AReVi/Utils/utilSound.h"

namespace AReVi {

extern "C" bool Sound3DPlayer_init(const StlVector<StlString> *);
extern "C" void Sound3DPlayer_destroy(void);

class Scene3D;
class Listener3D;
class MemoryBlock;

//----------------------------------------------------------------------------
// Class Sound3D (abstract)
//----------------------------------------------------------------------------

class Sound3D : public Base3D
{
//-------- Type information --------------------------------------------------

public:

  AR_CLASS(Sound3D)

//-------- Sound properties --------------------------------------------------

  virtual
  void
  setRepeat(bool repeat);

  virtual
  bool // sound repeats
  getRepeat(void) const;

  virtual
  void
  setGain(double gain);

  virtual
  double // gain
  getGain(void) const;

  virtual
  void
  setPitch(double pitch);

  virtual
  bool // pitch
  getPitch(void) const;

//-------- Attenuation -------------------------------------------------------

  virtual
  void
  setAttenuated(bool attenuated);

  virtual
  bool
  getAttenuated(void) const;


  virtual
  void
  setDistanceAttenuation(double minDistance,
                         double maxDistance);

  virtual
  void
  getDistanceAttenuation(double & minDistanceOut,
                         double & maxDistanceOut) const;

  virtual
  void
  setDirectional(bool directional);

  virtual
  bool // sound is directional
  getDirectional(void) const;

  virtual
  void
  setDirectionalAttenuation(double innerAngle,
                            double outerAngle,
                            double outerGain);

  virtual
  void
  getDirectionalAttenuation(double & innerAngleOut,
                            double & outerAngleOut,
                            double & outerGainOut) const;

  virtual
  unsigned int // nb pending buffers
  getNbBuffers(void) const;

//-------- Playback control --------------------------------------------------

  virtual
  void
  play(void);

  virtual
  void
  stop(void);

  virtual
  void
  pause(void);

  virtual
  bool // sound is in playing mode
  getPlayMode(void) const;

  virtual
  bool // sound is stopped
  isStopped(void) const;

  virtual
  bool // sound is paused
  isPaused(void) const;

  static
  void
  playSounds(const StlVector<ArRef<Sound3D> > & sounds);

  static
  void
  stopSounds(const StlVector<ArRef<Sound3D> > & sounds);

  static
  void
  pauseSounds(const StlVector<ArRef<Sound3D> > & sounds);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Sound3D,Setup,_setupCB,
              /* no field */)

  AR_CALLBACK(Sound3D,Control,_controlCB,
              /* no field */) // play, stop, pause

//-------- Forbidden operations ----------------------------------------------
protected:

//-------- Construction / Destruction ----------------------------------------

  Sound3D(ArCW & arCW);

  struct SoundData
  {
  int bufferID;
  int freq;
  UtilSound::SoundFormat format;
  };

  friend class Scene3D;
  friend class Listener3D;
  friend bool Sound3DPlayer_init(const StlVector<StlString> *);
  friend void Sound3DPlayer_destroy(void);

  virtual
  void
  _onMotion(void);

  virtual
  void
  _mute(bool mute);

  virtual
  void
  _updateSoundLocation(void) const;

  virtual
  void
  _addScene(ArRef<Scene3D> scene);

  virtual
  void
  _removeScene(ArRef<Scene3D> scene);

  virtual
  void
  _queryState(void);

  static
  void
  _freeBuffer(int bufferID);

  virtual
  bool
  _pushBuffer(ArConstRef<MemoryBlock> buffer, // uncompressed data!
              unsigned int freq,
              UtilSound::SoundFormat format);

  virtual
  unsigned int// nb played buffers
  _freeBuffers(void);

  virtual
  void
  _onSetup(void);

  virtual
  void
  _onControl(void);

//-------- Member attributes -------------------------------------------------
protected:

  StlVector<ArPtr<Scene3D> > _scenes;
  int _sourceID;
  StlVector<SoundData> *_soundData;

  bool _repeat;
  double _gain;
  double _pitch;
  double _minDistance;
  double _maxDistance;
  bool _directional;
  double _innerAngle;
  double _cosInnerAngle;
  double _outerAngle;
  double _cosOuterAngle;
  double _outerGain;

  bool _stopped; // cache state data
  bool _paused;  // cache state data

  bool _playMode;

  bool _isMute;
  bool _isAttenuated;

  CallbackManager<Sound3D,SetupEvent> _setupCB;
  CallbackManager<Sound3D,ControlEvent> _controlCB;

//-------- Static attributes -------------------------------------------------
protected:
  static ArPtr<Listener3D> _currentListener;
  static bool _initSuccess;
  static int (* _loadBufferFnct)(const char *);
  static int (* _createBufferFromMemoryFnct)(const char *,
                                             unsigned int,unsigned int,
                                             UtilSound::SoundFormat);
  static void (* _freeBufferFnct)(int);
  static int (* _createSourceFnct)(void);
  static void (* _destroySourceFnct)(int);
  static void (* _setSourceBufferFnct)(int,int);
  static void (* _addSourceBufferFnct)(int,int);
  static void (* _removeSourceBufferFnct)(int,int,StlVector<unsigned int> &);
  static void (* _getSourceBuffersStateFnct)(int,int &,int &);
  static void (* _setSourceGainFnct)(int,double);
  static void (* _setSourcePitchFnct)(int,double);
  static void (* _setSourceRepeatFnct)(int,bool);
  static void (* _setSourceRelativeFnct)(int,bool);
  static void (* _setSourcePositionFnct)(int,float[3]);
  static void (* _setSourceVelocityFnct)(int,float[3]);
  static void (* _playSourcesFnct)(int *,unsigned int);
  static void (* _pauseSourcesFnct)(int *,unsigned int);
  static void (* _stopSourcesFnct)(int *,unsigned int);
  static int (* _getSourceStateFnct)(int);
  static void (* _setListenerPositionFnct)(float[3]);
  static void (* _setListenerVelocityFnct)(float[3]);
  static void (* _setListenerOrientationFnct)(float[6]);
  static void (* _setListenerGainFnct)(double);
};


//----------------------------------------------------------------------------
// Class Listener3D
//----------------------------------------------------------------------------

class Listener3D : public Base3D
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Listener3D)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR(Listener3D)

//-------- Rendering ---------------------------------------------------------

  virtual
  void
  selectScene(ArRef<Scene3D> scene);

  virtual
  ArConstRef<Scene3D> // eared scene
  getScene(void) const;

  virtual
  ArRef<Scene3D> // eared scene
  accessScene(void);

  static
  void
  setCurrent(ArRef<Listener3D> listener);

  static
  ArConstRef<Listener3D> // current listener
  getCurrent(void);

  static
  ArRef<Listener3D> // current listener
  accessCurrent(void);

//-------- Sound properties --------------------------------------------------

  virtual
  void
  setGain(double gain);

  virtual
  double // gain
  getGain(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Listener3D,Setup,_setupCB,
              /* no field */)

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class Sound3D;

  virtual
  void
  _onMotion(void);

  virtual
  void
  _updateListenerLocation(void) const;

  virtual
  void
  _onSetup(void);

//-------- Member attributes -------------------------------------------------
protected:

  ArRef<Scene3D> _scene;
  double _gain;
  CallbackManager<Listener3D,SetupEvent> _setupCB;
};

} // namespace AReVi

#endif // AREVI_LIB3D_SOUND3D_H

//----------------------------------------------------------------------------
