// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/texture.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_TEXTURE_H
#define AREVI_LIB3D_TEXTURE_H 1

#include "AReVi/arObject.h"

namespace AReVi {

class ShapePart3D;
class ParticleSystem;
class Scene3D;
class MemoryBlock;

//----------------------------------------------------------------------------
// Class Texture (abstract)
//----------------------------------------------------------------------------

class Texture : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Texture)

//-------- Inspection --------------------------------------------------------

  virtual
  unsigned int // texture width before rounding to 2^N
  getInitialWidth(void) const;

  virtual
  unsigned int // texture height before rounding to 2^N
  getInitialHeight(void) const;

  virtual
  unsigned int // texture width
  getWidth(void) const;

  virtual
  unsigned int // texture height
  getHeight(void) const;

  virtual
  unsigned int // 1: grey  2: grey-alpha  3: RGB  4: RGB-alpha
  getComponents(void) const;

  virtual
  bool // blending is needed
  getBlending(void) const;

  virtual
  bool // alpha test is needed
  getAlphaTest(void) const;

  virtual
  bool // mipmapping is used by OpenGL
  usesMipmap(void) const;

  virtual
  bool // cube mapping is used instead of normal 2D mapping by OpenGL(extension needed)
  usesCubemap(void) const;

  virtual
  unsigned int // OpenGL texture ID
  getTextureID(void) const;

  virtual
  ArConstRef<MemoryBlock> // data filled from bottom to top
  getData(void) const;

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Utilities operations ----------------------------------------------

  // Maps from vector order : Front, Back, Left, Right, Top, Bottom
  // NB : width must be equal to height, and all maps must have the same dims.
  static
  ArRef<MemoryBlock> // Cube map data
  buildCubeTexture(unsigned &widthInOut, unsigned &heightInOut, unsigned &componentsInOut,
                   const StlVector<ArRef<MemoryBlock> > &maps, double attenuation=1.0);

  static
  bool // success
  buildCubeTexture(const StlString &frontUrl,const StlString &backUrl,
                   const StlString &leftUrl,const StlString &rightUrl,
                   const StlString &topUrl,const StlString &bottomUrl,
                   const StlString &outUrl, double attenuation=1.0);

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class ShapePart3D;
  friend class Surface3D;
  friend class ParticleSystem;
  friend class Scene3D;
  friend class ArSystem;

  Texture(ArCW & arCW);

  static
  void
  _init(void);

  static
  void
  _destroy(void);

  static
  unsigned int // size rounded up to 2^N
  _roundSize(unsigned int size);

  static
  void
  _downScale(unsigned int inWidth,
             unsigned int inHeight,
             unsigned int components,
             const char * inData,
             unsigned int outWidth,
             unsigned int outHeight,
             char * outData);

  static
  void
  _upScale(unsigned int inWidth,
           unsigned int inHeight,
           unsigned int components,
           const char * inData,
           unsigned int outWidth,
           unsigned int outHeight,
           char * outData);

  virtual
  bool // success
  _load(bool checkAlpha,
        bool compress);

  virtual
  void
  _postRedisplay(void);

  virtual
  void
  _postAppearanceRebuild(void);

  virtual
  void
  _addShapePart(ShapePart3D * part);

  virtual
  void
  _removeShapePart(ShapePart3D * part);

  virtual
  void
  _addParticleSystem(ParticleSystem * particleSystem);

  virtual
  void
  _removeParticleSystem(ParticleSystem * particleSystem);

  virtual
  void
  _addScene(Scene3D * scene);

  virtual
  void
  _removeScene(Scene3D * scene);

//-------- Member attributes -------------------------------------------------
protected:

  struct TexData
    {
    unsigned int texId;
    unsigned int initWidth;
    unsigned int initHeight;
    unsigned int width;
    unsigned int height;
    unsigned int components;
    bool needsBlending;
    bool needsAlphaTest;
    bool useMipmap;
    bool ownTexture;
    bool useCubeMap;
    ArRef<MemoryBlock> data;
    TexData(unsigned int t,
            unsigned int iw,
            unsigned int ih,
            unsigned int w,
            unsigned int h,
            unsigned int c,
            bool m,
            bool cube,
            ArRef<MemoryBlock> d);
    ~TexData(void);
    };

  StlVector<ShapePart3D *> _parts;
  StlVector<ParticleSystem *> _particleSystems;
  StlVector<Scene3D *> _scenes;
  TexData * _texData;

  static TexData * _emptyTexData;
};

} // namespace AReVi

#endif // AREVI_LIB3D_TEXTURE_H

//----------------------------------------------------------------------------
