// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/util3D_priv.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_LIB3D_UTIL3D_PRIV_H
#define AREVI_LIB3D_UTIL3D_PRIV_H 1

#include "AReVi/Utils/stlVector_priv.h"
#include "AReVi/Utils/math_priv.h"

#include <string.h>

namespace AReVi {

//----------------------------------------------------------------------------
// Class Util3D (abstract)
//----------------------------------------------------------------------------

class Util3D
{
public:

  enum MatrixIndex { R00=0, R01=4, R02=8,   TX=12,  // COLUMN-major ordered
                     R10=1, R11=5, R12=9,   TY=13,
                     R20=2, R21=6, R22=10,  TZ=14,
                      Z0=3,  Z1=7,  Z2=11, ONE=15 };

  enum VertexLocation { MIDDLE_LOC=0,
                        XPOS_LOC=1,   XNEG_LOC=2,
                        YPOS_LOC=4,   YNEG_LOC=8,
                        ZPOS_LOC=16,  ZNEG_LOC=32 };

  struct Dbl2
  {
  double x,y;
  inline Dbl2(void); // members are not initialized !
  inline Dbl2(double inX,double inY);
  };

  struct Dbl3
  {
  double x,y,z;
  inline Dbl3(void); // members are not initialized !
  inline Dbl3(double inX,double inY,double inZ);
  };

  struct Dbl4
  {
  double x,y,z,a;
  inline Dbl4(void); // members are not initialized !
  inline Dbl4(double inX,double inY,double inZ,double inA);
  };

  struct Face
  {
  unsigned int p1,p2,p3;
  inline Face(void); // members are not initialized !
  inline Face(unsigned int inP1,unsigned int inP2,unsigned int inP3);
  };

  struct Material
  {
  inline Material(void);
  inline void setLighting(bool lighting);
  inline bool getLighting(void) const;
  inline void setShininess(double shininess);
  inline double getShininess(void) const;
  inline void setTransparency(double transparency);
  inline double getTransparency(void) const;
  inline void setAmbientIntensity(double ambientIntensity);
  inline double getAmbientIntensity(void) const;
  inline void setDiffuseColor(double r,double g,double b);
  inline void setEmissiveColor(double r,double g,double b);
  inline void setSpecularColor(double r,double g,double b);
  inline const float * getAmbientColor(void) const;
  inline const float * getDiffuseColor(void) const;
  inline const float * getEmissiveColor(void) const;
  inline const float * getSpecularColor(void) const;
  private:
    bool _lighting;
    double _shininess;
    double _transparency;
    double _ambientIntensity;
    float _ambient[4];
    float _diffuse[4];
    float _emissive[4];
    float _specular[4];
  };

  struct Base
  {
  double matrix[16]; // COLUMN-major ordered 4x4 matrix
  inline Base(bool initMatrix);
  inline bool adjustOrthonormality(void);
  inline void globalToLocalLocation(Util3D::Base & baseInOut) const;
  inline void localToGlobalLocation(Util3D::Base & baseInOut) const;
  inline void globalToLocalPosition(double & xInOut,
                                    double & yInOut,
                                    double & zInOut) const;
  inline void localToGlobalPosition(double & xInOut,
                                    double & yInOut,
                                    double & zInOut) const;
  inline void globalToLocalVector(double & vxInOut,
                                  double & vyInOut,
                                  double & vzInOut) const;
  inline void localToGlobalVector(double & vxInOut,
                                  double & vyInOut,
                                  double & vzInOut) const;
  inline void globalToLocalOrientation(double & rollInOut,
                                       double & pitchInOut,
                                       double & yawInOut) const;
  inline void localToGlobalOrientation(double & rollInOut,
                                       double & pitchInOut,
                                       double & yawInOut) const;
  inline void getLocalX(double & xOut,
                        double & yOut,
                        double & zOut) const;
  inline void getLocalY(double & xOut,
                        double & yOut,
                        double & zOut) const;
  inline void getLocalZ(double & xOut,
                        double & yOut,
                        double & zOut) const;
  inline void setMatrix(const double * mtx);
  inline void move(const Util3D::Base & base);
  inline void invert(void);
  inline void getPosition(double & pxOut,
                          double & pyOut,
                          double & pzOut) const;
  inline void setPosition(double px,
                          double py,
                          double pz);
  inline void translate(double dx,
                        double dy,
                        double dz);
  inline void extractOrientation(double & rollOut,
                                 double & pitchOut,
                                 double & yawOut) const;
  inline void extractRotation(double & xOut,
                              double & yOut,
                              double & zOut,
                              double & angleOut) const;
  inline void cancelRotation(void);
  inline void setOrientation(double roll,
                             double pitch,
                             double yaw);
  inline void setRotation(double x,
                          double y,
                          double z,
                          double angle);
  inline void rotate(double x,
                     double y,
                     double z,
                     double dAngle);
  inline void roll(double dRoll);
  inline void pitch(double dPitch);
  inline void yaw(double dYaw);
  };

  struct Transform
  {
  double matrix[16]; // COLUMN-major ordered 4x4 matrix
  inline Transform(bool initMembers);
  inline Transform(const Util3D::Base & base);
  inline void setMatrix(const double * mtx);
  inline void invert(void);
  inline void apply(double & xOut,double & yOut,double & zOut,
                    double xIn,double yIn,double zIn) const;
  inline void identity(void);
  inline bool isIdentity(bool checkCoefs) const;
  inline void preTransform(const Util3D::Transform & tr);
  inline void postTransform(const Util3D::Transform & tr);
  inline void preMove(const Util3D::Base & base);
  inline void postMove(const Util3D::Base & base);
  inline void preTranslate(double tx,double ty,double tz);
  inline void postTranslate(double tx,double ty,double tz);
  inline void preRotate(double x,double y,double z,double angle);
  inline void postRotate(double x,double y,double z,double angle);
  inline void preRoll(double roll);
  inline void postRoll(double roll);
  inline void prePitch(double pitch);
  inline void postPitch(double pitch);
  inline void preYaw(double yaw);
  inline void postYaw(double yaw);
  inline void preScale(double sx,double sy,double sz);
  inline void postScale(double sx,double sy,double sz);
  private:
    bool _identity;
  };

  struct BoundingBox
  {
  bool valid;
  double xmin,ymin,zmin;
  double xmax,ymax,zmax;
  inline BoundingBox(void); // not valid
  inline void addXYZ(double x,double y,double z);
  inline void addDbl3(const Util3D::Dbl3 & vertex);
  inline void addBox(const Util3D::BoundingBox & bbox);
  inline void addTransformedBox(const Util3D::Transform & transform,
                                const Util3D::BoundingBox & bbox);
  inline double sizeX(void) const;
  inline double sizeY(void) const;
  inline double sizeZ(void) const;
  inline double centerX(void) const;
  inline double centerY(void) const;
  inline double centerZ(void) const;
  inline Util3D::VertexLocation locateVertex(double x,
                                             double y,
                                             double z) const;
  int // 0: no collision   1: collision   -1: possible collision
  testHalfApproxCollision(const Util3D::Transform & transform,
                          const Util3D::BoundingBox & bbox) const;
  int // 0: no collision   1: collision   -1: possible collision
  testApproxCollision(const Util3D::Transform & transform,
                      const Util3D::BoundingBox & bbox) const;
  bool testCollision(const Util3D::Transform & transform,
                     const Util3D::BoundingBox & bbox) const;
  };

  struct Perspective
  {
  inline Perspective(double fieldOfView,double aspectRatio,
                     double nearDistance,double farDistance,
                     double nearYOffset,double nearZOffset);
  inline void setFieldOfView(double fieldOfView);
  inline double getFieldOfView(void) const;
  inline void setAspectRatio(double aspectRatio);
  inline double getAspectRatio(void) const;
  inline void setNearDistance(double nearDistance);
  inline double getNearDistance(void) const;
  inline void setFarDistance(double farDistance);
  inline double getFarDistance(void) const;
  inline void setNearYOffset(double nearYOffset);
  inline double getNearYOffset(void) const;
  inline void setNearZOffset(double nearZOffset);
  inline double getNearZOffset(void) const;
  inline double getLeftSlope(void) const;
  inline double getRightSlope(void) const;
  inline double getUpSlope(void) const;
  inline double getDownSlope(void) const;
  inline Util3D::VertexLocation locateVertex(double x,
                                             double y,
                                             double z) const;
  private:
    inline void _computeSlopes(void);
    double _fieldOfView;
    double _aspectRatio;
    double _nearDistance;
    double _farDistance;
    double _nearYOffset;
    double _nearZOffset;
    double _leftSlope;
    double _rightSlope;
    double _upSlope;
    double _downSlope;
  };

//-------- Intersection ------------------------------------------------------

  inline static
  bool // success
  closestPointOnTriangle(const Util3D::Dbl3 & p1,
                         const Util3D::Dbl3 & p2,
                         const Util3D::Dbl3 & p3,
                         const Util3D::Dbl3 & from,
                         Util3D::Dbl3 & closestOut,
                         double & sqrDistOut);

  inline static
  double // (interP-rayP)/dirV [0.0;+inf[
  rayTriangleIntersection(const Util3D::Dbl3 & p1,
                          const Util3D::Dbl3 & p2,
                          const Util3D::Dbl3 & p3,
                          const Util3D::Dbl3 & rayP,
                          const Util3D::Dbl3 & dirV);

  inline static
  double // (interP-rayP)/dirV [0.0;+inf[
  rayTrianglesIntersection(const Util3D::Dbl3 * vertices,
                           const Util3D::Face * triangles,
                           unsigned int nbTriangles,
                           const Util3D::Dbl3 & rayP,
                           const Util3D::Dbl3 & dirV,
                           bool findNearest,
                           unsigned int & triIndexOut);

  inline static
  double // (Pinter-Pray)/Vdir [0.0;+inf[
  rayBoxIntersection(const Util3D::BoundingBox & bbox,
                     const Util3D::Dbl3 & rayP,
                     const Util3D::Dbl3 & dirV,
                     bool findNearest);

  inline static
  bool // contact
  triangleTriangleContact(const Util3D::Dbl3 & ta1,
                          const Util3D::Dbl3 & ta2,
                          const Util3D::Dbl3 & ta3,
                          const Util3D::Dbl3 & tb1,
                          const Util3D::Dbl3 & tb2,
                          const Util3D::Dbl3 & tb3);

  struct TriInter
  {
  Util3D::Dbl3 ta1,ta2,ta3;
  Util3D::Dbl3 tb1,tb2,tb3;
  Util3D::Dbl3 i1,i2;
  };

  inline static
  bool // intersection
  triangleTriangleIntersection(const Util3D::Dbl3 & ta1,
                               const Util3D::Dbl3 & ta2,
                               const Util3D::Dbl3 & ta3,
                               const Util3D::Dbl3 & tb1,
                               const Util3D::Dbl3 & tb2,
                               const Util3D::Dbl3 & tb3,
                               Util3D::Dbl3 & i1Out,
                               Util3D::Dbl3 & i2Out);

  inline static
  bool // contact
  obbObbContact(double ax, double ay, double az,
		const double * M,
		double bx, double by, double bz,
		const double * N);

  inline static
  bool // contact
  ellipsoidEllipsoidContact(double ax, double ay, double az,
                            const double * M,
                            double bx, double by, double bz,
                            const double * N);

//-------- Triangulation -----------------------------------------------------

  static
  void
  triangulateXY(StlVector<Util3D::Dbl3> & vertexInOut,
                StlVector<Util3D::Face> & triangleOut,
                double precision);

//-------- Static attributes -------------------------------------------------
private:

    static double _idMtx[16];
};

} // namespace AReVi

//-------- Inline source code ------------------------------------------------

#include "AReVi/Lib3D/util3D_priv.Ci"

#endif // AREVI_LIB3D_UTIL3D_PRIV_H

//----------------------------------------------------------------------------
