// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/shape3D.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_SHAPES_SHAPE3D_H
#define AREVI_SHAPES_SHAPE3D_H 1

#include "AReVi/arObject.h"

#include "AReVi/Lib3D/util3D_priv.h"

namespace AReVi {

class Object3D;
class Renderer3D;
class ShapePart3D;
class Transform3D;
class BoundingBox3D;

//----------------------------------------------------------------------------
// Class Shape3D
//----------------------------------------------------------------------------

class Shape3D : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Shape3D)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR(Shape3D)

//-------- Geometry ----------------------------------------------------------

  virtual
  void
  writeTransformation(ArConstRef<Transform3D> transform);

  virtual
  void
  readTransformation(ArRef<Transform3D> transformOut) const;

  virtual
  void
  readBoundingBox(ArRef<BoundingBox3D> bboxOut) const;

  virtual
  unsigned int // number of triangles
  getNbTriangles(void) const;

  virtual
  unsigned int // number of lines
  getNbLines(void) const;

  virtual
  unsigned int // number of points
  getNbPoints(void) const;

  virtual
  bool // currently displayed by a renderer
  isCurrentlyDisplayed(void) const;

//-------- Detection ----------------------------------------------------------

  virtual
  void
  setTriangleDivision(unsigned int maxTriangles);

  virtual
  unsigned int // max triangles in divisions
  getTriangleDivision(void) const;

//-------- Billboard ----------------------------------------------------------

  virtual
  void
  setBillboard(bool isBillboard,
               bool aroundZ);

  virtual
  void
  getBillboard(bool & isBillboardOut,
               bool & aroundZOut) const;

//-------- Part management ---------------------------------------------------

  virtual
  void
  addRootPart(ArRef<ShapePart3D> part);

  virtual
  void
  removeRootPart(ArRef<ShapePart3D> part);

  virtual
  unsigned int // number of root parts
  getNbRootParts(void) const;

  virtual
  ArConstRef<ShapePart3D> // index'th root part
  getRootPart(unsigned int index) const;

  virtual
  ArRef<ShapePart3D> // index'th root part
  accessRootPart(unsigned int index);

  virtual
  unsigned int // number of parts/subparts
  getNbParts(void) const;

  virtual
  ArConstRef<ShapePart3D> // index'th part/subpart
  getPart(unsigned int index) const;

  virtual
  ArRef<ShapePart3D> // index'th part/subpart
  accessPart(unsigned int index);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(Shape3D,Transformation,_transformationCB,
              /* no field */)

  // add/remove root nodes, billboard
  AR_CALLBACK(Shape3D,StructureChange,_structureChangeCB,
              /* no field */)

//-------- Data access (internal) --------------------------------------------

  AR_UNSTUBBED inline virtual
  const Util3D::Transform &
  getTransformation(void) const;

  AR_UNSTUBBED inline virtual
  const Util3D::BoundingBox &
  getBoundingBox(void) const;

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class Object3D;
  friend class Renderer3D;
  friend class Scene3D;
  friend class ShapePart3D;

  virtual
  void
  _addFlatPart(ShapePart3D * part);

  virtual
  void
  _removeFlatPart(ShapePart3D * part);

  virtual
  void
  _postRedisplay();

  virtual
  void
  _addObject(Object3D * object);

  virtual
  void
  _removeObject(Object3D * object);

  virtual
  void
  _askVisibleParts(Renderer3D * renderer,
                   const Object3D * object,
                   bool selected) const;

  virtual
  void
  _findBillboardLocation(Util3D::Base & locationOut,
                                const Util3D::Base & location,
                                const Util3D::Base & viewPoint) const;

  virtual
  bool // bbox is visible in renderer
  _boundingBoxVisibility(const Renderer3D * renderer,
                         const Util3D::Base & base,
                         const Util3D::BoundingBox & bbox,
                         const Util3D::Transform * transform) const;

  virtual
  double // (Pinter-Pray)/(Pdir-Pray) [0.0;+inf[
  _intersectRay(unsigned int level,
                bool findNearest,
                const Util3D::Dbl3 & rayP,
                const Util3D::Dbl3 & dirV,
		ArRef<ShapePart3D> & partOut,
                unsigned int & triIndexOut) const;

  static
  bool // collision occured
  _testCollision(const Shape3D * shape1,
                 const Shape3D * shape2,
                 const Util3D::Base & base1,
                 const Util3D::Base & base2,
                 unsigned int level);

  static
  void
  _intersect(const Shape3D * shape1,
             const Shape3D * shape2,
             const Util3D::Base & base1,
             const Util3D::Base & base2,
             StlVector<Util3D::TriInter> & itOut);

  virtual
  void
  _postRebuild(void);

  virtual
  void
  _rebuildIfNeeded(void);

  virtual
  void
  _onTransformation(void);

  virtual
  void
  _onStructureChange(void);

//-------- Member attributes -------------------------------------------------
protected:

  StlVector<Object3D *> _objects;
  StlVector<ShapePart3D *> _flatParts;
  Util3D::Transform _transform;
  Util3D::BoundingBox _bbox;
  unsigned int _nbTriangles;
  unsigned int _nbLines;
  unsigned int _nbPoints;
  unsigned int _triangleDivision;
  bool _billboard;
  bool _billboardAroundZ;
  unsigned int _frameID;
  StlVector<ArRef<ShapePart3D> > _rootParts;
  bool _needsGeometryUpdate;
  bool _needsRebuild;
  CallbackManager<Shape3D,TransformationEvent> _transformationCB;
  CallbackManager<Shape3D,StructureChangeEvent> _structureChangeCB;
};

//-------- Inline code -------------------------------------------------------

inline
const Util3D::Transform &
Shape3D::getTransformation(void) const
{
return(_transform);
}

inline
const Util3D::BoundingBox &
Shape3D::getBoundingBox(void) const
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
return(_bbox);
}

} // namespace AReVi

#endif // AREVI_SHAPES_SHAPE3D_H

//----------------------------------------------------------------------------
