// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/shapePart3D.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_SHAPES_SHAPEPART3D_H
#define AREVI_SHAPES_SHAPEPART3D_H 1

#include "AReVi/arObject.h"

#include "AReVi/Lib3D/util3D_priv.h"

namespace AReVi {

class Transform3D;
class BoundingBox3D;
class Object3D;
class Shape3D;
class ShapePartRenderer;
class Renderer3D;
class Light3D;

//----------------------------------------------------------------------------
// Class ShapePart3D (abstract)
//----------------------------------------------------------------------------

class ShapePart3D : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(ShapePart3D)

//-------- Geometry ----------------------------------------------------------

  virtual
  void
  writeTransformation(ArConstRef<Transform3D> transform);

  virtual
  void
  readTransformation(ArRef<Transform3D> transformOut) const;

  virtual
  void
  readFlatTransformation(ArRef<Transform3D> transformOut) const;

  virtual
  void
  readBoundingBox(ArRef<BoundingBox3D> bboxOut) const;

//-------- Appearance --------------------------------------------------------

  virtual
  void
  setDepthOffset(double offset);

  virtual
  double // depth buffer offset
  getDepthOffset(void) const;

  virtual
  void
  setRenderer(ArRef<ShapePartRenderer> renderer);

  virtual
  ArConstRef<ShapePartRenderer>
  getRenderer(void) const;

  virtual
  ArRef<ShapePartRenderer>
  accessRenderer(void);

//-------- Display list usage ------------------------------------------------

  virtual
  void
  setDisplayListUsage(bool useDisplayList);

  virtual
  bool // display list usage is enabled
  getDisplayListUsage(void) const;

//-------- Tree structure ----------------------------------------------------

  virtual
  void
  addSubPart(ArRef<ShapePart3D> part);

  virtual
  void
  removeSubPart(ArRef<ShapePart3D> part);

  virtual
  unsigned int // number of sub-parts
  getNbSubParts(void) const;

  virtual
  ArConstRef<ShapePart3D> // index'th sub-part
  getSubPart(unsigned int index) const;

  virtual
  ArRef<ShapePart3D> // index'th sub-part
  accessSubPart(unsigned int index);

  virtual
  ArConstRef<ShapePart3D> // parent part
  getParentPart(void) const;

  virtual
  ArRef<ShapePart3D> // parent part
  accessParentPart(void);

  virtual
  ArConstRef<Shape3D> // owner shape
  getShape(void) const;

  virtual
  ArRef<Shape3D> // owner shape
  accessShape(void);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(ShapePart3D,Transformation,_transformationCB,
              /* no field */)

  // add/remove sub nodes, depth offset
  AR_CALLBACK(ShapePart3D,StructureChange,_structureChangeCB,
              /* no field */)

  AR_CALLBACK(ShapePart3D,AppearanceRebuild,_appearanceRebuildCB,
              /* no field */)

  AR_CALLBACK(ShapePart3D,GeometryRebuild,_geometryRebuildCB,
              /* no field */)

//-------- Data access (internal) --------------------------------------------

  AR_UNSTUBBED inline virtual
  const Util3D::Transform &
  getTransformation(void) const;

  AR_UNSTUBBED inline virtual
  const Util3D::Transform &
  getFlatTransformation(void) const;

  AR_UNSTUBBED inline virtual
  const Util3D::BoundingBox &
  getBoundingBox(void) const;

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class Shape3D;
  friend class Renderer3D;
  friend class Texture;
  friend class ShapePartRenderer;

  ShapePart3D(ArCW & arCW);

  virtual
  void
  _postAppearanceRebuild(void);

  virtual
  void
  _postGeometryRebuild(void);

  virtual
  void
  _postRedisplayIfVisible(void);

  virtual
  void
  _updateTransformation(void);

  virtual
  void
  _updateShape(ArPtr<Shape3D> shape);

  virtual
  void
  _setAppearance(void) const=0;

  virtual
  void
  _resetAppearance(void) const=0;

  virtual
  void
  _drawGeometry(void) const=0;

  virtual
  void
  _rebuildIfNeeded(void)=0;

  virtual
  void
  _render(void) const;

  virtual
  void
  _renderWithoutAppearance(void) const;

  virtual
  void
  _describeGeometry(const Util3D::BoundingBox & bbox,
                    const Util3D::Dbl3 * vertices,
                    unsigned int nbVertices,
                    const Util3D::Face * triangles,
                    unsigned int nbTriangles,
                    unsigned int nbLines,
                    unsigned int nbPoints);

  virtual
  void
  _setTriangleDivision(unsigned int maxTriangles);

  virtual
  double // (interP-rayP)/dirV [0.0;+inf[
  _rayIntersection(const Util3D::Dbl3 & rayP,
                   const Util3D::Dbl3 & dirV,
                   bool findNearest,
                   unsigned int & triIndexOut) const;

  virtual
  bool // collision occured
  _testCollision(const Util3D::Transform & transform,
                 const ShapePart3D * part) const;

  virtual
  void
  _intersect(const Util3D::Transform & transform,
             const ShapePart3D * part,
             StlVector<Util3D::TriInter> & itOut) const;

  virtual
  void
  _buildBoxTree(void);

  virtual
  void
  _onTransformation(void);

  virtual
  void
  _onStructureChange(void);

//-------- Member attributes -------------------------------------------------
protected:

  struct BoxTree
  {
  ShapePart3D * part;
  Util3D::BoundingBox bbox;
  StlVector<Util3D::Face> tri;
  BoxTree * sub[2];
  BoxTree(ShapePart3D * owner);
  ~BoxTree(void);
  void divide(unsigned int maxTriangles);
  double // (interP-rayP)/dirV [0.0;+inf[
  rayIntersection(const Util3D::Dbl3 & rayP,
                  const Util3D::Dbl3 & dirV,
                  bool findNearest,
                  unsigned int & triIndexOut) const;
  bool // success
  testBoxBox(const Util3D::Transform & transform,
             const ShapePart3D::BoxTree * tree,
             Util3D::Dbl3 * & trVertices,
             bool * & trFlags) const;
  bool // success
  testTriBox(const Util3D::Transform & transform,
             const ShapePart3D::BoxTree * tree,
             Util3D::Dbl3 * trVertices,
             bool * trFlags) const;
  bool // success
  testBoxTri(const Util3D::Transform & transform,
             const ShapePart3D::BoxTree * tree,
             Util3D::Dbl3 * trVertices,
             bool * trFlags) const;
  bool // success
  testTriTri(const Util3D::Transform & transform,
             const ShapePart3D::BoxTree * tree,
             Util3D::Dbl3 * trVertices,
             bool * trFlags) const;
  void
  interBoxBox(const Util3D::Transform & transform,
              const ShapePart3D::BoxTree * tree,
              Util3D::Dbl3 * & trVertices,
              bool * & trFlags,
              StlVector<Util3D::TriInter> & itOut) const;
  void
  interTriBox(const Util3D::Transform & transform,
              const ShapePart3D::BoxTree * tree,
              Util3D::Dbl3 * trVertices,
              bool * trFlags,
              StlVector<Util3D::TriInter> & itOut) const;
  void
  interBoxTri(const Util3D::Transform & transform,
              const ShapePart3D::BoxTree * tree,
              Util3D::Dbl3 * trVertices,
              bool * trFlags,
              StlVector<Util3D::TriInter> & itOut) const;
  void
  interTriTri(const Util3D::Transform & transform,
              const ShapePart3D::BoxTree * tree,
              Util3D::Dbl3 * trVertices,
              bool * trFlags,
              StlVector<Util3D::TriInter> & itOut) const;
  };

  Util3D::BoundingBox _bbox;
  const Util3D::Dbl3 * _vPtr;
  unsigned int _nbVertices;
  const Util3D::Face * _tPtr;
  unsigned int _nbTriangles;
  unsigned int _nbLines;
  unsigned int _nbPoints;
  unsigned int _triangleDivision;
  ShapePart3D::BoxTree * _tree;
  bool _useDisplayList;
  int _appearanceGeometryList;
  int _geometryList;
  bool _needsBlending;
  bool _needsAlphaTest;
  Util3D::Transform _localTransform;
  Util3D::Transform _flatTransform;
  double _offset;
  ArRef<ShapePartRenderer> _renderer;
  bool _needsAppearanceRebuild;
  bool _needsGeometryRebuild;
  StlVector<ArRef<ShapePart3D> > _subParts;
  ArPtr<ShapePart3D> _parentPart;
  ArPtr<Shape3D> _shape;
  CallbackManager<ShapePart3D,TransformationEvent> _transformationCB;
  CallbackManager<ShapePart3D,StructureChangeEvent> _structureChangeCB;
  CallbackManager<ShapePart3D,AppearanceRebuildEvent> _appearanceRebuildCB;
  CallbackManager<ShapePart3D,GeometryRebuildEvent> _geometryRebuildCB;
};

//-------- Inline code -------------------------------------------------------

inline
const Util3D::Transform &
ShapePart3D::getTransformation(void) const
{
return(_localTransform);
}

inline
const Util3D::Transform &
ShapePart3D::getFlatTransformation(void) const
{
return(_flatTransform);
}

inline
const Util3D::BoundingBox &
ShapePart3D::getBoundingBox(void) const
{
const_cast<ShapePart3D *>(this)->_rebuildIfNeeded();
return(_bbox);
}


//----------------------------------------------------------------------------
// Class ShapePartRenderer (abstract)
//----------------------------------------------------------------------------

class ShapePartRenderer : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(ShapePartRenderer)

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class ShapePart3D;
  friend class Renderer3D;

  ShapePartRenderer(ArCW & arCW);

  virtual
  void
  _render(const ShapePart3D * part,
          const Object3D * object,
          const Renderer3D * renderer) const=0;

  virtual
  void
  _renderWithoutAppearance(const ShapePart3D * part,
                           const Object3D * object,
                           const Renderer3D * renderer) const=0;

  static
  const Util3D::Base &
  _getLookAt(const Renderer3D * renderer);

  static
  unsigned int
  _getNbActiveLights(const Renderer3D * renderer);

  static
  void
  _renderLight(unsigned int lightNumber,
               const Light3D * light);

  static
  void
  _renderHeadLight(unsigned int lightNumber,
                   const Renderer3D * renderer);

  static
  void
  _renderPart(const ShapePart3D * part);

  static
  void
  _renderPartWithoutAppearance(const ShapePart3D * part);

  virtual
  void
  _addShapePart(ShapePart3D * part);

  virtual
  void
  _removeShapePart(ShapePart3D * part);

  virtual
  void
  _postRedisplay(void);

//-------- Member attributes -------------------------------------------------
protected :

  StlVector<ShapePart3D *> _parts;
};

} // namespace AReVi

#endif // AREVI_SHAPES_SHAPEPART3D_H

//----------------------------------------------------------------------------
