// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/tcpUdp.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_UTILS_TCPUDP_H
#define AREVI_UTILS_TCPUDP_H 1

#include "AReVi/arObject.h"

namespace AReVi {

class MemoryBlock;

//----------------------------------------------------------------------------
// Class IPSocket (abstract)
//----------------------------------------------------------------------------

class IPSocket : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(IPSocket)

//-------- Inspection --------------------------------------------------------

  virtual
  unsigned int // local ip address
  getIPAddress(void) const;

  virtual
  int // local port number
  getPortNumber(void) const;

  virtual
  int // file descriptor
  getSocketFD(void) const;

//-------- Blocking / Non-blocking -------------------------------------------

  virtual
  void
  setBlocking(bool blocking);

  virtual
  bool // blocks until data arrives
  isBlocking(void) const;

//-------- Forbidden operations ----------------------------------------------
protected:

  virtual
  void
  _cancel(void);

  IPSocket(ArCW & arCW);

//-------- Member attributes -------------------------------------------------
protected:

  int _socket;
  bool _blocking;
  unsigned int _ipAddress;
  int _portNumber;
};


//----------------------------------------------------------------------------
// Class TCPConnection
//----------------------------------------------------------------------------

class TCPConnection : public IPSocket
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(TCPConnection)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_3(TCPConnection,
                   unsigned int, distantIPAddress,
                   int, distantPortNumber,
                   bool, packetOriented)

//-------- Inspection --------------------------------------------------------

  virtual
  unsigned int // distant ip address
  getDistantIPAddress(void) const;

  virtual
  int // distant port number
  getDistantPortNumber(void) const;

  virtual
  bool // packets size transmitted
  isPacketOriented(void) const;

//-------- Binary communication ----------------------------------------------

  virtual
  bool // bytes available
  readBytes(ArRef<MemoryBlock> memory);

  virtual
  void
  writeBytes(ArConstRef<MemoryBlock> memory,
             unsigned int offset,
             unsigned int length);

  virtual
  void
  postBytes(ArConstRef<MemoryBlock> memory,
            unsigned int offset,
            unsigned int length,
            bool copyMemoryIfDelayed);

//-------- Textual communication ---------------------------------------------

  virtual
  bool // string available
  readString(StlString & result);

  virtual
  void
  writeString(const StlString & str);

  virtual
  void
  postString(const StlString & str);

//-------- Writing queue -----------------------------------------------------

  virtual
  void
  pushPost(void);

  virtual
  bool // posted data should be pushed
  pendingPost(void) const;

//-------- Forbidden operations ----------------------------------------------
protected:

  virtual
  int // -1: error  0: nothing  1: ok
  _extractPacket(void);

  struct PendingData
  {
  ArConstRef<MemoryBlock> memory;
  unsigned int offset;
  unsigned int length;
  unsigned int lengthRemaining;
  inline
  PendingData(ArConstRef<MemoryBlock> m,unsigned int o,
              unsigned int l,unsigned int lr)
    : memory(m), offset(o), length(l), lengthRemaining(lr) {}
  };

  virtual
  int // -1: error  0: non-blocking  1: fully written
  _partialWrite(PendingData & data);

  friend class TCPListener;

  TCPConnection(ArCW & arCW,
                int socket,
                unsigned int distantIPAddress,
                int distantPortNumber,
                bool packetOriented);

//-------- Member attributes -------------------------------------------------
protected:

  unsigned int _distantIPAddress;
  int _distantPortNumber;
  bool _packetOriented;
  unsigned int _packetSize;
  unsigned int _nbExtracted;
  ArRef<MemoryBlock> _memory;
  StlList<PendingData> _pendingData;
};


//----------------------------------------------------------------------------
// Class TCPListener
//----------------------------------------------------------------------------

class TCPListener : public IPSocket
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(TCPListener)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_2(TCPListener,
                   int, portNumber,
                   bool, packetOriented)

//-------- Inspection --------------------------------------------------------

  virtual
  bool // creates packet oriented connections
  isPacketOriented(void) const;

//-------- Connection control ------------------------------------------------

  virtual
  ArRef<TCPConnection> // distant connection
  nextConnection(void);

//-------- Member attributes -------------------------------------------------
protected:

  bool _packetOriented;
};


//----------------------------------------------------------------------------
// Class UDPTransmitter
//----------------------------------------------------------------------------

class UDPTransmitter : public IPSocket
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(UDPTransmitter)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(UDPTransmitter,
                   int, portNumber)

//-------- Binary communication ----------------------------------------------

  virtual
  bool // bytes available
  receiveBytes(ArRef<MemoryBlock> memory,
               unsigned int & distantIPAddressOut,
               int & distantPortNumberOut);

  virtual
  void
  sendBytes(ArConstRef<MemoryBlock> memory,
            unsigned int offset,
            unsigned int length,
            unsigned int distantIPAddress,
            int distantPortNumber);

  virtual
  void
  broadcastBytes(ArConstRef<MemoryBlock> memory,
                 unsigned int offset,
                 unsigned int length,
                 int distantPortNumber);

//-------- Textual communication ---------------------------------------------

  virtual
  bool // string available
  receiveString(StlString & resultOut,
                unsigned int & distantIPAddressOut,
                int & distantPortNumberOut);

  virtual
  void
  sendString(const StlString & str,
             unsigned int distantIPAddress,
             int distantPortNumber);

  virtual
  void
  broadcastString(const StlString & str,
                  int distantPortNumber);

//-------- Member attributes -------------------------------------------------
protected:

  ArRef<MemoryBlock> _memory;
};

} // namespace AReVi

#endif // AREVI_UTILS_TCPUDP_H

//---------------------------------------------------------------------------
