// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/arObject.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_AROBJECT_H
#define AREVI_AROBJECT_H 1

#include "AReVi/Utils/platform_priv.h"
#include "AReVi/Utils/math_priv.h"
#include "AReVi/Utils/allocator_priv.h"
#include "AReVi/Utils/stlString_priv.h"
#include "AReVi/Utils/stlVector_priv.h"
#include "AReVi/Utils/stlDeque_priv.h"
#include "AReVi/Utils/stlList_priv.h"
#include "AReVi/Utils/stlSet_priv.h"
#include "AReVi/Utils/stlMap_priv.h"

#include <iostream>
#include <stdio.h>
#include <assert.h>

namespace AReVi {

#define AR_UNSTUBBED /* useless in C++, only for stub generation */

class ArObject;
class ArClass;
class Message;
class AbstractIStream;
class AbstractOStream;
template <class T> class ArRef;
template <class T> class ArConstRef;
template <class T> class ArPtr;
template <class T> class ArConstPtr;

//----------------------------------------------------------------------------
// ArRef/ArConstRef/ArPtr/ArConstPtr type conversions
//----------------------------------------------------------------------------

template <class T1,class T2> inline
ArRef<T1>
ar_down_cast(ArRef<T2> ref);

template <class T1,class T2> inline
ArPtr<T1>
ar_down_cast(ArPtr<T2> ptr);

template <class T1,class T2> inline
ArConstRef<T1>
ar_down_cast(ArConstRef<T2> ref);

template <class T1,class T2> inline
ArConstPtr<T1>
ar_down_cast(ArConstPtr<T2> ptr);

template <class T> inline
ArRef<T>
ar_const_cast(ArConstRef<T> ref);

template <class T> inline
ArPtr<T>
ar_const_cast(ArConstPtr<T> ptr);

template <class T,class T1,class T2> inline
void
ar_down_cast_append(StlVector<T1> & targetInOut,
                    const StlVector<T2> & source);

template <class T1,class T2> inline
void
ar_const_cast_append(StlVector<T1> & targetInOut,
                     const StlVector<T2> & source);

//----------------------------------------------------------------------------
// Class ArRef
//----------------------------------------------------------------------------

// The destructor of ``ArRef'' is not virtual ! (to preserve memory usage)
//   --> DO NOT inherit from ``ArRef''.

template <class T=ArObject>
class ArRef
{
//-------- Construction / Destruction ----------------------------------------
public:

  inline
  ArRef(void);

  inline
  ArRef(const ArRef<T> & ref);

  template <class T1> inline
  ArRef(const ArRef<T1> & ref);

  template <class T1> inline
  ArRef(const ArPtr<T1> & ptr);

  inline explicit
  ArRef(T * ref);

  inline
  ~ArRef(void); // NOT virtual !

//-------- Affectation -------------------------------------------------------

  inline
  ArRef<T> & // itself
  operator=(const ArRef<T> & ref);

  template <class T1> inline
  ArRef<T> & // itself
  operator=(const ArRef<T1> & ref);

  template <class T1> inline
  ArRef<T> & // itself
  operator=(const ArPtr<T1> & ptr);

//-------- Comparison --------------------------------------------------------

  inline
  bool // same object
  operator==(const ArRef<T> & ref) const;

  inline
  bool // different
  operator!=(const ArRef<T> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstPtr<T1> & ptr) const;

//-------- Operators ---------------------------------------------------------

  inline
  void
  clear(void);

  inline
  void
  destroy(void) const;

  inline
  bool // valid reference (object is not under destruction)
  valid(void) const;

  inline
  operator const void *(void) const; // non null reference (bool)

  inline
  bool // null reference
  null(void) const;

  inline
  bool // null reference
  operator!(void) const;

  inline
  T * // object
  operator->(void) const;

  inline
  T & // object
  operator*(void) const;

  inline
  T * // object
  c_ptr(void) const;

//-------- Member attributes -------------------------------------------------
private:
  T * _ref;
};

template <class T> inline
ostream & // output stream
operator<<(ostream & output,
           const ArRef<T> & ref);


//----------------------------------------------------------------------------
// Class ArConstRef
//----------------------------------------------------------------------------

// The destructor of ``ArConstRef'' is not virtual ! (to preserve memory usage)
//   --> DO NOT inherit from ``ArConstRef''.

template <class T=ArObject>
class ArConstRef
{
//-------- Construction / Destruction ----------------------------------------
public:

  inline
  ArConstRef(void);

  inline
  ArConstRef(const ArConstRef<T> & ref);

  template <class T1> inline
  ArConstRef(const ArConstRef<T1> & ref);

  template <class T1> inline
  ArConstRef(const ArRef<T1> & ref);

  template <class T1> inline
  ArConstRef(const ArConstPtr<T1> & ptr);

  template <class T1> inline
  ArConstRef(const ArPtr<T1> & ptr);

  inline explicit
  ArConstRef(const T * ref);

  inline
  ~ArConstRef(void); // NOT virtual !

//-------- Affectation -------------------------------------------------------

  inline
  ArConstRef<T> & // itself
  operator=(const ArConstRef<T> & ref);

  template <class T1> inline
  ArConstRef<T> & // itself
  operator=(const ArConstRef<T1> & ref);

  template <class T1> inline
  ArConstRef<T> & // itself
  operator=(const ArRef<T1> & ref);

  template <class T1> inline
  ArConstRef<T> & // itself
  operator=(const ArConstPtr<T1> & ptr);

  template <class T1> inline
  ArConstRef<T> & // itself
  operator=(const ArPtr<T1> & ptr);

//-------- Comparison --------------------------------------------------------

  inline
  bool // same object
  operator==(const ArConstRef<T> & ref) const;

  inline
  bool // different
  operator!=(const ArConstRef<T> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstPtr<T1> & ptr) const;

//-------- Operators ---------------------------------------------------------

  inline
  void
  clear(void);

  inline
  bool // valid reference (object is not under destruction)
  valid(void) const;

  inline
  operator const void *(void) const; // non null reference (bool)

  inline
  bool // null reference
  null(void) const;

  inline
  bool // null reference
  operator!(void) const;

  inline
  const T * // object
  operator->(void) const;

  inline
  const T & // object
  operator*(void) const;

  inline
  const T * // object
  c_ptr(void) const;

//-------- Member attributes -------------------------------------------------
private:
  const T * _ref;
};

template <class T> inline
ostream & // output stream
operator<<(ostream & output,
           const ArConstRef<T> & ref);


//----------------------------------------------------------------------------
// Class ArPtr
//----------------------------------------------------------------------------

// The destructor of ``ArPtr'' is not virtual ! (to preserve memory usage)
//   --> DO NOT inherit from ``ArPtr''.

template <class T=ArObject>
class ArPtr
{
//-------- Construction / Destruction ----------------------------------------
public:

  inline
  ArPtr(void);

  inline
  ArPtr(const ArPtr<T> & ptr);

  template <class T1> inline
  ArPtr(const ArPtr<T1> & ptr);

  template <class T1> inline
  ArPtr(const ArRef<T1> & ref);

  inline explicit
  ArPtr(T * ptr);

  inline
  ~ArPtr(void); // NOT virtual !

//-------- Affectation -------------------------------------------------------

  inline
  ArPtr<T> & // itself
  operator=(const ArPtr<T> & ptr);

  template <class T1> inline
  ArPtr<T> & // itself
  operator=(const ArPtr<T1> & ptr);

  template <class T1> inline
  ArPtr<T> & // itself
  operator=(const ArRef<T1> & ref);

//-------- Comparison --------------------------------------------------------

  inline
  bool // same object
  operator==(const ArPtr<T> & ptr) const;

  inline
  bool // different
  operator!=(const ArPtr<T> & ptr) const;

  template <class T1> inline
  bool // same object
  operator==(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstPtr<T1> & ptr) const;

//-------- Operators ---------------------------------------------------------

  inline
  void
  clear(void);

  inline
  void
  destroy(void) const;

  inline
  bool // valid pointer (object is not under destruction)
  valid(void) const;

  inline
  operator const void *(void) const; // non null pointer (bool)

  inline
  bool // null pointer
  null(void) const;

  inline
  bool // null pointer
  operator!(void) const;

  inline
  T * // object
  operator->(void) const;

  inline
  T & // object
  operator*(void) const;

  inline
  T * // object
  c_ptr(void) const;

//-------- Member attributes -------------------------------------------------
private:
  T * _ptr;
};

template <class T> inline
ostream & // output stream
operator<<(ostream & output,
           const ArPtr<T> & ptr);


//----------------------------------------------------------------------------
// Class ArConstPtr
//----------------------------------------------------------------------------

// The destructor of ``ArConstPtr'' is not virtual ! (to preserve memory usage)
//   --> DO NOT inherit from ``ArConstPtr''.

template <class T=ArObject>
class ArConstPtr
{
//-------- Construction / Destruction ----------------------------------------
public:

  inline
  ArConstPtr(void);

  inline
  ArConstPtr(const ArConstPtr<T> & ptr);

  template <class T1> inline
  ArConstPtr(const ArConstPtr<T1> & ptr);

  template <class T1> inline
  ArConstPtr(const ArPtr<T1> & ptr);

  template <class T1> inline
  ArConstPtr(const ArConstRef<T1> & ref);

  template <class T1> inline
  ArConstPtr(const ArRef<T1> & ref);

  inline explicit
  ArConstPtr(const T * ptr);

  inline
  ~ArConstPtr(void); // NOT virtual !

//-------- Affectation -------------------------------------------------------

  inline
  ArConstPtr<T> & // itself
  operator=(const ArConstPtr<T> & ptr);

  template <class T1> inline
  ArConstPtr<T> & // itself
  operator=(const ArConstPtr<T1> & ptr);

  template <class T1> inline
  ArConstPtr<T> & // itself
  operator=(const ArPtr<T1> & ptr);

  template <class T1> inline
  ArConstPtr<T> & // itself
  operator=(const ArConstRef<T1> & ref);

  template <class T1> inline
  ArConstPtr<T> & // itself
  operator=(const ArRef<T1> & ref);

//-------- Comparison --------------------------------------------------------

  inline
  bool // same object
  operator==(const ArConstPtr<T> & ptr) const;

  inline
  bool // different
  operator!=(const ArConstPtr<T> & ptr) const;

  template <class T1> inline
  bool // same object
  operator==(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstRef<T1> & ref) const;

  template <class T1> inline
  bool // same object
  operator==(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArPtr<T1> & ptr) const;

  template <class T1> inline
  bool // same object
  operator==(const ArConstPtr<T1> & ptr) const;

  template <class T1> inline
  bool // different
  operator!=(const ArConstPtr<T1> & ptr) const;

//-------- Operators ---------------------------------------------------------

  inline
  void
  clear(void);

  inline
  bool // valid pointer (object is not under destruction)
  valid(void) const;

  inline
  operator const void *(void) const; // non null pointer (bool)

  inline
  bool // null pointer
  null(void) const;

  inline
  bool // null pointer
  operator!(void) const;

  inline
  const T * // object
  operator->(void) const;

  inline
  const T & // object
  operator*(void) const;

  inline
  const T * // object
  c_ptr(void) const;

//-------- Member attributes -------------------------------------------------
private:
  const T * _ptr;
};

template <class T> inline
ostream & // output stream
operator<<(ostream & output,
           const ArConstPtr<T> & ptr);

} // namespace AReVi

#include "AReVi/Utils/callback_priv.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class SerializationDependencies
//----------------------------------------------------------------------------

class SerializationDependencies
{
//-------- Construction / Destruction ----------------------------------------
public:

  inline
  SerializationDependencies(void);

  inline
  ~SerializationDependencies(void); // NOT virtual !

//-------- Stream operations -------------------------------------------------

  void
  writeToStream(ArRef<AbstractOStream> output) const;

  bool // success
  readFromStream(ArRef<AbstractIStream> input);


//-------- Operators ---------------------------------------------------------
  inline
  void
  clear(void);

//-------- Reference(s) management -------------------------------------------

  inline
  unsigned int // reference serialization id
  addReference(ArRef<ArObject> ref); // NOT virtual !

  inline
  unsigned int // reference serialization id
  addConstReference(ArConstRef<ArObject> ref); // NOT virtual !

  inline
  ArRef<ArObject> // ref on success, nullRef on failure
  findReference(unsigned int refId) const; // NOT virtual !

  inline
  const StlVector<ArRef<ArObject> > &
  getReferences(void) const; // NOT virtual !

//-------- URL(s) management -------------------------------------------------

  inline
  unsigned int // url serialization id
  addURL(const StlString & url); // NOT virtual !

  inline
  StlString // url on success, "" on failure
  findURL(unsigned int urlId) const; // NOT virtual !

  inline
  const StlVector<StlString > &
  getURLs(void) const; // NOT virtual !

//-------- Forbidden operations ----------------------------------------------
private:

  SerializationDependencies(const SerializationDependencies & obj);
                                                          // no implementation

  SerializationDependencies & // itself
  operator=(const SerializationDependencies & obj); // no implementation

  void *
  operator new(size_t size); // no implementation

  void
  operator delete(void * /* address */,
                  size_t /* size */) {} // unused but mandatory !

//-------- Member attributes -------------------------------------------------
private:

  StlVector<ArRef<ArObject> > _references;
  StlVector<StlString > _urls;
};

typedef ArClass * ArCW;

//----------------------------------------------------------------------------
// Class ArObject
//----------------------------------------------------------------------------

// Several methods in class ``ArObject'' are NOT virtual to preserve the
// garbage collection control.
//   --> DO NOT try to provide a surdefinition for these methods.

class ArObject
{
friend class ArClass;
template <class T> friend class ArRef;
template <class T> friend class ArConstRef;
template <class T> friend class ArPtr;
template <class T> friend class ArConstPtr;
friend class MessageService;

//-------- Type information --------------------------------------------------
public:

// Classes derived from ``ArObject'' MUST use, in their public interface,
// the following macro.
//
#define AR_CLASS(_CLASS)                                        \
  private:                                                      \
    _CLASS(void); /* no implementation */                       \
    _CLASS(const _CLASS &); /* no implementation */             \
    _CLASS & operator=(const _CLASS &); /* no implementation */ \
    static ArCW _CLS;                                           \
    static ArObject * _makeInstance(void);                      \
  protected:                                                    \
    virtual ~_CLASS(void);                                      \
  public:                                                       \
    static void REGISTER_CLASS(void);                           \
    static ArConstRef<ArClass> CLASS(void);                     \
    inline static ArRef<_CLASS> nullRef(void)                   \
      { return(ArRef<_CLASS>((_CLASS *)0)); }                   \
    inline static ArPtr<_CLASS> nullPtr(void)                   \
      { return(ArPtr<_CLASS>((_CLASS *)0)); }                   \
    inline ArRef<_CLASS> thisRef(void)                          \
      { return(ArRef<_CLASS>(this)); }                          \
    inline ArConstRef<_CLASS> thisRef(void) const               \
      { return(ArConstRef<_CLASS>(this)); }                     \
    inline ArPtr<_CLASS> thisPtr(void)                          \
      { return(ArPtr<_CLASS>(this)); }                          \
    inline ArConstPtr<_CLASS> thisPtr(void) const               \
      { return(ArConstPtr<_CLASS>(this)); }

  static
  void
  REGISTER_CLASS(void);

  static
  ArConstRef<ArClass> // ArClass objet for ArObject
  CLASS(void);

  virtual
  ArConstRef<ArClass> // class of this instance
  getClass(void) const;

  virtual
  unsigned int // instance ID in its class
  getClassID(void) const;

  inline static
  ArRef<ArObject> // null reference
  nullRef(void);

  inline static
  ArPtr<ArObject> // null pointer
  nullPtr(void);

  inline
  ArRef<ArObject> // this
  thisRef(void);

  inline
  ArConstRef<ArObject> // const this
  thisRef(void) const;

  inline
  ArPtr<ArObject> // this
  thisPtr(void);

  inline
  ArConstPtr<ArObject> // const this
  thisPtr(void) const;

//-------- Construction / Destruction ----------------------------------------

// Non-abstract classes derived from ``ArObject'' can use, in their public
// interface, the following macros.
//
#define AR_CONSTRUCTOR(_CLASS)                      \
  protected:                                        \
    _CLASS(ArCW & arCW);                            \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(void)                                       \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS));  \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_1(_CLASS,                    \
                         _ARGTYPE,_ARGNAME)         \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE _ARGNAME);                      \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE _ARGNAME)                          \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME));      \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_2(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_3(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2,       \
                         _ARGTYPE3,_ARGNAME3)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2,                     \
           _ARGTYPE3 _ARGNAME3);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2,                        \
        _ARGTYPE3 _ARGNAME3)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2,       \
                                   _ARGNAME3));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_4(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2,       \
                         _ARGTYPE3,_ARGNAME3,       \
                         _ARGTYPE4,_ARGNAME4)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2,                     \
           _ARGTYPE3 _ARGNAME3,                     \
           _ARGTYPE4 _ARGNAME4);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2,                        \
        _ARGTYPE3 _ARGNAME3,                        \
        _ARGTYPE4 _ARGNAME4)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2,       \
                                   _ARGNAME3,       \
                                   _ARGNAME4));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_5(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2,       \
                         _ARGTYPE3,_ARGNAME3,       \
                         _ARGTYPE4,_ARGNAME4,       \
                         _ARGTYPE5,_ARGNAME5)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2,                     \
           _ARGTYPE3 _ARGNAME3,                     \
           _ARGTYPE4 _ARGNAME4,                     \
           _ARGTYPE5 _ARGNAME5);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2,                        \
        _ARGTYPE3 _ARGNAME3,                        \
        _ARGTYPE4 _ARGNAME4,                        \
        _ARGTYPE5 _ARGNAME5)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2,       \
                                   _ARGNAME3,       \
                                   _ARGNAME4,       \
                                   _ARGNAME5));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_6(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2,       \
                         _ARGTYPE3,_ARGNAME3,       \
                         _ARGTYPE4,_ARGNAME4,       \
                         _ARGTYPE5,_ARGNAME5,       \
                         _ARGTYPE6,_ARGNAME6)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2,                     \
           _ARGTYPE3 _ARGNAME3,                     \
           _ARGTYPE4 _ARGNAME4,                     \
           _ARGTYPE5 _ARGNAME5,                     \
           _ARGTYPE6 _ARGNAME6);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2,                        \
        _ARGTYPE3 _ARGNAME3,                        \
        _ARGTYPE4 _ARGNAME4,                        \
        _ARGTYPE5 _ARGNAME5,                        \
        _ARGTYPE6 _ARGNAME6)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2,       \
                                   _ARGNAME3,       \
                                   _ARGNAME4,       \
                                   _ARGNAME5,       \
                                   _ARGNAME6));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_7(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2,       \
                         _ARGTYPE3,_ARGNAME3,       \
                         _ARGTYPE4,_ARGNAME4,       \
                         _ARGTYPE5,_ARGNAME5,       \
                         _ARGTYPE6,_ARGNAME6,       \
                         _ARGTYPE7,_ARGNAME7)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2,                     \
           _ARGTYPE3 _ARGNAME3,                     \
           _ARGTYPE4 _ARGNAME4,                     \
           _ARGTYPE5 _ARGNAME5,                     \
           _ARGTYPE6 _ARGNAME6,                     \
           _ARGTYPE7 _ARGNAME7);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2,                        \
        _ARGTYPE3 _ARGNAME3,                        \
        _ARGTYPE4 _ARGNAME4,                        \
        _ARGTYPE5 _ARGNAME5,                        \
        _ARGTYPE6 _ARGNAME6,                        \
        _ARGTYPE7 _ARGNAME7)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2,       \
                                   _ARGNAME3,       \
                                   _ARGNAME4,       \
                                   _ARGNAME5,       \
                                   _ARGNAME6,       \
                                   _ARGNAME7));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

#define AR_CONSTRUCTOR_8(_CLASS,                    \
                         _ARGTYPE1,_ARGNAME1,       \
                         _ARGTYPE2,_ARGNAME2,       \
                         _ARGTYPE3,_ARGNAME3,       \
                         _ARGTYPE4,_ARGNAME4,       \
                         _ARGTYPE5,_ARGNAME5,       \
                         _ARGTYPE6,_ARGNAME6,       \
                         _ARGTYPE7,_ARGNAME7,       \
                         _ARGTYPE8,_ARGNAME8)       \
  protected:                                        \
    _CLASS(ArCW & arCW,                             \
           _ARGTYPE1 _ARGNAME1,                     \
           _ARGTYPE2 _ARGNAME2,                     \
           _ARGTYPE3 _ARGNAME3,                     \
           _ARGTYPE4 _ARGNAME4,                     \
           _ARGTYPE5 _ARGNAME5,                     \
           _ARGTYPE6 _ARGNAME6,                     \
           _ARGTYPE7 _ARGNAME7,                     \
           _ARGTYPE8 _ARGNAME8);                    \
  public:                                           \
    inline static                                   \
    ArRef<_CLASS> /* created instance */            \
    NEW(_ARGTYPE1 _ARGNAME1,                        \
        _ARGTYPE2 _ARGNAME2,                        \
        _ARGTYPE3 _ARGNAME3,                        \
        _ARGTYPE4 _ARGNAME4,                        \
        _ARGTYPE5 _ARGNAME5,                        \
        _ARGTYPE6 _ARGNAME6,                        \
        _ARGTYPE7 _ARGNAME7,                        \
        _ARGTYPE8 _ARGNAME8)                        \
      {                                             \
      if(!_CLASS::_CLS)                             \
        {                                           \
        cerr << "!!! Don't forget to call " #_CLASS \
                "::REGISTER_CLASS() !!!" << endl;   \
        _CLASS::REGISTER_CLASS();                   \
        }                                           \
      ArRef<_CLASS> ref(new _CLASS(_CLASS::_CLS,    \
                                   _ARGNAME1,       \
                                   _ARGNAME2,       \
                                   _ARGNAME3,       \
                                   _ARGNAME4,       \
                                   _ARGNAME5,       \
                                   _ARGNAME6,       \
                                   _ARGNAME7,       \
                                   _ARGNAME8));     \
      ref->setTransient(true);                      \
      return(ref);                                  \
      }

  inline static
  ArRef<ArObject> // created instance
  NEW(void);

//-------- Garbage collection control ----------------------------------------

  void
  setTransient(bool transient); // NOT virtual !

  bool // transient
  isTransient(void) const; // NOT virtual !

  unsigned int // nb references
  getNbReferences(void) const; // NOT virtual !

  AR_CALLBACK(ArObject,Destroy,_destroyCB,
              /* no field */)

//-------- Alias management --------------------------------------------------

  virtual
  bool // success
  setAlias(const StlString & alias);

  virtual
  const StlString & // alias or className.ID
  getAlias(void) const;

//-------- Message management ------------------------------------------------

  virtual
  unsigned int // number of pending messages
  getNbMessages(void) const;

  virtual
  ArConstRef<Message> // next message or 0
  getNextMessage(void);

  AR_CALLBACK(ArObject,Message,_messageCB,
              /* no field */)

//-------- Stream operations -------------------------------------------------

  virtual
  void
  writeToStream(ArRef<AbstractOStream> output) const;

  virtual
  bool // success
  readFromStream(ArRef<AbstractIStream> input);

//-------- Serialization operations ------------------------------------------

  AR_UNSTUBBED virtual
  void
  extractDependencies(SerializationDependencies & dependenciesOut,
                      bool recursive=true) const;

  AR_UNSTUBBED virtual
  void
  serialize(SerializationDependencies & dependenciesOut,
            ArRef<AbstractOStream> output) const;

  AR_UNSTUBBED virtual
  bool // success
  unserialize(const SerializationDependencies & dependencies,
              ArRef<AbstractIStream> input);

//-------- Error management --------------------------------------------------

  virtual
  void
  setErrorMessage(const StlString & errorMessage);

  virtual
  bool // problem occured
  fail(void) const;

  virtual
  const StlString & // problem description
  getErrorMessage(void) const;

//-------- User data ---------------------------------------------------------

  virtual
  void
  setUserData(void * userData);

  virtual
  void * // user data
  getUserData(void) const;

//-------- Forbidden operations ----------------------------------------------
protected:

  ArObject(ArCW & arCW);

  virtual
  ~ArObject(void);

  inline
  void *
  operator new(size_t size);

  inline
  void
  operator delete(void * address,
                  size_t size);

private:

  ArObject(void); // only used by ArClass

  ArObject(const ArObject & obj); // no implementation

  ArObject & // itself
  operator=(const ArObject & obj); // no implementation

  inline
  static
  void
  _manageRef(ArObject ** ref);

  inline
  static
  void
  _unmanageRef(ArObject ** ref,
               ArObject * obj);

  inline
  static
  void
  _managePtr(ArObject ** ptr);

  inline
  static
  void
  _unmanagePtr(ArObject ** ptr,
               ArObject * obj);

  inline
  void
  _performDestruction(void); // NOT virtual !

  static
  ArObject * // created instance
  _makeInstance(void);

//-------- Member attributes -------------------------------------------------
private:
  ArCW & _arCW;
  unsigned int _id;
  StlVector<ArObject **> _refs;
  StlVector<ArObject **> _ptrs;
  bool _valid;
  bool _transient;
  CallbackManager<ArObject,DestroyEvent> _destroyCB;
  StlString _alias;
  StlList<ArConstRef<Message> > _messages;
  CallbackManager<ArObject,MessageEvent> _messageCB;
  StlString _errorMessage;
  void * _userData;

//-------- Static attributes -------------------------------------------------
protected:
  static ArCW _CLS;
};

} // namespace AReVi

//-------- Inline source code ------------------------------------------------

#include "AReVi/arObject.Ci"

#include "AReVi/arClass.h" // make ArClass usable everywhere ...

#endif // AREVI_AROBJECT_H

//----------------------------------------------------------------------------
