// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/arSystem.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_ARSYSTEM_H
#define AREVI_ARSYSTEM_H 1

#include "AReVi/scheduler.h"

namespace AReVi {

class ThreadMutex;

//----------------------------------------------------------------------------
// Class ArSystem (singleton)
//----------------------------------------------------------------------------

class ArSystem
{
//-------- Construction / Destruction ----------------------------------------
public:

  ArSystem(int argc,
           char ** argv,
           bool use3D=true);

  virtual
  ~ArSystem(void);

//-------- Initialization ----------------------------------------------------

  static
  void
  init(int argc,
       char ** argv,
       bool use3D=true);

  static
  void
  destroy(void);

  static
  const StlVector<StlString> & // command line
  getCommandLine(void);

  static
  bool // 3D is used
  uses3D(void);

  static
  bool // success
  loadPlugin(const StlString & pluginName,
             const StlVector<StlString> * params
                                          =(const StlVector<StlString> *)0);

  static
  bool // plugin is loaded
  usesPlugin(const StlString & pluginName);

  enum InitState { AR_DOWN, AR_GOING_UP, AR_UP, AR_GOING_DOWN };

  static
  ArSystem::InitState // init state
  getInitState(void);

  static
  unsigned int // current AReVi version
  getVersion(void);

//-------- Simulation --------------------------------------------------------

  static
  bool // go on
  enterSimulation(ArRef<Scheduler> (*simulationInit)(void));

  static
  bool // go on
  simulationStep(void);

  static
  void
  leaveSimulation(void);

  static
  void
  simulationLoop(ArRef<Scheduler> (*simulationInit)(void));

//-------- Error Messages ----------------------------------------------------

  static
  StlString // error description message
  printError(int errorNumber);

//-------- Memory allocation -------------------------------------------------

  template <typename T> inline static
  T *
  allocate(unsigned int nbElems);

  template <typename T> inline static
  void
  deallocate(T * elems);

  template <typename T> inline static
  T *
  reallocate(T * elems,
             unsigned int newNbElems);

//-------- Random Generator --------------------------------------------------

  static
  void
  setRandomSeed(unsigned int seed,
                unsigned long long count);

  static
  void
  getRandomSeed(unsigned int & seedOut,
                unsigned long long & countOut);

  static
  unsigned int // random value in [0;limit-1]
  integerRand(unsigned int limit);

  static
  double // random value in [0.0;1.0[
  realRand(void);

//-------- Little/Big Endian -------------------------------------------------

  inline static
  bool // little endian detected
  isLittleEndian(void);

  inline static
  bool // big endian detected
  isBigEndian(void);

  inline static
  void
  invert16(void * ptr);

  inline static
  void
  invert32(void * ptr);

  inline static
  void
  invert64(void * ptr);

  inline static
  bool // inverted
  inOut16(void * ptr);

  inline static
  bool // inverted
  inOut32(void * ptr);

  inline static
  bool // inverted
  inOut64(void * ptr);

//-------- Signal Management -------------------------------------------------

  typedef const int SignalNumber;
  typedef int ProcessID;
  typedef void (*SignalHandler)(ArSystem::SignalNumber);

  static const ArSystem::SignalHandler SIGHANDLER_ERROR;
  static const ArSystem::SignalHandler SIGHANDLER_DEFAULT;
  static const ArSystem::SignalHandler SIGHANDLER_IGNORE;

  static const ArSystem::SignalNumber
         SIGNUM_NONE   ,  // No signal
         SIGNUM_HUP    ,  // Hangup                            (POSIX)
         SIGNUM_INT    ,  // Interrupt                          (ANSI)
         SIGNUM_QUIT   ,  // Quit                              (POSIX)
         SIGNUM_ILL    ,  // Illegal instruction                (ANSI)
         SIGNUM_TRAP   ,  // Trace/breakpoint trap             (POSIX)
         SIGNUM_ABRT   ,  // Abort                              (ANSI)
         SIGNUM_BUS    ,  // Bus error                       (4.2 BSD)
         SIGNUM_FPE    ,  // Arithmetic exception               (ANSI)
         SIGNUM_KILL   ,  // Killed, unblockable               (POSIX)
         SIGNUM_USR1   ,  // User defined signal 1             (POSIX)
         SIGNUM_SEGV   ,  // Segmentation violation             (ANSI)
         SIGNUM_USR2   ,  // User defined signal 2             (POSIX)
         SIGNUM_PIPE   ,  // Broken pipe                       (POSIX)
         SIGNUM_ALRM   ,  // Alarm clock                       (POSIX)
         SIGNUM_TERM   ,  // Terminated                         (ANSI)
         SIGNUM_CHLD   ,  // Child status changed              (POSIX)
         SIGNUM_CLD    ,  // A synonym for CHLD             (System V)
         SIGNUM_CONT   ,  // Continue                          (POSIX)
         SIGNUM_STOP   ,  // Stopped, unblockable              (POSIX)
         SIGNUM_TSTP   ,  // Stopped from tty                  (POSIX)
         SIGNUM_TTIN   ,  // Stopped, background read from tty (POSIX)
         SIGNUM_TTOU   ,  // Stopped, background write to tty  (POSIX)
         SIGNUM_URG    ,  // Urgent I/O condition            (4.2 BSD)
         SIGNUM_XCPU   ,  // CPU time limit exceeded         (4.2 BSD)
         SIGNUM_XFSZ   ,  // File size limit exceeded        (4.2 BSD)
         SIGNUM_VTALRM ,  // Virtual alarm clock             (4.2 BSD)
         SIGNUM_PROF   ,  // Profiling alarm clock           (4.2 BSD)
         SIGNUM_WINCH  ,  // Window size change         (4.3 BSD, Sun)
         SIGNUM_POLL   ,  // Pollable event                 (System V)
         SIGNUM_IO     ,  // I/O now possible                (4.2 BSD)
         SIGNUM_SYS    ;  // Bad system call                (System V)

  static
  StlString // signal description message
  printSignal(ArSystem::SignalNumber signum);

  static
  ArSystem::ProcessID // process identifier
  processID(void);

  static
  ArSystem::ProcessID // parent identifier
  parentID(void);

  static
  bool // process reached
  sendSignal(ArSystem::ProcessID processID,
             ArSystem::SignalNumber signum);

  static
  ArSystem::SignalHandler // old handler
  installSignalHandler(ArSystem::SignalNumber signum,
                       ArSystem::SignalHandler handler);

  static
  void
  ignoreAllSignals(void);

  static
  void
  restoreAllDefaultHandlers(void);

//-------- Time Management ---------------------------------------------------

  static
  void
  yield(void);

  static
  void
  sleep(unsigned int millisecond);

  static
  unsigned int // time in millisecond
  getTime(void);

  static
  void
  setRealTimer(unsigned int millisecond,
               bool repeat);

  static
  void
  setVirtualTimer(unsigned int millisecond,
                  bool repeat);

  static
  void
  setProfilingTimer(unsigned int millisecond,
                    bool repeat);

//-------- Library loading ---------------------------------------------------

  static
  void * // !=0: library handle  0: error
  openLibrary(const StlString & libName,
              bool convertLibName);

  static
  void * // !=0: symbol  0: error
  getLibrarySymbol(void * library,
                   const StlString & symbolName);

  static
  StlString // library error description message
  printLibraryError(void);

  static
  void
  closeLibrary(void * library);

//-------- File Management ---------------------------------------------------

  static
  int // >=0: file descriptor  -1: error
  openInFile(const StlString & fileName);

  static
  int // >=0: file descriptor  -1: error
  openOutFile(const StlString & fileName,
              bool append);

  static
  int // >=0: file descriptor  -1: error
  openInOutFile(const StlString & fileName);

  static
  int // >=0: file descriptor  -1: error
  createTmpFile(const StlString & prefix,
                StlString & fileNameOut);

  static
  bool // success
  getFileKind(const StlString & fileName,
              StlString & kindOut);

  static
  bool // success
  removeFile(const StlString & fileName);

//-------- Directory Management ----------------------------------------------

  static
  bool // success
  getDirContent(const StlString & dirName,
                StlVector<StlString> & contentOut);

  static
  int // >=0: file descriptor  -1: error
  createDir(const StlString & dirName);

  static
  int // >=0: file descriptor  -1: error
  createTmpDir(const StlString & prefix,
               StlString & dirNameOut);

  static
  bool // success
  removeDir(const StlString & dirName);

//-------- IP Services -------------------------------------------------------

  static
  StlString // local host name
  getHostName(void);

  static
  unsigned int // >0: IP address  0: error
  getIPAddress(const StlString & hostName);

  static
  void
  setHTTPProxy(const StlString & proxyHost,
               int proxyPort);

  static
  void
  getHTTPProxy(StlString & proxyHostOut,
               int & proxyPortOut);

  static
  bool // success
  decodeURL(const StlString & url,
            StlString & protocolOut,
            StlString & hostOut,
            int & portOut,
            StlString & fileOut);

//-------- Socket Management -------------------------------------------------

  static
  int // >=0: file descriptor  -1: error
  connectTCP(unsigned int ipAddress,
             int portNumber);

  static
  int // >=0: file descriptor  -1: error
  listenTCP(int portNumber);

  static
  int // >=0: file descriptor  -1: error  -2: non-blocking
  acceptTCP(int listenSocket,
            unsigned int & ipAddressOut,
            int & portNumberOut);

  static
  bool // success
  getSocketAddress(int sock,
                   unsigned int & localIPAddressOut,
                   int & localPortNumberOut);

  static
  int // >=0: file descriptor  -1: error
  openUDP(int portNumber);

  static
  const StlVector<unsigned int> & // UDP broadcast addresses
  getUDPBroadcastAddresses(void);

  static
  int // >0: nb received  0: EOF  -1: error  -2: non-blocking
  receiveUDP(int sock,
             void * buffer,
             unsigned int size,
             unsigned int & ipAddressOut,
             int & portNumberOut);

  static
  int // >=0: nb written  -1: error
  sendUDP(int sock,
          const void * buffer,
          unsigned int size,
          unsigned int ipAddress,
          int portNumber);

//-------- File Descriptor Management ----------------------------------------

  static
  bool // success
  closeFD(int fileDescriptor);

  static
  bool // success
  syncFD(int fileDescriptor);

  static
  bool // success
  setBlockingFD(int fileDescriptor,
                bool blocking);

  static
  bool // blocking
  isBlockingFD(int fileDescriptor);

  static
  int // >0: nb read  0: EOF  -1: error  -2: non-blocking
  partialReadFD(int fileDescriptor,
                void * buffer,
                unsigned int size);

  static
  int // >0: nb read  0: EOF  -1: error
  readFD(int fileDescriptor,
         void * buffer,
         unsigned int size);

  static
  int // >=0: nb written  -1: error  -2: non-blocking
  partialWriteFD(int fileDescriptor,
                 const void * buffer,
                 unsigned int size);

  static
  int // >=0: nb written  -1: error
  writeFD(int fileDescriptor,
          const void * buffer,
          unsigned int size);

//-------- Polling Management ------------------------------------------------

  static
  void
  addReadFDPolling(int fileDescriptor);

  static
  void
  removeReadFDPolling(int fileDescriptor);

  static
  bool // fileDescriptor is polled for reading
  hasReadFDPolling(int fileDescriptor);

  static
  void
  addWriteFDPolling(int fileDescriptor);

  static
  void
  removeWriteFDPolling(int fileDescriptor);

  static
  bool // fileDescriptor is polled for writing
  hasWriteFDPolling(int fileDescriptor);

  static
  int // >=0: nb ready  -1: error
  waitForPolledFD(StlVector<int> & readingFDOut,
                  StlVector<int> & writingFDOut,
                  unsigned int millisecond);

//-------- Forbidden operations ----------------------------------------------
private:

  static
  void
  _forceSettings(void);

  static
  void
  _exitFunc(void);

  static
  void
  _setTimer(int which,
            unsigned int millisec,
            bool repeat);

  static
  void
  _exitHandler(ArSystem::SignalNumber signum);

  ArSystem(void); // no implementation

  ArSystem(const ArSystem & obj); // no implementation

  ArSystem & // itself
  operator=(const ArSystem & obj); // no implementation

  void *
  operator new(size_t size); // no implementation

  void
  operator delete(void * /* address */,
                  size_t /* size */) {} // unused but mandatory !

//-------- Static attributes -------------------------------------------------
protected:

  static ArSystem::InitState _initState;
  static StlVector<StlString> * _commandLine;
  static bool _use3D;
  static ArSystem::ProcessID _pid;
  static ArSystem::ProcessID _ppid;
  static ArRef<ThreadMutex> _mtx;
  static StlString _httpProxyHost;
  static int _httpProxyPort;
  static StlVector<unsigned int> * _broadcast;
  static StlVector<int> * _readingPoll;
  static StlVector<int> * _writingPoll;
  static ArRef<Scheduler> _sched;
  static bool _goOn;
  struct PluginInfo
    {
    StlString name;
    void * library;
    bool (*initFnct)(const StlVector<StlString> *);
    void (*updateFnct)(void);
    void (*destroyFnct)(void);
    };
  static StlVector<ArSystem::PluginInfo> * _plugins;
  static unsigned int _randomSeed;
  static unsigned long long _randomCount;
};

} // namespace AReVi

//-------- Inline source code ------------------------------------------------

#include "AReVi/arSystem.Ci"

#endif // AREVI_ARSYSTEM_H

//----------------------------------------------------------------------------
