// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/scheduler.h
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVI_SCHEDULER_H
#define AREVI_SCHEDULER_H 1

#include "AReVi/activity.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Scheduler (abstract, singleton)
//----------------------------------------------------------------------------

class Scheduler : public ArObject
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(Scheduler)

//-------- Singleton ---------------------------------------------------------

  static
  ArConstRef<Scheduler> // the unique scheduler
  get(void);

  static
  ArRef<Scheduler> // the unique scheduler
  access(void);

//-------- Scheduling properties ---------------------------------------------

  virtual
  double // current time
  getTime(void) const;

  virtual
  double // time rounding precision
  getPrecision(void) const;

  virtual
  double // time rounded to precision
  roundTime(double time) const;

//-------- Activation --------------------------------------------------------

  virtual
  void
  setSuspended(bool suspended);

  virtual
  bool // suspended
  isSuspended(void) const;

//-------- File descriptor polling -------------------------------------------

  AR_CALLBACK(Scheduler,Polling,_pollingCB,
              StlVector<int> readingFD;
              StlVector<int> writingFD;)

//-------- Forbidden operations ----------------------------------------------
protected:

  friend class Activity;
  friend class ArSystem;

  Scheduler(ArCW & arCW,
            double _precision);

  virtual
  void
  _initialize(void);

  virtual
  bool // remaining activities
  _activate(void);

  virtual
  bool // remaining activities
  _jumpToNextTime(void)=0;

  virtual
  void
  _insertActivity(ArRef<Activity> activity);

  virtual
  void
  _removeActivity(ArRef<Activity> activity);

  virtual
  void
  _changeInterval(ArRef<Activity> activity,
                  double interval);

  virtual
  void
  _changePriority(ArRef<Activity> activity,
                  int priority);

  inline static
  double // rounded time
  _roundTime(const Scheduler * sched,
             double time);

  inline static
  double
  _computeNextTime(const Scheduler * sched,
                   ArRef<Activity> activity);

  inline static
  bool // go on
  _activityActivate(ArRef<Activity> activity,
                    double interval);

  inline static
  double & // activation interval
  _activityInterval(ArRef<Activity> activity);

  inline static
  int & // priority ( -1 0 +1 )
  _activityPriority(ArRef<Activity> activity);

  inline static
  double & // last activation
  _activityLastTime(ArRef<Activity> activity);

  inline static
  double & // next activation
  _activityNextTime(ArRef<Activity> activity);

  inline static
  bool & // suspended
  _activitySuspended(ArRef<Activity> activity);

  inline static
  bool & // ended
  _activityEnded(ArRef<Activity> activity);

  struct ActivityInfo
    {
    double nextTime;
    StlVector<ArRef<Activity> > activities;
    };

  inline static
  Scheduler::ActivityInfo *
  _useInfo(Scheduler * sched,
           double nextTime,
           ArRef<Activity> activity);

  inline static
  void
  _releaseInfo(Scheduler * sched,
               Scheduler::ActivityInfo * info);
  
//-------- Member attributes -------------------------------------------------
protected:

  double _time;
  double _precision;
  bool _suspended;
  CallbackManager<Scheduler,PollingEvent> _pollingCB;
  StlList<Scheduler::ActivityInfo *> _activities[3];
  StlVector<Scheduler::ActivityInfo *> _stock;
  StlVector<ArRef<Activity> > _current;
  ArRef<Activity> _currentActivity;

//-------- Static attributes -------------------------------------------------
protected:

  static ArPtr<Scheduler> _sched;
};


//----------------------------------------------------------------------------
// Class VirtualTimeScheduler (singleton)
//----------------------------------------------------------------------------

class VirtualTimeScheduler : public Scheduler
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(VirtualTimeScheduler)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(VirtualTimeScheduler,
                   double, precision)

//-------- Forbidden operations ----------------------------------------------
protected:

  virtual
  bool // remaining activities
  _jumpToNextTime(void);
};


//----------------------------------------------------------------------------
// Class RealTimeScheduler (singleton)
//----------------------------------------------------------------------------

class RealTimeScheduler : public Scheduler
{
//-------- Type information --------------------------------------------------
public:

  AR_CLASS(RealTimeScheduler)

//-------- Construction / Destruction ----------------------------------------

  AR_CONSTRUCTOR_1(RealTimeScheduler,
                   double, precision)

//-------- Scheduling properties ---------------------------------------------

  virtual
  void
  setTimeFactor(double timeFactor);

  virtual
  double // time factor
  getTimeFactor(void) const;

//-------- Activation --------------------------------------------------------

  virtual
  void
  setSuspended(bool suspended);

//-------- Forbidden operations ----------------------------------------------
protected:

  virtual
  void
  _initialize(void);

  virtual
  bool // remaining activities
  _jumpToNextTime(void);

//-------- Member attributes -------------------------------------------------
protected:

  double _realTime;
  double _timeFactor;
};

} // namespace AReVi

//-------- Inline source code ------------------------------------------------

#include "AReVi/scheduler.Ci"

#endif // AREVI_SCHEDULER_H

//----------------------------------------------------------------------------
