// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/networkService.h
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVINS_NETWORKSERVICE_H
#define AREVINS_NETWORKSERVICE_H 1

//--- Imported classes -------------------------------------------------------
#include "AReViNS/Events/nsEvent.h"

#include "AReVi/scheduler.h"

namespace AReVi
{
  class Activity;
  class TCPConnection;
  class TCPListener;
  class MemoryBlock;
}

#define NS_REQUEST_CONNECTION_MESSAGE         "RequestNSConnection"

//--- Constants --------------------------------------------------------------
#define NS_ERROR_UNKNOW             1  // Non typed error.
#define NS_ERROR_WARNING            2  // Process can continue
#define NS_ERROR_FATAL              3  // Serious problem
#define NS_ERROR_PROCESSEVENT       4  // Error during event process
#define NS_ERROR_INVALIDEVENT       5  // Error during reading event infos
#define NS_ERROR_INSTANCIATEEVENT   6  // Event can't be instancied
#define NS_ERROR_UNKNOWEVENT        7  // EventClass is unknown (here....)

#define NS_ERROR_NETWORKSERVICE_ID  10 // Derived class use this as base

//----------------------------------------------------------------------------
// CLASS       : NetworkService {Abstract}
//----------------------------------------------------------------------------
namespace AReVi {

void registerAReViNSClasses(void);

class NSEvent;

/*! \class NetworkService
  * Class managing local or remote service
  */
class NetworkService : public ArObject
{
public:
  typedef NSEvent::EventProcess
          (ArObject::*EventMethod)(ArConstRef<NSEvent> event);

public:
  AR_UNSTUBBED AR_CLASS(NetworkService)

//---------------------------- Public methods --------------------------------
public:
  //! Close connection with specified service or all
  virtual bool //success
  close(int serviceId=-1);

  unsigned int
  getNbServices(void) const;

  void
  getServices(StlVector<int> &servicesIdOut) const;

  //! Period of network refresh => <0 => stop watching to network
  void
  setUpdateInterval(double dt);

  //! Set session name
  virtual void
  setSessionName(const StlString &sessionName);

  //! Get session name
  virtual const StlString &
  getSessionName(void) const;

//-------- NSEvents management -----------------------------------------------
  //! Pending events => process pending events return true if success
  bool //success
  processEvents(void);

  //! Add an event to send to services (service=-1 <=> All)
  bool
  sendEvent(ArRef<NSEvent> event,int serviceId=-1);

  //! Register and unregister events
  virtual bool
  registerEvent(const StlString &eventClass,
                ArRef<ArObject> performer,
                EventMethod     method);

  virtual bool
  unregisterEvent(const StlString &eventClass);

//-------- Event management --------------------------------------------------

  AR_CALLBACK(NetworkService,ServiceDeconnect,_serviceDeconnectCB,
             int serviceID;)

  AR_CALLBACK(NetworkService,NetworkServiceError,_networkServiceErrorCB,
              unsigned int errorID;)

//------------------------- Forbiden operations ------------------------------
protected:
  NetworkService(ArCW & arCW);


  // Open a new connection into service
  virtual int //serviceID
  _open(ArRef<TCPConnection> connectionRead,
        ArRef<TCPConnection> connectionWrite,
        const StlString &sessionName,
        const StlString &hostName);

  //! Port number of serviceId.
  virtual void
  _getPortNumber(int serviceId, int &readPort, int &writePort) const;

  //! Port number of remote serviceId.
  virtual void
  _getDistantPortNumber(int serviceId, int &readPort, int &writePort) const;

  //! Port number of remote serviceId.
  virtual unsigned int
  _getDistantIPAdress(int serviceId) const;

  //! Set Connection message
  virtual void
  _setConnectionMessage(const StlString &connectionMessage);

  //! Service session name
  virtual StlString  // "" on error
  _getServiceSessionName(int serviceId) const;

  //! Get service host name
  virtual StlString
  _getServiceHostName(int serviceId) const;

//------------------------- Processing Events ------------------------------
  //! Tools Event<=>MemoryBlock
  bool
  _makeMemoryBlockFromEvent(ArRef<NSEvent> event, ArRef<MemoryBlock> memory, bool compress=true);

  ArRef<NSEvent>
  _makeEventFromMemoryBlock(ArConstRef<MemoryBlock> memory, int serviceId);

  //! write to network (write messages).
  bool
  _setNetworkEvents(void);

  //! watch network (look for news messages).
  bool
  _getNetworkEvents(void);

  virtual bool
  _updateNetwork(ArRef<Activity> act, double dt);

  virtual void
  _onPolling(const Scheduler::PollingEvent & evt);

  //! Connection method
  virtual
  ArRef<TCPConnection>
  _connect(unsigned int distantIPAddress,
           int           distantPort);

  //! Get next valid connection request
  virtual
  ArRef<TCPConnection>
  _acceptConnection(ArRef<TCPListener> listener, const StlString &requestedMessage);

//------------------------- Network Statistics ------------------------------
  //! Read data from this service
  virtual unsigned int
  _getDataReadCount(int serviceId) const;

  //! Write data to this service
  virtual unsigned int
  _getDataWriteCount(int serviceId) const;

  //! Ping specified service (or all with -1)
  virtual bool
  _ping(int serviceId=-1);

  //! CB : reception of a pingRequest Event
  virtual NSEvent::EventProcess
  _onRequestPingEvent(ArConstRef<NSEvent> eventObj);

  //! CB : reception of a ping Event
  virtual NSEvent::EventProcess
  _onPingEvent(ArConstRef<NSEvent> eventObj);

  //! CB : notify distant session name changed
  NSEvent::EventProcess
  _onSessionNameEvent(ArConstRef<NSEvent> eventObj);

  //! handle error
  virtual void
  _onError(unsigned int errorId);

private:
//----------------------------- Attributes -----------------------------------
protected:
  struct EventPerformer
  {
    ArPtr<ArObject> performer;
    EventMethod     method;
    bool            autoref;
  };
  // Link event type with a performer
  StlMap<StlString, EventPerformer> _eventsTable;

  //! Message that must be send by remote service to accept connection
  StlString _connectionMessage;

  //! Session Name => sent at connection time
  StlString _sessionName;

  // List of received events
  StlList<ArRef<NSEvent> > _eventsReceived;
  StlList<ArRef<NSEvent> > _eventsPosted;

  CallbackManager<NetworkService,ServiceDeconnectEvent> _serviceDeconnectCB;
  CallbackManager<NetworkService,NetworkServiceErrorEvent> _networkServiceErrorCB;

  ArRef<Activity> _networkAct;

  //!  Ping <=>time for emission/reception of a single message.
  StlMap<int, unsigned int>        _pingTimes;

private:
  //! List of connected remote services (private : use open and close)!
  typedef struct ServiceInfo
  {
    ArRef<TCPConnection> _connectionRead;
    ArRef<TCPConnection> _connectionWrite;
    unsigned int _dataReadCount;
    unsigned int _dataWriteCount;
    StlString     _sessionName;
    StlString     _hostName;
  } ServiceInfo;

  StlMap<int, ServiceInfo> _services;


  int _serviceId; // incremented at each connection.
//-------- Static attributes -------------------------------------------------
public:
  static
  bool
  getCompressData(void);

  static
  void
  setCompressData(bool compressData);

  static
  unsigned int
  getTimeOut(void);

  static
  void
  setTimeOut(unsigned int timeOut);


protected:
  static bool _compressData; // Usefull only for nbClient > 3
  static unsigned int _timeOut; // on connection, waiting for reply
};

} // end of namespace AReVi
#endif // end of AREVINS_NETWORKSERVICE_H
//---End-Of-File---------------------------------------------------------------
