// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/remoteDepService.h
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#ifndef AREVINS_REMOTEDEPSERVICE_H
#define AREVINS_REMOTEDEPSERVICE_H 1

//--- Imported classes -------------------------------------------------------
#include "AReViNS/remoteService.h"

//--- Constants --------------------------------------------------------------
// Interactor fail to associated with shared ArObject
// - Interactor class not exist
// - Interactor class have not default constructor
#define NS_ERROR_INTERACTOR_FAIL          (NS_ERROR_REMOTESERVICE_ID+1)

// Can't create a remote object
// - ArObject class not exist
// - ArObject class have not default constructor
#define NS_ERROR_REMOTEINSTANCIATION_FAIL (NS_ERROR_REMOTESERVICE_ID+2)

// Can't resolve dependencies for a remote ArObject
// - Ref Dependence(s) can't be instanciated (next an NS_ERROR_REMOTEINSTANCIATION_FAIL)
// - URL Dependence(s) can't be resolved.
#define NS_ERROR_RESOLVEDEPENDENCIES_FAIL (NS_ERROR_REMOTESERVICE_ID+3)

// Error during deserialization of an object (plugin not loaded?,bug?)
#define NS_ERROR_UNSERIALIZE_OBJECT       (NS_ERROR_REMOTESERVICE_ID+3)

// Error creating temporary file (setRemoteURLsPrefix correct? write protect?)
#define NS_ERROR_CREATETEMPFILE           (NS_ERROR_REMOTESERVICE_ID+4)

// Error writing temporary file (write protect?, disk full?)
#define NS_ERROR_WRITINGTEMPFILE          (NS_ERROR_REMOTESERVICE_ID+5)

// Error with an event controler (invalid message from listener...?)
#define NS_ERROR_EVENTCONTROLER           (NS_ERROR_REMOTESERVICE_ID+6)

// Derived class use this as base
#define NS_ERROR_REMOTEDEPSERVICE_ID      (NS_ERROR_REMOTESERVICE_ID+10)

//----------------------------------------------------------------------------
// CLASS       : RemoteService
//----------------------------------------------------------------------------
namespace AReVi {
  class NSInteractor;

/*! \class RemoteDepService
  * Class managing remote service
  */
class RemoteDepService : public RemoteService
{
public:
  AR_UNSTUBBED AR_CLASS(RemoteDepService)
//---------------------- Constructors / Destructor ---------------------------
  // Service will be local (isRemote=false)
  AR_CONSTRUCTOR_2(RemoteDepService,
                   unsigned int , distantIPAddress,
                   int           , distantPort)

//---------------------------- Public methods --------------------------------
public:
//----------------------- Remote objects manager -----------------------------
  const StlMap<StlString, ArRef<ArObject> > &
  getRemoteRefs(void) const;

  //! return in instanceOut reference to serverAlias
  //! if ref no exist instanceOut=T1::nullRef()
  template <class T1>
  inline void
  findRemoteObject(const StlString & serverAlias,
                   ArRef<T1> &instanceOut);

  //! return in instanceOut reference to serverAlias
  //! if ref no exist remote service will instanciate ref(with className)
  //! and request its data to server.
  template <class T1>
  inline void
  requestRemoteObject(const StlString & serverAlias,
                      const StlString &className,
                      ArRef<T1> &instanceOut);

  //! Comment
  virtual
  bool
  linkRemoteObjectToServer(const StlString & serverAlias,
                           ArRef<ArObject> remoteInstance);

  const StlMap<StlString, StlString> &
  getRemoteURLs(void) const;

  bool
  findRemoteURL(const StlString & serverUrl,
                StlString &remoteUrlOut) const;

  //! Same that findRemoteURL exept that add request to server if url not
  //! already exist (and return false).
  bool
  requestRemoteURL(const StlString & serverUrl,
                StlString &remoteUrlOut);

  //! Request serialisation of specified objects.
  virtual bool // success
  addRemoteReference(const StlString &serverRef, const StlString &className);

  //! Request serialisation of specified objects.
  virtual bool // success
  updateRemoteReference(const StlString &serverRef);

  //! Request suppression of element
  virtual bool // success
  removeRemoteReference(const StlString &serverRef);

  //! Set remote temps url prefix
  //! eg : /tmp/remoteFile => /tmp/remoteFile02d0e
  //!                      => /tmp/remoteFilefeggr ...
  virtual void
  setRemoteURLsPrefix(const StlString &prefix);

  //! Request serialisation of specified objects.
  virtual bool // true if success
  addRemoteURL(const StlString &url);

  //! Get list of waiting for references
  inline const StlList<StlString> &
  getWaitingURLs(void) const;

  //! Get list of waiting for urls
  inline const StlList<StlString> &
  getWaitingRefs(void) const;

  bool // true if all are resolved
  resolveDependencies(const StlVector<StlString> &refs,
                      const StlVector<StlString> &urls,
                      SerializationDependencies &depsOut,
                      StlVector<unsigned int> &missingRefsOut,
                      StlVector<unsigned int> &missingUrlsOut);

//------------------------  Controlers management ----------------------------
//! Register controlers (and all subclass boolean is true).
  virtual bool
  registerInteractor(const StlString &objectClass,
                     const StlString &interactorClass,
                     bool subClass);

//----------------------------Event management -------------------------------
  AR_CALLBACK(RemoteDepService,RemoteFileReception,_remoteFileReceptionCB,
              StlString serverUrl;)

  AR_CALLBACK(RemoteDepService,RemoteRefReception,_remoteRefReceptionCB,
              StlString serverAlias;)

//------------------------- Forbiden operations ------------------------------
protected:
  void
  _initRemoteDepService(unsigned int  distantIPAddress,int port);

  //! Pre-Create an object => It can be referenced...
  bool
  _addRemoteReference(const StlString &remoteRef,const StlString &className);

  //! CB : inform that a object is sent by server
  virtual NSEvent::EventProcess
  _onSetRemoteObject(ArConstRef<NSEvent> eventObj);

  //! CB : set downcast map
  virtual NSEvent::EventProcess
  _onSetDownCastMap(ArConstRef<NSEvent> eventObj);

  //! CB : inform that an URL is sent by server
  virtual NSEvent::EventProcess
  _onSetRemoteURL(ArConstRef<NSEvent> eventObj);

  //! CB : inform of message from a server controler
  virtual NSEvent::EventProcess
  _onControlerEvent(ArConstRef<NSEvent> eventObj);

  //! CB : inform of message pack from a server controler
  virtual NSEvent::EventProcess
  _onControlerEventPack(ArConstRef<NSEvent> eventObj);

  virtual bool
  _updateNetwork(ArRef<Activity> act, double dt);

  virtual void
  _onPolling(const Scheduler::PollingEvent & evt);

//----------------------------- Attributes -----------------------------------
protected:
  //! Map of remote objects ServerAlias=>LocalRemoteRef
  StlMap<StlString, ArRef<ArObject> > _remoteRefs;

  //! List of waiting for references
  StlList<StlString> _waitRefs;

  //! Path where urls will be stored (eg "tmp/RemoteFile")
  StlString _path;

  //! Map of remote URLs ServerURL=>LocalRemoteURL
  StlMap<StlString, StlString> _remoteUrls;

  //! List of waiting for urls
  StlList<StlString> _waitUrls;

  // Link shared object class with controlers class
  StlMap<StlString, StlString> _interactorsMap;

  // Map of shared objects => object controler
  StlMap<StlString, ArRef<NSInteractor> > _interactors;

  // Map serverClass to clientClass (serverClass must derive clientClass!)
  StlMap<StlString, StlString> _downCastMap;

  // State variable : true on ControlersPack processing
  bool _processingControlerPacks;

  CallbackManager<RemoteDepService,RemoteRefReceptionEvent> _remoteRefReceptionCB;
  CallbackManager<RemoteDepService,RemoteFileReceptionEvent> _remoteFileReceptionCB;
};

//----------------------------------------------------------------------------
// CLASS       : RemoteRef
//----------------------------------------------------------------------------
class RemoteRef
{
public:
  RemoteRef(ArConstRef<ArObject> ref=ArObject::nullRef());

  template <class T1>
  inline bool
  request(ArRef<RemoteDepService> depService,
          ArRef<T1> &instanceOut) const;

  inline bool
  valid(void) const;
//-------------------------- Stream operations -------------------------------
  void
  writeToStream(ArRef<AbstractOStream> output) const;

  bool // success
  readFromStream(ArRef<AbstractIStream> input);

protected:
  StlString _alias,_className;
};

// -- Inline implementations--
#include "AReViNS/remoteDepService.Ci"

} // end of namespace AReVi
#endif // end of AREVINS_REMOTEDEPSERVICE_H
//---End-Of-File---------------------------------------------------------------
