// LI2/ENIB - AReVi-Contrib - 3D file loader/saver
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/File3D/file3D.cpp
// author(s)   : C. Septseault
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "AReVi/VRML/vrmlShape3D.h"
#include "AReVi/Lib3D/material3D.h"
#include "AReVi/Shapes/shape3D.h"
#include "AReVi/Utils/urlStream.h"
#include "AReVi/Utils/fileStream.h"
#include "AReVi/Lib3D/transform3D.h"
#include "AReVi/Shapes/shapePart3D.h"

#include "AReVi/Contrib/File3D/file3D.h"
#include "AReVi/Contrib/File3D/3ds.h"
#include "AReVi/Contrib/File3D/vrmlSaver.h"
#include "AReVi/Contrib/File3D/wavefront.h"

#define FUNCTION_NOT_IMPLEMENTED(func)  func" : Function Not Implemented\n"

#define MAX_LINE_SIZE 255

namespace AReVi {

	
	
bool File3D::_failed=false;
StlString File3D::_message="";

ArRef<Shape3D>
File3D::load(const string& name)
{
	char head[8];
	_message="";
	ArRef<URLStream> input=URLStream::NEW(name);
	if(input->fail())
	{
		_message="Cannot read URL '"+name+"'";
		_failed=true;
		return Shape3D::nullRef();
	}
	input->readData(head,8);
	head[7]=0;


	if (head[0]==0x4D && head[1]==0x4D)
	{
		//3DS File
		ArRef<Shape3D> shapeOut;
		shapeOut=load3DS(name);
		return shapeOut;
	}
	
	else if (strstr(head,"#VRML")!=NULL)
	{
		ArRef<VrmlShape3D> shapeOut=VrmlShape3D::NEW();
		if(!shapeOut->parseURL(name))
		{
			_failed=true;
			_message=shapeOut->readParserMessages();
		}
		else _failed=false;
			
		return shapeOut;
	}
	else
	{
		//Wavefront file ?
		ArRef<Shape3D> shapeOut;
		shapeOut=loadWaveFront(name);
		return shapeOut;
	}
}

ArRef<Shape3D>
File3D::load3DS(const string& name)
{
	ArRef<Shape3D> shapeOut=Shape3D::NEW();
	_message="";
	ArRef<URLStream> input=URLStream::NEW(name);
	if(input->fail())
	{
		_message="Cannot read URL '"+name+"'";
		_failed=true;
		return Shape3D::nullRef();
	}
	else
	{
		Loader3DS loader3ds;
		loader3ds.setUrlDir(strDirname(name));
		if (!loader3ds.read3DSFile(input,shapeOut))
		{
			_message=loader3ds.getMessage();
			_failed=true;
			return shapeOut;
		}	
		else
		{
			_failed=false;
			return shapeOut;
		}
	}
}

ArRef<Shape3D>
File3D::loadWaveFront(const string& name)
{
	ArRef<Shape3D> shapeOut=Shape3D::NEW();
	_message="";
	ArRef<URLStream> input=URLStream::NEW(name);
	if(input->fail())
	{
		_message="Cannot read URL '"+name+"'";
		_failed=true;
		return shapeOut;
	}
	else
	{
		WFOParser parser(&input->stream());
		parser.setShape(shapeOut);
		parser.setUrlDir(strDirname(name));
			
		_failed=(!parser.parse())|| parser.fail();
		
		if (_failed)
			_message=parser.getMessage();
		return shapeOut;
	}
}

ArRef<Shape3D>
File3D::loadLightWave(const string& /*name*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("loadLightWave");
	_failed=true;
	return Shape3D::nullRef();
}

ArRef<Shape3D>
File3D::loadImagine(const string& /*name*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("loadImagine");
	_failed=true;
	return Shape3D::nullRef();
}

ArRef<Shape3D>
File3D::loadAutocad(const string& /*name*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("loadAutocad");
	_failed=true;
	return Shape3D::nullRef();
}

bool
File3D::save3DS(const string& name, ArConstRef<Shape3D> shapeIn)
{
//	_message=FUNCTION_NOT_IMPLEMENTED("save3DS");
//	return false;
#if 1// EN TEST
	ArRef<FileOStream> output=FileOStream::NEW(name,false);
	if(output->fail())
	{
		_message="Cannot write File '"+name+"'";
		_failed=true;
		return false;
	}
	else
	{
	
		Saver3DS s;
		s.save3DS(output,shapeIn);
		_failed=false;
		return true;
	}
#endif
}

bool
File3D::saveWaveFront(const string& name, ArConstRef<Shape3D> shapeIn)
{
	FILE * file=fopen(name.c_str(),"w");
	if (!file)
	{
		_message="Cannot write file \'"+name+"\'";
		_failed=true;
		return false;
	}
	
	SaverWaveFront swf;
	if (!swf.saveWaveFront(file,shapeIn))
	{
		_message=swf.getMessage();
		_failed=true;
		return false;
	}
	else
	{
		_failed=false;
		return true;
	}
	
}
bool
File3D::saveLightWave(const string& /*name*/, ArConstRef<Shape3D> /*shapeIn*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("saveLightWave");
	_failed=true;
	return false;
}
bool
File3D::saveImagine(const string& /*name*/, ArConstRef<Shape3D> /*shapeIn*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("saveImagine");
	_failed=true;
	return false;
}
bool
File3D::saveAutocad(const string& /*name*/, ArConstRef<Shape3D> /*shapeIn*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("saveAutocad");
	_failed=true;
	return false;
}

bool
File3D::saveVrml1(const string& /*name*/, ArConstRef<Shape3D> /*shapeIn*/)
{
	_message=FUNCTION_NOT_IMPLEMENTED("saveVrml1");
	_failed=true;
	return false;
}

bool
File3D::saveVrml2(const string& name, ArConstRef<Shape3D> shapeIn)
{
	ArRef<FileOStream> output=FileOStream::NEW(name,false);
	if(output->fail())
	{
		_message="Cannot write file \'"+name+"\'";
		_failed=true;
	}
	else
	{
		
		VrmlSaver vrmlSaver;
		vrmlSaver.saveVrml(output,shapeIn);
		_failed=false;
	}
		return !_failed;
}

bool
File3D::fail()
{
	return _failed;
}

StlString
File3D::readMessage()
{
	return (_message+"\n");
}

StlVector<ArRef<Shape3D> >
File3D::divideShape(ArRef<Shape3D> shape)
{
	StlVector<ArRef<Shape3D> > vec;
	ArRef<Shape3D> sh;
	ArRef<ShapePart3D> shp;
	while(shape->getNbRootParts())
	{		
		shp=shape->accessRootPart(0);
		sh=Shape3D::NEW();
		sh->addRootPart(shp);
		vec.push_back(sh);
	}
			

	return vec;
}

ArRef<Shape3D>
File3D::mergeShapes(StlVector<ArRef<Shape3D> > shapes)
{
  StlVector<ArRef<Shape3D> >::iterator itSh = shapes.begin();
  ArRef<Shape3D> sh = Shape3D::NEW();
  ArRef<Transform3D> tr = Transform3D::NEW();

  while(itSh!=shapes.end()){
    (*itSh)->readTransformation(tr);

    ArRef<Mesh3D> part = Mesh3D::NEW();

    while((*itSh)->getNbRootParts()) {
      part->addSubPart((*itSh)->accessRootPart(0));
    }

    part->writeTransformation(tr);

    sh->addRootPart(part);
    itSh++;
  }

  return sh;

}

void
File3D::smooth(ArRef<Shape3D> shape)
{
	for(size_t i=shape->getNbRootParts();i--;)
	{
		_smoothShapePart(shape->accessRootPart(i));
	}
	
}
void
File3D::_smoothShapePart(ArRef<ShapePart3D> shp)
{
	int erasedPoints=0;
	ArRef<Mesh3D> mesh=Mesh3D::nullRef();
	if (shp->getClass()->isA(Mesh3D::CLASS()) &&
		(mesh=ar_down_cast<Mesh3D>(shp)) )
	{
		for (size_t i=0;i<mesh->getVertices().size();i++)
		{
			for (size_t j=i+1;j<mesh->getVertices().size();j++)
			{
				if(mesh->getVertices()[i].x==mesh->getVertices()[j].x && mesh->getVertices()[i].y==mesh->getVertices()[j].y &&
				mesh->getVertices()[i].z==mesh->getVertices()[j].z)
				{
					mesh->accessVertices().erase(mesh->accessVertices().begin()+j);
					for (size_t k=0;k<mesh->getVertexFaces().size();k++)
					{
						if (mesh->getVertexFaces()[k].p1==j)
							mesh->accessVertexFaces()[k].p1=i;
						if (mesh->getVertexFaces()[k].p2==j)
							mesh->accessVertexFaces()[k].p2=i;
						if (mesh->getVertexFaces()[k].p3==j)
							mesh->accessVertexFaces()[k].p3=i;
						if(mesh->getVertexFaces()[k].p1>j)
							mesh->accessVertexFaces()[k].p1--;
						if(mesh->getVertexFaces()[k].p2>j)
							mesh->accessVertexFaces()[k].p2--;
						if(mesh->getVertexFaces()[k].p3>j)
							mesh->accessVertexFaces()[k].p3--;

					}
					erasedPoints++;
					j--;	
				}
				
			}
		}
		if (mesh->getTextureFaces().empty())
		{
			mesh->accessTextureFaces()=mesh->accessVertexFaces();
		}
		for (size_t i=0;i<mesh->getTextures().size();i++)
		{
			for (size_t j=i+1;j<mesh->getTextures().size();j++)
			{
				if(mesh->getTextures()[i].x==mesh->getTextures()[j].x && mesh->getTextures()[i].y==mesh->getTextures()[j].y)
				{
					mesh->accessTextures().erase(mesh->accessTextures().begin()+j);
					for (size_t k=0;k<mesh->getTextureFaces().size();k++)
					{
						if (mesh->getTextureFaces()[k].p1==j)
							mesh->accessTextureFaces()[k].p1=i;
						if (mesh->getTextureFaces()[k].p2==j)
							mesh->accessTextureFaces()[k].p2=i;
						if (mesh->getTextureFaces()[k].p3==j)
							mesh->accessTextureFaces()[k].p3=i;
						if(mesh->getTextureFaces()[k].p1>j)
							mesh->accessTextureFaces()[k].p1--;
						if(mesh->getTextureFaces()[k].p2>j)
							mesh->accessTextureFaces()[k].p2--;
						if(mesh->getTextureFaces()[k].p3>j)
							mesh->accessTextureFaces()[k].p3--;

					}
					erasedPoints++;
					j--;	
				}
				
			}
		}

		mesh->computeDefaultNormals(true);
		mesh->applyChanges(false);

	}
	for(int i=shp->getNbSubParts();i--;)
	{
		_smoothShapePart(shp->accessSubPart(i));
	}
}

} // namespace AReVi

//----------------------------------------------------------------------------
