// LI2/ENIB - AReVi-Contrib - 3D file loader/saver
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/File3D/vrmlSaver.cpp
// author(s)   : C. Septseault
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/shape3D.h"
#include "AReVi/Shapes/surface3D.h"
#include "AReVi/Lib3D/transform3D.h"
#include "AReVi/Lib3D/transform2D.h"
#include "AReVi/Utils/fdStream.h"
#include "AReVi/Lib3D/urlTexture.h"
#include "AReVi/Lib3D/material3D.h"

#include "AReVi/Contrib/File3D/vrmlSaver.h"

#define VRML_HEADER "#VRML V2.0 utf8\n\n#Created with AReVi\n\n"

#define KW_TRANSFORM "Transform {\n"
#define KW_TRANSLATION "  translation "
#define KW_ROTATION "  rotation "
#define KW_SCALE "  scale "
#define KW_CHILDREN "  children [ \n"
#define KW_SHAPE "   Shape {\n"

#define KW_END_TRANSFORM "}\n"
#define KW_END_CHILDREN "]\n"
#define KW_END_SHAPE "}\n"

//Appearance
#define KW_APPEARANCE "appearance Appearance {\n"
//MATERIAL
#define KW_MATERIAL "material Material {\n"
#define KW_AMBIENT_INTENSITY "ambientIntensity "
#define KW_DIFFUSE_COLOR "diffuseColor "
#define KW_SPECULAR_COLOR "specularColor "
#define KW_SHININESS "shininess "
#define KW_TRANSPARENCY "transparency "
#define KW_END_MATERIAL "}\n"

//Texture
#define KW_TEXTURE "texture ImageTexture {\n"
#define KW_REPEAT_S "repeatS "
#define KW_REPEAT_T "repeatT "
#define KW_END_TEXTURE "}\n"

#define KW_TEXTURE_TRANSFORM "textureTransform TextureTransform {\n"
#define KW_TEXTURE_SCALE "scale "
#define KW_TEXTURE_ROT "rotation "
#define KW_TEXTURE_TRANS "translation "
#define KW_END_TEXTURE_TRANSFORM "}\n"

#define KW_END_APPEARANCE "}\n"
//Geometry
#define KW_GEOMETRY_FACE_SET "geometry IndexedFaceSet { \n"
#define KW_COORD "coord Coordinate {\npoint [\n"
#define KW_END_COORD "]\n}\n"
#define KW_COORD_INDEX "coordIndex [ \n"
#define KW_END_COORD_INDEX "]\n"
#define KW_NORMAL "normal Normal {\nvector[\n"
#define KW_END_NORMAL "] \n}\n"
#define KW_TEX_COORD "texCoord TextureCoordinate {\n point [\n"
#define KW_END_TEX_COORD "]\n}\n"
#define KW_NORMAL_INDEX "normalIndex [\n"
#define KW_END_NORMAL_INDEX "]\n"
#define KW_TEX_COORD_INDEX "texCoordIndex [\n"
#define KW_END_TEX_COORD_INDEX "]\n"
#define KW_DOUBLE_SIDED "solid FALSE \n"
#define KW_END_GEOMETRY_FACE_SET "}\n"

namespace AReVi
{


void VrmlSaver::saveVrml(ArRef<FDOStream> arStream,ArConstRef<Shape3D> shape)
{
	ArConstRef<ShapePart3D> shp;
	ostream &stream=arStream->stream();

	stream<<VRML_HEADER;
	for (int i=shape->getNbRootParts();i--;)
	{
		shp=shape->getRootPart(i);
		saveVrmlSurface3D(stream,shp);

	}
}

void VrmlSaver::saveVrmlSurface3D(ostream& stream,ArConstRef<ShapePart3D> shp)
{
	ArConstRef<Surface3D> surface3D;
	ArRef<Transform3D> tr=Transform3D::NEW();
	shp->readTransformation(tr);

	bool identity=tr->isIdentity(true);
	if (!identity || shp->getNbSubParts())
	{
		stream<<KW_TRANSFORM;
		stream<<KW_CHILDREN;
	}
	if (shp->getClass()->isA(Surface3D::CLASS()) &&
			(surface3D= ar_down_cast<Surface3D>(shp)) )
	{
	
		if (!surface3D->getVertexFaces().empty())
		{
			stream<<KW_SHAPE;
			//Appearance

			stream<<KW_APPEARANCE;
			ArRef<Material3D> material=Material3D::NEW();
			surface3D->readMaterial(material);
			saveVrmlMaterial(stream,material);
			if (surface3D->getTexture() && surface3D->getTexture()->getClass()->isA(URLTexture::CLASS()))
				saveVrmlTexture(stream,ar_down_cast<URLTexture>(surface3D->getTexture()));
			ArRef<Transform2D> texTransform=Transform2D::NEW();
			surface3D->readTextureTransformation(texTransform);
			if (texTransform)
				saveVrmlTextureTransform(stream,texTransform);
			stream<<KW_END_APPEARANCE;
			// geometry
			saveVrmlGeometry(stream,surface3D);

			stream<<KW_END_SHAPE;
		}

	}
		if (shp->getNbSubParts())
		{
			ArConstRef<ShapePart3D> subPart;
			for (int i=shp->getNbSubParts();i--;)
			{
				saveVrmlSurface3D(stream,shp->getSubPart(i));

			}

		}
	if (!identity || shp->getNbSubParts())
	{
		stream<<KW_END_CHILDREN;
		double x,y,z,a,sx,sy,sz;
		Util3D::Base base(false);
		base.setMatrix(tr->getTransformation().matrix);
		double * matrix=base.matrix;
		sx=sqrt(matrix[Util3D::R00]*matrix[Util3D::R00]+matrix[Util3D::R01]*matrix[Util3D::R01]+matrix[Util3D::R02]*matrix[Util3D::R02]);
		sy=sqrt(matrix[Util3D::R10]*matrix[Util3D::R10]+matrix[Util3D::R11]*matrix[Util3D::R11]+matrix[Util3D::R12]*matrix[Util3D::R12]);
		sz=sqrt(matrix[Util3D::R20]*matrix[Util3D::R20]+matrix[Util3D::R21]*matrix[Util3D::R21]+matrix[Util3D::R22]*matrix[Util3D::R22]);

		matrix[Util3D::R00]*=1/sx;
		matrix[Util3D::R01]*=1/sx;
		matrix[Util3D::R02]*=1/sx;
//		matrix[Util3D::TX]*=1/sx;
		matrix[Util3D::R10]*=1/sy;
		matrix[Util3D::R11]*=1/sy;
		matrix[Util3D::R12]*=1/sy;
//		matrix[Util3D::TY]*=1/sy;
		matrix[Util3D::R20]*=1/sz;
		matrix[Util3D::R21]*=1/sz;
		matrix[Util3D::R22]*=1/sz;
//		matrix[Util3D::TZ]*=1/sz;
		
		string scale=KW_SCALE;
		scale+=doubleToStr(sx)+" "+
			doubleToStr(sy)+" "+
			doubleToStr(sz)+"\n";
		stream<<scale;

		base.extractRotation(x,y,z,a);
		if (x==y && x==z && x==0.)	
			x=1;
		string rotation=KW_ROTATION;
		rotation+=doubleToStr(x)+" "+
			doubleToStr(y)+" "+
			doubleToStr(z)+" "+
			doubleToStr(a)+"\n";
		stream<<rotation;
		base.getPosition(x,y,z);
		string translation=KW_TRANSLATION;
		translation+=doubleToStr(x)+" "+
			doubleToStr(y)+" "+
			doubleToStr(z)+"\n";
		stream<<translation;
		stream<<KW_END_TRANSFORM;
	}
}

void VrmlSaver::saveVrmlMaterial(ostream& stream,ArConstRef<Material3D> material)
{
	double r,g,b,p,t;
//TODO : Emissive & Transparency
	p=material->getAmbientIntensity();
	string ambientIntensity=KW_AMBIENT_INTENSITY;
	ambientIntensity+=doubleToStr(p)+"\n";

	material->getDiffuseColor(r,g,b);
	string diffuseColor=KW_DIFFUSE_COLOR;
	diffuseColor+=doubleToStr(r)+" "+
		doubleToStr(g)+" "+
		doubleToStr(b)+"\n";
	
	material->getSpecularColor(r,g,b);
	string specularColor=KW_SPECULAR_COLOR;
	specularColor+=doubleToStr(r)+" "+
		doubleToStr(g)+" "+
		doubleToStr(b)+"\n";
	
	p=material->getShininess();
	string shininess=KW_SHININESS;
	shininess+=doubleToStr(p)+"\n";

	t=material->getTransparency();
	string transparency=KW_TRANSPARENCY;
	transparency+=doubleToStr(t)+"\n";

	
	stream<<KW_MATERIAL;
	stream<<ambientIntensity;
	stream<<diffuseColor;
	stream<<specularColor;
	stream<<shininess;
	stream<<transparency;
	stream<<KW_END_MATERIAL;
}
void VrmlSaver::saveVrmlGeometry(ostream& stream,ArConstRef<Surface3D> surface3D)
{
	stream<<KW_GEOMETRY_FACE_SET;
	
	stream<<KW_COORD;
	saveVrmlDbl3(stream,surface3D->getVertices());
	stream<<KW_END_COORD;

	
	
	stream<<KW_COORD_INDEX;
	saveVrmlFaces(stream,surface3D->getVertexFaces());
	stream<<KW_END_COORD_INDEX;

	
	if (!surface3D->getNormals().empty())	
	{
		stream<<KW_NORMAL;
		saveVrmlDbl3(stream,surface3D->getNormals());
		stream<<KW_END_NORMAL;
	}
	
	if (!surface3D->getTextures().empty())	
	{
	stream<<KW_TEX_COORD;
	saveVrmlDbl2(stream,surface3D->getTextures());
	stream<<KW_END_TEX_COORD;
	}
	if (!surface3D->getNormalFaces().empty())	
	{
		stream<<KW_NORMAL_INDEX;
		saveVrmlFaces(stream,surface3D->getNormalFaces());
		stream<<KW_END_NORMAL_INDEX;
	}

	if (!surface3D->getTextureFaces().empty())	
	{
	stream<<KW_TEX_COORD_INDEX;
	saveVrmlFaces(stream,surface3D->getTextureFaces());
	stream<<KW_END_TEX_COORD_INDEX;
	}
	if (surface3D->getDoubleSided())	
	{
		stream<<KW_DOUBLE_SIDED;
	}

	stream<<KW_END_GEOMETRY_FACE_SET;
	
}
void VrmlSaver::saveVrmlTexture(ostream& stream,ArConstRef<URLTexture> urlTexture)
{
	if (urlTexture)
	{
		stream<<KW_TEXTURE;

		string url="url \"";
		url+=urlTexture->getURL()+"\"\n";
		stream<<url;
		stream<<KW_END_TEXTURE;
		
	}
}
void VrmlSaver::saveVrmlTextureTransform(ostream& stream,ArConstRef<Transform2D> textureTransform)
{
	if (textureTransform && !textureTransform->isIdentity(true))
	{
		const double * matrix=textureTransform->getTransformation().matrix;
		double scaleX,scaleY,transX,transY,angle;
		scaleX=sqrt(matrix[Util3D::R00]*matrix[Util3D::R00]+matrix[Util3D::R01]*matrix[Util3D::R01]+matrix[Util3D::R02]*matrix[Util3D::R02]);
		scaleY=sqrt(matrix[Util3D::R10]*matrix[Util3D::R10]+matrix[Util3D::R11]*matrix[Util3D::R11]+matrix[Util3D::R12]*matrix[Util3D::R12]);
		transX=matrix[Util3D::TX];
		transY=matrix[Util3D::TY];
		angle= atan2(matrix[Util3D::R10],matrix[Util3D::R00]);

		stream<<KW_TEXTURE_TRANSFORM;
		string scale=KW_TEXTURE_SCALE;
		scale+=doubleToStr(scaleX)+" "+doubleToStr(scaleY)+"\n";
		stream<<scale;
		string rotate=KW_TEXTURE_ROT;
		rotate+=doubleToStr(angle)+"\n";
		stream<<rotate;
		string translation=KW_TEXTURE_TRANS;
		translation+=doubleToStr(transX)+" "+doubleToStr(transY)+"\n";
		stream<<translation;
		stream<<KW_END_TEXTURE_TRANSFORM;
	}
}

void VrmlSaver::saveVrmlFaces(ostream& stream,const vector<Util3D::Face> & faces)
{
	size_t i=0;
	for (;i<faces.size()-1;i++)
	{
		stream<<doubleToStr(faces[i].p1)+", "+doubleToStr(faces[i].p2)+", "+doubleToStr(faces[i].p3)+",-1,\n";
	}
	stream<<doubleToStr(faces[i].p1)+", "+doubleToStr(faces[i].p2)+", "+doubleToStr(faces[i].p3)+",-1\n";
}
void VrmlSaver::saveVrmlDbl3(ostream& stream,const vector<Util3D::Dbl3> & double3)
{
	size_t i=0;
	for (;i<double3.size()-1;i++)
	{
		stream<<doubleToStr(double3[i].x)+" "+doubleToStr(double3[i].y)+" "+doubleToStr(double3[i].z)+",\n";
	}
	stream<<doubleToStr(double3[i].x)+" "+doubleToStr(double3[i].y)+" "+doubleToStr(double3[i].z)+"\n";
}
void VrmlSaver::saveVrmlDbl2(ostream& stream,const vector<Util3D::Dbl2> & double2)
{
	size_t i=0;
	for (;i<double2.size()-1;i++)
	{
		stream<<doubleToStr(double2[i].x)+" "+doubleToStr(double2[i].y)+",\n";
	}
	stream<<doubleToStr(double2[i].x)+", "+doubleToStr(double2[i].y)+"\n";
}

} // namespace AReVi

//----------------------------------------------------------------------------
