// LI2/ENIB - AReVi-Contrib - ArTweaker ???
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/arTweaker.h
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include <iostream>
#include <sstream>

#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/viewer3D.h"

#include "AReVi/Contrib/arTweaker.h"
#include "AReVi/Contrib/arTree.h"

namespace AReVi {

AR_CLASS_DEF(TV, ArObject)

TV::TV(ArCW& arCW)
: ArObject(arCW),
  _valueCB(thisPtr()) {
}

TV::~TV(void) {
}

ostream& operator<<(ostream& output, const ArRef<TV>& ref) {
  return output << ref->_desc;
}

void TV::_onValueChanged(void) {
  if(!_valueCB.empty()) {
    TV::ValueEvent evt;
    _valueCB.fire(evt);
  }
}

// -------------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TVBool, TV)

TVBool::TVBool(ArCW& arCW, bool& ref)
: TV(arCW),
  _data(&ref) {
}

TVBool::~TVBool() {
}

void TVBool::set(const StlString& input) {
  if((input == "true") || (input == "True") || (input == "TRUE")) {
    *_data = true;
  } else if((input == "false") || (input == "False") || (input == "FALSE")) {
    *_data = false;
  }
  _onValueChanged();
}

bool TVBool::value(void) {
  return *_data;
}

void TVBool::incr(void) {
  *_data = *_data == true ? false : true;
  _onValueChanged();
}

void TVBool::decr(void) {
  incr();
}

ostream& operator<<(ostream& output, const ArRef<TVBool>& ref) {
  output << ref->getDescription() << endl;
  output << "Type = boolean" << endl;
  output << "Value = ";
  if(*(ref->_data) == true) {
    output << "true";
  } else {
    output << "false";
  }
  return output << endl;
}

// -------------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TVChar, TV)

TVChar::TVChar(ArCW& arCW, char& ref)
: TV(arCW),
  _data(&ref) {
}

TVChar::~TVChar() {
}

void TVChar::set(const StlString& input) {
  *_data = input[0];
  _onValueChanged();
}

char TVChar::value(void) {
  return *_data;
}

void TVChar::incr(void) {
  *_data += 1;
  _onValueChanged();
}

void TVChar::decr(void) {
  *_data -= 1;
  _onValueChanged();
}

ostream& operator<<(ostream& output, const ArRef<TVChar>& ref) {
  output << ref->getDescription() << endl;
  output << "Type = char" << endl;
  output << "Value = " << ref->_data << endl;
  return output;
}

// -------------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TVInt, TV)

TVInt::TVInt(ArCW& arCW, int& ref)
: TV(arCW),
  _data(&ref),
  _incr(1),
  _low(0),
  _up(0) {
}

TVInt::TVInt(ArCW& arCW, int& ref, int incr)
: TV(arCW),
  _data(&ref),
  _incr(incr),
  _low(0),
  _up(0) {
}

TVInt::TVInt(ArCW& arCW, int& ref, int incr, int lowerBound, int upperBound)
: TV(arCW),
  _data(&ref),
  _incr(incr),
  _low(lowerBound),
  _up(upperBound) {
}

TVInt::~TVInt() {
}

void TVInt::set(const StlString& input) {
  int value;
  if(sscanf(input.c_str(), "%d", &value)==1) {
    *_data = value;
    _onValueChanged();
  }
}

int TVInt::value(void) {
  return *_data;
}

void TVInt::incr(void) {
  *_data += _incr;
  if(_low != _up) {
    *_data = CLAMP(*_data, _low, _up);
  }
  _onValueChanged();
}

void TVInt::decr(void) {
  *_data -= _incr;
  if(_low != _up) {
    *_data = CLAMP(*_data, _low, _up);
  }
  _onValueChanged();
}

ostream& operator<<(ostream& output, const ArRef<TVInt>& ref) {
  output << ref->getDescription() << endl;
  output << "Type = int" << endl;
  output << "Increment = " << ref->_incr << endl;
  if(ref->_low == ref->_up) {
    output << "Range = no range" << endl;
  } else {
    output << "Range = [" << ref->_low << " ; " << ref->_up << "]" << endl;
  }
  output << "Value = " << *(ref->_data) << endl;
  return output;
}

// -------------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TVFloat, TV)

TVFloat::TVFloat(ArCW& arCW, float& ref)
: TV(arCW),
  _data(&ref),
  _incr(1.f),
  _low(0.f),
  _up(0.f) {
}

TVFloat::TVFloat(ArCW& arCW, float& ref, float incr)
: TV(arCW),
  _data(&ref),
  _incr(incr),
  _low(0.f),
  _up(0.f) {
}

TVFloat::TVFloat(ArCW& arCW, float& ref, float incr, float lowerBound, float upperBound)
: TV(arCW),
  _data(&ref),
  _incr(incr),
  _low(lowerBound),
  _up(upperBound) {
}

TVFloat::~TVFloat() {
}

void TVFloat::set(const StlString& input) {
  float value;
  if(sscanf(input.c_str(), "%f", &value)==1) {
    *_data = value;
    _onValueChanged();
  }
}

float TVFloat::value(void) {
  return *_data;
}

void TVFloat::incr(void) {
  *_data += _incr;
  if(_low != _up) {
    *_data = CLAMP(*_data, _low, _up);
  }
  _onValueChanged();
}

void TVFloat::decr(void) {
  *_data -= _incr;
  if(_low != _up) {
    *_data = CLAMP(*_data, _low, _up);
  }
  _onValueChanged();
}

ostream& operator<<(ostream& output, const ArRef<TVFloat>& ref) {
  output << ref->getDescription() << endl;
  output << "Type = float" << endl;
  output << "Increment = " << ref->_incr << endl;
  if(ref->_low == ref->_up) {
    output << "Range = no range" << endl;
  } else {
    output << "Range = [" << ref->_low << " ; " << ref->_up << "]" << endl;
  }
  output << "Value = " << *(ref->_data) << endl;
  return output;
}

// -------------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TVDouble, TV)

TVDouble::TVDouble(ArCW& arCW, double& ref)
: TV(arCW),
  _data(&ref),
  _incr(1.0),
  _low(0.0),
  _up(0.0) {
}

TVDouble::TVDouble(ArCW& arCW, double& ref, double incr)
: TV(arCW),
  _data(&ref),
  _incr(incr),
  _low(0.0),
  _up(0.0) {
}

TVDouble::TVDouble(ArCW& arCW, double& ref, double incr, double lowerBound, double upperBound)
: TV(arCW),
  _data(&ref),
  _incr(incr),
  _low(lowerBound),
  _up(upperBound) {
}

TVDouble::~TVDouble() {
}

void TVDouble::set(const StlString& input) {
  double value;
  if(sscanf(input.c_str(), "%lf", &value)==1) {
    *_data = value;
    _onValueChanged();
  }
}

double TVDouble::value(void) {
  return *_data;
}

void TVDouble::incr(void) {
  *_data += _incr;
  if(_low != _up) {
    *_data = CLAMP(*_data, _low, _up);
  }
  _onValueChanged();
}

void TVDouble::decr(void) {
  *_data -= _incr;
  if(_low != _up) {
    *_data = CLAMP(*_data, _low, _up);
  }
  _onValueChanged();
}

ostream& operator<<(ostream& output, const ArRef<TVDouble>& ref) {
  output << ref->getDescription() << endl;
  output << "Type = double" << endl;
  output << "Increment = " << ref->_incr << endl;
  if(ref->_low == ref->_up) {
    output << "Range = no range" << endl;
  } else {
    output << "Range = [" << ref->_low << " ; " << ref->_up << "]" << endl;
  }
  output << "Value = " << *(ref->_data) << endl;
  return output;
}

// -------------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TVStlString, TV)

TVStlString::TVStlString(ArCW& arCW, StlString& ref)
: TV(arCW),
  _data(&ref) {
}

TVStlString::~TVStlString() {
}

void TVStlString::set(const StlString& input) {
  *_data = input;
  _onValueChanged();
}

StlString TVStlString::value(void) {
  return *_data;
}

ostream& operator<<(ostream& output, const ArRef<TVStlString>& ref) {
  output << ref->getDescription() << endl;
  output << "Type = string" << endl;
  output << "Value = " << *(ref->_data) << endl;
  return output;
}

// -------------------------------------------------------------------------------
// ArTweaker : Singleton
// -------------------------------------------------------------------------------

AR_CLASS_DEF(ArTweaker, ArConsole)

ArTweaker::ArTweaker(ArCW& arCW)
: ArConsole(arCW, true),
  _tree(ArTree::NEW()),
  _editing(false) {
  setBackgroundColor(0.f, 0.f, 0.f, 0.f);

  _selection = _tree->getRoot();
  _selection->setName("/");

  _buildText();
}

ArTweaker::~ArTweaker(void) {
}

// ------
// Path is in the form :
//   for a variable : /root/group/.../varIdent
//   for a group    : /root/group/.../child_group
// ------

bool
ArTweaker::insert(const StlString& path, ArRef<TV> var) {
  if(_lookup(path) != ArTreeNode::nullRef()) {
    return false;
  }

  StlString base;
  int i = path.size();
  do {
    i--;
  } while(i != 0 && path[i] != '/');
  base.assign(path, 0, i + 1);

  ArRef<ArTreeNode> parent = _lookup(base);
  if(parent == ArTreeNode::nullRef()) {
    return false;
  }

  ArRef<ArTreeNode> child;
  if(var == TV::nullRef()) {
    /** Insert a new group */
    child = ArTreeNode::NEW();
  } else {
    /** Insert a new var */
    child = ArTreeLeaf::NEW();
    ar_down_cast<ArTreeLeaf>(child)->setData(var);
  }
  child->setName(_eatBasePath(path));
  parent->insert(child);

  if(_selection == _tree->getRoot()) {
    _selection = child;
  }

  _buildText();

  return true;
}

bool
ArTweaker::erase(const StlString& path) {
  ArRef<ArTreeNode> node = _lookup(path);

  if(node == ArTreeNode::nullRef()) {
    return false;
  }

  ArRef<ArTreeNode> parent = node->parent();
  if(_selection == node) {
    _selection = parent;
  }
  node->clear();
  parent->erase(node);

  _buildText();

  return true;
}

ArRef<ArTreeNode>
ArTweaker::_lookup(const StlString& absolutePath) {
  StlVector<StlString> elems;
  _extractPath(absolutePath, elems);

  ArRef<ArTreeNode> currentNode = _tree->getRoot();

  for(int i = 0; i < (int)elems.size(); i++) {
    ArTreeNodeIter it = currentNode->begin();
    while(it != currentNode->end() && (*it)->name() != elems[i]) {
      it++;
    }

    if(it == currentNode->end()) {
      return ArTreeNode::nullRef();
    } else {
      currentNode = *it;
    }
  }

  return currentNode;
}

StlString
ArTweaker::_eatBasePath(const StlString& path) {
  StlVector<StlString> elems;
  _extractPath(path, elems);
  if(!elems.empty()) {
    return elems[elems.size() - 1];
  } else {
    return path;
  }
}

void
ArTweaker::_extractPath(const StlString& path, StlVector<StlString>& elems) {
  if(path.empty()) {
    return;
  }

  StlString currentElem;

  elems.clear();
  for(size_t i = 0; i < path.size(); i++) {
    if(path[i] == '/') {
      if(i != 0) {
	elems.push_back(currentElem);
	currentElem.clear();
      }
    } else {
      currentElem.push_back(path[i]);
    }
  }
  if(!currentElem.empty()) {
    elems.push_back(currentElem);
  }
}

void
ArTweaker::_buildText(void) {
  int sel, min, max, padding;
  min = 0;
  sel = 0;
  ostringstream output;
  ArConsoleAttr selectAttr;
  selectAttr.r = 1.f; selectAttr.g = 0.f; selectAttr.b = 0.f; selectAttr.a = 1.f;
  selectAttr.font = "";
  bool pop = false;

  ArRef<ArTreeNode> root = _tree->getRoot();

  if(!_selection.valid()) {
    if(root->size()) {
      _selection = *(root->begin());
    } else {
      _selection = root;
    }
    _editing = false;
  }

  clear();

  if(_editing) {
    ArRef<ArTreeLeaf> currentLeaf = ar_down_cast<ArTreeLeaf>(_selection);
    ArRef<TV> currentVar = ar_down_cast<TV>(currentLeaf->getData());
    output << _eatBasePath(currentLeaf->name()) << " -> ";
    if(currentVar->getClass()->isA(TVBool::CLASS())) {
      output << ar_down_cast<TVBool>(currentVar) << endl;
    } else if(currentVar->getClass()->isA(TVChar::CLASS())) {
      output << ar_down_cast<TVChar>(currentVar) << endl;
    } else if(currentVar->getClass()->isA(TVInt::CLASS())) {
      output << ar_down_cast<TVInt>(currentVar) << endl;
    } else if(currentVar->getClass()->isA(TVFloat::CLASS())) {
      output << ar_down_cast<TVFloat>(currentVar) << endl;
    } else if(currentVar->getClass()->isA(TVDouble::CLASS())) {
      output << ar_down_cast<TVDouble>(currentVar) << endl;
    } else if(currentVar->getClass()->isA(TVStlString::CLASS())) {
      output << ar_down_cast<TVStlString>(currentVar) << endl;
    } else {
      output << currentVar << endl;
    }
    output << "Input buffer : " << _buffer << endl;
    push(output.str());
  } else {
    if(_selection == root) {
      pushAttrib(selectAttr);
      push("- No variables -");
      popAttrib();
    } else {
      ArRef<ArTreeNode> parent = _selection->parent();
      ArTreeNodeIter lower = parent->begin();
      ArTreeNodeIter upper = parent->end();

      ArTreeNodeIter itSel = parent->begin();
      while(*itSel != _selection) {
	sel++;
	itSel++;
      }

      if(parent->size() > 10) {
	min = sel - 5 < 0 ? 0 : sel - 5;
	max = sel + 5 > (int)parent->size() ? parent->size() : sel + 5;
	padding = 10 - (max - min);
	min = (min == 0) ? 0 : min - padding;
	max = (max == (int)parent->size()) ? parent->size() : max + padding;

	for(int i = 0; i < min; i++) {
	  lower++;
	}

	for(int i = 0; i < (int)parent->size() - max; i++) {
	  upper--;
	}
      }

      for(ArTreeNodeIter it = lower;
	  it != upper;
	  it++) {
	ArTreeLeaf* leafPtr;
	try {
	  leafPtr = dynamic_cast<ArTreeLeaf*>(it->c_ptr());
	} catch(bad_cast) {
	  leafPtr = NULL;
	}

	if(*it == _selection) {
	  pushAttrib(selectAttr);
	  pop = true;
	}

	if(leafPtr) {
	  push("- " + (*it)->name());
	} else {
	  push("[ " + (*it)->name() + " ]");
	}

	if(pop) {
	  popAttrib();
	  pop = false;
	}
      }
    }
  }
}

void
ArTweaker::_keyboardCB(const Renderer3D::KeyboardEvent & evt) {
  ArRef<ArTreeNode> root = _tree->getRoot();

  if(!_selection.valid()) {
    if(root->size()) {
      _selection = *(root->begin());
    } else {
      _selection = root;
    }
    _editing = false;
  }

  if(evt.pressed) {
    if(evt.key == _focusKey) {
      releaseKeyboardFocus(evt);
    } else if(_editing) {
      ArRef<ArTreeLeaf> currentLeaf = ar_down_cast<ArTreeLeaf>(_selection);
      ArRef<TV> currentVar = ar_down_cast<TV>(currentLeaf->getData());

      if(evt.key == "Left" || evt.key == "Down") {
	currentVar->decr();
      } else if(evt.key == "Right" || evt.key == "Up") {
	currentVar->incr();
      } else if(evt.key[0] == 10) {
	currentVar->set(_buffer);
	_buffer.clear();
      } else if(evt.key[0] == 8) {
	if(!_buffer.empty()) {
	  _buffer.erase(_buffer.size() - 1, 1);
	}
      } else if(evt.key == "\x1b") {
	_editing = false;
      } else if(evt.key == "Prior"    ||
		evt.key == "Next"     ||
		evt.key == "Home"     ||
		evt.key == "End"      ||
		evt.key == "Begin"    ||
		evt.key == "Insert"   ||
		evt.key == "Delete"   ||
		evt.key == "Shift"    ||
		evt.key == "Control"  ||
		evt.key == "CapsLock" ||
		evt.key == "NumLock"  ||
		evt.key == "F1"       ||
		evt.key == "F2"       ||
		evt.key == "F3"       ||
		evt.key == "F4"       ||
		evt.key == "F5"       ||
		evt.key == "F6"       ||
		evt.key == "F7"       ||
		evt.key == "F8"       ||
		evt.key == "F9"       ||
		evt.key == "F10"      ||
		evt.key == "F11"      ||
		evt.key == "F12") {
	// Do nothing
      } else {
	_buffer += evt.key;
      }
    } else {
      if(_selection != root) {
	ArRef<ArTreeNode> parent = _selection->parent();
	ArTreeNodeIter it = parent->find(_selection);
	
	if(evt.key == "Up") {
	  if(it == parent->begin()) {
	    it = parent->end();
	  }
	  it--;
	  _selection = *it;
	} else if(evt.key == "Down") {
	  it++;
	  if(it == parent->end()) {
	    it = parent->begin();
	  }
	  _selection = *it;
	} else if(evt.key[0] == 10) {
	  // Enter
	  ArTreeLeaf* leafPtr;
	  try {
	    leafPtr = dynamic_cast<ArTreeLeaf*>(_selection.c_ptr());
	  } catch(bad_cast) {
	    leafPtr = NULL;
	  }
	  if(leafPtr) {
	    /** The selection was a leaf so start editing it */
	    _editing = true;
	  } else {
	    /** The selection was a node so find its first child */
	    if(_selection->size()) {
	      ArTreeNodeIter it = _selection->begin();
	      _selection = *it;
	    }
	  }
	} else if(evt.key == "\x1b") {
	  // Escape
	  if(parent != root) {
	    _selection = parent;
	  }
	}
      } // selection != root
    } // editing
  } // evt.pressed

  _buildText();
}

} // namespace AReVi

//----------------------------------------------------------------------------
