// LI2/ENIB - AReVi-Contrib - Free form deformation
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/ffdInteractor.cpp
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/arSystem.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/transform3D.h"
#include "AReVi/Shapes/mesh3D.h"
#include "AReVi/Shapes/shape3D.h"
#include "AReVi/Utils/abstractStream.h"
#include "AReVi/Contrib/ffdInteractor.h"

namespace AReVi {

static inline
double
_pow(double v, int p) {
  double result = 1.0;
  for(int i = 0; i < p; i++) {
    result *= v;
  }
  return result;
}

static inline
double
_combin(int n, int p) {
  double n_fact = 1.0;
  double p_fact = 1.0;
  double n_less_p_fact = 1.0;

  for(size_t i = n; i > 0; i--) {
    n_fact *= i;
  }

  for(size_t i = p; i > 0; i--) {
    p_fact *= i;
  }

  for(size_t i = n - p; i > 0; i--) {
    n_less_p_fact *= i;
  }

  return n_fact / (p_fact * n_less_p_fact);
}

//----------------------------------------------------------------------------
// Class FfdInteractor
//----------------------------------------------------------------------------

AR_CLASS_DEF(FfdInteractor, ArObject)

//-------- Construction / Destruction ----------------------------------------

FfdInteractor::FfdInteractor(ArCW& arCW)
: ArObject(arCW),
  _shape(),
  _parentFfd(),
  _rebuildCB(thisRef()),
  _structureChangeCB(thisRef()),
  _l(2),
  _m(2),
  _n(2),
  _combinL((double*)NULL),
  _combinM((double*)NULL),
  _combinN((double*)NULL) {
}

FfdInteractor::~FfdInteractor(void) {
}

//-------- Changes -----------------------------------------------------------

void
FfdInteractor::applyChanges(void) {
  if(_parentFfd) {
    _parentFfd->applyChanges();
  } else {
    StlVector<MeshEntry*> m;
    _rebuild(m);
    _invertLocation();
    _onRebuild();
  }
}

//-------- Deformation Box ---------------------------------------------------

void
FfdInteractor::writeDeformationBox(ArRef<BoundingBox3D> dbox) {
  _dbox = dbox->getBoundingBox();
  _updateDeformationBox();
}

void
FfdInteractor::readDeformationBox(ArRef<BoundingBox3D> dboxOut) {
  dboxOut->accessBoundingBox() = _dbox;
}

// ------- Control points ----------------------------------------------------

void
FfdInteractor::setDivision(unsigned int nbx,
			   unsigned int nby,
			   unsigned int nbz) {
  _l = nbx + 2;
  _m = nby + 2;
  _n = nbz + 2;
  _updateDeformationBox();
}

void
FfdInteractor::getDivision(unsigned int& nbx,
			   unsigned int& nby,
			   unsigned int& nbz) {
  nbx = _l - 2;
  nby = _m - 2;
  nbz = _n - 2;
}

Util3D::Dbl3&
FfdInteractor::accessControlPoint(int idx,
				  int idy,
				  int idz) {
  return _ctrlPoints[idx][idy][idz];
}

//-------- Tree structure ----------------------------------------------------

void
FfdInteractor::addSubFfdInteractor(ArRef<FfdInteractor> ffd) {
  if(ffd->_parentFfd) {
    ffd->_parentFfd->removeSubFfdInteractor(ffd);
  }
  _subFfd.push_back(ffd);
  ffd->setShape(_shape);
  ffd->_parentFfd = thisPtr();
  _onStructureChange();
}

bool
FfdInteractor::removeSubFfdInteractor(ArRef<FfdInteractor> ffd) {
  for(unsigned int i = _subFfd.size(); i--;) {
    if(_subFfd[i] == ffd) {
      StlVectorFastErase(_subFfd, i);
      ffd->_parentFfd.clear();
      _onStructureChange();
      return true;
    }
  }

  return false;
}

unsigned int
FfdInteractor::getNbSubFfdInteractors(void) const {
  return _subFfd.size();
}

ArConstRef<FfdInteractor>
FfdInteractor::getSubFfdInteractor(unsigned int index) const {
  return _subFfd[index];
}

ArRef<FfdInteractor>
FfdInteractor::accessSubFfdInteractor(unsigned int index) {
  return _subFfd[index];
}

ArConstRef<FfdInteractor>
FfdInteractor::getParentFfdInteractor(void) const {
  return _parentFfd;
}

ArRef<FfdInteractor>
FfdInteractor::accessParentFfdInteractor(void) {
  return _parentFfd;
}

//-------- Mesh management ---------------------------------------------------

void
FfdInteractor::setShape(ArConstRef<Shape3D> shape) {
  if(_parentFfd) {
    _parentFfd->setShape(shape);
  } else {
    _updateShape(shape);
  }
}

ArConstRef<Shape3D>
FfdInteractor::getShape(void) const {
  return _shape;
}

bool
FfdInteractor::addMesh(ArRef<Mesh3D> mesh) {
  if(mesh->getShape() == _shape) {
    // Store mesh
    MeshEntry entry;
    entry.mesh = mesh;
    entry.vertices = mesh->getVertices();
    entry.normals = mesh->getNormals();
    _meshes.push_back(entry);

    _computeCoordinates(_meshes.size() - 1);

    mesh->addTransformationCB(thisRef(),
			      &FfdInteractor::_onMeshTransformationCB);
    mesh->addStructureChangeCB(thisRef(),
			       &FfdInteractor::_onMeshStructureChangeCB);

    _onStructureChange();
    return true;
  } else {
    return false;
  }
}

bool
FfdInteractor::removeMesh(ArRef<Mesh3D> mesh) {
  for(unsigned int i = _meshes.size(); i--;) {
    if(_meshes[i].mesh == mesh) {
      // Restore default vertices & normals
      mesh->accessVertices() = _meshes[i].vertices;
      mesh->accessNormals() = _meshes[i].normals;
      mesh->applyChanges(true);

      mesh->removeTransformationCB(thisRef(),
				   &FfdInteractor::_onMeshTransformationCB);
      mesh->removeStructureChangeCB(thisRef(),
				    &FfdInteractor::_onMeshStructureChangeCB);

      // Erase
      StlVectorFastErase(_meshes, i);

      _onStructureChange();
      return true;
    }
  }

  return false;
}

unsigned int
FfdInteractor::getNbMeshes(void) const {
  return _meshes.size();
}

ArConstRef<Mesh3D>
FfdInteractor::getMesh(unsigned int index) const {
  return _meshes[index].mesh;
}

ArRef<Mesh3D>
FfdInteractor::accessMesh(unsigned int index) {
  return _meshes[index].mesh;
}

//-------- Serialization operations ------------------------------------------

void
FfdInteractor::extractDependencies(SerializationDependencies& /* dependenciesOut */,
				   bool /* recursive */) const {
}

void
FfdInteractor::serialize(SerializationDependencies& /* dependenciesOut */,
			 ArRef<AbstractOStream> /* output */) const {
}

bool // success
FfdInteractor::unserialize(const SerializationDependencies& /*dependencies */,
			   ArRef<AbstractIStream> /* input */) {
  return false;
}


//------- Callbacks ---------------------------------------------------------

void
FfdInteractor::_onMeshTransformationCB(const ShapePart3D::TransformationEvent& evt) {
  // Recompute mesh's vertices coordinates in deformation box location
  ArRef<Mesh3D> m = ar_down_cast<Mesh3D>(evt.source);

  for(size_t i = 0; i < _meshes.size(); i++) {
    if(_meshes[i].mesh == m) {
      _computeCoordinates(i);
      applyChanges();
      return ;
    }
  }
}

void
FfdInteractor::_onMeshStructureChangeCB(const ShapePart3D::StructureChangeEvent& evt) {
  ArRef<Mesh3D> m = ar_down_cast<Mesh3D>(evt.source);

  // If mesh is in a new shape remove it
  if(m->getShape() != _shape) {
    removeMesh(m);
    return;
  }

  // else if it has a new parent recompute its vertices coordinates
  // in deformation box location
  for(size_t i = 0; i < _meshes.size(); i++) {
    if(_meshes[i].mesh == m) {
      _computeCoordinates(i);
      applyChanges();
      return ;
    }
  }
}

//------- Utilities functions -----------------------------------------------

void
FfdInteractor::_onStructureChange(void) {
  if(!_structureChangeCB.empty()) {
    StructureChangeEvent evt;
    _structureChangeCB.fire(evt);
  }
}

void
FfdInteractor::_updateShape(ArConstRef<Shape3D> shape) {
  _shape = shape;

  // Remove mesh if they are in another shape
  for(size_t i = 0; i < _meshes.size(); i++) {
    if(_meshes[i].mesh->getShape() != _shape) {
      removeMesh(_meshes[i].mesh);
    }
  }

  // Set shape on sub ffd interactors
  for(size_t i = 0; i < _subFfd.size(); i++) {
    _subFfd[i]->_updateShape(shape);
  }
}

void
FfdInteractor::_updateDeformationBox(void) {
  // Compute new control points
  _ctrlPoints = StlVector<StlVector<StlVector<Util3D::Dbl3> > >(_l, StlVector<StlVector<Util3D::Dbl3> >(_m, StlVector<Util3D::Dbl3>(_n)));
  for(int i = 0; i < _l; i++) {
    for(int j = 0; j < _m; j++) {
      for(int k = 0; k < _n; k++) {
	_ctrlPoints[i][j][k] = Util3D::Dbl3(_dbox.xmin + i * _dbox.sizeX() / (_l - 1),
					    _dbox.ymin + j * _dbox.sizeY() / (_m - 1),
					    _dbox.zmin + k * _dbox.sizeZ() / (_n - 1));

      }
    }
  }

  // Compute combinaisons
  if(_combinL) delete [] _combinL;
  if(_combinM) delete [] _combinM;
  if(_combinN) delete [] _combinN;

  _combinL = new double[_l];
  _combinM = new double[_m];
  _combinN = new double[_n];

  for(int i = 0; i < _l; i++) {
    _combinL[i] = _combin(_l - 1, i);
  }

  for(int i = 0; i < _m; i++) {
    _combinM[i] = _combin(_m - 1, i);
  }

  for(int i = 0; i < _n; i++) {
    _combinN[i] = _combin(_n - 1, i);
  }
}

void
FfdInteractor::_computeCoordinates(unsigned int index) {
  ArRef<Transform3D> tr = Transform3D::NEW();
  _meshes[index].mesh->readFlatTransformation(tr);

  _meshes[index].transformedVertices.clear();

  Util3D::Dbl3 out;
  for(size_t i = 0; i < _meshes[index].vertices.size(); i++) {
    // Express vertex in shape location
    Util3D::Dbl3& in = _meshes[index].vertices[i];
    Util3D::Dbl3 tmp;
    tr->apply(tmp.x, tmp.y, tmp.z, in.x, in.y, in.z);

    // Then in deformation box location
    out.x = tmp.x;
    out.y = tmp.y;
    out.z = tmp.z;

    // Store it
    _meshes[index].transformedVertices.push_back(out);
  }
}

void
FfdInteractor::_rebuild(StlVector<MeshEntry*>& meshes) {
  // First rebuild chidlren
  for(size_t i = 0; i < _subFfd.size(); i++) {
    _subFfd[i]->_rebuild(meshes);
  }

  for(size_t i = 0; i < _meshes.size(); i++) {
    _meshes[i].verticesPtr = &(_meshes[i].transformedVertices);
    meshes.push_back(&(_meshes[i]));
  }

  // Then apply ffd on all meshes
  for(size_t a = 0; a < meshes.size(); a++) {
    StlVector<Util3D::Dbl3>& v = *(meshes[a]->verticesPtr);
    StlVector<Util3D::Dbl3>& mv = meshes[a]->mesh->accessVertices();

    for(size_t b = 0; b < v.size(); b++) {
      Util3D::Dbl3 dv = v[b];

      double s, t, u;

      s = (dv.x - _dbox.xmin) / _dbox.sizeX();
      t = (dv.y - _dbox.ymin) / _dbox.sizeY();
      u = (dv.z - _dbox.zmin) / _dbox.sizeZ();

      dv.x = s;
      dv.y = t;
      dv.z = u;

      Util3D::Dbl3  cv = Util3D::Dbl3(0.0, 0.0, 0.0);
	
      for(int i = 0; i < _l; i++) {
	//double tmp = _combin(_l - 1, i) * _pow(1.0 - dv.x, _l - 1 - i) * _pow(dv.x, i);
	double tmp = _combinL[i] * _pow(1.0 - dv.x, _l - 1 - i) * _pow(dv.x, i);
	Util3D::Dbl3 dva = Util3D::Dbl3(0.0, 0.0, 0.0);

	for(int j = 0; j < _m; j++) {
	  //double tmp = _combin(_m - 1, j) * _pow(1.0 - dv.y, _m - 1 - j) * _pow(dv.y, j);
	  double tmp = _combinM[j] * _pow(1.0 - dv.y, _m - 1 - j) * _pow(dv.y, j);
	  Util3D::Dbl3 dvb = Util3D::Dbl3(0.0, 0.0, 0.0);
	
	  for(int k = 0; k < _n; k++) {
	    //double tmp = _combin(_n - 1, k) * _pow(1.0 - dv.z, _n - 1 - k) * _pow(dv.z, k);
	    double tmp = _combinN[k] * _pow(1.0 - dv.z, _n - 1 - k) * _pow(dv.z, k);
	    dvb.x += tmp * _ctrlPoints[i][j][k].x;
	    dvb.y += tmp * _ctrlPoints[i][j][k].y;
	    dvb.z += tmp * _ctrlPoints[i][j][k].z;
	  }

	  dva.x += tmp * dvb.x;
	  dva.y += tmp * dvb.y;
	  dva.z += tmp * dvb.z;
	}
	
	cv.x += tmp * dva.x;
	cv.y += tmp * dva.y;
	cv.z += tmp * dva.z;
      }

      mv[b] = Util3D::Dbl3(cv.x, cv.y, cv.z);
    }
    meshes[a]->verticesPtr = &(meshes[a]->mesh->accessVertices());
  }
}

void
FfdInteractor::_onRebuild(void) {
  if(!_rebuildCB.empty()) {
    RebuildEvent evt;
    _rebuildCB.fire(evt);
  }
}

void
FfdInteractor::_invertLocation(void) {
  for(size_t i = 0; i < _meshes.size(); i++) {
    Util3D::Transform tr = _meshes[i].mesh->getFlatTransformation();
    tr.invert();

    // Convert new mesh coordinates from shape location to mesh location
    StlVector<Util3D::Dbl3>& v = _meshes[i].mesh->accessVertices();
    for(size_t j = v.size(); j--;) {
      Util3D::Dbl3 tmp;
      tr.apply(tmp.x, tmp.y, tmp.z, v[j].x, v[j].y, v[j].z);

      v[j].x = tmp.x;
      v[j].y = tmp.y;
      v[j].z = tmp.z;
    }

    // Apply changes
    _meshes[i].mesh->computeDefaultNormals(true);
    _meshes[i].mesh->applyChanges(true);
  }

  for(size_t i = 0; i < _subFfd.size(); i++) {
    _subFfd[i]->_invertLocation();
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
