// LI2/ENIB - AReVi-Contrib - Nurbs shapes
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/nurbsCurve3d.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Contrib/nurbsCurve3D.h"

#include "AReVi/Lib3D/gl_priv.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class NurbsCurve3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(NurbsCurve3D,ShapePart3D)

//-------- Construction / Destruction ----------------------------------------

NurbsCurve3D::NurbsCurve3D(ArCW & arCW) // Protected
: ShapePart3D(arCW),
  _thickness(1),
  _precision(10),
  _curve(),
  _color(0.8,0.8,0.8),
  _vertices()
{
ArNurbs::makeLine(4,2,Vector3d(0,0,0),Vector3d(1,0,0),_curve);
_describeGeometry(_bbox,
                  (const Util3D::Dbl3 *)0,0,
                  (const Util3D::Face *)0,0,
                  0,0);
}

NurbsCurve3D::~NurbsCurve3D(void) // Protected
{
  // Nothing to be done
}

//-------- Changes -----------------------------------------------------------

bool // success
NurbsCurve3D::applyChanges(bool checkConsistency)
{
if(checkConsistency)
  {
  if(!_curve.valid())
    {
    return(false);
    }
  }
_postGeometryRebuild();
return(true);
}

//-------- Appearance --------------------------------------------------------

bool // success
NurbsCurve3D::setThickness(unsigned int thickness)
{
if(thickness>0)
  {
  _thickness=thickness;
  _postAppearanceRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // lines thickness
NurbsCurve3D::getThickness(void) const
{
return(_thickness);
}

bool // success
NurbsCurve3D::setPrecision(unsigned int precision)
{
if(precision>0)
  {
  _precision=precision;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // nurbs interpolation precision
NurbsCurve3D::getPrecision(void) const
{
return(_precision);
}

//-------- Vertices ----------------------------------------------------------

const ArNurbs::Curve3d & // nurbs curve
NurbsCurve3D::getCurve(void) const
{
return(_curve);
}

ArNurbs::Curve3d & // nurbs curve
NurbsCurve3D::accessCurve(void)
{
return(_curve);
}

//-------- Color -------------------------------------------------------------

void
NurbsCurve3D::setColor(double r,
                       double g,
                       double b)
{
_color.x=r;
_color.y=g;
_color.z=b;
_postGeometryRebuild();
}

void
NurbsCurve3D::getColor(double & rOut,
                       double & gOut,
                       double & bOut) const
{
rOut=_color.x;
gOut=_color.y;
bOut=_color.z;
}

//-------- Serialization operations ------------------------------------------

void
NurbsCurve3D::serialize(SerializationDependencies & dependenciesOut,
                        ArRef<AbstractOStream> output) const
{
ShapePart3D::serialize(dependenciesOut,output);
output->writeUInt(_precision);
output->writeDouble(_color.x);
output->writeDouble(_color.y);
output->writeDouble(_color.z);
_curve.writeToStream(output);
}

bool // success
NurbsCurve3D::unserialize(const SerializationDependencies & dependencies,
                          ArRef<AbstractIStream> input)
{
unsigned int precision;
ArNurbs::Curve3d curve;
Util3D::Dbl3 color;
if(!ShapePart3D::unserialize(dependencies,input)||
   !input->readUInt(precision)||
   !input->readDouble(color.x)||
   !input->readDouble(color.y)||
   !input->readDouble(color.z)||
   !curve.readFromStream(input))
  {
  return(false);
  }
_precision=precision;
_curve=curve;
_color=color;
_postGeometryRebuild();
_postAppearanceRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
NurbsCurve3D::_setAppearance(void) const // Protected
{
if(_offset)
  {
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(_offset,_offset);
  }
glDisable(GL_LIGHTING);
glLineWidth(_thickness);
}

void
NurbsCurve3D::_resetAppearance(void) const // Protected
{
glLineWidth(1.0);
glEnable(GL_LIGHTING);
if(_offset)
  {
  glDisable(GL_POLYGON_OFFSET_FILL);
  }
}

void
NurbsCurve3D::_drawGeometry(void) const // Protected
{
glColor3f(_color.x,_color.y,_color.z); // curve color
glBegin(GL_LINE_STRIP);
for(unsigned int i=_vertices.size();i--;)
  {
  glVertex3f(_vertices[i].x,_vertices[i].y,_vertices[i].z);
  }
glEnd(); // GL_LINE_STRIP
}

void
NurbsCurve3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  _vertices.clear();
  Util3D::BoundingBox bbox;
  if(_curve.valid())
    {
    double start=_curve.begin();
    double step=(_curve.end()-start)/(_precision-1);
    for(unsigned int l=0;l<_precision;l++)
      {
      Vector3d vertex;
      _curve.computePoint(start+(step*l),vertex);
      _vertices.push_back(Util3D::Dbl3(vertex.x(),vertex.y(),vertex.z()));
      bbox.addDbl3(_vertices.back());
      }
    }
  if(_useDisplayList)
    {
    glNewList(_geometryList,GL_COMPILE);
    _drawGeometry();
    glEndList();
    }
  _describeGeometry(bbox,
                    (const Util3D::Dbl3 *)0,0,
                    (const Util3D::Face *)0,0,
                    _precision-1,0);
  _needsGeometryRebuild=false;
  }
if(_needsAppearanceRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_appearanceGeometryList,GL_COMPILE);
    _setAppearance();
    glCallList(_geometryList);
    _resetAppearance();
    glEndList();
    }
  _needsAppearanceRebuild=false;
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
