// LI2/ENIB - AReVi-Contrib - Nurbs shapes
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/nurbsSurface3D.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Contrib/nurbsSurface3D.h"

#include "AReVi/Utils/abstractStream.h"

#include "AReVi/arSystem.h"
#include "AReVi/Lib3D/llw_priv.h"
#include "AReVi/Lib3D/material3D.h"
#include "AReVi/Lib3D/texture.h"
#include "AReVi/Lib3D/transform2D.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class NurbsSurface3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(NurbsSurface3D,Surface3D)

//-------- Construction / Destruction ----------------------------------------

NurbsSurface3D::NurbsSurface3D(ArCW & arCW) // Protected
: Surface3D(arCW),
  _precisionU(10),
  _precisionV(10),
  _surface()
{
  // Nothing to be done
}

NurbsSurface3D::~NurbsSurface3D(void) // Protected
{
  // Nothing to be done
}

//-------- Changes -----------------------------------------------------------

bool // success
NurbsSurface3D::applyChanges(bool checkConsistency)
{
if(checkConsistency)
  {
  if(!_surface.valid())
    {
    return(false);
    }
  }
_postGeometryRebuild();
return(true);
}

//-------- Appearance --------------------------------------------------------

bool // success
NurbsSurface3D::setPrecision(unsigned int precisionU,
                             unsigned int precisionV)
{
if(precisionU&&precisionV)
  {
  _precisionU=precisionU;
  _precisionV=precisionV;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

void
NurbsSurface3D::getPrecision(unsigned int &precisionUOut,
                             unsigned int &precisionVOut) const
{
precisionUOut=_precisionU;
precisionVOut=_precisionV;
}

//-------- Vertices ----------------------------------------------------------

const ArNurbs::Surface3d & // nurbs surface
NurbsSurface3D::getSurface(void) const
{
return(_surface);
}

ArNurbs::Surface3d & // nurbs surface
NurbsSurface3D::accessSurface(void)
{
return(_surface);
}

//-------- Serialization operations ------------------------------------------

void
NurbsSurface3D::serialize(SerializationDependencies & dependenciesOut,
                          ArRef<AbstractOStream> output) const
{
Surface3D::serialize(dependenciesOut,output);
output->writeUInt(_precisionU);
output->writeUInt(_precisionV);
_surface.writeToStream(output);
}

bool // success
NurbsSurface3D::unserialize(const SerializationDependencies & dependencies,
                            ArRef<AbstractIStream> input)
{
unsigned int precisionU,precisionV;
ArNurbs::Surface3d surface;
if(!Surface3D::unserialize(dependencies,input)||
   !input->readUInt(precisionU)||
   !input->readUInt(precisionV)||
   !surface.readFromStream(input))
  {
  return(false);
  }
_surface=surface;
setPrecision(precisionU,precisionV);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
NurbsSurface3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild&&_surface.valid())
  {
  if((_vertices.size()!=_precisionU*_precisionV)|| // check mesh size change
     (!_facesV.empty()&&(_facesV[0].p3!=_precisionU+1))) // check alignment
    {
    _vertices.resize(_precisionU*_precisionV);
    _normals.resize(_precisionU*_precisionV);
    _textures.resize(_precisionU*_precisionV);
    _facesV.resize((_precisionU-1)*2*(_precisionV-1));
    for(unsigned faceIndex=0, index=0, v=0;v<_precisionV-1;v++)
      {
      for (unsigned u=0;u<_precisionU-1;u++,index++)
        {
        _facesV[faceIndex]=Util3D::Face(index,
                                        index+1,
                                        index+_precisionU+1);
        faceIndex++;
        _facesV[faceIndex]=Util3D::Face(index,
                                        index+_precisionU+1,
                                        index+_precisionU);
        faceIndex++;
        }
      index++;
      }
    }
  Vector3d vertex,normal;
  double startU=_surface.beginU();
  double stepU=(_surface.endU()-startU)/(_precisionU-1);
  double startV=_surface.beginV();
  double stepV=(_surface.endV()-startV)/(_precisionV-1);
  int vspan;
  for(unsigned v=0,index=0;v<_precisionV;v++)
    {
    double v0=startV+(stepV*v);
    vspan=_surface.findSpanV(v0);
    for(unsigned u=0;u<_precisionU;u++,index++)
      {
      double u0=startU+(stepU*u);
      _textures[index].x=u0;
      _textures[index].y=v0;
      _surface.computeVertice(u0,
                              _surface.findSpanU(u0),
                              v0, vspan,
                              vertex, normal);
      _vertices[index].x=vertex.x();
      _vertices[index].y=vertex.y();
      _vertices[index].z=vertex.z();
      _normals[index].x=normal.x();
      _normals[index].y=normal.y();
      _normals[index].z=normal.z();
      }
    }
  }
Surface3D::_rebuildIfNeeded();
}

} // namespace AReVi

//----------------------------------------------------------------------------
