// LI2/ENIB - AReVi-Contrib - Path ???
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/path.cpp
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Contrib/path.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

#define PATH_MAX_ITER  10
#define PATH_PRECISION 0.001

AR_CLASS_DEF(Path, ArObject)

Path::Path(ArCW& arCW)
: ArObject(arCW),
  _name(),
  _precision(PATH_PRECISION),
  _movingBase(Base3D::NEW()),
  _sd(Vector3d(1.0, 1.0, 1.0)),
  _ed(Vector3d(- 1.0, - 1.0, - 1.0)) {
}

Path::~Path(void) {
}

void
Path::applyChanges(void) {
  _points.resize(_wayPoints.size());
  _orientations.resize(_wayPoints.size());

  if(_points.empty()) {
    return;
  }

  while(_autoOrientations.size() < _wayPoints.size()) {
    _autoOrientations.push_back(true);
  }

  // Update positions
  for(size_t i = 0; i < _wayPoints.size(); i++) {
    double x, y, z;
    _wayPoints[i]->getPosition(x, y, z);
    _points[i] = Vector3d(x, y, z);
  }

  _curve.resize(_points.size() + 2, 3);
  _computeParameters();
  _computeKnotVector();

  StlVector<Vector3d>& ctrlPoints = _curve.accessCtrlPoints();
  StlVector<double>& knotVector = _curve.accessKnotVector();
  StlVector<double>& weightVector = _curve.accessWeights();
  /*
  for(size_t i = 0; i < weightVector.size(); i++) {
    weightVector[i] = 1.0;
  }
  */
  weightVector.clear();

  ctrlPoints[0] = _points.front();
  ctrlPoints[1] = knotVector[4] * _sd / 3 + _points[0];
  ctrlPoints[_points.size() + 1] = _points.back();
  ctrlPoints[_points.size()] = ctrlPoints[_points.size() + 1] -
    _ed * (1.0 - knotVector[_points.size() + 1]) / 3;

  if(_wayPoints.size() > 3) {
    _solveTridiagonal(ctrlPoints);
  }

  // Update orientations
  for(size_t i = 0; i < _wayPoints.size(); i++) {
    if(_autoOrientations[i]) {
      Vector3d d;
      _curve.computeDerive(_parameters[i], d);
      d.normalize();
      Vector3d axis = Vector3d(1.0, 0.0, 0.0) ^ d;
      double cosAngle = Vector3d(1.0, 0.0, 0.0) * d;
      double sinAngle = axis.length();
      double angle = atan2(sinAngle, cosAngle);
      axis.normalize();
      _orientations[i] = Quaterniond(axis, angle);
      //_wayPoints[i]->setRotation(axis.x(), axis.y(), axis.z(), angle);
      _wayPoints[i]->accessBase().setRotation(axis.x(), axis.y(), axis.z(), angle);
    } else {
      double roll, pitch, yaw;
      _wayPoints[i]->extractOrientation(roll, pitch, yaw);
      _orientations[i] = Quaterniond(roll, pitch, yaw);
    }
  }

  _precompLength();
}

void
Path::setName(const StlString& name) {
  _name = name;
}

const StlString&
Path::getName(void) const {
  return _name;
}

const StlVector<ArRef<Base3D> >&
Path::getWayPoints(void) const {
  return _wayPoints;
}

StlVector<ArRef<Base3D> >&
Path::accessWayPoints(void) {
  return _wayPoints;
}

const StlVector<bool>&
Path::getAutoOrientations(void) const {
  return _autoOrientations;
}

StlVector<bool>&
Path::accessAutoOrientations(void) {
  return _autoOrientations;
}

void
Path::setStartDerivate(const Vector3d& d) {
  if(d.length()) {
    _sd = d;
  }
}

Vector3d
Path::getStartDerivate(void) const {
  return _sd;
}

void
Path::setEndDerivate(const Vector3d& d) {
  if(d.length()) {
    _ed = d;
  }
}

Vector3d
Path::getEndDerivate(void) const {
  return _ed;
}

double
Path::getLength(void) const {
  return _pLength.back();
}

double
Path::getArcLength(double p, bool fromEnd) {
  if(fromEnd) {
        return (_pLength.back() -_fastArcLength(0.0,p,_precision));
	//return _fastArcLength(p, 1.0, _precision);
  } else {
    return _fastArcLength(0.0, p, _precision);
  }
}

double
Path::getParameter(double length, bool fromEnd) {
  if(fromEnd) {
    return _getP(_pLength.back() - length);
  } else {
    return _getP(length);
  }
}

double
Path::getPrecision(void) const {
  return _precision;
}

void
Path::setPrecision(double p) {
  _precision = p;
}

ArConstRef<Base3D>
Path::getMovingBase(void) const {
  return _movingBase;
}

ArRef<Base3D>
Path::accessMovingBase(void) {
  return _movingBase;
}

ArConstRef<Base3D>
Path::evaluate(double p) {
  if(_wayPoints.empty()) {
    // identity
    _movingBase->setPosition(0.0, 0.0, 0.0);
    _movingBase->cancelRotation();
  } else if(_wayPoints.size() == 1) {
    // front
    _movingBase->setLocation(_wayPoints.front());
  } else {
    if(p <= 0.0) {
      // front
      _movingBase->setLocation(_wayPoints.front());
    } else if(p >= 1.0) {
      // back
      _movingBase->setLocation(_wayPoints.back());
    } else {
      pair<Vector3d, Quaterniond> r = _evaluate(p);

      // Update interpolated base
      Vector3d axis;
      double angle=0.0;
      r.second.getAxisAngle(axis, angle);

      _movingBase->setPosition(r.first.x(),
			       r.first.y(),
			       r.first.z());
      _movingBase->setRotation(axis.x(),
			       axis.y(),
			       axis.z(),
			       angle);
    }
  }

  return _movingBase;
}

ArNurbs::Curve3d&
Path::getCurve(void) {
  return _curve;
}

void
Path::serialize(SerializationDependencies& dependenciesOut,
		ArRef<AbstractOStream> output) const {
  ArObject::serialize(dependenciesOut, output);

  output->writeString(_name);
  output->writeDouble(_precision);
  output->writeDouble(_sd.x());
  output->writeDouble(_sd.y());
  output->writeDouble(_sd.z());
  output->writeDouble(_ed.x());
  output->writeDouble(_ed.y());
  output->writeDouble(_ed.z());

  output->writeUInt(_wayPoints.size());
  for(size_t i = 0; i < _wayPoints.size(); i++) {
    output->writeUInt(dependenciesOut.addReference(_wayPoints[i]));
    _wayPoints[i]->serialize(dependenciesOut, output);
    output->writeChar(_autoOrientations[i]);
  }
}

bool
Path::unserialize(const SerializationDependencies& dependencies,
		  ArRef<AbstractIStream> input) {
  if(!ArObject::unserialize(dependencies, input)) return false;

  StlString name;
  if(!input->readString(name)) return false;

  double precision;
  if(!input->readDouble(precision)) return false;

  double sdx, sdy, sdz;
  if(!input->readDouble(sdx) ||
     !input->readDouble(sdy) ||
     !input->readDouble(sdz)) return false;

  double edx, edy, edz;
  if(!input->readDouble(edx) ||
     !input->readDouble(edy) ||
     !input->readDouble(edz)) return false;

  unsigned int nbWayPoints;
  if(!input->readUInt(nbWayPoints)) return false;

  StlVector<ArRef<Base3D> > wayPoints;
  StlVector<bool> autoOrientations;
  for(size_t i = 0; i < nbWayPoints; i++) {
    unsigned int wayPointID;
    if(!input->readUInt(wayPointID)) return false;
    ArRef<Base3D> tmpWayPoint =
      ar_down_cast<Base3D>(dependencies.findReference(wayPointID));
    if(!tmpWayPoint->unserialize(dependencies, input)) return false;
    wayPoints.push_back(tmpWayPoint);
    char ao;
    if(!input->readChar(ao)) return false;
    autoOrientations.push_back((bool)ao);
  }

  _name = name;
  _precision = precision;
  _sd = Vector3d(sdx, sdy, sdz);
  _ed = Vector3d(edx, edy, edz);
  _wayPoints = wayPoints;
  _autoOrientations = autoOrientations;

  applyChanges();

  return true;
}

void
Path::_computeParameters(void) {
  // Compute total chord length
  double d = 0.0;
  StlVector<double> dist;
  for(size_t i = 1; i < _points.size(); i++) {
    Vector3d t = _points[i] - _points[i - 1];
    dist.push_back(t.length());
    d += dist.back();
  }

  // Compute parameters
  _parameters.resize(_points.size());
  _parameters.front() = 0.0;
  _parameters.back()  = 1.0;
  for(size_t i = 1; i < _points.size() - 1; i++) {
    _parameters[i] = _parameters[i - 1] + dist[i - 1] / d;
  }
}

void
Path::_computeKnotVector(void) {
  StlVector<double>& knotVector = _curve.accessKnotVector();

  for(size_t i = 0; i < 4; i++) {
    knotVector[i] = 0.0;
    knotVector[_points.size() + 2 + i] = 1.0;
  }

  for(size_t j = 1; j < _points.size(); j++) {
    knotVector[j + 3] = _parameters[j];
  }
}

void
Path::_solveTridiagonal(StlVector<Vector3d>& ctrlPoints) {
  size_t n = _points.size() - 1;

  StlVector<Vector3d> R(n + 1);
  StlVector<double>   dd(n + 1);
  StlVector<double>   abc(4);
  StlVector<double>&  knotVector = _curve.accessKnotVector();
  double den;

  for(size_t i = 3; i < n; i++) {
    R[i] = _points[i - 1];
  }

  ArNurbs::basisFuns(4, knotVector[4], 3, knotVector, abc);
  den = abc[1];
  ctrlPoints[2] = (_points[1] - abc[0] * ctrlPoints[1]) / den;

  for(size_t i = 3; i < n; i++) {
    dd[i] = abc[2] / den;
    ArNurbs::basisFuns(i + 2, knotVector[i + 2], 3, knotVector, abc);
    den = abc[1] - abc[0] * dd[i];
    ctrlPoints[i] = (R[i] - abc[0] * ctrlPoints[i - 1]) / den;
  }

  dd[n] = abc[2] / den;
  ArNurbs::basisFuns(n + 2, knotVector[n + 2], 3, knotVector, abc);
  den = abc[1] - abc[0] * dd[n];
  ctrlPoints[n] = (_points[n - 1] - abc[2] * ctrlPoints[n + 1]
		   - abc[0] * ctrlPoints[n - 1]) / den;

  for(size_t i = n - 1; i >= 2; i--) {
    ctrlPoints[i] = ctrlPoints[i] - dd[i + 1] * ctrlPoints[i + 1];
  }
}

pair<Vector3d, Quaterniond>
Path::_evaluate(double p) {
  // Compute Position
  Vector3d pos;
  _curve.computePoint(_getP(p * _pLength.back()), pos);

  // Find orientations to interpolate
  size_t i = 1;
  while(p > _parameters[i]) {
    i++;
  }
  double p2 = (p - _parameters[i - 1]) /
    (_parameters[i] - _parameters[i - 1]);
  Quaterniond o = slerp(_orientations[i - 1],
			_orientations[i],
			p2);

  return make_pair<Vector3d, Quaterniond>(pos, o);
}

void
Path::_precompLength(void) {
  _pLength.clear();

  double l = 0.0;
  double lowBound, upperBound;
  double step = _precision;

  lowBound = 0.0;
  upperBound = lowBound + step;

  while(upperBound <= 1.0) {
    l += _arcLength(lowBound, upperBound, step);
    _pLength.push_back(l);

    lowBound = upperBound;
    upperBound += step;
  }
}

double
Path::_arcLength(double a, double b, double step) {
  // Compute arc length between a and b
  double l = 0.0;

  Vector3d v0, v1;
  _curve.computePoint(a, v0);

  while(a < b) {
    a += step;
    _curve.computePoint(a, v1);
    l += (v1 - v0).length();
    v0 = v1;
  }

  _curve.computePoint(b, v1);
  l += (v1 - v0).length();

  return l;
}

double
Path::_fastArcLength(double a, double b, double step) {
     double x = fmod(a, _precision);
     double y = fmod(b, _precision);

     unsigned int ni, nj, nk, nl;
     ni = (int)((a - x) / _precision);
     nj = ni + 1;
     nk = (int)((b - y) / _precision);
     nl = nk + 1;

     if(ni>_pLength.size() ||
        nk>_pLength.size()
       ) return 0.0;

     double low, middle, high;

     if(a > (ni+0.5) * _precision ) {
       low = _arcLength(a,nj * _precision, step);
     } else {
       low = _pLength[nj] - _pLength[ni] - _arcLength(ni * _precision,a, step);
     }

     middle = _pLength[nk] - _pLength[nj];

     if(b >  (nk+0.5) * _precision ) {
         high = _pLength[nl] - _pLength[nk] - _arcLength(b, nl * _precision, step);
     } else {
       high = _arcLength(nk * _precision, b, step);
     }

     return low + middle + high;
}

double
Path::_getP(double s) {
  double p = s / _pLength.back();
  for(size_t i = 0; i < PATH_MAX_ITER; i++) {
    double f = _fastArcLength(0.0, p, _precision) - s;
    if(fabs(f) < _precision) {
      return p;
    } else {
      double df = _speed(p);
      p -= f / df;
    }
  }
  return p;
}

double
Path::_speed(double s) {
  Vector3d d;
  _curve.computeDerive(s, d);
  return d.length();
}

} // namespace AReVi

