// LI2/ENIB - AReVi-Contrib - Sound player
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/soundPlayer.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Contrib/soundPlayer.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class SoundPlayer
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(SoundPlayer,ArObject)

bool SoundPlayer::_initSuccess=false;


void * ( *SoundPlayer::_openSoundOutputFnct)(SoundPlayer *, const StlString &, int &)=
  (void *( *)(SoundPlayer *, const StlString &, int &))0;

void (* SoundPlayer::_closeSoundOutputFnct)(void *)=
  (void (* )(void *))0;

bool (* SoundPlayer::_setPlayFnct)(void *, bool )=
  (bool (* )(void *, bool ))0;

bool (* SoundPlayer::_isPlayingFnct)(const void *)=
  (bool (* )(const void *))0;

bool (* SoundPlayer::_setOutputVolumeFnct)(void *)=
  (bool (* )(void *))0;

int (* SoundPlayer::_pushBufferFnct)(void *, ArConstRef<MemoryBlock> )=
  (int (* )(void *, ArConstRef<MemoryBlock> ))0;

unsigned (* SoundPlayer::_popBuffersFnct)(void *, bool all)=
  (unsigned (* )(void *, bool all))0;

unsigned (* SoundPlayer::_getNbBuffersFnct)(const void *)=
  (unsigned (* )(const void *))0;

bool (* SoundPlayer::_setCurrentBufferFnct)(void *, int )=
  (bool (* )(void *, int ))0;

int (* SoundPlayer::_getCurrentBufferFnct)(const void *)=
  (int (* )(const void *))0;

unsigned int  (* SoundPlayer::_getDurationFnct)(const void *)=
  (unsigned int  (* )(const void *))0;

bool           (* SoundPlayer::_setTimeFnct)(void *, unsigned int )=
  (bool           (* )(void *, unsigned int ))0;

unsigned int  (* SoundPlayer::_getTimeFnct)(const void *)=
  (unsigned int  (* )(const void *))0;


//-------- Construction / Destruction ----------------------------------------

SoundPlayer::SoundPlayer(ArCW & arCW,
                         UtilSound::SoundFormat format,
                         unsigned int rate) // Protected
: ArObject(arCW),
  _deviceID(-1),
  _pluginData(NULL),
  _playing(false),
  _soundFormat(format),
  _sampleRate(rate),
  _autoErase(true),
  _leftVolume(1.0),
  _rightVolume(1.0),
  _seekCB(thisRef()),
  _setupCB(thisRef()),
  _controlCB(thisRef())
{
if(!SoundPlayer::_initSuccess)
  {
  setErrorMessage("Plugin not initialized");
  }
else
  {
  _pluginData=((*SoundPlayer::_openSoundOutputFnct)(thisRef().c_ptr(), "", _deviceID));

  if ((!_pluginData) || _deviceID<0)
    {
    _deviceID=-1;
    _pluginData=NULL;
    setErrorMessage("Can not open audio device for output");
    }
  else
    {
    ArSystem::setBlockingFD(_deviceID,false);
    Scheduler::access()->addPollingCB(thisRef(),&SoundPlayer::_onPolling);
    ArSystem::addReadFDPolling(_deviceID);
    }
  }
}

SoundPlayer::~SoundPlayer(void) // Protected
{
if(SoundPlayer::_initSuccess)
  {
  Scheduler::access()->removePollingCB(thisRef(),&SoundPlayer::_onPolling);
  ArSystem::removeReadFDPolling(_deviceID);

  if (_deviceID!=-1)
    {
    ((*SoundPlayer::_closeSoundOutputFnct)(_pluginData));
    _deviceID=-1;
    _pluginData=NULL;
    }
  }
}

//-------- Sound properties --------------------------------------------------

UtilSound::SoundFormat // sound format
SoundPlayer::getFormat(void) const
{
return(_soundFormat);
}

unsigned int // sound rate (commons rate are 11025, 22050, 44100)
SoundPlayer::getRate(void) const
{
return(_sampleRate);
}

void
SoundPlayer::setAutoPop(bool autoPop)
{
_autoErase=autoPop;
_onSetup();
}

bool // buffers are auto erased from queue when played
SoundPlayer::getAutoPop(void) const
{
return(_autoErase);
}

bool // success (volume [0,1])
SoundPlayer::setVolume(double leftVolume, double rightVolume)
{
_leftVolume=(leftVolume<=0.0 ? 0.0 : leftVolume);
_rightVolume=(rightVolume<=0.0 ? 0.0 : rightVolume);

_onSetup();

if(SoundPlayer::_initSuccess)
  {
  if (_playing)
    return ((*SoundPlayer::_setOutputVolumeFnct)(_pluginData));
  else
    return(true);
  }
else
  return(false);
}

void
SoundPlayer::getVolume(double &leftVolumeOut, double &rightVolumeOut) const
{
leftVolumeOut=_leftVolume;
rightVolumeOut=_rightVolume;
}


//-------- Playback control --------------------------------------------------

bool
SoundPlayer::play(void)
{
_playing=true;
_onControl();

if(SoundPlayer::_initSuccess)
  return((*SoundPlayer::_setPlayFnct)(_pluginData,true));
else
  return(false);
}

bool
SoundPlayer::stop(void)
{
_playing=false;
_onControl();

if(SoundPlayer::_initSuccess)
  {
  return((*SoundPlayer::_setPlayFnct)(_pluginData,false));
  }
else
  {
  return(false);
  }
}


bool
SoundPlayer::seek(int bufferID)
{
if(!_seekCB.empty())
  {
  SoundPlayer::SeekEvent evt;
  _seekCB.fire(evt);
  }

if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_setCurrentBufferFnct)(_pluginData,bufferID));
  }
else
  return(false);
}

bool // sound is playing
SoundPlayer::isPlaying(void) const
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_isPlayingFnct)(_pluginData));
  }
else
  return(false);
}

bool // sound is in play mode (auto-run when buffers available)
SoundPlayer::getPlay(void) const
{
return(_playing);
}

//-------- Buffers control --------------------------------------------------

// Copy buffer into list of fixed size buffers.
int
SoundPlayer::pushBuffer(ArConstRef<MemoryBlock> buffer)
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_pushBufferFnct)(_pluginData,buffer));
  }
else
  return(-1);
}

unsigned int
SoundPlayer::popBuffers(bool all)
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_popBuffersFnct)(_pluginData,all));
  }
else
  return(0);
}

int
SoundPlayer::getCurrentBuffer(void) const
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_getCurrentBufferFnct)(_pluginData));
  }
else
  return(-1);
}

unsigned int
SoundPlayer::getNbBuffers(void) const
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_getNbBuffersFnct)(_pluginData));
  }
else
  return(0);
}

//-------- Time control --------------------------------------------------

unsigned int // time in millisecond
SoundPlayer::getTime(void) const
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_getTimeFnct)(_pluginData));
  }
else
  return(0);
}



bool
SoundPlayer::setTime(unsigned int newTimePosition)
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_setTimeFnct)(_pluginData,newTimePosition));
  }
else
  return(false);
}

unsigned int // total buffers duration
SoundPlayer::getDuration(void) const
{
if(SoundPlayer::_initSuccess)
  {
  return ((*SoundPlayer::_getDurationFnct)(_pluginData));
  }
else
  return(0);
}

//-------- Forbidden operations ----------------------------------------------

void
SoundPlayer::_onPolling(const Scheduler::PollingEvent & evt)
{
if (find(evt.readingFD.begin(),evt.readingFD.end(),_deviceID)!=evt.readingFD.end())
  {
  char data[10];

  if (_autoErase)
    popBuffers(false); // only played buffers

  while(ArSystem::partialReadFD(_deviceID,&data,10*sizeof(char))>0)
  {
  // Partial read of pipe...juste empty content
  }
  // Read
  if(!_seekCB.empty())
    {
    SoundPlayer::SeekEvent evt;
    _seekCB.fire(evt);
    }
  }
}


void
SoundPlayer::_onSetup(void) // Protected
{
if(!_setupCB.empty())
  {
  SoundPlayer::SetupEvent evt;
  _setupCB.fire(evt);
  }
}

void
SoundPlayer::_onControl(void) // Protected
{
if(!_controlCB.empty())
  {
  SoundPlayer::ControlEvent evt;
  _controlCB.fire(evt);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
