// LI2/ENIB - AReVi-Contrib - Sound recorder
//----------------------------------------------------------------------------
// file        : AReVi/Contrib/soundRecorder.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Contrib/soundRecorder.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class SoundRecorder
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(SoundRecorder,ArObject)

bool SoundRecorder::_initSuccess=false;

void * (* SoundRecorder::_openSoundInputFnct)(SoundRecorder *, const StlString &, int &)=
  (void * ( *)(SoundRecorder *, const StlString &, int &))0;

void (* SoundRecorder::_closeSoundInputFnct)(void  *)=
  (void (* )(void *))0;

bool (* SoundRecorder::_setRecordFnct)(void *, bool )=
  (bool (* )(void *, bool ))0;

bool (* SoundRecorder::_popPendingBufferFnct)(void *, ArRef<MemoryBlock> )=
  (bool (* )(void *, ArRef<MemoryBlock> ))0;

unsigned (* SoundRecorder::_getNbPendingBuffersFnct)(const void *)=
  (unsigned (* )(const void *))0;

//-------- Construction / Destruction ----------------------------------------

SoundRecorder::SoundRecorder(ArCW & arCW,
                             UtilSound::SoundFormat format,
                             unsigned int rate) // Protected
: ArObject(arCW),
  _deviceID(-1),
  _pluginData(NULL),
  _recording(false),
  _backupData(false),
  _soundFormat(format),
  _sampleRate(rate),
  _bufferReadyCB(thisRef()),
  _controlCB(thisRef())
{
if(!SoundRecorder::_initSuccess)
  {
  setErrorMessage("Plugin not initialized");
  }
else
  {

  _pluginData=((*SoundRecorder::_openSoundInputFnct)(thisRef().c_ptr(), "", _deviceID));

  if ((!_pluginData) || _deviceID<0)
    {
    _deviceID=-1;
    setErrorMessage("Can not open audio device for input");
    }
  else
    {
    ArSystem::setBlockingFD(_deviceID,false);
    Scheduler::access()->addPollingCB(thisRef(),&SoundRecorder::_onPolling);
    ArSystem::addReadFDPolling(_deviceID);
    }
  }
}


SoundRecorder::SoundRecorder(ArCW & arCW,
                             UtilSound::SoundFormat format,
                             unsigned int rate,
                             bool recordOutput) // Protected
: ArObject(arCW),
  _deviceID(-1),
  _pluginData(NULL),
  _recording(false),
  _backupData(false),
  _soundFormat(format),
  _sampleRate(rate),
  _bufferReadyCB(thisRef()),
  _controlCB(thisRef())
{
if(!SoundRecorder::_initSuccess)
  {
  setErrorMessage("Plugin not initialized");
  }
else
  {
  _pluginData=((*SoundRecorder::_openSoundInputFnct)(thisRef().c_ptr(), (recordOutput?"DeviceOutput":""), _deviceID));

  if ((!_pluginData) || _deviceID<0)
    {
    _deviceID=-1;
    setErrorMessage("Can not open audio device for input");
    }
  else
    {
    ArSystem::setBlockingFD(_deviceID,false);
    Scheduler::access()->addPollingCB(thisRef(),&SoundRecorder::_onPolling);
    ArSystem::addReadFDPolling(_deviceID);
    }
  }
}

SoundRecorder::~SoundRecorder(void) // Protected
{
if(SoundRecorder::_initSuccess)
  {
  Scheduler::access()->removePollingCB(thisRef(),&SoundRecorder::_onPolling);
  ArSystem::removeReadFDPolling(_deviceID);

  if (_deviceID!=-1)
    {
    ((*SoundRecorder::_closeSoundInputFnct)(_pluginData));
    _pluginData=NULL;
    _deviceID=-1;
    }
  }
}

//-------- Sound properties --------------------------------------------------

UtilSound::SoundFormat// sound format
SoundRecorder::getFormat(void) const
{
return(_soundFormat);
}

unsigned int // sound rate (commons rate are 11025, 22050, 44100)
SoundRecorder::getRate(void) const
{
return(_sampleRate);
}

//-------- Record control --------------------------------------------------

bool
SoundRecorder::start(bool backupData)
{
_backupData=backupData;

if (_recording)
  return(true);

_recording=true;
_onControl();

if(SoundRecorder::_initSuccess)
  return((*SoundRecorder::_setRecordFnct)(_pluginData,true));
else
  return(false);
}

bool
SoundRecorder::stop(void)
{
if (!_recording)
  return(true);

_recording=false;
_onControl();

if(SoundRecorder::_initSuccess)
  {
  return((*SoundRecorder::_setRecordFnct)(_pluginData,false));
  }
else
  {
  return(false);
  }
}

bool // sound is recording
SoundRecorder::isRecording(void) const
{
return _recording;
}

//-------- Data management  --------------------------------------------------

const StlList<ArRef<MemoryBlock> > &
SoundRecorder::getBuffers(void) const
{
return(_buffers);
}

StlList<ArRef<MemoryBlock> > &
SoundRecorder::accessBuffers(void)
{
return(_buffers);
}

// Return all recorded but not yet read buffers in queue list
bool // success
SoundRecorder::flushPendingBuffers(StlList<ArRef<MemoryBlock> > &buffersOut)
{
if(!SoundRecorder::_initSuccess)
  return(false);

ArRef<MemoryBlock> buffer=MemoryBlock::NEW();
while ((*SoundRecorder::_popPendingBufferFnct)(_pluginData,buffer))
  {
  buffersOut.push_back(buffer);

  if (_backupData)
    _buffers.push_back(buffer);

  buffer=MemoryBlock::NEW();
  }

return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
SoundRecorder::_onPolling(const Scheduler::PollingEvent & evt)
{
if(!SoundRecorder::_initSuccess)
  return;

if (find(evt.readingFD.begin(),evt.readingFD.end(),_deviceID)!=evt.readingFD.end())
  {
  char data[10];

  while(ArSystem::partialReadFD(_deviceID,&data,10*sizeof(char))>0)
  {
  // Partial read of pipe...juste empty content
  }

  ArRef<MemoryBlock> buffer=MemoryBlock::NEW();
  while ((*SoundRecorder::_popPendingBufferFnct)(_pluginData,buffer))
    {
    if (_backupData)
      {
      _buffers.push_back(buffer);

      if(!_bufferReadyCB.empty())
        {
        SoundRecorder::BufferReadyEvent evt;
        evt.buffer=buffer;
        _bufferReadyCB.fire(evt);
        }
      buffer=MemoryBlock::NEW();
      }
    else if(!_bufferReadyCB.empty())
      {
      SoundRecorder::BufferReadyEvent evt;
      evt.buffer=buffer;
      _bufferReadyCB.fire(evt);
      }
    }
  }
}

void
SoundRecorder::_onControl(void) // Protected
{
if(!_controlCB.empty())
  {
  SoundRecorder::ControlEvent evt;
  _controlCB.fire(evt);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
