// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/abstractInteractor.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/abstractInteractor.h"

#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/osd.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class AbstractInteractor (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(AbstractInteractor,ArObject)

//-------- Construction / Destruction ----------------------------------------

AbstractInteractor::AbstractInteractor(ArCW & arCW) // Protected
: ArObject(arCW),
  _renderer(),
  _ignoredObjects(),
  _ignoredOSD()
{
  // Nothing to be done
}

AbstractInteractor::~AbstractInteractor(void) // Protected
{
setRenderer(Renderer3D::nullRef());
}

//-------- Renderer management -----------------------------------------------

void
AbstractInteractor::setRenderer(ArRef<Renderer3D> renderer)
{
if(_renderer!=renderer)
  {
  _clearInteraction();
  if(_renderer)
    {
    _renderer->removeMouseButtonCB(thisRef(),
                                   &AbstractInteractor::_onMouseButton);
    _renderer->removeMouseMotionCB(thisRef(),
                                   &AbstractInteractor::_onMouseMotion);
    _renderer->removeTimeOutCB(thisRef(),
                               &AbstractInteractor::_onTimeOut);
    _renderer->removeKeyboardCB(thisRef(),
                                &AbstractInteractor::_onKeyboard);
    }
  _renderer=renderer;
  if(_renderer)
    {
    _renderer->addMouseButtonCB(thisRef(),
                                &AbstractInteractor::_onMouseButton);
    _renderer->addMouseMotionCB(thisRef(),
                                &AbstractInteractor::_onMouseMotion);
    _renderer->addTimeOutCB(thisRef(),
                            &AbstractInteractor::_onTimeOut);
    _renderer->addKeyboardCB(thisRef(),
                             &AbstractInteractor::_onKeyboard);
    }
  }
}

ArConstRef<Renderer3D> // interacted renderer
AbstractInteractor::getRenderer(void) const
{
return(_renderer);
}

ArRef<Renderer3D> // interacted renderer
AbstractInteractor::accessRenderer(void)
{
return(_renderer);
}

//-------- Object interaction ------------------------------------------------

void
AbstractInteractor::addIgnoredObject(ArRef<Object3D> object)
{
if(!ignoresObject(object))
  {
  _ignoredObjects.push_back(object->thisPtr());
  }
}

void
AbstractInteractor::removeIgnoredObject(ArRef<Object3D> object)
{
for(unsigned int i=_ignoredObjects.size();i--;)
  {
  if(_ignoredObjects[i])
    {
    if(_ignoredObjects[i]==object)
      {
      StlVectorFastErase(_ignoredObjects,i);
      return;
      }
    }
  else
    {
    StlVectorFastErase(_ignoredObjects,i);
    }
  }
}

bool // object is ignored for interactions
AbstractInteractor::ignoresObject(ArConstRef<Object3D> object) const
{
for(unsigned int i=_ignoredObjects.size();i--;)
  {
  if(_ignoredObjects[i])
    {
    if(_ignoredObjects[i]==object)
      {
      return(true);
      }
    }
  else
    {
    StlVectorFastErase(const_cast<AbstractInteractor *>(this)->_ignoredObjects,
                       i);
    }
  }
return(false);
}

void
AbstractInteractor::getIgnoredObjects(StlVector<ArRef<Object3D> > & objectsOut)
                                                                         const
{
objectsOut.clear();
for(unsigned int i=_ignoredObjects.size();i--;)
  {
  if(_ignoredObjects[i])
    {
    objectsOut.push_back(_ignoredObjects[i]->thisRef());
    }
  else
    {
    StlVectorFastErase(const_cast<AbstractInteractor *>(this)->_ignoredObjects,
                       i);
    }
  }
}

//-------- OSD interaction ---------------------------------------------------

void
AbstractInteractor::addIgnoredOSD(ArRef<OSD> osd)
{
if(!ignoresOSD(osd))
  {
  _ignoredOSD.push_back(osd->thisPtr());
  }
}

void
AbstractInteractor::removeIgnoredOSD(ArRef<OSD> osd)
{
for(unsigned int i=_ignoredOSD.size();i--;)
  {
  if(_ignoredOSD[i])
    {
    if(_ignoredOSD[i]==osd)
      {
      StlVectorFastErase(_ignoredOSD,i);
      return;
      }
    }
  else
    {
    StlVectorFastErase(_ignoredOSD,i);
    }
  }
}

bool // osd is ignored for interactions
AbstractInteractor::ignoresOSD(ArConstRef<OSD> osd) const
{
for(unsigned int i=_ignoredOSD.size();i--;)
  {
  if(_ignoredOSD[i])
    {
    if(_ignoredOSD[i]==osd)
      {
      return(true);
      }
    }
  else
    {
    StlVectorFastErase(const_cast<AbstractInteractor*>(this)->_ignoredOSD,i);
    }
  }
return(false);
}

void
AbstractInteractor::getIgnoredOSD(StlVector<ArRef<OSD> > & osdOut) const
{
osdOut.clear();
for(unsigned int i=_ignoredOSD.size();i--;)
  {
  if(_ignoredOSD[i])
    {
    osdOut.push_back(_ignoredOSD[i]->thisRef());
    }
  else
    {
    StlVectorFastErase(const_cast<AbstractInteractor*>(this)->_ignoredOSD,i);
    }
  }
}

//------- Forbidden operations ----------------------------------------------

void
AbstractInteractor::_clearInteraction(void) // Protected
{
  // Nothing to be done
}

void
AbstractInteractor::_onMouseButton(const Renderer3D::MouseButtonEvent &
                                                       /* evt */) // Protected
{
  // Nothing to be done
}

void
AbstractInteractor::_onMouseMotion(const Renderer3D::MouseMotionEvent &
                                                       /* evt */) // Protected
{
  // Nothing to be done
}

void
AbstractInteractor::_onKeyboard(const Renderer3D::KeyboardEvent &
                                                       /* evt */) // Protected
{
  // Nothing to be done
}

void
AbstractInteractor::_onTimeOut(const Renderer3D::TimeOutEvent &
                                                       /* evt */) // Protected
{
  // Nothing to be done
}

ArRef<Object3D> // found object
AbstractInteractor::_findObject(Util3D::Dbl3 & localPointOut,
                                ArRef<ShapePart3D> & interPartOut,
                                unsigned int & triIndexOut,
                                int xMouse,
                                int yMouse,
                                bool onlyMainScene) // Protected
{
if(_renderer)
  {
  double yaw,pitch;
  _renderer->computeCursorDirection(xMouse,yMouse,yaw,pitch);
  ArRef<Base3D> ray=Base3D::NEW();
  ray->setLocation(_renderer);
  ray->yaw(yaw);
  ray->pitch(pitch);
  ArConstRef<Scene3D> scenes[3];
  scenes[1]=_renderer->getScene();
  if(!onlyMainScene)
    {
    scenes[0]=_renderer->getBackScene();
    scenes[2]=_renderer->getFrontScene();
    }
  for(unsigned int i=3;i--;)
    {
    if(scenes[i])
      {
      StlVector<double> dist;
      StlVector<ArRef<Object3D> > found;
      StlVector<ArRef<ShapePart3D> > parts;
      StlVector<unsigned int> triIndices;
      scenes[i]->rayIntersection(ray,3,true,Object3D::CLASS(),
                                 found,parts,triIndices,dist,_renderer);
      unsigned int nb=found.size();
      for(unsigned int j=0;j<nb;j++)
        {
        if(found[j]&&!ignoresObject(found[j]))
          {
          localPointOut.x=dist[j];
          localPointOut.y=0.0;
          localPointOut.z=0.0;
          ray->localToGlobalVector(localPointOut.x,
                                   localPointOut.y,
                                   localPointOut.z);
          _renderer->globalToLocalVector(localPointOut.x,
                                         localPointOut.y,
                                         localPointOut.z);
          if((localPointOut.x>=_renderer->getNearDistance())&&
             (localPointOut.x<=_renderer->getFarDistance()))
            {
            interPartOut=parts[j];
            triIndexOut=triIndices[j];
            return(found[j]);
            }
          found[j].clear();
          }
        }
      }
    }
  }
interPartOut.clear();
triIndexOut=UINT_MAX;
return(Object3D::nullRef());
}

ArRef<OSD> // found osd
AbstractInteractor::_findOSD(int xMouse,
                             int yMouse)
{
if(_renderer) // FIXME: what about ignored OSDs !?!?
  {
  unsigned int w,h;
  _renderer->getRenderSize(w,h);
  if(_renderer->getStereoMode()==Renderer3D::STEREO_SPLIT)
    {
    w>>=1;
    if(xMouse>(int)w)
      {
      xMouse-=w;
      }
    }
  ArRef<BoundingBox3D> bbox=BoundingBox3D::NEW();
  StlVector<ArRef<OSD> > candidates;
  for(unsigned int i=_renderer->getNbOSD();i--;)
    {
    ArRef<OSD> osd=_renderer->accessOSD(i);
    osd->readBoundingBox(_renderer,bbox);
    if(bbox->getBoundingBox().locateVertex(xMouse,yMouse,
                                           0.0)==Util3D::MIDDLE_LOC)
      {
      candidates.push_back(osd);
      }
    }
  if(!candidates.empty())
    {
    sort(candidates.begin(),candidates.end(),OSD::DepthPrecedence());
    return(candidates.front());
    }
  }
return(OSD::nullRef());
}

} // namespace AReVi

//----------------------------------------------------------------------------
