// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/dataTexture.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/dataTexture.h"

#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/abstractStream.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/llw_priv.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class DataTexture
//----------------------------------------------------------------------------

AR_CLASS_DEF(DataTexture,Texture)

unsigned int DataTexture::_quality=75;

//-------- Construction / Destruction ----------------------------------------
DataTexture::DataTexture(ArCW & arCW) // Protected
: Texture(arCW),
  _reloadCB(thisPtr()),
  _setupCB(thisPtr())
{
_texData=Texture::_emptyTexData;
}

DataTexture::DataTexture(ArCW & arCW,
                         unsigned int width,
                         unsigned int height,
                         unsigned int components,
                         ArRef<MemoryBlock> data,
                         bool checkAlpha,
                         bool useMipmap) // Protected
: Texture(arCW),
  _reloadCB(thisPtr()),
  _setupCB(thisPtr())
{
_loadDataTexture(width,height,components,data,checkAlpha,useMipmap,false);
}


DataTexture::DataTexture(ArCW & arCW,
                         unsigned int width,
                         unsigned int height,
                         unsigned int components,
                         ArRef<MemoryBlock> data,
                         bool checkAlpha,
                         bool useMipmap,
                         bool useCubemap) // Protected
: Texture(arCW),
  _reloadCB(thisPtr()),
  _setupCB(thisPtr())
{
_loadDataTexture(width,height,components,data,checkAlpha,useMipmap,useCubemap);
}

DataTexture::~DataTexture(void) // Protected
{
  // Nothing to be done
}

//-------- Texture changes ---------------------------------------------------

ArRef<MemoryBlock> // data filled from bottom to top
DataTexture::accessData(void)
{
return(_texData ? _texData->data : MemoryBlock::nullRef());
}

void
DataTexture::setBlending(bool needsBlending)
{
if(_texData)
  {
  _texData->needsBlending=needsBlending;
  _onSetup();
  }
}

void
DataTexture::setAlphaTest(bool needsAlphaTest)
{
if(_texData)
  {
  _texData->needsAlphaTest=needsAlphaTest;
  _onSetup();
  }
}

bool // success
DataTexture::reload(bool checkAlpha,
                    bool useMipmap)
{
if(_texData)
  {
  unsigned int size=_texData->width*_texData->height*_texData->components;
  if(_texData->data->getSize()>=size)
    {
    _texData->useMipmap=useMipmap;
    if(!_load(checkAlpha,false))
      {
      return(false);
      }
    if(!_reloadCB.empty())
      {
      DataTexture::ReloadEvent event;
      _reloadCB.fire(event);
      }
    return(true);
    }
  else
    {
    return(false);
    }
  }
else
  {
  return(false);
  }
}

//-------- Serialization operations ------------------------------------------

void
DataTexture::serialize(SerializationDependencies & dependenciesOut,
                      ArRef<AbstractOStream> output) const
{
Texture::serialize(dependenciesOut,output);
if(_texData)
  {
  output->writeUInt(_texData->width);
  output->writeUInt(_texData->height);
  output->writeUInt(_texData->components);
  bool compressed;
  ArRef<MemoryBlock> buffer=MemoryBlock::NEW();
  if(_texData->components==3 &&
     DataTexture::_quality<=100 &&
     LLW::saveRGBImageToJPEG(_texData->data,
                             _texData->width,
                             _texData->height,
                             DataTexture::_quality,
                             buffer))
    {
    compressed=true;
    }
  else // Data will be not compressed
    {
    compressed=false;
    }
  unsigned int flags=(_texData->needsBlending  ? 1<<0 : 0)+
                     (_texData->needsAlphaTest ? 1<<1 : 0)+
                     (_texData->useMipmap      ? 1<<2 : 0)+
                     (compressed               ? 1<<3 : 0)+
                     (_texData->useCubeMap     ? 1<<4 : 0);
  output->writeUInt(flags);
  if(compressed)
    {
    output->writeUInt(buffer->getSize());
    output->writeData(buffer->getMemory(),buffer->getSize());
    }
  else
    {
    unsigned int size=_texData->width*_texData->height*_texData->components;
    output->writeData(_texData->data->getMemory(),size);
    }
  }
else
  {
  output->writeUInt(0);
  }
}

bool // success
DataTexture::unserialize(const SerializationDependencies & dependencies,
                         ArRef<AbstractIStream> input)
{
unsigned int width,height,components,size,flags;
ArRef<MemoryBlock> data;
if(!Texture::unserialize(dependencies,input)||
   !input->readUInt(width))
  {
  return(false);
  }
if(width)
  {
  if(!input->readUInt(height)||
     !input->readUInt(components)||
     !input->readUInt(flags))
    {
    return(false);
    }
  if(flags&(1<<3)) // => compressed
    {
    if(!input->readUInt(size))
      {
      return(false);
      }
    }
  else
    {
    size=width*height*components;
    }
  data=MemoryBlock::NEW();
  data->setSize(size);
  if(!input->readData(data->accessMemory(),size))
    {
    return(false);
    }
  }
if(flags&(1<<3)) // Look for compressed data
  {
  ArRef<MemoryBlock> buffer=MemoryBlock::NEW();
  unsigned w,h;
  if(!LLW::loadRGBImageFromJPEG(data,w,h,buffer))
    {
    setErrorMessage(buffer->getErrorMessage());
    return(false);
    }
  else if((w!=width)||(h!=height))
    {
    return (false);
    }
  buffer->swapBlock(data);
  }
if(_texData&&(_texData!=Texture::_emptyTexData))
  {
  delete _texData; // non shared data!
  }
_texData=(Texture::TexData *)0;
if(width)
  {
  if(_loadDataTexture(width,height,components,data,false,flags&(1<<2),flags&(1<<4)))
    {
    _texData->needsBlending=(flags&(1<<0));
    _texData->needsAlphaTest=(flags&(1<<1));
    }
  }
else
  {
  _texData=Texture::_emptyTexData;
  }
_postAppearanceRebuild();
_postRedisplay();
// Fire all CB
_onSetup();
if(!_reloadCB.empty())
  {
  DataTexture::ReloadEvent event;
  _reloadCB.fire(event);
  }
return(true);
}

//-------- Forbidden operations ----------------------------------------------

bool // success
DataTexture::_loadDataTexture(unsigned int width,
                              unsigned int height,
                              unsigned int components,
                              ArRef<MemoryBlock> data,
                              bool checkAlpha,
                              bool useMipmap,
                              bool useCubemap)
{
unsigned int size=components*height*width;
if(!size||(components>4))
  {
  StlString msg="bad texture dimensions ";
  msg+=uintToStr(width);
  msg+="x";
  msg+=uintToStr(height);
  msg+="x";
  msg+=uintToStr(components);
  setErrorMessage(msg);
  return(false);
  }
if(data->getSize()<size)
  {
  setErrorMessage("not enough data providen");
  return(false);
  }

unsigned int newWidth=Texture::_roundSize(width);
unsigned int newHeight=useCubemap?Texture::_roundSize(height/6)*6:Texture::_roundSize(height);
if((newWidth!=width)||(newHeight!=height))
  {
  ArRef<MemoryBlock> oldData=MemoryBlock::NEW();
  if(!oldData->ensureMinSize(newWidth*newHeight*components)||
      !oldData->swapBlock(data))
    {
    setErrorMessage("Cannot adjust data for a suitable texture size");
    return(false);
    }
  Texture::_upScale(width,height,components,oldData->getMemory(),
                    newWidth,newHeight,data->accessMemory());
  }
_texData=new Texture::TexData(0,width,height,newWidth,newHeight,
                              components,useMipmap,useCubemap,data);
if(!_load(checkAlpha,false))
  {
  delete _texData;
  _texData=(Texture::TexData *)0;
  return(false);
  }
return(true);
}

void
DataTexture::_onSetup(void) // Protected
{
if(!_setupCB.empty())
  {
  DataTexture::SetupEvent evt;
  _setupCB.fire(evt);
  }
}

unsigned int
DataTexture::getQuality(void) // Static
{
return(DataTexture::_quality);
}

void
DataTexture::setQuality(unsigned int quality) // Static
{
DataTexture::_quality=quality;
}

} // namespace AReVi

//----------------------------------------------------------------------------
