// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/gl_priv.cpp
// author(s)   : F. Harrouet, T. Jourdan, E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/gl_priv.h"

#include "AReVi/arSystem.h"

#include <string.h>

#ifdef __CYGWIN__
#include <windows.h>
#endif // __CYGWIN__

namespace AReVi {

//----------------------------------------------------------------------------
// Class GL (abstract)
//----------------------------------------------------------------------------

//-------- OpenGL initialization ---------------------------------------------

bool // success
GL::hasExtension(const char * extension,
                 const char * extensionList)
{
if(extensionList)
  {
  unsigned int len=strlen(extension);
  const char * endList=extensionList+strlen(extensionList);
  while(extensionList<endList)
    {
    unsigned int extLen=strcspn(extensionList," ");
    if((len==extLen)&&
       !strncmp(extension,extensionList,extLen))
      {
      return(true);
      }
    extensionList+=extLen+1;
    }
  }
return(false);
}

void
GL::init(void)
{
static bool firstInit=true;
if(firstInit)
  {
  const char * glExt=(const char *)glGetString(GL_EXTENSIONS);
  #ifndef __CYGWIN__
  GL::loadFunction(GL::_getProcAddress,"glXGetProcAddress","ARB");
  #endif // __CYGWIN__
  //---- properties ----
  GLint nb;
  glGetIntegerv(GL_MAX_LIGHTS,&nb);
  GL::_nbLights=nb;
  glGetIntegerv(GL_MAX_TEXTURE_SIZE,&nb);
  GL::_maxTextureSize=nb;
  //---- separate specular color ----
  GL::_hasSeparateSpecularColor=false;
  if(GL::hasExtension("GL_EXT_separate_specular_color",glExt) ||
     GL::hasExtension("GL_ARB_separate_specular_color",glExt))
    {
    GL::_hasSeparateSpecularColor=true;
    }
  else
    {
    cerr << "!!! GL_EXT_separate_specular_color extension is not supported !!!"
         << endl;
    }
  //---- texture compression ----
  GL::_hasTextureCompression=false;
  if(GL::hasExtension("GL_ARB_texture_compression",glExt))
    {
  #ifdef __APPLE__
    cerr << "!!! GL_ARB_texture_compression extension is supported"
         << " but intentionally disabled !!!"
         << endl;
  #else // not __APPLE__
    GL::_hasTextureCompression=true;
  #endif // not __APPLE__
    }
  else
    {
    cerr << "!!! GL_ARB_texture_compression extension is not supported !!!"
         << endl;
    }
  //---- cube texture ----
  GL::_hasTextureCubeMap=false;
  if(GL::hasExtension("GL_ARB_texture_cube_map",glExt))
    {
    GL::_hasTextureCubeMap=true;
    }
  else
    {
    cerr << "!!! GL_EXT_texture_cube_map extension is not supported !!!"
         << endl;
    }
  //---- multi-texturing ----
  if(GL::hasExtension("GL_ARB_multitexture",glExt))
    {
    GL::loadFunction(GL::glActiveTexture,"glActiveTexture","ARB");
    GL::loadFunction(GL::glClientActiveTexture,"glClientActiveTexture","ARB");
    GL::loadFunction(GL::glMultiTexCoord2dv,"glMultiTexCoord2dv","ARB");
    }
  if(GL::glActiveTexture&&
     GL::glClientActiveTexture&&
     GL::glMultiTexCoord2dv)
    {
    glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB,&nb);
    GL::_nbTextureUnits=nb;
    }
  else
    {
    GL::_nbTextureUnits=1;
    cerr << "!!! GL_ARB_multitexture extension is not supported !!!"
         << endl;
    }
  //---- automatic mipmap generation ----
  GL::_hasMipmapGeneration=false;
  if(GL::hasExtension("GL_SGIS_generate_mipmap",glExt))
    {
    GL::_hasMipmapGeneration=true;
    }
  else
    {
    cerr << "!!! GL_SGIS_generate_mipmap extension is not supported !!!"
         << endl;
    }
  //---- vertex array ranges ----
  if(GL::hasExtension("GL_EXT_draw_range_elements",glExt))
    {
    GL::loadFunction(GL::glDrawRangeElements,"glDrawRangeElements","EXT");
    }
  if(!GL::glDrawRangeElements)
    {
    cerr << "!!! GL_EXT_draw_range_elements extension is not supported !!!"
         << endl;
    }
  //---- OpenGL shading language : GLSL ----
  GL::_hasShadingLanguage=false;
  if(GL::hasExtension("GL_ARB_vertex_shader",glExt)&&
     GL::hasExtension("GL_ARB_fragment_shader",glExt)&&
     GL::hasExtension("GL_ARB_shader_objects",glExt)&&
     GL::hasExtension("GL_ARB_shading_language_100",glExt))
    {
    //---- GL_ARB_vertex_shader functions ----
    GL::loadFunction(GL::glVertexAttrib1f,"glVertexAttrib1f","ARB");
    GL::loadFunction(GL::glVertexAttrib1s,"glVertexAttrib1s","ARB");
    GL::loadFunction(GL::glVertexAttrib1d,"glVertexAttrib1d","ARB");
    GL::loadFunction(GL::glVertexAttrib2f,"glVertexAttrib2f","ARB");
    GL::loadFunction(GL::glVertexAttrib2s,"glVertexAttrib2s","ARB");
    GL::loadFunction(GL::glVertexAttrib2d,"glVertexAttrib2d","ARB");
    GL::loadFunction(GL::glVertexAttrib3f,"glVertexAttrib3f","ARB");
    GL::loadFunction(GL::glVertexAttrib3s,"glVertexAttrib3s","ARB");
    GL::loadFunction(GL::glVertexAttrib3d,"glVertexAttrib3d","ARB");
    GL::loadFunction(GL::glVertexAttrib4f,"glVertexAttrib4f","ARB");
    GL::loadFunction(GL::glVertexAttrib4s,"glVertexAttrib4s","ARB");
    GL::loadFunction(GL::glVertexAttrib4d,"glVertexAttrib4d","ARB");
    GL::loadFunction(GL::glVertexAttrib4Nub,"glVertexAttrib4Nub","ARB");
    GL::loadFunction(GL::glVertexAttrib1fv,"glVertexAttrib1fv","ARB");
    GL::loadFunction(GL::glVertexAttrib1sv,"glVertexAttrib1sv","ARB");
    GL::loadFunction(GL::glVertexAttrib1dv,"glVertexAttrib1dv","ARB");
    GL::loadFunction(GL::glVertexAttrib2fv,"glVertexAttrib2fv","ARB");
    GL::loadFunction(GL::glVertexAttrib2sv,"glVertexAttrib2sv","ARB");
    GL::loadFunction(GL::glVertexAttrib2dv,"glVertexAttrib2dv","ARB");
    GL::loadFunction(GL::glVertexAttrib3fv,"glVertexAttrib3fv","ARB");
    GL::loadFunction(GL::glVertexAttrib3sv,"glVertexAttrib3sv","ARB");
    GL::loadFunction(GL::glVertexAttrib3dv,"glVertexAttrib3dv","ARB");
    GL::loadFunction(GL::glVertexAttrib4fv,"glVertexAttrib4fv","ARB");
    GL::loadFunction(GL::glVertexAttrib4sv,"glVertexAttrib4sv","ARB");
    GL::loadFunction(GL::glVertexAttrib4dv,"glVertexAttrib4dv","ARB");
    GL::loadFunction(GL::glVertexAttrib4iv,"glVertexAttrib4iv","ARB");
    GL::loadFunction(GL::glVertexAttrib4bv,"glVertexAttrib4bv","ARB");
    GL::loadFunction(GL::glVertexAttrib4ubv,"glVertexAttrib4ubv","ARB");
    GL::loadFunction(GL::glVertexAttrib4usv,"glVertexAttrib4usv","ARB");
    GL::loadFunction(GL::glVertexAttrib4uiv,"glVertexAttrib4uiv","ARB");
    GL::loadFunction(GL::glVertexAttrib4Nbv,"glVertexAttrib4Nbv","ARB");
    GL::loadFunction(GL::glVertexAttrib4Nsv,"glVertexAttrib4Nsv","ARB");
    GL::loadFunction(GL::glVertexAttrib4Niv,"glVertexAttrib4Niv","ARB");
    GL::loadFunction(GL::glVertexAttrib4Nubv,"glVertexAttrib4Nubv","ARB");
    GL::loadFunction(GL::glVertexAttrib4Nusv,"glVertexAttrib4Nusv","ARB");
    GL::loadFunction(GL::glVertexAttrib4Nuiv,"glVertexAttrib4Nuiv","ARB");
    GL::loadFunction(GL::glVertexAttribPointer,"glVertexAttribPointer","ARB");
    GL::loadFunction(GL::glEnableVertexAttribArray,
                     "glEnableVertexAttribArray","ARB");
    GL::loadFunction(GL::glDisableVertexAttribArray,
                     "glDisableVertexAttribArray","ARB");
    GL::loadFunction(GL::glBindAttribLocation,"glBindAttribLocation","ARB");
    GL::loadFunction(GL::glGetActiveAttrib,"glGetActiveAttrib","ARB");
    GL::loadFunction(GL::glGetAttribLocation,"glGetAttribLocation","ARB");
    GL::loadFunction(GL::glGetVertexAttribfv,"glGetVertexAttribfv","ARB");
    GL::loadFunction(GL::glGetVertexAttribiv,"glGetVertexAttribiv","ARB");
    GL::loadFunction(GL::glGetVertexAttribdv,"glGetVertexAttribdv","ARB");
    GL::loadFunction(GL::glGetVertexAttribPointerv,
                     "glGetVertexAttribPointerv","ARB");
    //---- GL_ARB_shader_objects functions ----
    GL::loadFunction(GL::glDeleteObject,"glDeleteObject","ARB");
    GL::loadFunction(GL::glGetHandle,"glGetHandle","ARB");
    GL::loadFunction(GL::glDetachObject,"glDetachObject","ARB");
    GL::loadFunction(GL::glCreateShaderObject,"glCreateShaderObject","ARB");
    GL::loadFunction(GL::glShaderSource,"glShaderSource","ARB");
    GL::loadFunction(GL::glCompileShader,"glCompileShader","ARB");
    GL::loadFunction(GL::glCreateProgramObject,"glCreateProgramObject","ARB");
    GL::loadFunction(GL::glAttachObject,"glAttachObject","ARB");
    GL::loadFunction(GL::glLinkProgram,"glLinkProgram","ARB");
    GL::loadFunction(GL::glUseProgramObject,"glUseProgramObject","ARB");
    GL::loadFunction(GL::glValidateProgram,"glValidateProgram","ARB");
    GL::loadFunction(GL::glUniform1f,"glUniform1f","ARB");
    GL::loadFunction(GL::glUniform2f,"glUniform2f","ARB");
    GL::loadFunction(GL::glUniform3f,"glUniform3f","ARB");
    GL::loadFunction(GL::glUniform4f,"glUniform4f","ARB");
    GL::loadFunction(GL::glUniform1i,"glUniform1i","ARB");
    GL::loadFunction(GL::glUniform2i,"glUniform2i","ARB");
    GL::loadFunction(GL::glUniform3i,"glUniform3i","ARB");
    GL::loadFunction(GL::glUniform4i,"glUniform4i","ARB");
    GL::loadFunction(GL::glUniform1fv,"glUniform1fv","ARB");
    GL::loadFunction(GL::glUniform2fv,"glUniform2fv","ARB");
    GL::loadFunction(GL::glUniform3fv,"glUniform3fv","ARB");
    GL::loadFunction(GL::glUniform4fv,"glUniform4fv","ARB");
    GL::loadFunction(GL::glUniform1iv,"glUniform1iv","ARB");
    GL::loadFunction(GL::glUniform2iv,"glUniform2iv","ARB");
    GL::loadFunction(GL::glUniform3iv,"glUniform3iv","ARB");
    GL::loadFunction(GL::glUniform4iv,"glUniform4iv","ARB");
    GL::loadFunction(GL::glUniformMatrix2fv,"glUniformMatrix2fv","ARB");
    GL::loadFunction(GL::glUniformMatrix3fv,"glUniformMatrix3fv","ARB");
    GL::loadFunction(GL::glUniformMatrix4fv,"glUniformMatrix4fv","ARB");
    GL::loadFunction(GL::glGetObjectParameterfv,
                     "glGetObjectParameterfv","ARB");
    GL::loadFunction(GL::glGetObjectParameteriv,
                     "glGetObjectParameteriv","ARB");
    GL::loadFunction(GL::glGetInfoLog,"glGetInfoLog","ARB");
    GL::loadFunction(GL::glGetAttachedObjects,"glGetAttachedObjects","ARB");
    GL::loadFunction(GL::glGetUniformLocation,"glGetUniformLocation","ARB");
    GL::loadFunction(GL::glGetActiveUniform,"glGetActiveUniform","ARB");
    GL::loadFunction(GL::glGetUniformfv,"glGetUniformfv","ARB");
    GL::loadFunction(GL::glGetUniformiv,"glGetUniformiv","ARB");
    GL::loadFunction(GL::glGetShaderSource,"glGetShaderSource","ARB");
    }
  if(GL::glVertexAttrib1f&&
     GL::glVertexAttrib1s&&
     GL::glVertexAttrib1d&&
     GL::glVertexAttrib2f&&
     GL::glVertexAttrib2s&&
     GL::glVertexAttrib2d&&
     GL::glVertexAttrib3f&&
     GL::glVertexAttrib3s&&
     GL::glVertexAttrib3d&&
     GL::glVertexAttrib4f&&
     GL::glVertexAttrib4s&&
     GL::glVertexAttrib4d&&
     GL::glVertexAttrib4Nub&&
     GL::glVertexAttrib1fv&&
     GL::glVertexAttrib1sv&&
     GL::glVertexAttrib1dv&&
     GL::glVertexAttrib2fv&&
     GL::glVertexAttrib2sv&&
     GL::glVertexAttrib2dv&&
     GL::glVertexAttrib3fv&&
     GL::glVertexAttrib3sv&&
     GL::glVertexAttrib3dv&&
     GL::glVertexAttrib4fv&&
     GL::glVertexAttrib4sv&&
     GL::glVertexAttrib4dv&&
     GL::glVertexAttrib4iv&&
     GL::glVertexAttrib4bv&&
     GL::glVertexAttrib4ubv&&
     GL::glVertexAttrib4usv&&
     GL::glVertexAttrib4uiv&&
     GL::glVertexAttrib4Nbv&&
     GL::glVertexAttrib4Nsv&&
     GL::glVertexAttrib4Niv&&
     GL::glVertexAttrib4Nubv&&
     GL::glVertexAttrib4Nusv&&
     GL::glVertexAttrib4Nuiv&&
     GL::glVertexAttribPointer&&
     GL::glEnableVertexAttribArray&&
     GL::glDisableVertexAttribArray&&
     GL::glBindAttribLocation&&
     GL::glGetActiveAttrib&&
     GL::glGetAttribLocation&&
     GL::glGetVertexAttribfv&&
     GL::glGetVertexAttribiv&&
     GL::glGetVertexAttribdv&&
     GL::glGetVertexAttribPointerv&&
     GL::glDeleteObject&&
     GL::glGetHandle&&
     GL::glDetachObject&&
     GL::glCreateShaderObject&&
     GL::glShaderSource&&
     GL::glCompileShader&&
     GL::glCreateProgramObject&&
     GL::glAttachObject&&
     GL::glLinkProgram&&
     GL::glUseProgramObject&&
     GL::glValidateProgram&&
     GL::glUniform1f&&
     GL::glUniform2f&&
     GL::glUniform3f&&
     GL::glUniform4f&&
     GL::glUniform1i&&
     GL::glUniform2i&&
     GL::glUniform3i&&
     GL::glUniform4i&&
     GL::glUniform1fv&&
     GL::glUniform2fv&&
     GL::glUniform3fv&&
     GL::glUniform4fv&&
     GL::glUniform1iv&&
     GL::glUniform2iv&&
     GL::glUniform3iv&&
     GL::glUniform4iv&&
     GL::glUniformMatrix2fv&&
     GL::glUniformMatrix3fv&&
     GL::glUniformMatrix4fv&&
     GL::glGetObjectParameterfv&&
     GL::glGetObjectParameteriv&&
     GL::glGetInfoLog&&
     GL::glGetAttachedObjects&&
     GL::glGetUniformLocation&&
     GL::glGetActiveUniform&&
     GL::glGetUniformfv&&
     GL::glGetUniformiv&&
     GL::glGetShaderSource)
    {
    GL::_hasShadingLanguage=true;
    glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB,&nb);
    GL::_nbVertexUniformComponents=nb;
    glGetIntegerv(GL_MAX_VARYING_FLOATS_ARB,&nb);
    GL::_nbVaryingFloats=nb;
    glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB,&nb);
    GL::_nbVertexAttribs=nb;
    glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS_ARB,&nb);
    GL::_nbTextureImageUnits=nb;
    glGetIntegerv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB,&nb);
    GL::_nbVertexTextureImageUnits=nb;
    glGetIntegerv(GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS_ARB,&nb);
    GL::_nbCombinedTextureImageUnits=nb;
    glGetIntegerv(GL_MAX_TEXTURE_COORDS_ARB,&nb);
    GL::_nbTextureCoords=nb;
    glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB,&nb);
    GL::_nbFragmentUniformComponents=nb;
    }
  else
    {
    cerr << "!!! GL_ARB_shading_language_100 extension is not supported !!!"
         << endl;
    }
  //---- EXT_framebuffer_object ----
  GL::_hasFrameBufferObject=false;
  if(GL::hasExtension("GL_EXT_framebuffer_object",glExt))
    {
    GL::loadFunction(GL::glGenFramebuffers,"glGenFramebuffers","EXT");
    GL::loadFunction(GL::glDeleteFramebuffers,"glDeleteFramebuffers","EXT");
    GL::loadFunction(GL::glGenRenderbuffers,"glGenRenderbuffers","EXT");
    GL::loadFunction(GL::glDeleteRenderbuffers,"glDeleteRenderbuffers","EXT");
    GL::loadFunction(GL::glBindFramebuffer,"glBindFramebuffer","EXT");
    GL::loadFunction(GL::glFramebufferTexture2D,
                     "glFramebufferTexture2D","EXT");
    GL::loadFunction(GL::glBindRenderbuffer,"glBindRenderbuffer","EXT");
    GL::loadFunction(GL::glRenderbufferStorage,"glRenderbufferStorage","EXT");
    GL::loadFunction(GL::glFramebufferRenderbuffer,
                     "glFramebufferRenderbuffer","EXT");
    GL::loadFunction(GL::glCheckFramebufferStatus,
                     "glCheckFramebufferStatus","EXT");
    }
  if(GL::glGenFramebuffers&&
     GL::glDeleteFramebuffers&&
     GL::glGenRenderbuffers&&
     GL::glDeleteRenderbuffers&&
     GL::glBindFramebuffer&&
     GL::glFramebufferTexture2D&&
     GL::glBindRenderbuffer&&
     GL::glRenderbufferStorage&&
     GL::glFramebufferRenderbuffer&&
     GL::glCheckFramebufferStatus)
    {
    GL::_hasFrameBufferObject=true;
    }
  else
    {
    cerr << "!!! GL_EXT_framebuffer_object extension is not supported !!!"
         << endl;
    }
  //---- ARB_vertex_buffer_object ----
  GL::_hasVertexBufferObject=false;
  if(GL::hasExtension("GL_ARB_vertex_buffer_object",glExt))
    {
    GL::loadFunction(GL::glBindBuffer,"glBindBuffer","ARB");
    GL::loadFunction(GL::glDeleteBuffers,"glDeleteBuffers","ARB");
    GL::loadFunction(GL::glGenBuffers,"glGenBuffers","ARB");
    GL::loadFunction(GL::glIsBuffer,"glIsBuffer","ARB");
    GL::loadFunction(GL::glBufferData,"glBufferData","ARB");
    GL::loadFunction(GL::glBufferSubData,"glBufferSubData","ARB");
    GL::loadFunction(GL::glGetBufferSubData,"glGetBufferSubData","ARB");
    GL::loadFunction(GL::glMapBuffer,"glMapBuffer","ARB");
    GL::loadFunction(GL::glUnmapBuffer,"glUnmapBuffer","ARB");
    GL::loadFunction(GL::glGetBufferParameteriv,
                     "glGetBufferParameteriv","ARB");
    GL::loadFunction(GL::glGetBufferPointerv,"glGetBufferPointerv","ARB");
    }
  if(GL::glBindBuffer&&
     GL::glDeleteBuffers&&
     GL::glGenBuffers&&
     GL::glIsBuffer&&
     GL::glBufferData&&
     GL::glBufferSubData&&
     GL::glGetBufferSubData&&
     GL::glMapBuffer&&
     GL::glUnmapBuffer&&
     GL::glGetBufferParameteriv&&
     GL::glGetBufferPointerv)
    {
    GL::_hasVertexBufferObject=true;
    }
  else
    {
    cerr << "!!! GL_ARB_vertex_buffer_object extension is not supported !!!"
         << endl;
    }
  //---- ARB_point_sprite ----
  GL::_hasPointSprite=false;
  if(GL::hasExtension("GL_ARB_point_sprite",glExt) &&
     GL::hasExtension("GL_ARB_point_parameters",glExt))
    {
    GL::loadFunction(GL::glPointParameterf,"glPointParameterf","ARB");
    GL::loadFunction(GL::glPointParameterfv,"glPointParameterfv","ARB");
    }
  if(GL::glPointParameterf&&
     GL::glPointParameterfv)
    {
    GL::_hasPointSprite=true;
    }
  else
    {
    cerr << "!!! GL_ARB_point_sprite extension is not supported !!!"
         << endl;
    }
  firstInit=false;
  }
//---- general purpose OpenGL settings ----
glEnable(GL_DEPTH_TEST);
glDepthRange(0.0,1.0);
glDepthFunc(GL_LESS);
glDepthMask(GL_TRUE);
glEnable(GL_CULL_FACE);
glCullFace(GL_BACK);
glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
if(GL::hasSeparateSpecularColor())
  {
  glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL_EXT,
                GL_SEPARATE_SPECULAR_COLOR_EXT);
  }
glEnable(GL_LIGHTING);
glEnable(GL_NORMALIZE);
glShadeModel(GL_SMOOTH);
glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
glAlphaFunc(GL_GEQUAL,0.01);
glPixelStorei(GL_PACK_ALIGNMENT,1);
glPixelStorei(GL_UNPACK_SWAP_BYTES,GL_FALSE);
glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
glPixelStorei(GL_UNPACK_ALIGNMENT,1);
}

//-------- Properties --------------------------------------------------------

unsigned int GL::_nbLights=1;

unsigned int GL::_maxTextureSize=1;

//-------- GL_EXT_separate_specular_color ------------------------------------

bool GL::_hasSeparateSpecularColor=false;

//-------- GL_ARB_texture_compression ----------------------------------------

bool GL::_hasTextureCompression=false;

//-------- GL_EXT_texture_cube_map -------------------------------------------

bool GL::_hasTextureCubeMap=false;

//-------- GL_ARB_multitexture -----------------------------------------------

unsigned int GL::_nbTextureUnits=1;

void APIENTRY (* GL::glActiveTexture)(GLenum)=
  (void (APIENTRY *)(GLenum))0;

void APIENTRY (* GL::glClientActiveTexture)(GLenum)=
  (void (APIENTRY *)(GLenum))0;

void APIENTRY (* GL::glMultiTexCoord2dv)(GLenum,const GLdouble *)=
  (void (APIENTRY *)(GLenum,const GLdouble *))0;

//-------- GL_SGIS_generate_mipmap -------------------------------------------

bool GL::_hasMipmapGeneration=false;

//-------- GL_EXT_draw_range_elements ----------------------------------------

void APIENTRY (* GL::glDrawRangeElements)(GLenum,GLuint,GLuint,GLsizei,
                                          GLenum,const GLvoid *)=
  (void (APIENTRY *)(GLenum,GLuint,GLuint,GLsizei,GLenum,const GLvoid *))0;

//-------- GL_ARB_shading_language_100 ---------------------------------------

bool GL::_hasShadingLanguage=false;

//-------- GL_ARB_vertex_shader ----------------------------------------------

unsigned int GL::_nbVertexUniformComponents=0;

unsigned int GL::_nbVaryingFloats=0;

unsigned int GL::_nbVertexAttribs=0;

unsigned int GL::_nbTextureImageUnits=0;

unsigned int GL::_nbVertexTextureImageUnits=0;

unsigned int GL::_nbCombinedTextureImageUnits=0;

unsigned int GL::_nbTextureCoords=0;

void APIENTRY (* GL::glVertexAttrib1f)(GLuint,GLfloat)=
  (void (APIENTRY *)(GLuint,GLfloat))0;

void APIENTRY (* GL::glVertexAttrib1s)(GLuint,GLshort)=
  (void (APIENTRY *)(GLuint,GLshort))0;

void APIENTRY (* GL::glVertexAttrib1d)(GLuint,GLdouble)=
  (void (APIENTRY *)(GLuint,GLdouble))0;

void APIENTRY (* GL::glVertexAttrib2f)(GLuint,GLfloat,GLfloat)=
  (void (APIENTRY *)(GLuint,GLfloat,GLfloat))0;

void APIENTRY (* GL::glVertexAttrib2s)(GLuint,GLshort,GLshort)=
  (void (APIENTRY *)(GLuint,GLshort,GLshort))0;

void APIENTRY (* GL::glVertexAttrib2d)(GLuint,GLdouble,GLdouble)=
  (void (APIENTRY *)(GLuint,GLdouble,GLdouble))0;

void APIENTRY (* GL::glVertexAttrib3f)(GLuint,GLfloat,GLfloat,GLfloat)=
  (void (APIENTRY *)(GLuint,GLfloat,GLfloat,GLfloat))0;

void APIENTRY (* GL::glVertexAttrib3s)(GLuint,GLshort,GLshort,GLshort)=
  (void (APIENTRY *)(GLuint,GLshort,GLshort,GLshort))0;

void APIENTRY (* GL::glVertexAttrib3d)(GLuint,GLdouble,GLdouble,GLdouble)=
  (void (APIENTRY *)(GLuint,GLdouble,GLdouble,GLdouble))0;

void APIENTRY (* GL::glVertexAttrib4f)(GLuint,GLfloat,GLfloat,GLfloat,GLfloat)=
  (void (APIENTRY *)(GLuint,GLfloat,GLfloat,GLfloat,GLfloat))0;

void APIENTRY (* GL::glVertexAttrib4s)(GLuint,GLshort,GLshort,GLshort,GLshort)=
  (void (APIENTRY *)(GLuint,GLshort,GLshort,GLshort,GLshort))0;

void APIENTRY (* GL::glVertexAttrib4d)(GLuint,
                                       GLdouble,GLdouble,GLdouble,GLdouble)=
  (void (APIENTRY *)(GLuint,GLdouble,GLdouble,GLdouble,GLdouble))0;

void APIENTRY (* GL::glVertexAttrib4Nub)(GLuint,
                                         GLubyte,GLubyte,GLubyte,GLubyte)=
  (void (APIENTRY *)(GLuint,GLubyte,GLubyte,GLubyte,GLubyte))0;

void APIENTRY (* GL::glVertexAttrib1fv)(GLuint,const GLfloat *)=
  (void (APIENTRY *)(GLuint,const GLfloat *))0;

void APIENTRY (* GL::glVertexAttrib1sv)(GLuint,const GLshort *)=
  (void (APIENTRY *)(GLuint,const GLshort *))0;

void APIENTRY (* GL::glVertexAttrib1dv)(GLuint,const GLdouble *)=
  (void (APIENTRY *)(GLuint,const GLdouble *))0;

void APIENTRY (* GL::glVertexAttrib2fv)(GLuint,const GLfloat *)=
  (void (APIENTRY *)(GLuint,const GLfloat *))0;

void APIENTRY (* GL::glVertexAttrib2sv)(GLuint,const GLshort *)=
  (void (APIENTRY *)(GLuint,const GLshort *))0;

void APIENTRY (* GL::glVertexAttrib2dv)(GLuint,const GLdouble *)=
  (void (APIENTRY *)(GLuint,const GLdouble *))0;

void APIENTRY (* GL::glVertexAttrib3fv)(GLuint,const GLfloat *)=
  (void (APIENTRY *)(GLuint,const GLfloat *))0;

void APIENTRY (* GL::glVertexAttrib3sv)(GLuint,const GLshort *)=
  (void (APIENTRY *)(GLuint,const GLshort *))0;

void APIENTRY (* GL::glVertexAttrib3dv)(GLuint,const GLdouble *)=
  (void (APIENTRY *)(GLuint,const GLdouble *))0;

void APIENTRY (* GL::glVertexAttrib4fv)(GLuint,const GLfloat *)=
  (void (APIENTRY *)(GLuint,const GLfloat *))0;

void APIENTRY (* GL::glVertexAttrib4sv)(GLuint,const GLshort *)=
  (void (APIENTRY *)(GLuint,const GLshort *))0;

void APIENTRY (* GL::glVertexAttrib4dv)(GLuint,const GLdouble *)=
  (void (APIENTRY *)(GLuint,const GLdouble *))0;

void APIENTRY (* GL::glVertexAttrib4iv)(GLuint,const GLint *)=
  (void (APIENTRY *)(GLuint,const GLint *))0;

void APIENTRY (* GL::glVertexAttrib4bv)(GLuint,const GLbyte *)=
  (void (APIENTRY *)(GLuint,const GLbyte *))0;

void APIENTRY (* GL::glVertexAttrib4ubv)(GLuint,const GLubyte *)=
  (void (APIENTRY *)(GLuint,const GLubyte *))0;

void APIENTRY (* GL::glVertexAttrib4usv)(GLuint,const GLushort *)=
  (void (APIENTRY *)(GLuint,const GLushort *))0;

void APIENTRY (* GL::glVertexAttrib4uiv)(GLuint,const GLuint *)=
  (void (APIENTRY *)(GLuint,const GLuint *))0;

void APIENTRY (* GL::glVertexAttrib4Nbv)(GLuint,const GLbyte *)=
  (void (APIENTRY *)(GLuint,const GLbyte *))0;

void APIENTRY (* GL::glVertexAttrib4Nsv)(GLuint,const GLshort *)=
  (void (APIENTRY *)(GLuint,const GLshort *))0;

void APIENTRY (* GL::glVertexAttrib4Niv)(GLuint,const GLint *)=
  (void (APIENTRY *)(GLuint,const GLint *))0;

void APIENTRY (* GL::glVertexAttrib4Nubv)(GLuint,const GLubyte *)=
  (void (APIENTRY *)(GLuint,const GLubyte *))0;

void APIENTRY (* GL::glVertexAttrib4Nusv)(GLuint,const GLushort *)=
  (void (APIENTRY *)(GLuint,const GLushort *))0;

void APIENTRY (* GL::glVertexAttrib4Nuiv)(GLuint,const GLuint *)=
  (void (APIENTRY *)(GLuint,const GLuint *))0;

void APIENTRY (* GL::glVertexAttribPointer)(GLuint,GLint,GLenum,
                                            GLboolean,GLsizei,const GLvoid *)=
  (void (APIENTRY *)(GLuint,GLint,GLenum,GLboolean,GLsizei,const GLvoid *))0;

void APIENTRY (* GL::glEnableVertexAttribArray)(GLuint)=
  (void (APIENTRY *)(GLuint))0;

void APIENTRY (* GL::glDisableVertexAttribArray)(GLuint)=
  (void (APIENTRY *)(GLuint))0;

void APIENTRY (* GL::glBindAttribLocation)(GL::SLhandle,GLuint,
                                           const GL::SLchar *)=
  (void (APIENTRY *)(GL::SLhandle,GLuint,const GL::SLchar *))0;

void APIENTRY (* GL::glGetActiveAttrib)(GL::SLhandle,GLuint,GLsizei,GLsizei *,
                                        GLint *,GLenum *,GL::SLchar *)=
  (void (APIENTRY *)(GL::SLhandle,GLuint,GLsizei,GLsizei *,
                     GLint *,GLenum *,GL::SLchar *))0;

GLint APIENTRY (* GL::glGetAttribLocation)(GL::SLhandle,const GL::SLchar *)=
  (GLint (APIENTRY *)(GL::SLhandle,const GL::SLchar *))0;

void APIENTRY (* GL::glGetVertexAttribfv)(GLuint,GLenum,GLfloat *)=
  (void (APIENTRY *)(GLuint,GLenum,GLfloat *))0;

void APIENTRY (* GL::glGetVertexAttribiv)(GLuint,GLenum,GLint *)=
  (void (APIENTRY *)(GLuint,GLenum,GLint *))0;

void APIENTRY (* GL::glGetVertexAttribdv)(GLuint,GLenum,GLdouble *)=
  (void (APIENTRY *)(GLuint,GLenum,GLdouble *))0;

void APIENTRY (* GL::glGetVertexAttribPointerv)(GLuint,GLenum,GLvoid **)=
  (void (APIENTRY *)(GLuint,GLenum,GLvoid **))0;

//-------- GL_ARB_fragment_shader --------------------------------------------

unsigned int GL::_nbFragmentUniformComponents=0;

//-------- GL_ARB_shader_objects ---------------------------------------------

void APIENTRY (* GL::glDeleteObject)(GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle))0;

GL::SLhandle APIENTRY (* GL::glGetHandle)(GLenum)=
  (GL::SLhandle (APIENTRY *)(GLenum))0;

void APIENTRY (* GL::glDetachObject)(GL::SLhandle, GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle, GL::SLhandle))0;

GL::SLhandle APIENTRY (* GL::glCreateShaderObject)(GLenum)=
  (GL::SLhandle (APIENTRY *)(GLenum))0;

void APIENTRY (* GL::glShaderSource)(GL::SLhandle,GLsizei,
                                     const GL::SLchar **,const GLint *)=
  (void (APIENTRY *)(GL::SLhandle,GLsizei,const GL::SLchar **,const GLint *))0;

void APIENTRY (* GL::glCompileShader)(GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle))0;

GL::SLhandle APIENTRY (* GL::glCreateProgramObject)(void)=
  (GL::SLhandle (APIENTRY *)(void))0;

void APIENTRY (* GL::glAttachObject)(GL::SLhandle,GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle,GL::SLhandle))0;

void APIENTRY (* GL::glLinkProgram)(GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle))0;

void APIENTRY (* GL::glUseProgramObject)(GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle))0;

void APIENTRY (* GL::glValidateProgram)(GL::SLhandle)=
  (void (APIENTRY *)(GL::SLhandle))0;

void APIENTRY (* GL::glUniform1f)(GLint,GLfloat)=
  (void (APIENTRY *)(GLint,GLfloat))0;

void APIENTRY (* GL::glUniform2f)(GLint,GLfloat,GLfloat)=
  (void (APIENTRY *)(GLint,GLfloat,GLfloat))0;

void APIENTRY (* GL::glUniform3f)(GLint,GLfloat,GLfloat,GLfloat)=
  (void (APIENTRY *)(GLint,GLfloat,GLfloat,GLfloat))0;

void APIENTRY (* GL::glUniform4f)(GLint,GLfloat,GLfloat,GLfloat,GLfloat)=
  (void (APIENTRY *)(GLint,GLfloat,GLfloat,GLfloat,GLfloat))0;

void APIENTRY (* GL::glUniform1i)(GLint,GLint)=
  (void (APIENTRY *)(GLint,GLint))0;

void APIENTRY (* GL::glUniform2i)(GLint,GLint,GLint)=
  (void (APIENTRY *)(GLint,GLint,GLint))0;

void APIENTRY (* GL::glUniform3i)(GLint,GLint,GLint,GLint)=
  (void (APIENTRY *)(GLint,GLint,GLint,GLint))0;

void APIENTRY (* GL::glUniform4i)(GLint,GLint,GLint,GLint,GLint)=
  (void (APIENTRY *)(GLint,GLint,GLint,GLint,GLint))0;

void APIENTRY (* GL::glUniform1fv)(GLint,GLsizei,GLfloat *)=
  (void (APIENTRY *)(GLint,GLsizei,GLfloat *))0;

void APIENTRY (* GL::glUniform2fv)(GLint,GLsizei,GLfloat *)=
  (void (APIENTRY *)(GLint,GLsizei,GLfloat *))0;

void APIENTRY (* GL::glUniform3fv)(GLint,GLsizei,GLfloat *)=
  (void (APIENTRY *)(GLint,GLsizei,GLfloat *))0;

void APIENTRY (* GL::glUniform4fv)(GLint,GLsizei,GLfloat *)=
  (void (APIENTRY *)(GLint,GLsizei,GLfloat *))0;

void APIENTRY (* GL::glUniform1iv)(GLint,GLsizei,GLint *)=
  (void (APIENTRY *)(GLint,GLsizei,GLint *))0;

void APIENTRY (* GL::glUniform2iv)(GLint,GLsizei,GLint *)=
  (void (APIENTRY *)(GLint,GLsizei,GLint *))0;

void APIENTRY (* GL::glUniform3iv)(GLint,GLsizei,GLint *)=
  (void (APIENTRY *)(GLint,GLsizei,GLint *))0;

void APIENTRY (* GL::glUniform4iv)(GLint,GLsizei,GLint *)=
  (void (APIENTRY *)(GLint,GLsizei,GLint *))0;

void APIENTRY (* GL::glUniformMatrix2fv)(GLint,GLsizei,GLboolean,GLfloat *)=
  (void (APIENTRY *)(GLint,GLsizei,GLboolean,GLfloat *))0;

void APIENTRY (* GL::glUniformMatrix3fv)(GLint,GLsizei,GLboolean,GLfloat *)=
  (void (APIENTRY *)(GLint, GLsizei,GLboolean,GLfloat *))0;

void APIENTRY (* GL::glUniformMatrix4fv)(GLint,GLsizei,GLboolean,GLfloat *)=
  (void (APIENTRY *)(GLint,GLsizei,GLboolean,GLfloat *))0;

void APIENTRY (* GL::glGetObjectParameterfv)(GL::SLhandle,GLenum,GLfloat *)=
  (void (APIENTRY *)(GL::SLhandle,GLenum,GLfloat *))0;

void APIENTRY (* GL::glGetObjectParameteriv)(GL::SLhandle,GLenum,GLint *)=
  (void (APIENTRY *)(GL::SLhandle,GLenum,GLint *))0;

void APIENTRY (* GL::glGetInfoLog)(GL::SLhandle,GLsizei,
                                   GLsizei *,GL::SLchar *)=
  (void (APIENTRY *)(GL::SLhandle,GLsizei,GLsizei *,GL::SLchar *))0;

void APIENTRY (* GL::glGetAttachedObjects)(GL::SLhandle,GLsizei,
                                           GLsizei *,GL::SLhandle *)=
  (void (APIENTRY *)(GL::SLhandle,GLsizei,GLsizei *,GL::SLhandle *))0;

GLint APIENTRY (* GL::glGetUniformLocation)(GL::SLhandle,const GL::SLchar *)=
  (GLint (APIENTRY *)(GL::SLhandle,const GL::SLchar *))0;

void APIENTRY (* GL::glGetActiveUniform)(GL::SLhandle,GLuint,GLsizei,GLsizei *,
                                         GLint *,GLenum *,GL::SLchar *)=
  (void (APIENTRY *)(GL::SLhandle,GLuint,GLsizei,GLsizei *,
                     GLint *,GLenum *,GL::SLchar *))0;

void APIENTRY (* GL::glGetUniformfv)(GL::SLhandle,GLint,GLfloat *)=
  (void (APIENTRY *)(GL::SLhandle,GLint,GLfloat *))0;

void APIENTRY (* GL::glGetUniformiv)(GL::SLhandle,GLint,GLint *)=
  (void (APIENTRY *)(GL::SLhandle,GLint,GLint *))0;

void APIENTRY (* GL::glGetShaderSource)(GL::SLhandle,GLsizei,
                                        GLsizei *,GL::SLchar *)=
  (void (APIENTRY *)(GL::SLhandle,GLsizei,GLsizei *,GL::SLchar *))0;

//-------- GL_EXT_framebuffer_object -----------------------------------------

bool GL::_hasFrameBufferObject=false;

void APIENTRY (* GL::glGenFramebuffers)(GLsizei,GLuint *)=
  (void (APIENTRY *)(GLsizei,GLuint *))0;

void APIENTRY (* GL::glDeleteFramebuffers)(GLsizei, const GLuint *)=
  (void (APIENTRY *)(GLsizei, const GLuint *))0;

void APIENTRY (* GL::glGenRenderbuffers)(GLsizei,GLuint *)=
  (void (APIENTRY *)(GLsizei,GLuint *))0;

void APIENTRY (* GL::glDeleteRenderbuffers)(GLsizei,const GLuint *)=
  (void (APIENTRY *)(GLsizei,const GLuint *))0;

void APIENTRY (* GL::glBindFramebuffer)(GLenum,GLuint)=
  (void (APIENTRY *)(GLenum,GLuint))0;

void APIENTRY (* GL::glFramebufferTexture2D)(GLenum,GLenum,GLenum,
                                             GLuint,GLint)=
  (void (APIENTRY *)(GLenum,GLenum,GLenum,GLuint,GLint))0;

void APIENTRY (* GL::glBindRenderbuffer)(GLenum,GLuint)=
  (void (APIENTRY *)(GLenum,GLuint))0;

void APIENTRY (* GL::glRenderbufferStorage)(GLenum,GLenum,GLsizei,GLsizei)=
  (void (APIENTRY *)(GLenum,GLenum,GLsizei,GLsizei))0;

void APIENTRY (* GL::glFramebufferRenderbuffer)(GLenum,GLenum,GLenum,GLuint)=
  (void (APIENTRY *)(GLenum,GLenum,GLenum,GLuint))0;

GLenum APIENTRY (* GL::glCheckFramebufferStatus)(GLenum)=
  (GLenum (APIENTRY *)(GLenum))0;


//-------- GL_ARB_vertex_buffer_object ---------------------------------------

bool GL::_hasVertexBufferObject=false;

void APIENTRY (* GL::glBindBuffer)(GLenum, GLuint)=
  (void (APIENTRY *)(GLenum, GLuint))0;

void APIENTRY (* GL::glDeleteBuffers)(GLsizei,const GLuint *)=
  (void (APIENTRY *)(GLsizei,const GLuint *))0;

void APIENTRY (* GL::glGenBuffers)(GLsizei,GLuint *)=
  (void (APIENTRY *)(GLsizei,GLuint *))0;

GLboolean APIENTRY (* GL::glIsBuffer)(GLuint)=
  (GLboolean (APIENTRY *)(GLuint))0;

void APIENTRY (* GL::glBufferData)(GLenum,VBOsizeiptr,const GLvoid *,GLenum)=
  (void (APIENTRY *)(GLenum,VBOsizeiptr,const GLvoid *,GLenum))0;

void APIENTRY (* GL::glBufferSubData)(GLenum,VBOintptr,VBOsizeiptr,
                                      const GLvoid *)=
  (void (APIENTRY *)(GLenum,VBOintptr,VBOsizeiptr,const GLvoid *))0;

void APIENTRY (* GL::glGetBufferSubData)(GLenum,VBOintptr,VBOsizeiptr,
                                         GLvoid *)=
  (void (APIENTRY *)(GLenum, VBOintptr, VBOsizeiptr, GLvoid *))0;

GLvoid * APIENTRY (* GL::glMapBuffer)(GLenum,GLenum)=
  (GLvoid * (APIENTRY *)(GLenum,GLenum))0;

GLboolean APIENTRY (* GL::glUnmapBuffer)(GLenum)=
  (GLboolean (APIENTRY *)(GLenum))0;

void APIENTRY (* GL::glGetBufferParameteriv)(GLenum,GLenum,GLint *)=
  (void (APIENTRY *)(GLenum,GLenum,GLint *))0;

void APIENTRY (* GL::glGetBufferPointerv)(GLenum,GLenum,GLvoid **)=
  (void (APIENTRY *)(GLenum,GLenum,GLvoid **))0;

//-------- GL_ARB_point_parameters -------------------------------------------

bool GL::_hasPointSprite=false;

void APIENTRY (* GL::glPointParameterf)(GLenum,GLfloat)=
  (void (APIENTRY *)(GLenum,GLfloat))0;

void APIENTRY (* GL::glPointParameterfv)(GLenum,GLfloat *)=
  (void (APIENTRY *)(GLenum,GLfloat *))0;

//-------- Forbidden operations ----------------------------------------------

void * // GL function pointer
GL::_getFunctionPtr(const char * procName) // Private
{
#ifdef __CYGWIN__
return((void *)wglGetProcAddress(procName));
#else // not __CYGWIN__
if(GL::_getProcAddress)
  {
  void (*fnct)(void)=(*GL::_getProcAddress)(procName);
  void * result;
  memcpy(&result,&fnct,sizeof(void *));
  return(result);
  }
else
  {
  void * proc=(void *)0;
  void * currentBinary=ArSystem::openLibrary("",false);
  if(currentBinary)
    {
    proc=ArSystem::getLibrarySymbol(currentBinary,procName);
    ArSystem::closeLibrary(currentBinary);
    }
  if(!proc) // not found in the current binary... may be the AReVi application
    {       // is a dynamic library (which is linked against GL)...
    currentBinary=ArSystem::openLibrary("GL",true);
    if(currentBinary)
      {
      proc=ArSystem::getLibrarySymbol(currentBinary,procName);
      ArSystem::closeLibrary(currentBinary);
      }
    }
  return(proc);
  }
#endif // not __CYGWIN__
}

void (*(*GL::_getProcAddress)(const char *))(void)=
  (void (*(*)(const char *))(void))0;

} // namespace AReVi;

//----------------------------------------------------------------------------
