// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/index3D.cpp
// author(s)   : F. Harrouet & K. Manac'h
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/index3D.h"

#include "AReVi/Lib3D/base3D.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Index3D (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Index3D,ArObject)

//-------- Construction / Destruction ----------------------------------------

Index3D::Index3D(ArCW & arCW,
                 double unitSize)
: ArObject(arCW),
  _unitFactor(unitSize>0 ? 1.0/unitSize : 1.0),
  _grids(),
  _cellChangeCB(thisRef())
{
  // Nothing to be done
}

Index3D::~Index3D(void)
{
// _removeAllElements() should be called from derived classes !
}

//-------- Detection ---------------------------------------------------------

double // reference cell size
Index3D::getUnitSize(void) const
{
return(1.0/_unitFactor);
}

void
Index3D::getCellContent(ArConstRef<ArClass> elemClass,
                        double x,
                        double y,
                        double z,
                        double precision,
                        StlVector<ArRef<Base3D> > & elemsOut) const
{
elemsOut.clear();
if(precision<=0.0)
  {
  return;
  }
int level=(int)ceil(log(_unitFactor*precision)/M_LN2);
double factor=_unitFactor/pow(2.0,level);
int xi=(int)floor(x*factor);
int yi=(int)floor(y*factor);
int zi=(int)floor(z*factor);
if(!elemClass)
  {
  elemClass=Base3D::CLASS();
  }
StlVector<ArConstRef<ArClass> > derivedClasses;
elemClass->getDescendants(derivedClasses);
derivedClasses.push_back(elemClass);
for(unsigned int i=derivedClasses.size();i--;)
  {
  Index3D::GridMap::const_iterator it=_grids.find(derivedClasses[i].c_ptr());
  if(it!=_grids.end())
    {
    const Index3D::Grid::Cell * cell=(*it).second
                                     ->getCellIfExists(xi,yi,zi,level);
    if(cell)
      {
      for(unsigned int j=cell->size();j--;)
        {
        elemsOut.push_back((*cell)[j]->thisRef());
        }
      }
    }
  }
}

void
Index3D::getAll(ArConstRef<ArClass> elemClass,
                StlVector<ArRef<Base3D> > & elemsOut) const
{
elemsOut.clear();
if(!elemClass)
  {
  elemClass=Base3D::CLASS();
  }
StlVector<ArConstRef<ArClass> > derivedClasses;
elemClass->getDescendants(derivedClasses);
derivedClasses.push_back(elemClass);
for(unsigned int i=derivedClasses.size();i--;)
  {
  Index3D::GridMap::const_iterator it=_grids.find(derivedClasses[i].c_ptr());
  if(it!=_grids.end())
    {
    const Index3D::Grid * grid=(*it).second;
    for(unsigned int j=grid->elems.size();j--;)
      {
      elemsOut.push_back(grid->elems[j]);
      }
    }
  }
}

//-------- Content management ------------------------------------------------

// bool // success
// Index3D::addElement(ArRef<Base3D> elem,
//                     double precision)=0

// void
// Index3D::removeElement(ArRef<Base3D> elem)=0;

double // precision with which element is contained or 0
Index3D::getElementPrecision(ArConstRef<Base3D> elem) const
{
if(elem)
  {
  for(unsigned int i=elem->_indexes.size();i--;)
    {
    if(elem->_indexes[i].index.c_ptr()==this)
      {
      return(elem->_indexes[i].precision);
      }
    }
  }
return(0.0);
}

//-------- Forbidden operations ----------------------------------------------

void
Index3D::_removeAllElements(void) // Protected
{
while(_grids.empty())
  {
  Index3D::GridMap::iterator it=_grids.begin();
  Index3D::Grid * grid=(*it).second;
  while(!grid->elems.empty())
    {
    removeElement(grid->elems.back()); // cells/axis are destroyed here
    }
  delete grid;
  _grids.erase(it);
  }
}

// void
// Index3D::_onMotion(Base3D * elem,
//                    void * gridData,
//                    double factor,
//                    int level)=0 // Protected


//----------------------------------------------------------------------------
// Class PointIndex3D
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(PointIndex3D,Index3D)

//-------- Construction / Destruction ----------------------------------------

PointIndex3D::PointIndex3D(ArCW & arCW,
                           double unitSize) // Protected
: Index3D(arCW,unitSize)
{
  // Nothing to be done
}

PointIndex3D::~PointIndex3D(void) // Protected
{
_removeAllElements();
}

//-------- Detection ---------------------------------------------------------

void
PointIndex3D::getApproxNeighborhood(ArConstRef<ArClass> elemClass,
                                    double x,
                                    double y,
                                    double z,
                                    double range,
                                    StlVector<ArRef<Base3D> > & elemsOut) const
{
elemsOut.clear();
if(range<=0.0)
  {
  return;
  }
double xIn=x;
double yIn=y;
double zIn=z;
double rangeIn=range;
if(!elemClass)
  {
  elemClass=Base3D::CLASS();
  }
StlVector<ArConstRef<ArClass> > derivedClasses;
elemClass->getDescendants(derivedClasses);
derivedClasses.push_back(elemClass);
for(unsigned int i=derivedClasses.size();i--;)
  {
  Index3D::GridMap::const_iterator it=_grids.find(derivedClasses[i].c_ptr());
  if(it!=_grids.end())
    {
    const Index3D::Grid::GAxis * gAxis=&(*it).second->grid;
    if(!gAxis->g.empty())
      {
      double factor=_unitFactor/pow(2.0,gAxis->offset);
      x=xIn*factor;
      y=yIn*factor;
      z=zIn*factor;
      range=rangeIn*factor;
      int xMin=(int)floor(x-range);
      int yMin=(int)floor(y-range);
      int zMin=(int)floor(z-range);
      int xMax=(int)floor(x+range);
      int yMax=(int)floor(y+range);
      int zMax=(int)floor(z+range);
      unsigned int endLevel=gAxis->g.size();
      for(unsigned int g=0;
          g<endLevel;
          ++g,xMin>>=1,yMin>>=1,zMin>>=1,xMax>>=1,yMax>>=1,zMax>>=1)
        {
        const Index3D::Grid::ZAxis * zAxis=gAxis->g[g];
        if(zAxis&&!zAxis->z.empty())
          {
          for(int z=zMax;z>=zMin;--z)
            {
            if(z<zAxis->offset)
              {
              break; // other z will be even smaller
              }
            if(z>=zAxis->offset+(int)zAxis->z.size())
              {
              continue;
              }
            const Index3D::Grid::YAxis * yAxis=zAxis->z[z-zAxis->offset];
            if(!yAxis||yAxis->y.empty())
              {
              continue;
              }
            for(int y=yMax;y>=yMin;--y)
              {
              if(y<yAxis->offset)
                {
                break; // other y will be even smaller
                }
              if(y>=yAxis->offset+(int)yAxis->y.size())
                {
                continue;
                }
              const Index3D::Grid::XAxis * xAxis=yAxis->y[y-yAxis->offset];
              if(!xAxis||xAxis->x.empty())
                {
                continue;
                }
              for(int x=xMax;x>=xMin;--x)
                {
                if(x<xAxis->offset)
                  {
                  break; // other x will be even smaller
                  }
                if(x>=xAxis->offset+(int)xAxis->x.size())
                  {
                  continue;
                  }
                const Index3D::Grid::Cell * cell=xAxis->x[x-xAxis->offset];
                if(cell)
                  {
                  for(unsigned int j=cell->size();j--;)
                    {
                    elemsOut.push_back((*cell)[j]->thisRef());
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}

void
PointIndex3D::viewFirst(ArConstRef<Base3D> origin,
                        double yawVision,
                        double pitchVision,
                        double range,
                        ArConstRef<ArClass> elemClass,
                        ArRef<Base3D> & elemOut,
                        double & yawAngleOut,
                        double & pitchAngleOut,
                        double & distanceOut) const
{
double px,py,pz;
origin->_base.getPosition(px,py,pz);
StlVector<ArRef<Base3D> > candidates;
getApproxNeighborhood(elemClass,px,py,pz,range,candidates);
origin->viewFirst(yawVision,pitchVision,range,candidates,
                  elemOut,yawAngleOut,pitchAngleOut,distanceOut);
}

void
PointIndex3D::view(ArConstRef<Base3D> origin,
                   double yawVision,
                   double pitchVision,
                   double range,
                   ArConstRef<ArClass> elemClass,
                   StlVector<ArRef<Base3D> > & elemsOut,
                   StlVector<double> & yawAnglesOut,
                   StlVector<double> & pitchAnglesOut,
                   StlVector<double> & distancesOut) const
{
double px,py,pz;
origin->_base.getPosition(px,py,pz);
StlVector<ArRef<Base3D> > candidates;
getApproxNeighborhood(elemClass,px,py,pz,range,candidates);
origin->view(yawVision,pitchVision,range,candidates,
             elemsOut,yawAnglesOut,pitchAnglesOut,distancesOut);
}

//-------- Content management ------------------------------------------------

bool // success
PointIndex3D::addElement(ArRef<Base3D> elem,
                         double precision)
{
if(precision<=0.0)
  {
  return(false);
  }
const ArClass * elemClass=elem->getClass().c_ptr();
Index3D::GridMap::iterator it=_grids.find(elemClass);
if(it==_grids.end())
  {
  it=_grids.insert(make_pair(elemClass,new Index3D::Grid)).first;
  }
Index3D::Grid * grid=(*it).second;
int level=(int)ceil(log(_unitFactor*precision)/M_LN2);
for(unsigned int i=elem->_indexes.size();i--;)
  {
  if(elem->_indexes[i].index.c_ptr()==this)
    {
    if(elem->_indexes[i].level==level)
      {
      return(true);
      }
    else
      {
      removeElement(elem);
      }
    break;
    }
  }
double factor=_unitFactor/pow(2.0,level);
elem->_addIndex(thisRef(),grid,precision,factor,level);
grid->elems.push_back(elem);
elem->_base.getPosition(elem->_oldPos.x,elem->_oldPos.y,elem->_oldPos.z);
grid->getOrCreateCell((int)floor(elem->_oldPos.x*factor),
                      (int)floor(elem->_oldPos.y*factor),
                      (int)floor(elem->_oldPos.z*factor),
                      level)->push_back(elem.c_ptr());
if(!_cellChangeCB.empty())
  {
  Index3D::CellChangeEvent evt;
  evt.element=elem;
  evt.oldX=elem->_oldPos.x;
  evt.oldY=elem->_oldPos.y;
  evt.oldZ=elem->_oldPos.z;
  evt.enterIndex=true;
  evt.leaveIndex=false;
  _cellChangeCB.fire(evt);
  }
return(true);
}

void
PointIndex3D::removeElement(ArRef<Base3D> elem)
{
Index3D::GridMap::iterator it=_grids.find(elem->getClass().c_ptr());
if(it!=_grids.end())
  {
  Index3D::Grid * grid=(*it).second;
  for(unsigned int i=grid->elems.size();i--;)
    {
    if(grid->elems[i]==elem)
      {
      double factor=0.0;
      int level=0;
      unsigned int j;
      for(j=elem->_indexes.size();j--;)
        {
        if(elem->_indexes[j].index.c_ptr()==this)
          {
          factor=elem->_indexes[j].factor;
          level=elem->_indexes[j].level;
          break;
          }
        }
      int x=(int)floor(elem->_oldPos.x*factor);
      int y=(int)floor(elem->_oldPos.y*factor);
      int z=(int)floor(elem->_oldPos.z*factor);
      Index3D::Grid::Cell * cell=grid->getExistingCell(x,y,z,level);
      for(j=cell->size();j--;)
        {
        if((*cell)[j]==elem.c_ptr())
          {
          StlVectorFastErase(*cell,j);
          if(cell->empty())
            {
            grid->removeCell(x,y,z,level);
            }
          break;
          }
        }
      StlVectorFastErase(grid->elems,i);
      ArRef<PointIndex3D> flag(this);
      elem->_removeIndex(thisRef());
      if(flag&&!_cellChangeCB.empty())
        {
        Index3D::CellChangeEvent evt;
        evt.element=elem;
        evt.oldX=elem->_oldPos.x;
        evt.oldY=elem->_oldPos.y;
        evt.oldZ=elem->_oldPos.z;
        evt.enterIndex=false;
        evt.leaveIndex=true;
        _cellChangeCB.fire(evt);
        }
      break;
      }
    }
  }
}

//-------- Forbidden operations ----------------------------------------------

void
PointIndex3D::_onMotion(Base3D * elem,
                        void * gridData,
                        double factor,
                        int level) // Protected
{
Index3D::Grid * grid=(Index3D::Grid *)gridData;
int x1=(int)floor(elem->_oldPos.x*factor);
int y1=(int)floor(elem->_oldPos.y*factor);
int z1=(int)floor(elem->_oldPos.z*factor);
double xe,ye,ze;
elem->_base.getPosition(xe,ye,ze);
int x2=(int)floor(xe*factor);
int y2=(int)floor(ye*factor);
int z2=(int)floor(ze*factor);
if((x1!=x2)||(y1!=y2)||(z1!=z2))
  {
  grid->getOrCreateCell(x2,y2,z2,level)->push_back(elem);
  Index3D::Grid::Cell * cell=grid->getExistingCell(x1,y1,z1,level);
  for(unsigned int i=cell->size();i--;)
    {
    if((*cell)[i]==elem)
      {
      StlVectorFastErase(*cell,i);
      if(cell->empty())
        {
        grid->removeCell(x1,y1,z1,level);
        }
      break;
      }
    }
  if(!_cellChangeCB.empty())
    {
    Index3D::CellChangeEvent evt;
    evt.element=elem->thisRef();
    evt.oldX=elem->_oldPos.x;
    evt.oldY=elem->_oldPos.y;
    evt.oldZ=elem->_oldPos.z;
    evt.enterIndex=false;
    evt.leaveIndex=false;
    _cellChangeCB.fire(evt);
    }
  }
}


//----------------------------------------------------------------------------
// Class VolumeIndex3D
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(VolumeIndex3D,Index3D)

//-------- Construction / Destruction ----------------------------------------

VolumeIndex3D::VolumeIndex3D(ArCW & arCW,
                             double unitSize) // Protected
: Index3D(arCW,unitSize)
{
  // Nothing to be done
}

VolumeIndex3D::~VolumeIndex3D(void) // Protected
{
_removeAllElements();
}

//-------- Detection ---------------------------------------------------------

void
VolumeIndex3D::getApproxCoverage(ArConstRef<ArClass> elemClass,
                                 double x,
                                 double y,
                                 double z,
                                 StlVector<ArRef<Base3D> > & elemsOut) const
{
elemsOut.clear();
if(!elemClass)
  {
  elemClass=Base3D::CLASS();
  }
StlVector<ArConstRef<ArClass> > derivedClasses;
elemClass->getDescendants(derivedClasses);
derivedClasses.push_back(elemClass);
for(unsigned int i=derivedClasses.size();i--;)
  {
  Index3D::GridMap::const_iterator it=_grids.find(derivedClasses[i].c_ptr());
  if(it!=_grids.end())
    {
    Index3D::Grid * grid=(*it).second;
    const Index3D::Grid::GAxis * gAxis=&grid->grid;
    if(!gAxis->g.empty())
      {
      double factor=_unitFactor/pow(2.0,gAxis->offset);
      int xi=(int)floor(x*factor);
      int yi=(int)floor(y*factor);
      int zi=(int)floor(z*factor);
      unsigned int endLevel=gAxis->offset+gAxis->g.size();
      for(unsigned int g=gAxis->offset;
          g<endLevel;
          ++g,xi>>=1,yi>>=1,zi>>=1)
        {
        const Index3D::Grid::Cell * cell=grid->getCellIfExists(xi,yi,zi,g);
        if(cell)
          {
          for(unsigned int j=cell->size();j--;)
            {
            elemsOut.push_back((*cell)[j]->thisRef());
            }
          }
        }
      }
    }
  }
}

void
VolumeIndex3D::getApproxNeighborhood(ArConstRef<ArClass> elemClass,
                                     double x,
                                     double y,
                                     double z,
                                     double range,
                                     StlVector<ArRef<Base3D> > & elemsOut)
                                                                         const
{
elemsOut.clear();
if(range<=0.0)
  {
  return;
  }
double xIn=x;
double yIn=y;
double zIn=z;
double rangeIn=range;
if(!elemClass)
  {
  elemClass=Base3D::CLASS();
  }
StlVector<ArConstRef<ArClass> > derivedClasses;
elemClass->getDescendants(derivedClasses);
derivedClasses.push_back(elemClass);
for(unsigned int i=derivedClasses.size();i--;)
  {
  Index3D::GridMap::const_iterator it=_grids.find(derivedClasses[i].c_ptr());
  if(it!=_grids.end())
    {
    const Index3D::Grid::GAxis * gAxis=&(*it).second->grid;
    if(!gAxis->g.empty())
      {
      double factor=_unitFactor/pow(2.0,gAxis->offset);
      x=xIn*factor;
      y=yIn*factor;
      z=zIn*factor;
      range=rangeIn*factor;
      int xMin=(int)floor(x-range);
      int yMin=(int)floor(y-range);
      int zMin=(int)floor(z-range);
      int xMax=(int)floor(x+range);
      int yMax=(int)floor(y+range);
      int zMax=(int)floor(z+range);
      unsigned int endLevel=gAxis->g.size();
      for(unsigned int g=0;
          g<endLevel;
          ++g,xMin>>=1,yMin>>=1,zMin>>=1,xMax>>=1,yMax>>=1,zMax>>=1)
        {
        const Index3D::Grid::ZAxis * zAxis=gAxis->g[g];
        if(zAxis&&!zAxis->z.empty())
          {
          for(int z=zMax;z>=zMin;--z)
            {
            if(z<zAxis->offset)
              {
              break; // other z will be even smaller
              }
            if(z>=zAxis->offset+(int)zAxis->z.size())
              {
              continue;
              }
            const Index3D::Grid::YAxis * yAxis=zAxis->z[z-zAxis->offset];
            if(!yAxis||yAxis->y.empty())
              {
              continue;
              }
            for(int y=yMax;y>=yMin;--y)
              {
              if(y<yAxis->offset)
                {
                break; // other y will be even smaller
                }
              if(y>=yAxis->offset+(int)yAxis->y.size())
                {
                continue;
                }
              const Index3D::Grid::XAxis * xAxis=yAxis->y[y-yAxis->offset];
              if(!xAxis||xAxis->x.empty())
                {
                continue;
                }
              for(int x=xMax;x>=xMin;--x)
                {
                if(x<xAxis->offset)
                  {
                  break; // other x will be even smaller
                  }
                if(x>=xAxis->offset+(int)xAxis->x.size())
                  {
                  continue;
                  }
                const Index3D::Grid::Cell * cell=xAxis->x[x-xAxis->offset];
                if(cell)
                  {
                  for(unsigned int j=cell->size();j--;)
                    {
                    Base3D * elem=(*cell)[j];
                    int k;
                    for(k=elemsOut.size();k--;)
                      {
                      if(elemsOut[k].c_ptr()==elem)
                        {
                        break;
                        }
                      }
                    if(k==-1)
                      {
                      elemsOut.push_back(elem->thisRef());
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}

//-------- Content management ------------------------------------------------

bool // success
VolumeIndex3D::addElement(ArRef<Base3D> elem,
                          double precision)
{
if(precision<=0.0)
  {
  return(false);
  }
const ArClass * elemClass=elem->getClass().c_ptr();
Index3D::GridMap::iterator it=_grids.find(elemClass);
if(it==_grids.end())
  {
  it=_grids.insert(make_pair(elemClass,new Index3D::Grid)).first;
  }
Index3D::Grid * grid=(*it).second;
int level=(int)ceil(log(_unitFactor*precision)/M_LN2);
for(unsigned int i=elem->_indexes.size();i--;)
  {
  if(elem->_indexes[i].index.c_ptr()==this)
    {
    if(elem->_indexes[i].level==level)
      {
      return(true);
      }
    else
      {
      removeElement(elem);
      }
    break;
    }
  }
double factor=_unitFactor/pow(2.0,level);
elem->_addIndex(thisRef(),grid,precision,factor,level);
grid->elems.push_back(elem);
elem->_base.getPosition(elem->_oldPos.x,elem->_oldPos.y,elem->_oldPos.z);
int xMax=1+(int)floor(elem->_oldPos.x*factor);
int yMax=1+(int)floor(elem->_oldPos.y*factor);
int zMax=1+(int)floor(elem->_oldPos.z*factor);
int xMin=xMax-2;
int yMin=yMax-2;
int zMin=zMax-2;
for(int z=zMax;z>=zMin;--z)
  {
  for(int y=yMax;y>=yMin;--y)
    {
    for(int x=xMax;x>=xMin;--x)
      {
      grid->getOrCreateCell(x,y,z,level)->push_back(elem.c_ptr());
      }
    }
  }
if(!_cellChangeCB.empty())
  {
  Index3D::CellChangeEvent evt;
  evt.element=elem;
  evt.oldX=elem->_oldPos.x;
  evt.oldY=elem->_oldPos.y;
  evt.oldZ=elem->_oldPos.z;
  evt.enterIndex=true;
  evt.leaveIndex=false;
  _cellChangeCB.fire(evt);
  }
return(true);
}

void
VolumeIndex3D::removeElement(ArRef<Base3D> elem)
{
Index3D::GridMap::iterator it=_grids.find(elem->getClass().c_ptr());
if(it!=_grids.end())
  {
  Index3D::Grid * grid=(*it).second;
  for(unsigned int i=grid->elems.size();i--;)
    {
    if(grid->elems[i]==elem)
      {
      double factor=0.0;
      int level=0;
      unsigned int j;
      for(j=elem->_indexes.size();j--;)
        {
        if(elem->_indexes[j].index.c_ptr()==this)
          {
          factor=elem->_indexes[j].factor;
          level=elem->_indexes[j].level;
          break;
          }
        }
      int xMax=1+(int)floor(elem->_oldPos.x*factor);
      int yMax=1+(int)floor(elem->_oldPos.y*factor);
      int zMax=1+(int)floor(elem->_oldPos.z*factor);
      int xMin=xMax-2;
      int yMin=yMax-2;
      int zMin=zMax-2;
      for(int z=zMax;z>=zMin;--z)
        {
        for(int y=yMax;y>=yMin;--y)
          {
          for(int x=xMax;x>=xMin;--x)
            {
            Index3D::Grid::Cell * cell=grid->getExistingCell(x,y,z,level);
            for(j=cell->size();j--;)
              {
              if((*cell)[j]==elem.c_ptr())
                {
                StlVectorFastErase(*cell,j);
                if(cell->empty())
                  {
                  grid->removeCell(x,y,z,level);
                  }
                break;
                }
              }
            }
          }
        }
      StlVectorFastErase(grid->elems,i);
      ArRef<VolumeIndex3D> flag(this);
      elem->_removeIndex(thisRef());
      if(flag&&!_cellChangeCB.empty())
        {
        Index3D::CellChangeEvent evt;
        evt.element=elem;
        evt.oldX=elem->_oldPos.x;
        evt.oldY=elem->_oldPos.y;
        evt.oldZ=elem->_oldPos.z;
        evt.enterIndex=false;
        evt.leaveIndex=true;
        _cellChangeCB.fire(evt);
        }
      break;
      }
    }
  }
}

//-------- Forbidden operations ----------------------------------------------

void
VolumeIndex3D::_onMotion(Base3D * elem,
                         void * gridData,
                         double factor,
                         int level) // Protected
{
Index3D::Grid * grid=(Index3D::Grid *)gridData;
int x1Max=1+(int)floor(elem->_oldPos.x*factor);
int y1Max=1+(int)floor(elem->_oldPos.y*factor);
int z1Max=1+(int)floor(elem->_oldPos.z*factor);
double xe,ye,ze;
elem->_base.getPosition(xe,ye,ze);
int x2Max=1+(int)floor(xe*factor);
int y2Max=1+(int)floor(ye*factor);
int z2Max=1+(int)floor(ze*factor);
if((x1Max!=x2Max)||(y1Max!=y2Max)||(z1Max!=z2Max))
  {
  int x1Min=x1Max-2;
  int y1Min=y1Max-2;
  int z1Min=z1Max-2;
  int x2Min=x2Max-2;
  int y2Min=y2Max-2;
  int z2Min=z2Max-2;
  int x,y,z;
  for(z=z1Min;z<=z1Max;++z)
    {
    if((z<z2Min)||(z>z2Max))
      {
      for(y=y1Min;y<=y1Max;++y)
        {
        for(x=x1Min;x<=x1Max;++x)
          {
          Index3D::Grid::Cell * cell=grid->getExistingCell(x,y,z,level);
          for(unsigned int i=cell->size();i--;)
            {
            if((*cell)[i]==elem)
              {
              StlVectorFastErase(*cell,i);
              if(cell->empty())
                {
                grid->removeCell(x,y,z,level);
                }
              break;
              }
            }
          }
        }
      }
    else
      {
      for(y=y1Min;y<=y1Max;++y)
        {
        if((y<y2Min)||(y>y2Max))
          {
          for(x=x1Min;x<=x1Max;++x)
            {
            Index3D::Grid::Cell * cell=grid->getExistingCell(x,y,z,level);
            for(unsigned int i=cell->size();i--;)
              {
              if((*cell)[i]==elem)
                {
                StlVectorFastErase(*cell,i);
                if(cell->empty())
                  {
                  grid->removeCell(x,y,z,level);
                  }
                break;
                }
              }
            }
          }
        else
          {
          for(x=x1Min;x<=x1Max;++x)
            {
            if((x<x2Min)||(x>x2Max))
              {
              Index3D::Grid::Cell * cell=grid->getExistingCell(x,y,z,level);
              for(unsigned int i=cell->size();i--;)
                {
                if((*cell)[i]==elem)
                  {
                  StlVectorFastErase(*cell,i);
                  if(cell->empty())
                    {
                    grid->removeCell(x,y,z,level);
                    }
                  break;
                  }
                }
              }
            }
          }
        }
      }
    }
  for(z=z2Min;z<=z2Max;++z)
    {
    if((z<z1Min)||(z>z1Max))
      {
      for(y=y2Min;y<=y2Max;++y)
        {
        for(x=x2Min;x<=x2Max;++x)
          {
          grid->getOrCreateCell(x,y,z,level)->push_back(elem);
          }
        }
      }
    else
      {
      for(y=y2Min;y<=y2Max;++y)
        {
        if((y<y1Min)||(y>y1Max))
          {
          for(x=x2Min;x<=x2Max;++x)
            {
            grid->getOrCreateCell(x,y,z,level)->push_back(elem);
            }
          }
        else
          {
          for(x=x2Min;x<=x2Max;++x)
            {
            if((x<x1Min)||(x>x1Max))
              {
              grid->getOrCreateCell(x,y,z,level)->push_back(elem);
              }
            }
          }
        }
      }
    }
  if(!_cellChangeCB.empty())
    {
    Index3D::CellChangeEvent evt;
    evt.element=elem->thisRef();
    evt.oldX=elem->_oldPos.x;
    evt.oldY=elem->_oldPos.y;
    evt.oldZ=elem->_oldPos.z;
    evt.enterIndex=false;
    evt.leaveIndex=false;
    _cellChangeCB.fire(evt);
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
