// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/light3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/light3D.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/scene3D.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Light3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Light3D,Base3D)

//-------- Construction / Destruction ----------------------------------------

Light3D::Light3D(ArCW & arCW) // Protected
: Base3D(arCW),
  _scenes(),
  _on(true),
  _directional(true),
  _constantAtt(1.0),
  _linearAtt(0.0),
  _quadraticAtt(0.0),
  _spotExponent(0.0),
  _spotCutOff(180.0),
  _color(1.0,1.0,1.0),
  _setupCB(thisRef())
{
Scene3D::_globalScene->addLight(thisRef());
}

Light3D::~Light3D(void) // Protected
{
while(!_scenes.empty())
  {
  _scenes.back()->_removeLight(thisRef());
  }
}

//-------- Switching ---------------------------------------------------------

void
Light3D::setSwitch(bool onOff)
{
_on=onOff;
_postRedisplay();
_onSetup();
}

bool // light is switched on
Light3D::getSwitch(void) const
{
return(_on);
}

//-------- Light model -------------------------------------------------------

void
Light3D::setDirectional(bool directional)
{
_directional=directional;
_postRedisplay();
_onSetup();
}

bool // light is directional
Light3D::getDirectional(void) const
{
return(_directional);
}

void
Light3D::setAttenuation(double constant,
                        double linear,
                        double quadratic)
{
_constantAtt=(constant<1.0 ? 1.0 : constant);
_linearAtt=(linear<0.0 ? 0.0 : linear);
_quadraticAtt=(quadratic<0.0 ? 0.0 : quadratic);
_postRedisplay();
_onSetup();
}

void
Light3D::getAttenuation(double & constantOut,
                        double & linearOut,
                        double & quadraticOut) const
{
constantOut=_constantAtt;
linearOut=_linearAtt;
quadraticOut=_quadraticAtt;
}

void
Light3D::setSpotParameters(double exponent,
                           double cutOff)
{
_spotExponent=128.0*exponent;
if(_spotExponent<0.0)
  {
  _spotExponent=0.0;
  }
else if(_spotExponent>128.0)
  {
  _spotExponent=128.0;
  }
_spotCutOff=M_180_PI*cutOff;
if(_spotCutOff<0.0)
  {
  _spotCutOff=180.0;
  }
else if(_spotCutOff>90.0)
  {
  _spotCutOff=90.0;
  }
_postRedisplay();
_onSetup();
}

void
Light3D::getSpotParameters(double & exponentOut,
                           double & cutOffOut) const
{
exponentOut=_spotExponent/128.0;
if(_spotCutOff>90.0)
  {
  cutOffOut=-1.0;
  }
else
  {
  cutOffOut=_spotCutOff*M_PI_180;
  }
}

//-------- Color -------------------------------------------------------------

void
Light3D::setColor(double red,
                  double green,
                  double blue)
{
_color.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_color.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_color.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onSetup();
}

void
Light3D::getColor(double & redOut,
                  double & greenOut,
                  double & blueOut) const
{
redOut=_color.x;
greenOut=_color.y;
blueOut=_color.z;
}

//-------- Serialization operations ------------------------------------------

void
Light3D::serialize(SerializationDependencies & dependenciesOut,
                   ArRef<AbstractOStream> output) const
{
Base3D::serialize(dependenciesOut,output);
output->writeDouble(_constantAtt);
output->writeDouble(_linearAtt);
output->writeDouble(_quadraticAtt);
output->writeDouble(_spotExponent);
output->writeDouble(_spotCutOff);
output->writeDouble(_color.x);
output->writeDouble(_color.y);
output->writeDouble(_color.z);
unsigned int flags=(_on          ? 1<<0 : 0)+
                   (_directional ? 1<<1 : 0);
output->writeUInt(flags);
}

bool // success
Light3D::unserialize(const SerializationDependencies & dependencies,
                     ArRef<AbstractIStream> input)
{
double constantAtt,linearAtt,quadraticAtt,spotExponent,spotCutOff;
Util3D::Dbl3 color;
unsigned int flags;
if(!Base3D::unserialize(dependencies,input)||
   !input->readDouble(constantAtt)||
   !input->readDouble(linearAtt)||
   !input->readDouble(quadraticAtt)||
   !input->readDouble(spotExponent)||
   !input->readDouble(spotCutOff)||
   !input->readDouble(color.x)||
   !input->readDouble(color.y)||
   !input->readDouble(color.z)||
   !input->readUInt(flags))
  {
  return(false);
  }
_constantAtt=constantAtt;
_linearAtt=linearAtt;
_quadraticAtt=quadraticAtt;
_spotExponent=spotExponent;
_spotCutOff=spotCutOff;
_color=color;
_on=flags&(1<<0);
_directional=flags&(1<<1);
_postRedisplay(); // FIXME: needs _onSetup() ???
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Light3D::_onMotion(void) // Protected
{
_postRedisplay();
Base3D::_onMotion();
}

void
Light3D::_addScene(ArRef<Scene3D> scene) // Protected
{
_scenes.push_back(scene);
}

void
Light3D::_removeScene(ArRef<Scene3D> scene) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  if(_scenes[i]==scene)
    {
    StlVectorFastErase(_scenes,i);
    break;
    }
  }
}

void
Light3D::_postRedisplay(void) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  _scenes[i]->_postRedisplay();
  }
}

void
Light3D::_render(unsigned int lightNumber) const // Protected
{
GLenum l=(GLenum)(GL_LIGHT0+lightNumber);
const double * mtx=_base.matrix;
if(_directional)
  {
  float lightDir[]={-mtx[Util3D::R00],
                    -mtx[Util3D::R10],
                    -mtx[Util3D::R20],
                    0.0};
  glLightfv(l,GL_POSITION,lightDir);
  glLightf(l,GL_SPOT_CUTOFF,180.0);
  }
else
  {
  float lightPos[]={mtx[Util3D::TX],
                    mtx[Util3D::TY],
                    mtx[Util3D::TZ],
                    1.0};
  float lightDir[]={mtx[Util3D::R00],
                    mtx[Util3D::R10],
                    mtx[Util3D::R20]};
  glLightfv(l,GL_POSITION,lightPos);
  glLightfv(l,GL_SPOT_DIRECTION,lightDir);
  glLightf(l,GL_SPOT_EXPONENT,_spotExponent);
  glLightf(l,GL_SPOT_CUTOFF,_spotCutOff);
  glLightf(l,GL_CONSTANT_ATTENUATION,_constantAtt);
  glLightf(l,GL_LINEAR_ATTENUATION,_linearAtt);
  glLightf(l,GL_QUADRATIC_ATTENUATION,_quadraticAtt);
  }
float lightColor[]={_color.x,_color.y,_color.z,1.0};
glLightfv(l,GL_DIFFUSE,lightColor);
glLightfv(l,GL_SPECULAR,lightColor);
glLightfv(l,GL_AMBIENT,lightColor);
}

void
Light3D::_renderHeadLight(unsigned int lightNumber,
                          const Util3D::Base & base,
                          double red,
                          double green,
                          double blue) // Protected, Static
{
GLenum l=(GLenum)(GL_LIGHT0+lightNumber);
float lightDir[]={-base.matrix[Util3D::R00],
                  -base.matrix[Util3D::R10],
                  -base.matrix[Util3D::R20],
                  0.0};
glLightfv(l,GL_POSITION,lightDir);
glLightf(l,GL_SPOT_CUTOFF,180.0);
float lightColor[]={red,green,blue,1.0};
glLightfv(l,GL_DIFFUSE,lightColor);
glLightfv(l,GL_SPECULAR,lightColor);
glLightfv(l,GL_AMBIENT,lightColor);
}

void
Light3D::_onSetup(void) // Protected
{
if(!_setupCB.empty())
  {
  Light3D::SetupEvent evt;
  _setupCB.fire(evt);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
