// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/object3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/object3D.h"

#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/material3D.h"
#include "AReVi/Lib3D/abstractInteractor.h"

#include "AReVi/Shapes/shape3D.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Object3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Object3D,Base3D)

//-------- Construction / Destruction ----------------------------------------

Object3D::Object3D(ArCW & arCW) // Protected
: Base3D(arCW),
  _scenes(),
  _distances(),
  _shapes(),
  _interShape(),
  _wireFrame(false),
  _drawShapeBBox(false),
  _drawShapePartBBox(false),
  _selectedMaterial(),
  _intersectionLevel(3),
  _frameID(Renderer3D::_currentFrameID),
  _mouseButtonInteraction(true),
  _keyboardInteraction(true),
  _motionInteraction(true),
  _shapeChangeCB(thisPtr()),
  _appearanceCB(thisPtr())
{
_selectedMaterial.setTransparency(0.666);
_selectedMaterial.setDiffuseColor(0.8,0.0,0.0);
_selectedMaterial.setEmissiveColor(0.25,0.0,0.0);
_selectedMaterial.setAmbientIntensity(0.4);
Scene3D::_globalScene->addObject(thisRef());
}

Object3D::~Object3D(void) // Protected
{
for(unsigned int i=_shapes.size();i--;)
  {
  _shapes[i]->_removeObject(this);
  }
while(!_scenes.empty())
  {
  _scenes.back()->_removeObject(thisRef());
  }
}

//-------- Shape management --------------------------------------------------

void
Object3D::setShape(ArRef<Shape3D> shape)
{
while(!_shapes.empty())
  {
  _shapes.back()->_removeObject(this);
  _shapes.pop_back();
  }
_distances.clear();
if(shape)
  {
  _shapes.push_back(shape);
  shape->_addObject(this);
  }
_onShapeChange();
_postRedisplay();
}

void
Object3D::setLODShapes(const StlVector<double> & distances,
                       const StlVector<ArRef<Shape3D> > & shapes)
{
if(distances.size()==shapes.size())
  {
  while(!_shapes.empty())
    {
    _shapes.back()->_removeObject(this);
    _shapes.pop_back();
    }
  _distances=distances;
  _shapes=shapes;
  for(unsigned int i=_shapes.size();i--;)
    {
    _shapes[i]->_addObject(this);
    }
  _onShapeChange();
  _postRedisplay();
  }
}

ArConstRef<Shape3D> // first shape in LOD
Object3D::getShape(void) const
{
return(_shapes.empty() ? Shape3D::nullRef() : _shapes.front());
}

ArRef<Shape3D> // first shape in LOD
Object3D::accessShape(void)
{
return(_shapes.empty() ? Shape3D::nullRef() : _shapes.front());
}

void
Object3D::queryShapes(StlVector<double> & distancesOut,
                      StlVector<ArRef<Shape3D> > & shapesOut) const
{
distancesOut=_distances;
shapesOut=_shapes;
}

bool // currently displayed by a renderer
Object3D::isCurrentlyDisplayed(void) const
{
return(_frameID==Renderer3D::_currentFrameID);
}

//-------- Appearance --------------------------------------------------------

void
Object3D::setWireFrame(bool wireFrameOn)
{
_wireFrame=wireFrameOn;
_postRedisplay();
if(!_appearanceCB.empty())
  {
  Object3D::AppearanceEvent evt;
  _appearanceCB.fire(evt);
  }
}

bool // wire frame is on
Object3D::getWireFrame(void) const
{
return(_wireFrame);
}

void
Object3D::setShapeBoundingBox(bool shapeBoundingBoxOn)
{
_drawShapeBBox=shapeBoundingBoxOn;
_postRedisplay();
if(!_appearanceCB.empty())
  {
  Object3D::AppearanceEvent evt;
  _appearanceCB.fire(evt);
  }
}

bool // shapes bounding boxes are drawn
Object3D::getShapeBoundingBox(void) const
{
return(_drawShapeBBox);
}

void
Object3D::setShapePartBoundingBox(bool shapePartBoundingBoxOn)
{
_drawShapePartBBox=shapePartBoundingBoxOn;
_postRedisplay();
if(!_appearanceCB.empty())
  {
  Object3D::AppearanceEvent evt;
  _appearanceCB.fire(evt);
  }
}

bool // shape parts bounding boxes are drawn
Object3D::getShapePartBoundingBox(void) const
{
return(_drawShapePartBBox);
}

void
Object3D::writeSelectedMaterial(ArConstRef<Material3D> material)
{
_selectedMaterial=material->getMaterial();
_postRedisplay();
if(!_appearanceCB.empty())
  {
  Object3D::AppearanceEvent evt;
  _appearanceCB.fire(evt);
  }
}

void
Object3D::readSelectedMaterial(ArRef<Material3D> materialOut) const
{
materialOut->accessMaterial()=_selectedMaterial;
}

//-------- Detection ---------------------------------------------------------

void
Object3D::computeBoundingBox(ArRef<BoundingBox3D> bboxOut,
                             ArConstRef<Base3D> viewPoint) const
{
Util3D::BoundingBox & bbox=bboxOut->accessBoundingBox();
bbox.valid=false;
const Shape3D * shape=_findShape(viewPoint.c_ptr());
if(shape)
  {
  bbox.addTransformedBox(shape->getTransformation(),
                         shape->getBoundingBox());
  }
}

void
Object3D::setIntersectionShape(ArRef<Shape3D> shape)
{
_interShape=shape;
}

ArConstRef<Shape3D> // intersection shape
Object3D::getIntersectionShape(void) const
{
return(_interShape);
}

ArRef<Shape3D> // intersection shape
Object3D::accessIntersectionShape(void)
{
return(_interShape);
}

void
Object3D::setIntersectionLevel(unsigned int level)
{
_intersectionLevel=level;
if(_intersectionLevel>3)
  {
  _intersectionLevel=3;
  }
}

unsigned int // 0: nothing  1: shape bbox   2: shape part bbox  3: detail
Object3D::getIntersectionLevel(void) const
{
return(_intersectionLevel);
}

double // intersection distance
Object3D::intersectRay(ArConstRef<Base3D> ray,
                       unsigned int level,
                       bool findNearest,
                       ArConstRef<Base3D> viewPoint,
		       ArRef<ShapePart3D> & interPartOut,
                       unsigned int & triIndexOut) const
{
interPartOut.clear();
triIndexOut=UINT_MAX;
if(!level)
  {
  level=_intersectionLevel;
  }
if(!level)
  {
  return(-1.0);
  }
const Shape3D * shape=_findShape(viewPoint.c_ptr());
if(!shape)
  {
  return(-1.0);
  }
Util3D::Base base(false);
if(viewPoint)
  {
  shape->_findBillboardLocation(base,_base,viewPoint->getBase());
  }
else
  {
  base=_base;
  }
Util3D::Dbl3 rayP,dirP(1.0,0.0,0.0);
ray->getPosition(rayP.x,rayP.y,rayP.z);
ray->localToGlobalPosition(dirP.x,dirP.y,dirP.z);
base.globalToLocalPosition(rayP.x,rayP.y,rayP.z);
base.globalToLocalPosition(dirP.x,dirP.y,dirP.z);
return(shape->_intersectRay(level,findNearest,rayP,dirP,
                            interPartOut,triIndexOut));
}

double // intersection distance
Object3D::simpleIntersectRay(double sourceX,
                             double sourceY,
                             double sourceZ,
                             double unitDirX,
                             double unitDirY,
                             double unitDirZ,
                             unsigned int level,
                             bool findNearest,
			     ArRef<ShapePart3D> & interPartOut,
                             unsigned int & triIndexOut) const
{
interPartOut.clear();
triIndexOut=UINT_MAX;
if(!level)
  {
  level=_intersectionLevel;
  }
if(!level)
  {
  return(-1.0);
  }
const Shape3D * shape=_findShape((const Base3D *)0);
if(!shape||shape->_billboard)
  {
  return(-1.0);
  }
Util3D::Dbl3 rayP(sourceX,sourceY,sourceZ);
Util3D::Dbl3 dirP(sourceX+unitDirX,sourceY+unitDirY,sourceZ+unitDirZ);
_base.globalToLocalPosition(rayP.x,rayP.y,rayP.z);
_base.globalToLocalPosition(dirP.x,dirP.y,dirP.z);
return(shape->_intersectRay(level,findNearest,rayP,dirP,
                            interPartOut,triIndexOut));
}

bool // collision occured
Object3D::testCollision(ArConstRef<Object3D> target,
                        unsigned int level,
                        ArConstRef<Base3D> viewPoint) const
{
if(!level)
  {
  level=(target->_intersectionLevel>_intersectionLevel
         ? target->_intersectionLevel : _intersectionLevel);
  }
if(!level)
  {
  return(false);
  }
const Shape3D * shape1=_findShape(viewPoint.c_ptr());
if(shape1)
  {
  const Shape3D * shape2=target->_findShape(viewPoint.c_ptr());
  if(shape2)
    {
    if(viewPoint)
      {
      const Util3D::Base & view=viewPoint->getBase();
      Util3D::Base base1(false),base2(false);
      shape1->_findBillboardLocation(base1,_base,view);
      shape2->_findBillboardLocation(base2,target->_base,view);
      return(Shape3D::_testCollision(shape1,shape2,base1,base2,level));
      }
    else
      {
      return(Shape3D::_testCollision(shape1,shape2,_base,target->_base,level));
      }
    }
  }
return(false);
}

bool // collision occured
Object3D::intersect(ArConstRef<Object3D> target,
                    StlVector<Util3D::TriInter> & itOut,
                    ArConstRef<Base3D> viewPoint) const
{
itOut.clear();
const Shape3D * shape1=_findShape(viewPoint.c_ptr());
if(shape1)
  {
  const Shape3D * shape2=target->_findShape(viewPoint.c_ptr());
  if(shape2)
    {
    if(viewPoint)
      {
      const Util3D::Base & view=viewPoint->getBase();
      Util3D::Base base1(false),base2(false);
      shape1->_findBillboardLocation(base1,_base,view);
      shape2->_findBillboardLocation(base2,target->_base,view);
      Shape3D::_intersect(shape1,shape2,base1,base2,itOut);
      }
    else
      {
      Shape3D::_intersect(shape1,shape2,_base,target->_base,itOut);
      }
    }
  }
return(!itOut.empty());
}

//-------- Interaction -------------------------------------------------------

void
Object3D::setMouseButtonInteraction(bool mouseButtonInteraction)
{
_mouseButtonInteraction=mouseButtonInteraction;
}

bool // object is sensible to mouse button interaction
Object3D::getMouseButtonInteraction(void) const
{
return(_mouseButtonInteraction);
}

void
Object3D::onMouseButtonInteraction(ArRef<AbstractInteractor> /* source */,
		                   ArRef<ShapePart3D> /* part */,
				   unsigned int /* triIndex */,
                                   int /* button */,
                                   bool /* pressed */)
{
/*
cerr << getAlias() << ": mouse button [" << button
     << (pressed ? "] pressed " : "] released ")
     << "from " << source->getAlias() << "/"
     << source->getWindow()->getAlias() << endl;
*/
}

void
Object3D::setKeyboardInteraction(bool keyboardInteraction)
{
_keyboardInteraction=keyboardInteraction;
}

bool // object is sensible to keyboard interaction
Object3D::getKeyboardInteraction(void) const
{
return(_keyboardInteraction);
}

void
Object3D::onKeyboardInteraction(ArRef<AbstractInteractor> /* source */,
		                ArRef<ShapePart3D> /* part */,
				unsigned int /* triIndex */,
                                const StlString & /* key */,
                                bool /* pressed */)
{
/*
StlString str;
if((key.size()==1)&&!isprint(key[0]))
  {
  char buffer[0x10];
  sprintf(buffer,"\\x%.2x",(int)key[0]);
  str=buffer;
  }
else
  {
  str=key;
  }
cerr << getAlias() << ": key [" << str
     << (pressed ? "] pressed " : "] released ")
     << "from " << source->getAlias() << "/"
     << source->getWindow()->getAlias() << endl;
*/
}

void
Object3D::setMotionInteraction(bool motionInteraction)
{
_motionInteraction=motionInteraction;
}

bool // object is sensible to motion interaction
Object3D::getMotionInteraction(void) const
{
return(_motionInteraction);
}

void
Object3D::onMotionInteraction(ArRef<AbstractInteractor> /* source */,
		              ArRef<ShapePart3D> /* part */,
			      unsigned int /* triIndex */,
                              ArRef<Base3D> newLocation,
                              bool /* begin */,
                              bool /* end */)
{
/*
if(begin)
  {
  cerr << getAlias() << ": motion begin "
       << "from " << source->getAlias() << "/"
       << source->getWindow()->getAlias() << endl;
  }
if(end)
  {
  cerr << getAlias() << ": motion end "
       << "from " << source->getAlias() << "/"
       << source->getWindow()->getAlias() << endl;
  }
*/
setLocation(newLocation);
}

//-------- Serialization operations ------------------------------------------

void
Object3D::extractDependencies(SerializationDependencies & dependenciesOut,
                              bool recursive) const
{
Base3D::extractDependencies(dependenciesOut,recursive);
for(unsigned int i=_shapes.size();i--;)
  {
  if(recursive)
    {
    _shapes[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_shapes[i]);
  }
}

void
Object3D::serialize(SerializationDependencies & dependenciesOut,
                    ArRef<AbstractOStream> output) const
{
Base3D::serialize(dependenciesOut,output);
// output->writeUInt(_intersectionLevel); // FIXME: needed ?
unsigned int flags=(_wireFrame         ? (1<<0) : 0)+
                   (_drawShapeBBox     ? (1<<1) : 0)+
                   (_drawShapePartBBox ? (1<<2) : 0);
output->writeUInt(flags);
ArRef<Material3D> material=Material3D::NEW();
readSelectedMaterial(material);
material->writeToStream(output);
unsigned int nbShapes=_shapes.size();
output->writeUInt(nbShapes); // 0 => no shape
if(nbShapes)
  {
  unsigned int nbDistances=_distances.size();
  output->writeUInt(nbDistances); // 0 => no LOD
  for(unsigned int i=0;i<nbShapes;i++)
    {
    output->writeUInt(dependenciesOut.addReference(_shapes[i]));
    if(nbDistances)
      {
      output->writeDouble(_distances[i]);
      }
    }
  }
}

bool // success
Object3D::unserialize(const SerializationDependencies & dependencies,
                      ArRef<AbstractIStream> input)
{
// input->readUInt(_intersectionLevel); // FIXME: needed ?
unsigned int i,flags,nbShapes,nbDistances;
StlString shapeAlias;
ArRef<Material3D> material=Material3D::NEW();
ArRef<Shape3D> shape;
StlVector<unsigned int> shapes;
StlVector<double> distances;
if(!Base3D::unserialize(dependencies,input)||
   !input->readUInt(flags)||
   !material->readFromStream(input)||
   !input->readUInt(nbShapes))
  {
  return(false);
  }
if(nbShapes)
  {
  if(!input->readUInt(nbDistances))
    {
    return(false);
    }
  for(i=0;i<nbShapes;i++)
    {
    shapes.push_back(0);
    if(!input->readUInt(shapes.back()))
      {
      return(false);
      }
    if(nbDistances)
      {
      distances.push_back(0.0);
      if(!input->readDouble(distances.back()))
        {
        return(false);
        }
      }
    }
  StlVector<ArRef<Shape3D> > lodShapes;
  for(i=0;i<nbShapes;i++)
    {
    lodShapes.push_back(ar_down_cast<Shape3D>
                        (dependencies.findReference(shapes[i])));
    }
  if(distances.size())
    {
    setLODShapes(distances,lodShapes);
    }
  else
    {
    setShape(lodShapes.front());
    }
  }
else
  {
  setShape(Shape3D::nullRef());
  }
_wireFrame=flags&(1<<0);
_drawShapeBBox=flags&(1<<1);
_drawShapePartBBox=flags&(1<<2);
writeSelectedMaterial(material);
_postRedisplay();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Object3D::_onMotion(void) // Protected
{
_postRedisplay();
Base3D::_onMotion();
}

void
Object3D::_rebuildShapesIfNeeded(void) // Protected
{
for(unsigned int i=_shapes.size();i--;)
  {
  _shapes[i]->_rebuildIfNeeded();
  }
if(_interShape)
  {
  _interShape->_rebuildIfNeeded();
  }
}

void
Object3D::_shapeDestroyed(Shape3D * shape) // Protected
{
for(unsigned int i=_shapes.size();i--;)
  {
  if(_shapes[i].c_ptr()==shape)
    {
    _shapes.erase(_shapes.begin()+i);
    if(_distances.size()>i)
      {
      _distances.erase(_distances.begin()+i);
      }
    }
  }
}

void
Object3D::_addScene(ArRef<Scene3D> scene) // Protected
{
_scenes.push_back(scene);
}

void
Object3D::_removeScene(ArRef<Scene3D> scene) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  if(_scenes[i]==scene)
    {
    StlVectorFastErase(_scenes,i);
    break;
    }
  }
}

void
Object3D::_postRedisplay(void) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  _scenes[i]->_postRedisplay();
  }
}

const Shape3D * // shape to be displayed from point of view (if any)
Object3D::_findShape(const Base3D * viewPoint) const // Protected
{
if(viewPoint)
  {
  unsigned int nb=_shapes.size();
  if(nb)
    {
    if(_distances.empty())
      {
      return(_shapes.front().c_ptr());
      }
    else
      {
      double dx,dy,dz;
      viewPoint->getPosition(dx,dy,dz);
      double x,y,z;
      getPosition(x,y,z);
      dx-=x;
      dy-=y;
      dz-=z;
      double squareDist=dx*dx+dy*dy+dz*dz;
      for(unsigned int i=0;i<nb;i++)
        {
        double limit=_distances[i];
        if(limit*limit>=squareDist)
          {
          return(_shapes[i].c_ptr());
          }
        }
      }
    }
  }
else
  {
  if(_interShape)
    {
    return(_interShape.c_ptr());
    }
  if(!_shapes.empty())
    {
    return(_shapes.front().c_ptr());
    }
  }
return((const Shape3D *)0);
}

void
Object3D::_onShapeChange(void) // Protected
{
if(!_shapeChangeCB.empty())
  {
  Object3D::ShapeChangeEvent evt;
  _shapeChangeCB.fire(evt);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
