// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/osd.cpp
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/osd.h"

#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/abstractInteractor.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class OSD (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(OSD, ArObject)

//-------- Construction / Destruction ----------------------------------------

OSD::OSD(ArCW& arCW) // Protected
: ArObject(arCW),
  _location(OSD::LOC_TOP_LEFT),
  _x(0),
  _y(0),
  _orientation(0.0),
  _depth(0.0),
  _alpha(1.0),
  _padding(0.0,0.0),
  _renderers(),
  _mouseButtonInteraction(true),
  _mouseMotionInteraction(true),
  _keyboardInteraction(true),
  _motionInteraction(true),
  _setupCB(thisRef()),
  _motionCB(thisRef()),
  _motionMask(0),
  _attachInfos()
{
  // Nothing to be done
}

OSD::~OSD(void) // Protected
{
while(!_attachInfos.empty())
  {
  if(_attachInfos.back()->src.c_ptr()==this)
    {
    _attachInfos.back()->dst->attachTo(thisRef(),OSD::ATTACH_NONE);
    }
  else
    {
    attachTo(_attachInfos.back()->src,OSD::ATTACH_NONE);
    }
  }
while(!_renderers.empty())
  {
  _renderers.back()->removeOSD(thisRef());
  }
}

//-------- Position ----------------------------------------------------------

void
OSD::setLocation(OSD::Location location)
{
_location=location;
_postRedisplay();
_onMotion();
}

OSD::Location
OSD::getLocation(void) const
{
return(_location);
}

void
OSD::setPosition(int x,
                 int y)
{
_x=x;
_y=y;
_postRedisplay();
_onMotion();
}

bool // success
OSD::getPosition(ArConstRef<Renderer3D> renderer,
                 int & xOut,
                 int & yOut) const
{
switch(_location)
  {
  case OSD::LOC_FREE:
    {
    xOut=_x;
    yOut=_y;
    return(true);
    }
  default:
    {
    if(renderer)
      {
      unsigned int rw,rh,w,h;
      renderer->getRenderSize(rw,rh);
      _getPositionAndSize(rw,rh,xOut,yOut,w,h);
      return(true);
      }
    else
      {
      return(false);
      }
    }
  }
}

void
OSD::setOrientation(double orientation)
{
_orientation=orientation;
_postRedisplay();
_onMotion();
}

double
OSD::getOrientation(void) const
{
return(_orientation);
}

void
OSD::setPadding(double x,
                double y)
{
_padding.x=(x>0.0 ? x : 0.0);
_padding.y=(y>0.0 ? y : 0.0);
_postRedisplay();
_onMotion();
}

void
OSD::getPadding(double & xOut,
                double & yOut) const
{
xOut=_padding.x;
yOut=_padding.y;
}

void
OSD::setDepth(double depth)
{
_depth=(depth<-1.0 ? -1.0 : (depth>1.0 ? 1.0 : depth));
_postRedisplay();
_onSetup();
}

double
OSD::getDepth(void) const
{
return(_depth);
}

//-------- Transparency ------------------------------------------------------

void
OSD::setAlpha(double alpha)
{
_alpha=(alpha<0.0 ? 0.0 : (alpha>1.0 ? 1.0 : alpha));
_postRedisplay();
_onSetup();
}

double
OSD::getAlpha(void) const
{
return(_alpha);
}

//-------- Size --------------------------------------------------------------

void
OSD::readBoundingBox(ArConstRef<Renderer3D> renderer,
                     ArRef<BoundingBox3D> bboxOut) const
{
unsigned int rw,rh,w,h;
int x,y;
renderer->getRenderSize(rw,rh);
if(renderer->getStereoMode()==Renderer3D::STEREO_SPLIT)
  {
  rw>>=1;
  }
_getPositionAndSize(rw,rh,x,y,w,h);

double c=cos(_orientation);
double s=sin(_orientation);
double w2=0.5*w;
double h2=0.5*h;
Util3D::BoundingBox & bbox=bboxOut->accessBoundingBox();
bbox.valid=false;
double x1(x+w2),y1(y+h2);

bbox.addXYZ(x1+((w2*c)-(h2*s)),y1+((h2*c)+(w2*s)),0.0);
bbox.addXYZ(x1-((w2*c)+(h2*s)),y1+((h2*c)-(w2*s)),0.0);
bbox.addXYZ(x1+(-(w2*c)+(h2*s)),y1-((h2*c)+(w2*s)),0.0);
bbox.addXYZ(x1+((w2*c)+(h2*s)),y1+(-(h2*c)+(w2*s)),0.0);
}

//-------- Interaction -------------------------------------------------------

void
OSD::setMouseButtonInteraction(bool mouseButtonInteraction)
{
_mouseButtonInteraction=mouseButtonInteraction;
}

bool // osd is sensible to mouse button interaction
OSD::getMouseButtonInteraction(void) const
{
return(_mouseButtonInteraction);
}

void
OSD::onMouseButtonInteraction(ArRef<AbstractInteractor> /* source */,
                              int /* button */,
                              bool /* pressed */)
{
/*
cerr << getAlias() << ": mouse button [" << button
     << (pressed ? "] pressed " : "] released ")
     << "from " << source->getAlias() << "/"
     << source->getWindow()->getAlias() << endl;
*/
}

void
OSD::setMouseMotionInteraction(bool mouseMotionInteraction)
{
_mouseMotionInteraction=mouseMotionInteraction;
}

bool // osd is sensible to motion button interaction
OSD::getMouseMotionInteraction(void) const
{
return(_mouseMotionInteraction);
}

void
OSD::onMouseMotionInteraction(ArRef<AbstractInteractor> /*source */,
                              int /* xMouse */,
                              int /* yMouse */)
{
/*
cerr << getAlias() << ": mouse motion [" << xMouse << " , " << yMouse << "] "
     << "from " << source->getAlias() << "/"
     << source->getWindow()->getAlias() << endl;
*/
}

void
OSD::setKeyboardInteraction(bool keyboardInteraction)
{
_keyboardInteraction=keyboardInteraction;
}

bool // osd is sensible to keyboard interaction
OSD::getKeyboardInteraction(void) const
{
return(_keyboardInteraction);
}

void
OSD::onKeyboardInteraction(ArRef<AbstractInteractor> /* source */,
                           const StlString & /* key */,
                           bool /* pressed */)
{
/*
StlString str;
if((key.size()==1)&&!isprint(key[0]))
  {
  char buffer[0x10];
  sprintf(buffer,"\\x%.2x",(int)key[0]);
  str=buffer;
  }
else
  {
  str=key;
  }
cerr << getAlias() << ": key [" << str
     << (pressed ? "] pressed " : "] released ")
     << "from " << source->getAlias() << "/"
     << source->getWindow()->getAlias() << endl;
*/
}

void
OSD::setMotionInteraction(bool motionInteraction)
{
_motionInteraction=motionInteraction;
}

bool // osd is sensible to motion interaction
OSD::getMotionInteraction(void) const
{
return(_motionInteraction);
}

void
OSD::onMotionInteraction(ArRef<AbstractInteractor> /* source */,
                         int newX,
                         int newY,
                         double newOrientation,
                         bool /* begin */,
                         bool /* end */)
{
/*
if(begin)
  {
  cerr << getAlias() << ": motion begin "
       << "from " << source->getAlias() << "/"
       << source->getWindow()->getAlias() << endl;
  }
if(end)
  {
  cerr << getAlias() << ": motion end "
       << "from " << source->getAlias() << "/"
       << source->getWindow()->getAlias() << endl;
  }
*/
setPosition(newX,newY);
setOrientation(newOrientation);
}

//-------- Attachment --------------------------------------------------------

bool // success if location is OSD::LOC_FREE
OSD::attachTo(ArRef<OSD> osd,
              int attachment)
{
if(osd.c_ptr()!=this)
  {
  attachment&=OSD::ATTACH_ALL;
  OSD::AttachInfo * info=(OSD::AttachInfo*)0;
  unsigned int i=_attachInfos.size();
  while(i--)
    {
    if(_attachInfos[i]->src==osd)
      {
      info=_attachInfos[i];
      break;
      }
    }
  if(info)
    {
    if(attachment)
      {
      info->attachment=attachment;
      }
    else
      {
      StlVectorFastErase(_attachInfos,i);
      for(i=osd->_attachInfos.size();i--;)
        {
        if(osd->_attachInfos[i]==info)
          {
          StlVectorFastErase(osd->_attachInfos,i);
          break;
          }
        }
      delete info;
      }
    }
  else if(attachment)
    {
    OSD::AttachInfo * info=new OSD::AttachInfo(osd,thisPtr(),attachment);
    _attachInfos.push_back(info);
    osd->_attachInfos.push_back(info);
    }
  }
return(true);
}

int // attachement to osd
OSD::getAttachmentTo(ArConstRef<OSD> osd) const
{
for(unsigned int i=_attachInfos.size();i--;)
  {
  if(_attachInfos[i]->src==osd)
    {
    return(_attachInfos[i]->attachment);
    }
  }
return(OSD::ATTACH_NONE);
}

void
OSD::getAttachments(StlVector<ArRef<OSD> >& srcOut,
                    StlVector<ArRef<OSD> >& dstOut) const
{
srcOut.clear();
dstOut.clear();
for(unsigned int i=_attachInfos.size();i--;)
  {
  if(_attachInfos[i]->src.c_ptr()==this)
    {
    dstOut.push_back(_attachInfos[i]->dst);
    }
  else
    {
    srcOut.push_back(_attachInfos[i]->src);
    }
  }
}

//-------- Event management --------------------------------------------------

void
OSD::maskMotion(void)
{
++_motionMask;
}

void
OSD::unmaskMotion(void)
{
--_motionMask;
_onMotion();
}

//-------- Serialization operations ------------------------------------------

void
OSD::serialize(SerializationDependencies & dependenciesOut,
               ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
output->writeInt(_location);
output->writeInt(_x);
output->writeInt(_y);
output->writeDouble(_depth);
output->writeDouble(_alpha);
output->writeDouble(_padding.x);
output->writeDouble(_padding.y);
output->writeDouble(_orientation);
output->writeChar(_mouseButtonInteraction); // FIXME : Use on flag rather than 4 chars
output->writeChar(_mouseMotionInteraction);
output->writeChar(_keyboardInteraction);
output->writeChar(_motionInteraction);
}

bool // success
OSD::unserialize(const SerializationDependencies & dependencies,
                 ArRef<AbstractIStream> input)
{
int location;
int x,y;
double depth,alpha,orientation;
Util3D::Dbl2 padding;
char mouseButtonInteraction; // FIXME : Use on flag rather than 4 chars
char mouseMotionInteraction;
char keyboardInteraction;
char motionInteraction;
if(!ArObject::unserialize(dependencies,input)||
   !input->readInt(location)||
   !input->readInt(x)||
   !input->readInt(y)||
   !input->readDouble(depth)||
   !input->readDouble(alpha)||
   !input->readDouble(padding.x)||
   !input->readDouble(padding.y)||
   !input->readDouble(orientation) ||
   !input->readChar(mouseButtonInteraction)||
   !input->readChar(mouseMotionInteraction)||
   !input->readChar(keyboardInteraction)||
   !input->readChar(motionInteraction))
  {
  return(false);
  }
_location=(OSD::Location)location;
_x=x;
_y=y;
_depth=depth;
_alpha=alpha;
_padding=padding;
_orientation=orientation;
_mouseButtonInteraction=(bool)mouseButtonInteraction;
_mouseMotionInteraction=(bool)mouseMotionInteraction;
_keyboardInteraction=(bool)keyboardInteraction;
_motionInteraction=(bool)motionInteraction;
_postRedisplay();
_onSetup(); // notify change
_onMotion();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
OSD::_addRenderer(ArPtr<Renderer3D> renderer) // Protected
{
for(unsigned int i=_renderers.size();i--;)
  {
  if(_renderers[i]==renderer)
    {
    return;
    }
  }
_renderers.push_back(renderer);
}

void
OSD::_removeRenderer(ArPtr<Renderer3D> renderer) // Protected
{
for(unsigned int i=_renderers.size();i--;)
  {
  if(_renderers[i]==renderer)
    {
    StlVectorFastErase(_renderers,i);
    return;
    }
  }
}

// void
// OSD::_getSize(unsigned int rendererWidth,
//               unsigned int rendererHeight,
//               unsigned int& widthOut,
//               unsigned int& heightOut) const=0 // Protected

void
OSD::_getPositionAndSize(unsigned int rendererWidth,
                         unsigned int rendererHeight,
                         int & xOut,
                         int & yOut,
                         unsigned int & widthOut,
                         unsigned int & heightOut) const // Protected
{
_getSize(rendererWidth,rendererHeight,widthOut,heightOut);
int padx=(int)(_padding.x*rendererWidth);
int pady=(int)(_padding.y*rendererHeight);
switch(_location)
  {
  case OSD::LOC_TOP_LEFT:
    {
    xOut=padx;
    yOut=pady;
    break;
    }
  case OSD::LOC_TOP_CENTER:
    {
    xOut=(int)(((int)rendererWidth-(int)widthOut)/2.0);
    yOut=pady;
    break;
    }
  case OSD::LOC_TOP_RIGHT:
    {
    xOut=(int)rendererWidth-(int)widthOut-padx;
    yOut=pady;
    break;
    }
  case OSD::LOC_CENTER_LEFT:
    {
    xOut=padx;
    yOut=(int)(((int)rendererHeight-(int)heightOut)/2.0);
    break;
    }
  case OSD::LOC_CENTER:
    {
    xOut=(int)(((int)rendererWidth-(int)widthOut)/ 2.0);
    yOut=(int)(((int)rendererHeight-(int)heightOut)/2.0);
    break;
    }
  case OSD::LOC_CENTER_RIGHT:
    {
    xOut=(int)rendererWidth-(int)widthOut-padx;
    yOut=(int)(((int)rendererHeight-(int)heightOut)/2.0);
    break;
    }
  case OSD::LOC_BOTTOM_LEFT:
    {
    xOut=padx;
    yOut=(int)rendererHeight-(int)heightOut-pady;
    break;
    }
  case OSD::LOC_BOTTOM_CENTER:
    {
    xOut=(int)(((int)rendererWidth-(int)widthOut)/2.0);
    yOut=(int)rendererHeight-(int)heightOut-pady;
    break;
    }
  case OSD::LOC_BOTTOM_RIGHT:
    {
    xOut=(int)rendererWidth-(int)widthOut-padx;
    yOut=(int)rendererHeight-(int)heightOut-pady;
    break;
    }
  case OSD::LOC_FREE:
    {
    xOut=_x;
    yOut=_y;
    break;
    }
  }
}

void
OSD::_postRedisplay() // Protected
{
for(unsigned int i=_renderers.size();i--;)
  {
  _renderers[i]->_postRedisplay();
  }
}

// void
// OSD::_render(unsigned int width,
//              unsigned int height) const=0 // Protected

void
OSD::_onSetup(void) // Protected
{
if(!_setupCB.empty())
  {
  OSD::SetupEvent event;
  _setupCB.fire(event);
  }
}

void
OSD::_onMotion(void) // Protected
{
if(!_motionMask)
  {
  ArRef<OSD> flag(this);
  _moving=true;
  if(!_motionCB.empty())
    {
    OSD::MotionEvent evt;
    _motionCB.fire(evt);
    }
  if(flag&&!_attachInfos.empty())
    {
    for(unsigned int i=_attachInfos.size();i--;)
      {
      if(i<_attachInfos.size()) // not removed
        {
        OSD::AttachInfo & info=*_attachInfos[i];
        if((info.src.c_ptr()==this)&&!info.dst->_moving)
          {
          info.propagating=true;
          info.dst->maskMotion();
          info.dst->setPosition(_x+info.relTx,_y+info.relTy);
          info.dst->setOrientation(_orientation+info.relR);
          if(info.attachment!=OSD::ATTACH_ALL)
            {
            if((~info.attachment)&OSD::ATTACH_TRANSLATION)
              {
              int x, y;
              info.dst->getPosition(Renderer3D::nullRef(),x,y);
              if(!(info.attachment&OSD::ATTACH_TX))
                {
                x=info.absTx;
                }
              if(!(info.attachment&OSD::ATTACH_TY))
                {
                y=info.absTy;
                }
              info.dst->setPosition(x,y);
              }
            if((~info.attachment)&OSD::ATTACH_ROTATION)
              {
              info.dst->setOrientation(info.absR);
              }
            }
          info.dst->unmaskMotion();
          if(flag)
            {
            info.propagating=false;
            }
          else
            {
            break;
            }
          }
        else if(!info.propagating)
          {
          int srcX,srcY;
          if(info.src->getPosition(Renderer3D::nullRef(),srcX,srcY)&&
             info.dst->getPosition(Renderer3D::nullRef(),
                                   info.absTx,info.absTy))
            {
            info.relTx=info.absTx-srcX;
            info.relTy=info.absTy-srcY;
            }
          else
            {
            info.relTx=info.absTx=0;
            info.relTy=info.absTy=0;
            }
          info.absR=info.dst->_orientation;
          info.relR=info.absR-info.src->_orientation;
          }
        }
      }
    }
  if(flag)
    {
    _moving=false;
    }
  }
}

OSD::AttachInfo::AttachInfo(ArPtr<OSD> s,
                            ArRef<OSD> d,
                            int a)
: src(s),
  dst(d),
  attachment(a),
  propagating(false)
{
int srcX,srcY;
if(src->getPosition(Renderer3D::nullRef(),srcX,srcY)&&
   dst->getPosition(Renderer3D::nullRef(),absTx,absTy))
  {
  relTx=absTx-srcX;
  relTy=absTy-srcY;
  }
else
  {
  relTx=absTx=0;
  relTy=absTy=0;
  }
absR=dst->_orientation;
relR=absR-src->_orientation;
}

} // namespace AReVi

