// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/osdSprite.cpp
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/osdSprite.h"

#include "AReVi/Utils/abstractStream.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/texture.h"
#include "AReVi/Lib3D/dataTexture.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class OSDSprite
//----------------------------------------------------------------------------

AR_CLASS_DEF(OSDSprite, OSD)

//-------- Construction / Destruction ----------------------------------------

OSDSprite::OSDSprite(ArCW& arCW, ArConstRef<Texture> texture) // Protected
: OSD(arCW),
  _resize(true),
  _ratio(1.0),
  _scale(1.0,1.0),
  _color(1.0,1.0,1.0),
  _w(0),
  _h(0),
  _texture(texture)
{
  // Nothing to be done
}

OSDSprite::OSDSprite(ArCW& arCW) // Protected, for serialization
: OSD(arCW),
  _resize(true),
  _ratio(1.0),
  _scale(1.0,1.0),
  _color(1.0,1.0,1.0),
  _w(0),
  _h(0),
  _texture()
{
ArRef<Texture> texture;
DataTexture::CLASS()->newInstance(texture);
_texture=texture;
}


OSDSprite::~OSDSprite() // Protected
{
  // Nothing to be done
}

//-------- Aspect ------------------------------------------------------------

void
OSDSprite::resizeWithRenderer(bool resize)
{
_resize=resize;
_onSetup();
_postRedisplay();
}

bool // resized with renderer
OSDSprite::isResizedWithRenderer(void) const
{
return(_resize);
}

void
OSDSprite::setRatio(double ratio)
{
_ratio=(ratio>0.0 ? ratio : 0.0);
_onSetup();
_postRedisplay();
}

double // > 0.0 => use ratio, else no ratio
OSDSprite::getRatio(void) const
{
return(_ratio);
}

void
OSDSprite::setScale(double x,
                    double y)
{
if(x>0.0&&y>0.0)
  {
  _scale.x=x;
  _scale.y=y;
  }
_onSetup();
_postRedisplay();
}

void
OSDSprite::getScale(double & xOut,
                    double & yOut) const
{
xOut=_scale.x;
yOut=_scale.y;
}

void
OSDSprite::setSize(unsigned int width,
                   unsigned int height)
{
_w=width;
_h=height;
_onSetup();
_postRedisplay();
}

bool
OSDSprite::getSize(ArRef<Renderer3D> renderer,
		   unsigned int & widthOut,
		   unsigned int & heightOut)
{
if(_resize)
  {
  if(renderer)
    {
    unsigned int rw,rh;
    renderer->getRenderSize(rh,rw);
    _getSize(rw,rh,widthOut,heightOut);
    return(true);
    }
  else
    {
    return(false);
    }
  }
else
  {
  widthOut=_w;
  heightOut=_h;
  return(true);
  }
}

ArConstRef<Texture>
OSDSprite::getTexture(void) const
{
return(_texture);
}

void
OSDSprite::setTexture(ArConstRef<Texture> texture)
{
_texture=texture;
_onSetup();
_postRedisplay();
}

void
OSDSprite::setColorFilter(double r,
                          double g,
                          double b)
{
_color.x=r;
_color.y=g;
_color.z=b;
_onSetup();
_postRedisplay();
}

void
OSDSprite::getColorFilter(double & rOut,
                          double & gOut,
                          double & bOut) const
{
rOut=_color.x;
gOut=_color.y;
bOut=_color.z;
}

//-------- Serialization operations ------------------------------------------

void
OSDSprite::extractDependencies(SerializationDependencies & dependenciesOut,
                               bool recursive) const
{
OSD::extractDependencies(dependenciesOut,recursive);
if(_texture.valid())
  {
  dependenciesOut.addConstReference(_texture);
  }
}

void
OSDSprite::serialize(SerializationDependencies & dependenciesOut,
                     ArRef<AbstractOStream> output) const
{
OSD::serialize(dependenciesOut,output);

if(_texture.valid())
  {
  unsigned index=(dependenciesOut.addConstReference(_texture));
  output->writeChar((char)true); // texture is valid
  output->writeUInt(index);
  }
else
  output->writeChar((char)false); // texture is not valid

output->writeChar((char)_resize);
output->writeDouble(_ratio);
output->writeDouble(_scale.x);
output->writeDouble(_scale.y);
output->writeDouble(_color.x);
output->writeDouble(_color.y);
output->writeDouble(_color.z);
output->writeUInt(_w);
output->writeUInt(_h);
}

bool // success
OSDSprite::unserialize(const SerializationDependencies & dependencies,
                       ArRef<AbstractIStream> input)
{
char resize,textureValid;
Util3D::Dbl2 scale;
Util3D::Dbl3 color;

double ratio;
unsigned int index,w,h;
if(!OSD::unserialize(dependencies,input)||
  !input->readChar(textureValid))
  {
  return(false);
  }

if (textureValid)
  {
  if(!input->readUInt(index))
    return(false);
  }

if(!input->readChar(resize)||
   !input->readDouble(ratio)||
   !input->readDouble(scale.x)||
   !input->readDouble(scale.y)||
   !input->readDouble(color.x)||
   !input->readDouble(color.y)||
   !input->readDouble(color.z)||
   !input->readUInt(w) ||
   !input->readUInt(h))
   {
   return(false);
   }

if (textureValid)
  {
  ArRef<Texture> texture=ar_down_cast<Texture>
                        (dependencies.findReference(index));
  if(texture)
    {
    _texture=texture;
    }
  }
else
  _texture.clear();

_resize=(bool)resize;
_ratio=ratio;
_scale=scale;
_w=w;
_h=h;
setColorFilter(color.x,color.y,color.z);
_onSetup(); // notify change.
_postRedisplay();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
OSDSprite::_getSize(unsigned int rendererWidth,
		    unsigned int rendererHeight,
		    unsigned int & widthOut,
		    unsigned int & heightOut) const // Protected
{
if(_resize)
  {
  // size depends on renderer resolution
  widthOut=(unsigned int)(rendererWidth*_scale.x);
  if(!_ratio)
    {
    // don't care of any ratio
    heightOut=(unsigned int)(rendererHeight*_scale.y);
    }
  else
    {
    heightOut=(unsigned int)(rendererWidth*_scale.x/_ratio);
    }
  }
else
  {
  // size is fixed
  widthOut=_w;
  if(!_ratio)
    {
    heightOut=_h;
    }
  else
    {
    heightOut=(unsigned int)(_w/_ratio);
    }
  }
}

void
OSDSprite::_render(unsigned int rendererWidth,
		   unsigned int rendererHeight) const // Protected
{
unsigned int w,h;
int x,y;
_getPositionAndSize(rendererWidth,rendererHeight,x,y,w,h);
y=(int)rendererHeight-y;
glPushAttrib(GL_ENABLE_BIT|GL_DEPTH_BUFFER_BIT);
glDisable(GL_LIGHTING);

if (_texture.valid())
  {
  glEnable(GL_TEXTURE_2D);

  if (_texture->getBlending() || _alpha<=0.99)
    {
    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    }
  else if (_texture->getAlphaTest())
    glEnable(GL_ALPHA_TEST);

  GL::selectTexture(_texture->getTextureID(),false,false,true);
  }
else if (_alpha<=0.99)
  {
  glDepthMask(GL_FALSE);
  glEnable(GL_BLEND);
  }

glPushMatrix();
glLoadIdentity();
// glColor4f(1.0,1.0,1.0,_alpha);
glColor4f(_color.x,_color.y,_color.z,_alpha);
double c=cos(_orientation);
double s=sin(_orientation);
double w2=0.5*w;
double h2=0.5*h;
glBegin(GL_QUADS);
  glTexCoord2f(0.0,0.0);
  glVertex3f(x+w2*(1.0-c)+h2*s,y-h2*(1.0+c)-w2*s,_depth);
  glTexCoord2f(1.0,0.0);
  glVertex3f(x+w2*(1.0+c)+h2*s,y-h2*(1.0+c)+w2*s,_depth);
  glTexCoord2f(1.0,1.0);
  glVertex3f(x+w2*(1.0+c)-h2*s,y-h2*(1.0-c)+w2*s,_depth);
  glTexCoord2f(0.0,1.0);
  glVertex3f(x+w2*(1.0-c)-h2*s,y-h2*(1.0-c)-w2*s,_depth);
glEnd();
glPopMatrix();

glPopAttrib();
}

} // namespace AReVi

//----------------------------------------------------------------------------
