// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/osdText.cpp
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/osdText.h"

#include "AReVi/Utils/abstractStream.h"

#include "AReVi/Lib3D/gl_priv.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class OSDSprite
//----------------------------------------------------------------------------

StlVector<StlString> OSDText::_fontDir;
StlMap<StlString,LLW::BMFontID> OSDText::_font;

AR_CLASS_DEF(OSDText,OSD)

//-------- Construction / Destruction ----------------------------------------

OSDText::OSDText(ArCW& arCW)
: OSD(arCW),
  _color(1.0,1.0,1.0,1.0),
  _justification(OSDText::JUSTIF_LEFT),
  _textChangedCB(thisPtr())
{
if(_font.empty())
  {
  _fontDir.push_back("BMFonts/");
  _loadFont("9x15bold");
  }
StlMap<StlString,LLW::BMFontID>::iterator itF=_font.begin();
if(itF==_font.end())
  {
  // No font available
  _fontName="";
  _curFont=NULL;
  }
else
  {
  // By default, use the first font available
  _fontName=itF->first;
  _curFont=itF->second;
  }
}

OSDText::~OSDText(void)
{
  // Nothing to be done
}

//-------- Aspect ------------------------------------------------------------

bool // success
OSDText::setFont(const StlString & fontName)
{
LLW::BMFontID f=_loadFont(fontName);
if(f)
  {
  _curFont=f;
  _fontName=fontName;
  _onSetup();
  _postRedisplay();
  return(true);
  }
else
  {
  return(false);
  }
}

const StlString &
OSDText::getFont(void) const
{
return(_fontName);
}

void
OSDText::setJustification(OSDText::Justification justification)
{
_justification=justification;
_onSetup();
_postRedisplay();
}

OSDText::Justification
OSDText::getJustification(void) const
{
return(_justification);
}

void
OSDText::setText(const StlString & text)
{
_text.clear();
size_t lastPos=0;
size_t curPos=0;
size_t nextCR,nextEOL;
while(curPos!=text.size())
  {
  nextCR=text.find_first_of('\n',lastPos);
  nextEOL=text.find_first_of('\0',lastPos);
  curPos=(nextCR<nextEOL ? nextCR : nextEOL);
  if(curPos==StlString::npos)
    {
    curPos=text.size();
    }
  _text.push_back(StlString(text,lastPos,curPos-lastPos));
  lastPos=curPos+1;
  }
_onTextChanged();
_postRedisplay();
}

StlString
OSDText::getText(void) const
{
StlString text;
unsigned int nbLines=_text.size();
for(unsigned int i=0;i<nbLines;i++)
  {
  if(i)
    {
    text+='\n';
    }
  text+=_text[i];
  }
return(text);
}

void
OSDText::setColor(double r,
                  double g,
                  double b,
                  double a)
{
OSD::setAlpha(a);
_color.x=r;
_color.y=g;
_color.z=b;
_onSetup();
_postRedisplay();
}

void
OSDText::getColor(double & rOut,
                  double & gOut,
                  double & bOut,
                  double & aOut) const
{
rOut=_color.x;
gOut=_color.y;
bOut=_color.z;
aOut=_alpha;
}

void
OSDText::setOrientation(double /* orientation */)
{
  // Nothing to be done
}

//-------- Serialization operations ------------------------------------------

void
OSDText::serialize(SerializationDependencies & dependenciesOut,
                   ArRef<AbstractOStream> output) const
{
OSD::serialize(dependenciesOut,output);
output->writeDouble(_color.x);
output->writeDouble(_color.y);
output->writeDouble(_color.z);
output->writeDouble(_color.a);
output->writeUInt((unsigned int)_justification);
output->writeString(_fontName);
output->writeUInt(_text.size());
for(unsigned int i=0;i<_text.size();i++)
  {
  output->writeString(_text[i]);
  }
}

bool // success
OSDText::unserialize(const SerializationDependencies & dependencies,
                     ArRef<AbstractIStream> input)
{
StlString fontName;
StlVector<StlString> text;
Util3D::Dbl4 color;
unsigned int nbTexts,justification;
if(!OSD::unserialize(dependencies,input)||
   !input->readDouble(color.x)||
   !input->readDouble(color.y)||
   !input->readDouble(color.z)||
   !input->readDouble(color.a)||
   !input->readUInt(justification)||
   !input->readString(fontName)||
   !input->readUInt(nbTexts))
  {
  return(false);
  }
for(unsigned int i=0;i<nbTexts;i++)
  {
  text.push_back("");
  if(!input->readString(text[i]))
    {
    return(false);
    }
  }
_text=text;
setColor(color.x,color.y,color.z,color.a);
setFont(fontName);
setJustification((OSDText::Justification)justification);
// NO CALLBACK HERE! (setColor/setFont/set... does it.)
return(true);
}

//-------- Forbidden operations ----------------------------------------------

LLW::BMFontID
OSDText::_loadFont(const StlString & fontName) // Protected
{
if(fontName=="")
  {
  return((LLW::BMFontID)0);
  }
StlMap<StlString,LLW::BMFontID>::iterator itFont=_font.find(fontName);
if(itFont!=_font.end())
  {
  return(itFont->second);
  }
StlString tmpFontName;
LLW::BMFontID fontPtr;
for(size_t i=0;i<_fontDir.size();i++)
  {
  tmpFontName=_fontDir[i]+fontName+".arbmf";
  fontPtr=LLW::loadBMFont(tmpFontName);
  if(fontPtr)
    {
    _font.insert(make_pair(fontName,fontPtr));
    return(fontPtr);
    }
  }
return((LLW::BMFontID)0);
}

void
OSDText::_getSize(unsigned int /* rendererWidth */,
		  unsigned int /* rendererHeight */,
		  unsigned int & widthOut,
		  unsigned int & heightOut) const // Protected
{
ar_const_cast<OSDText>(thisRef())->_lineWidth.resize(_text.size(),0);
unsigned int ascent,descent;
widthOut=0;
heightOut=0;
LLW::getBMFontHeight(_curFont,ascent,descent);
heightOut=(ascent+descent+4)*_text.size();
for(size_t i=0;i<_text.size();i++)
  {
  ar_const_cast<OSDText>
  (thisRef())->_lineWidth[i]=LLW::getBMTextWidth(_curFont,_text[i]);
  if(_lineWidth[i]>widthOut)
    {
    widthOut=_lineWidth[i];
    }
  }
}

void
OSDText::_render(unsigned int rendererWidth,
		 unsigned int rendererHeight) const // Protected
{
if(!_curFont)
  {
  return;
  }
unsigned int w,h;
int x,y;
_getPositionAndSize(rendererWidth,rendererHeight,x,y,w,h);
y=(int)rendererHeight-y-(int)h;
double offset=0.0;
double curHeight=0.0;
unsigned int ascent,descent;
LLW::getBMFontHeight(_curFont,ascent,descent);
glPushAttrib(GL_ENABLE_BIT);
glEnable(GL_BLEND);
glDisable(GL_LIGHTING);
glPushMatrix();
glLoadIdentity();
glColor4f(_color.x,_color.y,_color.z,_alpha);
for(size_t line=_text.size();line--;)
  {
  switch(_justification)
    {
    case OSDText::JUSTIF_LEFT:
      {
      offset=0.0;
      break;
      }
    case OSDText::JUSTIF_CENTER:
      {
      offset=(w-_lineWidth[line])/2.0;
      break;
      }
    case OSDText::JUSTIF_RIGHT:
      {
      offset=w-_lineWidth[line];
      break;
      }
    }
  int advance=0;
  for(size_t ch=0;ch<_text[line].size();ch++)
    {
    glRasterPos3f(x+offset+advance,y+curHeight,_depth);
    LLW::drawBMSymbol(_curFont,_text[line][ch]);
    advance+=LLW::getBMSymbolWidth(_curFont,_text[line][ch]);
    }
  curHeight+=ascent+descent+4;
  }
glPopMatrix();
glPopAttrib();
}

void
OSDText::_onTextChanged(void) // Protected
{
if(!_textChangedCB.empty())
  {
  OSDText::TextChangedEvent event;
  _textChangedCB.fire(event);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
