// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/particleSystem.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/particleSystem.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/texture.h"

#include "AReVi/Utils/abstractStream.h"

#include "AReVi/arSystem.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class ParticleCollider (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(ParticleCollider,ArObject)

//-------- Construction / Destruction ----------------------------------------

ParticleCollider::ParticleCollider(ArCW & arCW) // Protected
: ArObject(arCW),
  _location(Base3D::NEW()),
  _particleSystems(),
  _destroyFlag(false),
  _nFactor(0.5),
  _tFactor(1.0),
  _collisionCount(0),
  _nSpeed(0.0,0.0,0.0),
  _tSpeed(0.0,0.0,0.0),
  _nSpeedSum(0.0,0.0,0.0),
  _tSpeedSum(0.0,0.0,0.0),
  _particleCollisionCB(thisPtr()),
  _setupCB(thisRef())
{
  // Nothing to be done
}

ParticleCollider::ParticleCollider(ArCW & arCW,
                                   ArRef<Base3D> location) // Protected
: ArObject(arCW),
  _location(location),
  _particleSystems(),
  _destroyFlag(false),
  _nFactor(0.5),
  _tFactor(1.0),
  _collisionCount(0),
  _nSpeed(0.0,0.0,0.0),
  _tSpeed(0.0,0.0,0.0),
  _nSpeedSum(0.0,0.0,0.0),
  _tSpeedSum(0.0,0.0,0.0),
  _particleCollisionCB(thisRef()),
  _setupCB(thisRef())
{
  // Nothing to be done
}

ParticleCollider::~ParticleCollider(void) // Protected
{
while(!_particleSystems.empty())
  {
  _particleSystems.back()->removeCollider(thisRef());
  }
}

//-------- Location ----------------------------------------------------------

ArConstRef<Base3D> // location
ParticleCollider::getLocation(void) const
{
return(_location);
}

ArRef<Base3D> // location
ParticleCollider::accessLocation(void)
{
return(_location);
}

//-------- Collision management ----------------------------------------------

void
ParticleCollider::setDestroyFlag(bool destroyFlag)
{
_destroyFlag=destroyFlag;
_onSetup();
}

bool // destroy flag
ParticleCollider::getDestroyFlag(void) const
{
return(_destroyFlag);
}

//-------- Surface parameters ------------------------------------------------

bool // success
ParticleCollider::setNormalFactor(double normalFactor)
{
if(normalFactor>=0.0)
  {
  _nFactor=normalFactor;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // normal factor
ParticleCollider::getNormalFactor(void) const
{
return(_nFactor);
}

bool // success
ParticleCollider::setTangentialFactor(double tangentialFactor)
{
if(tangentialFactor>=0.0)
  {
  _tFactor=tangentialFactor;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // tangential factor
ParticleCollider::getTangentialFactor(void) const
{
return(_tFactor);
}

//-------- Serialization operations ------------------------------------------

void
ParticleCollider::extractDependencies(SerializationDependencies &
                                      dependenciesOut,
                                      bool recursive) const
{
ArObject::extractDependencies(dependenciesOut,recursive);
if(recursive)
  {
  _location->extractDependencies(dependenciesOut,recursive);
  }
dependenciesOut.addReference(_location);
}

void
ParticleCollider::serialize(SerializationDependencies & dependenciesOut,
                            ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
output->writeUInt(dependenciesOut.addReference(_location));
output->writeDouble(_nFactor);
output->writeDouble(_tFactor);
output->writeUInt(_destroyFlag ? 1 : 0);
}

bool // success
ParticleCollider::unserialize(const SerializationDependencies & dependencies,
                              ArRef<AbstractIStream> input)
{
unsigned int index;
double nFactor,tFactor;
unsigned int destroyFlag;
if(!ArObject::unserialize(dependencies,input)||
   !input->readUInt(index)||
   !input->readDouble(nFactor)||
   !input->readDouble(tFactor)||
   !input->readUInt(destroyFlag))
  {
  return(false);
  }
_location=ar_down_cast<Base3D>(dependencies.findReference(index));
_nFactor=nFactor;
_tFactor=tFactor;
_destroyFlag=destroyFlag;
_onSetup();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
ParticleCollider::_addParticleSystem(ArPtr<ParticleSystem> particleSystem)
{
_particleSystems.push_back(particleSystem);
}

void
ParticleCollider::_removeParticleSystem(ArPtr<ParticleSystem> particleSystem)
{
for(unsigned int i=_particleSystems.size();i--;)
  {
  if(_particleSystems[i]==particleSystem)
    {
    StlVectorFastErase(_particleSystems,i);
    break;
    }
  }
}

void
ParticleCollider::_preUpdate(void) // Protected
{
_collisionCount=0;
_nSpeedSum.x=_nSpeedSum.y=_nSpeedSum.z=0.0;
_tSpeedSum.x=_tSpeedSum.y=_tSpeedSum.z=0.0;
}

// void
// ParticleCollider::_collide(ParticleCollider::Particle & partInOut)=0
//                                                                // Protected

void
ParticleCollider::_onSetup(void) // Protected
{
if(!_setupCB.empty())
  {
  ParticleCollider::SetupEvent evt;
  _setupCB.fire(evt);
  }
}


//----------------------------------------------------------------------------
// Class PlaneParticleCollider
//----------------------------------------------------------------------------

AR_CLASS_DEF(PlaneParticleCollider,ParticleCollider)

//-------- Construction / Destruction ----------------------------------------

PlaneParticleCollider::PlaneParticleCollider(ArCW & arCW) // Protected
: ParticleCollider(arCW)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

PlaneParticleCollider::PlaneParticleCollider(ArCW & arCW,
                                             ArRef<Base3D> location)
: ParticleCollider(arCW,location)                                 // Protected
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

PlaneParticleCollider::~PlaneParticleCollider(void) // Protected
{
  // Nothing to be done
}

//-------- Forbidden operations ----------------------------------------------

void
PlaneParticleCollider::_preUpdate(void) // Protected
{
ParticleCollider::_preUpdate();
Util3D::Dbl3 center;
_location->getBase().getPosition(center.x,center.y,center.z);
_location->getBase().getLocalZ(_a,_b,_c); // normal=Z , a*x+b*y+c*z+d=0;
_d=-(_a*center.x+
     _b*center.y+
     _c*center.z);
}

void
PlaneParticleCollider::_collide(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
double d1=_a*partInOut.oldPos.x+
          _b*partInOut.oldPos.y+
          _c*partInOut.oldPos.z+
          _d;
double d2=_a*partInOut.pos.x+
          _b*partInOut.pos.y+
          _c*partInOut.pos.z+
          _d;
if(((d1*d2)<=0.0)&&d2) // crosses plane
  {
  Util3D::Dbl3 dPos(partInOut.pos.x-partInOut.oldPos.x,
                    partInOut.pos.y-partInOut.oldPos.y,
                    partInOut.pos.z-partInOut.oldPos.z);
  double k=-d1/(_a*dPos.x+
                _b*dPos.y+
                _c*dPos.z);
  if(k<partInOut.kmin)
    {
    partInOut.lastCollider=this;
    partInOut.kmin=k;
    double n=_a*partInOut.speed.x+
             _b*partInOut.speed.y+
             _c*partInOut.speed.z;
    _nSpeed.x=_a*n;
    _nSpeed.y=_b*n;
    _nSpeed.z=_c*n;
    _tSpeed.x=partInOut.speed.x-_nSpeed.x;
    _tSpeed.y=partInOut.speed.y-_nSpeed.y;
    _tSpeed.z=partInOut.speed.z-_nSpeed.z;
    }
  }
}


//----------------------------------------------------------------------------
// Class RectangleParticleCollider
//----------------------------------------------------------------------------

AR_CLASS_DEF(RectangleParticleCollider,ParticleCollider)

//-------- Construction / Destruction ----------------------------------------

RectangleParticleCollider::RectangleParticleCollider(ArCW & arCW) // Protected
: ParticleCollider(arCW),
  _xSize(2.0),
  _halfXSize(1.0),
  _ySize(2.0),
  _halfYSize(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

RectangleParticleCollider::RectangleParticleCollider(ArCW & arCW,
                                                  ArRef<Base3D> location)
: ParticleCollider(arCW,location),                                // Protected
  _xSize(2.0),
  _halfXSize(1.0),
  _ySize(2.0),
  _halfYSize(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

RectangleParticleCollider::~RectangleParticleCollider(void) // Protected
{
  // Nothing to be done
}

//-------- Size management ---------------------------------------------------

bool // success
RectangleParticleCollider::setSize(double sizeX,
                                   double sizeY)
{
if((sizeX>0.0)&&(sizeY>0.0))
  {
  _xSize=sizeX;
  _halfXSize=0.5*_xSize;
  _ySize=sizeY;
  _halfYSize=0.5*_ySize;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

void
RectangleParticleCollider::getSize(double & sizeXOut,
                                   double & sizeYOut) const
{
sizeXOut=_xSize;
sizeYOut=_ySize;
}


//-------- Serialization operations ------------------------------------------

void
RectangleParticleCollider::serialize(SerializationDependencies &
                                     dependenciesOut,
                                     ArRef<AbstractOStream> output) const
{
ParticleCollider::serialize(dependenciesOut,output);
output->writeDouble(_xSize);
output->writeDouble(_ySize);
}

bool // success
RectangleParticleCollider::unserialize(const SerializationDependencies &
                                       dependencies,
                                       ArRef<AbstractIStream> input)
{
double xSize,ySize;
if(!ParticleCollider::unserialize(dependencies,input)||
   !input->readDouble(xSize)||
   !input->readDouble(ySize))
  {
  return(false);
  }
setSize(xSize,ySize);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
RectangleParticleCollider::_preUpdate(void) // Protected
{
ParticleCollider::_preUpdate();
_location->getBase().getPosition(_center.x,_center.y,_center.z);
_location->getBase().getLocalX(_xDir.x,_xDir.y,_xDir.z);
_location->getBase().getLocalY(_yDir.x,_yDir.y,_yDir.z);
_location->getBase().getLocalZ(_a,_b,_c); // normal=Z , a*x+b*y+c*z+d=0;
_d=-(_a*_center.x+
     _b*_center.y+
     _c*_center.z);
}

void
RectangleParticleCollider::_collide(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
double d1=_a*partInOut.oldPos.x+
          _b*partInOut.oldPos.y+
          _c*partInOut.oldPos.z+
          _d;
double d2=_a*partInOut.pos.x+
          _b*partInOut.pos.y+
          _c*partInOut.pos.z+
          _d;
if(((d1*d2)<=0.0)&&d2) // crosses plane
  {
  Util3D::Dbl3 dPos(partInOut.pos.x-partInOut.oldPos.x,
                    partInOut.pos.y-partInOut.oldPos.y,
                    partInOut.pos.z-partInOut.oldPos.z);
  double k=-d1/(_a*dPos.x+
                _b*dPos.y+
                _c*dPos.z);
  if(k<partInOut.kmin)
    {
    Util3D::Dbl3 inter(partInOut.oldPos.x+k*dPos.x-_center.x,
                       partInOut.oldPos.y+k*dPos.y-_center.y,
                       partInOut.oldPos.z+k*dPos.z-_center.z);
    double xProj=inter.x*_xDir.x+
                 inter.y*_xDir.y+
                 inter.z*_xDir.z;
    if((xProj<=_halfXSize)&&(xProj>=-_halfXSize))
      {
      double yProj=inter.x*_yDir.x+
                   inter.y*_yDir.y+
                   inter.z*_yDir.z;
      if((yProj<=_halfYSize)&&(yProj>=-_halfYSize))
        {
        partInOut.lastCollider=this;
        partInOut.kmin=k;
        double n=_a*partInOut.speed.x+
                 _b*partInOut.speed.y+
                 _c*partInOut.speed.z;
        _nSpeed.x=_a*n;
        _nSpeed.y=_b*n;
        _nSpeed.z=_c*n;
        _tSpeed.x=partInOut.speed.x-_nSpeed.x;
        _tSpeed.y=partInOut.speed.y-_nSpeed.y;
        _tSpeed.z=partInOut.speed.z-_nSpeed.z;
        }
      }
    }
  }
}


//----------------------------------------------------------------------------
// Class DiskParticleCollider
//----------------------------------------------------------------------------

AR_CLASS_DEF(DiskParticleCollider,ParticleCollider)

//-------- Construction / Destruction ----------------------------------------

DiskParticleCollider::DiskParticleCollider(ArCW & arCW) // Protected
: ParticleCollider(arCW),
  _radius(1.0),
  _r2(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

DiskParticleCollider::DiskParticleCollider(ArCW & arCW,
                                           ArRef<Base3D> location)
: ParticleCollider(arCW,location),                                // Protected
  _radius(1.0),
  _r2(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

DiskParticleCollider::~DiskParticleCollider(void) // Protected
{
  // Nothing to be done
}

//-------- Radius management -------------------------------------------------

bool // success
DiskParticleCollider::setRadius(double radius)
{
if(radius>0.0)
  {
  _radius=radius;
  _r2=_radius*_radius;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // radius
DiskParticleCollider::getRadius(void) const
{
return(_radius);
}

//-------- Serialization operations ------------------------------------------

void
DiskParticleCollider::serialize(SerializationDependencies & dependenciesOut,
                                ArRef<AbstractOStream> output) const
{
ParticleCollider::serialize(dependenciesOut,output);
output->writeDouble(_radius);
}

bool // success
DiskParticleCollider::unserialize(const SerializationDependencies &
                                  dependencies,
                                  ArRef<AbstractIStream> input)
{
double radius;
if(!ParticleCollider::unserialize(dependencies,input)||
   !input->readDouble(radius))
  {
  return(false);
  }
setRadius(radius);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
DiskParticleCollider::_preUpdate(void) // Protected
{
ParticleCollider::_preUpdate();
_location->getBase().getPosition(_center.x,_center.y,_center.z);
_location->getBase().getLocalZ(_a,_b,_c); // normal=Z , a*x+b*y+c*z+d=0;
_d=-(_a*_center.x+
     _b*_center.y+
     _c*_center.z);
}

void
DiskParticleCollider::_collide(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
double d1=_a*partInOut.oldPos.x+
          _b*partInOut.oldPos.y+
          _c*partInOut.oldPos.z+
          _d;
double d2=_a*partInOut.pos.x+
          _b*partInOut.pos.y+
          _c*partInOut.pos.z+
          _d;
if(((d1*d2)<=0.0)&&d2) // crosses plane
  {
  Util3D::Dbl3 dPos(partInOut.pos.x-partInOut.oldPos.x,
                    partInOut.pos.y-partInOut.oldPos.y,
                    partInOut.pos.z-partInOut.oldPos.z);
  double k=-d1/(_a*dPos.x+
                _b*dPos.y+
                _c*dPos.z);
  if(k<partInOut.kmin)
    {
    Util3D::Dbl3 inter(partInOut.oldPos.x+k*dPos.x-_center.x,
                       partInOut.oldPos.y+k*dPos.y-_center.y,
                       partInOut.oldPos.z+k*dPos.z-_center.z);
    if((inter.x*inter.x+inter.y*inter.y+inter.z*inter.z)<=_r2)
      {
      partInOut.lastCollider=this;
      partInOut.kmin=k;
      double n=_a*partInOut.speed.x+
               _b*partInOut.speed.y+
               _c*partInOut.speed.z;
      _nSpeed.x=_a*n;
      _nSpeed.y=_b*n;
      _nSpeed.z=_c*n;
      _tSpeed.x=partInOut.speed.x-_nSpeed.x;
      _tSpeed.y=partInOut.speed.y-_nSpeed.y;
      _tSpeed.z=partInOut.speed.z-_nSpeed.z;
      }
    }
  }
}


//----------------------------------------------------------------------------
// Class SphereParticleCollider
//----------------------------------------------------------------------------

AR_CLASS_DEF(SphereParticleCollider,ParticleCollider)

//-------- Construction / Destruction ----------------------------------------

SphereParticleCollider::SphereParticleCollider(ArCW & arCW) // Protected
: ParticleCollider(arCW),
  _radius(1.0),
  _r2(1.0),
  _rInv(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

SphereParticleCollider::SphereParticleCollider(ArCW & arCW,
                                               ArRef<Base3D> location)
: ParticleCollider(arCW,location),                                // Protected
  _radius(1.0),
  _r2(1.0),
  _rInv(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

SphereParticleCollider::~SphereParticleCollider(void)
{
  // Nothing to be done
}

//-------- Radius management -------------------------------------------------

bool // success
SphereParticleCollider::setRadius(double radius)
{
if(radius>0.0)
  {
  _radius=radius;
  _r2=_radius*_radius;
  _rInv=1.0/_radius;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // radius
SphereParticleCollider::getRadius(void) const
{
return(_radius);
}

//-------- Serialization operations ------------------------------------------

void
SphereParticleCollider::serialize(SerializationDependencies & dependenciesOut,
                                  ArRef<AbstractOStream> output) const
{
ParticleCollider::serialize(dependenciesOut,output);
output->writeDouble(_radius);
}

bool // success
SphereParticleCollider::unserialize(const SerializationDependencies &
                                    dependencies,
                                    ArRef<AbstractIStream> input)
{
double radius;
if(!ParticleCollider::unserialize(dependencies,input)||
   !input->readDouble(radius))
  {
  return(false);
  }
setRadius(radius);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
SphereParticleCollider::_preUpdate(void) // Protected
{
ParticleCollider::_preUpdate();
_location->getBase().getPosition(_center.x,_center.y,_center.z);
}

void
SphereParticleCollider::_collide(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
Util3D::Dbl3 dPos(partInOut.pos.x-partInOut.oldPos.x,
                  partInOut.pos.y-partInOut.oldPos.y,
                   partInOut.pos.z-partInOut.oldPos.z);
Util3D::Dbl3 pLoc(partInOut.oldPos.x-_center.x,
                  partInOut.oldPos.y-_center.y,
                  partInOut.oldPos.z-_center.z);
double A=dPos.x*dPos.x+dPos.y*dPos.y+dPos.z*dPos.z;
double B=2.0*(pLoc.x*dPos.x+pLoc.y*dPos.y+pLoc.z*dPos.z);
double C=pLoc.x*pLoc.x+pLoc.y*pLoc.y+pLoc.z*pLoc.z-_r2;
double delta=B*B-4.0*A*C;
if(delta>0.0)
  {
  delta=sqrt(delta);
  A*=2.0;
  double k=-1.0;
  double k1=(-B-delta)/A;
  double k2=(-B+delta)/A;
  if(k1>=0.0)
    {
    k=(((k2>=0.0)&&(k2<k1)) ? k2 : k1);
    }
  else if(k2>=0.0)
    {
    k=k2;
    }
  if((k>=0.0)&&(k<1.0)&&(k<partInOut.kmin))
    {
    partInOut.lastCollider=this;
    partInOut.kmin=k;
    pLoc.x+=k*dPos.x;
    pLoc.y+=k*dPos.y;
    pLoc.z+=k*dPos.z;
    double a=pLoc.x*_rInv;
    double b=pLoc.y*_rInv;
    double c=pLoc.z*_rInv;
    double n=a*partInOut.speed.x+
             b*partInOut.speed.y+
             c*partInOut.speed.z;
    _nSpeed.x=a*n;
    _nSpeed.y=b*n;
    _nSpeed.z=c*n;
    _tSpeed.x=partInOut.speed.x-_nSpeed.x;
    _tSpeed.y=partInOut.speed.y-_nSpeed.y;
    _tSpeed.z=partInOut.speed.z-_nSpeed.z;
    }
  }
}


//----------------------------------------------------------------------------
// Class CylinderParticleCollider
//----------------------------------------------------------------------------

AR_CLASS_DEF(CylinderParticleCollider,ParticleCollider)

//-------- Construction / Destruction ----------------------------------------

CylinderParticleCollider::CylinderParticleCollider(ArCW & arCW) // Protected
: ParticleCollider(arCW),
  _radius(1.0),
  _r2(1.0),
  _rInv(1.0),
  _height(2.0),
  _halfHeight(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

CylinderParticleCollider::CylinderParticleCollider(ArCW & arCW,
                                                   ArRef<Base3D> location)
: ParticleCollider(arCW,location),                                // Protected
  _radius(1.0),
  _r2(1.0),
  _rInv(1.0),
  _height(2.0),
  _halfHeight(1.0)
  // remaining members initialized in _preUpdate()
{
  // Nothing to be done
}

CylinderParticleCollider::~CylinderParticleCollider(void)
{
  // Nothing to be done
}

//-------- Size management ---------------------------------------------------

bool // success
CylinderParticleCollider::setRadius(double radius)
{
if(radius>0.0)
  {
  _radius=radius;
  _r2=_radius*_radius;
  _rInv=1.0/_radius;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // radius
CylinderParticleCollider::getRadius(void) const
{
return(_radius);
}

bool // success
CylinderParticleCollider::setHeight(double height)
{
if(height>0.0)
  {
  _height=height;
  _halfHeight=0.5*_height;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // height
CylinderParticleCollider::getHeight(void) const
{
return(_height);
}

//-------- Serialization operations ------------------------------------------

void
CylinderParticleCollider::serialize(SerializationDependencies & dependenciesOut,
                                    ArRef<AbstractOStream> output) const
{
ParticleCollider::serialize(dependenciesOut,output);
output->writeDouble(_radius);
output->writeDouble(_height);
}

bool // success
CylinderParticleCollider::unserialize(const SerializationDependencies &
                                      dependencies,
                                      ArRef<AbstractIStream> input)
{
double radius,height;
if(!ParticleCollider::unserialize(dependencies,input)||
   !input->readDouble(radius)||
   !input->readDouble(height))
  {
  return(false);
  }
setRadius(radius);
setHeight(height);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
CylinderParticleCollider::_collide(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
Util3D::Dbl3 pLoc=partInOut.oldPos;
Util3D::Dbl3 dPos(partInOut.pos.x-partInOut.oldPos.x,
                  partInOut.pos.y-partInOut.oldPos.y,
                  partInOut.pos.z-partInOut.oldPos.z);
_location->getBase().globalToLocalPosition(pLoc.x,pLoc.y,pLoc.z);
_location->getBase().globalToLocalVector(dPos.x,dPos.y,dPos.z);
double A=dPos.x*dPos.x+dPos.y*dPos.y;
double B=2.0*(pLoc.x*dPos.x+pLoc.y*dPos.y);
double C=pLoc.x*pLoc.x+pLoc.y*pLoc.y-_r2;
double delta=B*B-4.0*A*C;
if(delta>0.0)
  {
  delta=sqrt(delta);
  A*=2.0;
  double k;
  double k1=(-B-delta)/A;
  double k2=(-B+delta)/A;
  double zk1=pLoc.z+k1*dPos.z;
  double zk2=pLoc.z+k2*dPos.z;
  if((k1>=0.0)&&(k1<1.0)&&
     (zk1<=_halfHeight)&&(zk1>=-_halfHeight)) // k1 valid
    {
    if((k2>=0.0)&&(k2<1.0)&&
       (zk2<=_halfHeight)&&(zk2>=-_halfHeight)&&
       (k2<k1))                                  // k2 valid and closer
      {
      k=k2;
      pLoc.z=zk2;
      }
    else
      {
      k=k1;
      pLoc.z=zk1;
      }
    }
  else
    {
    if((k2>=0.0)&&(k2<1.0)&&
       (zk2<=_halfHeight)&&(zk2>=-_halfHeight)) // k2 valid
      {
      k=k2;
      pLoc.z=zk2;
      }
    else
      {
      return; // no collision
      }
    }
  if(k<partInOut.kmin)
    {
    partInOut.lastCollider=this;
    partInOut.kmin=k;
    pLoc.x+=k*dPos.x;
    pLoc.y+=k*dPos.y;
    double a=pLoc.x*_rInv;
    double b=pLoc.y*_rInv;
    double c=0.0;
    _location->getBase().localToGlobalVector(a,b,c);
    double n=a*partInOut.speed.x+
             b*partInOut.speed.y+
             c*partInOut.speed.z;
    _nSpeed.x=a*n;
    _nSpeed.y=b*n;
    _nSpeed.z=c*n;
    _tSpeed.x=partInOut.speed.x-_nSpeed.x;
    _tSpeed.y=partInOut.speed.y-_nSpeed.y;
    _tSpeed.z=partInOut.speed.z-_nSpeed.z;
    }
  }
}


//----------------------------------------------------------------------------
// Class ConeParticleCollider
//----------------------------------------------------------------------------

AR_CLASS_DEF(ConeParticleCollider,ParticleCollider)

//-------- Construction / Destruction ----------------------------------------

ConeParticleCollider::ConeParticleCollider(ArCW & arCW) // Protected
: ParticleCollider(arCW),
  _radius(1.0),
  _r2(1.0),
  _rInv(1.0),
  _height(2.0),
  _hInv(0.5),
  _h2Inv(0.25),
  _hCoef(1.0/sqrt(1.25))
{
  // Nothing to be done
}

ConeParticleCollider::ConeParticleCollider(ArCW & arCW,
                                           ArRef<Base3D> location)
: ParticleCollider(arCW,location),                                // Protected
  _radius(1.0),
  _r2(1.0),
  _rInv(1.0),
  _height(2.0),
  _hInv(0.5),
  _h2Inv(0.25),
  _hCoef(1.0/sqrt(1.25))
{
  // Nothing to be done
}

ConeParticleCollider::~ConeParticleCollider(void)
{
  // Nothing to be done
}

//-------- Size management ---------------------------------------------------

bool // success
ConeParticleCollider::setRadius(double radius)
{
if(radius>0.0)
  {
  _radius=radius;
  _r2=_radius*_radius;
  _rInv=1.0/_radius;
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // radius
ConeParticleCollider::getRadius(void) const
{
return(_radius);
}

bool // success
ConeParticleCollider::setHeight(double height)
{
if(height>0.0)
  {
  _height=height;
  _hInv=1.0/_height;
  _h2Inv=_hInv*_hInv;
  _hCoef=1.0/sqrt(1.0+_h2Inv);
  _onSetup();
  return(true);
  }
else
  {
  return(false);
  }
}

double // height
ConeParticleCollider::getHeight(void) const
{
return(_height);
}

//-------- Serialization operations ------------------------------------------

void
ConeParticleCollider::serialize(SerializationDependencies & dependenciesOut,
                                ArRef<AbstractOStream> output) const
{
ParticleCollider::serialize(dependenciesOut,output);
output->writeDouble(_radius);
output->writeDouble(_height);
}

bool // success
ConeParticleCollider::unserialize(const SerializationDependencies &
                                  dependencies,
                                  ArRef<AbstractIStream> input)
{
double radius,height;
if(!ParticleCollider::unserialize(dependencies,input)||
   !input->readDouble(radius)||
   !input->readDouble(height))
  {
  return(false);
  }
setRadius(radius);
setHeight(height);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
ConeParticleCollider::_collide(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
Util3D::Dbl3 pLoc=partInOut.oldPos;
Util3D::Dbl3 dPos(partInOut.pos.x-partInOut.oldPos.x,
                  partInOut.pos.y-partInOut.oldPos.y,
                  partInOut.pos.z-partInOut.oldPos.z);
_location->getBase().globalToLocalPosition(pLoc.x,pLoc.y,pLoc.z);
_location->getBase().globalToLocalVector(dPos.x,dPos.y,dPos.z);
pLoc.z+=0.5*_height; // FIXME: change formulas !
double A=dPos.x*dPos.x+dPos.y*dPos.y-
         dPos.z*dPos.z*_r2*_h2Inv;
double B=2.0*(pLoc.x*dPos.x+pLoc.y*dPos.y-
              pLoc.z*dPos.z*_r2*_h2Inv+dPos.z*_r2*_hInv);
double C=pLoc.x*pLoc.x+pLoc.y*pLoc.y-_r2-
         pLoc.z*pLoc.z*_r2*_h2Inv+2.0*pLoc.z*_r2*_hInv;
double delta=B*B-4.0*A*C;
if(delta>0.0)
  {
  delta=sqrt(delta);
  A*=2.0;
  double k;
  double k1=(-B-delta)/A;
  double k2=(-B+delta)/A;
  double zk1=pLoc.z+k1*dPos.z;
  double zk2=pLoc.z+k2*dPos.z;
  if((k1>=0.0)&&(k1<1.0)&&
     (zk1>=0.0)&&(zk1<_height)) // k1 valid
    {
    if((k2>=0.0)&&(k2<1.0)&&
       (zk2>=0.0)&&(zk2<_height)&&
       (k2<k1))                     // k2 valid and closer
      {
      k=k2;
      pLoc.z=zk2;
      }
    else
      {
      k=k1;
      pLoc.z=zk1;
      }
    }
  else
    {
    if((k2>=0.0)&&(k2<1.0)&&
       (zk2>=0.0)&&(zk2<_height)) // k2 valid
      {
      k=k2;
      pLoc.z=zk2;
      }
    else
      {
      return; // no collision
      }
    }
  if(k<partInOut.kmin)
    {
    partInOut.lastCollider=this;
    partInOut.kmin=k;
    pLoc.x+=k*dPos.x;
    pLoc.y+=k*dPos.y;
    double r=1.0/(1.0-pLoc.z*_hInv);
    double a=pLoc.x*_rInv*_hCoef*r;
    double b=pLoc.y*_rInv*_hCoef*r;
    double c=_radius*_hInv*_hCoef;
    _location->getBase().localToGlobalVector(a,b,c);
    double n=a*partInOut.speed.x+
             b*partInOut.speed.y+
             c*partInOut.speed.z;
    _nSpeed.x=a*n;
    _nSpeed.y=b*n;
    _nSpeed.z=c*n;
    _tSpeed.x=partInOut.speed.x-_nSpeed.x;
    _tSpeed.y=partInOut.speed.y-_nSpeed.y;
    _tSpeed.z=partInOut.speed.z-_nSpeed.z;
    }
  }
}


//----------------------------------------------------------------------------
// Class ParticleSystem
//----------------------------------------------------------------------------

AR_CLASS_DEF(ParticleSystem,Base3D)

//-------- Construction / Destruction ----------------------------------------

ParticleSystem::ParticleSystem(ArCW & arCW) // Protected
: Base3D(arCW),
  _scenes(),
  _width(0.1),
  _height(0.1),
  _widthAngle(M_PI/8.0),
  _heightAngle(M_PI/8.0),
  _rounded(true),
  _particleRate(1000.0),
  _emissionSpeed(1.0),
  _emissionSpeedDisp(0.0),
  _particleDuration(3.0),
  _gravity(0.0,0.0,-9.81),
  _lines(true),
  _thickness(1.0),
  _texture(),
  _initAngle(0.0),
  _initAngleDisp(0.0),
  _angularSpeed(0.0),
  _angularSpeedDisp(0.0),
  _size(1.0,1.0),
  _sizeDisp(0.0,0.0),
  _color(0.8,0.8,0.8,1.0),
  _lighting(false),
  _avgDir(0.0,0.0,0.0),
  _bbox(),
  _nbBlended(0),
  _particles(),
  _colliders(),
  _frameID(Renderer3D::_currentFrameID),
  _updateCB(thisRef()),
  _emissionCB(thisRef()),
  _evolutionCB(thisRef()),
  _renderingCB(thisRef()),
  _addColliderCB(thisRef()),
  _removeColliderCB(thisRef())
{
Scene3D::_globalScene->addParticleSystem(thisRef());
}

ParticleSystem::~ParticleSystem(void) // Protected
{
setTexture(Texture::nullRef());
while(!_scenes.empty())
  {
  _scenes.back()->_removeParticleSystem(thisRef());
  }
while(!_colliders.empty())
  {
  removeCollider(_colliders.back());
  }
}

//-------- Emission ----------------------------------------------------------

bool // success
ParticleSystem::setSection(double width,
                           double height,
                           double widthAngle,
                           double heightAngle,
                           bool rounded)
{
if((width>=0.0)&&(height>=0.0)&&
   (widthAngle>=0.0)&&(widthAngle<=M_2PI)&&
   (heightAngle>=0.0)&&(heightAngle<=M_2PI))
  {
  _width=width;
  _height=height;
  _widthAngle=widthAngle;
  _heightAngle=heightAngle;
  _rounded=rounded;
  _onEmission();
  return(true);
  }
else
  {
  return(false);
  }
}

void
ParticleSystem::getSection(double & widthOut,
                           double & heightOut,
                           double & widthAngleOut,
                           double & heightAngleOut,
                           bool & roundedOut) const
{
widthOut=_width;
heightOut=_height;
widthAngleOut=_widthAngle;
heightAngleOut=_heightAngle;
roundedOut=_rounded;
}

bool // success
ParticleSystem::setParticleRate(double particleRate)
{
if(particleRate>=0.0)
  {
  _particleRate=particleRate;
  _onEmission();
  return(true);
  }
else
  {
  return(false);
  }
}

double // particle rate
ParticleSystem::getParticleRate(void) const
{
return(_particleRate);
}

void
ParticleSystem::setEmissionSpeed(double speed)
{
_emissionSpeed=speed;
_onEmission();
}

double // particle emission speed
ParticleSystem::getEmissionSpeed(void) const
{
return(_emissionSpeed);
}

void
ParticleSystem::setEmissionSpeedDispersion(double dispersion)
{
_emissionSpeedDisp=dispersion;
_onEmission();
}

double // dispersion around particle emission speed
ParticleSystem::getEmissionSpeedDispersion(void) const
{
return(_emissionSpeedDisp);
}

//-------- Evolution ---------------------------------------------------------

bool // success
ParticleSystem::setParticleDuration(double particleDuration)
{
if(particleDuration>=0.0)
  {
  _particleDuration=particleDuration;
  _onEvolution();
  return(true);
  }
else
  {
  return(false);
  }
}

double // particle duration
ParticleSystem::getParticleDuration(void) const
{
return(_particleDuration);
}

unsigned int // current particle number
ParticleSystem::getNbParticles(void) const
{
return(_particles.size());
}

void
ParticleSystem::setGravity(double gx,
                           double gy,
                           double gz)
{
_gravity.x=gx;
_gravity.y=gy;
_gravity.z=gz;
_onEvolution();
}

void
ParticleSystem::getGravity(double & gxOut,
                           double & gyOut,
                           double & gzOut) const
{
gxOut=_gravity.x;
gyOut=_gravity.y;
gzOut=_gravity.z;
}

void
ParticleSystem::readBoundingBox(ArRef<BoundingBox3D> bboxOut) const
{
bboxOut->accessBoundingBox()=_bbox;
}

void
ParticleSystem::update(double dt)
{
ArRef<ParticleSystem> flag(this);
if(!_updateCB.empty())
  {
  ParticleSystem::UpdateEvent event;
  event.dt=dt;
  _updateCB.fire(event);
  }
if(!flag)
  {
  return;
  }
unsigned int i;
for(i=_colliders.size();i--;)
  {
  _colliders[i]->_preUpdate();
  }
_bbox.valid=false;
_avgDir.x=_avgDir.y=_avgDir.z=0.0;
StlList<ParticleCollider::Particle>::iterator it;
for(it=_particles.begin();it!=_particles.end();++it)
  {
  ParticleCollider::Particle & part=*it;
  bool wasBlended=part.color.a<0.99;
  bool destroyParticle=true;
  if((part.duration+=dt)<=_particleDuration)
    {
    if(_updateParticle(dt,static_cast<ParticleSystem::Particle &>(part))&&
       _manageCollisions(part))
      {
      if(part.color.a<0.99)
        {
        if(!wasBlended)
          {
          ++_nbBlended;
          }
        }
      else
        {
        if(wasBlended)
          {
          --_nbBlended;
          }
        }
      _avgDir.x+=part.speed.x;
      _avgDir.y+=part.speed.y;
      _avgDir.z+=part.speed.z;
      _bbox.addXYZ(part.pos.x,part.pos.y,part.pos.z);
      destroyParticle=false;
      }
    }
  if(destroyParticle)
    {
    if(wasBlended)
      {
      --_nbBlended;
      }
    _particles.erase(it--);
    }
  }
if(_particleRate)
  {
  unsigned int nbToAdd=1+(unsigned int)(_particleRate*dt);
  double offset=dt/nbToAdd;
  ParticleCollider::Particle part;
  double initDt=dt;
  while(nbToAdd--)
    {
    initDt-=offset;
    part.duration=0.0;
    part.lastCollider=(ParticleCollider *)0;
    if(_initParticle(static_cast<ParticleSystem::Particle &>(part))&&
       _updateParticle(initDt,static_cast<ParticleSystem::Particle &>(part))&&
       _manageCollisions(part))
      {
      if(part.color.a<0.99)
        {
        ++_nbBlended;
        }
      _particles.push_back(part);
      _avgDir.x+=part.speed.x;
      _avgDir.y+=part.speed.y;
      _avgDir.z+=part.speed.z;
      _bbox.addXYZ(part.pos.x,part.pos.y,part.pos.z);
      }
    }
  }
_postRedisplay();
StlVector<ParticleCollider::ParticleCollisionEvent > events;
events.reserve(_colliders.size());
for(i=_colliders.size();i--;)
  {
  if(_colliders[i]->_collisionCount&&
     !_colliders[i]->_particleCollisionCB.empty())
    {
    ParticleCollider::ParticleCollisionEvent event;
    events.push_back(event);
    events.back().source=_colliders[i];
    events.back().count=_colliders[i]->_collisionCount;
    events.back().normalSpeedX=_colliders[i]->_nSpeedSum.x;
    events.back().normalSpeedY=_colliders[i]->_nSpeedSum.y;
    events.back().normalSpeedZ=_colliders[i]->_nSpeedSum.z;
    events.back().tangentialSpeedX=_colliders[i]->_tSpeedSum.x;
    events.back().tangentialSpeedY=_colliders[i]->_tSpeedSum.y;
    events.back().tangentialSpeedZ=_colliders[i]->_tSpeedSum.z;
    events.back().dt=dt;
    events.back().particleSystem=thisRef();
    }
  }
for(i=events.size();i--;)
  {
  if(events[i].source)
    {
    events[i].source->_particleCollisionCB.fire(events[i]);
    }
  }
}

//-------- Rendering ---------------------------------------------------------

void
ParticleSystem::setLineAppearance(bool drawLines)
{
_lines=drawLines;
_postRedisplay();
_onRendering();
}

bool // draws lines
ParticleSystem::getLineAppearance(void) const
{
return(_lines);
}

bool // success
ParticleSystem::setThickness(unsigned int thickness)
{
if(thickness)
  {
  _thickness=thickness;
  _postRedisplay();
  _onRendering();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // line/point thickness
ParticleSystem::getThickness(void) const
{
return((unsigned int)_thickness);
}

void
ParticleSystem::setTexture(ArRef<Texture> texture)
{
if((texture!=_texture)&&!(texture&&texture->fail()))
  {
  ArRef<Texture> old=_texture;
  _texture=texture;
  if(old)
    {
    old->_removeParticleSystem(this);
    }
  if(_texture)
    {
    _texture->_addParticleSystem(this);
    }
  _postRedisplay();
  _onRendering();
  }
}

ArConstRef<Texture> // particle texture
ParticleSystem::getTexture(void) const
{
return(_texture);
}

ArRef<Texture> // particle texture
ParticleSystem::accessTexture(void)
{
return(_texture);
}

void
ParticleSystem::setInitialAngle(double angle)
{
_initAngle=angle;
_onRendering();
}

double // initial particle angle
ParticleSystem::getInitialAngle(void) const
{
return(_initAngle);
}

void
ParticleSystem::setInitialAngleDispersion(double dispersion)
{
_initAngleDisp=dispersion;
_onRendering();
}

double // dispersion around initial particle angle
ParticleSystem::getInitialAngleDispersion(void) const
{
return(_initAngleDisp);
}

void
ParticleSystem::setAngularSpeed(double speed)
{
_angularSpeed=speed;
_onRendering();
}

double // angular particle speed
ParticleSystem::getAngularSpeed(void) const
{
return(_angularSpeed);
}

void
ParticleSystem::setAngularSpeedDispersion(double dispersion)
{
_angularSpeedDisp=dispersion;
_onRendering();
}

double // dispersion around angular particle speed
ParticleSystem::getAngularSpeedDispersion(void) const
{
return(_angularSpeedDisp);
}

void
ParticleSystem::setSize(double width,
                        double height)
{
_size.x=width;
_size.y=height;
_onRendering();
}

void
ParticleSystem::getSize(double & widthOut,
                        double & heightOut) const
{
widthOut=_size.x;
heightOut=_size.y;
}

void
ParticleSystem::setSizeDispersion(double widthDispersion,
                                  double heightDispersion)
{
_sizeDisp.x=widthDispersion;
_sizeDisp.y=heightDispersion;
_onRendering();
}

void
ParticleSystem::getSizeDispersion(double & widthDispersionOut,
                                  double & heightDispersionOut) const
{
widthDispersionOut=_sizeDisp.x;
heightDispersionOut=_sizeDisp.y;
}

bool // success
ParticleSystem::setColor(double r,
                         double g,
                         double b,
                         double a)
{
if((r>=0.0)&&(r<=1.0)&&(g>=0.0)&&(g<=1.0)&&
   (b>=0.0)&&(b<=1.0)&&(a>=0.0)&&(a<=1.0))
  {
  _color.x=r;
  _color.y=g;
  _color.z=b;
  _color.a=a;
  _onRendering();
  return(true);
  }
else
  {
  return(false);
  }
}

void
ParticleSystem::getColor(double & rOut,
                         double & gOut,
                         double & bOut,
                         double & aOut) const
{
rOut=_color.x;
gOut=_color.y;
bOut=_color.z;
aOut=_color.a;
}

void
ParticleSystem::setLighting(bool lighting)
{
_lighting=lighting;
_postRedisplay();
_onRendering();
}

bool // lighting is used for color
ParticleSystem::getLighting(void) const
{
return(_lighting);
}

bool // currently displayed by a renderer
ParticleSystem::isCurrentlyDisplayed(void) const
{
return(_frameID==Renderer3D::_currentFrameID);
}

//-------- Collider management -----------------------------------------------

void
ParticleSystem::addCollider(ArRef<ParticleCollider> collider)
{
if(!hasCollider(collider))
  {
  collider->_addParticleSystem(thisPtr());
  _colliders.push_back(collider);
  if(!_addColliderCB.empty())
    {
    ParticleSystem::AddColliderEvent event;
    event.collider=collider;
    _addColliderCB.fire(event);
    }
  }
}

void
ParticleSystem::removeCollider(ArRef<ParticleCollider> collider)
{
for(unsigned int i=_colliders.size();i--;)
  {
  if(_colliders[i]==collider)
    {
    StlVectorFastErase(_colliders,i);
    collider->_removeParticleSystem(thisPtr());
    if(!_removeColliderCB.empty())
      {
      ParticleSystem::RemoveColliderEvent event;
      event.collider=collider;
      _removeColliderCB.fire(event);
      }
    break;
    }
  }
}

bool // collider is contained
ParticleSystem::hasCollider(ArConstRef<ParticleCollider> collider) const
{
for(unsigned int i=_colliders.size();i--;)
  {
  if(_colliders[i]==collider)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of colliders
ParticleSystem::getNbColliders(void) const
{
return(_colliders.size());
}

ArConstRef<ParticleCollider> // index'th collider
ParticleSystem::getCollider(unsigned int index) const
{
return(_colliders[index]);
}

ArRef<ParticleCollider> // index'th collider
ParticleSystem::accessCollider(unsigned int index)
{
return(_colliders[index]);
}

//-------- Serialization operations ------------------------------------------

void
ParticleSystem::extractDependencies(SerializationDependencies & dependenciesOut,
                                    bool recursive) const
{
Base3D::extractDependencies(dependenciesOut,recursive);
if(_texture)
  {
  dependenciesOut.addReference(_texture);
  }
unsigned int nbColliders=_colliders.size();
for(unsigned int i=0;i<nbColliders;i++)
  {
  dependenciesOut.addReference(_colliders[i]);
  }
}

void
ParticleSystem::serialize(SerializationDependencies & dependenciesOut,
                          ArRef<AbstractOStream> output) const
{
Base3D::serialize(dependenciesOut,output);
output->writeDouble(_width);
output->writeDouble(_height);
output->writeDouble(_widthAngle);
output->writeDouble(_heightAngle);
output->writeDouble(_particleRate);
output->writeDouble(_emissionSpeed);
output->writeDouble(_emissionSpeedDisp);
output->writeDouble(_particleDuration);
output->writeDouble(_gravity.x);
output->writeDouble(_gravity.y);
output->writeDouble(_gravity.z);
output->writeDouble(_thickness);
output->writeDouble(_initAngle);
output->writeDouble(_initAngleDisp);
output->writeDouble(_angularSpeed);
output->writeDouble(_angularSpeedDisp);
output->writeDouble(_size.x);
output->writeDouble(_size.y);
output->writeDouble(_sizeDisp.x);
output->writeDouble(_sizeDisp.y);
output->writeDouble(_color.x);
output->writeDouble(_color.y);
output->writeDouble(_color.z);
output->writeDouble(_color.a);
unsigned int flags=(_rounded  ? 1<<0 : 0)+
                   (_lines    ? 1<<1 : 0)+
                   (_texture  ? 1<<2 : 0)+
                   (_lighting ? 1<<3 : 0);
output->writeUInt(flags);
unsigned int nbColliders=_colliders.size();
output->writeUInt(nbColliders);
unsigned int nbParticles=_particles.size();
output->writeUInt(nbParticles);
if(_texture)
  {
  output->writeUInt(dependenciesOut.addReference(_texture));
  }
unsigned int i;
for(i=0;i<nbColliders;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_colliders[i]));
  }
StlList<ParticleCollider::Particle>::const_iterator it=_particles.begin();
for(;it!=_particles.end();it++) // Serialize particles state
  {
  output->writeDouble((*it).duration);
  output->writeDouble((*it).oldPos.x);
  output->writeDouble((*it).oldPos.y);
  output->writeDouble((*it).oldPos.z);
  output->writeDouble((*it).pos.x);
  output->writeDouble((*it).pos.y);
  output->writeDouble((*it).pos.z);
  output->writeDouble((*it).pos.a);
  output->writeDouble((*it).speed.x);
  output->writeDouble((*it).speed.y);
  output->writeDouble((*it).speed.z);
  output->writeDouble((*it).speed.a);
  output->writeDouble((*it).color.x);
  output->writeDouble((*it).color.y);
  output->writeDouble((*it).color.z);
  output->writeDouble((*it).color.a);
  output->writeDouble((*it).size.x);
  output->writeDouble((*it).size.y);
  }
}

bool // success
ParticleSystem::unserialize(const SerializationDependencies & dependencies,
                            ArRef<AbstractIStream> input)
{
double width,height,widthAngle,heightAngle;
double particleRate,emissionSpeed,emissionSpeedDisp,particleDuration;
Util3D::Dbl3 gravity;
double thickness,initAngle,initAngleDisp,angularSpeed,angularSpeedDisp;
Util3D::Dbl2 size,sizeDisp;
Util3D::Dbl4 color;
unsigned int flags;
unsigned int nbColliders;
unsigned int texture;
StlVector<unsigned int> colliders;
unsigned int nbParticles;
StlList<ParticleCollider::Particle> particles;
if(!Base3D::unserialize(dependencies,input)||
   !input->readDouble(width)||
   !input->readDouble(height)||
   !input->readDouble(widthAngle)||
   !input->readDouble(heightAngle)||
   !input->readDouble(particleRate)||
   !input->readDouble(emissionSpeed)||
   !input->readDouble(emissionSpeedDisp)||
   !input->readDouble(particleDuration)||
   !input->readDouble(gravity.x)||
   !input->readDouble(gravity.y)||
   !input->readDouble(gravity.z)||
   !input->readDouble(thickness)||
   !input->readDouble(initAngle)||
   !input->readDouble(initAngleDisp)||
   !input->readDouble(angularSpeed)||
   !input->readDouble(angularSpeedDisp)||
   !input->readDouble(size.x)||
   !input->readDouble(size.y)||
   !input->readDouble(sizeDisp.x)||
   !input->readDouble(sizeDisp.y)||
   !input->readDouble(color.x)||
   !input->readDouble(color.y)||
   !input->readDouble(color.z)||
   !input->readDouble(color.a)||
   !input->readUInt(flags)||
   !input->readUInt(nbColliders)||
   !input->readUInt(nbParticles))
  {
  return(false);
  }
bool hasTexture=flags&(1<<2);
if(hasTexture)
  {
  if(!input->readUInt(texture))
    {
    return(false);
    }
  }
unsigned int i;
for(i=0;i<nbColliders;i++)
  {
  colliders.push_back(0);
  if(!input->readUInt(colliders.back()))
    {
    return(false);
    }
  }
for(i=0;i<nbParticles;i++)
  {
  particles.push_back(ParticleCollider::Particle());
  particles.back().lastCollider=(ParticleCollider *)0;
  if(!input->readDouble(particles.back().duration)||
     !input->readDouble(particles.back().oldPos.x)||
     !input->readDouble(particles.back().oldPos.y)||
     !input->readDouble(particles.back().oldPos.z)||
     !input->readDouble(particles.back().pos.x)||
     !input->readDouble(particles.back().pos.y)||
     !input->readDouble(particles.back().pos.z)||
     !input->readDouble(particles.back().pos.a)||
     !input->readDouble(particles.back().speed.x)||
     !input->readDouble(particles.back().speed.y)||
     !input->readDouble(particles.back().speed.z)||
     !input->readDouble(particles.back().speed.a)||
     !input->readDouble(particles.back().color.x)||
     !input->readDouble(particles.back().color.y)||
     !input->readDouble(particles.back().color.z)||
     !input->readDouble(particles.back().color.a)||
     !input->readDouble(particles.back().size.x)||
     !input->readDouble(particles.back().size.y))
    {
    return(false);
    }
  }
_width=width;
_height=height;
_widthAngle=widthAngle;
_heightAngle=heightAngle;
_particleRate=particleRate;
_emissionSpeed=emissionSpeed;
_emissionSpeedDisp=emissionSpeedDisp;
_particleDuration=particleDuration;
_gravity=gravity;
_thickness=thickness;
_initAngle=initAngle;
_initAngleDisp=initAngleDisp;
_angularSpeed=angularSpeed;
_angularSpeedDisp=angularSpeedDisp;
_size=size;
_sizeDisp=sizeDisp;
_color=color;
_rounded=flags&(1<<0);
_lines=flags&(1<<1);
_lighting=flags&(1<<3);
if(hasTexture)
  {
  setTexture(ar_down_cast<Texture>(dependencies.findReference(texture)));
  }
for(i=0;i<nbColliders;i++)
  {
  addCollider(ar_down_cast<ParticleCollider>
                          (dependencies.findReference(colliders[i])));
  }
_particles.swap(particles);
_onEmission();
_onEvolution();
_onRendering();
_postRedisplay();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
ParticleSystem::_addScene(ArRef<Scene3D> scene) // Protected
{
_scenes.push_back(scene);
}

void
ParticleSystem::_removeScene(ArRef<Scene3D> scene) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  if(_scenes[i]==scene)
    {
    StlVectorFastErase(_scenes,i);
    break;
    }
  }
}

void
ParticleSystem::_postRedisplay(void) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  _scenes[i]->_postRedisplay();
  }
}

bool // blending is needed for rendering
ParticleSystem::_needsBlending(void) const // Protected
{
return(_nbBlended||(_texture&&_texture->getBlending()));
}

bool // alpha-test is needed for rendering
ParticleSystem::_needsAlphaTest(void) const // Protected
{
return(_texture&&_texture->getAlphaTest());
}

bool // keep particle
ParticleSystem::_manageCollisions(ParticleCollider::Particle & partInOut)
{                                                                 // Protected
partInOut.lastCollider=(ParticleCollider *)0;
partInOut.kmin=DBL_MAX;
for(unsigned int i=_colliders.size();i--;)
  {
  _colliders[i]->_collide(partInOut);
  }
if(partInOut.lastCollider)
  {
  ParticleCollider * coll=partInOut.lastCollider;
  ++(coll->_collisionCount);
  coll->_nSpeedSum.x+=coll->_nSpeed.x;
  coll->_nSpeedSum.y+=coll->_nSpeed.y;
  coll->_nSpeedSum.z+=coll->_nSpeed.z;
  coll->_tSpeedSum.x+=coll->_tSpeed.x;
  coll->_tSpeedSum.y+=coll->_tSpeed.y;
  coll->_tSpeedSum.z+=coll->_tSpeed.z;
  if(coll->_destroyFlag)
    {
    return(false);
    }
  else
    {
    partInOut.speed.x=coll->_tSpeed.x*coll->_tFactor-
                      coll->_nSpeed.x*coll->_nFactor;
    partInOut.speed.y=coll->_tSpeed.y*coll->_tFactor-
                      coll->_nSpeed.y*coll->_nFactor;
    partInOut.speed.z=coll->_tSpeed.z*coll->_tFactor-
                      coll->_nSpeed.z*coll->_nFactor;
    }
  }
return(true);
}

bool // keep particle
ParticleSystem::_initParticle(ParticleSystem::Particle & particleInOut)
{
double dy,dz;
Util3D::Base b(_base);
if(_rounded)
  {
  double a=ArSystem::realRand()*M_2PI;
  double r=sqrt(ArSystem::realRand())*0.5;
  dy=cos(a)*r;
  dz=sin(a)*r;
  }
else
  {
  dy=ArSystem::realRand()-0.5;
  dz=ArSystem::realRand()-0.5;
  }
b.translate(0.0,dy*_width,dz*_height);
b.yaw(dy*_widthAngle);
b.pitch(-dz*_heightAngle);
b.getPosition(particleInOut.accessOldPosition().x,
              particleInOut.accessOldPosition().y,
              particleInOut.accessOldPosition().z);
particleInOut.accessPosition().x=particleInOut.accessOldPosition().x;
particleInOut.accessPosition().y=particleInOut.accessOldPosition().y;
particleInOut.accessPosition().z=particleInOut.accessOldPosition().z;
particleInOut.accessPosition().a=ParticleSystem::_randomize(_initAngle,
                                                            _initAngleDisp);
Util3D::Dbl3 p;
b.getLocalX(p.x,p.y,p.z);
double d=ParticleSystem::_randomize(_emissionSpeed,_emissionSpeedDisp);
particleInOut.accessSpeed().x=p.x*d;
particleInOut.accessSpeed().y=p.y*d;
particleInOut.accessSpeed().z=p.z*d;
particleInOut.accessSpeed().a=ParticleSystem::_randomize(_angularSpeed,
                                                         _angularSpeedDisp);
particleInOut.accessColor()=_color;
particleInOut.accessSize().x=ParticleSystem::_randomize(_size.x,
                                                        _sizeDisp.x);
particleInOut.accessSize().y=ParticleSystem::_randomize(_size.y,
                                                        _sizeDisp.y);
return(true);
}

bool // keep particle
ParticleSystem::_updateParticle(double dt,
                                ParticleSystem::Particle & particleInOut)
{
if(particleInOut.getCollider()) // on collision try new speed
  {
  particleInOut.accessPosition().x=particleInOut.getOldPosition().x
                                  +particleInOut.getSpeed().x*dt;
  particleInOut.accessPosition().y=particleInOut.getOldPosition().y
                                  +particleInOut.getSpeed().y*dt;
  particleInOut.accessPosition().z=particleInOut.getOldPosition().z
                                  +particleInOut.getSpeed().z*dt;
  }
else
  {
  particleInOut.accessOldPosition().x=particleInOut.getPosition().x;
  particleInOut.accessOldPosition().y=particleInOut.getPosition().y;
  particleInOut.accessOldPosition().z=particleInOut.getPosition().z;
  double d=0.5*dt*dt;
  particleInOut.accessPosition().x+=particleInOut.getSpeed().x*dt+_gravity.x*d;
  particleInOut.accessPosition().y+=particleInOut.getSpeed().y*dt+_gravity.y*d;
  particleInOut.accessPosition().z+=particleInOut.getSpeed().z*dt+_gravity.z*d;
  particleInOut.accessSpeed().x+=_gravity.x*dt;
  particleInOut.accessSpeed().y+=_gravity.y*dt;
  particleInOut.accessSpeed().z+=_gravity.z*dt;
  }
particleInOut.accessPosition().a+=particleInOut.getSpeed().a*dt;
return(true);
}

double
ParticleSystem::_randomize(double value,
                           double dispersion) // Protected, Static
{
return(dispersion ? value+dispersion*(ArSystem::realRand()-0.5) : value);
}

void
ParticleSystem::_render(const Renderer3D * renderer) const // Protected
{
const Util3D::Base & base=renderer->_base;
Util3D::Dbl3 p;
base.getLocalX(p.x,p.y,p.z);
if(_lighting)
  {
  float defaultMaterial[]={0.0,0.0,0.0,1.0};
  glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,defaultMaterial);
  glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,defaultMaterial);
  glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,defaultMaterial);
  glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,0.0);
  glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  }
else
  {
  glDisable(GL_LIGHTING);
  glShadeModel(GL_FLAT);
  }
StlList<ParticleCollider::Particle>::const_iterator it,start,end;
bool up;
if(p.x*_avgDir.x+p.y*_avgDir.y+p.z*_avgDir.z>0.0)
  {
  start=_particles.begin();
  end=_particles.end();
  up=true;
  }
else
  {
  start=--_particles.end();
  end=--_particles.begin();
  up=false;
  }
base.getPosition(p.x,p.y,p.z);
if(_texture)
  {
  glEnable(GL_TEXTURE_2D);
  GL::selectTexture(_texture->getTextureID(),true,true,true);
  glBegin(GL_QUADS);
  for(it=start;it!=end;up ? ++it : --it)
    {
    const ParticleCollider::Particle & part=*it;
    double s=sin(part.pos.a);
    double c=cos(part.pos.a);
    Util3D::Dbl3 p1(0.0,+c*part.size.x-s*part.size.y,
                        -s*part.size.x-c*part.size.y);
    Util3D::Dbl3 p2(0.0,-c*part.size.x-s*part.size.y,
                        +s*part.size.x-c*part.size.y);
    Util3D::Dbl3 p3(0.0,-c*part.size.x+s*part.size.y,
                        +s*part.size.x+c*part.size.y);
    Util3D::Dbl3 p4(0.0,+c*part.size.x+s*part.size.y,
                        -s*part.size.x+c*part.size.y);
    base.localToGlobalVector(p1.x,p1.y,p1.z);
    base.localToGlobalVector(p2.x,p2.y,p2.z);
    base.localToGlobalVector(p3.x,p3.y,p3.z);
    base.localToGlobalVector(p4.x,p4.y,p4.z);
    glNormal3f(p.x-part.pos.x,p.y-part.pos.y,p.z-part.pos.z);
    glColor4f(part.color.x,part.color.y,part.color.z,part.color.a);
    glTexCoord2f(0.0,0.0);
    glVertex3f(part.pos.x+p1.x,part.pos.y+p1.y,part.pos.z+p1.z);
    glTexCoord2f(1.0,0.0);
    glVertex3f(part.pos.x+p2.x,part.pos.y+p2.y,part.pos.z+p2.z);
    glTexCoord2f(1.0,1.0);
    glVertex3f(part.pos.x+p3.x,part.pos.y+p3.y,part.pos.z+p3.z);
    glTexCoord2f(0.0,1.0);
    glVertex3f(part.pos.x+p4.x,part.pos.y+p4.y,part.pos.z+p4.z);
    }
  glEnd(); // GL_QUADS
  glDisable(GL_TEXTURE_2D);
  }
else if(_lines)
  {
  glLineWidth(_thickness);
  glBegin(GL_LINES);
  for(it=start;it!=end;up ? ++it : --it)
    {
    const ParticleCollider::Particle & part=*it;
    if(!part.lastCollider) // avoid zero length lines
      {
      glNormal3f(p.x-part.pos.x,p.y-part.pos.y,p.z-part.pos.z);
      glColor4f(part.color.x,part.color.y,part.color.z,part.color.a);
      glVertex3dv(&part.oldPos.x);
      glVertex3dv(&part.pos.x);
      }
    }
  glEnd(); // GL_LINES
  glLineWidth(1.0);
  }
else
  {
  glPointSize(_thickness);
  glBegin(GL_POINTS);
  for(it=start;it!=end;up ? ++it : --it)
    {
    const ParticleCollider::Particle & part=*it;
    glNormal3f(p.x-part.pos.x,p.y-part.pos.y,p.z-part.pos.z);
    glColor4f(part.color.x,part.color.y,part.color.z,part.color.a);
    glVertex3dv(&part.pos.x);
    }
  glEnd(); // GL_POINTS
  glPointSize(1.0);
  }
if(_lighting)
  {
  glDisable(GL_COLOR_MATERIAL);
  }
else
  {
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  }
const_cast<ParticleSystem *>(this)->_frameID=Renderer3D::_currentFrameID;
}

void
ParticleSystem::_onEmission(void)
{
if(!_emissionCB.empty())
  {
  ParticleSystem::EmissionEvent event;
  _emissionCB.fire(event);
  }
}

void
ParticleSystem::_onEvolution(void)
{
if(!_evolutionCB.empty())
  {
  ParticleSystem::EvolutionEvent event;
  _evolutionCB.fire(event);
  }
}

void
ParticleSystem::_onRendering(void)
{
if(!_renderingCB.empty())
  {
  ParticleSystem::RenderingEvent event;
  _renderingCB.fire(event);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
