// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/renderer3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/renderer3D.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/light3D.h"
#include "AReVi/Lib3D/particleSystem.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/texture.h"
#include "AReVi/Lib3D/material3D.h"
#include "AReVi/Lib3D/osd.h"

#include "AReVi/Shapes/shape3D.h"
#include "AReVi/Shapes/shapePart3D.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//-------- Initialization ----------------------------------------------------

static ArSystem::InitState _initState=ArSystem::AR_DOWN;

void
Renderer3D::_init(void) // Static, Protected
{
if(_initState==ArSystem::AR_DOWN)
  {
  _initState=ArSystem::AR_GOING_UP;
  Renderer3D::_needsRedisplay=false;
  Renderer3D::_lastFrameRate=ArSystem::getTime();
  Renderer3D::_currentFrameID=0;
  Renderer3D::_nbFrames=0;
  Renderer3D::_frameRate=0.0;
  Renderer3D::_lastRedraw=ArSystem::getTime();
  Renderer3D::_redrawInterval=0;
  Renderer3D::_renderers=new StlVector<ArPtr<Renderer3D> >;
  Renderer3D::_renderInfoPool=new StlVector<Renderer3D::RenderInfo *>;
  _initState=ArSystem::AR_UP;
  }
}

void
Renderer3D::_destroy(void) // Static, Protected
{
if(_initState==ArSystem::AR_UP)
  {
  _initState=ArSystem::AR_GOING_DOWN;
  delete Renderer3D::_renderers;
  Renderer3D::_renderers=(StlVector<ArPtr<Renderer3D> > *)0;
  while(!Renderer3D::_renderInfoPool->empty())
    {
    delete Renderer3D::_renderInfoPool->back();
    Renderer3D::_renderInfoPool->pop_back();
    }
  delete Renderer3D::_renderInfoPool;
  Renderer3D::_renderInfoPool=(StlVector<Renderer3D::RenderInfo *> *)0;
  _initState=ArSystem::AR_DOWN;
  }
}

//----------------------------------------------------------------------------
// Class Renderer3D (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Renderer3D,Base3D)

bool
Renderer3D::_needsRedisplay=false;

unsigned int
Renderer3D::_lastFrameRate=0;

unsigned int
Renderer3D::_currentFrameID=0;

unsigned int
Renderer3D::_nbFrames=0;

double
Renderer3D::_frameRate=0.0;

unsigned int
Renderer3D::_lastRedraw=0;

unsigned int
Renderer3D::_redrawInterval=0;

StlVector<ArPtr<Renderer3D> > *
Renderer3D::_renderers=(StlVector<ArPtr<Renderer3D> > *)0;

StlVector<Renderer3D::RenderInfo *> *
Renderer3D::_renderInfoPool=(StlVector<Renderer3D::RenderInfo *> *)0;

//-------- Construction / Destruction ----------------------------------------

Renderer3D::Renderer3D(ArCW & arCW) // Protected
: Base3D(arCW),
  _perspective(M_PI/3.0, // fieldOfView
               1.0,      // aspectRatio
               0.1,      // nearDistance
               100.0,    // farDistance
               0.0,      // nearYOffset
               0.0),     // nearZOffset
  _fixedAspectRatio(0.0),
  _backgroundColor(0x60/255.0,0x60/255.0,0x60/255.0),
  _headLightColor(0.8,0.8,0.8),
  _bboxColor(0.9,0.9,0.9),
  _wireFrame(false),
  _headLight(true),
  _originVisible(false),
  _drawShapeBBox(false),
  _drawShapePartBBox(false),
  _stereoMode(Renderer3D::STEREO_NONE),
  _leftColor(1.0,1.0,0.0),
  _rightColor(0.0,0.0,1.0),
  _stereoMaxWidth(GL::getMaxTextureSize()>>2),
  _stereoMaxHeight(GL::getMaxTextureSize()>>1),
  _drawStereoGlance(false),
  _stereoViewDistance(1.0),
  _stereoEyeSeparation(0.1),
  _lookAt(true),
  _scenes(),
  _opaqueInfos(),
  _alphaTestInfos(),
  _transparentInfos(),
  _lights(),
  _osd(),
  _nbSeen(0),
  _xSeen(0.0),
  _ySeen(0.0),
  _zSeen(0.0),
  _renderArea((LLW::AreaID)0), // initialized in sub-classes
  _stereoBuffer((LLW::GLBufID)0),
  _closeAction(Renderer3D::CLOSE_DESTROY),
  _modifiers(Renderer3D::MDF_NONE),
  _lastMouseX(0),
  _lastMouseY(0),
  _closeCB(thisRef()),
  _keyboardCB(thisRef()),
  _mouseButtonCB(thisRef()),
  _mouseMotionCB(thisRef()),
  _timeOutCB(thisRef()),
  _viewParametersCB(thisRef()),
  _renderingSetupCB(thisRef()),
  _appearanceSetupCB(thisRef()),
  _addOSDCB(thisRef()),
  _removeOSDCB(thisRef()),
  _aboutToRenderCB(thisRef())
{
Renderer3D::_renderers->push_back(thisPtr());
_scenes[1]=Scene3D::_globalScene;
_scenes[1]->_addRenderer(thisPtr());
}

Renderer3D::~Renderer3D(void) // Protected
{
unsigned int i;
while(!_osd.empty())
  {
  removeOSD(_osd.back());
  }
for(i=3;i--;)
  {
  if(_scenes[i])
    {
    _scenes[i]->_removeRenderer(thisPtr());
    }
  }
for(i=Renderer3D::_renderers->size();i--;)
  {
  if((*Renderer3D::_renderers)[i].c_ptr()==this)
    {
    StlVectorFastErase(*Renderer3D::_renderers,i);
    break;
    }
  }
if(_stereoBuffer)
  {
  LLW::destroyGLBuffer(_stereoBuffer);
  _stereoBuffer=(LLW::GLBufID)0;
  }
}

//-------- View parameters ---------------------------------------------------

const Util3D::Perspective &
Renderer3D::getPerspective(void) const
{
return(_perspective);
}

void
Renderer3D::setFieldOfView(double fieldOfView)
{
if((fieldOfView>0.0)&&(fieldOfView<M_PI))
  {
  _perspective.setFieldOfView(fieldOfView);
  _postRedisplay();
  _onViewParameters();
  }
}

double // field of view
Renderer3D::getFieldOfView(void) const
{
return(_perspective.getFieldOfView());
}

double // current aspect ratio
Renderer3D::getAspectRatio(void) const
{
return(_perspective.getAspectRatio());
}

void
Renderer3D::setFixedAspectRatio(double aspectRatio)
{
if(aspectRatio>=0.0)
  {
  _fixedAspectRatio=aspectRatio;
  if(_fixedAspectRatio)
    {
    _perspective.setAspectRatio(_fixedAspectRatio);
    }
  else
    {
    unsigned int width,height;
    LLW::getAreaSize(_renderArea,width,height);
    if(_stereoMode==Renderer3D::STEREO_SPLIT)
      {
      width>>=1;
      }
    _perspective.setAspectRatio(((double)width)/((double)height));
    }
  _postRedisplay();
  _onViewParameters();
  }
}

double // fixed aspect ratio (whatever width/height) or 0.0 if not fixed
Renderer3D::getFixedAspectRatio(void) const
{
return(_fixedAspectRatio);
}

void
Renderer3D::setNearDistance(double nearDistance)
{
if(nearDistance>0.0)
  {
  if(nearDistance>=_perspective.getFarDistance()) // avoid crossing planes !
    {
    double factor=_perspective.getFarDistance()/_perspective.getNearDistance();
    _perspective.setFarDistance(nearDistance*factor);
    }
  _perspective.setNearDistance(nearDistance);
  _postRedisplay();
  _onViewParameters();
  }
}

double // near clipping plane distance
Renderer3D::getNearDistance(void) const
{
return(_perspective.getNearDistance());
}

void
Renderer3D::setFarDistance(double farDistance)
{
if(farDistance>0.0)
  {
  if(farDistance<=_perspective.getNearDistance()) // avoid crossing planes !
    {
    double factor=_perspective.getNearDistance()/_perspective.getFarDistance();
    _perspective.setNearDistance(farDistance*factor);
    }
  _perspective.setFarDistance(farDistance);
  _postRedisplay();
  _onViewParameters();
  }
}

double // far clipping plane distance
Renderer3D::getFarDistance(void) const
{
return(_perspective.getFarDistance());
}

void
Renderer3D::setNearYOffset(double nearYOffset)
{
_perspective.setNearYOffset(nearYOffset);
_postRedisplay();
_onViewParameters();
}

double // near clipping plane horizontal offset
Renderer3D::getNearYOffset(void) const
{
return(_perspective.getNearYOffset());
}

void
Renderer3D::setNearZOffset(double nearZOffset)
{
_perspective.setNearZOffset(nearZOffset);
_postRedisplay();
_onViewParameters();
}

double // near clipping plane vertical offset
Renderer3D::getNearZOffset(void) const
{
return(_perspective.getNearZOffset());
}

double // left clipping plane slope
Renderer3D::getLeftSlope(void) const
{
return(_perspective.getLeftSlope());
}

double // right clipping plane slope
Renderer3D::getRightSlope(void) const
{
return(_perspective.getRightSlope());
}

double // up clipping plane slope
Renderer3D::getUpSlope(void) const
{
return(_perspective.getUpSlope());
}

double // down clipping plane slope
Renderer3D::getDownSlope(void) const
{
return(_perspective.getDownSlope());
}

void
Renderer3D::setWireFrame(bool wireFrameOn)
{
_wireFrame=wireFrameOn;
_postRedisplay();
_onViewParameters();
}

bool // wire frame is on
Renderer3D::getWireFrame(void) const
{
return(_wireFrame);
}

void
Renderer3D::setHeadLight(bool headLightOn)
{
_headLight=headLightOn;
_postRedisplay();
_onViewParameters();
}

bool // head light is on
Renderer3D::getHeadLight(void) const
{
return(_headLight);
}

void
Renderer3D::setOriginVisible(bool originVisible)
{
_originVisible=originVisible;
_postRedisplay();
_onViewParameters();
}

bool // origine is visible
Renderer3D::getOriginVisible(void) const
{
return(_originVisible);
}

bool // something is visible
Renderer3D::getCenterOfView(double & xOut,
                            double & yOut,
                            double & zOut) const
{
if(_nbSeen)
  {
  xOut=_xSeen/_nbSeen;
  yOut=_ySeen/_nbSeen;
  zOut=_zSeen/_nbSeen;
  return(true);
  }
else
  {
  xOut=0.0;
  yOut=0.0;
  zOut=0.0;
  return(false);
  }
}

void
Renderer3D::setShapeBoundingBox(bool shapeBoundingBoxOn)
{
_drawShapeBBox=shapeBoundingBoxOn;
_postRedisplay();
_onViewParameters();
}

bool // shapes bounding boxes are drawn
Renderer3D::getShapeBoundingBox(void) const
{
return(_drawShapeBBox);
}

void
Renderer3D::setShapePartBoundingBox(bool shapePartBoundingBoxOn)
{
_drawShapePartBBox=shapePartBoundingBoxOn;
_postRedisplay();
_onViewParameters();
}

bool // shape parts bounding boxes are drawn
Renderer3D::getShapePartBoundingBox(void) const
{
return(_drawShapePartBBox);
}

void
Renderer3D::setStereoMode(Renderer3D::StereoMode stereoMode)
{
_stereoMode=stereoMode;
Renderer3D::_afterRenderAreaSetup();
_onViewParameters();
}

Renderer3D::StereoMode // current stereo mode
Renderer3D::getStereoMode(void) const
{
return(_stereoMode);
}

void
Renderer3D::setStereoLeftColor(double red,
                               double green,
                               double blue)
{
_leftColor.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_leftColor.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_leftColor.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onViewParameters();
}

void
Renderer3D::getStereoLeftColor(double & redOut,
                               double & greenOut,
                               double & blueOut) const
{
redOut=_leftColor.x;
greenOut=_leftColor.y;
blueOut=_leftColor.z;
}

void
Renderer3D::setStereoRightColor(double red,
                                double green,
                                double blue)
{
_rightColor.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_rightColor.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_rightColor.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onViewParameters();
}

void
Renderer3D::getStereoRightColor(double & redOut,
                                double & greenOut,
                                double & blueOut) const
{
redOut=_rightColor.x;
greenOut=_rightColor.y;
blueOut=_rightColor.z;
}

void
Renderer3D::setColorStereoMaxResolution(unsigned int maxWidth,
                                        unsigned int maxHeight)
{
if((maxWidth>0)&&(maxWidth<=GL::getMaxTextureSize()))
  {
  _stereoMaxWidth=maxWidth;
  }
if((maxHeight>0)&&(maxHeight<=GL::getMaxTextureSize()))
  {
  _stereoMaxHeight=maxHeight;
  }
Renderer3D::_afterRenderAreaSetup();
_onViewParameters();
}

void
Renderer3D::getColorStereoMaxResolution(unsigned int & maxWidthOut,
                                        unsigned int & maxHeightOut) const
{
maxWidthOut=_stereoMaxWidth;
maxHeightOut=_stereoMaxHeight;
}

void
Renderer3D::setStereoViewDistance(double viewDistance)
{
if(viewDistance>0.0)
  {
  _stereoViewDistance=viewDistance;
  _postRedisplay();
  _onViewParameters();
  }
}

double // view distance for stereo
Renderer3D::getStereoViewDistance(void) const
{
return(_stereoViewDistance);
}

void
Renderer3D::setStereoEyeSeparation(double eyeSeparation)
{
if(eyeSeparation>0.0)
  {
  _stereoEyeSeparation=eyeSeparation;
  _postRedisplay();
  _onViewParameters();
  }
}

double // eye separation for stereo
Renderer3D::getStereoEyeSeparation(void) const
{
return(_stereoEyeSeparation);
}

void
Renderer3D::setStereoGlance(bool stereoGlanceOn)
{
_drawStereoGlance=stereoGlanceOn;
_postRedisplay();
_onViewParameters();
}

bool // stereo glance is drawn
Renderer3D::getStereoGlance(void) const
{
return(_drawStereoGlance);
}

//-------- On Screen Display -------------------------------------------------

void
Renderer3D::addOSD(ArRef<OSD> osd)
{
if(!hasOSD(osd))
  {
  osd->_addRenderer(thisPtr());
  _osd.push_back(osd);
  if(!_addOSDCB.empty())
    {
    Renderer3D::AddOSDEvent event;
    event.osd=osd;
    _addOSDCB.fire(event);
    }
  _postRedisplay();
  }
}

void
Renderer3D::removeOSD(ArRef<OSD> osd)
{
for(unsigned int i=_osd.size();i--;)
  {
  if(_osd[i]==osd)
    {
    StlVectorFastErase(_osd,i);
    if(!_removeOSDCB.empty())
      {
      Renderer3D::RemoveOSDEvent event;
      event.osd=osd;
      _removeOSDCB.fire(event);
      }
    _postRedisplay();
    break;
    }
  }
osd->_removeRenderer(thisPtr());
}

bool // osd is contained
Renderer3D::hasOSD(ArConstRef<OSD> osd) const
{
for(unsigned int i=_osd.size();i--;)
  {
  if(_osd[i]==osd)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of OSD contained
Renderer3D::getNbOSD(void) const
{
return(_osd.size());
}

ArConstRef<OSD> // index'th OSD contained
Renderer3D::getOSD(unsigned int index) const
{
return(_osd[index]);
}

ArRef<OSD> // index'th OSD contained
Renderer3D::accessOSD(unsigned int index)
{
return(_osd[index]);
}

//-------- Rendering ---------------------------------------------------------

void
Renderer3D::selectScene(ArRef<Scene3D> scene)
{
_scenes[1]->_removeRenderer(thisPtr());
_scenes[1]=(scene ? scene : Scene3D::_globalScene);
_scenes[1]->_addRenderer(thisPtr());
_postRedisplay();
_onRenderingSetup();
}

ArConstRef<Scene3D> // main scene
Renderer3D::getScene(void) const
{
return(_scenes[1]);
}

ArRef<Scene3D> // main scene
Renderer3D::accessScene(void)
{
return(_scenes[1]);
}

void
Renderer3D::setBackScene(ArRef<Scene3D> scene)
{
if(_scenes[2])
  {
  _scenes[2]->_removeRenderer(thisPtr());
  }
_scenes[2]=scene;
if(_scenes[2])
  {
  _scenes[2]->_addRenderer(thisPtr());
  }
_postRedisplay();
_onRenderingSetup();
}

ArConstRef<Scene3D> // back scene
Renderer3D::getBackScene(void) const
{
return(_scenes[2]);
}

ArRef<Scene3D> // back scene
Renderer3D::accessBackScene(void)
{
return(_scenes[2]);
}

void
Renderer3D::setFrontScene(ArRef<Scene3D> scene)
{
if(_scenes[0])
  {
  _scenes[0]->_removeRenderer(thisPtr());
  }
_scenes[0]=scene;
if(_scenes[0])
  {
  _scenes[0]->_addRenderer(thisPtr());
  }
_postRedisplay();
_onRenderingSetup();
}

ArConstRef<Scene3D> // front scene
Renderer3D::getFrontScene(void) const
{
return(_scenes[0]);
}

ArRef<Scene3D> // front scene
Renderer3D::accessFrontScene(void)
{
return(_scenes[0]);
}

double // frames / second
Renderer3D::getFrameRate(void) // Static
{
return(Renderer3D::_frameRate);
}

void
Renderer3D::setFrameRateLimit(double limit) // Static
{
if(limit>=1e3) // max: 1 frame per millisecond
  {
  Renderer3D::_redrawInterval=0;
  }
else if(limit<=1.0/60.0) // min: 1 frame per minute
  {
  Renderer3D::_redrawInterval=60000;
  }
else
  {
  Renderer3D::_redrawInterval=static_cast<unsigned int>(rint(1000.0/limit));
  }
}

double // max frames / second allowed
Renderer3D::getFrameRateLimit(void) // Static
{
return(Renderer3D::_redrawInterval ? 1000.0/Renderer3D::_redrawInterval : 1e3);
}

//-------- Appearance --------------------------------------------------------

void
Renderer3D::setBackgroundColor(double red,
                               double green,
                               double blue)
{
_backgroundColor.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_backgroundColor.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_backgroundColor.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onAppearanceSetup();
}

void
Renderer3D::getBackgroundColor(double & redOut,
                               double & greenOut,
                               double & blueOut) const
{
redOut=_backgroundColor.x;
greenOut=_backgroundColor.y;
blueOut=_backgroundColor.z;
}

void
Renderer3D::setHeadLightColor(double red,
                              double green,
                              double blue)
{
_headLightColor.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_headLightColor.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_headLightColor.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onAppearanceSetup();
}

void
Renderer3D::getHeadLightColor(double & redOut,
                              double & greenOut,
                              double & blueOut) const
{
redOut=_headLightColor.x;
greenOut=_headLightColor.y;
blueOut=_headLightColor.z;
}

void
Renderer3D::setBoundingBoxColor(double red,
                                double green,
                                double blue)
{
_bboxColor.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_bboxColor.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_bboxColor.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onAppearanceSetup();
}

void
Renderer3D::getBoundingBoxColor(double & redOut,
                                double & greenOut,
                                double & blueOut) const
{
redOut=_bboxColor.x;
greenOut=_bboxColor.y;
blueOut=_bboxColor.z;
}

//-------- Rendering area ----------------------------------------------------

void
Renderer3D::getRenderSize(unsigned int & widthOut,
                          unsigned int & heightOut) const
{
if(_renderArea)
  {
  LLW::getAreaSize(_renderArea,widthOut,heightOut);
  }
else
  {
  widthOut=heightOut=0;
  }
}

void
Renderer3D::getScreenSize(unsigned int & widthOut,
                          unsigned int & heightOut) // Static
{
LLW::getScreenSize(widthOut,heightOut);
}

//-------- Snapshot ----------------------------------------------------------

bool // success
Renderer3D::snapshotToFile(const StlString & fileName) const
{
if(_renderArea)
  {
  unsigned int w,h,c;
  ArRef<MemoryBlock> data=MemoryBlock::NEW();
  return(LLW::dumpArea(_renderArea,w,h,c,data)&&
         LLW::saveImage(fileName,w,h,c,data));
  }
else
  {
  return(false);
  }
}

bool // success (dataOut filled from bottom to top)
Renderer3D::takeSnapshot(unsigned int & widthOut,
                         unsigned int & heightOut,
                         unsigned int & componentsOut,
                         ArRef<MemoryBlock> dataOut) const
{
return(_renderArea&&
       LLW::dumpArea(_renderArea,widthOut,heightOut,componentsOut,dataOut));
}

bool // success
Renderer3D::saveSnapshot(const StlString & fileName,
                         unsigned int width,
                         unsigned int height,
                         unsigned int components,
                         ArConstRef<MemoryBlock> data) // Static
{
return(LLW::saveImage(fileName,width,height,components,data));
}

//-------- Interaction -------------------------------------------------------

void
Renderer3D::moveSystemCursor(int dx,
                             int dy) // Static
{
LLW::moveCursor(dx,dy);
}

void
Renderer3D::computeCursorDirection(int xMouse,
                                   int yMouse,
                                   double & localYawOut,
                                   double & localPitchOut) const
{
unsigned int w,h;
LLW::getAreaSize(_renderArea,w,h);
if(_stereoMode==Renderer3D::STEREO_SPLIT)
  {
  w>>=1;
  if(xMouse>(int)w)
    {
    xMouse-=w;
    }
  }
double y=_perspective.getLeftSlope()+
         xMouse*(_perspective.getRightSlope()-_perspective.getLeftSlope())/w;
double z=_perspective.getUpSlope()+
         yMouse*(_perspective.getDownSlope()-_perspective.getUpSlope())/h;
double dist;
cart3DToSpherical(dist,localYawOut,localPitchOut,1.0,y,z);
}

void
Renderer3D::onMotionInteraction(ArRef<AbstractInteractor> /* source */,
                                ArRef<Base3D> newLocation,
                                bool /* begin */,
                                bool /* end */)
{
/*
if(begin)
  {
  cerr << getAlias() << ": motion begin "
       << "from " << source->getAlias() << endl;
  }
if(end)
  {
  cerr << getAlias() << ": motion end "
       << "from " << source->getAlias() << endl;
  }
*/
setLocation(newLocation);
}

void
Renderer3D::setCloseAction(Renderer3D::CloseAction closeAction)
{
_closeAction=closeAction;
}


Renderer3D::CloseAction // action performed on close event
Renderer3D::getCloseAction(void) const
{
return(_closeAction);
}

Renderer3D::Modifier // key/button modifiers
Renderer3D::getModifiers(void) const
{
return(_modifiers);
}

bool // [Shift] key is pressed
Renderer3D::shiftPressed(void) const
{
return(_modifiers&Renderer3D::MDF_SHIFT);
}

bool // [Control] key is pressed
Renderer3D::controlPressed(void) const
{
return(_modifiers&Renderer3D::MDF_CONTROL);
}

bool // [CapsLock] key is locked
Renderer3D::capsLockPressed(void) const
{
return(_modifiers&Renderer3D::MDF_CAPSLOCK);
}

bool // [NumLock] key is locked
Renderer3D::numLockPressed(void) const
{
return(_modifiers&Renderer3D::MDF_NUMLOCK);
}

bool // left button is pressed
Renderer3D::leftPressed(void) const
{
return(_modifiers&Renderer3D::MDF_LEFT);
}

bool // middle button is pressed
Renderer3D::middlePressed(void) const
{
return(_modifiers&Renderer3D::MDF_MIDDLE);
}

bool // right button is pressed
Renderer3D::rightPressed(void) const
{
return(_modifiers&Renderer3D::MDF_RIGHT);
}

void
Renderer3D::getLastMousePosition(int & xOut,
                                 int & yOut) const
{
xOut=_lastMouseX;
yOut=_lastMouseY;
}

//-------- Event management --------------------------------------------------

void
Renderer3D::simulateClose(void)
{
_manageClose();
}

void
Renderer3D::simulateKeyboard(Renderer3D::Modifier modifiers,
                             const StlString & key,
                             bool pressed,
                             int xMouse,
                             int yMouse)
{
_modifiers=modifiers;
_lastMouseX=xMouse;
_lastMouseY=yMouse;
if(!_keyboardCB.empty())
  {
  Renderer3D::KeyboardEvent event;
  event.key=key;
  event.pressed=pressed;
  event.xMouse=xMouse;
  event.yMouse=yMouse;
  _keyboardCB.fire(event);
  }
}

void
Renderer3D::simulateMouseButton(Renderer3D::Modifier modifiers,
                                int button,
                                bool pressed,
                                int xMouse,
                                int yMouse)
{
_modifiers=modifiers;
_lastMouseX=xMouse;
_lastMouseY=yMouse;
if(!_mouseButtonCB.empty())
  {
  Renderer3D::MouseButtonEvent event;
  event.button=button;
  event.pressed=pressed;
  event.xMouse=xMouse;
  event.yMouse=yMouse;
  _mouseButtonCB.fire(event);
  }
}

void
Renderer3D::simulateMouseMotion(Renderer3D::Modifier modifiers,
                                int xMouse,
                                int yMouse)
{
_modifiers=modifiers;
_lastMouseX=xMouse;
_lastMouseY=yMouse;
if(!_mouseMotionCB.empty())
  {
  Renderer3D::MouseMotionEvent event;
  event.xMouse=xMouse;
  event.yMouse=yMouse;
  _mouseMotionCB.fire(event);
  }
}

unsigned int // expiration time
Renderer3D::postTimeOut(unsigned int delay)
{
return(LLW::postTimeOut(_renderArea,delay));
}

void
Renderer3D::cancelTimeOut(unsigned int timeOut)
{
LLW::cancelTimeOut(_renderArea,timeOut);
}

//-------- Serialization operations ------------------------------------------

void
Renderer3D::extractDependencies(SerializationDependencies & dependenciesOut,
                                bool recursive) const
{
Base3D::extractDependencies(dependenciesOut,recursive);
if(_scenes[0])
  {
  dependenciesOut.addReference(_scenes[0]);
  }
if(_scenes[1])
  {
  dependenciesOut.addReference(_scenes[1]);
  }
if(_scenes[2])
  {
  dependenciesOut.addReference(_scenes[2]);
  }
unsigned int i,nb=_osd.size();
for(i=0;i<nb;i++)
  {
  if(recursive)
    {
    _osd[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_osd[i]);
  }
}

void
Renderer3D::serialize(SerializationDependencies & dependenciesOut,
                      ArRef<AbstractOStream> output) const
{
Base3D::serialize(dependenciesOut,output);
unsigned int width,height;
getRenderSize(width,height);
output->writeUInt(width);
output->writeUInt(height);
output->writeDouble(_perspective.getFieldOfView());
output->writeDouble(_perspective.getNearDistance());
output->writeDouble(_perspective.getFarDistance());
output->writeDouble(_perspective.getNearYOffset());
output->writeDouble(_perspective.getNearZOffset());
output->writeDouble(_fixedAspectRatio);
output->writeDouble(_backgroundColor.x);
output->writeDouble(_backgroundColor.y);
output->writeDouble(_backgroundColor.z);
output->writeDouble(_headLightColor.x);
output->writeDouble(_headLightColor.y);
output->writeDouble(_headLightColor.z);
output->writeDouble(_bboxColor.x);
output->writeDouble(_bboxColor.y);
output->writeDouble(_bboxColor.z);
output->writeUInt(_stereoMode);
output->writeDouble(_leftColor.x);
output->writeDouble(_leftColor.y);
output->writeDouble(_leftColor.z);
output->writeDouble(_rightColor.x);
output->writeDouble(_rightColor.y);
output->writeDouble(_rightColor.z);
output->writeUInt(_stereoMaxWidth);
output->writeUInt(_stereoMaxHeight);
output->writeDouble(_stereoViewDistance);
output->writeDouble(_stereoEyeSeparation);
unsigned int flags=(_wireFrame         ? 1<<0 : 0)+
                   (_headLight         ? 1<<1 : 0)+
                   (_originVisible     ? 1<<2 : 0)+
                   (_drawShapeBBox     ? 1<<3 : 0)+
                   (_drawShapePartBBox ? 1<<4 : 0)+
                   (_drawStereoGlance  ? 1<<5 : 0)+
                   (_scenes[0]         ? 1<<6 : 0)+
                   (_scenes[1]         ? 1<<7 : 0)+
                   (_scenes[2]         ? 1<<8 : 0);
output->writeUInt(flags);
if(_scenes[0])
  {
  output->writeUInt(dependenciesOut.addReference(_scenes[0]));
  }
if(_scenes[1])
  {
  output->writeUInt(dependenciesOut.addReference(_scenes[1]));
  }
if(_scenes[2])
  {
  output->writeUInt(dependenciesOut.addReference(_scenes[2]));
  }
unsigned int nb=_osd.size();
output->writeUInt(_osd.size());
for(unsigned int i=0;i<nb;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_osd[i]));
  }
}

bool // success
Renderer3D::unserialize(const SerializationDependencies & dependencies,
                        ArRef<AbstractIStream> input)
{
double fieldOfView,nearDistance,farDistance,
       nearYOffset,nearZOffset,fixedAspectRatio;
Util3D::Dbl3 backgroundColor;
Util3D::Dbl3 headLightColor;
Util3D::Dbl3 bboxColor;
unsigned int stereoMode;
Util3D::Dbl3 leftColor;
Util3D::Dbl3 rightColor;
unsigned int stereoMaxWidth;
unsigned int stereoMaxHeight;
double stereoViewDistance,stereoEyeSeparation;
unsigned int flags;
unsigned int scenes[3];
unsigned int nbOsd;
StlVector<unsigned int> osd;
unsigned int width,height;
if(!Base3D::unserialize(dependencies,input)||
   !input->readUInt(width)||
   !input->readUInt(height)||
   !input->readDouble(fieldOfView)||
   !input->readDouble(nearDistance)||
   !input->readDouble(farDistance)||
   !input->readDouble(nearYOffset)||
   !input->readDouble(nearZOffset)||
   !input->readDouble(fixedAspectRatio)||
   !input->readDouble(backgroundColor.x)||
   !input->readDouble(backgroundColor.y)||
   !input->readDouble(backgroundColor.z)||
   !input->readDouble(headLightColor.x)||
   !input->readDouble(headLightColor.y)||
   !input->readDouble(headLightColor.z)||
   !input->readDouble(bboxColor.x)||
   !input->readDouble(bboxColor.y)||
   !input->readDouble(bboxColor.z)||
   !input->readUInt(stereoMode)||
   !input->readDouble(leftColor.x)||
   !input->readDouble(leftColor.y)||
   !input->readDouble(leftColor.z)||
   !input->readDouble(rightColor.x)||
   !input->readDouble(rightColor.y)||
   !input->readDouble(rightColor.z)||
   !input->readUInt(stereoMaxWidth)||
   !input->readUInt(stereoMaxHeight)||
   !input->readDouble(stereoViewDistance)||
   !input->readDouble(stereoEyeSeparation)||
   !input->readUInt(flags)||
   ((flags&(1<<6))&&!input->readUInt(scenes[0]))||
   ((flags&(1<<7))&&!input->readUInt(scenes[1]))||
   ((flags&(1<<8))&&!input->readUInt(scenes[2]))||
   !input->readUInt(nbOsd))
  {
  return(false);
  }
unsigned int i;
for(i=0;i<nbOsd;i++)
  {
  osd.push_back(0);
  if(!input->readUInt(osd.back()))
    {
    return(false);
    }
  }
_perspective.setFieldOfView(fieldOfView);
_perspective.setNearDistance(nearDistance);
_perspective.setFarDistance(farDistance);
_perspective.setNearYOffset(nearYOffset);
_perspective.setNearZOffset(nearZOffset);
_backgroundColor=backgroundColor;
_headLightColor=headLightColor;
_bboxColor=bboxColor;
_leftColor=leftColor;
_rightColor=rightColor;
_stereoViewDistance=stereoViewDistance;
_stereoEyeSeparation=stereoEyeSeparation;
_wireFrame=flags&(1<<0);
_headLight=flags&(1<<1);
_originVisible=flags&(1<<2);
_drawShapeBBox=flags&(1<<3);
_drawShapePartBBox=flags&(1<<4);
_drawStereoGlance=flags&(1<<5);
setFrontScene(flags&(1<<6)
              ? ar_down_cast<Scene3D>(dependencies.findReference(scenes[0]))
              : Scene3D::nullRef());
selectScene(flags&(1<<7)
            ? ar_down_cast<Scene3D>(dependencies.findReference(scenes[1]))
            : Scene3D::nullRef());
setBackScene(flags&(1<<8)
             ? ar_down_cast<Scene3D>(dependencies.findReference(scenes[2]))
             : Scene3D::nullRef());
setRenderSize(width,height);
setColorStereoMaxResolution(stereoMaxWidth,stereoMaxHeight);
setStereoMode((Renderer3D::StereoMode)stereoMode);
setFixedAspectRatio(fixedAspectRatio);
for(i=0;i<nbOsd;i++)
  {
  addOSD(ar_down_cast<OSD>
            (dependencies.findReference(osd[i])));
  }
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Renderer3D::_afterRenderAreaSetup(void) // Protected
{
if(_stereoBuffer)
  {
  LLW::destroyGLBuffer(_stereoBuffer);
  _stereoBuffer=(LLW::GLBufID)0;
  }
if(_fixedAspectRatio)
  {
  _perspective.setAspectRatio(_fixedAspectRatio);
  }
else
  {
  unsigned int width,height;
  LLW::getAreaSize(_renderArea,width,height);
  if(_stereoMode==Renderer3D::STEREO_SPLIT)
    {
    width>>=1;
    }
  _perspective.setAspectRatio(((double)width)/((double)height));
  }
_postRedisplay();
}

void
Renderer3D::_checkEvents(bool blocking) // Protected, Static
{
const LLW::Event * evt=LLW::nextEvent(blocking);
if(evt)
  {
  for(unsigned int i=Renderer3D::_renderers->size();i--;)
    {
    if((*Renderer3D::_renderers)[i]->_processEvent(evt))
      {
      break;
      }
    }
  }
}

bool // event processed
Renderer3D::_processEvent(const LLW::Event * evt) // Protected
{
if(evt->area==_renderArea)
  {
  switch(evt->type)
    {
    case LLW::EVT_TIMEOUT:
      {
      // nb: modifiers, x and y are not relevant with LLW::EVT_TIMEOUT
      if(!_timeOutCB.empty())
        {
        Renderer3D::TimeOutEvent event;
        event.timeOut=evt->time;
        _timeOutCB.fire(event);
        }
      break;
      }
    default:
      break;
    }
  return(true);
  }
else
  {
  return(false);
  }
}

void
Renderer3D::_manageClose(void) // Protected
{
// nb: modifiers, x and y are not relevant with LLW::EVT_CLOSE
ArRef<Renderer3D> flag(this);
if(!_closeCB.empty())
  {
  Renderer3D::CloseEvent event;
  _closeCB.fire(event);
  }
if(flag)
  {
  switch(_closeAction)
    {
    case Renderer3D::CLOSE_NOP:
      {
      break; // Nothing to be done
      }
    case Renderer3D::CLOSE_DESTROY:
      {
      thisRef().destroy(); // UGLY !
      break;
      }
    case Renderer3D::CLOSE_LEAVE:
      {
      ArSystem::leaveSimulation();
      break;
      }
    }
  }
}

void
Renderer3D::_onMotion(void) // Protected
{
_postRedisplay();
Base3D::_onMotion();
}

void
Renderer3D::_rebuildShapesIfNeeded(void) // Protected
{
if(_scenes[0])
  {
  _scenes[0]->_rebuildShapesIfNeeded();
  }
_scenes[1]->_rebuildShapesIfNeeded();
if(_scenes[2])
  {
  _scenes[2]->_rebuildShapesIfNeeded();
  }
}

void
Renderer3D::_postRedisplay(void) // Protected
{
Renderer3D::_needsRedisplay=true;
}

bool // redisplay was needed
Renderer3D::_redisplay(void) // Protected, Static
{
unsigned int t=ArSystem::getTime();
if(t>=Renderer3D::_lastRedraw+Renderer3D::_redrawInterval)
  {
  Renderer3D::_lastRedraw=t;
  if(Renderer3D::_needsRedisplay)
    {
    ++Renderer3D::_currentFrameID;
    for(unsigned int i=Renderer3D::_renderers->size();i--;)
      {
      Renderer3D * renderer=(*Renderer3D::_renderers)[i].c_ptr();
      if(!renderer->_aboutToRenderCB.empty())
        {
        Renderer3D::AboutToRenderEvent event;
        renderer->_aboutToRenderCB.fire(event);
        }
      renderer->_render();
      }
    ++Renderer3D::_nbFrames;
    unsigned int dt=t-Renderer3D::_lastFrameRate;
    if(dt>2000)
      {
      Renderer3D::_frameRate=(Renderer3D::_nbFrames*1000.0)/((double)dt);
      Renderer3D::_lastFrameRate=t;
      Renderer3D::_nbFrames=0;
      }
    Renderer3D::_needsRedisplay=false;
    return(true);
    }
  else
    {
    return(false);
    }
  }
else
  {
  return(false);
  }
}

void
Renderer3D::_updateLookAt(double eyeX,
                          double eyeY,
                          double eyeZ) // Protected
{
const double * rot=_base.matrix;
double * look=_lookAt.matrix;
look[Util3D::R00]=-rot[Util3D::R01]; // -y --> x
look[Util3D::R01]=-rot[Util3D::R11];
look[Util3D::R02]=-rot[Util3D::R21];
look[Util3D::R10]=rot[Util3D::R02];  // z --> y
look[Util3D::R11]=rot[Util3D::R12];
look[Util3D::R12]=rot[Util3D::R22];
look[Util3D::R20]=-rot[Util3D::R00]; // -x --> z
look[Util3D::R21]=-rot[Util3D::R10];
look[Util3D::R22]=-rot[Util3D::R20];
look[Util3D::TX]=-eyeX*look[Util3D::R00]
                 -eyeY*look[Util3D::R01]
                 -eyeZ*look[Util3D::R02];
look[Util3D::TY]=-eyeX*look[Util3D::R10]
                 -eyeY*look[Util3D::R11]
                 -eyeZ*look[Util3D::R12];
look[Util3D::TZ]=-eyeX*look[Util3D::R20]
                 -eyeY*look[Util3D::R21]
                 -eyeZ*look[Util3D::R22];
}

void
Renderer3D::_preRender(void) // Protected
{
_nbSeen=0;
_xSeen=0.0;
_ySeen=0.0;
_zSeen=0.0;
if((_stereoMode==Renderer3D::STEREO_COLOR)&&!_stereoBuffer)
  {
  unsigned int areaWidth,areaHeight;
  LLW::getAreaSize(_renderArea,areaWidth,areaHeight);
  _stereoBuffer=LLW::createGLBuffer((areaWidth<=_stereoMaxWidth
                                     ? areaWidth : _stereoMaxWidth)<<1,
                                    (areaHeight<=_stereoMaxHeight
                                     ? areaHeight : _stereoMaxHeight));
  }
LLW::beginGLRendering(_renderArea);
}

void
Renderer3D::_postRender(void) // Protected
{
LLW::endGLRendering();
}

void
Renderer3D::_render(void) // Protected
{
_preRender();
_renderNoSwap();
_postRender();
}

void
Renderer3D::_renderNoSwap(void) // Protected
{
unsigned int areaWidth,areaHeight;
LLW::getAreaSize(_renderArea,areaWidth,areaHeight);
unsigned int viewportWidth,viewportHeight;
if(_stereoBuffer)
  {
  LLW::getAreaSize(_stereoBuffer,viewportWidth,viewportHeight);
  LLW::beginGLRendering(_stereoBuffer);
  }
else
  {
  viewportWidth=areaWidth;
  viewportHeight=areaHeight;
  }
glClearColor(_backgroundColor.x,_backgroundColor.y,_backgroundColor.z,1.0);
glClear(GL_COLOR_BUFFER_BIT);
const double * mtx=_base.matrix;
if(_wireFrame)
  {
  glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  }
bool drawOsd=true;
if(_stereoBuffer||(_stereoMode==Renderer3D::STEREO_SPLIT))
  {
  double eyeOffset=0.5*_stereoEyeSeparation;
  Util3D::Dbl3 leftEye(mtx[Util3D::TX]+eyeOffset*mtx[Util3D::R01],
                       mtx[Util3D::TY]+eyeOffset*mtx[Util3D::R11],
                       mtx[Util3D::TZ]+eyeOffset*mtx[Util3D::R21]);
  Util3D::Dbl3 rightEye(mtx[Util3D::TX]-eyeOffset*mtx[Util3D::R01],
                        mtx[Util3D::TY]-eyeOffset*mtx[Util3D::R11],
                        mtx[Util3D::TZ]-eyeOffset*mtx[Util3D::R21]);
  double yNearLeft=_perspective.getNearDistance()*
                   _perspective.getLeftSlope();
  double yNearRight=_perspective.getNearDistance()*
                    _perspective.getRightSlope();
  double dYNear=_perspective.getNearDistance()/_stereoViewDistance;
  if(dYNear>1.0)
    {
    dYNear=eyeOffset;
    }
  else
    {
    dYNear*=eyeOffset;
    }
  double zNearUp=_perspective.getNearDistance()*
                 _perspective.getUpSlope();
  double zNearDown=_perspective.getNearDistance()*
                   _perspective.getDownSlope();
  unsigned int midwidth=viewportWidth>>1;
  for(unsigned int i=3;i--;)
    {
    const Scene3D * scene=_scenes[i].c_ptr();
    if(scene)
      {
      _beginScene(scene);
      glViewport(0,0,midwidth,viewportHeight);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glFrustum(-yNearLeft+dYNear, // our Y is used as -X
                -yNearRight+dYNear,
                zNearDown,
                zNearUp,
                _perspective.getNearDistance(),
                _perspective.getFarDistance());
      glMatrixMode(GL_MODELVIEW);
      _updateLookAt(leftEye.x,leftEye.y,leftEye.z);
      glLoadMatrixd(_lookAt.matrix);
      _renderScene(scene);
      glViewport(viewportWidth-midwidth,0,midwidth,viewportHeight);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glFrustum(-yNearLeft-dYNear, // our Y is used as -X
                -yNearRight-dYNear,
                zNearDown,
                zNearUp,
                _perspective.getNearDistance(),
                _perspective.getFarDistance());
      glMatrixMode(GL_MODELVIEW);
      _updateLookAt(rightEye.x,rightEye.y,rightEye.z);
      glLoadMatrixd(_lookAt.matrix);
      _renderScene(scene);
      _endScene(scene);
      }
    }
  if(!_stereoBuffer)
    {
    drawOsd=false;
    if(!_osd.empty())
      {
      sort(_osd.begin(),_osd.end(),OSD::DepthPrecedence());
      glClear(GL_DEPTH_BUFFER_BIT);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0.0,midwidth,0.0,viewportHeight,-1.0,1.0);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      unsigned int i;
      glViewport(0,0,midwidth,viewportHeight);
      for(i=_osd.size();i--;)
        {
        _osd[i]->_render(midwidth,viewportHeight);
        }
      glViewport(viewportWidth-midwidth,0,midwidth,viewportHeight);
      for(i=_osd.size();i--;)
        {
        _osd[i]->_render(midwidth,viewportHeight);
        }
      }
    }
  }
else
  {
  glViewport(0,0,areaWidth,areaHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-_perspective.getNearDistance()*_perspective.getLeftSlope(),
            -_perspective.getNearDistance()*_perspective.getRightSlope(),
            _perspective.getNearDistance()*_perspective.getDownSlope(),
            _perspective.getNearDistance()*_perspective.getUpSlope(),
            _perspective.getNearDistance(),
            _perspective.getFarDistance());
  glMatrixMode(GL_MODELVIEW);
  _updateLookAt(mtx[Util3D::TX],mtx[Util3D::TY],mtx[Util3D::TZ]);
  glLoadMatrixd(_lookAt.matrix);
  for(unsigned int i=3;i--;)
    {
    const Scene3D * scene=_scenes[i].c_ptr();
    if(scene)
      {
      _beginScene(scene);
      _renderScene(scene);
      _endScene(scene);
      }
    }
  }
if(_stereoBuffer)
  {
  LLW::endGLRendering();
  LLW::beginGLRendering(_renderArea);
  GL::selectTexture(LLW::getGLBufferTexture(_stereoBuffer),false,false,true);
  glViewport(0,0,areaWidth,areaHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0,1.0,0.0,1.0,-1.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  if(_wireFrame)
    {
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    }
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glShadeModel(GL_FLAT);
  glEnable(GL_TEXTURE_2D);
  glColor3f(_leftColor.x,_leftColor.y,_leftColor.z);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(0.0,0.0); glVertex3f(0.0,0.0,0.0);
  glTexCoord2f(0.5,0.0); glVertex3f(1.0,0.0,0.0);
  glTexCoord2f(0.5,1.0); glVertex3f(1.0,1.0,0.0);
  glTexCoord2f(0.0,1.0); glVertex3f(0.0,1.0,0.0);
  glEnd();
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE);
  glColor3f(_rightColor.x,_rightColor.y,_rightColor.z);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(0.5,0.0); glVertex3f(0.0,0.0,0.0);
  glTexCoord2f(1.0,0.0); glVertex3f(1.0,0.0,0.0);
  glTexCoord2f(1.0,1.0); glVertex3f(1.0,1.0,0.0);
  glTexCoord2f(0.5,1.0); glVertex3f(0.0,1.0,0.0);
  glEnd();
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);
  if(_wireFrame)
    {
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    }
  }
if(drawOsd&&!_osd.empty())
  {
  sort(_osd.begin(),_osd.end(),OSD::DepthPrecedence());
  glClear(GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0,areaWidth,0.0,areaHeight,-1.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  for(unsigned int i=_osd.size();i--;)
    {
    _osd[i]->_render(areaWidth,areaHeight);
    }
  }
if(_wireFrame)
  {
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  }
const char * err=GL::lastError();
if(err)
  {
  cerr << "!!! OpenGL error while rendering " << getAlias() << ": "
       << err << " !!!" << endl;
  }
}

void
Renderer3D::_beginScene(const Scene3D * scene) // Protected
{
glClear(GL_DEPTH_BUFFER_BIT);
if(scene->_fogDensity)
  {
  glEnable(GL_FOG);
  float fogColor[4]={scene->_fogColor.x,scene->_fogColor.y,
                     scene->_fogColor.z,1.0};
  glFogfv(GL_FOG_COLOR,fogColor);
  glHint(GL_FOG_HINT,GL_DONT_CARE);
  switch(scene->_fogDegree)
    {
    case 0:
      {
      glFogi(GL_FOG_MODE,GL_LINEAR);
      double start,end;
      if(scene->_fogDensity>0.0)
        {
        start=_perspective.getNearDistance();
        end=start+(_perspective.getFarDistance()-start)/scene->_fogDensity;
        }
      else
        {
        start=_perspective.getFarDistance();
        end=start+(start-_perspective.getNearDistance())/scene->_fogDensity;
        }
      glFogf(GL_FOG_START,start);
      glFogf(GL_FOG_END,end);
      break;
      }
    case 1:
      {
      glFogi(GL_FOG_MODE,GL_EXP);
      glFogf(GL_FOG_DENSITY,
             scene->_fogDensity/_perspective.getFarDistance());
      break;
      }
    case 2:
      {
      glFogi(GL_FOG_MODE,GL_EXP2);
      glFogf(GL_FOG_DENSITY,
             scene->_fogDensity/_perspective.getFarDistance());
      break;
      }
    }
  }
if(_headLight)
  {
  _lights.push_back((const Light3D *)0); // use one light
  }
unsigned int i;
for(i=scene->getNbLights();i--;)
  {
  const Light3D * l=scene->getLight(i).c_ptr();
  if(l->getSwitch()&&(_lights.size()<GL::getNbLights()))
    {
    _lights.push_back(l);
    }
  }
for(i=scene->getNbParticleSystems();i--;)
  {
  const ParticleSystem * p=scene->getParticleSystem(i).c_ptr();
  if(p->_bbox.valid)
    {
    unsigned int outFront=8;
    unsigned int outBack=8;
    unsigned int outLeft=8;
    unsigned int outRight=8;
    unsigned int outUp=8;
    unsigned int outDown=8;
    for(unsigned int i=8;i--;)
      {
      double x=(i&1 ? p->_bbox.xmax : p->_bbox.xmin);
      double y=(i&2 ? p->_bbox.ymax : p->_bbox.ymin);
      double z=(i&4 ? p->_bbox.zmax : p->_bbox.zmin);
      _base.globalToLocalPosition(x,y,z);
      Util3D::VertexLocation loc=_perspective.locateVertex(x,y,z);
      if(loc==Util3D::MIDDLE_LOC)
        {
        break; // inside perspective
        }
      if(loc&Util3D::XPOS_LOC)
        {
        --outFront;
        }
      else if(loc&Util3D::XNEG_LOC)
        {
        --outBack;
        }
      if(loc&Util3D::YPOS_LOC)
        {
        --outLeft;
        }
      if(loc&Util3D::YNEG_LOC)
        {
        --outRight;
        }
      if(loc&Util3D::ZPOS_LOC)
        {
        --outUp;
        }
      if(loc&Util3D::ZNEG_LOC)
        {
        --outDown;
        }
      }
    if(outFront&&outBack&&
       outLeft&&outRight&&
       outUp&&outDown) // not on the same side
      {
      Util3D::Dbl3 c(p->_bbox.centerX(),p->_bbox.centerY(),p->_bbox.centerZ());
      ++_nbSeen;
      _xSeen+=c.x;
      _ySeen+=c.y;
      _zSeen+=c.z;
      Renderer3D::RenderInfo * info=Renderer3D::_makeParticleSystemInfo(p);
      if(p->_needsBlending())
        {
        c.x-=_base.matrix[Util3D::TX];
        c.y-=_base.matrix[Util3D::TY];
        c.z-=_base.matrix[Util3D::TZ];
        info->squareDistance=c.x*c.x+c.y*c.y+c.z*c.z;
        _insertTransparentInfo(info);
        }
      else if(p->_needsAlphaTest())
        {
        _alphaTestInfos.push_back(info);
        }
      else
        {
        _opaqueInfos.push_back(info);
        }
      }
    }
  }
scene->_askVisibleParts(this);
}

void
Renderer3D::_renderScene(const Scene3D * scene) // Protected
{
if(scene->_backTexture)
  {
  double x,y,z;
  _base.getPosition(x,y,z);
  glDepthMask(GL_FALSE);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  double r1=0.9*_perspective.getFarDistance();
  double r2=r1*0.5*sqrt(2.0);
  double r3=0.5*r2;
  glEnable(GL_TEXTURE_2D);
  GL::selectTexture(scene->_backTexture->getTextureID(),true,true,false);
  if(!scene->_backTransform.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
      glLoadMatrixd(scene->_backTransform.matrix);
    glMatrixMode(GL_MODELVIEW);
    }
  glBegin(GL_TRIANGLE_FAN);
    glTexCoord2f(0.5,0.5); glVertex3f(x,y,z+r3);
    glTexCoord2f(0.0,0.0); glVertex3f(x-r2,y-r2,z);
    glTexCoord2f(0.0,0.5); glVertex3f(x-r1,y,z);
    glTexCoord2f(0.0,1.0); glVertex3f(x-r2,y+r2,z);
    glTexCoord2f(0.5,1.0); glVertex3f(x,y+r1,z);
    glTexCoord2f(1.0,1.0); glVertex3f(x+r2,y+r2,z);
    glTexCoord2f(1.0,0.5); glVertex3f(x+r1,y,z);
    glTexCoord2f(1.0,0.0); glVertex3f(x+r2,y-r2,z);
    glTexCoord2f(0.5,0.0); glVertex3f(x,y-r1,z);
    glTexCoord2f(0.0,0.0); glVertex3f(x-r2,y-r2,z);
  glEnd(); // GL_TRIANGLE_FAN
  glBegin(GL_TRIANGLE_FAN);
    glTexCoord2f(0.5,0.5); glVertex3f(x,y,z-r3);
    glTexCoord2f(0.0,0.0); glVertex3f(x-r2,y-r2,z);
    glTexCoord2f(0.5,0.0); glVertex3f(x,y-r1,z);
    glTexCoord2f(1.0,0.0); glVertex3f(x+r2,y-r2,z);
    glTexCoord2f(1.0,0.5); glVertex3f(x+r1,y,z);
    glTexCoord2f(1.0,1.0); glVertex3f(x+r2,y+r2,z);
    glTexCoord2f(0.5,1.0); glVertex3f(x,y+r1,z);
    glTexCoord2f(0.0,1.0); glVertex3f(x-r2,y+r2,z);
    glTexCoord2f(0.0,0.5); glVertex3f(x-r1,y,z);
    glTexCoord2f(0.0,0.0); glVertex3f(x-r2,y-r2,z);
  glEnd(); // GL_TRIANGLE_FAN
  if(!scene->_backTransform.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
      glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    }
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  }
if(_originVisible)
  {
  if(scene->_fogDensity)
    {
    glDisable(GL_FOG);
    }
  glDisable(GL_LIGHTING);
  // glDisable(GL_DEPTH_TEST);
  glLineWidth(2.0);
  glBegin(GL_LINES);
    glColor3f(1.0,0.0,0.0);
    glVertex3f(0.0,0.0,0.0); glVertex3f(1.0,0.0,0.0);
    glColor3f(0.0,1.0,0.0);
    glVertex3f(0.0,0.0,0.0); glVertex3f(0.0,1.0,0.0);
    glColor3f(0.0,0.0,1.0);
    glVertex3f(0.0,0.0,0.0); glVertex3f(0.0,0.0,1.0);
  glEnd(); // GL_LINES
  glLineWidth(1.0);
  // glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  if(scene->_fogDensity)
    {
    glEnable(GL_FOG);
    }
  }
if(_drawStereoGlance)
  {
  _drawGlance(scene);
  }
unsigned int i;
for(i=_lights.size();i--;)
  {
  glEnable((GLenum)(GL_LIGHT0+i));
  if(_lights[i])
    {
    _lights[i]->_render(i);
    }
  else
    {
    Light3D::_renderHeadLight(i,_base,
                              _headLightColor.x,
                              _headLightColor.y,
                              _headLightColor.z);
    }
  }
_drawVisibleInfos(_opaqueInfos,scene);
if(!_alphaTestInfos.empty())
  {
  glEnable(GL_ALPHA_TEST);
  _drawVisibleInfos(_alphaTestInfos,scene);
  glDisable(GL_ALPHA_TEST);
  }
if(!_transparentInfos.empty())
  {
  glDepthMask(GL_FALSE);
  glEnable(GL_BLEND);
  _drawVisibleInfos(_transparentInfos,scene);
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  }
for(i=_lights.size();i--;)
  {
  glDisable((GLenum)(GL_LIGHT0+i));
  }
}

void
Renderer3D::_endScene(const Scene3D * scene) // Protected
{
while(!_opaqueInfos.empty())
  {
  Renderer3D::_renderInfoPool->push_back(_opaqueInfos.back());
  _opaqueInfos.pop_back();
  }
while(!_alphaTestInfos.empty())
  {
  Renderer3D::_renderInfoPool->push_back(_alphaTestInfos.back());
  _alphaTestInfos.pop_back();
  }
while(!_transparentInfos.empty())
  {
  Renderer3D::_renderInfoPool->push_back(_transparentInfos.back());
  _transparentInfos.pop_back();
  }
_lights.clear();
if(scene->_fogDensity)
  {
  glDisable(GL_FOG);
  }
}

void
Renderer3D::_drawVisibleInfos(StlVector<Renderer3D::RenderInfo *> & info,
                              const Scene3D * scene) // Protected
{
const Object3D * lastObject=(const Object3D *)0;
for(unsigned i=info.size();i--;)
  {
  const ParticleSystem * particleSystem=info[i]->particleSystem;
  if(particleSystem)
    {
    if(lastObject)
      {
      if(!_wireFrame&&lastObject->_wireFrame)
        {
        glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
        }
      glPopMatrix();
      }
    lastObject=(const Object3D *)0;
    particleSystem->_render(this);
    if(_drawShapeBBox||_drawShapePartBBox)
      {
      _drawBoundingBox(particleSystem->_bbox,scene);
      }
    }
  else
    {
    const Util3D::Base & base=info[i]->location;
    const Object3D * object=info[i]->object;
    const Shape3D * shape=info[i]->shape;
    const ShapePart3D * part=info[i]->part;
    double xc=part->_bbox.centerX();
    double yc=part->_bbox.centerY();
    double zc=part->_bbox.centerZ();
    part->_flatTransform.apply(xc,yc,zc,xc,yc,zc);
    shape->getTransformation().apply(xc,yc,zc,xc,yc,zc);
    base.localToGlobalPosition(xc,yc,zc);
    ++_nbSeen;
    _xSeen+=xc;
    _ySeen+=yc;
    _zSeen+=zc;
    if(object!=lastObject)
      {
      if(!_wireFrame)
        {
        if(object->_wireFrame)
          {
          if(!(lastObject&&lastObject->_wireFrame))
            {
            glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
            }
          }
        else
          {
          if(lastObject&&lastObject->_wireFrame)
            {
            glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
            }
          }
        }
      if(lastObject)
        {
        glPopMatrix();
        }
      glPushMatrix();
      glMultMatrixd(base.matrix);
      if(!shape->getTransformation().isIdentity(false))
        {
        glMultMatrixd(shape->getTransformation().matrix);
        }
      if(_drawShapeBBox||object->_drawShapeBBox)
        {
        _drawBoundingBox(shape->getBoundingBox(),scene);
        }
      }
    if(!part->_flatTransform.isIdentity(false))
      {
      glPushMatrix();
      glMultMatrixd(part->_flatTransform.matrix);
      }
    if(info[i]->selected)
      {
      glDepthFunc(GL_LEQUAL);
      if(object->_selectedMaterial.getLighting())
        {
        glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,
                     object->_selectedMaterial.getAmbientColor());
        glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,
                     object->_selectedMaterial.getDiffuseColor());
        glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,
                     object->_selectedMaterial.getEmissiveColor());
        glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,
                     object->_selectedMaterial.getSpecularColor());
        glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,
                    128.0*object->_selectedMaterial.getShininess());
        if(part->_renderer)
          {
          part->_renderer->_renderWithoutAppearance(part,object,this);
          }
        else
          {
          part->_renderWithoutAppearance();
          }
        }
      else
        {
        glDisable(GL_LIGHTING);
        glColor4fv(object->_selectedMaterial.getEmissiveColor());
        if(part->_renderer)
          {
          part->_renderer->_renderWithoutAppearance(part,object,this);
          }
        else
          {
          part->_renderWithoutAppearance();
          }
        glEnable(GL_LIGHTING);
        }
      glDepthFunc(GL_LESS);
      }
    else
      {
      if(part->_renderer)
        {
        part->_renderer->_render(part,object,this);
        }
      else
        {
        part->_render();
        }
      }
    if(_drawShapePartBBox||object->_drawShapePartBBox)
      {
      _drawBoundingBox(part->_bbox,scene);
      }
    if(!part->_flatTransform.isIdentity(false))
      {
      glPopMatrix();
      }
    lastObject=object;
    }
  }
if(lastObject)
  {
  if(!_wireFrame&&lastObject->_wireFrame)
    {
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    }
  glPopMatrix();
  }
}

Renderer3D::RenderInfo * // (re)allocated render info
Renderer3D::_makeParticleSystemInfo(const ParticleSystem * particleSystem)
{                                                         // Static, Protected
Renderer3D::RenderInfo * info;
if(Renderer3D::_renderInfoPool->empty())
  {
  info=new Renderer3D::RenderInfo;
  }
else
  {
  info=Renderer3D::_renderInfoPool->back();
  Renderer3D::_renderInfoPool->pop_back();
  }
info->particleSystem=particleSystem;
info->object=(const Object3D *)0;
return(info);
}

Renderer3D::RenderInfo * // (re)allocated render info
Renderer3D::_makeObject3DInfo(const Util3D::Base & location,
                              const Object3D * object,
                              const Shape3D * shape,
                              const ShapePart3D * part,
                              bool selected)
{                                                         // Static, Protected
Renderer3D::RenderInfo * info;
if(Renderer3D::_renderInfoPool->empty())
  {
  info=new Renderer3D::RenderInfo;
  }
else
  {
  info=Renderer3D::_renderInfoPool->back();
  Renderer3D::_renderInfoPool->pop_back();
  }
info->particleSystem=(const ParticleSystem *)0;
info->location=location;
info->object=object;
info->shape=shape;
info->part=part;
info->selected=selected;
return(info);
}

void
Renderer3D::_insertTransparentInfo(Renderer3D::RenderInfo * info)
{                                                                 // Protected
unsigned int min=0;
unsigned int max=_transparentInfos.size();
while(min!=max)
  {
  unsigned int mid=(min+max)>>1;
  if(info->squareDistance<_transparentInfos[mid]->squareDistance)
    {
    max=mid;
    }
  else
    {
    min=mid+1;
    }
  }
_transparentInfos.insert(_transparentInfos.begin()+min,info);
}

void
Renderer3D::_drawBoundingBox(const Util3D::BoundingBox & bbox,
                             const Scene3D * scene) // Protected
{
if(scene->_fogDensity)
  {
  glDisable(GL_FOG);
  }
glDisable(GL_LIGHTING);
glLineWidth(2.0);
glBegin(GL_LINES);
glColor3f(_bboxColor.x,_bboxColor.y,_bboxColor.z);
glVertex3f(bbox.xmin,bbox.ymin,bbox.zmin);
glVertex3f(bbox.xmax,bbox.ymin,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymin,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymax,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymin,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymin,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymax,bbox.zmax);
glVertex3f(bbox.xmin,bbox.ymax,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymax,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymin,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymax,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymax,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymin,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymin,bbox.zmax);
glVertex3f(bbox.xmin,bbox.ymin,bbox.zmax);
glVertex3f(bbox.xmin,bbox.ymax,bbox.zmax);
glVertex3f(bbox.xmax,bbox.ymax,bbox.zmin);
glVertex3f(bbox.xmax,bbox.ymin,bbox.zmin);
glVertex3f(bbox.xmax,bbox.ymax,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymax,bbox.zmin);
glVertex3f(bbox.xmax,bbox.ymin,bbox.zmin);
glVertex3f(bbox.xmax,bbox.ymin,bbox.zmax);
glVertex3f(bbox.xmin,bbox.ymax,bbox.zmin);
glVertex3f(bbox.xmin,bbox.ymax,bbox.zmax);
glEnd();
glLineWidth(1.0);
glEnable(GL_LIGHTING);
if(scene->_fogDensity)
  {
  glEnable(GL_FOG);
  }
}

void
Renderer3D::_drawGlance(const Scene3D * scene) // Protected
{
const double * mtx=_base.matrix;
double eyeOffset=0.5*_stereoEyeSeparation;
double zNearUp=_perspective.getNearDistance()*_perspective.getUpSlope();
double zNearDown=_perspective.getNearDistance()*_perspective.getDownSlope();
double yNearLeft=_perspective.getNearDistance()*_perspective.getLeftSlope();
double yNearRight=_perspective.getNearDistance()*_perspective.getRightSlope();
yNearLeft*=2.0;
yNearRight*=2.0;
Util3D::Dbl3 p0(mtx[Util3D::TX]+
                mtx[Util3D::R00]*_stereoViewDistance+
                mtx[Util3D::R01]*eyeOffset+
                mtx[Util3D::R02]*zNearDown,
                mtx[Util3D::TY]+
                mtx[Util3D::R10]*_stereoViewDistance+
                mtx[Util3D::R11]*eyeOffset+
                mtx[Util3D::R12]*zNearDown,
                mtx[Util3D::TZ]+
                mtx[Util3D::R20]*_stereoViewDistance+
                mtx[Util3D::R21]*eyeOffset+
                mtx[Util3D::R22]*zNearDown);
Util3D::Dbl3 p1(p0.x-mtx[Util3D::R01]*eyeOffset*2.0,
                p0.y-mtx[Util3D::R11]*eyeOffset*2.0,
                p0.z-mtx[Util3D::R21]*eyeOffset*2.0);
Util3D::Dbl3 p2(p1.x+mtx[Util3D::R02]*(zNearUp-zNearDown),
                p1.y+mtx[Util3D::R12]*(zNearUp-zNearDown),
                p1.z+mtx[Util3D::R22]*(zNearUp-zNearDown));
Util3D::Dbl3 p3(p0.x+mtx[Util3D::R02]*(zNearUp-zNearDown),
                p0.y+mtx[Util3D::R12]*(zNearUp-zNearDown),
                p0.z+mtx[Util3D::R22]*(zNearUp-zNearDown));
Util3D::Dbl3 p4(p0.x-mtx[Util3D::R00]*_stereoViewDistance+
                mtx[Util3D::R01]*yNearLeft,
                p0.y-mtx[Util3D::R10]*_stereoViewDistance+
                mtx[Util3D::R11]*yNearLeft,
                p0.z-mtx[Util3D::R20]*_stereoViewDistance+
                mtx[Util3D::R21]*yNearLeft);
Util3D::Dbl3 p5(p1.x-mtx[Util3D::R00]*_stereoViewDistance+
                mtx[Util3D::R01]*yNearRight,
                p1.y-mtx[Util3D::R10]*_stereoViewDistance+
                mtx[Util3D::R11]*yNearRight,
                p1.z-mtx[Util3D::R20]*_stereoViewDistance+
                mtx[Util3D::R21]*yNearRight);
Util3D::Dbl3 p6(p2.x-mtx[Util3D::R00]*_stereoViewDistance+
                mtx[Util3D::R01]*yNearRight,
                p2.y-mtx[Util3D::R10]*_stereoViewDistance+
                mtx[Util3D::R11]*yNearRight,
                p2.z-mtx[Util3D::R20]*_stereoViewDistance+
                mtx[Util3D::R21]*yNearRight);
Util3D::Dbl3 p7(p3.x-mtx[Util3D::R00]*_stereoViewDistance+
                mtx[Util3D::R01]*yNearLeft,
                p3.y-mtx[Util3D::R10]*_stereoViewDistance+
                mtx[Util3D::R11]*yNearLeft,
                p3.z-mtx[Util3D::R20]*_stereoViewDistance+
                mtx[Util3D::R21]*yNearLeft);
if(scene->_fogDensity)
  {
  glDisable(GL_FOG);
  }
glDisable(GL_LIGHTING);
glLineWidth(2.0);
glBegin(GL_LINES);
glColor3f(_bboxColor.x,_bboxColor.y,_bboxColor.z);
glVertex3f(p0.x,p0.y,p0.z); glVertex3f(p1.x,p1.y,p1.z);
glVertex3f(p3.x,p3.y,p3.z); glVertex3f(p2.x,p2.y,p2.z);
glVertex3f(p0.x,p0.y,p0.z); glVertex3f(p3.x,p3.y,p3.z);
glVertex3f(p1.x,p1.y,p1.z); glVertex3f(p2.x,p2.y,p2.z);
glVertex3f(p0.x,p0.y,p0.z); glVertex3f(p2.x,p2.y,p2.z);
glVertex3f(p1.x,p1.y,p1.z); glVertex3f(p3.x,p3.y,p3.z);
glVertex3f(p0.x,p0.y,p0.z); glVertex3f(p4.x,p4.y,p4.z);
glVertex3f(p1.x,p1.y,p1.z); glVertex3f(p5.x,p5.y,p5.z);
glVertex3f(p2.x,p2.y,p2.z); glVertex3f(p6.x,p6.y,p6.z);
glVertex3f(p3.x,p3.y,p3.z); glVertex3f(p7.x,p7.y,p7.z);
glEnd(); // GL_LINES
glLineWidth(1.0);
glEnable(GL_LIGHTING);
if(scene->_fogDensity)
  {
  glEnable(GL_FOG);
  }
}

void
Renderer3D::_onViewParameters(void)
{
if(!_viewParametersCB.empty())
  {
  Renderer3D::ViewParametersEvent event;
  _viewParametersCB.fire(event);
  }
}

void
Renderer3D::_onRenderingSetup(void)
{
if(!_renderingSetupCB.empty())
  {
  Renderer3D::RenderingSetupEvent event;
  _renderingSetupCB.fire(event);
  }
}

void
Renderer3D::_onAppearanceSetup(void)
{
if(!_appearanceSetupCB.empty())
  {
  Renderer3D::AppearanceSetupEvent event;
  _appearanceSetupCB.fire(event);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
