// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/scene3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/scene3D.h"

#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/light3D.h"
#include "AReVi/Lib3D/particleSystem.h"
#include "AReVi/Lib3D/sound3D.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/texture.h"
#include "AReVi/Lib3D/transform2D.h"

#include "AReVi/Shapes/shape3D.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//-------- Initialization ----------------------------------------------------

static ArSystem::InitState _initState=ArSystem::AR_DOWN;

void
Scene3D::_init(void) // Static, Protected
{
if(_initState==ArSystem::AR_DOWN)
  {
  _initState=ArSystem::AR_GOING_UP;
  Scene3D::_globalScene=Scene3D::NEW();
  Scene3D::_globalScene->setAlias("Global_Scene3D");
  _initState=ArSystem::AR_UP;
  }
}

void
Scene3D::_destroy(void) // Static, Protected
{
if(_initState==ArSystem::AR_UP)
  {
  _initState=ArSystem::AR_GOING_DOWN;
  Scene3D::_globalScene.destroy(); // UGLY !
  _initState=ArSystem::AR_DOWN;
  }
}

//----------------------------------------------------------------------------
// Class Scene3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Scene3D,ArObject)

ArRef<Scene3D>
Scene3D::_globalScene;

//-------- Construction / Destruction ----------------------------------------

Scene3D::Scene3D(ArCW & arCW) // Protected
: ArObject(arCW),
  _objects(),
  _lights(),
  _particleSystems(),
  _sounds(),
  _renderers(),
  _listeners(),
  _fogColor(),
  _fogDegree(1),
  _fogDensity(0.0),
  _backTexture(),
  _backTransform(true),
  _addObjectCB(thisRef()),
  _removeObjectCB(thisRef()),
  _addLightCB(thisRef()),
  _removeLightCB(thisRef()),
  _addParticleSystemCB(thisRef()),
  _removeParticleSystemCB(thisRef()),
  _addSoundCB(thisRef()),
  _removeSoundCB(thisRef()),
  _sceneSetupCB(thisRef()),
  _selectObjectCB(thisRef()),
  _unselectObjectCB(thisRef())
{
setFogColor(0.8,0.8,0.8);
}

Scene3D::~Scene3D(void) // Protected
{
setBackground(Texture::nullRef());
if(Scene3D::_globalScene.c_ptr()==this)
  {
  while(!_objects.empty())
    {
    _objects.back()->thisRef().destroy(); // UGLY !
    }
  while(!_lights.empty())
    {
    _lights.back()->thisRef().destroy(); // UGLY !
    }
  while(!_particleSystems.empty())
    {
    _particleSystems.back()->thisRef().destroy(); // UGLY !
    }
  while(!_sounds.empty())
    {
    _sounds.back()->thisRef().destroy(); // UGLY !
    }
  while(!_renderers.empty())
    {
    _renderers.back()->thisRef().destroy(); // UGLY !
    }
  while(!_listeners.empty())
    {
    _listeners.back()->thisRef().destroy(); // UGLY !
    }
  if(_initState==ArSystem::AR_UP)
    {
    cerr << "Explicit destruction of the global Scene3D is forbidden !"
         << endl;
    exit(-1);
    }
  }
else
  {
  while(!_objects.empty())
    {
    _removeObject(_objects.back()->thisRef());
    }
  while(!_lights.empty())
    {
    _removeLight(_lights.back()->thisRef());
    }
  while(!_particleSystems.empty())
    {
    _removeParticleSystem(_particleSystems.back()->thisRef());
    }
  while(!_sounds.empty())
    {
    _removeSound(_sounds.back()->thisRef());
    }
  while(!_renderers.empty())
    {
    ArRef<Renderer3D> r=_renderers.back();
    r->selectScene(Scene3D::nullRef());
    r->setBackScene(Scene3D::nullRef());
    r->setFrontScene(Scene3D::nullRef());
    }
  while(!_listeners.empty())
    {
    _listeners.back()->selectScene(Scene3D::nullRef());
    }
  }
}

//-------- Global scene ------------------------------------------------------

ArConstRef<Scene3D> // global scene
Scene3D::getGlobalScene(void) // Static
{
return(Scene3D::_globalScene);
}

ArRef<Scene3D> // global scene
Scene3D::accessGlobalScene(void) // Static
{
return(Scene3D::_globalScene);
}

//-------- Object3D management -----------------------------------------------

void
Scene3D::addObject(ArRef<Object3D> obj)
{
if(!hasObject(obj))
  {
  obj->_addScene(thisRef());
  _objects.push_back(obj->thisPtr());
  _postRedisplay();
  if(!_addObjectCB.empty())
    {
    Scene3D::AddObjectEvent event;
    event.object=obj;
    _addObjectCB.fire(event);
    }
  }
}

void
Scene3D::removeObject(ArRef<Object3D> obj)
{
if(Scene3D::_globalScene.c_ptr()!=this)
  {
  _removeObject(obj);
  if(!_removeObjectCB.empty())
    {
    Scene3D::RemoveObjectEvent event;
    event.object=obj;
    _removeObjectCB.fire(event);
    }
  }
}

bool // object is contained
Scene3D::hasObject(ArConstRef<Object3D> obj) const
{
for(unsigned int i=_objects.size();i--;)
  {
  if(_objects[i]==obj)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of objects contained
Scene3D::getNbObjects(void) const
{
return(_objects.size());
}

ArConstRef<Object3D> // index'th object contained
Scene3D::getObject(unsigned int index) const
{
return(_objects[index]->thisRef());
}

ArRef<Object3D> // index'th object contained
Scene3D::accessObject(unsigned int index)
{
return(_objects[index]->thisRef());
}

//-------- Light3D management ------------------------------------------------

void
Scene3D::addLight(ArRef<Light3D> light)
{
if(!hasLight(light))
  {
  light->_addScene(thisRef());
  _lights.push_back(light->thisPtr());
  _postRedisplay();
  if(!_addLightCB.empty())
    {
    Scene3D::AddLightEvent event;
    event.light=light;
    _addLightCB.fire(event);
    }
  }
}

void
Scene3D::removeLight(ArRef<Light3D> light)
{
if(Scene3D::_globalScene.c_ptr()!=this)
  {
  _removeLight(light);
  if(!_removeLightCB.empty())
    {
    Scene3D::RemoveLightEvent event;
    event.light=light;
    _removeLightCB.fire(event);
    }
  }
}

bool // light is contained
Scene3D::hasLight(ArConstRef<Light3D> light) const
{
for(unsigned int i=_lights.size();i--;)
  {
  if(_lights[i]==light)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of lights contained
Scene3D::getNbLights(void) const
{
return(_lights.size());
}

ArConstRef<Light3D> // index'th light contained
Scene3D::getLight(unsigned int index) const
{
return(_lights[index]->thisRef());
}

ArRef<Light3D> // index'th light contained
Scene3D::accessLight(unsigned int index)
{
return(_lights[index]->thisRef());
}

//-------- ParticleSystem management -----------------------------------------

void
Scene3D::addParticleSystem(ArRef<ParticleSystem> particleSystem)
{
if(!hasParticleSystem(particleSystem))
  {
  particleSystem->_addScene(thisRef());
  _particleSystems.push_back(particleSystem->thisPtr());
  _postRedisplay();
  if(!_addParticleSystemCB.empty())
    {
    Scene3D::AddParticleSystemEvent event;
    event.particleSystem=particleSystem;
    _addParticleSystemCB.fire(event);
    }
  }
}

void
Scene3D::removeParticleSystem(ArRef<ParticleSystem> particleSystem)
{
if(Scene3D::_globalScene.c_ptr()!=this)
  {
  _removeParticleSystem(particleSystem);
  if(!_removeParticleSystemCB.empty())
    {
    Scene3D::RemoveParticleSystemEvent event;
    event.particleSystem=particleSystem;
    _removeParticleSystemCB.fire(event);
    }
  }
}

bool // particle system is contained
Scene3D::hasParticleSystem(ArConstRef<ParticleSystem> particleSystem) const
{
for(unsigned int i=_particleSystems.size();i--;)
  {
  if(_particleSystems[i]==particleSystem)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of particle systems contained
Scene3D::getNbParticleSystems(void) const
{
return(_particleSystems.size());
}

ArConstRef<ParticleSystem> // index'th particle system contained
Scene3D::getParticleSystem(unsigned int index) const
{
return(_particleSystems[index]->thisRef());
}

ArRef<ParticleSystem> // index'th particle system contained
Scene3D::accessParticleSystem(unsigned int index)
{
return(_particleSystems[index]->thisRef());
}

//-------- Sound3D management ------------------------------------------------

void
Scene3D::addSound(ArRef<Sound3D> sound)
{
if(!hasSound(sound))
  {
  sound->_addScene(thisRef());
  _sounds.push_back(sound->thisPtr());
  if(!_addSoundCB.empty())
    {
    Scene3D::AddSoundEvent event;
    event.sound=sound;
    _addSoundCB.fire(event);
    }
  }
}

void
Scene3D::removeSound(ArRef<Sound3D> sound)
{
if(Scene3D::_globalScene.c_ptr()!=this)
  {
  _removeSound(sound);
  if(!_removeSoundCB.empty())
    {
    Scene3D::RemoveSoundEvent event;
    event.sound=sound;
    _removeSoundCB.fire(event);
    }
  }
}

bool // sound is contained
Scene3D::hasSound(ArConstRef<Sound3D> sound) const
{
for(unsigned int i=_sounds.size();i--;)
  {
  if(_sounds[i]==sound)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of sounds contained
Scene3D::getNbSounds(void) const
{
return(_sounds.size());
}

ArConstRef<Sound3D> // index'th sound contained
Scene3D::getSound(unsigned int index) const
{
return(_sounds[index]->thisRef());
}

ArRef<Sound3D> // index'th sound contained
Scene3D::accessSound(unsigned int index)
{
return(_sounds[index]->thisRef());
}

//-------- Selection management ----------------------------------------------

void
Scene3D::selectObject(ArRef<Object3D> obj)
{
if(!isSelected(obj)&&hasObject(obj))
  {
  _selection.push_back(obj->thisPtr());
  _postRedisplay();
  if(!_selectObjectCB.empty())
    {
    Scene3D::SelectObjectEvent event;
    event.object=obj;
    _selectObjectCB.fire(event);
    }
  }
}

void
Scene3D::unselectObject(ArRef<Object3D> obj)
{
for(unsigned int i=_selection.size();i--;)
  {
  if(_selection[i]==obj)
    {
    StlVectorFastErase(_selection,i);
    if(!_unselectObjectCB.empty())
      {
      Scene3D::UnselectObjectEvent event;
      event.object=obj;
      _unselectObjectCB.fire(event);
      }
    _postRedisplay();
    break;
    }
  }
}

bool // object is selected
Scene3D::isSelected(ArConstRef<Object3D> obj) const
{
for(unsigned int i=_selection.size();i--;)
  {
  if(_selection[i]==obj)
    {
    return(true);
    }
  }
return(false);
}

unsigned int // number of objects selected
Scene3D::getSelectionSize(void) const
{
return(_selection.size());
}

ArConstRef<Object3D> // index'th object selected
Scene3D::getSelectedObject(unsigned int index) const
{
return(_selection[index]->thisRef());
}

ArRef<Object3D> // index'th object selected
Scene3D::accessSelectedObject(unsigned int index)
{
return(_selection[index]->thisRef());
}


//-------- Fog management ----------------------------------------------------

void
Scene3D::setFogColor(double red,
                     double green,
                     double blue)
{
_fogColor.x=(red<0.0 ? 0.0 : (red>1.0 ? 1.0 : red));
_fogColor.y=(green<0.0 ? 0.0 : (green>1.0 ? 1.0 : green));
_fogColor.z=(blue<0.0 ? 0.0 : (blue>1.0 ? 1.0 : blue));
_postRedisplay();
_onSceneSetup();
}

void
Scene3D::getFogColor(double & redOut,
                     double & greenOut,
                     double & blueOut) const
{
redOut=_fogColor.x;
greenOut=_fogColor.y;
blueOut=_fogColor.z;
}

void
Scene3D::setFog(unsigned int degree,
                double density)
{
_fogDegree=(degree>2 ? 2 : degree);
_fogDensity=(_fogDegree&&(density<0.0) ? 0.0 : density);
_postRedisplay();
_onSceneSetup();
}

void
Scene3D::getFog(unsigned int & degreeOut,
                double & densityOut) const
{
degreeOut=_fogDegree;
densityOut=_fogDensity;
}

//-------- Background --------------------------------------------------------

void
Scene3D::setBackground(ArRef<Texture> texture)
{
if((texture!=_backTexture)&&!(texture&&texture->fail()))
  {
  ArRef<Texture> old=_backTexture;
  if(old)
    {
    old->_removeScene(this);
    }
  _backTexture=texture;
  if(_backTexture)
    {
    _backTexture->_addScene(this);
    }
  _postRedisplay();
  _onSceneSetup();
  }
}

ArConstRef<Texture> // background texture
Scene3D::getBackground(void) const
{
return(_backTexture);
}

ArRef<Texture> // background texture
Scene3D::accessBackground(void)
{
return(_backTexture);
}

void
Scene3D::writeBackgroundTransformation(ArConstRef<Transform2D> transform)
{
_backTransform=transform->getTransformation();
_postRedisplay();
_onSceneSetup();
}

void
Scene3D::readBackgroundTransformation(ArRef<Transform2D> transformInOut) const
{
transformInOut->accessTransformation()=_backTransform;
}

//-------- Detection ---------------------------------------------------------

void
Scene3D::computeBoundingBox(ArRef<BoundingBox3D> bboxOut,
                            ArConstRef<Base3D> viewPoint) const
{
Util3D::BoundingBox & bbox=bboxOut->accessBoundingBox();
bbox.valid=false;
ArRef<BoundingBox3D> objBox=BoundingBox3D::NEW();
for(unsigned int i=_objects.size();i--;)
  {
  Util3D::Transform tr(_objects[i]->getBase());
  _objects[i]->computeBoundingBox(objBox,viewPoint);
  bbox.addTransformedBox(tr,objBox->getBoundingBox());
  }
}

void
Scene3D::getAll(ArConstRef<ArClass> objectClass,
                StlVector<ArRef<Object3D> > & objectsOut) const
{
objectsOut.clear();
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if(obj->getClass()->isA(objectClass))
    {
    objectsOut.push_back(obj->thisRef());
    }
  }
}

void
Scene3D::viewFirst(ArConstRef<Base3D> origin,
                   double yawVision,
                   double pitchVision,
                   double range,
                   ArConstRef<ArClass> objectClass,
                   ArRef<Object3D> & objectOut,
                   double & yawAngleOut,
                   double & pitchAngleOut,
                   double & distanceOut) const
{
objectOut.clear();
yawAngleOut=0.0;
pitchAngleOut=0.0;
distanceOut=-1.0;
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
yawVision=(yawVision>0.0 ? yawVision/2.0 : M_PI);
pitchVision=(pitchVision>0.0 ? pitchVision/2.0 : M_PI_2);
range=(range>0.0 ? range*range : DBL_MAX);
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if((obj!=origin.c_ptr())&&obj->getClass()->isA(objectClass))
    {
    double x,y,z;
    obj->getPosition(x,y,z);
    origin->globalToLocalPosition(x,y,z);
    double d2=x*x+y*y+z*z;
    if((d2<range)&&(!objectOut||(d2<distanceOut)))
      {
      double dxy=sqrt(x*x+y*y);
      double yaw=(dxy ? atan2(y,x) : 0.0);
      if(fabs(yaw)<=yawVision)
        {
        double pitch=(d2 ? -atan2(z,dxy) : 0.0);
        if(fabs(pitch)<=pitchVision)
          {
          objectOut=obj->thisRef();
          yawAngleOut=yaw;
          pitchAngleOut=pitch;
          distanceOut=d2;
          }
        }
      }
    }
  }
if(objectOut)
  {
  distanceOut=sqrt(distanceOut);
  }
}

void
Scene3D::view(ArConstRef<Base3D> origin,
              double yawVision,
              double pitchVision,
              double range,
              ArConstRef<ArClass> objectClass,
              StlVector<ArRef<Object3D> > & objectsOut,
              StlVector<double> & yawAnglesOut,
              StlVector<double> & pitchAnglesOut,
              StlVector<double> & distancesOut) const
{
objectsOut.clear();
yawAnglesOut.clear();
pitchAnglesOut.clear();
distancesOut.clear();
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
yawVision=(yawVision>0.0 ? yawVision/2.0 : M_PI);
pitchVision=(pitchVision>0.0 ? pitchVision/2.0 : M_PI_2);
range=(range>0.0 ? range*range : DBL_MAX);
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if((obj!=origin.c_ptr())&&obj->getClass()->isA(objectClass))
    {
    double x,y,z;
    obj->getPosition(x,y,z);
    origin->globalToLocalPosition(x,y,z);
    double d2=x*x+y*y+z*z;
    if(d2<range)
      {
      double dxy=sqrt(x*x+y*y);
      double yaw=(dxy ? atan2(y,x) : 0.0);
      if(fabs(yaw)<=yawVision)
        {
        double pitch=(d2 ? -atan2(z,dxy) : 0.0);
        if(fabs(pitch)<=pitchVision)
          {
          d2=sqrt(d2);
          unsigned int j;
          unsigned int nb=distancesOut.size();
          for(j=0;j<nb;j++)
            {
            if(d2<=distancesOut[j])
              {
              yawAnglesOut.insert(yawAnglesOut.begin()+j,yaw);
              pitchAnglesOut.insert(pitchAnglesOut.begin()+j,pitch);
              distancesOut.insert(distancesOut.begin()+j,d2);
              objectsOut.insert(objectsOut.begin()+j,obj->thisRef());
              break;
              }
            }
          if(j==nb)
            {
            yawAnglesOut.push_back(yaw);
            pitchAnglesOut.push_back(pitch);
            distancesOut.push_back(d2);
            objectsOut.push_back(obj->thisRef());
            }
          }
        }
      }
    }
  }
}

void
Scene3D::firstPicking(ArConstRef<Renderer3D> viewPoint,
                      int xMouse,
                      int yMouse,
                      unsigned int level,
                      bool findNearest,
                      ArConstRef<ArClass> objectClass,
                      ArRef<Object3D> & objectOut,
                      ArRef<ShapePart3D> & interPartOut,
                      unsigned int & triIndexOut,
                      double & distanceOut) const
{
double yaw,pitch;
viewPoint->computeCursorDirection(xMouse,yMouse,yaw,pitch);
ArRef<Base3D> ray=Base3D::NEW();
ray->setLocation(viewPoint);
ray->yaw(yaw);
ray->pitch(pitch);
firstRayIntersection(ray,level,findNearest,objectClass,
                     objectOut,interPartOut,triIndexOut,distanceOut,viewPoint);
}

void
Scene3D::firstRayIntersection(ArConstRef<Base3D> ray,
                              unsigned int level,
                              bool findNearest,
                              ArConstRef<ArClass> objectClass,
                              ArRef<Object3D> & objectOut,
                              ArRef<ShapePart3D> & interPartOut,
                              unsigned int & triIndexOut,
                              double & distanceOut,
                              ArConstRef<Base3D> viewPoint) const
{
objectOut.clear();
interPartOut.clear();
triIndexOut=UINT_MAX;
distanceOut=-1.0;
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if((obj!=ray.c_ptr())&&obj->getClass()->isA(objectClass))
    {
    ArRef<ShapePart3D> part;
    unsigned int triIndex;
    double d=obj->intersectRay(ray,level,findNearest,viewPoint,part,triIndex);
    if((d>=0.0)&&(!objectOut||(d<distanceOut)))
      {
      objectOut=obj->thisRef();
      interPartOut=part;
      triIndexOut=triIndex;
      distanceOut=d;
      }
    }
  }
}

void
Scene3D::firstSimpleRayIntersection(double sourceX,
                                    double sourceY,
                                    double sourceZ,
                                    double unitDirX,
                                    double unitDirY,
                                    double unitDirZ,
                                    unsigned int level,
                                    bool findNearest,
                                    ArConstRef<ArClass> objectClass,
                                    ArRef<Object3D> & objectOut,
                                    ArRef<ShapePart3D> & interPartOut,
                                    unsigned int & triIndexOut,
                                    double & distanceOut) const
{
objectOut.clear();
interPartOut.clear();
triIndexOut=UINT_MAX;
distanceOut=-1.0;
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if(obj->getClass()->isA(objectClass))
    {
    ArRef<ShapePart3D> part;
    unsigned int triIndex;
    double d=obj->simpleIntersectRay(sourceX,sourceY,sourceZ,
                                     unitDirX,unitDirY,unitDirZ,
                                     level,findNearest,part,triIndex);
    if((d>=0.0)&&(!objectOut||(d<distanceOut)))
      {
      objectOut=obj->thisRef();
      interPartOut=part;
      triIndexOut=triIndex;
      distanceOut=d;
      }
    }
  }
}

void
Scene3D::picking(ArConstRef<Renderer3D> viewPoint,
                 int xMouse,
                 int yMouse,
                 unsigned int level,
                 bool findNearest,
                 ArConstRef<ArClass> objectClass,
                 StlVector<ArRef<Object3D> > & objectsOut,
                 StlVector<ArRef<ShapePart3D> > & interPartsOut,
                 StlVector<unsigned int> & triIndicesOut,
                 StlVector<double> & distancesOut) const
{
double yaw,pitch;
viewPoint->computeCursorDirection(xMouse,yMouse,yaw,pitch);
ArRef<Base3D> ray=Base3D::NEW();
ray->setLocation(viewPoint);
ray->yaw(yaw);
ray->pitch(pitch);
rayIntersection(ray,level,findNearest,objectClass,
                objectsOut,interPartsOut,triIndicesOut,distancesOut,viewPoint);
}

void
Scene3D::rayIntersection(ArConstRef<Base3D> ray,
                         unsigned int level,
                         bool findNearest,
                         ArConstRef<ArClass> objectClass,
                         StlVector<ArRef<Object3D> > & objectsOut,
                         StlVector<ArRef<ShapePart3D> > & interPartsOut,
                         StlVector<unsigned int> & triIndicesOut,
                         StlVector<double> & distancesOut,
                         ArConstRef<Base3D> viewPoint) const
{
objectsOut.clear();
interPartsOut.clear();
triIndicesOut.clear();
distancesOut.clear();
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if((obj!=ray.c_ptr())&&obj->getClass()->isA(objectClass))
    {
    ArRef<ShapePart3D> part;
    unsigned int triIndex;
    double d=obj->intersectRay(ray,level,findNearest,viewPoint,part,triIndex);
    if(d>=0.0)
      {
      unsigned int j;
      unsigned int nb=distancesOut.size();
      for(j=0;j<nb;j++)
        {
        if(d<=distancesOut[j])
          {
          objectsOut.insert(objectsOut.begin()+j,obj->thisRef());
          interPartsOut.insert(interPartsOut.begin()+j,part);
          triIndicesOut.insert(triIndicesOut.begin()+j,triIndex);
          distancesOut.insert(distancesOut.begin()+j,d);
          break;
          }
        }
      if(j==nb)
        {
        objectsOut.push_back(obj->thisRef());
        interPartsOut.push_back(part);
        triIndicesOut.push_back(triIndex);
        distancesOut.push_back(d);
        }
      }
    }
  }
}

void
Scene3D::simpleRayIntersection(double sourceX,
                               double sourceY,
                               double sourceZ,
                               double unitDirX,
                               double unitDirY,
                               double unitDirZ,
                               unsigned int level,
                               bool findNearest,
                               ArConstRef<ArClass> objectClass,
                               StlVector<ArRef<Object3D> > & objectsOut,
                               StlVector<ArRef<ShapePart3D> > & interPartsOut,
                               StlVector<unsigned int> & triIndicesOut,
                               StlVector<double> & distancesOut) const
{
objectsOut.clear();
interPartsOut.clear();
triIndicesOut.clear();
distancesOut.clear();
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  Object3D * obj=_objects[i].c_ptr();
  if(obj->getClass()->isA(objectClass))
    {
    ArRef<ShapePart3D> part;
    unsigned int triIndex;
    double d=obj->simpleIntersectRay(sourceX,sourceY,sourceZ,
                                     unitDirX,unitDirY,unitDirZ,
                                     level,findNearest,part,triIndex);
    if(d>=0.0)
      {
      unsigned int j;
      unsigned int nb=distancesOut.size();
      for(j=0;j<nb;j++)
        {
        if(d<=distancesOut[j])
          {
          objectsOut.insert(objectsOut.begin()+j,obj->thisRef());
          interPartsOut.insert(interPartsOut.begin()+j,part);
          triIndicesOut.insert(triIndicesOut.begin()+j,triIndex);
          distancesOut.insert(distancesOut.begin()+j,d);
          break;
          }
        }
      if(j==nb)
        {
        objectsOut.push_back(obj->thisRef());
        interPartsOut.push_back(part);
        triIndicesOut.push_back(triIndex);
        distancesOut.push_back(d);
        }
      }
    }
  }
}

bool // collision occured
Scene3D::testCollision(ArConstRef<Object3D> object,
                       unsigned int level,
                       ArConstRef<ArClass> objectClass,
                       StlVector<ArRef<Object3D> > & objectsOut,
                       ArConstRef<Base3D> viewPoint) const
{
objectsOut.clear();
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  ArRef<Object3D> obj=_objects[i];
  if((obj!=object)&&obj->getClass()->isA(objectClass))
    {
    if(object->testCollision(obj,level,viewPoint))
      {
      objectsOut.push_back(obj);
      }
    }
  }
return(!objectsOut.empty());
}

bool // collision occured
Scene3D::intersection(ArConstRef<Object3D> object,
                      ArConstRef<ArClass> objectClass,
                      StlVector<ArRef<Object3D> > & objectsOut,
                      StlVector<StlVector<Util3D::TriInter> > & itOut,
                      ArConstRef<Base3D> viewPoint) const
{
objectsOut.clear();
itOut.clear();
if(!objectClass)
  {
  objectClass=Object3D::CLASS();
  }
for(unsigned int i=_objects.size();i--;)
  {
  ArRef<Object3D> obj=_objects[i];
  if((obj!=object)&&obj->getClass()->isA(objectClass))
    {
    itOut.push_back(StlVector<Util3D::TriInter>());
    if(object->intersect(obj,itOut.back(),viewPoint))
      {
      objectsOut.push_back(obj);
      }
    else
      {
      itOut.pop_back();
      }
    }
  }
return(!objectsOut.empty());
}

//-------- Serialization operations ------------------------------------------

void
Scene3D::extractDependencies(SerializationDependencies & dependenciesOut,
                             bool recursive) const
{
ArObject::extractDependencies(dependenciesOut,recursive);
unsigned int i,nb=_objects.size();
for(i=0;i<nb;i++)
  {
  if(recursive)
    {
    _objects[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_objects[i]);
  }
nb=_selection.size();
for(i=0;i<nb;i++)
  {
  if(recursive)
    {
    _selection[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_selection[i]);
  }
nb=_particleSystems.size();
for(i=0;i<nb;i++)
  {
  if(recursive)
    {
    _particleSystems[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_particleSystems[i]);
  }
nb=_sounds.size();
for(i=0;i<nb;i++)
  {
  if(recursive)
    {
    _sounds[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_sounds[i]);
  }
nb=_lights.size();
for(i=0;i<nb;i++)
  {
  if(recursive)
    {
    _lights[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_lights[i]);
  }
if(_backTexture)
  {
  dependenciesOut.addReference(_backTexture);
  }
}

void
Scene3D::serialize(SerializationDependencies & dependenciesOut,
                   ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
unsigned int i,nb=_objects.size();
output->writeUInt(nb);
for(i=0;i<nb;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_objects[i]));
  }
nb=_selection.size();
output->writeUInt(nb);
for(i=0;i<nb;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_selection[i]));
  }
nb=_particleSystems.size();
output->writeUInt(nb);
for(i=0;i<nb;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_particleSystems[i]));
  }
nb=_sounds.size();
output->writeUInt(nb);
for(i=0;i<nb;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_sounds[i]));
  }
nb=_lights.size();
output->writeUInt(nb);
for(i=0;i<nb;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_lights[i]));
  }
output->writeDouble(_fogColor.x);
output->writeDouble(_fogColor.y);
output->writeDouble(_fogColor.z);
output->writeUInt(_fogDegree);
output->writeDouble(_fogDensity);
bool hasBackTransform=!_backTransform.isIdentity(false);
unsigned int flags=(_backTexture     ? 1<<0 : 0)+
                   (hasBackTransform ? 1<<1 : 0);
output->writeUInt(flags);
if(_backTexture)
  {
  output->writeUInt(dependenciesOut.addReference(_backTexture));
  }
if(hasBackTransform)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_backTransform.matrix[i]);
    }
  }
}

bool // success
Scene3D::unserialize(const SerializationDependencies & dependencies,
                     ArRef<AbstractIStream> input)
{
unsigned int nbObjects,nbSelected,nbParticleSystems,nbSounds,nbLights;
StlVector<unsigned int> objects,selection,particleSystems,sounds,lights;
Util3D::Dbl3 fogColor;
unsigned int fogDegree;
double fogDensity;
unsigned int flags,backTexture;
double matrix[16];
if(!ArObject::unserialize(dependencies,input)||
   !input->readUInt(nbObjects))
  {
  return(false);
  }
unsigned int i;
for(i=0;i<nbObjects;i++)
  {
  objects.push_back(0);
  if(!input->readUInt(objects.back()))
    {
    return(false);
    }
  }
if(!input->readUInt(nbSelected))
  {
  return(false);
  }
for(i=0;i<nbSelected;i++)
  {
  selection.push_back(0);
  if(!input->readUInt(selection.back()))
    {
    return(false);
    }
  }
if(!input->readUInt(nbParticleSystems))
  {
  return(false);
  }
for(i=0;i<nbParticleSystems;i++)
  {
  particleSystems.push_back(0);
  if(!input->readUInt(particleSystems.back()))
    {
    return(false);
    }
  }
if(!input->readUInt(nbSounds))
  {
  return(false);
  }
for(i=0;i<nbSounds;i++)
  {
  sounds.push_back(0);
  if(!input->readUInt(sounds.back()))
    {
    return(false);
    }
  }
if(!input->readUInt(nbLights))
  {
  return(false);
  }
for(i=0;i<nbLights;i++)
  {
  lights.push_back(0);
  if(!input->readUInt(lights.back()))
    {
    return(false);
    }
  }
if(!input->readDouble(fogColor.x)||
   !input->readDouble(fogColor.y)||
   !input->readDouble(fogColor.z)||
   !input->readUInt(fogDegree)||
   !input->readDouble(fogDensity)||
   !input->readUInt(flags)||
   ((flags&(1<<0))&&!input->readUInt(backTexture)))
  {
  return(false);
  }
bool hasBackTransform=flags&(1<<1);
if(hasBackTransform)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix[i]))
      {
      return(false);
      }
    }
  }
for(i=0;i<nbObjects;i++)
  {
  addObject(ar_down_cast<Object3D>
            (dependencies.findReference(objects[i])));
  }
for(i=0;i<nbSelected;i++)
  {
  selectObject(ar_down_cast<Object3D>
               (dependencies.findReference(selection[i])));
  }
for(i=0;i<nbParticleSystems;i++)
  {
  addParticleSystem(ar_down_cast<ParticleSystem>
                    (dependencies.findReference(particleSystems[i])));
  }
for(i=0;i<nbSounds;i++)
  {
  addSound(ar_down_cast<Sound3D>
           (dependencies.findReference(sounds[i])));
  }
for(i=0;i<nbLights;i++)
  {
  addLight(ar_down_cast<Light3D>
           (dependencies.findReference(lights[i])));
  }
_fogColor=fogColor;
_fogDegree=fogDegree;
_fogDensity=fogDensity;
if(flags&(1<<0))
  {
  setBackground(ar_down_cast<Texture>
                (dependencies.findReference(backTexture)));
  }
else
  {
  setBackground(Texture::nullRef());
  }
if(hasBackTransform)
  {
  _backTransform.setMatrix(matrix);
  }
else
  {
  _backTransform.identity();
  }
_postRedisplay();
_onSceneSetup();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Scene3D::_rebuildShapesIfNeeded(void) // Protected
{
for(unsigned int i=_objects.size();i--;)
  {
  _objects[i]->_rebuildShapesIfNeeded();
  }
}

void
Scene3D::_removeObject(ArRef<Object3D> obj) // Protected
{
for(unsigned int i=_objects.size();i--;)
  {
  if(_objects[i]==obj)
    {
    unselectObject(obj);
    StlVectorFastErase(_objects,i);
    _postRedisplay();
    obj->_removeScene(thisRef());
    break;
    }
  }
}

void
Scene3D::_removeLight(ArRef<Light3D> light) // Protected
{
for(unsigned int i=_lights.size();i--;)
  {
  if(_lights[i]==light)
    {
    StlVectorFastErase(_lights,i);
    _postRedisplay();
    light->_removeScene(thisRef());
    break;
    }
  }
}

void
Scene3D::_removeParticleSystem(ArRef<ParticleSystem> particleSystem)
{                                                                 // Protected
for(unsigned int i=_particleSystems.size();i--;)
  {
  if(_particleSystems[i]==particleSystem)
    {
    StlVectorFastErase(_particleSystems,i);
    _postRedisplay();
    particleSystem->_removeScene(thisRef());
    break;
    }
  }
}

void
Scene3D::_removeSound(ArRef<Sound3D> sound) // Protected
{
for(unsigned int i=_sounds.size();i--;)
  {
  if(_sounds[i]==sound)
    {
    StlVectorFastErase(_sounds,i);
    sound->_removeScene(thisRef());
    break;
    }
  }
}

void
Scene3D::_addRenderer(ArPtr<Renderer3D> renderer) // Protected
{
_renderers.push_back(renderer);
}

void
Scene3D::_removeRenderer(ArPtr<Renderer3D> renderer) // Protected
{
for(unsigned int i=_renderers.size();i--;)
  {
  if(_renderers[i]==renderer)
    {
    StlVectorFastErase(_renderers,i);
    break;
    }
  }
}

void
Scene3D::_addListener(ArPtr<Listener3D> listener) // Protected
{
_listeners.push_back(listener);
}

void
Scene3D::_removeListener(ArPtr<Listener3D> listener) // Protected
{
for(unsigned int i=_listeners.size();i--;)
  {
  if(_listeners[i]==listener)
    {
    StlVectorFastErase(_listeners,i);
    break;
    }
  }
}

void
Scene3D::_postRedisplay(void) // Protected
{
for(unsigned int i=_renderers.size();i--;)
  {
  _renderers[i]->_postRedisplay();
  }
}

void
Scene3D::_askVisibleParts(Renderer3D * renderer) const // Protected
{
for(unsigned int i=_selection.size();i--;)
  {
  const Shape3D * shape=_selection[i]->_findShape(renderer);
  if(shape)
    {
    shape->_askVisibleParts(renderer,_selection[i].c_ptr(),true);
    }
  }
for(unsigned int i=_objects.size();i--;)
  {
  const Shape3D * shape=_objects[i]->_findShape(renderer);
  if(shape)
    {
    shape->_askVisibleParts(renderer,_objects[i].c_ptr(),false);
    }
  }
}

void
Scene3D::_onSceneSetup(void)
{
if(!_sceneSetupCB.empty())
  {
  Scene3D::SceneSetupEvent event;
  _sceneSetupCB.fire(event);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
