// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/simpleInteractor.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/simpleInteractor.h"

#include "AReVi/arSystem.h"

#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/viewer3D.h"
#include "AReVi/Lib3D/osd.h"
#include "AReVi/Lib3D/boundingBox3D.h"

namespace AReVi {

#define NAVIGATION_DT 20
#define ROTATION_DT 20
#define ROTATION_TRESHOLD 100

//----------------------------------------------------------------------------
// Class SimpleInteractor
//----------------------------------------------------------------------------

AR_CLASS_DEF(SimpleInteractor,AbstractInteractor)

//-------- Construction / Destruction ----------------------------------------

SimpleInteractor::SimpleInteractor(ArCW & arCW) // Protected
: AbstractInteractor(arCW),
  _useObjectInteraction(true),
  _useOSDInteraction(true),
  _useShortcuts(true),
  _useExamination(true),
  _useNavigation(false),
  _mouseX(0),
  _mouseY(0),
  _mouseBtn(0),
  _localPoint(0.0,0.0,0.0),
  _restrictToMainScene(false),
  _canRotate(false),
  _timeOut(0),
  _lastTime(ArSystem::getTime()),
  _newLocation(),
  _newX(0),
  _newY(0),
  _lastXc(0.0),
  _lastDx(0.0),
  _lastDy(0.0),
  _lastDz(0.0),
  _lastAngle(0.0),
  _linearSpeed(1.0),
  _angularSpeed(M_PI),
  _clickX(0),
  _clickY(0),
  _pendingInteraction(false),
  _objInteracted(),
  _partInteracted(),
  _triInteracted(0),
  _osdInteracted(),
  _pendingExamination(false),
  _pendingNavigation(false)
{
  // Nothing to be done
}

SimpleInteractor::~SimpleInteractor(void) // Protected
{
setRenderer(Renderer3D::nullRef());
}

//-------- Interaction with objects ------------------------------------------

void
SimpleInteractor::setObjectInteraction(bool useObjectInteraction)
{
if(_useObjectInteraction!=useObjectInteraction)
  {
  _useObjectInteraction=useObjectInteraction;
  _clearInteraction();
  }
}

bool // object interaction is used
SimpleInteractor::getObjectInteraction(void) const
{
return(_useObjectInteraction);
}

void
SimpleInteractor::setMainSceneRestriction(bool restrictToMainScene)
{
_restrictToMainScene=restrictToMainScene;
}

bool // object interaction is restricted to main scene
SimpleInteractor::getMainSceneRestriction(void) const
{
return(_restrictToMainScene);
}

//-------- Interaction with osd ----------------------------------------------

void
SimpleInteractor::setOSDInteraction(bool useOSDInteraction)
{
if(_useOSDInteraction!=useOSDInteraction)
  {
  _useOSDInteraction=useOSDInteraction;
  _clearInteraction();
  }
}

bool // osd interaction is used
SimpleInteractor::getOSDInteraction(void) const
{
return(_useOSDInteraction);
}

//-------- Usual shortcuts ---------------------------------------------------

void
SimpleInteractor::setShortcuts(bool useShortcuts)
{
if(_useShortcuts!=useShortcuts)
  {
  _useShortcuts=useShortcuts;
  _clearInteraction();
  }
}

bool // keyboard shorcuts are used
SimpleInteractor::getShortcuts(void) const
{
return(_useShortcuts);
}

//-------- Scene examination -------------------------------------------------

void
SimpleInteractor::setExamination(bool useExamination)
{
if(_useExamination!=useExamination)
  {
  _useExamination=useExamination;
  if(_useExamination)
    {
    _useNavigation=false;
    }
  _clearInteraction();
  }
}

bool // examination is used
SimpleInteractor::getExamination(void) const
{
return(_useExamination);
}

//-------- Scene navigation --------------------------------------------------

void
SimpleInteractor::setNavigation(bool useNavigation)
{
if(_useNavigation!=useNavigation)
  {
  _useNavigation=useNavigation;
  if(_useNavigation)
    {
    _useExamination=false;
    }
  _clearInteraction();
  }
}

bool // navigation is used
SimpleInteractor::getNavigation(void) const
{
return(_useNavigation);
}

void
SimpleInteractor::setLinearSpeed(double linearSpeed)
{
_linearSpeed=fabs(linearSpeed);
}

double // half-screen linear speed in 1/s
SimpleInteractor::getLinearSpeed(void) const
{
return(_linearSpeed);
}

void
SimpleInteractor::setAngularSpeed(double angularSpeed)
{
_angularSpeed=fabs(angularSpeed);
}

double // half-screen angular speed in rad/s
SimpleInteractor::getAngularSpeed(void) const
{
return(_angularSpeed);
}

//------- Forbidden operations ----------------------------------------------

Util3D::Base & // current location
SimpleInteractor::_setupLocation(ArRef<Base3D> base)
{
if(!_newLocation)
  {
  _newLocation=Base3D::NEW();
  }
Util3D::Base & loc=_newLocation->accessBase();
loc=base->accessBase();
return(loc);
}

void
SimpleInteractor::_chooseCursor(void) // Protected
{
if(_renderer&&_renderer->getClass()->isA(Window3D::CLASS()))
  {
  Window3D::Cursor cursor=Window3D::CSR_ARROW;
  if(_pendingInteraction)
    {
    if(_osdInteracted)
      {
      cursor=Window3D::CSR_HAND;
      }
    else if(_objInteracted)
      {
      cursor=Window3D::CSR_HAND;
      }
    else if(_pendingExamination)
      {
      cursor=Window3D::CSR_MOVE;
      }
    else if(_pendingNavigation)
      {
      cursor=Window3D::CSR_CROSS;
      }
    }
  else
    {
    ArRef<OSD> osd;
    if(_useOSDInteraction)
      {
      ArRef<OSD> osd=_findOSD(_mouseX,_mouseY);
      if(osd)
	{
        if(osd->getMotionInteraction()||
           osd->getMouseButtonInteraction())
          {
          cursor=Window3D::CSR_HAND;
          }
        else if(osd->getKeyboardInteraction())
          {
          cursor=Window3D::CSR_TEXT;
          }	
	}
      }
    if(_useObjectInteraction&&!osd)
      {
      Util3D::Dbl3 localPoint;
      ArRef<ShapePart3D> part;
      unsigned int triIndex;
      ArRef<Object3D> found=_findObject(localPoint,part,triIndex,
                                        _mouseX,_mouseY,
                                        _restrictToMainScene);
      if(found)
        {
        if(found->getMotionInteraction()||
           found->getMouseButtonInteraction())
          {
          cursor=Window3D::CSR_HAND;
          }
        else if(found->getKeyboardInteraction())
          {
          cursor=Window3D::CSR_TEXT;
          }
        }
      }
    }
  ar_down_cast<Window3D>(_renderer)->setWindowCursor(cursor);
  }
}

void
SimpleInteractor::_clearTimeOut(void) // Protected
{
if(_timeOut)
  {
  if(_renderer)
    {
    _renderer->cancelTimeOut(_timeOut);
    }
  _timeOut=0;
  }
}

void
SimpleInteractor::_clearInteraction(void) // Protected
{
ArRef<SimpleInteractor> flag(this);
_canRotate=false;
_clearTimeOut();
if(_pendingInteraction)
  {
  if(_osdInteracted&&_renderer)
    {
    int x,y;
    _osdInteracted->getPosition(_renderer,x,y);
    _osdInteracted->onMotionInteraction(thisRef(),
                                        x,y,_osdInteracted->getOrientation(),
					false,true); // motion end
    if(flag &&
       _osdInteracted &&
       _osdInteracted->getMouseButtonInteraction())
      {
      _osdInteracted->onMouseButtonInteraction(thisRef(),
					       _mouseBtn,
                                               false); // fake release
      }
    }
  else if(_objInteracted)
    {
    _setupLocation(_objInteracted); // do not move !
    _objInteracted->onMotionInteraction(thisRef(),
                                        _partInteracted,_triInteracted,
                                        _newLocation,
					false,true); // motion end
    if(flag&&_objInteracted&&_objInteracted->getMouseButtonInteraction())
      {
      _objInteracted->onMouseButtonInteraction(thisRef(),
                                               _partInteracted,_triInteracted,
					       _mouseBtn,
                                               false); // fake release
      }
    }
  else if(_pendingExamination&&_renderer)
    {
    _setupLocation(_renderer); // do not move !
    _renderer->onMotionInteraction(thisRef(),_newLocation,
                                   false,true); // motion end
    }
  else if(_pendingNavigation&&_renderer)
    {
    _setupLocation(_renderer); // do not move !
    _renderer->onMotionInteraction(thisRef(),_newLocation,
                                   false,true); // motion end
    }
  }
if(flag)
  {
  _pendingInteraction=false;
  _pendingNavigation=false;
  _pendingExamination=false;
  _objInteracted.clear();
  _partInteracted.clear();
  _triInteracted=0;
  }
if(flag)
  {
  _osdInteracted.clear();
  }
if(flag)
  {
  _chooseCursor();
  }
}

void
SimpleInteractor::_onMouseButton(const Renderer3D::MouseButtonEvent & evt)
{                                                                 // Protected
ArRef<SimpleInteractor> flag(this);
_mouseX=evt.xMouse;
_mouseY=evt.yMouse;
if(_pendingInteraction)
  {
  if(_osdInteracted)
    {
    if(evt.button>3) // mouse wheel
      {
      if(_osdInteracted->getMouseButtonInteraction())
        {
        _osdInteracted->onMouseButtonInteraction(thisRef(),
						 evt.button,
						 evt.pressed);
        }
      }
    else if(!evt.pressed &&
	    (evt.button == _mouseBtn)) // end of interaction
      {
      _clearInteraction();
      }
    }
  else if(_objInteracted)
    {
    if(evt.button>3) // mouse wheel
      {
      if(_objInteracted->getMouseButtonInteraction())
        {
        _objInteracted->onMouseButtonInteraction(thisRef(),
                                                 _partInteracted,
                                                 _triInteracted,
						 evt.button,evt.pressed);
        }
      }
    else if(!evt.pressed&&(evt.button==_mouseBtn)) // end of interaction
      {
      _clearInteraction();
      }
    }
  else if(_pendingExamination)
    {
    if(!evt.pressed&&(evt.button==_mouseBtn))
      {
      bool couldRotate=_canRotate;
      _clearInteraction();
      if(couldRotate&&flag&&(ArSystem::getTime()-_lastTime<ROTATION_TRESHOLD))
        {
        _canRotate=true;
        _timeOut=evt.source->postTimeOut(ROTATION_DT);
        }
      }
    }
  else if(_pendingNavigation)
    {
    if(!evt.pressed&&(evt.button==_mouseBtn))
      {
      _clearInteraction();
      }
    }
  }
else
  {
  ArRef<OSD> osd;
  if(_useOSDInteraction)
    {
    osd=_findOSD(_mouseX,_mouseY);
    if(osd&&
       !osd->getMotionInteraction()&&
       !osd->getMouseButtonInteraction())
      {
      osd.clear();
      }
    }
  ArRef<Object3D> obj;
  ArRef<ShapePart3D> part;
  unsigned int triIndex;
  Util3D::Dbl3 localPoint;
  if(_useObjectInteraction&&!osd)
    {
    obj=_findObject(localPoint,
                    part,triIndex,
                    _mouseX,_mouseY,
                    _restrictToMainScene);
    if(obj&&
       !obj->getMotionInteraction()&&
       !obj->getMouseButtonInteraction())
      {
      obj.clear();
      }
    }
  if(osd)
    {
    if(evt.pressed)
      {
      if(osd->getMouseButtonInteraction())
	{
        osd->onMouseButtonInteraction(thisRef(),evt.button,evt.pressed);
	}
      if((evt.button<=3)&&flag&&osd->getMotionInteraction())
	{
        _mouseBtn=evt.button;
	_osdInteracted=osd;
	_pendingInteraction=true;
	_clearTimeOut();
	_chooseCursor();
	int x,y;
	osd->getPosition(evt.source,x,y);
	_osdInteracted->onMotionInteraction(thisRef(),x,y,
                                            _osdInteracted->getOrientation(),
                                            true,false); // motion begin
	}
      }
    else
      {
      if(osd->getMouseButtonInteraction())
	{
        osd->onMouseButtonInteraction(thisRef(),evt.button,evt.pressed);
	}
      }
    }
  else if(obj)
    {
    if(evt.pressed)
      {
      if(obj->getMouseButtonInteraction())
	{
        obj->onMouseButtonInteraction(thisRef(),
                                      part,triIndex,
                                      evt.button,evt.pressed);
	}
      if((evt.button<=3)&&flag&&obj->getMotionInteraction())
	{
        _mouseBtn=evt.button;
	_objInteracted=obj;
	_partInteracted=part;
	_triInteracted=triIndex;
	_localPoint=localPoint;
	_pendingInteraction=true;
	_canRotate=false;
	_clearTimeOut();
	_chooseCursor();
	_setupLocation(_objInteracted); // do not move !
	_objInteracted->onMotionInteraction(thisRef(),
                                            _partInteracted,_triInteracted,
                                            _newLocation,
					    true,false); // motion begin
	}
      }
    else
      {
      if(obj->getMouseButtonInteraction())
	{
        obj->onMouseButtonInteraction(thisRef(),
                                      part,triIndex,
                                      evt.button,evt.pressed);
	}
      }
    }
  else if(_useExamination)
    {
    if((evt.button<=3)&&evt.pressed)
      {
      _mouseBtn=evt.button;
      _canRotate=false;
      _clearTimeOut();
      _lastTime=ArSystem::getTime();
      _pendingExamination=true;
      _pendingInteraction=true;
      _chooseCursor();
      _setupLocation(evt.source); // do not move !
      evt.source->onMotionInteraction(thisRef(),_newLocation,
                                      true,false); // motion begin
      }
    }
  else if(_useNavigation)
    {
    if((evt.button<=3)&&evt.pressed)
      {
      _clickX=_mouseX;
      _clickY=_mouseY;
      _mouseBtn=evt.button;
      _canRotate=false;
      _clearTimeOut();
      _pendingNavigation=true;
      _pendingInteraction=true;
      _chooseCursor();
      _lastTime=ArSystem::getTime();
      _timeOut=evt.source->postTimeOut(NAVIGATION_DT);
      _setupLocation(evt.source); // do not move !
      evt.source->onMotionInteraction(thisRef(),_newLocation,
                                      true,false); // motion begin
      }
    }
  }
}

void
SimpleInteractor::_onMouseMotion(const Renderer3D::MouseMotionEvent & evt)
{                                                                 // Protected
int oldX=_mouseX;
int oldY=_mouseY;
_mouseX=evt.xMouse;
_mouseY=evt.yMouse;
if(_pendingInteraction)
  {
  if(_osdInteracted)
    {
    if(evt.source->shiftPressed())
      {
      double orientation;
      int x,y;
      _osdInteracted->getPosition(evt.source,x,y);
      orientation=_osdInteracted->getOrientation();
      unsigned int w,h;
      evt.source->getRenderSize(w,h);
      orientation+=M_PI*4.0*((double)(oldY-_mouseY))/(double)h;
      _osdInteracted->onMotionInteraction(thisRef(),x,y,orientation,
                                          false,false);
      }
    else
      {
      int x,y;
      _osdInteracted->getPosition(evt.source,x,y);
      x+=_mouseX-oldX;
      y+=_mouseY-oldY;
      _osdInteracted->onMotionInteraction(thisRef(),x,y,
                                          _osdInteracted->getOrientation(),
                                          false,false);
      }
    }
  else if(_objInteracted)
    {
    unsigned int w,h;
    evt.source->getRenderSize(w,h);
    if(evt.source->getStereoMode()==Renderer3D::STEREO_SPLIT)
      {
      w>>=1;
      }
    switch(_mouseBtn)
      {
      case 1:
        {
        if(evt.source->shiftPressed())
          {
          double dx=0.0;
          double dy=4.0*((double)(oldY-_mouseY))/(double)h;
          double dz=-4.0*((double)(oldX-_mouseX))/(double)w;
          double r=sqrt(dy*dy+dz*dz);
          if(r)
            {
            dy/=r;
            dz/=r;
            }
          evt.source->localToGlobalVector(dx,dy,dz);
          Util3D::Base & loc=_setupLocation(_objInteracted);
          loc.globalToLocalVector(dx,dy,dz);
          loc.rotate(dx,dy,dz,r*M_PI);
          _objInteracted->onMotionInteraction(thisRef(),
                                              _partInteracted,_triInteracted,
                                              _newLocation,false,false);
          }
        else
          {
          const Util3D::Perspective & persp=evt.source->getPerspective();
          double dx=0.0;
          double dy=_localPoint.x*(persp.getLeftSlope()-persp.getRightSlope())*
                    ((double)(oldX-_mouseX))/(double)w;
          double dz=_localPoint.x*(persp.getUpSlope()-persp.getDownSlope())*
                    ((double)(oldY-_mouseY))/(double)h;
          evt.source->localToGlobalVector(dx,dy,dz);
          Util3D::Base & loc=_setupLocation(_objInteracted);
          loc.globalToLocalVector(dx,dy,dz);
          loc.translate(dx,dy,dz);
          _objInteracted->onMotionInteraction(thisRef(),
                                              _partInteracted,_triInteracted,
                                              _newLocation,false,false);
          }
        break;
        }
      case 2:
      case 3:
        {
        double factor=2.0*((double)(oldY-_mouseY))/(double)h;
        double dx=_localPoint.x*factor;
        double dy=_localPoint.y*factor;
        double dz=_localPoint.z*factor;
        evt.source->localToGlobalVector(dx,dy,dz);
        Util3D::Base & loc=_setupLocation(_objInteracted);
        loc.globalToLocalVector(dx,dy,dz);
        loc.translate(dx,dy,dz);
        _objInteracted->onMotionInteraction(thisRef(),
                                            _partInteracted,_triInteracted,
                                            _newLocation,false,false);
        break;
        }
      }
    }
  else if(_pendingExamination)
    {
    _canRotate=false;
    unsigned int w,h;
    evt.source->getRenderSize(w,h);
    if(evt.source->getStereoMode()==Renderer3D::STEREO_SPLIT)
      {
      w>>=1;
      }
    double dx=2.0*((double)(_mouseX-oldX))/w;
    double dy=2.0*((double)(_mouseY-oldY))/h;
    double xc,yc,zc;
    if(!evt.source->getCenterOfView(xc,yc,zc))
      {
      xc=yc=zc=0.0;
      }
    evt.source->globalToLocalPosition(xc,yc,zc);
    if(xc<evt.source->getNearDistance())
      {
      xc=0.5*(evt.source->getFarDistance()+evt.source->getNearDistance());
      }
    switch(_mouseBtn)
      {
      case 1:
        {
        if(evt.source->shiftPressed())
          {
          Util3D::Base & loc=_setupLocation(evt.source);
          loc.translate(0.0,0.5*xc*dx,0.5*xc*dy);
          evt.source->onMotionInteraction(thisRef(),_newLocation,
                                          false,false);
          }
        else
          {
          _lastDx=0.0;
          _lastDy=dy;
          _lastDz=-dx;
          double r=sqrt(_lastDy*_lastDy+_lastDz*_lastDz);
          if(r)
            {
            _lastDy/=r;
            _lastDz/=r;
            }
          _lastAngle=r*M_PI;
          _lastXc=xc;
          Util3D::Base & loc=_setupLocation(evt.source);
          loc.translate(_lastXc,0.0,0.0);
          loc.rotate(_lastDx,_lastDy,_lastDz,_lastAngle);
          loc.translate(-_lastXc,0.0,0.0);
          evt.source->localToGlobalVector(_lastDx,_lastDy,_lastDz);
          unsigned int dt=ArSystem::getTime()-_lastTime;
          _lastTime+=dt;
          if(dt)
            {
            _lastAngle/=dt;
            }
          _canRotate=true;
          evt.source->onMotionInteraction(thisRef(),_newLocation,
                                          false,false);
          }
        break;
        }
      case 2:
      case 3:
        {
        Util3D::Base & loc=_setupLocation(evt.source);
        loc.translate(xc*dy,0.0,0.0);
        evt.source->onMotionInteraction(thisRef(),_newLocation,
                                        false,false);
        break;
        }
      }
    }
  else if(_pendingNavigation)
    {
      // Nothing to be done
    }
  }
else
  {
  _chooseCursor();
  }
}

void
SimpleInteractor::_onKeyboard(const Renderer3D::KeyboardEvent & evt)
{                                                                 // Protected
_mouseX=evt.xMouse;
_mouseY=evt.yMouse;
ArRef<SimpleInteractor> flag(this);
if(_useShortcuts)
  {
  if(evt.pressed)
    {
    if(evt.key=="Home")
      {
      Util3D::Base & loc=_setupLocation(evt.source);
      loc.cancelRotation();
      loc.setPosition(0.0,0.0,0.0);
      evt.source->onMotionInteraction(thisRef(),_newLocation,
                                      true,true); // motion begins and ends !
      }
    else if((evt.key=="d")||(evt.key=="D"))
      {
      if(evt.source->getClass()->isA(Viewer3D::CLASS()))
        {
        ArRef<Viewer3D> v=ar_down_cast<Viewer3D>(evt.source);
        v->setDecoration(!v->getDecoration());
        }
      }
    else if((evt.key=="w")||(evt.key=="W"))
      {
      ArRef<Object3D> found;
      if(_useObjectInteraction)
        {
        Util3D::Dbl3 localPoint;
        ArRef<ShapePart3D> part;
        unsigned int triIndex;
        found=_findObject(localPoint,part,triIndex,
                          _mouseX,_mouseY,
                          _restrictToMainScene);
        }
      if(found)
        {
        found->setWireFrame(!found->getWireFrame());
        }
      else
        {
        evt.source->setWireFrame(!evt.source->getWireFrame());
        }
      }
    else if((evt.key=="l")||(evt.key=="L"))
      {
      evt.source->setHeadLight(!evt.source->getHeadLight());
      }
    else if((evt.key=="o")||(evt.key=="O"))
      {
      evt.source->setOriginVisible(!evt.source->getOriginVisible());
      }
    else if((evt.key=="b")||(evt.key=="B"))
      {
      ArRef<Object3D> found;
      if(_useObjectInteraction)
        {
        Util3D::Dbl3 localPoint;
        ArRef<ShapePart3D> part;
        unsigned int triIndex;
        found=_findObject(localPoint,part,triIndex,
                          _mouseX,_mouseY,
                          _restrictToMainScene);
        }
      if(found)
        {
        if(found->getShapeBoundingBox())
          {
          if(found->getShapePartBoundingBox())
            {
            found->setShapeBoundingBox(false);
            }
          else
            {
            found->setShapePartBoundingBox(true);
            }
          }
        else
          {
          if(found->getShapePartBoundingBox())
            {
            found->setShapePartBoundingBox(false);
            }
          else
            {
            found->setShapeBoundingBox(true);
            }
          }
        }
      else
        {
        if(evt.source->getShapeBoundingBox())
          {
          if(evt.source->getShapePartBoundingBox())
            {
            evt.source->setShapeBoundingBox(false);
            }
          else
            {
            evt.source->setShapePartBoundingBox(true);
            }
          }
        else
          {
          if(evt.source->getShapePartBoundingBox())
            {
            evt.source->setShapePartBoundingBox(false);
            }
          else
            {
            evt.source->setShapeBoundingBox(true);
            }
          }
        }
      }
    else if((evt.key=="s")||(evt.key=="S"))
      {
      switch(evt.source->getStereoMode())
        {
        case Renderer3D::STEREO_NONE:
          {
          evt.source->setStereoMode(Renderer3D::STEREO_SPLIT);
          break;
          }
        case Renderer3D::STEREO_SPLIT:
          {
          evt.source->setStereoMode(Renderer3D::STEREO_COLOR);
          break;
          }
        case Renderer3D::STEREO_COLOR:
          {
          evt.source->setStereoMode(Renderer3D::STEREO_NONE);
          break;
          }
        }
      }
    else if((evt.key=="g")||(evt.key=="G"))
      {
      evt.source->setStereoGlance(!evt.source->getStereoGlance());
      }
    else if((evt.key=="r")||(evt.key=="R"))
      {
      if(evt.source->getClass()->isA(Window3D::CLASS()))
        {
        ArRef<Window3D> w=ar_down_cast<Window3D>(evt.source);
        w->setKeyRepeat(!w->getKeyRepeat());
        }
      }
    else if(evt.key=="Up")
      {
      if(evt.source->shiftPressed())
        {
        if(evt.source->controlPressed())
          {
          evt.source->setNearZOffset(evt.source->getNearZOffset()+
                                     0.1*evt.source->getNearDistance());
          }
        else
          {
          double dist=evt.source->getStereoViewDistance()*1.1;
          if(dist>evt.source->getFarDistance())
            {
            dist=evt.source->getFarDistance();
            }
          evt.source->setStereoViewDistance(dist);
          }
        }
      }
    else if(evt.key=="Down")
      {
      if(evt.source->shiftPressed())
        {
        if(evt.source->controlPressed())
          {
          evt.source->setNearZOffset(evt.source->getNearZOffset()-
                                     0.1*evt.source->getNearDistance());
          }
        else
          {
          double dist=evt.source->getStereoViewDistance()/1.1;
          if(dist<evt.source->getNearDistance())
            {
            dist=evt.source->getNearDistance();
            }
          evt.source->setStereoViewDistance(dist);
          }
        }
      }
    else if(evt.key=="Right")
      {
      if(evt.source->shiftPressed())
        {
        if(evt.source->controlPressed())
          {
          evt.source->setNearYOffset(evt.source->getNearYOffset()-
                                     0.1*evt.source->getNearDistance());
          }
        else
          {
          double separ=evt.source->getStereoEyeSeparation()*1.1;
          if(separ>evt.source->getFarDistance())
            {
            separ=evt.source->getFarDistance();
            }
          evt.source->setStereoEyeSeparation(separ);
          }
        }
      }
    else if(evt.key=="Left")
      {
      if(evt.source->shiftPressed())
        {
        if(evt.source->controlPressed())
          {
          evt.source->setNearYOffset(evt.source->getNearYOffset()+
                                     0.1*evt.source->getNearDistance());
          }
        else
          {
          double separ=evt.source->getStereoEyeSeparation()/1.1;
          if(separ<0.01*evt.source->getNearDistance())
            {
            separ=0.01*evt.source->getNearDistance();
            }
          evt.source->setStereoEyeSeparation(separ);
          }
        }
      }
    else if((evt.key=="n")||(evt.key=="N"))
      {
      if(getExamination())
        {
        setNavigation(true);
        }
      else if(getNavigation())
        {
        setNavigation(false);
        }
      else
        {
        setExamination(true);
        }
      }
    else if((evt.key=="i")||(evt.key=="I"))
      {
      if(getObjectInteraction())
        {
        if(getOSDInteraction())
          {
          setObjectInteraction(false);
          }
        else
          {
          setOSDInteraction(true);
          }
        }
      else
        {
        if(getOSDInteraction())
          {
          setOSDInteraction(false);
          }
        else
          {
          setObjectInteraction(true);
          }
        }
      }
    else if((evt.key=="h")||(evt.key=="H"))
      {
      cerr << endl;
      cerr << "Keyboard interaction for ``" << getAlias() << "'':" << endl;
      cerr << "          [h] display this help" << endl;
      cerr << "          [l] toggle head light" << endl;
      cerr << "          [o] toggle origin visibility" << endl;
      cerr << "          [b] toggle scene/object shape/part bounding boxes"
           << endl;
      cerr << "          [w] toggle scene/object wire frame" << endl;
      if(evt.source->getClass()->isA(Viewer3D::CLASS()))
        {
        cerr << "          [d] toggle viewer decoration" << endl;
        }
      cerr << "          [n] toggle examination/navigation" << endl;
      cerr << "          [i] toggle object/OSD interaction" << endl;
      cerr << "          [s] toggle stereo view" << endl;
      cerr << "          [g] toggle stereo glance" << endl;
      if(evt.source->getClass()->isA(Window3D::CLASS()))
        {
        cerr << "          [r] toggle key repeat" << endl;
        }
      cerr << "   [Shift+Up] increase stereo view distance" << endl;
      cerr << " [Shift+Down] decrease stereo view distance" << endl;
      cerr << "[Shift+Right] increase stereo eye separation" << endl;
      cerr << " [Shift+Left] decrease stereo eye separation" << endl;
      cerr << "       [Home] reset location" << endl;
      cerr << "        [Esc] close" << endl;
      cerr << endl;
      }
    else if(evt.key=="\x1b")
      {
      evt.source->simulateClose();
      }
    }
  }
ArRef<OSD> osd;
if(flag)
  {
  if(_useOSDInteraction)
    {
    osd=_findOSD(_mouseX,_mouseY);
    if(osd&&osd->getKeyboardInteraction())
      {
      osd->onKeyboardInteraction(thisRef(),evt.key,evt.pressed);
      }
    }
  }
if(flag)
  {
  if(_useObjectInteraction&&!osd)
    {
    Util3D::Dbl3 localPoint;
    ArRef<ShapePart3D> part;
    unsigned int triIndex;
    ArRef<Object3D> found=_findObject(localPoint,part,triIndex,
                                      _mouseX,_mouseY,
				      _restrictToMainScene);
    if(found&&found->getKeyboardInteraction())
      {
      found->onKeyboardInteraction(thisRef(),
                                   part,triIndex,
                                   evt.key,evt.pressed);
      }
    }
  }
}

void
SimpleInteractor::_onTimeOut(const Renderer3D::TimeOutEvent & evt)
{                                                                 // Protected
if(evt.timeOut==_timeOut)
  {
  if(_canRotate)
    {
    unsigned int dt=ArSystem::getTime()-_lastTime;
    _lastTime+=dt;
    double dx=_lastDx;
    double dy=_lastDy;
    double dz=_lastDz;
    evt.source->globalToLocalVector(dx,dy,dz);
    Util3D::Base & loc=_setupLocation(evt.source);
    loc.translate(_lastXc,0.0,0.0);
    loc.rotate(dx,dy,dz,_lastAngle*dt*0.25);
    loc.translate(-_lastXc,0.0,0.0);
    _timeOut=evt.source->postTimeOut(ROTATION_DT);
    evt.source->onMotionInteraction(thisRef(),_newLocation,
                                    true,true); // motion begins and ends !
    }
  else if(_pendingNavigation)
    {
    unsigned int dt=ArSystem::getTime()-_lastTime;
    _lastTime+=dt;
    unsigned int w,h;
    evt.source->getRenderSize(w,h);
    if(evt.source->getStereoMode()==Renderer3D::STEREO_SPLIT)
      {
      w>>=1;
      }
    double xRatio=(2.0*(_mouseX-_clickX))/w;
    double yRatio=(2.0*(_mouseY-_clickY))/h;
    Util3D::Base & loc=_setupLocation(evt.source);
    switch(_mouseBtn)
      {
      case 1:
        {
        loc.yaw(-xRatio*_angularSpeed*dt*1e-3);
        loc.translate(-yRatio*_linearSpeed*dt*1e-3,0.0,0.0);
        break;
        }
      case 2:
      case 3:
        {
        loc.translate(0.0,-xRatio*_linearSpeed*dt*1e-3,
                          -yRatio*_linearSpeed*dt*1e-3);
        break;
        }
      }
    _timeOut=evt.source->postTimeOut(NAVIGATION_DT);
    evt.source->onMotionInteraction(thisRef(),_newLocation,
                                    false,false);
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
