// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/texture.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/texture.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"

#include "AReVi/Shapes/shapePart3D.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/llw_priv.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/particleSystem.h"

namespace AReVi {

//-------- Initialization ----------------------------------------------------

static ArSystem::InitState _initState=ArSystem::AR_DOWN;

void
Texture::_init(void) // Static, Protected
{
if(_initState==ArSystem::AR_DOWN)
  {
  _initState=ArSystem::AR_GOING_UP;
  unsigned int size=4; // 2 x 2 x 1
  ArRef<MemoryBlock> data=MemoryBlock::NEW();
  data->setSize(size);
  Texture::_emptyTexData=new Texture::TexData(0,2,2,2,2,1,false,false,data);
  _initState=ArSystem::AR_UP;
  }
}

void
Texture::_destroy(void) // Static, Protected
{
if(_initState==ArSystem::AR_UP)
  {
  _initState=ArSystem::AR_GOING_DOWN;
  if(Texture::_emptyTexData)
    {
    delete Texture::_emptyTexData;
    Texture::_emptyTexData=NULL;
    }
  _initState=ArSystem::AR_DOWN;
  }
}

//----------------------------------------------------------------------------
// Class Texture (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Texture,ArObject)

Texture::TexData *
Texture::_emptyTexData=(Texture::TexData *)0;

//-------- Construction / Destruction ----------------------------------------

Texture::Texture(ArCW & arCW) // Protected
: ArObject(arCW),
  _parts(),
  _particleSystems(),
  _scenes(),
  _texData((Texture::TexData *)0)
{
  // Nothing to be done
}

Texture::~Texture(void) // Protected
{
if(_texData&&(_texData!=Texture::_emptyTexData))
  {
  delete _texData;
  }
while(!_parts.empty())
  {
  _parts.back()->_postAppearanceRebuild();
  _parts.back()->_postGeometryRebuild();
  _parts.pop_back();
  }
while(!_particleSystems.empty())
  {
  _particleSystems.back()->_postRedisplay();
  _particleSystems.back()->_onRendering();
  _particleSystems.pop_back();
  }
while(!_scenes.empty())
  {
  _scenes.back()->_postRedisplay();
  _scenes.pop_back();
  }
}

//-------- Inspection --------------------------------------------------------

unsigned int // texture width before rounding to 2^N
Texture::getInitialWidth(void) const
{
return(_texData ? _texData->initWidth : 0);
}

unsigned int // texture height before rounding to 2^N
Texture::getInitialHeight(void) const
{
return(_texData ? _texData->initHeight : 0);
}

unsigned int // texture width
Texture::getWidth(void) const
{
return(_texData ? _texData->width : 0);
}

unsigned int // texture height
Texture::getHeight(void) const
{
return(_texData ? _texData->height : 0);
}

unsigned int // 1: grey  2: grey-alpha  3: RGB  4: RGB-alpha
Texture::getComponents(void) const
{
return(_texData ? _texData->components : 0);
}

bool // blending is needed
Texture::getBlending(void) const
{
return(_texData ? _texData->needsBlending : false);
}

bool // alpha test is needed
Texture::getAlphaTest(void) const
{
return(_texData ? _texData->needsAlphaTest&&!_texData->needsBlending : false);
}

bool // mipmapping is used by OpenGL
Texture::usesMipmap(void) const
{
return(_texData ? _texData->useMipmap : false);
}

bool // cube mapping is used instead of normal 2d mapping by opengl(extension needed)
Texture::usesCubemap(void) const
{
return(_texData ? _texData->useCubeMap : false);
}
unsigned int // OpenGL texture ID
Texture::getTextureID(void) const
{
return(_texData ? _texData->texId : 0);
}

ArConstRef<MemoryBlock> // data filled from bottom to top
Texture::getData(void) const
{
return(_texData ? _texData->data : MemoryBlock::nullRef());
}


//-------- Serialization operations ------------------------------------------

bool // success
Texture::unserialize(const SerializationDependencies & dependencies,
                     ArRef<AbstractIStream> input)
{
if(!ArObject::unserialize(dependencies,input))
  {
  return(false);
  }
return(true);
}

void
Texture::serialize(SerializationDependencies & dependenciesOut,
                   ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
}

//-------- Forbidden operations ----------------------------------------------

unsigned int // size rounded up to 2^N
Texture::_roundSize(unsigned int size) // Static, Protected
{
unsigned int newSize=1;
while((newSize<<1)<=size) // (newSize=2^N) <= size
  {
  newSize<<=1;
  }
if(newSize<size) // (newSize=2^N) >= size
  {
  newSize<<=1;
  }
return(newSize);
}

void
Texture::_downScale(unsigned int inWidth,
                    unsigned int inHeight,
                    unsigned int components,
                    const char * inData,
                    unsigned int outWidth,
                    unsigned int outHeight,
                    char * outData) // Static, Protected
{
double sx=((double)inWidth)/((double)outWidth);
double sy=((double)inHeight)/((double)outHeight);
for(unsigned int y=outHeight;y--;)
  {
  for(unsigned int x=outWidth;x--;)
    {
    unsigned int outIndex=(x+(outWidth*y))*components;
    unsigned int x1=(unsigned int)(sx*x);
    unsigned int x2=(unsigned int)(sx*(x+1));
    if(x2==x1)
      {
      ++x2;
      }
    if(x2>inWidth)
      {
      x2=inWidth;
      }
    unsigned int y1=(unsigned int)(sy*y);
    unsigned int y2=(unsigned int)(sy*(y+1));
    if(y2==y1)
      {
      ++y2;
      }
    if(y2>inHeight)
      {
      y2=inHeight;
      }
    double comp[4]={0.0,0.0,0.0,0.0};
    unsigned int nb=0;
    for(unsigned int inY=y1;inY<y2;inY++)
      {
      for(unsigned int inX=x1;inX<x2;inX++)
        {
        unsigned int inIndex=(inX+(inWidth*inY))*components;
        for(unsigned int c=components;c--;)
          {
          comp[c]+=inData[inIndex+c];
          }
        ++nb;
        }
      }
    for(unsigned int c=components;c--;)
      {
      outData[outIndex+c]=(unsigned int)(comp[c]/nb);
      }
    }
  }
}

void
Texture::_upScale(unsigned int inWidth,
                  unsigned int inHeight,
                  unsigned int components,
                  const char * inData,
                  unsigned int outWidth,
                  unsigned int outHeight,
                  char * outData) // Static, Protected
{
double sx=((double)inWidth)/((double)outWidth);
double sy=((double)inHeight)/((double)outHeight);
for(unsigned int y=outHeight;y--;)
  {
  for(unsigned int x=outWidth;x--;)
    {
    unsigned int outIndex=(x+(outWidth*y))*components;
    unsigned int inX=(unsigned int)(sx*x);
    unsigned int inY=(unsigned int)(sy*y);
    unsigned int inIndex=(inX+(inWidth*inY))*components;
    for(unsigned int c=components;c--;)
      {
      outData[outIndex+c]=inData[inIndex+c];
      }
    }
  }
}

bool // success
Texture::_load(bool checkAlpha,
               bool compress) // Protected
{
GLenum format=(GLenum)0;
GLenum internalFormat=(GLenum)0;
const unsigned char * ptr=(const unsigned char *)
                          _texData->data->getMemory();
if(checkAlpha)
  {
  _texData->needsBlending=false;
  _texData->needsAlphaTest=false;
  }
switch(_texData->components)
  {
  case 1: // grey
    {
    format=(GLenum)GL_LUMINANCE;
    internalFormat=(GLenum)GL_COMPRESSED_LUMINANCE_ARB;
    break;
    }
  case 2: // grey-alpha
    {
    format=(GLenum)GL_LUMINANCE_ALPHA;
    internalFormat=(GLenum)GL_COMPRESSED_LUMINANCE_ALPHA_ARB;
    if(checkAlpha)
      {
      for(unsigned int i=_texData->height*_texData->width;i--;)
        {
        unsigned char alpha=ptr[2*i+1];
        if(alpha!=(unsigned char)255)
          {
          if(alpha)
            {
            _texData->needsBlending=true;
            }
          else
            {
            _texData->needsAlphaTest=true;
            }
          }
        }
      }
    break;
    }
  case 3: // RGB
    {
    format=(GLenum)GL_RGB;
    internalFormat=(GLenum)GL_COMPRESSED_RGB_ARB;
    break;
    }
  case 4: // RGB-alpha
    {
    format=(GLenum)GL_RGBA;
    internalFormat=(GLenum)GL_COMPRESSED_RGBA_ARB;
    if(checkAlpha)
      {
      for(unsigned int i=_texData->height*_texData->width;i--;)
        {
        unsigned char alpha=ptr[4*i+3];
        if(alpha!=(unsigned char)255)
          {
          if(alpha)
            {
            _texData->needsBlending=true;
            }
          else
            {
            _texData->needsAlphaTest=true;
            }
          }
        }
      }
    break;
    }
  }
if(checkAlpha&&_texData->needsBlending)
  {
  _texData->needsAlphaTest=false;
  }
if(!(compress&&GL::hasTextureCompression()))
  {
  internalFormat=format;
  }
const char * glErr=GL::lastError(); // clears previous error if any

if (_texData->useCubeMap)
  {
  if (!GL::hasTextureCubeMap())
    {
    setErrorMessage("Cannot send texture data to OpenGL : Missing support for CubeMap Extension");
    return(false);
    }

  glBindTexture(GL_TEXTURE_CUBE_MAP_EXT,_texData->texId);
  unsigned visibleHeight=_texData->height/6;
  for (unsigned i=6;i--;)
    {
    if(_texData->useMipmap)
      {
      if(GL::hasMipmapGeneration())
        {
        glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT,GL_GENERATE_MIPMAP_SGIS,GL_TRUE);
        glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT+i,0,internalFormat,
                    _texData->width,visibleHeight,
                    0,format,GL_UNSIGNED_BYTE,ptr);
        }
      else
        {
        gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT+i,internalFormat,
                          _texData->width,visibleHeight,
                          format,GL_UNSIGNED_BYTE,ptr);
        }
      }
    else
      {
      if(GL::hasMipmapGeneration())
        {
        glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT,GL_GENERATE_MIPMAP_SGIS,GL_FALSE);
        }
      glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT+i,0,internalFormat,
                  _texData->width,visibleHeight,
                  0,format,GL_UNSIGNED_BYTE,ptr);
      }
    ptr+=visibleHeight*_texData->width*_texData->components;
    }
  }
else
  {
  glBindTexture(GL_TEXTURE_2D,_texData->texId);
  if(_texData->useMipmap)
    {
    if(GL::hasMipmapGeneration())
      {
      glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS,GL_TRUE);
      glTexImage2D(GL_TEXTURE_2D,0,internalFormat,
                  _texData->width,_texData->height,
                  0,format,GL_UNSIGNED_BYTE,ptr);
      }
    else
      {
      gluBuild2DMipmaps(GL_TEXTURE_2D,internalFormat,
                        _texData->width,_texData->height,
                        format,GL_UNSIGNED_BYTE,ptr);
      }
    }
  else
    {
    if(GL::hasMipmapGeneration())
      {
      glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS,GL_FALSE);
      }
    glTexImage2D(GL_TEXTURE_2D,0,internalFormat,
                _texData->width,_texData->height,
                0,format,GL_UNSIGNED_BYTE,ptr);
    }
  }

glErr=GL::lastError();
if(glErr)
  {
  setErrorMessage("Cannot send texture data to OpenGL : "+StlString(glErr));
  return(false);
  }
else
  {
  _postRedisplay();
  return(true);
  }
}

void
Texture::_postRedisplay(void) // Protected
{
unsigned int i;
for(i=_parts.size();i--;)
  {
  _parts[i]->_postRedisplayIfVisible();
  }
for(i=_particleSystems.size();i--;)
  {
  _particleSystems[i]->_postRedisplay();
  }
for(i=_scenes.size();i--;)
  {
  _scenes[i]->_postRedisplay();
  }
}

void
Texture::_postAppearanceRebuild(void) // Protected
{
for(unsigned int i=_parts.size();i--;)
  {
  _parts[i]->_postAppearanceRebuild();
  }
}

void
Texture::_addShapePart(ShapePart3D * part)
{
_parts.push_back(part);
}

void
Texture::_removeShapePart(ShapePart3D * part) // Protected
{
for(unsigned int i=_parts.size();i--;)
  {
  if(_parts[i]==part)
    {
    StlVectorFastErase(_parts,i);
    break;
    }
  }
}

void
Texture::_addParticleSystem(ParticleSystem * particleSystem)
{
_particleSystems.push_back(particleSystem);
}

void
Texture::_removeParticleSystem(ParticleSystem * particleSystem) // Protected
{
for(unsigned int i=_particleSystems.size();i--;)
  {
  if(_particleSystems[i]==particleSystem)
    {
    StlVectorFastErase(_particleSystems,i);
    break;
    }
  }
}

void
Texture::_addScene(Scene3D * scene) // Protected
{
_scenes.push_back(scene);
}

void
Texture::_removeScene(Scene3D * scene) // Protected
{
for(unsigned int i=_scenes.size();i--;)
  {
  if(_scenes[i]==scene)
    {
    StlVectorFastErase(_scenes,i);
    break;
    }
  }
}

Texture::TexData::TexData(unsigned int t,
                          unsigned int iw,
                          unsigned int ih,
                          unsigned int w,
                          unsigned int h,
                          unsigned int c,
                          bool m,
                          bool cube,
                          ArRef<MemoryBlock> d)
: texId(t),
  initWidth(iw),
  initHeight(ih),
  width(w),
  height(h),
  components(c),
  needsBlending(false),
  needsAlphaTest(false),
  useMipmap(m),
  ownTexture(!t),
  useCubeMap(cube),
  data(d)
{
if(ownTexture)
  {
  glGenTextures(1,&texId);
  }
if (useCubeMap && GL::hasTextureCubeMap())
  {
  glBindTexture(GL_TEXTURE_CUBE_MAP_EXT,texId);
  glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  // glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  }
else
  {
  glBindTexture(GL_TEXTURE_2D,texId);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  // glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  }
}

Texture::TexData::~TexData(void)
{
if(ownTexture)
  {
  glDeleteTextures(1,&texId);
  }
}



//-------- Utilities operations ----------------------------------------------

// Maps from vector order : Front, Back, Left, Right, Top, Bottom
// NB : width must be equal to height, and all maps must have the same dims.
ArRef<MemoryBlock>
Texture::buildCubeTexture(unsigned &width, unsigned &height, unsigned &components,  //static
                         const StlVector<ArRef<MemoryBlock> > &maps, double attenuation)
{
if(width!=height) return(MemoryBlock::nullRef());
ArRef<MemoryBlock> data=MemoryBlock::NEW();
if(!data->ensureMinSize(width*height*components*6)) return(MemoryBlock::nullRef());
attenuation=attenuation>1.0?1.0:attenuation;
attenuation=attenuation<0.0?0.0:attenuation;

char *ptr(data->accessMemory());
// Back
const char *src=maps[0]->getMemory();
for (unsigned y=height;y--;)
  {
  for (unsigned x=width;x--;)
    {
    for (unsigned c=0;c<components;c++)
      {
      *ptr=src[(y*components)+(x*width*components)+c];
      ptr++;
      }
    }
  }

// Front
src=maps[1]->getMemory();
for (unsigned y=0;y<height;y++)
  {
  for (unsigned x=0;x<width;x++)
    {
    for (unsigned c=0;c<components;c++)
      {
      *ptr=src[(y*components)+(x*width*components)+c];
      ptr++;
      }
    }
  }

// Right
src=maps[2]->getMemory();
for (unsigned y=height;y--;)
  {
  for (unsigned x=0;x<width;x++)
    {
    for (unsigned c=0;c<components;c++)
      {
      *ptr=src[(x*components)+(y*width*components)+c];
      ptr++;
      }
    }
  }

// Left

src=maps[3]->getMemory();
for (unsigned y=0;y<height;y++)
  {
  for (unsigned x=width;x--;)
    {
    for (unsigned c=0;c<components;c++)
      {
      *ptr=src[(x*components)+(y*width*components)+c];
      ptr++;
      }
    }
  }

// Top
src=maps[4]->getMemory();
for (unsigned y=0;y<height;y++)
  {
  for (unsigned x=0;x<width;x++)
    {
    for (unsigned c=0;c<components;c++)
      {
      *ptr=src[(y*components)+(x*width*components)+c];
      ptr++;
      }
    }
  }


// Bottom
src=maps[5]->getMemory();
for (unsigned y=0;y<height;y++)
  {
  for (unsigned x=0;x<width;x++)
    {
    for (unsigned c=0;c<components;c++)
      {
      *ptr=src[(y*components)+(x*width*components)+c];
      ptr++;
      }
    }
  }

height*=6;

if (attenuation<1.0)
  {
  unsigned char *ptr=(unsigned char *)data->accessMemory();
  unsigned char offset=(unsigned char )(255*(1.0-attenuation));
  for (unsigned i=height*width*components;i--;)
    {
    *ptr=(unsigned char)(((*ptr)*attenuation)+(offset));
    ptr++;
    }
  }
return(data);
}



bool// success
Texture::buildCubeTexture(const StlString &frontUrl,const StlString &backUrl,  //static
                          const StlString &leftUrl,const StlString &rightUrl,
                          const StlString &topUrl,const StlString &bottomUrl,
                          const StlString &outUrl, double attenuation)
{
StlVector<ArRef<MemoryBlock> > maps;
ArRef<MemoryBlock> data;
unsigned int width,height,components;

// back
data=MemoryBlock::NEW();
if(!LLW::loadImage(backUrl,width,height,components,data))
  {
  cerr << ("Cannot load texture `"+backUrl+"'") << endl;
  return(false);
  }
maps.push_back(data);

// Front
data=MemoryBlock::NEW();
if(!LLW::loadImage(frontUrl,width,height,components,data))
  {
  cerr << ("Cannot load texture `"+frontUrl+"'") << endl;
  return(false);
  }
maps.push_back(data);

// right
data=MemoryBlock::NEW();
if(!LLW::loadImage(rightUrl,width,height,components,data))
  {
  cerr << ("Cannot load texture `"+rightUrl+"'") << endl;
  return(false);
  }
maps.push_back(data);

// left
data=MemoryBlock::NEW();
if(!LLW::loadImage(leftUrl,width,height,components,data))
  {
  cerr << ("Cannot load texture `"+leftUrl+"'") << endl;
  return(false);
  }
maps.push_back(data);

// top
data=MemoryBlock::NEW();
if(!LLW::loadImage(topUrl,width,height,components,data))
  {
  cerr << ("Cannot load texture `"+topUrl+"'") << endl;
  return(false);
  }
maps.push_back(data);

// bottom
data=MemoryBlock::NEW();
if(!LLW::loadImage(bottomUrl,width,height,components,data))
  {
  cerr << ("Cannot load texture `"+bottomUrl+"'") << endl;
  return(false);
  }
maps.push_back(data);



data=buildCubeTexture(width,height,components,maps,attenuation);
if (data.valid())
  {
  // Swap Result Image
  ArRef<MemoryBlock> tmp=MemoryBlock::NEW();
  if(!tmp->ensureMinSize(width*components)) return(false);
  unsigned int lineSize(width*components);
  for (unsigned i=0;i<height>>1;i++)
    {
    memcpy(tmp->accessMemory(),data->getMemory()+(i*lineSize),lineSize);
    memcpy(data->accessMemory()+(i*lineSize),data->getMemory()+((height-i-1)*lineSize),lineSize);
    memcpy(data->accessMemory()+((height-i-1)*lineSize),tmp->getMemory(),lineSize);
    }

  LLW::saveImage(outUrl,width,height,components,data);
  return(true);
  }
return(false);
}


} // namespace AReVi

//----------------------------------------------------------------------------
