// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/urlTexture.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/urlTexture.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/llw_priv.h"

namespace AReVi {

//-------- Initialization ----------------------------------------------------

static ArSystem::InitState _initState=ArSystem::AR_DOWN;

void
URLTexture::_init(void) // Static, Protected
{
if(_initState==ArSystem::AR_DOWN)
  {
  _initState=ArSystem::AR_GOING_UP;
  URLTexture::_cache=new StlMap<StlString,Texture::TexData *>;
  _initState=ArSystem::AR_UP;
  }
}

void
URLTexture::_destroy(void) // Static, Protected
{
if(_initState==ArSystem::AR_UP)
  {
  _initState=ArSystem::AR_GOING_DOWN;
  while(!URLTexture::_cache->empty())
    {
    delete (*URLTexture::_cache->begin()).second;
    URLTexture::_cache->erase(URLTexture::_cache->begin());
    }
  delete URLTexture::_cache;
  URLTexture::_cache=(StlMap<StlString,Texture::TexData *> *)0;
  _initState=ArSystem::AR_DOWN;
  }
}

//----------------------------------------------------------------------------
// Class URLTexture
//----------------------------------------------------------------------------

AR_CLASS_DEF(URLTexture,Texture)

StlMap<StlString,Texture::TexData *> *
URLTexture::_cache=(StlMap<StlString,Texture::TexData *> *)0;

//-------- Construction / Destruction ----------------------------------------

URLTexture::URLTexture(ArCW & arCW) // Protected
: Texture(arCW),
  _url(),
  _useCache(true)
{
_texData=Texture::_emptyTexData;
}

URLTexture::URLTexture(ArCW & arCW,
                       const StlString & url,
                       bool useCache,
                       bool useMipmap) // Protected
: Texture(arCW),
  _url(url),
  _useCache(useCache)
{
_loadURL(url,useCache,useMipmap,false);
}


URLTexture::URLTexture(ArCW & arCW,
                       const StlString & url,
                       bool useCache,
                       bool useMipmap,
                       bool useCubemap) // Protected
: Texture(arCW),
  _url(url),
  _useCache(useCache)
{
_loadURL(url,useCache,useMipmap,useCubemap);
}

URLTexture::~URLTexture(void) // Protected
{
if(_useCache)
  {
  _texData=(Texture::TexData *)0; // avoid destruction of cached data
  }
}

//-------- Inspection --------------------------------------------------------

const StlString & // texture URL
URLTexture::getURL(void) const
{
return(_url);
}

bool // cache is used for same texture URLs
URLTexture::usesCache(void) const
{
return(_useCache);
}

//-------- Serialization operations ------------------------------------------

void
URLTexture::extractDependencies(SerializationDependencies & dependenciesOut,
                                bool recursive) const
{
Texture::extractDependencies(dependenciesOut,recursive);
dependenciesOut.addURL(_url);
}

void
URLTexture::serialize(SerializationDependencies & dependenciesOut,
                      ArRef<AbstractOStream> output) const
{
Texture::serialize(dependenciesOut,output);
output->writeUInt(dependenciesOut.addURL(_url));
unsigned int flags=(_useCache                       ? 1<<0 : 0)+
                   ((_texData&&_texData->useMipmap) ? 1<<1 : 0)+
                   ((_texData&&_texData->useCubeMap) ? 1<<2 :0);


output->writeUInt(flags);
}

bool // success
URLTexture::unserialize(const SerializationDependencies & dependencies,
                        ArRef<AbstractIStream> input)
{
unsigned int url,flags;
if (!Texture::unserialize(dependencies,input)||
    !input->readUInt(url)||
    !input->readUInt(flags))
  {
  return(false);
  }
bool useCache=flags&(1<<0);
bool useMipmap=flags&(1<<1);
bool useCubemap=flags&(1<<2);
if(!_loadURL(dependencies.findURL(url),useCache,useMipmap,useCubemap))
  {
  return(false);
  }
_postAppearanceRebuild();
_postRedisplay();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

bool // success
URLTexture::_loadURL(const StlString & url, // Protected
                     bool useCache,
                     bool useMipmap,
                     bool useCubemap)
{
if(_texData&&!_useCache)
  {
  delete _texData;
  }
_texData=(Texture::TexData *)0;
_url=url;
_useCache=useCache;
StlMap<StlString,Texture::TexData *>::iterator it=
  (_useCache ? URLTexture::_cache->find(_url) : URLTexture::_cache->end());
if(it==URLTexture::_cache->end())
  {
  ArRef<MemoryBlock> data=MemoryBlock::NEW();
  unsigned int width,height,components;
  if(!LLW::loadImage(_url,width,height,components,data))
    {
    setErrorMessage("Cannot load texture `"+_url+"'");
    return(false);
    }

  unsigned int newWidth=Texture::_roundSize(width);
  unsigned int newHeight=useCubemap?Texture::_roundSize(height/6)*6:Texture::_roundSize(height);
  if((newWidth!=width)||(newHeight!=height))
    {
    ArRef<MemoryBlock> oldData=MemoryBlock::NEW();
    if(!oldData->ensureMinSize(newWidth*newHeight*components)||
       !oldData->swapBlock(data))
      {
      setErrorMessage("Cannot adjust data for a suitable texture size");
      return(false);
      }
    Texture::_upScale(width,height,components,oldData->getMemory(),
                      newWidth,newHeight,data->accessMemory());
    }
  _texData=new Texture::TexData(0,width,height,newWidth,newHeight,
                                components,useMipmap,useCubemap,data);
  if(!_load(true,true))
    {
    delete _texData;
    _texData=(Texture::TexData *)0;
    return(false);
    }
  if(_useCache)
    {
    URLTexture::_cache->insert(
      StlMap<StlString,Texture::TexData *>::value_type(_url,_texData));
    }
  }
else
  {
  _texData=(*it).second;
  }
return(true);
}

} // namespace AReVi

//----------------------------------------------------------------------------
