// LI2/ENIB - AReVi - 3D services
//----------------------------------------------------------------------------
// file        : AReVi/Lib3D/window3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/window3D.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Window3D (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Window3D,Renderer3D)

//-------- Construction / Destruction ----------------------------------------

Window3D::Window3D(ArCW & arCW) // Protected
: Renderer3D(arCW),
  _geometryCB(thisRef()),
  _cursorCB(thisRef())
{
  // _renderArea is initialized as a LLW::WinID by sub-classes
}

Window3D::~Window3D(void) // Protected
{
  // Nothing to be done
}

//-------- Window geometry ---------------------------------------------------

void
Window3D::setWindowGeometry(int x,
                            int y,
                            unsigned int width,
                            unsigned int height)
{
LLW::setWindowGeometry(_renderArea,x,y,width,height);
}

void
Window3D::getWindowGeometry(int & xOut,
                            int & yOut,
                            unsigned int & widthOut,
                            unsigned int & heightOut) const
{
LLW::getWindowGeometry(_renderArea,xOut,yOut,widthOut,heightOut);
}

void
Window3D::setWindowPosition(int x,
                            int y)
{
LLW::setWindowPosition(_renderArea,x,y);
}

void
Window3D::getWindowPosition(int & xOut,
                            int & yOut) const
{
LLW::getWindowPosition(_renderArea,xOut,yOut);
}

void
Window3D::setWindowSize(unsigned int width,
                        unsigned int height)
{
LLW::setWindowSize(_renderArea,width,height);
}

void
Window3D::getWindowSize(unsigned int & widthOut,
                        unsigned int & heightOut) const
{
LLW::getWindowSize(_renderArea,widthOut,heightOut);
}

void
Window3D::setRenderSize(unsigned int width,
                        unsigned int height)
{
LLW::setWindowSize(_renderArea,width,height);
}

//-------- Cursor management -------------------------------------------------

void
Window3D::setWindowCursor(Window3D::Cursor cursor)
{
LLW::setWindowCursor(_renderArea,(LLW::CursorID)cursor);
if(!_cursorCB.empty())
  {
  Window3D::CursorEvent event;
  event.cursor=(Window3D::Cursor)LLW::getWindowCursor(_renderArea);
  _cursorCB.fire(event);
  }
}

Window3D::Cursor // used cursor
Window3D::getWindowCursor(void) const
{
return((Window3D::Cursor)LLW::getWindowCursor(_renderArea));
}

//-------- Event management --------------------------------------------------

void
Window3D::setKeyRepeat(bool keyRepeat)
{
LLW::setWindowKeyRepeat(_renderArea,keyRepeat);
}

bool // key repeat is enabled
Window3D::getKeyRepeat(void) const
{
return(LLW::getWindowKeyRepeat(_renderArea));
}

//-------- Window ID ---------------------------------------------------------

int // platform-specific render window ID
Window3D::getRenderWindowID(void) const
{
return(LLW::getSpecificWindowID(_renderArea));
}

//-------- Appearance --------------------------------------------------------

void
Window3D::setBackgroundColor(double red,
                             double green,
                             double blue)
{
Renderer3D::setBackgroundColor(red,green,blue);
LLW::setWindowBackgroundColor(_renderArea,
                              LLW::makeDoubleColor(_backgroundColor.x,
                                                   _backgroundColor.y,
                                                   _backgroundColor.z));
}

//-------- Forbidden operations ----------------------------------------------

void
Window3D::_afterRenderAreaSetup(void) // Protected
{
Renderer3D::_afterRenderAreaSetup();
LLW::setWindowBackgroundColor(_renderArea,
                              LLW::makeDoubleColor(_backgroundColor.x,
                                                   _backgroundColor.y,
                                                   _backgroundColor.z));
LLW::setWindowEventMask(_renderArea,LLW::EVT_EXPOSE |
                                    LLW::EVT_GEOMETRY |
                                    LLW::EVT_KEYBOARD |
                                    LLW::EVT_MOUSE_BUTTON |
                                    LLW::EVT_MOUSE_MOTION |
                                    LLW::EVT_TIMEOUT);
LLW::mapWindow(_renderArea);
}

void
Window3D::_render(void) // Protected
{
if(LLW::isWindowViewable(_renderArea))
  {
  Renderer3D::_render();
  }
else
  {
  _rebuildShapesIfNeeded();
  }
}

bool // event processed
Window3D::_processEvent(const LLW::Event * evt) // Protected
{
if(evt->area==_renderArea)
  {
  switch(evt->type)
    {
    case LLW::EVT_EXPOSE:
      {
      _postRedisplay();
      break;
      }
    case LLW::EVT_MOUSE_MOTION:
      {
      _modifiers=(Renderer3D::Modifier)evt->modifiers;
      _lastMouseX=evt->x;
      _lastMouseY=evt->y;
      if(!_mouseMotionCB.empty())
        {
        Renderer3D::MouseMotionEvent event;
        event.xMouse=evt->x;
        event.yMouse=evt->y;
        _mouseMotionCB.fire(event);
        }
      break;
      }
    case LLW::EVT_GEOMETRY:
      {
      Renderer3D::_afterRenderAreaSetup();
      if(!_geometryCB.empty())
        {
        Window3D::GeometryEvent event;
        _geometryCB.fire(event);
        }
      break;
      }
    case LLW::EVT_KEYBOARD:
      {
      _modifiers=(Renderer3D::Modifier)evt->modifiers;
      _lastMouseX=evt->x;
      _lastMouseY=evt->y;
      if(!_keyboardCB.empty())
        {
        Renderer3D::KeyboardEvent event;
        event.key=evt->key;
        event.pressed=evt->onOff;
        event.xMouse=evt->x;
        event.yMouse=evt->y;
        _keyboardCB.fire(event);
        }
      break;
      }
    case LLW::EVT_MOUSE_BUTTON:
      {
      _modifiers=(Renderer3D::Modifier)evt->modifiers;
      _lastMouseX=evt->x;
      _lastMouseY=evt->y;
      if(!_mouseButtonCB.empty())
        {
        Renderer3D::MouseButtonEvent event;
        event.button=(int)evt->button;
        event.pressed=evt->onOff;
        event.xMouse=evt->x;
        event.yMouse=evt->y;
        _mouseButtonCB.fire(event);
        }
      break;
      }
    default:
      {
      Renderer3D::_processEvent(evt);
      }
    }
  return(true);
  }
else
  {
  return(false);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
