// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/bumping.cpp
// author(s)   : P.A. Beal and F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/bumping.h"

#include "AReVi/Lib3D/texture.h"

#include "AReVi/Utils/memoryBlock.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Bumping
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Bumping,Surface3D)

//-------- Construction / Destruction ----------------------------------------

Bumping::Bumping(ArCW & arCW,
                 ArRef<Texture> elevationTexture)
: Surface3D(arCW),
  _elevation(elevationTexture),
  _xMin(-1.0),
  _xMax(1.0),
  _yMin(-1.0),
  _yMax(1.0),
  _zMin(0.0),
  _zMax(1.0),
  _xPrecision(10),
  _yPrecision(10)
{
if(_elevation->fail())
  {
  setErrorMessage(_elevation->getErrorMessage());
  return;
  }
if(_elevation->getComponents()!=1)
  {
  setErrorMessage("Elevation texture should have only one component");
  return;
  }
}

Bumping::~Bumping(void)
{
  // Nothing to be done
}

//-------- Geometry ----------------------------------------------------------

void
Bumping::setHeight(double minHeight,
                   double maxHeight)
{
_zMin=minHeight;
_zMax=maxHeight;
_postGeometryRebuild();
}

void
Bumping::getHeight(double & minHeightOut,
                   double & maxHeightOut) const
{
minHeightOut=_zMin;
maxHeightOut=_zMax;
}

void
Bumping::setBounds(double xMin,
                   double xMax,
                   double yMin,
                   double yMax)
{
_xMin=xMin;
_xMax=xMax;
_yMin=yMin;
_yMax=yMax;
_postGeometryRebuild();
}

void
Bumping::getBounds(double & xMinOut,
                   double & xMaxOut,
                   double & yMinOut,
                   double & yMaxOut) const
{
xMinOut=_xMin;
xMaxOut=_xMax;
yMinOut=_yMin;
yMaxOut=_yMax;
}

//-------- Elevation ---------------------------------------------------------

ArConstRef<Texture>
Bumping::getElevationTexture(void) const
{
return(_elevation);
}

void
Bumping::setPrecision(unsigned int xPrecision,
                      unsigned int yPrecision)
{
_xPrecision=xPrecision;
_yPrecision=yPrecision;
_postGeometryRebuild();
}

void
Bumping::getPrecision(unsigned int & xPrecisionOut,
                      unsigned int & yPrecisionOut) const
{
xPrecisionOut=_xPrecision;
yPrecisionOut=_yPrecision;
}

//-------- Forbidden operations ----------------------------------------------

void
Bumping::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  _facesV.clear();
  _vertices.clear();
  _facesT.clear();
  _textures.clear();
  unsigned int x,y;
  double xFactor=_xMax-_xMin;
  double yFactor=_yMax-_yMin;
  double zFactor=(_zMax-_zMin)/255.0;
  double xStep=(_xPrecision ? 1.0/_xPrecision : 1.0);
  double yStep=(_yPrecision ? 1.0/_yPrecision : 1.0);
  const unsigned char * mem=reinterpret_cast<const unsigned char *>
                            (_elevation->getData()->getMemory());
  unsigned int w=_elevation->getWidth();
  unsigned int h=_elevation->getHeight();
  for(y=0;y<=_yPrecision;y++)
    {
    double yt=y*yStep;
    unsigned int py=static_cast<unsigned int>(yt*(h-1));
    for(x=0;x<=_xPrecision;x++)
      {
      double xt=x*xStep;
      unsigned int px=static_cast<unsigned int>(xt*(w-1));
      _vertices.push_back(Util3D::Dbl3(_xMin+xt*xFactor,
                                       _yMin+yt*yFactor,
                                       _zMin+mem[px+py*w]*zFactor));
      _textures.push_back(Util3D::Dbl2(xt,yt));
      }
    }
  for(y=0;y<_yPrecision;y++)
    {
    for(x=0;x<_xPrecision;x++)
      {
      _facesV.push_back(Util3D::Face(y*(_xPrecision+1)+x,
                                     y*(_xPrecision+1)+(x+1),
                                     (y+1)*(_xPrecision+1)+(x+1)));
      _facesV.push_back(Util3D::Face(y*(_xPrecision+1)+x,
                                     (y+1)*(_xPrecision+1)+(x+1),
                                     (y+1)*(_xPrecision+1)+x));
      }
    }
  // _facesV will be used instead of _facesT
  _computeDefaultNormals(true);
  }
Surface3D::_rebuildIfNeeded();
}

} // namespace AReVi

//----------------------------------------------------------------------------
