// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/cone3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/cone3D.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Cone3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Cone3D,Surface3D)

//-------- Construction / Destruction ----------------------------------------

Cone3D::Cone3D(ArCW & arCW) // Protected
: Surface3D(arCW),
  _radius(1.0),
  _height(2.0),
  _side(true),
  _bottom(true),
  _ccw(true),
  _precision(16)
{
  // Nothing to be done
}

Cone3D::~Cone3D(void) // Protected
{
  // Nothing to be done
}

//-------- Geometry ----------------------------------------------------------

bool // success
Cone3D::setRadius(double radius)
{
if(radius>0.0)
  {
  _radius=radius;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

double // radius
Cone3D::getRadius(void) const
{
return(_radius);
}

bool // success
Cone3D::setHeight(double height)
{
if(height>0.0)
  {
  _height=height;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

double // height
Cone3D::getHeight(void) const
{
return(_height);
}

void
Cone3D::setSide(bool buildSide)
{
_side=buildSide;
_postGeometryRebuild();
}

bool // side is built
Cone3D::getSide(void) const
{
return(_side);
}

void
Cone3D::setBottom(bool buildBottom)
{
_bottom=buildBottom;
_postGeometryRebuild();
}

bool // bottom is built
Cone3D::getBottom(void) const
{
return(_bottom);
}

void
Cone3D::setCCW(bool ccw)
{
_ccw=ccw;
_postGeometryRebuild();
}

bool // counter clock wise surface
Cone3D::getCCW(void) const
{
return(_ccw);
}

bool // success
Cone3D::setPrecision(unsigned int precision)
{
if(precision>0)
  {
  _precision=precision;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // precision
Cone3D::getPrecision(void) const
{
return(_precision);
}

//-------- Serialization operations ------------------------------------------

void
Cone3D::serialize(SerializationDependencies & dependenciesOut,
                  ArRef<AbstractOStream> output) const
{
Surface3D::serialize(dependenciesOut,output);
output->writeDouble(_radius);
output->writeDouble(_height);
output->writeUInt(_precision);
unsigned int flags=(_side   ? 1<<0 : 0)+
                   (_bottom ? 1<<1 : 0)+
                   (_ccw    ? 1<<2 : 0);
output->writeInt(flags);
}

bool // success
Cone3D::unserialize(const SerializationDependencies & dependencies,
                    ArRef<AbstractIStream> input)
{
double radius,height;
unsigned int precision,flags;
if(!Surface3D::unserialize(dependencies,input)||
   !input->readDouble(radius)||
   !input->readDouble(height)||
   !input->readUInt(precision)||
   !input->readUInt(flags))
  {
  return(false);
  }
_radius=radius;
_height=height;
_precision=precision;
_side=flags&(1<<0);
_bottom=flags&(1<<1);
_ccw=flags&(1<<2);
_postGeometryRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Cone3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  _vertices.clear();
  _facesV.clear();
  _normals.clear();
  _facesN.clear();
  _textures.clear();
  _facesT.clear();
  double height=0.5*_height;
  double normalZ=_radius*atan(_radius/_height);
  if(_bottom)
    {
    if(_side)
      {
      unsigned int i;
      for(i=0;i<_precision;i++)
        {
        double xt=(double)i/(double)_precision;
        double angle=M_2PI*(0.5+xt);
        double c=cos(angle);
        double s=sin(angle);
        double x=_radius*c;
        double y=_radius*s;
        if(_ccw)
          {
          _vertices.push_back(Util3D::Dbl3(x,y,-height));
          _normals.push_back(Util3D::Dbl3(x,y,normalZ));
          _textures.push_back(Util3D::Dbl2(xt,0.0));
          _textures.push_back(Util3D::Dbl2(0.5*(1.0+s),0.5*(1.0+c)));
          }
        else
          {
          _vertices.push_back(Util3D::Dbl3(-x,-y,-height));
          _normals.push_back(Util3D::Dbl3(x,y,-normalZ));
          _textures.push_back(Util3D::Dbl2(1.0-xt,0.0));
          _textures.push_back(Util3D::Dbl2(0.5*(1.0-s),0.5*(1.0+c)));
          }
        xt=((double)i+0.5)/(double)_precision;
        angle=M_2PI*(0.5+xt);
        c=cos(angle);
        s=sin(angle);
        x=_radius*c;
        y=_radius*s;
        if(_ccw)
          {
          _normals.push_back(Util3D::Dbl3(x,y,normalZ));
          _textures.push_back(Util3D::Dbl2(xt,1.0));
          }
        else
          {
          _normals.push_back(Util3D::Dbl3(-x,-y,-normalZ));
          _textures.push_back(Util3D::Dbl2(1.0-xt,1.0));
          }
        }
      if(_ccw)
        {
        _textures.push_back(Util3D::Dbl2(1.0,0.0)); // side end
        }
      else
        {
        _textures.push_back(Util3D::Dbl2(0.0,0.0)); // side end
        }
      _textures.push_back(Util3D::Dbl2(0.5,0.0)); // bottom end
      unsigned int center=_vertices.size();
      _vertices.push_back(Util3D::Dbl3(0.0,0.0,height));
      if(_ccw)
        {
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(center,i,next));
          _facesN.push_back(Util3D::Face(2*i+1,2*i,2*next));
          _facesT.push_back(Util3D::Face(3*i+2,3*i,3*(i+1)));
          }
        }
      else
        {
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(center,next,i));
          _facesN.push_back(Util3D::Face(2*i+1,2*next,2*i));
          _facesT.push_back(Util3D::Face(3*i+2,3*(i+1),3*i));
          }
        }
      center=_vertices.size();
      _vertices.push_back(Util3D::Dbl3(0.0,0.0,-height));
      unsigned int centerT=_textures.size();
      _textures.push_back(Util3D::Dbl2(0.5,0.5));
      unsigned int normal=_normals.size();
      if(_ccw)
        {
        _normals.push_back(Util3D::Dbl3(0.0,0.0,-1.0));
        _textures.push_back(Util3D::Dbl2(0.5,0.5));
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(i,center,next));
          _facesN.push_back(Util3D::Face(normal,normal,normal));
          _facesT.push_back(Util3D::Face(3*i+1,centerT,3*(i+1)+1));
          }
        }
      else
        {
        _normals.push_back(Util3D::Dbl3(0.0,0.0,1.0));
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(i,next,center));
          _facesN.push_back(Util3D::Face(normal,normal,normal));
          _facesT.push_back(Util3D::Face(3*i+1,3*(i+1)+1,centerT));
          }
        }
      }
    else // bottom only
      {
      unsigned int i;
      for(i=0;i<_precision;i++)
        {
        double xt=(double)i/(double)_precision;
        double angle=M_2PI*(0.5+xt);
        double c=cos(angle);
        double s=sin(angle);
        double x=_radius*c;
        double y=_radius*s;
        if(_ccw)
          {
          _vertices.push_back(Util3D::Dbl3(x,y,-height));
          _textures.push_back(Util3D::Dbl2(0.5*(1.0+s),0.5*(1.0+c)));
          }
        else
          {
          _vertices.push_back(Util3D::Dbl3(-x,-y,-height));
          _textures.push_back(Util3D::Dbl2(0.5*(1.0-s),0.5*(1.0+c)));
          }
        }
      _textures.push_back(Util3D::Dbl2(0.5,0.0)); // bottom end
      unsigned int center=_vertices.size();
      _vertices.push_back(Util3D::Dbl3(0.0,0.0,-height));
      unsigned int centerT=_textures.size();
      _textures.push_back(Util3D::Dbl2(0.5,0.5));
      unsigned int normal=_normals.size();
      if(_ccw)
        {
        _normals.push_back(Util3D::Dbl3(0.0,0.0,-1.0));
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(i,center,next));
          _facesN.push_back(Util3D::Face(normal,normal,normal));
          _facesT.push_back(Util3D::Face(i,centerT,i+1));
          }
        }
      else
        {
        _normals.push_back(Util3D::Dbl3(0.0,0.0,1.0));
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(i,next,center));
          _facesN.push_back(Util3D::Face(normal,normal,normal));
          _facesT.push_back(Util3D::Face(i,i+1,centerT));
          }
        }
      }
    }
  else
    {
    if(_side) // side only
      {
      unsigned int i;
      for(i=0;i<_precision;i++)
        {
        double xt=(double)i/(double)_precision;
        double angle=M_2PI*(0.5+xt);
        double c=cos(angle);
        double s=sin(angle);
        double x=_radius*c;
        double y=_radius*s;
        if(_ccw)
          {
          _vertices.push_back(Util3D::Dbl3(x,y,-height));
          _normals.push_back(Util3D::Dbl3(x,y,normalZ));
          _textures.push_back(Util3D::Dbl2(xt,0.0));
          }
        else
          {
          _vertices.push_back(Util3D::Dbl3(-x,-y,-height));
          _normals.push_back(Util3D::Dbl3(x,y,-normalZ));
          _textures.push_back(Util3D::Dbl2(1.0-xt,0.0));
          }
        xt=((double)i+0.5)/(double)_precision;
        angle=-M_2PI*xt;
        c=cos(angle);
        s=sin(angle);
        x=_radius*c;
        y=_radius*s;
        if(_ccw)
          {
          _normals.push_back(Util3D::Dbl3(x,y,normalZ));
          _textures.push_back(Util3D::Dbl2(xt,1.0));
          }
        else
          {
          _normals.push_back(Util3D::Dbl3(-x,-y,-normalZ));
          _textures.push_back(Util3D::Dbl2(1.0-xt,1.0));
          }
        }
      unsigned int center=_vertices.size();
      _vertices.push_back(Util3D::Dbl3(0.0,0.0,height));
      if(_ccw)
        {
        _textures.push_back(Util3D::Dbl2(1.0,0.0)); // side end
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(center,i,next));
          _facesN.push_back(Util3D::Face(2*i+1,2*i,2*next));
          _facesT.push_back(Util3D::Face(2*i+1,2*i,2*(i+1)));
          }
        }
      else
        {
        _textures.push_back(Util3D::Dbl2(0.0,0.0)); // side end
        for(i=0;i<_precision;i++)
          {
          unsigned int next=(i+1)%_precision;
          _facesV.push_back(Util3D::Face(center,next,i));
          _facesN.push_back(Util3D::Face(2*i+1,2*next,2*i));
          _facesT.push_back(Util3D::Face(2*i+1,2*(i+1),2*i));
          }
        }
      }
    }
  }
Surface3D::_rebuildIfNeeded();
}

} // namespace AReVi

//----------------------------------------------------------------------------
