// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/lineSet3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/lineSet3D.h"

#include "AReVi/Lib3D/gl_priv.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class LineSet3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(LineSet3D,ShapePart3D)

//-------- Construction / Destruction ----------------------------------------

LineSet3D::LineSet3D(ArCW & arCW) // Protected
: ShapePart3D(arCW),
  _thickness(1),
  _vertices(),
  _colors()
{
_describeGeometry(_bbox,
                  (const Util3D::Dbl3 *)0,0,
                  (const Util3D::Face *)0,0,
                  0,0);
}

LineSet3D::~LineSet3D(void) // Protected
{
  // Nothing to be done
}

//-------- Changes -----------------------------------------------------------

bool // success
LineSet3D::applyChanges(bool checkConsistency)
{
if(checkConsistency)
  {
  if(_colors.size()<_vertices.size())
    {
    return(false);
    }
  for(unsigned int l=_vertices.size();l--;)
    {
    unsigned int nbV=_vertices[l].size();
    unsigned int nbC=_colors[l].size();
    if(nbV&&(nbC>1)&&(nbC<nbV))
      {
      return(false);
      }
    }
  }
_postGeometryRebuild();
return(true);
}

//-------- Appearance --------------------------------------------------------

bool // success
LineSet3D::setThickness(unsigned int thickness)
{
if(thickness>0)
  {
  _thickness=thickness;
  _postAppearanceRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // lines thickness
LineSet3D::getThickness(void) const
{
return(_thickness);
}

//-------- Vertices ----------------------------------------------------------

const StlVector<StlVector<Util3D::Dbl3> > & // vertex coords
LineSet3D::getVertices(void) const
{
return(_vertices);
}

StlVector<StlVector<Util3D::Dbl3> > & // vertex coords
LineSet3D::accessVertices(void)
{
return(_vertices);
}

//-------- Colors ------------------------------------------------------------

const StlVector<StlVector<Util3D::Dbl3> > & // color components
LineSet3D::getColors(void) const
{
return(_colors);
}

StlVector<StlVector<Util3D::Dbl3> > & // color components
LineSet3D::accessColors(void)
{
return(_colors);
}

//-------- Serialization operations ------------------------------------------

void
LineSet3D::serialize(SerializationDependencies & dependenciesOut,
                     ArRef<AbstractOStream> output) const
{
ShapePart3D::serialize(dependenciesOut,output);
output->writeUInt(_thickness);
unsigned int nbVertices=_vertices.size(),nbColors=_colors.size();
output->writeUInt(nbVertices);
output->writeUInt(nbColors);
unsigned int i,j;
for(i=0;i<nbVertices;i++)
  {
  unsigned int nbV=_vertices[i].size();
  output->writeUInt(nbV);
  for(j=0;j<nbV;j++)
    {
    output->writeDouble(_vertices[i][j].x);
    output->writeDouble(_vertices[i][j].y);
    output->writeDouble(_vertices[i][j].z);
    }
  }
for(i=0;i<nbColors;i++)
  {
  unsigned int nbC=_colors[i].size();
  output->writeUInt(nbC);
  for(j=0;j<nbC;j++)
    {
    output->writeDouble(_colors[i][j].x);
    output->writeDouble(_colors[i][j].y);
    output->writeDouble(_colors[i][j].z);
    }
  }
}

bool // success
LineSet3D::unserialize(const SerializationDependencies & dependencies,
                       ArRef<AbstractIStream> input)
{
unsigned int thickness,nbVertices,nbColors;
StlVector<StlVector<Util3D::Dbl3> > vertices;
StlVector<StlVector<Util3D::Dbl3> > colors;
if(!ShapePart3D::unserialize(dependencies,input)||
   !input->readUInt(thickness)||
   !input->readUInt(nbVertices)||
   !input->readUInt(nbColors))
  {
  return(false);
  }
unsigned int i,j;
vertices.resize(nbVertices);
for(i=0;i<nbVertices;i++)
  {
  unsigned int nbV;
  if(!input->readUInt(nbV))
    {
    return(false);
    }
  vertices[i].resize(nbV);
  for(j=0;j<nbV;j++)
    {
    if(!input->readDouble(vertices[i][j].x)||
       !input->readDouble(vertices[i][j].y)||
       !input->readDouble(vertices[i][j].z))
      {
      return(false);
      }
    }
  }
colors.resize(nbColors);
for(i=0;i<nbColors;i++)
  {
  unsigned int nbC;
  if(!input->readUInt(nbC))
    {
    return(false);
    }
  colors[i].resize(nbC);
  for(j=0;j<nbC;j++)
    {
    if(!input->readDouble(colors[i][j].x)||
       !input->readDouble(colors[i][j].y)||
       !input->readDouble(colors[i][j].z))
      {
      return(false);
      }
    }
  }
_thickness=thickness;
_vertices.swap(vertices);
_colors.swap(colors);
_postGeometryRebuild();
_postAppearanceRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
LineSet3D::_setAppearance(void) const // Protected
{
if(_offset)
  {
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(_offset,_offset);
  }
glDisable(GL_LIGHTING);
glLineWidth(_thickness);
}

void
LineSet3D::_resetAppearance(void) const // Protected
{
glLineWidth(1.0);
glEnable(GL_LIGHTING);
if(_offset)
  {
  glDisable(GL_POLYGON_OFFSET_FILL);
  }
}

void
LineSet3D::_drawGeometry(void) const // Protected
{
if(!_vertices.empty())
  {
  for(unsigned int l=_vertices.size();l--;)
    {
    unsigned int nbV=_vertices[l].size();
    if(nbV>1)
      {
      unsigned int nbC=_colors[l].size();
      switch(nbC)
        {
        case 0:
          glColor3f(0.8,0.8,0.8); // default color
          break;
        case 1:
          glColor3dv(&_colors[l].front().x); // common color
          break;
        }
      bool hasColors=(nbC>1);
      glBegin(GL_LINE_STRIP);
      for(unsigned int i=nbV;i--;)
        {
        if(hasColors)
          {
          glColor3dv(&_colors[l][i].x);
          }
        const Util3D::Dbl3 & vertex=_vertices[l][i];
        glVertex3dv(&vertex.x);
        }
      glEnd(); // GL_LINE_STRIP
      }
    }
  }
}

void
LineSet3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_geometryList,GL_COMPILE);
    _drawGeometry();
    glEndList();
    }
  unsigned int nbLines=0;
  Util3D::BoundingBox bbox;
  for(unsigned int l=_vertices.size();l--;)
    {
    unsigned int nbV=_vertices[l].size();
    if(nbV>1)
      {
      nbLines+=nbV-1;
      for(unsigned int i=nbV;i--;)
        {
        const Util3D::Dbl3 & vertex=_vertices[l][i];
        bbox.addXYZ(vertex.x,vertex.y,vertex.z);
        }
      }
    }
  _describeGeometry(bbox,
                    (const Util3D::Dbl3 *)0,0,
                    (const Util3D::Face *)0,0,
                    nbLines,0);
  _needsGeometryRebuild=false;
  }
if(_needsAppearanceRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_appearanceGeometryList,GL_COMPILE);
    _setAppearance();
    glCallList(_geometryList);
    _resetAppearance();
    glEndList();
    }
  _needsAppearanceRebuild=false;
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
