// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/mesh3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/mesh3D.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Mesh3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Mesh3D,Surface3D)

//-------- Construction / Destruction ----------------------------------------

Mesh3D::Mesh3D(ArCW & arCW) // Protected
: Surface3D(arCW)
{
_rebuildIfNeeded(); // start with an empty shape
}

Mesh3D::~Mesh3D(void) // Protected
{
  // Nothing to be done
}

//-------- Changes -----------------------------------------------------------

bool // success
Mesh3D::applyChanges(bool checkConsistency)
{
if(checkConsistency&&
   (!checkVertexConsistency()||
    !checkNormalConsistency()||
    !checkTextureConsistency()||
    !checkSecondTextureConsistency()||
    !checkThirdTextureConsistency()||
    !checkFourthTextureConsistency()||
    !checkColorConsistency()))
  {
  return(false);
  }
_postGeometryRebuild();
return(true);
}

//-------- Vertices ----------------------------------------------------------

StlVector<Util3D::Dbl3> & // vertex coords
Mesh3D::accessVertices(void)
{
return(_vertices);
}

StlVector<Util3D::Face> & // vertex faces
Mesh3D::accessVertexFaces(void)
{
return(_facesV);
}

bool // success
Mesh3D::checkVertexConsistency(void) const
{
unsigned int nbV=_vertices.size();
for(unsigned int i=_facesV.size();i--;)
  {
  if((_facesV[i].p1>=nbV)||(_facesV[i].p2>=nbV)||(_facesV[i].p3>=nbV))
    {
    return(false);
    }
  }
return(true);
}

//-------- Normals -----------------------------------------------------------

StlVector<Util3D::Dbl3> & // normal coords
Mesh3D::accessNormals(void)
{
return(_normals);
}

StlVector<Util3D::Face> & // normal faces
Mesh3D::accessNormalFaces(void)
{
return(_facesN);
}

void
Mesh3D::computeDefaultNormals(bool smooth)
{
_computeDefaultNormals(smooth);
}

bool // success
Mesh3D::checkNormalConsistency(void) const
{
if(_normals.empty()&&!_vertices.empty())
  {
  if(_material.getLighting())
    {
    return(false);
    }
  }
else
  {
  const StlVector<Util3D::Face> & facesN=(_facesN.empty() ? _facesV
                                                          : _facesN);
  if(facesN.size()!=_facesV.size())
    {
    return(false);
    }
  unsigned int nbN=_normals.size();
  for(unsigned int i=facesN.size();i--;)
    {
    if((facesN[i].p1>=nbN)||(facesN[i].p2>=nbN)||(facesN[i].p3>=nbN))
      {
      return(false);
      }
    }
  }
return(true);
}

//-------- Textures ----------------------------------------------------------

StlVector<Util3D::Dbl2> & // texture coords
Mesh3D::accessTextures(void)
{
return(_textures);
}

StlVector<Util3D::Face> & // texture faces
Mesh3D::accessTextureFaces(void)
{
return(_facesT);
}

bool // success
Mesh3D::checkTextureConsistency(void) const
{
if(_textures.empty()&&!_vertices.empty())
  {
  if(_texture)
    {
    return(false);
    }
  }
else
  {
  const StlVector<Util3D::Face> & facesT=(_facesT.empty() ? _facesV
                                                          : _facesT);
  if(facesT.size()!=_facesV.size())
    {
    return(false);
    }
  unsigned int nbT=_textures.size();
  for(unsigned int i=facesT.size();i--;)
    {
    if((facesT[i].p1>=nbT)||(facesT[i].p2>=nbT)||(facesT[i].p3>=nbT))
      {
      return(false);
      }
    }
  }
return(true);
}

StlVector<Util3D::Dbl2> & // second texture coords
Mesh3D::accessSecondTextures(void)
{
return(_textures2);
}

StlVector<Util3D::Face> & // second texture faces
Mesh3D::accessSecondTextureFaces(void)
{
return(_facesT2);
}

bool // success
Mesh3D::checkSecondTextureConsistency(void) const
{
const StlVector<Util3D::Dbl2> & t2=(_textures2.empty() ? _textures
                                                       : _textures2);
if(t2.empty()&&!_vertices.empty())
  {
  if(_texture2)
    {
    return(false);
    }
  }
else
  {
  const StlVector<Util3D::Face> & facesT=(_facesT.empty() ? _facesV
                                                          : _facesT);
  const StlVector<Util3D::Face> & facesT2=(_facesT2.empty() ? facesT
                                                            : _facesT2);
  if(facesT2.size()!=_facesV.size())
    {
    return(false);
    }
  unsigned int nbT2=t2.size();
  for(unsigned int i=facesT2.size();i--;)
    {
    if((facesT2[i].p1>=nbT2)||(facesT2[i].p2>=nbT2)||(facesT2[i].p3>=nbT2))
      {
      return(false);
      }
    }
  }
return(true);
}

StlVector<Util3D::Dbl2> & // third texture coords
Mesh3D::accessThirdTextures(void)
{
return(_textures3);
}

StlVector<Util3D::Face> & // third texture faces
Mesh3D::accessThirdTextureFaces(void)
{
return(_facesT3);
}

bool // success
Mesh3D::checkThirdTextureConsistency(void) const
{
const StlVector<Util3D::Dbl2> & t3=(_textures3.empty() ? _textures
                                                       : _textures3);
if(t3.empty()&&!_vertices.empty())
  {
  if(_texture3)
    {
    return(false);
    }
  }
else
  {
  const StlVector<Util3D::Face> & facesT=(_facesT.empty() ? _facesV
                                                          : _facesT);
  const StlVector<Util3D::Face> & facesT3=(_facesT3.empty() ? facesT
                                                            : _facesT3);
  if(facesT3.size()!=_facesV.size())
    {
    return(false);
    }
  unsigned int nbT3=t3.size();
  for(unsigned int i=facesT3.size();i--;)
    {
    if((facesT3[i].p1>=nbT3)||(facesT3[i].p2>=nbT3)||(facesT3[i].p3>=nbT3))
      {
      return(false);
      }
    }
  }
return(true);
}

StlVector<Util3D::Dbl2> & // fourth texture coords
Mesh3D::accessFourthTextures(void)
{
return(_textures4);
}

StlVector<Util3D::Face> & // fourth texture faces
Mesh3D::accessFourthTextureFaces(void)
{
return(_facesT4);
}

bool // success
Mesh3D::checkFourthTextureConsistency(void) const
{
const StlVector<Util3D::Dbl2> & t4=(_textures4.empty() ? _textures
                                                       : _textures4);
if(t4.empty()&&!_vertices.empty())
  {
  if(_texture4)
    {
    return(false);
    }
  }
else
  {
  const StlVector<Util3D::Face> & facesT=(_facesT.empty() ? _facesV
                                                          : _facesT);
  const StlVector<Util3D::Face> & facesT4=(_facesT4.empty() ? facesT
                                                            : _facesT4);
  if(facesT4.size()!=_facesV.size())
    {
    return(false);
    }
  unsigned int nbT4=t4.size();
  for(unsigned int i=facesT4.size();i--;)
    {
    if((facesT4[i].p1>=nbT4)||(facesT4[i].p2>=nbT4)||(facesT4[i].p3>=nbT4))
      {
      return(false);
      }
    }
  }
return(true);
}

//-------- Colors ------------------------------------------------------------

StlVector<Util3D::Dbl3> & // color components
Mesh3D::accessColors(void)
{
return(_colors);
}

StlVector<Util3D::Face> & // color faces
Mesh3D::accessColorFaces(void)
{
return(_facesC);
}

bool // success
Mesh3D::checkColorConsistency(void) const
{
if(!_colors.empty())
  {
  const StlVector<Util3D::Face> & facesC=(_facesC.empty() ? _facesV
                                                          : _facesC);
  if(facesC.size()!=_facesV.size())
    {
    return(false);
    }
  unsigned int nbC=_colors.size();
  for(unsigned int i=facesC.size();i--;)
    {
    if((facesC[i].p1>=nbC)||(facesC[i].p2>=nbC)||(facesC[i].p3>=nbC))
      {
      return(false);
      }
    }
  }
return(true);
}

//-------- Serialization operations ------------------------------------------

void
Mesh3D::serialize(SerializationDependencies & dependenciesOut,
                  ArRef<AbstractOStream> output) const
{
Surface3D::serialize(dependenciesOut,output);
unsigned int i;
unsigned int nbVertices=_vertices.size(),nbFacesV=_facesV.size();
output->writeUInt(nbVertices);
output->writeUInt(nbFacesV);
for(i=0;i<nbVertices;i++)
  {
  output->writeDouble(_vertices[i].x);
  output->writeDouble(_vertices[i].y);
  output->writeDouble(_vertices[i].z);
  }
for(i=0;i<nbFacesV;i++)
  {
  output->writeUInt(_facesV[i].p1);
  output->writeUInt(_facesV[i].p2);
  output->writeUInt(_facesV[i].p3);
  }
unsigned int nbNormals=_normals.size(),nbFacesN=_facesN.size();
output->writeUInt(nbNormals);
output->writeUInt(nbFacesN);
for(i=0;i<nbNormals;i++)
  {
  output->writeDouble(_normals[i].x);
  output->writeDouble(_normals[i].y);
  output->writeDouble(_normals[i].z);
  }
for(i=0;i<nbFacesN;i++)
  {
  output->writeUInt(_facesN[i].p1);
  output->writeUInt(_facesN[i].p2);
  output->writeUInt(_facesN[i].p3);
  }
unsigned int nbTextures=_textures.size(),nbFacesT=_facesT.size();
output->writeUInt(nbTextures);
output->writeUInt(nbFacesT);
for(i=0;i<nbTextures;i++)
  {
  output->writeDouble(_textures[i].x);
  output->writeDouble(_textures[i].y);
  }
for(i=0;i<nbFacesT;i++)
  {
  output->writeUInt(_facesT[i].p1);
  output->writeUInt(_facesT[i].p2);
  output->writeUInt(_facesT[i].p3);
  }
unsigned int nbTextures2=_textures2.size(),nbFacesT2=_facesT2.size();
output->writeUInt(nbTextures2);
output->writeUInt(nbFacesT2);
for(i=0;i<nbTextures2;i++)
  {
  output->writeDouble(_textures2[i].x);
  output->writeDouble(_textures2[i].y);
  }
for(i=0;i<nbFacesT2;i++)
  {
  output->writeUInt(_facesT2[i].p1);
  output->writeUInt(_facesT2[i].p2);
  output->writeUInt(_facesT2[i].p3);
  }
unsigned int nbTextures3=_textures3.size(),nbFacesT3=_facesT3.size();
output->writeUInt(nbTextures3);
output->writeUInt(nbFacesT3);
for(i=0;i<nbTextures3;i++)
  {
  output->writeDouble(_textures3[i].x);
  output->writeDouble(_textures3[i].y);
  }
for(i=0;i<nbFacesT3;i++)
  {
  output->writeUInt(_facesT3[i].p1);
  output->writeUInt(_facesT3[i].p2);
  output->writeUInt(_facesT3[i].p3);
  }
unsigned int nbTextures4=_textures4.size(),nbFacesT4=_facesT4.size();
output->writeUInt(nbTextures4);
output->writeUInt(nbFacesT4);
for(i=0;i<nbTextures4;i++)
  {
  output->writeDouble(_textures4[i].x);
  output->writeDouble(_textures4[i].y);
  }
for(i=0;i<nbFacesT4;i++)
  {
  output->writeUInt(_facesT4[i].p1);
  output->writeUInt(_facesT4[i].p2);
  output->writeUInt(_facesT4[i].p3);
  }
unsigned int nbColors=_colors.size(),nbFacesC=_facesC.size();
output->writeUInt(nbColors);
output->writeUInt(nbFacesC);
for(i=0;i<nbColors;i++)
  {
  output->writeDouble(_colors[i].x);
  output->writeDouble(_colors[i].y);
  output->writeDouble(_colors[i].z);
  }
for(i=0;i<nbFacesC;i++)
  {
  output->writeUInt(_facesC[i].p1);
  output->writeUInt(_facesC[i].p2);
  output->writeUInt(_facesC[i].p3);
  }
}

bool // success
Mesh3D::unserialize(const SerializationDependencies & dependencies,
                    ArRef<AbstractIStream> input)
{
unsigned int i;
unsigned int nbVertices,nbFacesV; // vertices
StlVector<Util3D::Dbl3> vertices;
StlVector<Util3D::Face> facesV;
if(!Surface3D::unserialize(dependencies,input)||
   !input->readUInt(nbVertices)||
   !input->readUInt(nbFacesV))
  {
  return(false);
  }
vertices.resize(nbVertices);
for(i=0;i<nbVertices;i++)
  {
  if(!input->readDouble(vertices[i].x)||
     !input->readDouble(vertices[i].y)||
     !input->readDouble(vertices[i].z))
    {
    return(false);
    }
  }
facesV.resize(nbFacesV);
for(i=0;i<nbFacesV;i++)
  {
  if(!input->readUInt(facesV[i].p1)||
     !input->readUInt(facesV[i].p2)||
     !input->readUInt(facesV[i].p3))
    {
    return(false);
    }
  }
unsigned int nbNormals,nbFacesN; // normals
StlVector<Util3D::Dbl3> normals;
StlVector<Util3D::Face> facesN;
if(!input->readUInt(nbNormals)||
   !input->readUInt(nbFacesN))
  {
  return(false);
  }
normals.resize(nbNormals);
for(i=0;i<nbNormals;i++)
  {
  if(!input->readDouble(normals[i].x)||
     !input->readDouble(normals[i].y)||
     !input->readDouble(normals[i].z))
    {
    return(false);
    }
  }
facesN.resize(nbFacesN);
for(i=0;i<nbFacesN;i++)
  {
  if(!input->readUInt(facesN[i].p1)||
     !input->readUInt(facesN[i].p2)||
     !input->readUInt(facesN[i].p3))
    {
    return(false);
    }
  }
unsigned int nbTextures,nbFacesT; // texture
StlVector<Util3D::Dbl2> textures;
StlVector<Util3D::Face> facesT;
if(!input->readUInt(nbTextures)||
   !input->readUInt(nbFacesT))
  {
  return(false);
  }
textures.resize(nbTextures);
for(i=0;i<nbTextures;i++)
  {
  if(!input->readDouble(textures[i].x)||
     !input->readDouble(textures[i].y))
    {
    return(false);
    }
  }
facesT.resize(nbFacesT);
for(i=0;i<nbFacesT;i++)
  {
  if(!input->readUInt(facesT[i].p1)||
     !input->readUInt(facesT[i].p2)||
     !input->readUInt(facesT[i].p3))
    {
    return(false);
    }
  }
unsigned int nbTextures2,nbFacesT2; // second texture
StlVector<Util3D::Dbl2> textures2;
StlVector<Util3D::Face> facesT2;
if(!input->readUInt(nbTextures2)||
   !input->readUInt(nbFacesT2))
  {
  return(false);
  }
textures2.resize(nbTextures2);
for(i=0;i<nbTextures2;i++)
  {
  if(!input->readDouble(textures2[i].x)||
     !input->readDouble(textures2[i].y))
    {
    return(false);
    }
  }
facesT2.resize(nbFacesT2);
for(i=0;i<nbFacesT2;i++)
  {
  if(!input->readUInt(facesT2[i].p1)||
     !input->readUInt(facesT2[i].p2)||
     !input->readUInt(facesT2[i].p3))
    {
    return(false);
    }
  }
unsigned int nbTextures3,nbFacesT3; // third texture
StlVector<Util3D::Dbl2> textures3;
StlVector<Util3D::Face> facesT3;
if(!input->readUInt(nbTextures3)||
   !input->readUInt(nbFacesT3))
  {
  return(false);
  }
textures3.resize(nbTextures3);
for(i=0;i<nbTextures3;i++)
  {
  if(!input->readDouble(textures3[i].x)||
     !input->readDouble(textures3[i].y))
    {
    return(false);
    }
  }
facesT3.resize(nbFacesT3);
for(i=0;i<nbFacesT3;i++)
  {
  if(!input->readUInt(facesT3[i].p1)||
     !input->readUInt(facesT3[i].p2)||
     !input->readUInt(facesT3[i].p3))
    {
    return(false);
    }
  }
unsigned int nbTextures4,nbFacesT4; // fourth texture
StlVector<Util3D::Dbl2> textures4;
StlVector<Util3D::Face> facesT4;
if(!input->readUInt(nbTextures4)||
   !input->readUInt(nbFacesT4))
  {
  return(false);
  }
textures4.resize(nbTextures4);
for(i=0;i<nbTextures4;i++)
  {
  if(!input->readDouble(textures4[i].x)||
     !input->readDouble(textures4[i].y))
    {
    return(false);
    }
  }
facesT4.resize(nbFacesT4);
for(i=0;i<nbFacesT4;i++)
  {
  if(!input->readUInt(facesT4[i].p1)||
     !input->readUInt(facesT4[i].p2)||
     !input->readUInt(facesT4[i].p3))
    {
    return(false);
    }
  }
unsigned int nbColors,nbFacesC; // colors
StlVector<Util3D::Dbl3> colors;
StlVector<Util3D::Face> facesC;
if(!input->readUInt(nbColors)||
   !input->readUInt(nbFacesC))
  {
  return(false);
  }
colors.resize(nbColors);
for(i=0;i<nbColors;i++)
  {
  if(!input->readDouble(colors[i].x)||
     !input->readDouble(colors[i].y)||
     !input->readDouble(colors[i].z))
    {
    return(false);
    }
  }
facesC.resize(nbFacesC);
for(i=0;i<nbFacesC;i++)
  {
  if(!input->readUInt(facesC[i].p1)||
     !input->readUInt(facesC[i].p2)||
     !input->readUInt(facesC[i].p3))
    {
    return(false);
    }
  }
_vertices.swap(vertices);
_facesV.swap(facesV);
_normals.swap(normals);
_facesN.swap(facesN);
_textures.swap(textures);
_facesT.swap(facesT);
_textures2.swap(textures2);
_facesT2.swap(facesT2);
_textures3.swap(textures3);
_facesT3.swap(facesT3);
_textures4.swap(textures4);
_facesT4.swap(facesT4);
_colors.swap(colors);
_facesC.swap(facesC);
_postGeometryRebuild();
_postAppearanceRebuild();
return(true);
}

} // namespace AReVi

//----------------------------------------------------------------------------
