// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/metaball.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/metaball.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"
#include "AReVi/Lib3D/llw_priv.h"
#include "AReVi/Lib3D/texture.h"
#include <math.h>

#include "GL/gl.h"

namespace AReVi {

//#define DEBUG_METABALL
//#define COMPUTE_NORMALS
//#define COMPUTE_PRECISE_NORMALS

//          v7---------e6---------v6
//         /|                     /|
//        / |                    / |
//       /  |                   /  |
//      e11 |                 e10  |
//     /    |                 /    |
//    /     e7               /     e5
//   /      |               /      |
// v3---------e2-----------v2      |
//  |       |              |       |
//  |       |              |       |
//  |       |              |       |
//          v4---------e4----------v5
//  |      /               |      /
//  e3    /                e1    /
//  |    /                 |    /
//  |   e8                 |   e9
//  |  /                   |  /
//  | /                    | /
//  |/                     |/
// v0---------e0-----------v1
// Two tables below taken from following:
//    File Name: CIsoSurface.cpp
//    Last Modified: 5/8/2000
//    Author: Raghavendra Chandrashekara (based on source code provided
//    by Paul Bourke and Cory Gene Bloyd)
//    Email: rc99@doc.ic.ac.uk, rchandrashekara@hotmail.com
// Original code by Cory Gene Bloyd marked as public domain.
// File found at http://astronomy.swin.edu.au/pbourke/modelling/polygonise

static const int edgeTable[256] = {
        0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
        0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
        0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
        0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
        0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
        0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
        0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
        0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
        0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
        0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
        0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
        0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
        0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
        0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
        0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
        0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
        0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
        0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
        0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
        0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
        0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
        0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
        0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
        0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
        0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
        0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
        0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
        0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
        0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
        0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
        0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
        0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0
};


static const int triTable[256][16] = {
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
        {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
        {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
        {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
        {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
        {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
        {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
        {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
        {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
        {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
        {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
        {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
        {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
        {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
        {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
        {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
        {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
        {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
        {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
        {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
        {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
        {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
        {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
        {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
        {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
        {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
        {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
        {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
        {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
        {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
        {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
        {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
        {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
        {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
        {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
        {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
        {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
        {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
        {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
        {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
        {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
        {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
        {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
        {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
        {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
        {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
        {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
        {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
        {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
        {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
        {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
        {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
        {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
        {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
        {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
        {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
        {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
        {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
        {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
        {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
        {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
        {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
        {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
        {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
        {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
        {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
        {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
        {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
        {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
        {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
        {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
        {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
        {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
        {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
        {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
        {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
        {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
        {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
        {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
        {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
        {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
        {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
        {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
        {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
        {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
        {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
        {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
        {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
        {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
        {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
        {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
        {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
        {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
        {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
        {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
        {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
        {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
        {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
        {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
        {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
        {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
        {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
        {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
        {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
        {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
        {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
        {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
        {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
        {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
        {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
        {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
        {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
        {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
        {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
        {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
        {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
        {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
        {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
        {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
        {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
        {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
        {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
        {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
        {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
        {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
        {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
        {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
        {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
        {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
        {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
        {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
        {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
        {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
        {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
        {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
        {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
        {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
        {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
        {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
        {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
        {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
        {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
        {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
        {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
        {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
        {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
        {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
        {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
        {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
        {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
        {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
        {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
        {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
        {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};

//----------------------------------------------------------------------------
// Class Metaball
//----------------------------------------------------------------------------

AR_CLASS_DEF(Metaball,Surface3D)

//-------- Construction / Destruction ----------------------------------------

Metaball::Metaball(ArCW & arCW) // Protected
: Surface3D(arCW),
  _isolevel(0.45),
  _ccw(true),
  _smooth(true),
  _precision(0),
  _fieldFunction(Metaball::FIELD_FUNCTION_SOFTOBJECTS),
  _metaballs()
{
_useDisplayList=false;
//setDoubleSided(true);
/*  _normals.push_back(Util3D::Dbl3(+0.0,-1.0,+0.0));
  _normals.push_back(Util3D::Dbl3(+1.0,+0.0,+0.0));
  _normals.push_back(Util3D::Dbl3(+0.0,+1.0,+0.0));
  _normals.push_back(Util3D::Dbl3(-1.0,+0.0,+0.0));
  _normals.push_back(Util3D::Dbl3(+0.0,+0.0,+1.0));
  _normals.push_back(Util3D::Dbl3(+0.0,+0.0,-1.0));*/
setPrecision(30);
}

Metaball::~Metaball(void) // Protected
{
  // Nothing to be done
}

//-------- Changes -----------------------------------------------------------

bool // success
Metaball::applyChanges(void)
{
_postGeometryRebuild();
return true;
}

//-------- Geometry ----------------------------------------------------------

const Metaball::MetaballInfo &
Metaball::getMetaball(unsigned int index) const
{
return _metaballs[index];
}

Metaball::MetaballInfo &
Metaball::accessMetaball(unsigned int index)
{
return _metaballs[index];
}

unsigned int
Metaball::getNbMetaballs(void) const
{
return _metaballs.size();
}


void
Metaball::setCCW(bool ccw)
{
_ccw=ccw;
_postGeometryRebuild();
}

bool // counter clock wise surface
Metaball::getCCW(void) const
{
return(_ccw);
}

void
Metaball::setSmoothed(bool smooth)
{
_smooth=smooth;
_postGeometryRebuild();
}

bool // faces normals are smoothed
Metaball::getSmoothed(void) const
{
return(_smooth);
}


bool // success
Metaball::setIsoLevel(double isolevel)
{
_isolevel=isolevel;
_postGeometryRebuild();
return(true);
}

void
Metaball::addMetaball(const MetaballInfo &metaball)
{
_metaballs.push_back(metaball);

_postGeometryRebuild();
}

void
Metaball::removeMetaball(unsigned int index)
{
StlVectorFastErase(_metaballs,index);
_postGeometryRebuild();
}


double // isolevel
Metaball::getIsoLevel(void) const
{
return(_isolevel);
}

bool // success
Metaball::setPrecision(unsigned int precision)
{
if(precision)
  {
  _precision=precision;
  _sideSize=_precision+1;
  _cubeFields.resize(_sideSize*_sideSize*_sideSize);
  _cubePoints.resize(_sideSize*_sideSize*_sideSize);

#ifdef COMPUTE_NORMALS
  _cubeDFields.resize(_sideSize*_sideSize*_sideSize);
#endif

  _zeverts.resize(_sideSize*_sideSize*_sideSize);
  _yeverts.resize(_sideSize*_sideSize*_sideSize);
  _xeverts.resize(_sideSize*_sideSize*_sideSize);

  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // precision
Metaball::getPrecision(void) const
{return(_precision);
}

void
Metaball::setFieldFunction(Metaball::FieldFunction function)
{
_fieldFunction=function;
_postGeometryRebuild();
}

Metaball::FieldFunction // Select Field Fonctions
Metaball::getFieldFunction(void) const
{
return _fieldFunction;
}

//-------- Serialization operations ------------------------------------------

void
Metaball::serialize(SerializationDependencies & dependenciesOut,
                    ArRef<AbstractOStream> output) const
{
Surface3D::serialize(dependenciesOut,output);
output->writeDouble(_isolevel);
output->writeUInt(_precision);
output->writeUInt((int)_fieldFunction);
unsigned int flags=(_smooth ? 1<<0 : 0)+
                   (_ccw    ? 1<<1 : 0);
output->writeUInt(flags);

// Write metaballs ...
output->writeUInt(_metaballs.size());

for (unsigned i=_metaballs.size();i--;)
  {
  output->writeDouble(_metaballs[i].radius);
  output->writeDouble(_metaballs[i].innerSize);
  output->writeDouble(_metaballs[i].potential);
  output->writeDouble(_metaballs[i].position.x);
  output->writeDouble(_metaballs[i].position.y);
  output->writeDouble(_metaballs[i].position.z);
  output->writeUInt((int)_metaballs[i].type);
  }
}

bool // success
Metaball::unserialize(const SerializationDependencies & dependencies,
                      ArRef<AbstractIStream> input)
{
double isolevel;
unsigned int precision,flags, fieldFunction, nbMetaballs;
StlVector<MetaballInfo> metaballs;
if(!Surface3D::unserialize(dependencies,input)||
   !input->readDouble(isolevel)||
   !input->readUInt(precision)||
   !input->readUInt(fieldFunction)||
   !input->readUInt(flags) ||
   !input->readUInt(nbMetaballs))
  {
  return(false);
  }

// Read metaballs ...
metaballs.resize(nbMetaballs);

for (unsigned i=metaballs.size();i--;)
  {
  unsigned int type;
  if (!input->readDouble(metaballs[i].radius) ||
      !input->readDouble(metaballs[i].innerSize) ||
      !input->readDouble(metaballs[i].potential) ||
      !input->readDouble(metaballs[i].position.x) ||
      !input->readDouble(metaballs[i].position.y) ||
      !input->readDouble(metaballs[i].position.z) ||
      !input->readUInt(type))
    {
    return(false);
    }
  metaballs[i].type=(Metaball::MetaballType)type;
  }

_isolevel=isolevel;
_smooth=flags&(1<<0);
_ccw=flags&(1<<1);
_fieldFunction=(Metaball::FieldFunction)fieldFunction;
_metaballs=metaballs;
setPrecision(precision);
_postGeometryRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------


bool
Metaball::_rebuildGeometry(void) // Protected
{
_vertices.clear();
_facesV.clear();

if (_metaballs.empty())
  {
  return false;
  }

// Eval bouding box
_minX=(_metaballs[0].position.x-_metaballs[0].radius);
_minY=(_metaballs[0].position.y-_metaballs[0].radius);
_minZ=(_metaballs[0].position.z-_metaballs[0].radius);
_maxX=(_metaballs[0].position.x+_metaballs[0].radius);
_maxY=(_metaballs[0].position.y+_metaballs[0].radius);
_maxZ=(_metaballs[0].position.z+_metaballs[0].radius);

for (int index=_metaballs.size();(index--)>1; )
  {
  MetaballInfo &metaball(_metaballs[index]);

  if (metaball.position.x-metaball.radius<_minX)
    _minX=metaball.position.x-metaball.radius;
  if (metaball.position.y-metaball.radius<_minY)
    _minY=metaball.position.y-metaball.radius;
  if (metaball.position.z-metaball.radius<_minZ)
    _minZ=metaball.position.z-metaball.radius;

  if (metaball.position.x+metaball.radius>_maxX)
    _maxX=metaball.position.x+metaball.radius;
  if (metaball.position.y+metaball.radius>_maxY)
    _maxY=metaball.position.y+metaball.radius;
  if (metaball.position.z+metaball.radius>_maxZ)
    _maxZ=metaball.position.z+metaball.radius;
  }

// Eval Fields in all bouding box
double stepX((_maxX-_minX)/((double)_precision));
double stepY((_maxY-_minY)/((double)_precision));
double stepZ((_maxZ-_minZ)/((double)_precision));

if (_fieldFunction==FIELD_FUNCTION_SOFTOBJECTS)
  {
  for (unsigned x=0,index=0;x<_sideSize;x++)
    for (unsigned y=0;y<_sideSize;y++)
      for (unsigned z=0;z<_sideSize;z++,index++)
        {
        _cubePoints[index].x=_minX+(x*stepX);
        _cubePoints[index].y=_minY+(y*stepY);
        _cubePoints[index].z=_minZ+(z*stepZ);
        _cubeFields[index]=_evalFieldSoftObjects(_cubePoints[index].x,_cubePoints[index].y,_cubePoints[index].z);
        }
  }
else
  {
  for (unsigned x=0,index=0;x<_sideSize;x++)
    for (unsigned y=0;y<_sideSize;y++)
      for (unsigned z=0;z<_sideSize;z++,index++)
        {
        _cubePoints[index].x=_minX+(x*stepX);
        _cubePoints[index].y=_minY+(y*stepY);
        _cubePoints[index].z=_minZ+(z*stepZ);
        _cubeFields[index]=_evalFieldLinear(_cubePoints[index].x,_cubePoints[index].y,_cubePoints[index].z);
        }
  }

// Process all cubies, processed in order to no-recompute common edges intersections.
for (unsigned x=0,index=0;x<_precision;x++,index+=_sideSize)
  for (unsigned y=0;y<_precision;y++,index++)
    for (unsigned z=0;z<_precision;z++,index++)
      _processCubie(x,y,z,index);


_computeIsoSurfaceNormals(_smooth);
//cerr << "NbV=" << _vertices.size() << "   NbF=" << _facesV.size() << "   NbN=" << _normals.size() << "   NbFN=" << _facesN.size() << endl ;
return true;
}


//    Linearly interpolate the position where an isosurface cuts
//    an edge between two vertices, each with their own scalar value
void
vertexInterp(double isolevel,const Point3d &p1,const Point3d &p2,double valp1,double valp2, Util3D::Dbl3 &pointOut)
{
double mu;

if (fabs(isolevel-valp1) < 0.00001)
  {
  pointOut.x=p1.x;
  pointOut.y=p1.y;
  pointOut.z=p1.z;
  return;
  }

if (fabs(isolevel-valp2) < 0.00001)
  {
  pointOut.x=p2.x;
  pointOut.y=p2.y;
  pointOut.z=p2.z;
  return;
  }
if (fabs(valp1-valp2) < 0.00001)
  {
  pointOut.x=p1.x;
  pointOut.y=p1.y;
  pointOut.z=p1.z;
  return;
  }
mu = (isolevel - valp1) / (valp2 - valp1);
pointOut.x = p1.x + (mu * (p2.x - p1.x));
pointOut.y = p1.y + (mu * (p2.y - p1.y));
pointOut.z = p1.z + (mu * (p2.z - p1.z));
}

void
Metaball::_processCubie(unsigned xsub,unsigned  ysub,unsigned  zsub,unsigned  index)
{
static double cubePotentials[8];
static Point3d *cubePoints[8];
static Util3D::Dbl3 vertlist[12];
static int vertices[12];

unsigned squareSize(_sideSize*_sideSize);
unsigned cubeindex(0);
// Backup cube Potentials
cubePotentials[0]=_cubeFields[index];
cubePotentials[1]=_cubeFields[index+_sideSize];
cubePotentials[2]=_cubeFields[index+_sideSize+1];
cubePotentials[3]=_cubeFields[index+1];
cubePotentials[4]=_cubeFields[index+squareSize];
cubePotentials[5]=_cubeFields[index+squareSize+_sideSize];
cubePotentials[6]=_cubeFields[index+squareSize+_sideSize+1];
cubePotentials[7]=_cubeFields[index+squareSize+1];

// Backup cube Points
cubePoints[0]=&_cubePoints[index];
cubePoints[1]=&_cubePoints[index+_sideSize];
cubePoints[2]=&_cubePoints[index+_sideSize+1];
cubePoints[3]=&_cubePoints[index+1];
cubePoints[4]=&_cubePoints[index+squareSize];
cubePoints[5]=&_cubePoints[index+squareSize+_sideSize];
cubePoints[6]=&_cubePoints[index+squareSize+_sideSize+1];
cubePoints[7]=&_cubePoints[index+squareSize+1];

// Determine the index into the edge table which
// tells us which vertices are inside of the surface
if (cubePotentials[0] < _isolevel) cubeindex |= 1;
if (cubePotentials[1] < _isolevel) cubeindex |= 2;
if (cubePotentials[2] < _isolevel) cubeindex |= 4;
if (cubePotentials[3] < _isolevel) cubeindex |= 8;
if (cubePotentials[4] < _isolevel) cubeindex |= 16;
if (cubePotentials[5] < _isolevel) cubeindex |= 32;
if (cubePotentials[6] < _isolevel) cubeindex |= 64;
if (cubePotentials[7] < _isolevel) cubeindex |= 128;


//  Cube is entirely in/out of the surface
if (edgeTable[cubeindex] == 0)
  return;

//  Find the vertices where the surface intersects the cube
unsigned edgeIndex=edgeTable[cubeindex];

// For efficiency: compute only needed intersections
if (edgeIndex & 8)  // edge  3: xyz(0)-Xyz(3)
  {
  if (xsub==0 && ysub==0)
    {
    vertexInterp(_isolevel,*cubePoints[3],*cubePoints[0],cubePotentials[3],cubePotentials[0],vertlist[3]);

    _zeverts[index]=_vertices.size();
    vertices[3]=_zeverts[index];
    _vertices.push_back(Util3D::Dbl3(vertlist[3].x,vertlist[3].y,vertlist[3].z));
    _bbox.addXYZ(vertlist[3].x,vertlist[3].y,vertlist[3].z);
    }
  else
    {
    vertices[3]=_zeverts[index];
    assert(_zeverts[index]!=-1);
    vertlist[3]=_vertices[vertices[3]];
    }
  }

if (edgeIndex & 1)  // edge  0: xyz(0)-xYz(1)
  {
  if (xsub==0 && zsub==0)
    {
    vertexInterp(_isolevel,*cubePoints[0],*cubePoints[1],cubePotentials[0],cubePotentials[1],vertlist[0]);
    _yeverts[index]=_vertices.size();
    vertices[0]=_yeverts[index];
    _vertices.push_back(vertlist[0]);
    _bbox.addXYZ(vertlist[0].x,vertlist[0].y,vertlist[0].z);
    }
  else
    {
    vertices[0]=_yeverts[index];
    assert(_yeverts[index]!=-1);
    vertlist[0]=_vertices[vertices[0]];
    }
  }

if (edgeIndex &  256)  // edge  8: xyz(0)-xyZ(4)
  {
  if (ysub==0 && zsub==0)
    {
    vertexInterp(_isolevel,*cubePoints[0],*cubePoints[4],cubePotentials[0],cubePotentials[4],vertlist[8]);
    _xeverts[index]=_vertices.size();
    vertices[8]=_vertices.size();
    _vertices.push_back(vertlist[8]);
    _bbox.addXYZ(vertlist[8].x,vertlist[8].y,vertlist[8].z);
    }
  else
    {
    vertices[8]=_xeverts[index];
    assert(_xeverts[index]!=-1);
    vertlist[8] = _vertices[vertices[8]];
    }
  }

if (edgeIndex &  512)  // edge  9: xYz(1)-xYZ(5)
  {
  if (zsub==0)
    {
    vertexInterp(_isolevel,*cubePoints[1],*cubePoints[5],cubePotentials[1],cubePotentials[5],vertlist[9]);
    _xeverts[index+_sideSize]=_vertices.size();
    vertices[9]=_vertices.size();
    _vertices.push_back(vertlist[9]);
    _bbox.addXYZ(vertlist[9].x,vertlist[9].y,vertlist[9].z);
    }
  else
    {
    vertices[9]=_xeverts[index+_sideSize];
    assert(vertices[9]!=-1);
    vertlist[9] = _vertices[vertices[9]];
    }
  }
if (edgeIndex &   16)  // edge  4: xyZ(4)-xYZ(5)
  {
  if (zsub==0)
    {
    vertexInterp(_isolevel,*cubePoints[4],*cubePoints[5],cubePotentials[4],cubePotentials[5],vertlist[4]);
    _yeverts[index+squareSize]=_vertices.size();
    vertices[4]=_vertices.size();
      _vertices.push_back(vertlist[4]);
      _bbox.addXYZ(vertlist[4].x,vertlist[4].y,vertlist[4].z);
    }
  else
    {
    vertices[4]=_yeverts[index+squareSize];
    assert(vertices[4]!=-1);
    vertlist[4] =_vertices[vertices[4]];
    }
  }

if (edgeIndex & 2048)  // edge 11: Xyz(3)-XyZ(7)
  {
  if (ysub==0)
    {
    vertexInterp(_isolevel,*cubePoints[3],*cubePoints[7],cubePotentials[3],cubePotentials[7],vertlist[11]);
    _xeverts[index+1]=_vertices.size();
    vertices[11]=_vertices.size();
    _vertices.push_back(vertlist[11]);
    _bbox.addXYZ(vertlist[11].x,vertlist[11].y,vertlist[11].z);
    }
  else
    {
    vertices[11]=_xeverts[index+1];
    assert(vertices[11]!=-1);
    vertlist[11] = _vertices[vertices[11]];
    }
  }
if (edgeIndex &  128)  // edge  7: xyZ(4)-XyZ(7)
  {
  if (ysub == 0)
    {
    vertexInterp(_isolevel,*cubePoints[7],*cubePoints[4],cubePotentials[7],cubePotentials[4],vertlist[7]);
    _zeverts[index+squareSize]=_vertices.size();
    vertices[7]=_vertices.size();
    _vertices.push_back(vertlist[7]);
      _bbox.addXYZ(vertlist[7].x,vertlist[7].y,vertlist[7].z);
    }
  else
    {
    vertices[7]=_zeverts[index+squareSize];
    assert(vertices[7]!=-1);
    vertlist[7] = _vertices[vertices[7]];
    }
  }

if (edgeIndex &    4)  // edge  2: Xyz(3)-XYz(2)
  {
  if (zsub == 0 && xsub==0)
    {
    vertexInterp(_isolevel,*cubePoints[2],*cubePoints[3],cubePotentials[2],cubePotentials[3],vertlist[2]);
    _yeverts[index+1]=_vertices.size();
    vertices[2]=_vertices.size();
    _vertices.push_back(vertlist[2]);
      _bbox.addXYZ(vertlist[2].x,vertlist[2].y,vertlist[2].z);
    }
  else
    {
    vertices[2]=_yeverts[index+1];
    assert(vertices[2]!=-1);
    vertlist[2] = _vertices[vertices[2]];
    }
  }
if (edgeIndex &    2)  // edge  1: xYz(1)-XYz(2)
  {
  if (xsub==0)
    {
    vertexInterp(_isolevel,*cubePoints[1],*cubePoints[2],cubePotentials[1],cubePotentials[2],vertlist[1]);
    _zeverts[index+_sideSize]=_vertices.size();
    vertices[1]=_vertices.size();
    _vertices.push_back(vertlist[1]);
      _bbox.addXYZ(vertlist[1].x,vertlist[1].y,vertlist[1].z);
    }
  else
    {
    vertices[1]=_zeverts[index+_sideSize];
    assert(vertices[1]!=-1);
    vertlist[1] = _vertices[vertices[1]];
    }
  }

if (edgeIndex & 1024)  // edge 10: XYz(2)-XYZ(6)
  {
  vertexInterp(_isolevel,*cubePoints[2],*cubePoints[6],cubePotentials[2],cubePotentials[6],vertlist[10]);
  _xeverts[index+1+_sideSize]=_vertices.size();
  vertices[10]=_vertices.size();
  _vertices.push_back(vertlist[10]);
  _bbox.addXYZ(vertlist[10].x,vertlist[10].y,vertlist[10].z);
  }

if (edgeIndex &   64)  // edge  6: XyZ(7)-XYZ(6)
  {
  vertexInterp(_isolevel,*cubePoints[6],*cubePoints[7],cubePotentials[6],cubePotentials[7],vertlist[6]);
  _yeverts[index+squareSize+1]=_vertices.size();
  vertices[6]=_vertices.size();
  _vertices.push_back(vertlist[6]);
  _bbox.addXYZ(vertlist[6].x,vertlist[6].y,vertlist[6].z);
  }

if (edgeIndex &   32)  // edge  5: xYZ(5)-XYZ(6)
  {
  vertexInterp(_isolevel,*cubePoints[5],*cubePoints[6],cubePotentials[5],cubePotentials[6],vertlist[5]);
  _zeverts[index+_sideSize+squareSize]=_vertices.size();
  vertices[5]=_vertices.size();
  _vertices.push_back(vertlist[5]);
  _bbox.addXYZ(vertlist[5].x,vertlist[5].y,vertlist[5].z);
  }

// Add faces
if (_ccw)
  for (unsigned i=0;triTable[cubeindex][i]!=-1;i+=3)
    {
    _facesV.push_back(Util3D::Face(vertices[triTable[cubeindex][i  ]],
                                  vertices[triTable[cubeindex][i+1]],
                                  vertices[triTable[cubeindex][i+2]]));
    }
else
  for (unsigned i=0;triTable[cubeindex][i]!=-1;i+=3)
    {
    _facesV.push_back(Util3D::Face(vertices[triTable[cubeindex][i+2]],
                                  vertices[triTable[cubeindex][i+1]],
                                  vertices[triTable[cubeindex][i  ]]));
    }
}

double
Metaball::_evalFieldLinear(double x,double y,double z)
{
double f = 0.0;

StlVector<MetaballInfo>::const_iterator it=  _metaballs.begin();

while(it!=_metaballs.end())
  {
  double dx = x-it->position.x;
  double dy = y-it->position.y;
  double dz = z-it->position.z;

  if (it->type==METABALL_SPHERE)
    {
    double dist = ((dx*dx)+(dy*dy)+(dz*dz));
    if (dist<=it->radius*it->radius)
      f += it->potential * ((it->radius-sqrt(dist))/it->radius);
    }
  else if (it->type==METABALL_CUBE)
    {
    dx*=dx;
    dy*=dy;
    dz*=dz;

    double d4((dx*dx)+(dy*dy)+(dz*dz));
    double r2(it->radius*it->radius);
    double r4(r2*r2);

    if (d4<=r4)
      {
      f += it->potential * ((r4-(d4))/r4);// TODO : Verify that is correct (or do sqrt(sqrt(d4))/radius!
      }
    }
  else if (it->type==METABALL_TORUS)
    {
    double innerR2(it->innerSize*it->innerSize);
    double distInXZ = sqrt(dx*dx + dz*dz) - it->radius + it->innerSize;

    double distInY  = dy;
    double dist_from_track_squared = ((distInXZ*distInXZ) + (distInY*distInY));

    if (dist_from_track_squared < innerR2)
      {
      f+=it->potential * ((innerR2-dist_from_track_squared)/innerR2);
      }
    }
  it++;
  }


return f;
}

double
Metaball::_evalFieldSoftObjects(double x,double y,double z)
{
double f = 0.0;

StlVector<MetaballInfo>::const_iterator it=  _metaballs.begin();

while(it!=_metaballs.end())
  {
  double dx = x-it->position.x;
  double dy = y-it->position.y;
  double dz = z-it->position.z;

  if (it->type==METABALL_SPHERE)
    {
    double d2((dx*dx)+(dy*dy)+(dz*dz));
    double r2(it->radius*it->radius);
    if (d2<=r2)
      {
      double ratio2=d2/r2;
      double ratio4=ratio2*ratio2;
      double ratio6=ratio2*ratio4;

      f += it->potential * (1.0-((4.0/9.0)*ratio6)+((17.0/9.0)*ratio4)-((22.0/9.0)*ratio2));
      }
    }
  else if (it->type==METABALL_CUBE)
    {
    dx*=dx;
    dy*=dy;
    dz*=dz;

    double d4((dx*dx)+(dy*dy)+(dz*dz));
    double r2(it->radius*it->radius);
    double r4(r2*r2);

    if (d4<=r4)
      {
      double ratio2=sqrt(d4)/r2;
      double ratio4=d4/r4;
      double ratio6=ratio2*ratio4;
      f += it->potential * (1.0-((4.0/9.0)*ratio6)+((17.0/9.0)*ratio4)-((22.0/9.0)*ratio2));
      }
    }
  else if (it->type==METABALL_TORUS)
    {
    double innerR2(it->innerSize*it->innerSize);
    double distInXZ = sqrt(dx*dx + dz*dz) - it->radius + it->innerSize;
    double distInY  = dy;
    double dist_from_track_squared = ((distInXZ*distInXZ) + (distInY*distInY));

    if (dist_from_track_squared < innerR2)
      {
      double ratio2=(dist_from_track_squared/innerR2);
      double ratio4=ratio2*ratio2;
      double ratio6=ratio2*ratio4;
      f += it->potential * (1.0-((4.0/9.0)*ratio6)+((17.0/9.0)*ratio4)-((22.0/9.0)*ratio2));

      }
    }
  it++;
  }

return f;
}

void
Metaball::_computeIsoSurfaceNormals(bool smooth)
{
_normals.clear();
_facesN.clear();
if(smooth)
  {
  unsigned int i;
  unsigned int nb=_facesV.size();
  Util3D::Dbl3 * tmpN=ArSystem::allocate<Util3D::Dbl3>(nb);
  for(i=nb;i--;)
    {
    float dx1=_vertices[_facesV[i].p2].x-_vertices[_facesV[i].p1].x;
    float dy1=_vertices[_facesV[i].p2].y-_vertices[_facesV[i].p1].y;
    float dz1=_vertices[_facesV[i].p2].z-_vertices[_facesV[i].p1].z;
    float dx2=_vertices[_facesV[i].p3].x-_vertices[_facesV[i].p1].x;
    float dy2=_vertices[_facesV[i].p3].y-_vertices[_facesV[i].p1].y;
    float dz2=_vertices[_facesV[i].p3].z-_vertices[_facesV[i].p1].z;
    float dx3=dy1*dz2-dz1*dy2;
    float dy3=dx2*dz1-dz2*dx1;
    float dz3=dx1*dy2-dy1*dx2;
    float r=sqrt(dx3*dx3+dy3*dy3+dz3*dz3);
    r=(r? 1.0/r : 1.0);
    tmpN[i].x=dx3*r;
    tmpN[i].y=dy3*r;
    tmpN[i].z=dz3*r;
    }
  StlVector<unsigned int> * usedNormals;
  usedNormals=new StlVector<unsigned int>[_vertices.size()];
  for(i=nb;i--;)
    {
    usedNormals[_facesV[i].p1].push_back(i);
    usedNormals[_facesV[i].p2].push_back(i);
    usedNormals[_facesV[i].p3].push_back(i);
    }
  nb=_vertices.size();
  _normals.reserve(nb);
  _textures.resize(nb);
  double dx(0.5/(_maxX-_minX)),dy(0.5/(_maxY-_minY)),dz(0.5/(_maxZ-_minZ)),z;
  for(i=0;i<nb;i++)
    {
    Util3D::Dbl3 norm(0.0,0.0,0.0);
    for(unsigned int j=usedNormals[i].size();j--;)
      {
      norm.x+=tmpN[usedNormals[i][j]].x;
      norm.y+=tmpN[usedNormals[i][j]].y;
      norm.z+=tmpN[usedNormals[i][j]].z;
      }
    _normals.push_back(norm);

    z=(_vertices[i].z-_minZ)*dz;
    _textures[i].x=((_vertices[i].x-_minX)*dx)+z;
    _textures[i].y=((_vertices[i].y-_minY)*dy)+z;

    }
  delete[] usedNormals;
  ArSystem::deallocate(tmpN);
  // _facesV will be used instead of _facesN
  }
else
  {
  unsigned int nb=_facesV.size();
  _normals.reserve(nb);
  _facesN.reserve(nb);
  for(unsigned int i=0;i<nb;i++)
    {
    float dx1=_vertices[_facesV[i].p2].x-_vertices[_facesV[i].p1].x;
    float dy1=_vertices[_facesV[i].p2].y-_vertices[_facesV[i].p1].y;
    float dz1=_vertices[_facesV[i].p2].z-_vertices[_facesV[i].p1].z;
    float dx2=_vertices[_facesV[i].p3].x-_vertices[_facesV[i].p1].x;
    float dy2=_vertices[_facesV[i].p3].y-_vertices[_facesV[i].p1].y;
    float dz2=_vertices[_facesV[i].p3].z-_vertices[_facesV[i].p1].z;
    _normals.push_back(Util3D::Dbl3(dy1*dz2-dz1*dy2,
                                      dx2*dz1-dz2*dx1,
                                      dx1*dy2-dy1*dx2));
    _facesN.push_back(Util3D::Face(i,i,i));
    }

  nb=_vertices.size();
  _textures.reserve(nb);
  for(unsigned int i=0;i<nb;i++)
    _textures.push_back(Util3D::Dbl2((_vertices[i].x-_minX)/(_maxX-_minX),(_vertices[i].y-_minY)/(_maxY-_minY)));

  }
}


void
Metaball::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  _rebuildGeometry();
  }
Surface3D::_rebuildIfNeeded();
}

} // namespace AReVi

//----------------------------------------------------------------------------
