// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/points3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/pointSet3D.h"

#include "AReVi/Lib3D/gl_priv.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class PointSet3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(PointSet3D,ShapePart3D)

//-------- Construction / Destruction ----------------------------------------

PointSet3D::PointSet3D(ArCW & arCW) // Protected
: ShapePart3D(arCW),
  _thickness(1),
  _vertices(),
  _colors()
{
_describeGeometry(_bbox,
                  (const Util3D::Dbl3 *)0,0,
                  (const Util3D::Face *)0,0,
                  0,0);
}

PointSet3D::~PointSet3D(void) // Protected
{
  // Nothing to be done
}

//-------- Changes -----------------------------------------------------------

bool // success
PointSet3D::applyChanges(bool checkConsistency)
{
if(checkConsistency)
  {
  unsigned int nbV=_vertices.size();
  unsigned int nbC=_colors.size();
  if(nbV&&(nbC>1)&&(nbC<nbV))
    {
    return(false);
    }
  }
_postGeometryRebuild();
return(true);
}

//-------- Appearance --------------------------------------------------------

bool // success
PointSet3D::setThickness(unsigned int thickness)
{
if(thickness>0)
  {
  _thickness=thickness;
  _postAppearanceRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // points thickness
PointSet3D::getThickness(void) const
{
return(_thickness);
}

//-------- Vertices ----------------------------------------------------------

const StlVector<Util3D::Dbl3> & // vertex coords
PointSet3D::getVertices(void) const
{
return(_vertices);
}

StlVector<Util3D::Dbl3> & // vertex coords
PointSet3D::accessVertices(void)
{
return(_vertices);
}

//-------- Colors ------------------------------------------------------------

const StlVector<Util3D::Dbl3> & // color components
PointSet3D::getColors(void) const
{
return(_colors);
}

StlVector<Util3D::Dbl3> & // color components
PointSet3D::accessColors(void)
{
return(_colors);
}


//-------- Serialization operations ------------------------------------------

void
PointSet3D::serialize(SerializationDependencies & dependenciesOut,
                      ArRef<AbstractOStream> output) const
{
ShapePart3D::serialize(dependenciesOut,output);
output->writeUInt(_thickness);
unsigned int nbVertices=_vertices.size(),nbColors=_colors.size();
output->writeUInt(nbVertices);
output->writeUInt(nbColors);
unsigned int i;
for(i=0;i<nbVertices;i++)
  {
  output->writeDouble(_vertices[i].x);
  output->writeDouble(_vertices[i].y);
  output->writeDouble(_vertices[i].z);
  }
for(i=0;i<nbColors;i++)
  {
  output->writeDouble(_colors[i].x);
  output->writeDouble(_colors[i].y);
  output->writeDouble(_colors[i].z);
  }
}

bool // success
PointSet3D::unserialize(const SerializationDependencies & dependencies,
                        ArRef<AbstractIStream> input)
{
unsigned int i,thickness,nbVertices,nbColors;
StlVector<Util3D::Dbl3> vertices;
StlVector<Util3D::Dbl3> colors;
if(!ShapePart3D::unserialize(dependencies,input)||
   !input->readUInt(thickness)||
   !input->readUInt(nbVertices)||
   !input->readUInt(nbColors))
  {
  return(false);
  }
vertices.resize(nbVertices);
for(i=0;i<nbVertices;i++)
  {
  if(!input->readDouble(vertices[i].x)||
     !input->readDouble(vertices[i].y)||
     !input->readDouble(vertices[i].z))
    {
    return(false);
    }
  }
colors.resize(nbColors);
for(i=0;i<nbColors;i++)
  {
  if(!input->readDouble(colors[i].x)||
     !input->readDouble(colors[i].y)||
     !input->readDouble(colors[i].z))
    {
    return(false);
    }
  }
_thickness=thickness;
_vertices.swap(vertices);
_colors.swap(colors);
_postAppearanceRebuild();
_postGeometryRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
PointSet3D::_setAppearance(void) const // Protected
{
if(_offset)
  {
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(_offset,_offset);
  }
glDisable(GL_LIGHTING);
glPointSize(_thickness);
}

void
PointSet3D::_resetAppearance(void) const // Protected
{
glPointSize(1.0);
glEnable(GL_LIGHTING);
if(_offset)
  {
  glDisable(GL_POLYGON_OFFSET_FILL);
  }
}

void
PointSet3D::_drawGeometry(void) const // Protected
{
unsigned int nbV=_vertices.size();
if(nbV)
  {
  unsigned int nbC=_colors.size();
  switch(nbC)
    {
    case 0:
      glColor3f(0.8,0.8,0.8); // default color
      break;
    case 1:
      glColor3dv(&_colors.front().x); // common color
      break;
    }
  bool hasColors=(nbC>1);
  glBegin(GL_POINTS);
  for(unsigned int i=nbV;i--;)
    {
    if(hasColors)
      {
      glColor3dv(&_colors[i].x);
      }
    Util3D::Dbl3 vertex=_vertices[i];
    glVertex3dv(&vertex.x);
    }
  glEnd(); // GL_POINTS
  }
}

void
PointSet3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_geometryList,GL_COMPILE);
    _drawGeometry();
    glEndList();
    }
  Util3D::BoundingBox bbox;
  for(unsigned int i=_vertices.size();i--;)
    {
    const Util3D::Dbl3 & vertex=_vertices[i];
    bbox.addXYZ(vertex.x,vertex.y,vertex.z);
    }
  _describeGeometry(bbox,
                    (const Util3D::Dbl3 *)0,0,
                    (const Util3D::Face *)0,0,
                    0,_vertices.size());
  _needsGeometryRebuild=false;
  }
if(_needsAppearanceRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_appearanceGeometryList,GL_COMPILE);
    _setAppearance();
    glCallList(_geometryList);
    _resetAppearance();
    glEndList();
    }
  _needsAppearanceRebuild=false;
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
