// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/shader.cpp
// author(s)   : T. Jourdan, F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/shader.h"

#include "AReVi/Shapes/shapePart3D.h"

#include "AReVi/Lib3D/gl_priv.h"

#include "AReVi/Utils/urlStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Shader (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Shader,ArObject)

//-------- Construction / Destruction ----------------------------------------

Shader::Shader(ArCW & arCW) // Protected
: ArObject(arCW),
  _handle((GL::SLhandle)0),
  _code(),
  _programs()
{
if(!GL::hasShadingLanguage())
  {
  setErrorMessage("Shading language is not supported under this platform");
  }
}

Shader::~Shader(void) // Protected
{
if(_handle)
  {
  GL::glDeleteObject(_handle);
  }
}

//-------- Source code management --------------------------------------------

bool // success
Shader::load(const StlString & code)
{
if(GL::hasShadingLanguage())
  {
  _code=code;
  const char * codePtr=_code.c_str();
  int len=_code.size();
  GL::glShaderSource(_handle,1,&codePtr,&len);
  GL::glCompileShader(_handle);
  int compiled;
  GL::glGetObjectParameteriv(_handle,GL_OBJECT_COMPILE_STATUS_ARB,&compiled);
  if(compiled)
    {
    setErrorMessage("");
    }
  else
    {
    int len1;
    GL::glGetObjectParameteriv(_handle,GL_OBJECT_INFO_LOG_LENGTH_ARB,&len1);
    if(len1>1)
      {
      int len2;
      char * msgPtr=new char[len1];
      GL::glGetInfoLog(_handle,len1,&len2,msgPtr);
      setErrorMessage(msgPtr);
      delete[] msgPtr;
      }
    else
      {
      setErrorMessage("Shader compilation failure");
      }
    }
  _linkShaders();
  return(compiled);
  }
else
  {
  return(false);
  }
}

bool // success
Shader::loadURL(const StlString & url)
{
if(GL::hasShadingLanguage())
  {
  ArRef<URLStream> input=URLStream::NEW(url);
  if(input->fail())
    {
    setErrorMessage("Cannot load URL '"+url+"'");
    return(false);
    }
  else
    {
    return(loadStream(input));
    }
  }
else
  {
  return(false);
  }
}

bool // success
Shader::loadStream(ArRef<AbstractIStream> input)
{
if(GL::hasShadingLanguage())
  {
  if(input->fail())
    {
    setErrorMessage("Cannot read input stream");
    return(false);
    }
  StlString code;
  StlString line;
  while(input->readTxtLine(line))
    {
    code+=line;
    code+='\n';
    }
  return(load(code));
  }
else
  {
  return(false);
  }
}

unsigned int // OpenGL shader handle
Shader::getGLHandle(void) const
{
return(_handle);
}

//-------- Serialization operations ------------------------------------------

void
Shader::serialize(SerializationDependencies & dependenciesOut,
                  ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut, output);
output->writeString(_code);
}

bool // success
Shader::unserialize(const SerializationDependencies & dependencies,
                    ArRef<AbstractIStream> input)
{
if(!ArObject::unserialize(dependencies, input))
  {
  return(false);
  }
StlString code;
if(!input->readString(code))
  {
  return(false);
  }
load(code);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Shader::_addShaderProgram(ShaderProgram * shaderProgram) // Protected
{
_programs.push_back(shaderProgram);
}

void
Shader::_removeShaderProgram(ShaderProgram * shaderProgram) // Protected
{
for(unsigned int i=_programs.size();i--;)
  {
  if(_programs[i]==shaderProgram)
    {
    StlVectorFastErase(_programs,i);
    break;
    }
  }
}

void
Shader::_linkShaders(void) // Protected
{
for(unsigned int i=_programs.size();i--;)
  {
  _programs[i]->_linkShaders();
  }
}


//----------------------------------------------------------------------------
// Class VertexShader
//----------------------------------------------------------------------------

AR_CLASS_DEF(VertexShader,Shader)

//-------- Construction / Destruction ----------------------------------------

VertexShader::VertexShader(ArCW & arCW) // Protected
: Shader(arCW)
{
if(GL::hasShadingLanguage())
  {
  _handle=GL::glCreateShaderObject(GL_VERTEX_SHADER_ARB);
  }
}

VertexShader::~VertexShader(void) // Protected
{
while(!_programs.empty())
  {
  _programs.back()->setShaders(VertexShader::nullRef(),
                               _programs.back()->accessFragmentShader());
  }
}


//----------------------------------------------------------------------------
// Class FragmentShader
//----------------------------------------------------------------------------

AR_CLASS_DEF(FragmentShader,Shader)

//-------- Construction / Destruction ----------------------------------------

FragmentShader::FragmentShader(ArCW & arCW) // Protected
: Shader(arCW)
{
if(GL::hasShadingLanguage())
  {
  _handle=GL::glCreateShaderObject(GL_FRAGMENT_SHADER_ARB);
  }
}

FragmentShader::~FragmentShader(void) // Protected
{
while(!_programs.empty())
  {
  _programs.back()->setShaders(_programs.back()->accessVertexShader(),
                               FragmentShader::nullRef());
  }
}


//----------------------------------------------------------------------------
// Class ShaderProgram
//----------------------------------------------------------------------------

AR_CLASS_DEF(ShaderProgram,ShapePartRenderer)

//-------- Construction / Destruction ----------------------------------------

ShaderProgram::ShaderProgram(ArCW & arCW) // Protected
: ShapePartRenderer(arCW),
  _handle((GL::SLhandle)0),
  _linked(false),
  _vertexShader(),
  _fragmentShader()
{
if(GL::hasShadingLanguage())
  {
  _handle=GL::glCreateProgramObject();
  }
else
  {
  setErrorMessage("Shading language is not supported under this platform");
  }
}

ShaderProgram::~ShaderProgram(void)
{
setShaders(VertexShader::nullRef(),FragmentShader::nullRef());
if(_handle)
  {
  GL::glDeleteObject(_handle);
  }
}

//-------- Shader management -------------------------------------------------

bool // link success
ShaderProgram::setShaders(ArRef<VertexShader> vertexShader,
                          ArRef<FragmentShader> fragmentShader)
{
if(_vertexShader)
  {
  _vertexShader->_removeShaderProgram(this);
  }
_vertexShader=vertexShader;
if(_vertexShader)
  {
  _vertexShader->_addShaderProgram(this);
  }
if(_fragmentShader)
  {
  _fragmentShader->_removeShaderProgram(this);
  }
_fragmentShader=fragmentShader;
if(_fragmentShader)
  {
  _fragmentShader->_addShaderProgram(this);
  }
return(_linkShaders());
}

ArConstRef<VertexShader>
ShaderProgram::getVertexShader(void) const
{
return(_vertexShader);
}

ArRef<VertexShader>
ShaderProgram::accessVertexShader(void)
{
return(_vertexShader);
}

ArConstRef<FragmentShader>
ShaderProgram::getFragmentShader(void) const
{
return(_fragmentShader);
}

ArRef<FragmentShader>
ShaderProgram::accessFragmentShader(void)
{
return(_fragmentShader);
}

unsigned int // OpenGL program handle
ShaderProgram::getGLHandle(void) const
{
return(_handle);
}

//-------- Serialization operations ------------------------------------------

void
ShaderProgram::extractDependencies(SerializationDependencies & dependenciesOut,
                                   bool recursive) const
{
ArObject::extractDependencies(dependenciesOut,recursive);
if(_vertexShader)
  {
  dependenciesOut.addReference(_vertexShader);
  if(recursive)
    {
    _vertexShader->extractDependencies(dependenciesOut,recursive);
    }
  }
if(_fragmentShader)
  {
  dependenciesOut.addReference(_fragmentShader);
  if(recursive)
    {
    _fragmentShader->extractDependencies(dependenciesOut,recursive);
    }
  }
}

void
ShaderProgram::serialize(SerializationDependencies & dependenciesOut,
                         ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
unsigned int flags=(_vertexShader   ? (1<<0) : 0)+
                   (_fragmentShader ? (1<<1) : 0);
output->writeUInt(flags);
if(_vertexShader)
  {
  output->writeUInt(dependenciesOut.addReference(_vertexShader));
  }
if(_fragmentShader)
  {
  output->writeUInt(dependenciesOut.addReference(_fragmentShader));
  }
}

bool // success
ShaderProgram::unserialize(const SerializationDependencies & dependencies,
                           ArRef<AbstractIStream> input)
{
if(!ArObject::unserialize(dependencies,input))
  {
  return false;
  }
unsigned int flags;
if(!input->readUInt(flags))
  {
  return(false);
  }
bool hasVertexShader=flags&(1<<0);
unsigned int vertexShaderID;
if(hasVertexShader)
  {
  if(!input->readUInt(vertexShaderID))
    {
    return(false);
    }
  }
bool hasFragmentShader=flags&(1<<1);
unsigned int fragmentShaderID;
if(hasFragmentShader)
  {
  if(!input->readUInt(fragmentShaderID))
    {
    return(false);
    }
  }
ArRef<VertexShader> vertexShader;
ArRef<FragmentShader> fragmentShader;
if(hasVertexShader)
  {
  vertexShader=ar_down_cast<VertexShader>
               (dependencies.findReference(vertexShaderID));
  }
if(hasFragmentShader)
  {
  fragmentShader=ar_down_cast<FragmentShader>
                 (dependencies.findReference(fragmentShaderID));
  }
setShaders(vertexShader,fragmentShader);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
ShaderProgram::_render(const ShapePart3D * part,
                       const Object3D * /* object */,
                       const Renderer3D * /* renderer */) const // Protected
{
if(_linked)
  {
  GL::glUseProgramObject(_handle);
  _renderPart(part);
  GL::glUseProgramObject(0);
  }
else
  {
  _renderPart(part);
  }
}

void
ShaderProgram::_renderWithoutAppearance(const ShapePart3D * part,
                                        const Object3D * /* object */,
                                        const Renderer3D * /* renderer */)
                                                            const // Protected
{
if(_linked)
  {
  GL::glUseProgramObject(_handle);
  _renderPartWithoutAppearance(part);
  GL::glUseProgramObject(0);
  }
else
  {
  _renderPartWithoutAppearance(part);
  }
}

bool // success
ShaderProgram::_linkShaders(void) // Protected
{
if(GL::hasShadingLanguage())
  {
  int linked;
  if(_vertexShader)
    {
    GL::glAttachObject(_handle,_vertexShader->_handle);
    }
  if(_fragmentShader)
    {
    GL::glAttachObject(_handle,_fragmentShader->_handle);
    }
  GL::glLinkProgram(_handle);
  if(_vertexShader)
    {
    GL::glDetachObject(_handle,_vertexShader->_handle);
    }
  if(_fragmentShader)
    {
    GL::glDetachObject(_handle,_fragmentShader->_handle);
    }
  GL::glGetObjectParameteriv(_handle,GL_OBJECT_LINK_STATUS_ARB,&linked);
  _linked=linked;
  if(_linked)
    {
    setErrorMessage("");
    }
  else
    {
    int len1;
    GL::glGetObjectParameteriv(_handle,GL_OBJECT_INFO_LOG_LENGTH_ARB,&len1);
    if(len1>1)
      {
      int len2;
      char * msgPtr=new char[len1];
      GL::glGetInfoLog(_handle,len1,&len2,msgPtr);
      setErrorMessage(msgPtr);
      delete[] msgPtr;
      }
    else
      {
      setErrorMessage("Shader link failure");
      }
    }
  _postRedisplay();
  return(_linked);
  }
else
  {
  return(false);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
