// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/shape3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/shape3D.h"

#include "AReVi/Shapes/shapePart3D.h"

#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/transform3D.h"
#include "AReVi/Lib3D/boundingBox3D.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Shape3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Shape3D,ArObject)

//-------- Construction / Destruction ----------------------------------------

Shape3D::Shape3D(ArCW & arCW) // Protected
: ArObject(arCW),
  _objects(),
  _flatParts(),
  _transform(true),
  _bbox(),
  _nbTriangles(0),
  _nbLines(0),
  _nbPoints(0),
  _triangleDivision(0),
  _billboard(false),
  _billboardAroundZ(true),
  _frameID(Renderer3D::_currentFrameID),
  _rootParts(),
  _needsGeometryUpdate(true),
  _needsRebuild(true),
  _transformationCB(thisRef()),
  _structureChangeCB(thisRef())
{
  // Nothing to be done
}

Shape3D::~Shape3D(void) // Protected
{
while(!_rootParts.empty())
  {
  removeRootPart(_rootParts.back());
  }
for(unsigned int i=_objects.size();i--;)
  {
  _objects[i]->_shapeDestroyed(this);
  }
}

//-------- Geometry ----------------------------------------------------------

void
Shape3D::writeTransformation(ArConstRef<Transform3D> transform)
{
_transform=transform->getTransformation();
_postRedisplay();
_onTransformation();
}

void
Shape3D::readTransformation(ArRef<Transform3D> transformOut) const
{
transformOut->accessTransformation()=_transform;
}

void
Shape3D::readBoundingBox(ArRef<BoundingBox3D> bboxOut) const
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
bboxOut->accessBoundingBox()=_bbox;
}

unsigned int // number of triangles
Shape3D::getNbTriangles(void) const
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
return(_nbTriangles);
}

unsigned int // number of lines
Shape3D::getNbLines(void) const
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
return(_nbLines);
}

unsigned int // number of points
Shape3D::getNbPoints(void) const
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
return(_nbPoints);
}

bool // currently displayed by a renderer
Shape3D::isCurrentlyDisplayed(void) const
{
return(_frameID==Renderer3D::_currentFrameID);
}

//-------- Detection ----------------------------------------------------------

void
Shape3D::setTriangleDivision(unsigned int maxTriangles)
{
_triangleDivision=maxTriangles;
for(unsigned int i=_flatParts.size();i--;)
  {
  _flatParts[i]->_setTriangleDivision(_triangleDivision);
  }
}

unsigned int // max triangles in divisions
Shape3D::getTriangleDivision(void) const
{
return(_triangleDivision);
}

//-------- Billboard ---------------------------------------------------------

void
Shape3D::setBillboard(bool isBillboard,
                      bool aroundZ)
{
_billboard=isBillboard;
_billboardAroundZ=aroundZ;
_postRedisplay();
_onStructureChange();
}

void
Shape3D::getBillboard(bool & isBillboardOut,
                      bool & aroundZOut) const
{
isBillboardOut=_billboard;
aroundZOut=_billboardAroundZ;
}

//-------- Part management ---------------------------------------------------

void
Shape3D::addRootPart(ArRef<ShapePart3D> part)
{
if(part->_parentPart)
  {
  part->_parentPart->removeSubPart(part);
  }
else if(part->_shape)
  {
  part->_shape->removeRootPart(part);
  }
_rootParts.push_back(part);
part->_updateShape(thisRef());
part->_updateTransformation();
_onStructureChange();
}

void
Shape3D::removeRootPart(ArRef<ShapePart3D> part)
{
for(unsigned int i=_rootParts.size();i--;)
  {
  if(_rootParts[i]==part)
    {
    StlVectorFastErase(_rootParts,i);
    part->_updateShape(Shape3D::nullPtr());
    _onStructureChange();
    return;
    }
  }
}

unsigned int // number of root parts
Shape3D::getNbRootParts(void) const
{
return(_rootParts.size());
}

ArConstRef<ShapePart3D> // index'th root part
Shape3D::getRootPart(unsigned int index) const
{
return(_rootParts[index]);
}

ArRef<ShapePart3D> // index'th root part
Shape3D::accessRootPart(unsigned int index)
{
return(_rootParts[index]);
}

unsigned int // number of parts/subparts
Shape3D::getNbParts(void) const
{
return(_flatParts.size());
}

ArConstRef<ShapePart3D> // index'th part/subpart
Shape3D::getPart(unsigned int index) const
{
return(_flatParts[index]->thisRef());
}

ArRef<ShapePart3D> // index'th part/subpart
Shape3D::accessPart(unsigned int index)
{
return(_flatParts[index]->thisRef());
}

//-------- Serialization operations ------------------------------------------

void
Shape3D::extractDependencies(SerializationDependencies & dependenciesOut,
                             bool recursive) const
{
ArObject::extractDependencies(dependenciesOut,recursive);
unsigned int nbRootParts=_rootParts.size();
for(unsigned int i=0;i<nbRootParts;i++)
  {
  if(recursive)
    {
    _rootParts[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_rootParts[i]);
  }
}

void
Shape3D::serialize(SerializationDependencies & dependenciesOut,
                   ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
output->writeUInt(_triangleDivision);
bool hasTransform=!_transform.isIdentity(false);
unsigned int flags=(_billboard        ? 1<<0 : 0)+
                   (_billboardAroundZ ? 1<<1 : 0)+
                   (hasTransform      ? 1<<2 : 0);
output->writeUInt(flags);
unsigned int nbRootParts=_rootParts.size();
output->writeUInt(nbRootParts);
unsigned int i;
for(i=0;i<nbRootParts;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_rootParts[i]));
  }
if(hasTransform)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_transform.matrix[i]);
    }
  }
}

bool // success
Shape3D::unserialize(const SerializationDependencies & dependencies,
                     ArRef<AbstractIStream> input)
{
unsigned int triangleDivision,flags,nbRootParts;
StlVector<unsigned int> rootParts;
double matrix[16];
if(!ArObject::unserialize(dependencies,input)||
   !input->readUInt(triangleDivision)||
   !input->readUInt(flags)||
   !input->readUInt(nbRootParts))
  {
  return(false);
  }
unsigned int i;
for(i=0;i<nbRootParts;i++)
  {
  rootParts.push_back(0);
  if(!input->readUInt(rootParts.back()))
    {
    return(false);
    }
  }
bool hasTransform=flags&(1<<2);
if(hasTransform)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix[i]))
      {
      return(false);
      }
    }
  }
_triangleDivision=triangleDivision;
_billboard=flags&(1<<0);
_billboardAroundZ=flags&(1<<1);
for(i=0;i<nbRootParts;i++)
  {
  addRootPart(ar_down_cast<ShapePart3D>
              (dependencies.findReference(rootParts[i])));
  }
if(hasTransform)
  {
  _transform.setMatrix(matrix);
  }
else
  {
  _transform.identity();
  }
_postRedisplay();
return(true); // FIXME: needs _on...() ???
}

//-------- Forbidden operations ----------------------------------------------

void
Shape3D::_addFlatPart(ShapePart3D * part) // Protected
{
_flatParts.push_back(part);
_needsRebuild=true;
_needsGeometryUpdate=true;
_postRedisplay();
}

void
Shape3D::_removeFlatPart(ShapePart3D * part) // Protected
{
for(unsigned int i=_flatParts.size();i--;)
  {
  if(_flatParts[i]==part)
    {
    _needsRebuild=true;
    _needsGeometryUpdate=true;
    _postRedisplay();
    StlVectorFastErase(_flatParts,i);
    break;
    }
  }
}

void
Shape3D::_postRedisplay(void) // Protected
{
for(unsigned int i=_objects.size();i--;)
  {
  _objects[i]->_postRedisplay();
  }
}

void
Shape3D::_addObject(Object3D * object) // Protected
{
_objects.push_back(object);
}

void
Shape3D::_removeObject(Object3D * object) // Protected
{
for(unsigned int i=_objects.size();i--;)
  {
  if(_objects[i]==object)
    {
    StlVectorFastErase(_objects,i);
    break;
    }
  }
}

void
Shape3D::_askVisibleParts(Renderer3D * renderer,
                          const Object3D * object,
                          bool selected) const // Protected
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
Util3D::Base location(false);
_findBillboardLocation(location,object->getBase(),renderer->getBase());
unsigned int nbParts=_flatParts.size();
if((nbParts<2)||
   _boundingBoxVisibility(renderer,location,_bbox,
                          (const Util3D::Transform *)0))
  {
  bool updateFrameID=false;
  for(unsigned int i=nbParts;i--;)
    {
    const ShapePart3D * part=_flatParts[i];
    const Util3D::BoundingBox & bbox=part->_bbox;
    const Util3D::Transform & transform=part->_flatTransform;
    if(_boundingBoxVisibility(renderer,location,bbox,&transform))
      {
      Renderer3D::RenderInfo * info=Renderer3D::_makeObject3DInfo(location,
                                                                  object,
                                                                  this,
                                                                  part,
                                                                  selected);
      if(selected)
        {
        if(object->_selectedMaterial.getTransparency()>=0.01)
          {
          double x=bbox.centerX();
          double y=bbox.centerY();
          double z=bbox.centerZ();
          transform.apply(x,y,z,x,y,z);
          location.localToGlobalPosition(x,y,z);
          renderer->globalToLocalPosition(x,y,z);
          info->squareDistance=x*x+y*y+z*z;
          renderer->_insertTransparentInfo(info);
          }
        else
          {
          renderer->_opaqueInfos.push_back(info);
          }
        }
      else
        {
        if(part->_needsBlending)
          {
          double x=bbox.centerX();
          double y=bbox.centerY();
          double z=bbox.centerZ();
          transform.apply(x,y,z,x,y,z);
          location.localToGlobalPosition(x,y,z);
          renderer->globalToLocalPosition(x,y,z);
          info->squareDistance=x*x+y*y+z*z;
          renderer->_insertTransparentInfo(info);
          }
        else if(part->_needsAlphaTest)
          {
          renderer->_alphaTestInfos.push_back(info);
          }
        else
          {
          renderer->_opaqueInfos.push_back(info);
          }
        updateFrameID=true;
        }
      }
    }
  if(updateFrameID)
    {
    const_cast<Object3D *>(object)->_frameID=Renderer3D::_currentFrameID;
    const_cast<Shape3D *>(this)->_frameID=Renderer3D::_currentFrameID;
    }
  }
}

void
Shape3D::_findBillboardLocation(Util3D::Base & locationOut,
                                const Util3D::Base & location,
                                const Util3D::Base & viewPoint) const
{                                                                 // Protected
if(_billboard)
  {
  if(_billboardAroundZ)
    {
    locationOut=location;
    double x,y,z;
    viewPoint.getPosition(x,y,z);
    location.globalToLocalPosition(x,y,z);
    locationOut.yaw(cart2DToAngle(x,y));
    }
  else
    {
    locationOut=viewPoint;
    double x,y,z;
    location.getPosition(x,y,z);
    locationOut.setPosition(x,y,z);
    viewPoint.globalToLocalPosition(x,y,z);
    double yaw,pitch;
    cart3DToAngles(yaw,pitch,x,y,z);
    locationOut.yaw(M_PI+yaw);
    locationOut.pitch(-pitch);
    }
  }
else
  {
  locationOut=location;
  }
}

// #define BBOX_TRANSFORM_PRECOMP

bool // bbox is visible in renderer
Shape3D::_boundingBoxVisibility(const Renderer3D * renderer,
                                const Util3D::Base & base,
                                const Util3D::BoundingBox & bbox,
                                const Util3D::Transform * transform) const
{                                                                 // Protected
if(bbox.valid)
  {
#ifdef BBOX_TRANSFORM_PRECOMP
  Util3D::Base b(renderer->getBase());
  b.invert();
  Util3D::Transform tr(b);
  tr.postMove(base);
  tr.postTransform(_transform);
  if(transform) tr.postTransform(*transform);
#endif
  const Util3D::Perspective & perspective=renderer->getPerspective();
  unsigned int outFront=8;
  unsigned int outBack=8;
  unsigned int outLeft=8;
  unsigned int outRight=8;
  unsigned int outUp=8;
  unsigned int outDown=8;
  for(unsigned int i=8;i--;)
    {
    double x=(i&1 ? bbox.xmax : bbox.xmin);
    double y=(i&2 ? bbox.ymax : bbox.ymin);
    double z=(i&4 ? bbox.zmax : bbox.zmin);
#ifdef BBOX_TRANSFORM_PRECOMP
    tr.apply(x,y,z,x,y,z);
#else
    if(transform)
      {
      transform->apply(x,y,z,x,y,z);
      }
    _transform.apply(x,y,z,x,y,z);
    base.localToGlobalPosition(x,y,z);
    renderer->getBase().globalToLocalPosition(x,y,z);
#endif
    Util3D::VertexLocation loc=perspective.locateVertex(x,y,z);
    if(loc==Util3D::MIDDLE_LOC)
      {
      return(true); // inside perspective
      }
    if(loc&Util3D::XPOS_LOC)
      {
      --outFront;
      }
    else if(loc&Util3D::XNEG_LOC)
      {
      --outBack;
      }
    if(loc&Util3D::YPOS_LOC)
      {
      --outLeft;
      }
    if(loc&Util3D::YNEG_LOC)
      {
      --outRight;
      }
    if(loc&Util3D::ZPOS_LOC)
      {
      --outUp;
      }
    if(loc&Util3D::ZNEG_LOC)
      {
      --outDown;
      }
    }
  return(outFront&&outBack&&
         outLeft&&outRight&&
         outUp&&outDown); // not on the same side
  }
else
  {
  return(false);
  }
}

double // (Pinter-Pray)/(Pdir-Pray) [0.0;+inf[
Shape3D::_intersectRay(unsigned int level,
                       bool findNearest,
                       const Util3D::Dbl3 & rayP,
                       const Util3D::Dbl3 & dirP,
                       ArRef<ShapePart3D> & partOut,
                       unsigned int & triIndexOut) const // Protected
{
const_cast<Shape3D *>(this)->_rebuildIfNeeded();
Util3D::Transform tr=_transform;
tr.invert();
Util3D::Dbl3 tr1RayP,tr1DirP,tr2RayP,tr2DirV;
tr.apply(tr1RayP.x,tr1RayP.y,tr1RayP.z,rayP.x,rayP.y,rayP.z);
tr.apply(tr1DirP.x,tr1DirP.y,tr1DirP.z,dirP.x,dirP.y,dirP.z);
if(level==1)
  {
  return(Util3D::rayBoxIntersection(_bbox,tr1RayP,
                                    Util3D::Dbl3(tr1DirP.x-tr1RayP.x,
                                                 tr1DirP.y-tr1RayP.y,
                                                 tr1DirP.z-tr1RayP.z),
                                    findNearest));
  }
else
  {
  double k=Util3D::rayBoxIntersection(_bbox,tr1RayP,
                                      Util3D::Dbl3(tr1DirP.x-tr1RayP.x,
                                                   tr1DirP.y-tr1RayP.y,
                                                   tr1DirP.z-tr1RayP.z),
                                      false);
  if(k<0.0)
    {
    return(-1.0);
    }
  }
double kmin=DBL_MAX;
unsigned int i;
for(i=_flatParts.size();i--;)
  {
  ShapePart3D * part=_flatParts[i];
  tr=part->_flatTransform;
  tr.invert();
  tr.apply(tr2RayP.x,tr2RayP.y,tr2RayP.z,tr1RayP.x,tr1RayP.y,tr1RayP.z);
  tr.apply(tr2DirV.x,tr2DirV.y,tr2DirV.z,tr1DirP.x,tr1DirP.y,tr1DirP.z);
  tr2DirV.x-=tr2RayP.x;
  tr2DirV.y-=tr2RayP.y;
  tr2DirV.z-=tr2RayP.z;
  if(level==2)
    {
    double k=Util3D::rayBoxIntersection(part->_bbox,
                                        tr2RayP,tr2DirV,findNearest);
    if((k>=0.0)&&(k<kmin))
      {
      partOut=part->thisRef();
      if(findNearest)
        {
        kmin=k;
        }
      else
        {
        return(k);
        }
      }
    }
  else
    {
    double k=Util3D::rayBoxIntersection(part->_bbox,
                                        tr2RayP,tr2DirV,true);
    if((k>=0.0)&&(k<kmin))
      {
      unsigned int triIndex;
      k=part->_rayIntersection(tr2RayP,tr2DirV,findNearest,triIndex);
      if((k>=0.0)&&(k<kmin))
        {
        triIndexOut=triIndex;
        partOut=part->thisRef();
        if(findNearest)
          {
          kmin=k;
          }
        else
          {
          return(k);
          }
        }
      }
    }
  }
return(kmin==DBL_MAX ? -1.0 : kmin);
}

bool // collision occured
Shape3D::_testCollision(const Shape3D * shape1,
                        const Shape3D * shape2,
                        const Util3D::Base & base1,
                        const Util3D::Base & base2,
                        unsigned int level) // Protected, Static
{
const_cast<Shape3D *>(shape1)->_rebuildIfNeeded();
const_cast<Shape3D *>(shape2)->_rebuildIfNeeded();
if(shape1->_bbox.valid&&shape2->_bbox.valid)
  {
  Util3D::Transform tr21(base1);
  tr21.postTransform(shape1->getTransformation());
  tr21.invert();
  tr21.postMove(base2);
  tr21.postTransform(shape2->getTransformation());
  if(level==1)
    {
    if(shape1->_bbox.testCollision(tr21,shape2->_bbox))
      {
      return(true); // shape bbox collision
      }
    }
  else if(shape1->_bbox.testHalfApproxCollision(tr21,shape2->_bbox))
    {
    for(unsigned int i=shape1->_flatParts.size();i--;)
      {
      ShapePart3D * part1=shape1->_flatParts[i];
      if(part1->_bbox.valid)
        {
        Util3D::Transform tr2p1(part1->_flatTransform);
        tr2p1.invert();
        tr2p1.postTransform(tr21);
        for(unsigned int j=shape2->_flatParts.size();j--;)
          {
          ShapePart3D * part2=shape2->_flatParts[j];
          if(part2->_bbox.valid)
            {
            Util3D::Transform trp2p1(tr2p1);
            trp2p1.postTransform(part2->_flatTransform);
            if(level==2)
              {
              if(part1->_bbox.testCollision(trp2p1,part2->_bbox))
                {
                return(true); // shape part bbox collision
                }
              }
            else if(part1->_bbox.testHalfApproxCollision(trp2p1,part2->_bbox)&&
                    part1->_testCollision(trp2p1,part2))
              {
              return(true);
              }
            }
          }
        }
      }
    }
  }
return(false);
}

void
Shape3D::_intersect(const Shape3D * shape1,
                    const Shape3D * shape2,
                    const Util3D::Base & base1,
                    const Util3D::Base & base2,
                    StlVector<Util3D::TriInter> & itOut) // Protected, Static
{
const_cast<Shape3D *>(shape1)->_rebuildIfNeeded();
const_cast<Shape3D *>(shape2)->_rebuildIfNeeded();
if(shape1->_bbox.valid&&shape2->_bbox.valid)
  {
  Util3D::Transform tr21(base1);
  tr21.postTransform(shape1->getTransformation());
  tr21.invert();
  tr21.postMove(base2);
  tr21.postTransform(shape2->getTransformation());
  if(shape1->_bbox.testHalfApproxCollision(tr21,shape2->_bbox))
    {
    for(unsigned int i=shape1->_flatParts.size();i--;)
      {
      ShapePart3D * part1=shape1->_flatParts[i];
      if(part1->_bbox.valid)
        {
        Util3D::Transform tr2p1(part1->_flatTransform);
        tr2p1.invert();
        tr2p1.postTransform(tr21);
        unsigned int oldSize=itOut.size();
        for(unsigned int j=shape2->_flatParts.size();j--;)
          {
          ShapePart3D * part2=shape2->_flatParts[j];
          if(part2->_bbox.valid)
            {
            Util3D::Transform trp2p1(tr2p1);
            trp2p1.postTransform(part2->_flatTransform);
            if(part1->_bbox.testHalfApproxCollision(trp2p1,part2->_bbox))
              {
              part1->_intersect(trp2p1,part2,itOut);
              }
            }
          }
        unsigned int newSize=itOut.size();
        if(newSize!=oldSize)
          {
          Util3D::Transform global(base1);
          global.postTransform(shape1->getTransformation());
          global.postTransform(part1->_flatTransform);
          for(unsigned int k=oldSize;k<newSize;k++)
            {
            Util3D::TriInter & it=itOut[k];
            double x,y,z;
            global.apply(x,y,z,it.ta1.x,it.ta1.y,it.ta1.z);
            it.ta1.x=x; it.ta1.y=y; it.ta1.z=z;
            global.apply(x,y,z,it.ta2.x,it.ta2.y,it.ta2.z);
            it.ta2.x=x; it.ta2.y=y; it.ta2.z=z;
            global.apply(x,y,z,it.ta3.x,it.ta3.y,it.ta3.z);
            it.ta3.x=x; it.ta3.y=y; it.ta3.z=z;
            global.apply(x,y,z,it.tb1.x,it.tb1.y,it.tb1.z);
            it.tb1.x=x; it.tb1.y=y; it.tb1.z=z;
            global.apply(x,y,z,it.tb2.x,it.tb2.y,it.tb2.z);
            it.tb2.x=x; it.tb2.y=y; it.tb2.z=z;
            global.apply(x,y,z,it.tb3.x,it.tb3.y,it.tb3.z);
            it.tb3.x=x; it.tb3.y=y; it.tb3.z=z;
            global.apply(x,y,z,it.i1.x,it.i1.y,it.i1.z);
            it.i1.x=x; it.i1.y=y; it.i1.z=z;
            global.apply(x,y,z,it.i2.x,it.i2.y,it.i2.z);
            it.i2.x=x; it.i2.y=y; it.i2.z=z;
            }
          }
        }
      }
    }
  }
}

void
Shape3D::_postRebuild(void) // Protected
{
_needsRebuild=true;
_postRedisplay();
}

void
Shape3D::_rebuildIfNeeded(void) // Protected
{
if(_needsRebuild)
  {
  for(unsigned int i=_flatParts.size();i--;)
    {
    _flatParts[i]->_rebuildIfNeeded();
    }
  _needsRebuild=false;
  }
if(_needsGeometryUpdate)
  {
  _bbox.valid=false;
  _nbTriangles=0;
  _nbLines=0;
  _nbPoints=0;
  ShapePart3D * part;
  for(unsigned int i=_flatParts.size();i--;)
    {
    part=_flatParts[i];
    _bbox.addTransformedBox(part->_flatTransform,part->_bbox);
    _nbTriangles+=part->_nbTriangles;
    _nbLines+=part->_nbLines;
    _nbPoints+=part->_nbPoints;
    }
  _postRedisplay();
  _needsGeometryUpdate=false;
  }
}

void
Shape3D::_onTransformation(void)
{
if(!_transformationCB.empty())
  {
  Shape3D::TransformationEvent evt;
  _transformationCB.fire(evt);
  }
}

void
Shape3D::_onStructureChange(void)
{
if(!_structureChangeCB.empty())
  {
  Shape3D::StructureChangeEvent evt;
  _structureChangeCB.fire(evt);
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
