// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/shapePart3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/shapePart3D.h"

#include "AReVi/Shapes/shape3D.h"
#include "AReVi/Shapes/shader.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/transform3D.h"
#include "AReVi/Lib3D/boundingBox3D.h"
#include "AReVi/Lib3D/renderer3D.h"
#include "AReVi/Lib3D/light3D.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class ShapePart3D (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(ShapePart3D,ArObject)

//-------- Construction / Destruction ----------------------------------------

ShapePart3D::ShapePart3D(ArCW & arCW) // Protected
: ArObject(arCW),
  _bbox(),
  _vPtr((const Util3D::Dbl3 *)0),
  _nbVertices(0),
  _tPtr((const Util3D::Face *)0),
  _nbTriangles(0),
  _nbLines(0),
  _nbPoints(0),
  _triangleDivision(0),
  _tree((ShapePart3D::BoxTree *)0),
  _useDisplayList(true),
  _appearanceGeometryList(glGenLists(2)),
  _geometryList(_appearanceGeometryList+1),
  _needsBlending(false),
  _needsAlphaTest(false),
  _localTransform(true),
  _flatTransform(true),
  _offset(0.0),
  _renderer(),
  _needsAppearanceRebuild(true),
  _needsGeometryRebuild(true),
  _subParts(),
  _parentPart(),
  _shape(),
  _transformationCB(thisRef()),
  _structureChangeCB(thisRef()),
  _appearanceRebuildCB(thisRef()),
  _geometryRebuildCB(thisRef())
{
  // Nothing to be done
}

ShapePart3D::~ShapePart3D(void) // Protected
{
setRenderer(ShapePartRenderer::nullRef());
while(!_subParts.empty())
  {
  removeSubPart(_subParts.back());
  }
if(_parentPart)
  {
  _parentPart->removeSubPart(thisRef());
  }
else if(_shape)
  {
  _shape->removeRootPart(thisRef());
  }
if(_useDisplayList)
  {
  glDeleteLists(_appearanceGeometryList,2);
  }
if(_tree)
  {
  delete _tree;
  }
}

//-------- Geometry ----------------------------------------------------------

void
ShapePart3D::writeTransformation(ArConstRef<Transform3D> transform)
{
_localTransform=transform->getTransformation();
_updateTransformation();
if(_shape)
  {
  _shape->_needsGeometryUpdate=true;
  _shape->_postRedisplay();
  }
_onTransformation();
}

void
ShapePart3D::readTransformation(ArRef<Transform3D> transformOut) const
{
transformOut->accessTransformation()=_localTransform;
}

void
ShapePart3D::readFlatTransformation(ArRef<Transform3D> transformOut) const
{
transformOut->accessTransformation()=_flatTransform;
}

void
ShapePart3D::readBoundingBox(ArRef<BoundingBox3D> bboxOut) const
{
const_cast<ShapePart3D *>(this)->_rebuildIfNeeded();
bboxOut->accessBoundingBox()=_bbox;
}

//-------- Appearance --------------------------------------------------------

void
ShapePart3D::setDepthOffset(double offset)
{
_offset=-offset;
_postAppearanceRebuild();
}

double // depth buffer offset
ShapePart3D::getDepthOffset(void) const
{
return(-_offset);
}

void
ShapePart3D::setRenderer(ArRef<ShapePartRenderer> renderer)
{
if(_renderer)
  {
  _renderer->_removeShapePart(this);
  }
_renderer=renderer;
if(_renderer)
  {
  _renderer->_addShapePart(this);
  }
_postRedisplayIfVisible();
}

ArConstRef<ShapePartRenderer>
ShapePart3D::getRenderer(void) const
{
return(_renderer);
}

ArRef<ShapePartRenderer>
ShapePart3D::accessRenderer(void)
{
return(_renderer);
}

//-------- Display list usage ------------------------------------------------

void
ShapePart3D::setDisplayListUsage(bool useDisplayList)
{
if(_useDisplayList!=useDisplayList)
  {
  _useDisplayList=useDisplayList;
  if(_useDisplayList)
    {
    _appearanceGeometryList=glGenLists(2);
    _geometryList=_appearanceGeometryList+1;
    _postAppearanceRebuild();
    _postGeometryRebuild();
    }
  else
    {
    glDeleteLists(_appearanceGeometryList,2);
    }
  }
}

bool // display list usage is enabled
ShapePart3D::getDisplayListUsage(void) const
{
return(_useDisplayList);
}

//-------- Tree structure ----------------------------------------------------

void
ShapePart3D::addSubPart(ArRef<ShapePart3D> part)
{
if(part->_parentPart)
  {
  part->_parentPart->removeSubPart(part);
  }
else if(part->_shape)
  {
  part->_shape->removeRootPart(part);
  }
_subParts.push_back(part);
part->_parentPart=thisPtr();
part->_updateShape(_shape);
part->_updateTransformation();
_onStructureChange();
}

void
ShapePart3D::removeSubPart(ArRef<ShapePart3D> part)
{
for(unsigned int i=_subParts.size();i--;)
  {
  if(_subParts[i]==part)
    {
    StlVectorFastErase(_subParts,i);
    part->_parentPart.clear();
    part->_updateShape(Shape3D::nullPtr());
    _onStructureChange();
    return;
    }
  }
}

unsigned int // number of sub-parts
ShapePart3D::getNbSubParts(void) const
{
return(_subParts.size());
}

ArConstRef<ShapePart3D> // index'th sub-part
ShapePart3D::getSubPart(unsigned int index) const
{
return(_subParts[index]);
}

ArRef<ShapePart3D> // index'th sub-part
ShapePart3D::accessSubPart(unsigned int index)
{
return(_subParts[index]);
}

ArConstRef<ShapePart3D> // parent part
ShapePart3D::getParentPart(void) const
{
return(_parentPart);
}

ArRef<ShapePart3D> // parent part
ShapePart3D::accessParentPart(void)
{
return(_parentPart);
}

ArConstRef<Shape3D> // owner shape
ShapePart3D::getShape(void) const
{
return(_shape);
}

ArRef<Shape3D> // owner shape
ShapePart3D::accessShape(void)
{
return(_shape);
}

//-------- Serialization operations ------------------------------------------

void
ShapePart3D::extractDependencies(SerializationDependencies & dependenciesOut,
                                 bool recursive) const
{
ArObject::extractDependencies(dependenciesOut,recursive);
unsigned int nbSubParts=_subParts.size();
for(unsigned int i=0;i<nbSubParts;i++)
  {
  if(recursive)
    {
    _subParts[i]->extractDependencies(dependenciesOut,true);
    }
  dependenciesOut.addReference(_subParts[i]);
  }
if(_renderer)
  {
  dependenciesOut.addReference(_renderer);
  if(recursive)
    {
    _renderer->extractDependencies(dependenciesOut);
    }
  }
}

void
ShapePart3D::serialize(SerializationDependencies & dependenciesOut,
                       ArRef<AbstractOStream> output) const
{
ArObject::serialize(dependenciesOut,output);
output->writeDouble(_offset);
bool hasTransform=!_localTransform.isIdentity(false);
unsigned int flags=(hasTransform ? 1<<0 : 0)+
                   (_renderer    ? 1<<1 : 0);
output->writeUInt(flags);
unsigned int nbSubParts=_subParts.size();
output->writeUInt(nbSubParts);
unsigned int i;
for(i=0;i<nbSubParts;i++)
  {
  output->writeUInt(dependenciesOut.addReference(_subParts[i]));
  }
if(hasTransform)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_localTransform.matrix[i]);
    }
  }
if(_renderer)
  {
  output->writeUInt(dependenciesOut.addReference(_renderer));
  }
}

bool // success
ShapePart3D::unserialize(const SerializationDependencies & dependencies,
                         ArRef<AbstractIStream> input)
{
double offset;
unsigned int flags;
unsigned int nbSubParts;
StlVector<unsigned int> subParts;
double matrix[16];
if(!ArObject::unserialize(dependencies,input)||
   !input->readDouble(offset)||
   !input->readUInt(flags)||
   !input->readUInt(nbSubParts))
  {
  return(false);
  }
unsigned int i;
for(i=0;i<nbSubParts;i++)
  {
  subParts.push_back(0);
  if(!input->readUInt(subParts.back()))
    {
    return(false);
    }
  }
bool hasTransform=flags&(1<<0);
if(hasTransform)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix[i]))
      {
      return false;
      }
    }
  }
bool hasRenderer=flags&(1<<1);
unsigned int renderer;
if(hasRenderer&&!input->readUInt(renderer))
  {
  return(false);
  }
_offset=offset;
for(i=0;i<nbSubParts;i++)
  {
  addSubPart(ar_down_cast<ShapePart3D>
             (dependencies.findReference(subParts[i])));
  }
if(hasTransform)
  {
  _localTransform.setMatrix(matrix);
  }
else
  {
  _localTransform.identity();
  }
if(hasRenderer)
  {
  setRenderer(ar_down_cast<ShapePartRenderer>
              (dependencies.findReference(renderer)));
  }
else
  {
  setRenderer(ShapePartRenderer::nullRef());
  }
_updateTransformation();
_onTransformation();
_onStructureChange();
_postGeometryRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
ShapePart3D::_postAppearanceRebuild(void) // Protected
{
_needsAppearanceRebuild=true;
if(_shape)
  {
  _shape->_postRebuild();
  }
if(!_appearanceRebuildCB.empty())
  {
  ShapePart3D::AppearanceRebuildEvent evt;
  _appearanceRebuildCB.fire(evt);
  }
}

void
ShapePart3D::_postGeometryRebuild(void) // Protected
{
_needsGeometryRebuild=true;
if(_shape)
  {
  _shape->_postRebuild();
  }
if(!_geometryRebuildCB.empty())
  {
  ShapePart3D::GeometryRebuildEvent evt;
  _geometryRebuildCB.fire(evt);
  }
}

void
ShapePart3D::_postRedisplayIfVisible(void) // Protected
{
if(_shape&&_shape->isCurrentlyDisplayed())
  {
  _shape->_postRedisplay();
  }
}

void
ShapePart3D::_updateTransformation(void) // Protected
{
_flatTransform=_localTransform;
if(_parentPart)
  {
  _flatTransform.preTransform(_parentPart->_flatTransform);
  }
for(unsigned int i=_subParts.size();i--;)
  {
  _subParts[i]->_updateTransformation();
  }
}

void
ShapePart3D::_updateShape(ArPtr<Shape3D> shape) // Protected
{
if(shape!=_shape)
  {
  if(_shape)
    {
    _shape->_removeFlatPart(this);
    }
  _shape=shape;
  if(_shape)
    {
    _shape->_addFlatPart(this);
    _setTriangleDivision(_shape->getTriangleDivision());
    }
  for(unsigned int i=_subParts.size();i--;)
    {
    _subParts[i]->_updateShape(shape);
    }
  }
}

// void
// ShapePart3D::_setAppearance(void) const=0 // Protected

// void
// ShapePart3D::_resetAppearance(void) const=0 // Protected

// void
// ShapePart3D::_drawGeometry(void) const=0 // Protected

// void
// ShapePart3D::_rebuildIfNeeded(void) const=0 // Protected

void
ShapePart3D::_render(void) const // Protected
{
if(_useDisplayList)
  {
  glCallList(_appearanceGeometryList); // calls _geometryList
  }
else
  {
  _setAppearance();
  _drawGeometry();
  _resetAppearance();
  }
}

void
ShapePart3D::_renderWithoutAppearance(void) const // Protected
{
if(_useDisplayList)
  {
  glCallList(_geometryList);
  }
else
  {
  _drawGeometry();
  }
}

void
ShapePart3D::_describeGeometry(const Util3D::BoundingBox & bbox,
                               const Util3D::Dbl3 * vertices,
                               unsigned int nbVertices,
                               const Util3D::Face * triangles,
                               unsigned int nbTriangles,
                               unsigned int nbLines,
                               unsigned int nbPoints) // Protected
{
_bbox=bbox;
_vPtr=vertices;
_nbVertices=nbVertices;
_tPtr=triangles;
_nbTriangles=nbTriangles;
_nbLines=nbLines;
_nbPoints=nbPoints;
if(_tree)
  {
  delete _tree;
  _tree=(ShapePart3D::BoxTree *)0;
  }
if(_shape)
  {
  _shape->_needsGeometryUpdate=true;
  _shape->_postRedisplay();
  }
}

void
ShapePart3D::_setTriangleDivision(unsigned int maxTriangles) // Protected
{
if(_triangleDivision!=maxTriangles)
  {
  _triangleDivision=maxTriangles;
  if(_tree)
    {
    delete _tree;
    _tree=(ShapePart3D::BoxTree *)0;
    }
  }
}

double // (interP-rayP)/dirV [0.0;+inf[
ShapePart3D::_rayIntersection(const Util3D::Dbl3 & rayP,
                              const Util3D::Dbl3 & dirV,
                              bool findNearest,
                              unsigned int & triIndexOut) const // Protected
{
if(!_tree)
  {
  const_cast<ShapePart3D *>(this)->_buildBoxTree();
  }
return(_tree ? _tree->rayIntersection(rayP,dirV,findNearest,triIndexOut)
             : -1.0);
}

bool // collision occured
ShapePart3D::_testCollision(const Util3D::Transform & transform,
                            const ShapePart3D * part) const // Protected
{
if(!_tree)
  {
  const_cast<ShapePart3D *>(this)->_buildBoxTree();
  }
if(_tree)
  {
  if(!part->_tree)
    {
    const_cast<ShapePart3D *>(part)->_buildBoxTree();
    }
  if(part->_tree)
    {
    Util3D::Dbl3 * trVertices=(Util3D::Dbl3 *)0;
    bool * trFlags=(bool *)0;
    bool result=_tree->testBoxBox(transform,part->_tree,trVertices,trFlags);
    if(trVertices)
      {
      ArSystem::deallocate(trVertices);
      ArSystem::deallocate(trFlags);
      }
    return(result);
    }
  }
return(false);
}

void
ShapePart3D::_intersect(const Util3D::Transform & transform,
                        const ShapePart3D * part,
                        StlVector<Util3D::TriInter> & itOut) const // Protected
{
if(!_tree)
  {
  const_cast<ShapePart3D *>(this)->_buildBoxTree();
  }
if(_tree)
  {
  if(!part->_tree)
    {
    const_cast<ShapePart3D *>(part)->_buildBoxTree();
    }
  if(part->_tree)
    {
    Util3D::Dbl3 * trVertices=(Util3D::Dbl3 *)0;
    bool * trFlags=(bool *)0;
    _tree->interBoxBox(transform,part->_tree,trVertices,trFlags,itOut);
    if(trVertices)
      {
      ArSystem::deallocate(trVertices);
      ArSystem::deallocate(trFlags);
      }
    }
  }
}

void
ShapePart3D::_buildBoxTree(void) // Protected
{
if(_nbTriangles)
  {
  _tree=new ShapePart3D::BoxTree(this);
  _tree->bbox=_bbox;
  _tree->tri.insert(_tree->tri.end(), // FIXME: duplication in root node ?
                    _tPtr,_tPtr+_nbTriangles);
  _tree->divide(_triangleDivision);
  }
}

void
ShapePart3D::_onTransformation(void) // Protected
{
if(!_transformationCB.empty())
  {
  ShapePart3D::TransformationEvent evt;
  _transformationCB.fire(evt);
  }
}

void
ShapePart3D::_onStructureChange(void) // Protected
{
if(!_structureChangeCB.empty())
  {
  ShapePart3D::StructureChangeEvent evt;
  _structureChangeCB.fire(evt);
  }
}

ShapePart3D::BoxTree::BoxTree(ShapePart3D * owner) // Protected
: part(owner),
  bbox(),
  tri()
{
sub[0]=sub[1]=(ShapePart3D::BoxTree *)0;
}

ShapePart3D::BoxTree::~BoxTree(void) // Protected
{
if(sub[0])
  {
  delete sub[0];
  }
if(sub[1])
  {
  delete sub[1];
  }
}

void
ShapePart3D::BoxTree::divide(unsigned int maxTriangles) // Protected
{
if(maxTriangles&&(tri.size()>maxTriangles))
  {
  const Util3D::Dbl3 * vv=part->_vPtr;
  sub[0]=new ShapePart3D::BoxTree(part);
  sub[1]=new ShapePart3D::BoxTree(part);
  double sx=bbox.sizeX();
  double sy=bbox.sizeY();
  double sz=bbox.sizeZ();
  if((sx>=sy)&&(sx>=sz))
    {
    double cx=3.0*bbox.centerX();
    for(unsigned int i=tri.size();i--;)
      {
      const Util3D::Face & f=tri[i];
      if(vv[f.p1].x+vv[f.p2].x+vv[f.p3].x<=cx)
        {
        sub[0]->tri.push_back(f);
        }
      else
        {
        sub[1]->tri.push_back(f);
        }
      }
    }
  else if((sy>=sx)&&(sy>=sz))
    {
    double cy=3.0*bbox.centerY();
    for(unsigned int i=tri.size();i--;)
      {
      const Util3D::Face & f=tri[i];
      if(vv[f.p1].y+vv[f.p2].y+vv[f.p3].y<=cy)
        {
        sub[0]->tri.push_back(f);
        }
      else
        {
        sub[1]->tri.push_back(f);
        }
      }
    }
  else // (sz>=sx)&&(sz>=sy)
    {
    double cz=3.0*bbox.centerZ();
    for(unsigned int i=tri.size();i--;)
      {
      const Util3D::Face & f=tri[i];
      if(vv[f.p1].z+vv[f.p2].z+vv[f.p3].z<=cz)
        {
        sub[0]->tri.push_back(f);
        }
      else
        {
        sub[1]->tri.push_back(f);
        }
      }
    }
  if(sub[0]->tri.empty()||
     sub[1]->tri.empty())
    {
    delete sub[0];
    delete sub[1];
    sub[0]=sub[1]=(ShapePart3D::BoxTree *)0;
    }
  else
    {
    StlVectorFreeMemory(tri);
    for(unsigned int i=2;i--;)
      {
      for(unsigned int j=sub[i]->tri.size();j--;)
        {
        const Util3D::Face & f=sub[i]->tri[j];
        sub[i]->bbox.addXYZ(vv[f.p1].x,vv[f.p1].y,vv[f.p1].z);
        sub[i]->bbox.addXYZ(vv[f.p2].x,vv[f.p2].y,vv[f.p2].z);
        sub[i]->bbox.addXYZ(vv[f.p3].x,vv[f.p3].y,vv[f.p3].z);
        }
      sub[i]->divide(maxTriangles);
      }
    }
  }
}

double // (interP-rayP)/dirV [0.0;+inf[
ShapePart3D::BoxTree::rayIntersection(const Util3D::Dbl3 & rayP,
                                      const Util3D::Dbl3 & dirV,
                                      bool findNearest,
                                      unsigned int & triIndexOut) const
{                                                                 // Protected
if(tri.empty())
  {
  double k,kmin=DBL_MAX;
  for(unsigned int i=2;i--;)
    {
    if(Util3D::rayBoxIntersection(sub[i]->bbox,rayP,dirV,false)>=0.0)
      {
      unsigned int triIndex;
      k=sub[i]->rayIntersection(rayP,dirV,findNearest,triIndex);
      if((k>=0.0)&&(k<kmin))
        {
        triIndexOut=triIndex;
        if(findNearest)
          {
          kmin=k;
          }
        else
          {
          return(k);
          }
        }
      }
    }
  return(kmin==DBL_MAX ? -1.0 : kmin);
  }
else
  {
  return(Util3D::rayTrianglesIntersection(part->_vPtr,
                                          &tri.front(),tri.size(),
                                          rayP,dirV,findNearest,
                                          triIndexOut));
  }
}

bool // success
ShapePart3D::BoxTree::testBoxBox(const Util3D::Transform & transform,
                                 const ShapePart3D::BoxTree * tree,
                                 Util3D::Dbl3 * & trVertices,
                                 bool * & trFlags) const // Protected
{
if(tri.empty())
  {
  if(tree->tri.empty())
    {
    return((sub[0]->bbox.testHalfApproxCollision(transform,
                                                 tree->sub[0]->bbox)&&
            sub[0]->testBoxBox(transform,tree->sub[0],trVertices,trFlags))||
           (sub[0]->bbox.testHalfApproxCollision(transform,
                                                 tree->sub[1]->bbox)&&
            sub[0]->testBoxBox(transform,tree->sub[1],trVertices,trFlags))||
           (sub[1]->bbox.testHalfApproxCollision(transform,
                                                 tree->sub[0]->bbox)&&
            sub[1]->testBoxBox(transform,tree->sub[0],trVertices,trFlags))||
           (sub[1]->bbox.testHalfApproxCollision(transform,
                                                 tree->sub[1]->bbox)&&
            sub[1]->testBoxBox(transform,tree->sub[1],trVertices,trFlags)));
    }
  else
    {
    if(tree->tri.size()==1)
      {
      if(!trVertices)
        {
        unsigned int size=tree->part->_nbVertices;
        trVertices=ArSystem::allocate<Util3D::Dbl3>(size);
        trFlags=ArSystem::allocate<bool>(size);
        memset(trFlags,0,size*sizeof(bool));
        }
      return(testBoxTri(transform,tree,trVertices,trFlags));
      }
    else
      {
      return((sub[0]->bbox.testHalfApproxCollision(transform,tree->bbox)&&
              sub[0]->testBoxBox(transform,tree,trVertices,trFlags))||
             (sub[1]->bbox.testHalfApproxCollision(transform,tree->bbox)&&
              sub[1]->testBoxBox(transform,tree,trVertices,trFlags)));
      }
    }
  }
else
  {
  if(tree->tri.empty())
    {
    if(tri.size()==1)
      {
      if(!trVertices)
        {
        unsigned int size=tree->part->_nbVertices;
        trVertices=ArSystem::allocate<Util3D::Dbl3>(size);
        trFlags=ArSystem::allocate<bool>(size);
        memset(trFlags,0,size*sizeof(bool));
        }
      return(testTriBox(transform,tree,trVertices,trFlags));
      }
    else
      {
      return((bbox.testHalfApproxCollision(transform,tree->sub[0]->bbox)&&
              testBoxBox(transform,tree->sub[0],trVertices,trFlags))||
             (bbox.testHalfApproxCollision(transform,tree->sub[1]->bbox)&&
              testBoxBox(transform,tree->sub[1],trVertices,trFlags)));
      }
    }
  else
    {
    if(!trVertices)
      {
      unsigned int size=tree->part->_nbVertices;
      trVertices=ArSystem::allocate<Util3D::Dbl3>(size);
      trFlags=ArSystem::allocate<bool>(size);
      memset(trFlags,0,size*sizeof(bool));
      }
    return(testTriTri(transform,tree,trVertices,trFlags));
    }
  }
}

bool // success
ShapePart3D::BoxTree::testTriBox(const Util3D::Transform & transform,
                                 const ShapePart3D::BoxTree * tree,
                                 Util3D::Dbl3 * trVertices,
                                 bool * trFlags) const // Protected
{
if(tree->tri.empty())
  {
  return(testTriBox(transform,tree->sub[0],trVertices,trFlags)||
         testTriBox(transform,tree->sub[1],trVertices,trFlags));

  }
else
  {
  return(testTriTri(transform,tree,trVertices,trFlags));
  }
}

bool // success
ShapePart3D::BoxTree::testBoxTri(const Util3D::Transform & transform,
                                 const ShapePart3D::BoxTree * tree,
                                 Util3D::Dbl3 * trVertices,
                                 bool * trFlags) const // Protected
{
if(tri.empty())
  {
  return(sub[0]->testBoxTri(transform,tree,trVertices,trFlags)||
         sub[1]->testBoxTri(transform,tree,trVertices,trFlags));
  }
else
  {
  return(testTriTri(transform,tree,trVertices,trFlags));
  }
}

bool // success
ShapePart3D::BoxTree::testTriTri(const Util3D::Transform & transform,
                                 const ShapePart3D::BoxTree * tree,
                                 Util3D::Dbl3 * trVertices,
                                 bool * trFlags) const
{                                                                 // Protected
const Util3D::Dbl3 * v1=part->_vPtr;
const Util3D::Dbl3 * v2=tree->part->_vPtr;
for(unsigned int i=tri.size();i--;)
  {
  const Util3D::Face & t1=tri[i];
  for(unsigned int j=tree->tri.size();j--;)
    {
    const Util3D::Face & t2=tree->tri[j];
    if(!trFlags[t2.p1])
      {
      transform.apply(trVertices[t2.p1].x,
                      trVertices[t2.p1].y,
                      trVertices[t2.p1].z,
                      v2[t2.p1].x,v2[t2.p1].y,v2[t2.p1].z);
      trFlags[t2.p1]=true;
      }
    if(!trFlags[t2.p2])
      {
      transform.apply(trVertices[t2.p2].x,
                      trVertices[t2.p2].y,
                      trVertices[t2.p2].z,
                      v2[t2.p2].x,v2[t2.p2].y,v2[t2.p2].z);
      trFlags[t2.p2]=true;
      }
    if(!trFlags[t2.p3])
      {
      transform.apply(trVertices[t2.p3].x,
                      trVertices[t2.p3].y,
                      trVertices[t2.p3].z,
                      v2[t2.p3].x,v2[t2.p3].y,v2[t2.p3].z);
      trFlags[t2.p3]=true;
      }
    if(Util3D::triangleTriangleContact(v1[t1.p1],
                                       v1[t1.p2],
                                       v1[t1.p3],
                                       trVertices[t2.p1],
                                       trVertices[t2.p2],
                                       trVertices[t2.p3]))
      {
      return(true);
      }
    }
  }
return(false);
}

void
ShapePart3D::BoxTree::interBoxBox(const Util3D::Transform & transform,
                                  const ShapePart3D::BoxTree * tree,
                                  Util3D::Dbl3 * & trVertices,
                                  bool * & trFlags,
                                  StlVector<Util3D::TriInter> & itOut) const
{                                                                 // Protected
if(tri.empty())
  {
  if(tree->tri.empty())
    {
    if(sub[0]->bbox.testHalfApproxCollision(transform,tree->sub[0]->bbox))
      {
      sub[0]->interBoxBox(transform,tree->sub[0],trVertices,trFlags,itOut);
      }
    if(sub[0]->bbox.testHalfApproxCollision(transform,tree->sub[1]->bbox))
      {
      sub[0]->interBoxBox(transform,tree->sub[1],trVertices,trFlags,itOut);
      }
    if(sub[1]->bbox.testHalfApproxCollision(transform,tree->sub[0]->bbox))
      {
      sub[1]->interBoxBox(transform,tree->sub[0],trVertices,trFlags,itOut);
      }
    if(sub[1]->bbox.testHalfApproxCollision(transform,tree->sub[1]->bbox))
      {
      sub[1]->interBoxBox(transform,tree->sub[1],trVertices,trFlags,itOut);
      }
    }
  else
    {
    if(tree->tri.size()==1)
      {
      if(!trVertices)
        {
        unsigned int size=tree->part->_nbVertices;
        trVertices=ArSystem::allocate<Util3D::Dbl3>(size);
        trFlags=ArSystem::allocate<bool>(size);
        memset(trFlags,0,size*sizeof(bool));
        }
      interBoxTri(transform,tree,trVertices,trFlags,itOut);
      }
    else
      {
      if(sub[0]->bbox.testHalfApproxCollision(transform,tree->bbox))
        {
        sub[0]->interBoxBox(transform,tree,trVertices,trFlags,itOut);
        }
      if(sub[1]->bbox.testHalfApproxCollision(transform,tree->bbox))
        {
        sub[1]->interBoxBox(transform,tree,trVertices,trFlags,itOut);
        }
      }
    }
  }
else
  {
  if(tree->tri.empty())
    {
    if(tri.size()==1)
      {
      if(!trVertices)
        {
        unsigned int size=tree->part->_nbVertices;
        trVertices=ArSystem::allocate<Util3D::Dbl3>(size);
        trFlags=ArSystem::allocate<bool>(size);
        memset(trFlags,0,size*sizeof(bool));
        }
      interTriBox(transform,tree,trVertices,trFlags,itOut);
      }
    else
      {
      if(bbox.testHalfApproxCollision(transform,tree->sub[0]->bbox))
        {
        interBoxBox(transform,tree->sub[0],trVertices,trFlags,itOut);
        }
      if(bbox.testHalfApproxCollision(transform,tree->sub[1]->bbox))
        {
        interBoxBox(transform,tree->sub[1],trVertices,trFlags,itOut);
        }
      }
    }
  else
    {
    if(!trVertices)
      {
      unsigned int size=tree->part->_nbVertices;
      trVertices=ArSystem::allocate<Util3D::Dbl3>(size);
      trFlags=ArSystem::allocate<bool>(size);
      memset(trFlags,0,size*sizeof(bool));
      }
    interTriTri(transform,tree,trVertices,trFlags,itOut);
    }
  }
}

void
ShapePart3D::BoxTree::interTriBox(const Util3D::Transform & transform,
                                  const ShapePart3D::BoxTree * tree,
                                  Util3D::Dbl3 * trVertices,
                                  bool * trFlags,
                                  StlVector<Util3D::TriInter> & itOut) const
{                                                                 // Protected
if(tree->tri.empty())
  {
  interTriBox(transform,tree->sub[0],trVertices,trFlags,itOut);
  interTriBox(transform,tree->sub[1],trVertices,trFlags,itOut);
  }
else
  {
  interTriTri(transform,tree,trVertices,trFlags,itOut);
  }
}

void
ShapePart3D::BoxTree::interBoxTri(const Util3D::Transform & transform,
                                  const ShapePart3D::BoxTree * tree,
                                  Util3D::Dbl3 * trVertices,
                                  bool * trFlags,
                                  StlVector<Util3D::TriInter> & itOut) const
{                                                                 // Protected
if(tri.empty())
  {
  sub[0]->interBoxTri(transform,tree,trVertices,trFlags,itOut);
  sub[1]->interBoxTri(transform,tree,trVertices,trFlags,itOut);
  }
else
  {
  interTriTri(transform,tree,trVertices,trFlags,itOut);
  }
}

void
ShapePart3D::BoxTree::interTriTri(const Util3D::Transform & transform,
                                  const ShapePart3D::BoxTree * tree,
                                  Util3D::Dbl3 * trVertices,
                                  bool * trFlags,
                                  StlVector<Util3D::TriInter> & itOut) const
{                                                                 // Protected
Util3D::TriInter inter;
const Util3D::Dbl3 * v1=part->_vPtr;
const Util3D::Dbl3 * v2=tree->part->_vPtr;
for(unsigned int i=tri.size();i--;)
  {
  const Util3D::Face & t1=tri[i];
  for(unsigned int j=tree->tri.size();j--;)
    {
    const Util3D::Face & t2=tree->tri[j];
    if(!trFlags[t2.p1])
      {
      transform.apply(trVertices[t2.p1].x,
                      trVertices[t2.p1].y,
                      trVertices[t2.p1].z,
                      v2[t2.p1].x,v2[t2.p1].y,v2[t2.p1].z);
      trFlags[t2.p1]=true;
      }
    if(!trFlags[t2.p2])
      {
      transform.apply(trVertices[t2.p2].x,
                      trVertices[t2.p2].y,
                      trVertices[t2.p2].z,
                      v2[t2.p2].x,v2[t2.p2].y,v2[t2.p2].z);
      trFlags[t2.p2]=true;
      }
    if(!trFlags[t2.p3])
      {
      transform.apply(trVertices[t2.p3].x,
                      trVertices[t2.p3].y,
                      trVertices[t2.p3].z,
                      v2[t2.p3].x,v2[t2.p3].y,v2[t2.p3].z);
      trFlags[t2.p3]=true;
      }
    if(Util3D::triangleTriangleIntersection(v1[t1.p1],
                                            v1[t1.p2],
                                            v1[t1.p3],
                                            trVertices[t2.p1],
                                            trVertices[t2.p2],
                                            trVertices[t2.p3],
                                            inter.i1,
                                            inter.i2))
      {
      inter.ta1=v1[t1.p1];
      inter.ta2=v1[t1.p2];
      inter.ta3=v1[t1.p3];
      inter.tb1=trVertices[t2.p1];
      inter.tb2=trVertices[t2.p2];
      inter.tb3=trVertices[t2.p3];
      itOut.push_back(inter);
      }
    }
  }
}


//----------------------------------------------------------------------------
// Class ShapePartRenderer (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(ShapePartRenderer,ArObject)

//-------- Construction / Destruction ----------------------------------------

ShapePartRenderer::ShapePartRenderer(ArCW & arCW) // Protected
: ArObject(arCW),
  _parts()
{
  // Nothing to be done
}

ShapePartRenderer::~ShapePartRenderer(void)
{
while(!_parts.empty())
  {
  _parts.back()->_postAppearanceRebuild();
  _parts.back()->_postGeometryRebuild();
  _parts.pop_back();
  }
}

//-------- Forbidden operations ----------------------------------------------

// void
// ShapePartRenderer::_render(const ShapePart3D * part,
//                            const Object3D * object,
//                            const Renderer3D * renderer) const=0

// void
// ShapePartRenderer::_renderWithoutAppearance(const ShapePart3D * part,
//                                             const Object3D * object,
//                                             const Renderer3D * renderer)
//                                                                     const=0

const Util3D::Base &
ShapePartRenderer::_getLookAt(const Renderer3D * renderer) // Static, Protected
{
return(renderer->_lookAt);
}

unsigned int
ShapePartRenderer::_getNbActiveLights(const Renderer3D * renderer)
{                                                         // Static, Protected
return(renderer->_lights.size());
}

void
ShapePartRenderer::_renderLight(unsigned int lightNumber,
                                const Light3D * light) // Static, Protected
{
light->_render(lightNumber);
}

void
ShapePartRenderer::_renderHeadLight(unsigned int lightNumber,
                                    const Renderer3D * renderer)
{                                                         // Static, Protected
Light3D::_renderHeadLight(lightNumber,
                          renderer->_base,
                          renderer->_headLightColor.x,
                          renderer->_headLightColor.y,
                          renderer->_headLightColor.z);
}

void
ShapePartRenderer::_renderPart(const ShapePart3D * part) // Static, Protected
{
part->_render();
}

void
ShapePartRenderer::_renderPartWithoutAppearance(const ShapePart3D * part)
{                                                         // Static, Protected
part->_renderWithoutAppearance();
}

void
ShapePartRenderer::_addShapePart(ShapePart3D * part) // Protected
{
_parts.push_back(part);
}

void
ShapePartRenderer::_removeShapePart(ShapePart3D * part) // Protected
{
for(unsigned int i=_parts.size();i--;)
  {
  if(_parts[i]==part)
    {
    StlVectorFastErase(_parts,i);
    break;
    }
  }
}

void
ShapePartRenderer::_postRedisplay(void) // Protected
{
for(unsigned int i=_parts.size();i--;)
  {
  _parts[i]->_postRedisplayIfVisible();
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
