// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/sphere3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/sphere3D.h"

#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Sphere3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Sphere3D,Surface3D)

//-------- Construction / Destruction ----------------------------------------

Sphere3D::Sphere3D(ArCW & arCW) // Protected
: Surface3D(arCW),
  _radius(1.0),
  _bottom(true),
  _ccw(true),
  _precision(5)
{
  // Nothing to be done
}

Sphere3D::~Sphere3D(void) // Protected
{
  // Nothing to be done
}

//-------- Geometry ----------------------------------------------------------

bool // success
Sphere3D::setRadius(double radius)
{
if(radius>0.0)
  {
  _radius=radius;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

double // radius
Sphere3D::getRadius(void) const
{
return(_radius);
}

void
Sphere3D::setBottom(bool buildBottom)
{
_bottom=buildBottom;
_postGeometryRebuild();
}

bool // bottom is built
Sphere3D::getBottom(void) const
{
return(_bottom);
}

void
Sphere3D::setCCW(bool ccw)
{
_ccw=ccw;
_postGeometryRebuild();
}

bool // counter clock wise surface
Sphere3D::getCCW(void) const
{
return(_ccw);
}

bool // success
Sphere3D::setPrecision(unsigned int precision)
{
if(precision)
  {
  _precision=precision;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

unsigned int // precision
Sphere3D::getPrecision(void) const
{
return(_precision);
}

//-------- Serialization operations ------------------------------------------

void
Sphere3D::serialize(SerializationDependencies & dependenciesOut,
                    ArRef<AbstractOStream> output) const
{
Surface3D::serialize(dependenciesOut,output);
output->writeDouble(_radius);
output->writeUInt(_precision);
unsigned int flags=(_bottom ? 1<<0 : 0)+
                   (_ccw    ? 1<<1 : 0);
output->writeInt(flags);
}

bool // success
Sphere3D::unserialize(const SerializationDependencies & dependencies,
                      ArRef<AbstractIStream> input)
{
double radius;
unsigned int precision,flags;
if(!Surface3D::unserialize(dependencies,input)||
   !input->readDouble(radius)||
   !input->readUInt(precision)||
   !input->readUInt(flags))
  {
  return(false);
  }
_radius=radius;
_precision=precision;
_bottom=flags&(1<<0);
_ccw=flags&(1<<1);
_postGeometryRebuild();
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Sphere3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  _vertices.clear();
  _textures.clear();
  _facesV.clear();
  _facesT.clear();
  double xt;
  _vertices.push_back(Util3D::Dbl3(0.0,0.0,_radius));
  if(_bottom)
    {
    _vertices.push_back(Util3D::Dbl3(0.0,0.0,-_radius));
    }
  for(unsigned int i=0;i<4;i++)
    {
    xt=(0.5+(double)i)*0.25;
    if(!_ccw)
      {
      xt=1.0-xt;
      }
    _textures.push_back(Util3D::Dbl2(xt,1.0));
    if(_bottom)
      {
      _textures.push_back(Util3D::Dbl2(1.0-xt,0.0));
      }
    }
  unsigned int row;
  for(row=1;row<=_precision;row++)
    {
    double yt=0.5*(double)row/(double)_precision;
    double zAngle=M_PI*yt;
    double z=_radius*cos(zAngle);
    double radius=_radius*sin(zAngle);
    unsigned int nb=4*row;
    for(unsigned int i=0;i<nb;i++)
      {
      xt=(double)i/(double)nb;
      double xyAngle=M_2PI*(0.5+xt);
      double x=radius*cos(xyAngle);
      double y=radius*sin(xyAngle);
      if(!_ccw)
        {
        x=-x;
        y=-y;
        xt=1.0-xt;
        }
      _vertices.push_back(Util3D::Dbl3(x,y,z));
      _textures.push_back(Util3D::Dbl2(xt,1.0-yt));
      if(_bottom)
        {
        _vertices.push_back(Util3D::Dbl3(x,-y,-z));
        _textures.push_back(Util3D::Dbl2(1.0-xt,yt));
        }
      }
    xt=(_ccw ? 1.0 : 0.0);
    _textures.push_back(Util3D::Dbl2(xt,1.0-yt)); // row end
    if(_bottom)
      {
      _textures.push_back(Util3D::Dbl2(1.0-xt,yt));
      }
    }
  unsigned int delta=(_bottom ? 2 : 1);
  unsigned int p1=delta;
  unsigned int p2=0;
  unsigned int t1=4*delta;
  unsigned int t2=0;
  for(row=1;row<=_precision;row++)
    {
    unsigned int startP1=p1;
    unsigned int startP2=p2;
    unsigned int endP1=startP1+(delta*row*4);
    unsigned int endP2=startP2+(delta*(row-1)*4);
    for(unsigned int j=4;j--;)
      {
      for(unsigned int i=row;i--;)
        {
        if(i)
          {
          unsigned int nextP1=p1+delta;
          if(nextP1==endP1)
            {
            nextP1=startP1;
            }
          if(_ccw)
            {
            _facesV.push_back(Util3D::Face(p1,nextP1,p2));
            _facesT.push_back(Util3D::Face(t1,t1+delta,t2));
            if(_bottom)
              {
              _facesV.push_back(Util3D::Face(p1+1,nextP1+1,p2+1));
              _facesT.push_back(Util3D::Face(t1+1,t1+delta+1,t2+1));
              }
            }
          else
            {
            _facesV.push_back(Util3D::Face(p1,p2,nextP1));
            _facesT.push_back(Util3D::Face(t1,t2,t1+delta));
            if(_bottom)
              {
              _facesV.push_back(Util3D::Face(p1+1,p2+1,nextP1+1));
              _facesT.push_back(Util3D::Face(t1+1,t2+1,t1+delta+1));
              }
            }
          unsigned int nextP2=p2+delta;
          if(nextP2==endP2)
            {
            nextP2=startP2;
            }
          if(_ccw)
            {
            _facesV.push_back(Util3D::Face(p2,nextP1,nextP2));
            _facesT.push_back(Util3D::Face(t2,t1+delta,t2+delta));
            if(_bottom)
              {
              _facesV.push_back(Util3D::Face(p2+1,nextP1+1,nextP2+1));
              _facesT.push_back(Util3D::Face(t2+1,t1+delta+1,t2+delta+1));
              }
            }
          else
            {
            _facesV.push_back(Util3D::Face(p2,nextP2,nextP1));
            _facesT.push_back(Util3D::Face(t2,t2+delta,t1+delta));
            if(_bottom)
              {
              _facesV.push_back(Util3D::Face(p2+1,nextP2+1,nextP1+1));
              _facesT.push_back(Util3D::Face(t2+1,t2+delta+1,t1+delta+1));
              }
            }
          if((row!=1)||!j)
            {
            p2+=delta;
            }
          t2+=delta;
          }
        else
          {
          unsigned int nextP1=p1+delta;
          if(nextP1==endP1)
            {
            nextP1=startP1;
            }
          unsigned int nextP2=p2;
          if(nextP2==endP2)
            {
            nextP2=startP2;
            }
          if(_ccw)
            {
            _facesV.push_back(Util3D::Face(p1,nextP1,nextP2));
            _facesT.push_back(Util3D::Face(t1,t1+delta,t2));
            if(_bottom)
              {
              _facesV.push_back(Util3D::Face(p1+1,nextP1+1,nextP2+1));
              _facesT.push_back(Util3D::Face(t1+1,t1+delta+1,t2+1));
              }
            }
          else
            {
            _facesV.push_back(Util3D::Face(p1,nextP2,nextP1));
            _facesT.push_back(Util3D::Face(t1,t2,t1+delta));
            if(_bottom)
              {
              _facesV.push_back(Util3D::Face(p1+1,nextP2+1,nextP1+1));
              _facesT.push_back(Util3D::Face(t1+1,t2+1,t1+delta+1));
              }
            }
          if(row==1)
            {
            t2+=delta;
            }
          }
        p1+=delta;
        t1+=delta;
        }
      }
    t1+=delta;
    if(row==1)
      {
      p2+=delta;
      }
    else
      {
      t2+=delta;
      }
    }
  if(_ccw)
    {
    _normals=_vertices;
    }
  else
    {
    _normals.clear();
    unsigned int nb=_vertices.size();
    _normals.reserve(nb);
    for(unsigned int i=0;i<nb;i++)
      {
      _normals.push_back(Util3D::Dbl3(-_vertices[i].x,
                                      -_vertices[i].y,
                                      -_vertices[i].z));
      }
    }
  }
Surface3D::_rebuildIfNeeded();
}

} // namespace AReVi

//----------------------------------------------------------------------------
