// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/surface3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/surface3D.h"

#include "AReVi/Lib3D/gl_priv.h"
#include "AReVi/Lib3D/material3D.h"
#include "AReVi/Lib3D/texture.h"
#include "AReVi/Lib3D/transform2D.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Surface3D (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Surface3D,ShapePart3D)

//-------- Construction / Destruction ----------------------------------------

Surface3D::Surface3D(ArCW & arCW) // Protected
: ShapePart3D(arCW),
  _material(),
  _doubleSided(false),
  _texture(),
  _texture2(),
  _texture3(),
  _texture4(),
  _texTransform(true),
  _texTransform2(true),
  _texTransform3(true),
  _texTransform4(true),
  _repeatX(true),
  _repeatY(true),
  _cubeMapReflect(false),
  _repeatX2(true),
  _repeatY2(true),
  _cubeMapReflect2(false),
  _repeatX3(true),
  _repeatY3(true),
  _cubeMapReflect3(false),
  _repeatX4(true),
  _repeatY4(true),
  _cubeMapReflect4(false),
  _vertices(),
  _facesV(),
  _normals(),
  _facesN(),
  _textures(),
  _facesT(),
  _textures2(),
  _facesT2(),
  _textures3(),
  _facesT3(),
  _textures4(),
  _facesT4(),
  _colors(),
  _facesC()
{
_describeGeometry(_bbox,
                  (const Util3D::Dbl3 *)0,0,
                  (const Util3D::Face *)0,0,
                  0,0);
}

Surface3D::~Surface3D(void) // Protected
{
setTexture(Texture::nullRef());
setSecondTexture(Texture::nullRef());
setThirdTexture(Texture::nullRef());
setFourthTexture(Texture::nullRef());
}

//-------- Appearance --------------------------------------------------------

void
Surface3D::writeMaterial(ArConstRef<Material3D> material)
{
if(_material.getLighting()!=material->getMaterial().getLighting())
  {
  _postGeometryRebuild();
  }
_material=material->getMaterial();
_postAppearanceRebuild();
}

void
Surface3D::readMaterial(ArRef<Material3D> materialOut) const
{
materialOut->accessMaterial()=_material;
}

void
Surface3D::setDoubleSided(bool doubleSided)
{
_doubleSided=doubleSided;
_postAppearanceRebuild();
}

bool // double sided surface
Surface3D::getDoubleSided(void) const
{
return(_doubleSided);
}

//-------- Texture management ------------------------------------------------

void
Surface3D::setTexture(ArRef<Texture> texture)
{
if((texture!=_texture)&&!(texture&&texture->fail()))
  {
  ArRef<Texture> old=_texture;
  _texture=texture;
  if(old)
    {
    old->_removeShapePart(this);
    if(!_texture)
      {
      _postGeometryRebuild();
      }
    }
  else
    {
    if(_texture)
      {
      _postGeometryRebuild();
      }
    }
  if(_texture)
    {
    _texture->_addShapePart(this);
    }
  _postAppearanceRebuild();
  }
}

ArConstRef<Texture> // texture
Surface3D::getTexture(void) const
{
return(_texture);
}

ArRef<Texture> // texture
Surface3D::accessTexture(void)
{
return(_texture);
}

void
Surface3D::setTextureRepeat(bool repeatX,
                            bool repeatY)
{
_repeatX=repeatX;
_repeatY=repeatY;
_postAppearanceRebuild();
}

void
Surface3D::getTextureRepeat(bool & repeatXOut,
                            bool & repeatYOut) const
{
repeatXOut=_repeatX;
repeatYOut=_repeatY;
}

void
Surface3D::writeTextureTransformation(ArConstRef<Transform2D> transform)
{
_texTransform=transform->getTransformation();
_postAppearanceRebuild();
}

void
Surface3D::readTextureTransformation(ArRef<Transform2D> transformOut) const
{
transformOut->accessTransformation()=_texTransform;
}

void // Consistent only for CubeMapTextures
Surface3D::setTextureReflection(bool reflectMode)
{
_cubeMapReflect=reflectMode;
_postAppearanceRebuild();
}

bool
Surface3D::getTextureReflection(void) const
{
return(_cubeMapReflect);
}

void
Surface3D::setSecondTexture(ArRef<Texture> texture)
{
if((texture!=_texture2)&&!(texture&&texture->fail()))
  {
  ArRef<Texture> old=_texture2;
  _texture2=texture;
  if(old)
    {
    old->_removeShapePart(this);
    if(!texture)
      {
      _postGeometryRebuild();
      }
    }
  else
    {
    if(texture)
      {
      _postGeometryRebuild();
      }
    }
  if(_texture2)
    {
    _texture2->_addShapePart(this);
    }
  _postAppearanceRebuild();
  }
}

ArConstRef<Texture> // second texture
Surface3D::getSecondTexture(void) const
{
return(_texture2);
}

ArRef<Texture> // second texture
Surface3D::accessSecondTexture(void)
{
return(_texture2);
}

void
Surface3D::setSecondTextureRepeat(bool repeatX,
                                  bool repeatY)
{
_repeatX2=repeatX;
_repeatY2=repeatY;
_postAppearanceRebuild();
}

void
Surface3D::getSecondTextureRepeat(bool & repeatXOut,
                                  bool & repeatYOut) const
{
repeatXOut=_repeatX2;
repeatYOut=_repeatY2;
}

void
Surface3D::writeSecondTextureTransformation(ArConstRef<Transform2D> transform)
{
_texTransform2=transform->getTransformation();
_postAppearanceRebuild();
}

void
Surface3D::readSecondTextureTransformation(ArRef<Transform2D> transformOut)
                                                                         const
{
transformOut->accessTransformation()=_texTransform2;
}

void // Consistent only for CubeMapTextures
Surface3D::setSecondTextureReflection(bool reflectMode)
{
_cubeMapReflect2=reflectMode;
_postAppearanceRebuild();
}

bool
Surface3D::getSecondTextureReflection(void) const
{
return(_cubeMapReflect2);
}

void
Surface3D::setThirdTexture(ArRef<Texture> texture)
{
if((texture!=_texture3)&&!(texture&&texture->fail()))
  {
  ArRef<Texture> old=_texture3;
  _texture3=texture;
  if(old)
    {
    old->_removeShapePart(this);
    if(!texture)
      {
      _postGeometryRebuild();
      }
    }
  else
    {
    if(texture)
      {
      _postGeometryRebuild();
      }
    }
  if(_texture3)
    {
    _texture3->_addShapePart(this);
    }
  _postAppearanceRebuild();
  }
}

ArConstRef<Texture> // third texture
Surface3D::getThirdTexture(void) const
{
return(_texture3);
}

ArRef<Texture> // third texture
Surface3D::accessThirdTexture(void)
{
return(_texture3);
}

void
Surface3D::setThirdTextureRepeat(bool repeatX,
                                 bool repeatY)
{
_repeatX3=repeatX;
_repeatY3=repeatY;
_postAppearanceRebuild();
}

void
Surface3D::getThirdTextureRepeat(bool & repeatXOut,
                                 bool & repeatYOut) const
{
repeatXOut=_repeatX3;
repeatYOut=_repeatY3;
}

void
Surface3D::writeThirdTextureTransformation(ArConstRef<Transform2D> transform)
{
_texTransform3=transform->getTransformation();
_postAppearanceRebuild();
}

void
Surface3D::readThirdTextureTransformation(ArRef<Transform2D> transformOut)
                                                                         const
{
transformOut->accessTransformation()=_texTransform3;
}

void // Consistent only for CubeMapTextures
Surface3D::setThirdTextureReflection(bool reflectMode)
{
_cubeMapReflect3=reflectMode;
_postAppearanceRebuild();
}

bool
Surface3D::getThirdTextureReflection(void) const
{
return(_cubeMapReflect3);
}

void
Surface3D::setFourthTexture(ArRef<Texture> texture)
{
if((texture!=_texture4)&&!(texture&&texture->fail()))
  {
  ArRef<Texture> old=_texture4;
  _texture4=texture;
  if(old)
    {
    old->_removeShapePart(this);
    if(!texture)
      {
      _postGeometryRebuild();
      }
    }
  else
    {
    if(texture)
      {
      _postGeometryRebuild();
      }
    }
  if(_texture4)
    {
    _texture4->_addShapePart(this);
    }
  _postAppearanceRebuild();
  }
}

ArConstRef<Texture> // fourth texture
Surface3D::getFourthTexture(void) const
{
return(_texture4);
}

ArRef<Texture> // fourth texture
Surface3D::accessFourthTexture(void)
{
return(_texture4);
}

void
Surface3D::setFourthTextureRepeat(bool repeatX,
                                  bool repeatY)
{
_repeatX4=repeatX;
_repeatY4=repeatY;
_postAppearanceRebuild();
}

void
Surface3D::getFourthTextureRepeat(bool & repeatXOut,
                                  bool & repeatYOut) const
{
repeatXOut=_repeatX4;
repeatYOut=_repeatY4;
}

void
Surface3D::writeFourthTextureTransformation(ArConstRef<Transform2D> transform)
{
_texTransform4=transform->getTransformation();
_postAppearanceRebuild();
}

void
Surface3D::readFourthTextureTransformation(ArRef<Transform2D> transformOut)
                                                                         const
{
transformOut->accessTransformation()=_texTransform4;
}

void // Consistent only for CubeMapTextures
Surface3D::setFourthTextureReflection(bool reflectMode)
{
_cubeMapReflect4=reflectMode;
_postAppearanceRebuild();
}

bool
Surface3D::getFourthTextureReflection(void) const
{
return(_cubeMapReflect4);
}

//-------- Vertices ----------------------------------------------------------

const StlVector<Util3D::Dbl3> & // vertex coords
Surface3D::getVertices(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_vertices);
}

const StlVector<Util3D::Face> & // vertex faces
Surface3D::getVertexFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesV);
}

//-------- Normals -----------------------------------------------------------

const StlVector<Util3D::Dbl3> & // normal coords
Surface3D::getNormals(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_normals);
}

const StlVector<Util3D::Face> & // normal faces
Surface3D::getNormalFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesN);
}

//-------- Textures ----------------------------------------------------------

const StlVector<Util3D::Dbl2> & // texture coords
Surface3D::getTextures(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_textures);
}

const StlVector<Util3D::Face> & // texture faces
Surface3D::getTextureFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesT);
}

const StlVector<Util3D::Dbl2> & // second texture coords
Surface3D::getSecondTextures(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_textures2);
}

const StlVector<Util3D::Face> & // second texture faces
Surface3D::getSecondTextureFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesT2);
}

const StlVector<Util3D::Dbl2> & // third texture coords
Surface3D::getThirdTextures(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_textures3);
}

const StlVector<Util3D::Face> & // third texture faces
Surface3D::getThirdTextureFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesT3);
}

const StlVector<Util3D::Dbl2> & // fourth texture coords
Surface3D::getFourthTextures(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_textures4);
}

const StlVector<Util3D::Face> & // fourth texture faces
Surface3D::getFourthTextureFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesT4);
}

//-------- Colors ------------------------------------------------------------

const StlVector<Util3D::Dbl3> & // color components
Surface3D::getColors(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_colors);
}

const StlVector<Util3D::Face> & // color faces
Surface3D::getColorFaces(void) const
{
const_cast<Surface3D *>(this)->_rebuildIfNeeded();
return(_facesC);
}

//-------- Serialization operations ------------------------------------------

void
Surface3D::extractDependencies(SerializationDependencies & dependenciesOut,
                               bool recursive) const
{
ShapePart3D::extractDependencies(dependenciesOut,recursive);
if(_texture)
  {
  if(recursive)
    {
    _texture->extractDependencies(dependenciesOut,recursive);
    }
  dependenciesOut.addReference(_texture);
  }
if(_texture2)
  {
  if(recursive)
    {
    _texture2->extractDependencies(dependenciesOut,recursive);
    }
  dependenciesOut.addReference(_texture2);
  }
if(_texture3)
  {
  if(recursive)
    {
    _texture3->extractDependencies(dependenciesOut,recursive);
    }
  dependenciesOut.addReference(_texture3);
  }
if(_texture4)
  {
  if(recursive)
    {
    _texture4->extractDependencies(dependenciesOut,recursive);
    }
  dependenciesOut.addReference(_texture4);
  }
}

void
Surface3D::serialize(SerializationDependencies & dependenciesOut,
                     ArRef<AbstractOStream> output) const
{
ShapePart3D::serialize(dependenciesOut, output);
ArRef<Material3D> material=Material3D::NEW();
readMaterial(material);
material->writeToStream(output);
bool hasTexTransform=!_texTransform.isIdentity(false);
bool hasTexTransform2=!_texTransform2.isIdentity(false);
bool hasTexTransform3=!_texTransform3.isIdentity(false);
bool hasTexTransform4=!_texTransform4.isIdentity(false);
unsigned int flags=(_doubleSided     ? 1<<0  : 0)+
                   (_texture         ? 1<<1  : 0)+
                   (_texture2        ? 1<<2  : 0)+
                   (_texture3        ? 1<<3  : 0)+
                   (_texture4        ? 1<<4  : 0)+
                   (hasTexTransform  ? 1<<5  : 0)+
                   (hasTexTransform2 ? 1<<6  : 0)+
                   (hasTexTransform3 ? 1<<7  : 0)+
                   (hasTexTransform4 ? 1<<8  : 0)+
                   (_repeatX         ? 1<<9  : 0)+
                   (_repeatY         ? 1<<10 : 0)+
                   (_repeatX2        ? 1<<11 : 0)+
                   (_repeatY2        ? 1<<12 : 0)+
                   (_repeatX3        ? 1<<13 : 0)+
                   (_repeatY3        ? 1<<14 : 0)+
                   (_repeatX4        ? 1<<15 : 0)+
                   (_repeatY4        ? 1<<16 : 0)+
                   (_cubeMapReflect  ? 1<<17 : 0)+
                   (_cubeMapReflect2 ? 1<<18 : 0)+
                   (_cubeMapReflect3 ? 1<<19 : 0)+
                   (_cubeMapReflect4 ? 1<<20 : 0);
output->writeUInt(flags);
if(_texture)
  {
  output->writeUInt(dependenciesOut.addReference(_texture));
  }
if(_texture2)
  {
  output->writeUInt(dependenciesOut.addReference(_texture2));
  }
if(_texture3)
  {
  output->writeUInt(dependenciesOut.addReference(_texture3));
  }
if(_texture4)
  {
  output->writeUInt(dependenciesOut.addReference(_texture4));
  }
unsigned int i;
if(hasTexTransform)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_texTransform.matrix[i]);
    }
  }
if(hasTexTransform2)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_texTransform2.matrix[i]);
    }
  }
if(hasTexTransform3)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_texTransform3.matrix[i]);
    }
  }
if(hasTexTransform4)
  {
  for(i=16;i--;)
    {
    output->writeDouble(_texTransform4.matrix[i]);
    }
  }
}

bool // success
Surface3D::unserialize(const SerializationDependencies & dependencies,
                       ArRef<AbstractIStream> input)
{
ArRef<Material3D> material=Material3D::NEW();
unsigned int flags,texture,texture2,texture3,texture4;
double matrix[16];
double matrix2[16];
double matrix3[16];
double matrix4[16];
if(!ShapePart3D::unserialize(dependencies,input)||
   !material->readFromStream(input)||
   !input->readUInt(flags)||
   ((flags&(1<<1))&&!input->readUInt(texture))||
   ((flags&(1<<2))&&!input->readUInt(texture2))||
   ((flags&(1<<3))&&!input->readUInt(texture3))||
   ((flags&(1<<4))&&!input->readUInt(texture4)))
  {
  return(false);
  }
unsigned int i;
bool hasTexTransform=flags&(1<<5);
bool hasTexTransform2=flags&(1<<6);
bool hasTexTransform3=flags&(1<<7);
bool hasTexTransform4=flags&(1<<8);
if(hasTexTransform)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix[i]))
      {
      return(false);
      }
    }
  }
if(hasTexTransform2)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix2[i]))
      {
      return(false);
      }
    }
  }
if(hasTexTransform3)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix3[i]))
      {
      return(false);
      }
    }
  }
if(hasTexTransform4)
  {
  for(i=16;i--;)
    {
    if(!input->readDouble(matrix4[i]))
      {
      return(false);
      }
    }
  }
_repeatX=flags&(1<<9);
_repeatY=flags&(1<<10);
_cubeMapReflect=flags&(1<<17);
_repeatX2=flags&(1<<11);
_repeatY2=flags&(1<<12);
_cubeMapReflect2=flags&(1<<18);
_repeatX3=flags&(1<<13);
_repeatY3=flags&(1<<14);
_cubeMapReflect3=flags&(1<<19);
_repeatX4=flags&(1<<15);
_repeatY4=flags&(1<<16);
_cubeMapReflect4=flags&(1<<20);
_doubleSided=flags&(1<<0);
if(hasTexTransform)
  {
  _texTransform.setMatrix(matrix);
  }
else
  {
  _texTransform.identity();
  }
if(hasTexTransform2)
  {
  _texTransform2.setMatrix(matrix2);
  }
else
  {
  _texTransform2.identity();
  }
if(hasTexTransform3)
  {
  _texTransform3.setMatrix(matrix3);
  }
else
  {
  _texTransform3.identity();
  }
if(hasTexTransform4)
  {
  _texTransform4.setMatrix(matrix4);
  }
else
  {
  _texTransform4.identity();
  }
if(flags&(1<<1))
  {
  setTexture(ar_down_cast<Texture>(dependencies.findReference(texture)));
  }
else
  {
  setTexture(Texture::nullRef());
  }
if(flags&(1<<2))
  {
  setSecondTexture(ar_down_cast<Texture>
                               (dependencies.findReference(texture2)));
  }
else
  {
  setSecondTexture(Texture::nullRef());
  }
if(flags&(1<<3))
  {
  setThirdTexture(ar_down_cast<Texture>
                              (dependencies.findReference(texture3)));
  }
else
  {
  setThirdTexture(Texture::nullRef());
  }
if(flags&(1<<4))
  {
  setFourthTexture(ar_down_cast<Texture>
                               (dependencies.findReference(texture4)));
  }
else
  {
  setFourthTexture(Texture::nullRef());
  }
writeMaterial(material);
return(true);
}

//-------- Forbidden operations ----------------------------------------------


void
Surface3D::_setAppearance(void) const // Protected
{
if(_offset)
  {
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(_offset,_offset);
  }
if(_doubleSided)
  {
  glDisable(GL_CULL_FACE);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
  }
bool multiTexture=false;
if(_texture4&&(GL::getNbTextureUnits()>3))
  {
  multiTexture=true;
  GL::glActiveTexture(GL_TEXTURE3_ARB);
  if(_texture4->usesCubemap())
    {
    glEnable(GL_TEXTURE_CUBE_MAP_EXT);
    GL::selectCubeTexture(_texture4->getTextureID(),
                          _repeatX4,_repeatY4,true,_cubeMapReflect4);
    }
  else
    {
    glEnable(GL_TEXTURE_2D);
    GL::selectTexture(_texture4->getTextureID(),_repeatX4,_repeatY4,true);
    }
  if(!_texTransform4.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadMatrixd(_texTransform4.matrix);
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(_texture3&&(GL::getNbTextureUnits()>2))
  {
  multiTexture=true;
  GL::glActiveTexture(GL_TEXTURE2_ARB);
  if(_texture3->usesCubemap())
    {
    glEnable(GL_TEXTURE_CUBE_MAP_EXT);
    GL::selectCubeTexture(_texture3->getTextureID(),
                          _repeatX3,_repeatY3,true,_cubeMapReflect3);
    }
  else
    {
    glEnable(GL_TEXTURE_2D);
    GL::selectTexture(_texture3->getTextureID(),_repeatX3,_repeatY3,true);
    }
  if(!_texTransform3.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadMatrixd(_texTransform3.matrix);
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(_texture2&&(GL::getNbTextureUnits()>1))
  {
  multiTexture=true;
  GL::glActiveTexture(GL_TEXTURE1_ARB);
  if(_texture2->usesCubemap())
    {
    glEnable(GL_TEXTURE_CUBE_MAP_EXT);
    GL::selectCubeTexture(_texture2->getTextureID(),
                          _repeatX2,_repeatY2,true,_cubeMapReflect2);
    }
  else
    {
    glEnable(GL_TEXTURE_2D);
    GL::selectTexture(_texture2->getTextureID(),_repeatX2,_repeatY2,true);
    }
  if(!_texTransform2.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadMatrixd(_texTransform2.matrix);
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(multiTexture)
  {
  GL::glActiveTexture(GL_TEXTURE0_ARB);
  }
if(_texture)
  {
  if(_texture->usesCubemap())
    {
    glEnable(GL_TEXTURE_CUBE_MAP_EXT);
    GL::selectCubeTexture(_texture->getTextureID(),
                          _repeatX,_repeatY,true,_cubeMapReflect);
    }
  else
    {
    glEnable(GL_TEXTURE_2D);
    GL::selectTexture(_texture->getTextureID(),_repeatX,_repeatY,true);
    }
  if(!_texTransform.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadMatrixd(_texTransform.matrix);
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(_material.getLighting())
  {
  glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,_material.getAmbientColor());
  glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,_material.getDiffuseColor());
  glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,_material.getEmissiveColor());
  glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,_material.getSpecularColor());
  glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,128.0*_material.getShininess());
  }
else
  {
  glDisable(GL_LIGHTING);
  glColor4fv(_material.getEmissiveColor());
  }
}

void
Surface3D::_resetAppearance(void) const // Protected
{
if(!_material.getLighting())
  {
  glEnable(GL_LIGHTING);
  }
bool multiTexture=false;
if(_texture4&&(GL::getNbTextureUnits()>3))
  {
  multiTexture=true;
  GL::glActiveTexture(GL_TEXTURE3_ARB);
  if(_texture4->usesCubemap())
    {
    glDisable(GL_TEXTURE_CUBE_MAP_EXT);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
    }
  else
    {
    glDisable(GL_TEXTURE_2D);
    }
  if(!_texTransform4.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(_texture3&&(GL::getNbTextureUnits()>2))
  {
  multiTexture=true;
  GL::glActiveTexture(GL_TEXTURE2_ARB);
  if(_texture3->usesCubemap())
    {
    glDisable(GL_TEXTURE_CUBE_MAP_EXT);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
    }
  else
    {
    glDisable(GL_TEXTURE_2D);
    }
  if(!_texTransform3.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(_texture2&&(GL::getNbTextureUnits()>1))
  {
  multiTexture=true;
  GL::glActiveTexture(GL_TEXTURE1_ARB);
  if(_texture2->usesCubemap())
    {
    glDisable(GL_TEXTURE_CUBE_MAP_EXT);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
    }
  else
    {
    glDisable(GL_TEXTURE_2D);
    }
  if(!_texTransform2.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(multiTexture)
  {
  GL::glActiveTexture(GL_TEXTURE0_ARB);
  }
if(_texture)
  {
  if(_texture->usesCubemap())
    {
    glDisable(GL_TEXTURE_CUBE_MAP_EXT);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
    }
  else
    {
    glDisable(GL_TEXTURE_2D);
    }
  if(!_texTransform.isIdentity(false))
    {
    glMatrixMode(GL_TEXTURE);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    }
  }
if(_doubleSided)
  {
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
  glEnable(GL_CULL_FACE);
  }
if(_offset)
  {
  glDisable(GL_POLYGON_OFFSET_FILL);
  }
}

void
Surface3D::_drawGeometry(void) const // Protected
{
unsigned int nbV=_facesV.size()*3;
if(nbV)
  {
  const unsigned int * indexV=&_facesV.front().p1;
  const unsigned int * indexC=(_colors.empty()
                               ? (const unsigned int *)0
                               : (_facesC.empty()
                                  ? indexV
                                  : &_facesC.front().p1));
  const unsigned int * indexN=(_material.getLighting()
                               ? (_facesN.empty()
                                  ? indexV
                                  : &_facesN.front().p1)
                               : (const unsigned int *)0);
  const unsigned int * indexT=(_texture
                               ? (_facesT.empty()
                                  ? indexV
                                  : &_facesT.front().p1)
                               : (const unsigned int *)0);
  const StlVector<Util3D::Dbl2> & t2=(_textures2.empty() ? _textures
                                                         : _textures2);
  const unsigned int * indexT2=(_texture2&&(GL::getNbTextureUnits()>1)
                                ? (_facesT2.empty()
                                   ? indexT
                                   : &_facesT2.front().p1)
                                : (const unsigned int *)0);
  const StlVector<Util3D::Dbl2> & t3=(_textures3.empty() ? _textures
                                                         : _textures3);
  const unsigned int * indexT3=(_texture3&&(GL::getNbTextureUnits()>2)
                                ? (_facesT3.empty()
                                   ? indexT
                                   : &_facesT3.front().p1)
                                : (const unsigned int *)0);
  const StlVector<Util3D::Dbl2> & t4=(_textures4.empty() ? _textures
                                                         : _textures4);
  const unsigned int * indexT4=(_texture4&&(GL::getNbTextureUnits()>3)
                                ? (_facesT4.empty()
                                   ? indexT
                                   : &_facesT4.front().p1)
                                : (const unsigned int *)0);
  if(indexN)
    {
    if(indexC)
      {
      float defaultMaterial[]={0.0,0.0,0.0,1.0};
      glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,defaultMaterial);
      glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,defaultMaterial);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,defaultMaterial);
      glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,0.0);
      glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
      glEnable(GL_COLOR_MATERIAL);
      }
    }
  else
    {
    glDisable(GL_LIGHTING);
    }
  if((!indexN||(indexN==indexV))&&
     (!indexC||(indexC==indexV))&&
     (!indexT||(indexT==indexV))&&
     (!indexT2||(indexT2==indexV))&&
     (!indexT3||(indexT3==indexV))&&
     (!indexT4||(indexT4==indexV)))
    {
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3,GL_DOUBLE,0,&(_vertices[0].x));
    if(indexC)
      {
      glEnableClientState(GL_COLOR_ARRAY); // FIXME: alpha ?
      glColorPointer(3,GL_DOUBLE,sizeof(Util3D::Dbl3),&(_colors[0].x));
      }
    if(indexN)
      {
      glEnableClientState(GL_NORMAL_ARRAY);
      glNormalPointer(GL_DOUBLE,0,&(_normals[0].x));
      }
    bool multiTexture=false;
    if(indexT4)
      {
      multiTexture=true;
      GL::glClientActiveTexture(GL_TEXTURE3_ARB);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_DOUBLE,0,&(t4[0].x));
      }
    if(indexT3)
      {
      multiTexture=true;
      GL::glClientActiveTexture(GL_TEXTURE2_ARB);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_DOUBLE,0,&(t3[0].x));
      }
    if(indexT2)
      {
      multiTexture=true;
      GL::glClientActiveTexture(GL_TEXTURE1_ARB);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_DOUBLE,0,&(t2[0].x));
      }
    if(multiTexture)
      {
      GL::glClientActiveTexture(GL_TEXTURE0_ARB);
      }
    if(indexT)
      {
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(2,GL_DOUBLE,0,&(_textures[0].x));
      }
    if(GL::glDrawRangeElements)
      {
      GL::glDrawRangeElements(GL_TRIANGLES,0,_vertices.size()-1,
                              nbV,GL_UNSIGNED_INT,indexV);
      }
    else
      {
      glDrawElements(GL_TRIANGLES,nbV,GL_UNSIGNED_INT,indexV);
      }
    glDisableClientState(GL_VERTEX_ARRAY);
    if(indexC)
      {
      glDisableClientState(GL_COLOR_ARRAY);
      }
    if(indexN)
      {
      glDisableClientState(GL_NORMAL_ARRAY);
      }
    if(indexT4)
      {
      GL::glClientActiveTexture(GL_TEXTURE3_ARB);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
    if(indexT3)
      {
      GL::glClientActiveTexture(GL_TEXTURE2_ARB);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
    if(indexT2)
      {
      GL::glClientActiveTexture(GL_TEXTURE1_ARB);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
    if(multiTexture)
      {
      GL::glClientActiveTexture(GL_TEXTURE0_ARB);
      }
    if(indexT)
      {
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
    }
  else
    {
    unsigned int lastColor=UINT_MAX;
    unsigned int lastNormal=UINT_MAX;
    glBegin(GL_TRIANGLES);
    for(unsigned int i=0;i<nbV;i++)
      {
      // if(indexC&&(indexC[i]!=lastColor))
      if(indexC) // seems to be more efficient with redundancy !
        {
        lastColor=indexC[i];
        glColor4f(_colors[lastColor].x,
                  _colors[lastColor].y,
                  _colors[lastColor].z,
                  1.0-_material.getTransparency());
        }
      // if(indexN&&(indexN[i]!=lastNormal))
      if(indexN) // seems to be more efficient with redundancy !
        {
        lastNormal=indexN[i];
        glNormal3dv(&_normals[lastNormal].x);
        }
      if(indexT)
        {
        glTexCoord2dv(&_textures[indexT[i]].x);
        }
      if(indexT2)
        {
        GL::glMultiTexCoord2dv(GL_TEXTURE1_ARB,&t2[indexT2[i]].x);
        }
      if(indexT3)
        {
        GL::glMultiTexCoord2dv(GL_TEXTURE2_ARB,&t3[indexT3[i]].x);
        }
      if(indexT4)
        {
        GL::glMultiTexCoord2dv(GL_TEXTURE3_ARB,&t4[indexT4[i]].x);
        }
      const Util3D::Dbl3 & vertex=_vertices[indexV[i]];
      glVertex3dv(&vertex.x);
      }
    glEnd(); // GL_TRIANGLES
    }
  if(indexN)
    {
    if(indexC)
      {
      glDisable(GL_COLOR_MATERIAL);
      }
    }
  else
    {
    glEnable(GL_LIGHTING);
    }
  }
}

void
Surface3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_geometryList,GL_COMPILE);
    _drawGeometry();
    glEndList();
    }
  Util3D::BoundingBox bbox;
  for(unsigned int i=_vertices.size();i--;)
    {
    const Util3D::Dbl3 & vertex=_vertices[i];
    bbox.addXYZ(vertex.x,vertex.y,vertex.z);
    }
  _describeGeometry(bbox,
                    &_vertices.front(),_vertices.size(),
                    &_facesV.front(),_facesV.size(),
                    0,0);
  _needsGeometryRebuild=false;
  }
if(_needsAppearanceRebuild)
  {
  if(_useDisplayList)
    {
    glNewList(_appearanceGeometryList,GL_COMPILE);
    _setAppearance();
    glCallList(_geometryList);
    _resetAppearance();
    glEndList();
    }
  _needsBlending=(_material.getTransparency()>=0.01);
  _needsAlphaTest=false;
  if(_texture)
    {
    _needsBlending|=_texture->getBlending();
    _needsAlphaTest|=_texture->getAlphaTest();
    }
  if(_texture2&&(GL::getNbTextureUnits()>1))
    {
    _needsBlending|=_texture2->getBlending();
    _needsAlphaTest|=_texture2->getAlphaTest();
    }
  if(_texture3&&(GL::getNbTextureUnits()>2))
    {
    _needsBlending|=_texture3->getBlending();
    _needsAlphaTest|=_texture3->getAlphaTest();
    }
  if(_texture4&&(GL::getNbTextureUnits()>3))
    {
    _needsBlending|=_texture4->getBlending();
    _needsAlphaTest|=_texture4->getAlphaTest();
    }
  _needsAppearanceRebuild=false;
  }
}

void
Surface3D::_computeDefaultNormals(bool smooth) // protected
{
_normals.clear();
_facesN.clear();
if(smooth)
  {
  unsigned int i;
  unsigned int nb=_facesV.size();
  Util3D::Dbl3 * tmpN=ArSystem::allocate<Util3D::Dbl3>(nb);
  for(i=nb;i--;)
    {
    double dx1=_vertices[_facesV[i].p2].x-_vertices[_facesV[i].p1].x;
    double dy1=_vertices[_facesV[i].p2].y-_vertices[_facesV[i].p1].y;
    double dz1=_vertices[_facesV[i].p2].z-_vertices[_facesV[i].p1].z;
    double dx2=_vertices[_facesV[i].p3].x-_vertices[_facesV[i].p1].x;
    double dy2=_vertices[_facesV[i].p3].y-_vertices[_facesV[i].p1].y;
    double dz2=_vertices[_facesV[i].p3].z-_vertices[_facesV[i].p1].z;
    double dx3=dy1*dz2-dz1*dy2;
    double dy3=dx2*dz1-dz2*dx1;
    double dz3=dx1*dy2-dy1*dx2;
    double r=sqrt(dx3*dx3+dy3*dy3+dz3*dz3);
    r=(r? 1.0/r : 1.0);
    tmpN[i].x=dx3*r;
    tmpN[i].y=dy3*r;
    tmpN[i].z=dz3*r;
    }
  StlVector<unsigned int> * usedNormals;
  usedNormals=new StlVector<unsigned int>[_vertices.size()];
  for(i=nb;i--;)
    {
    usedNormals[_facesV[i].p1].push_back(i);
    usedNormals[_facesV[i].p2].push_back(i);
    usedNormals[_facesV[i].p3].push_back(i);
    }
  nb=_vertices.size();
  _normals.reserve(nb);
  for(i=0;i<nb;i++)
    {
    Util3D::Dbl3 norm(0.0,0.0,0.0);
    for(unsigned int j=usedNormals[i].size();j--;)
      {
      norm.x+=tmpN[usedNormals[i][j]].x;
      norm.y+=tmpN[usedNormals[i][j]].y;
      norm.z+=tmpN[usedNormals[i][j]].z;
      }
    _normals.push_back(norm);
    }
  delete[] usedNormals;
  ArSystem::deallocate(tmpN);
  // _facesV will be used instead of _facesN
  }
else
  {
  unsigned int nb=_facesV.size();
  _normals.reserve(nb);
  _facesN.reserve(nb);
  for(unsigned int i=0;i<nb;i++)
    {
    double dx1=_vertices[_facesV[i].p2].x-_vertices[_facesV[i].p1].x;
    double dy1=_vertices[_facesV[i].p2].y-_vertices[_facesV[i].p1].y;
    double dz1=_vertices[_facesV[i].p2].z-_vertices[_facesV[i].p1].z;
    double dx2=_vertices[_facesV[i].p3].x-_vertices[_facesV[i].p1].x;
    double dy2=_vertices[_facesV[i].p3].y-_vertices[_facesV[i].p1].y;
    double dz2=_vertices[_facesV[i].p3].z-_vertices[_facesV[i].p1].z;
    _normals.push_back(Util3D::Dbl3(dy1*dz2-dz1*dy2,
                                    dx2*dz1-dz2*dx1,
                                    dx1*dy2-dy1*dx2));
    _facesN.push_back(Util3D::Face(i,i,i));
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
