// LI2/ENIB - AReVi - 3D shapes
//----------------------------------------------------------------------------
// file        : AReVi/Shapes/text3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Shapes/text3D.h"

#include "AReVi/Lib3D/llw_priv.h"
#include "AReVi/Lib3D/material3D.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/abstractStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Text3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(Text3D,Surface3D)

//-------- Construction / Destruction ----------------------------------------

Text3D::Text3D(ArCW & arCW) // Protected
: Surface3D(arCW),
  _fontURL(),
  _font((LLW::GLFontID)0),
  _symbolSpace(0.1),
  _whiteSpace(0.5),
  _text(),
  _extrusion(0.0)
{
  // Nothing to be done
}

Text3D::~Text3D(void) // Protected
{
if(_font)
  {
  LLW::freeGLFont(_font);
  }
}

//-------- Text --------------------------------------------------------------

bool // success
Text3D::setFont(const StlString & fontURL)
{
LLW::FontID font=LLW::loadGLFont(fontURL);
if(font)
  {
  if(_font)
    {
    LLW::freeGLFont(_font);
    }
  _font=font;
  _fontURL=fontURL;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

const StlString & // font URL
Text3D::getFont(void) const
{
return(_fontURL);
}

bool // success
Text3D::setSpaces(double symbolSpace,
                  double whiteSpace)
{
if((symbolSpace>=0.0)&&(whiteSpace>=0.0))
  {
  _symbolSpace=symbolSpace;
  _whiteSpace=whiteSpace;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

void
Text3D::getSpaces(double & symbolSpaceOut,
                  double & whiteSpaceOut) const
{
symbolSpaceOut=_symbolSpace;
whiteSpaceOut=_whiteSpace;
}

void
Text3D::setText(const StlString & text)
{
_text=text;
_postGeometryRebuild();
}

const StlString &
Text3D::getText(void) const
{
return(_text);
}

bool // success
Text3D::setExtrusion(double extrusion)
{
if(extrusion>=0.0)
  {
  _extrusion=extrusion;
  _postGeometryRebuild();
  return(true);
  }
else
  {
  return(false);
  }
}

double // extrusion thickness
Text3D::getExtrusion(void) const
{
return(_extrusion);
}

//-------- Serialization operations ------------------------------------------

void
Text3D::extractDependencies(SerializationDependencies & dependenciesOut,
                            bool recursive) const
{
Surface3D::extractDependencies(dependenciesOut,recursive);
dependenciesOut.addURL(_fontURL);
}

void
Text3D::serialize(SerializationDependencies & dependenciesOut,
                  ArRef<AbstractOStream> output) const
{
Surface3D::serialize(dependenciesOut,output);
output->writeUInt(dependenciesOut.addURL(_fontURL));
output->writeString(_text);
output->writeDouble(_symbolSpace);
output->writeDouble(_whiteSpace);
output->writeDouble(_extrusion);
}

bool // success
Text3D::unserialize(const SerializationDependencies & dependencies,
                    ArRef<AbstractIStream> input)
{
unsigned int fontURL;
StlString text;
double symbolSpace,whiteSpace,extrusion;
if(!Surface3D::unserialize(dependencies,input)||
   !input->readUInt(fontURL)||
   !input->readString(text)||
   !input->readDouble(symbolSpace)||
   !input->readDouble(whiteSpace)||
   !input->readDouble(extrusion))
  {
  return(false);
  }
_text.swap(text);
_symbolSpace=symbolSpace;
_whiteSpace=whiteSpace;
_extrusion=extrusion;
setFont(dependencies.findURL(fontURL));
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
Text3D::_rebuildIfNeeded(void) // Protected
{
if(_needsGeometryRebuild)
  {
  _vertices.clear();
  _facesV.clear();
  _normals.clear();
  _facesN.clear();
  _textures.clear();
  _facesT.clear();
  if(_font)
    {
    LLW::buildGLText(_font,_symbolSpace,_whiteSpace,_text,_extrusion,
                     _vertices,_facesV,_normals,_facesN);
    double ascent,descent;
    LLW::getGLFontHeight(_font,ascent,descent);
    double yFactor=LLW::getGLTextWidth(_font,_symbolSpace,_whiteSpace,_text);
    double zFactor=ascent+descent;
    yFactor=(yFactor ? 1.0/yFactor : 0.0);
    zFactor=(zFactor ? 1.0/zFactor : 0.0);
    unsigned int nb=_vertices.size();
    for(unsigned int i=0;i<nb;i++)
      {
      const Util3D::Dbl3 & v=_vertices[i];
      _textures.push_back(Util3D::Dbl2(v.y*yFactor,(v.z+descent)*zFactor));
      }
    }
  }
Surface3D::_rebuildIfNeeded();
}

} // namespace AReVi

//----------------------------------------------------------------------------
