// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/abstractStream.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/abstractStream.h"

#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/arSystem.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class AbstractIStream (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(AbstractIStream,ArObject)

//-------- Construction / Destruction ----------------------------------------

AbstractIStream::AbstractIStream(ArCW & arCW) // Protected
: ArObject(arCW),
  _iBuffer(new AbstractIStream::IBuffer(this)),
  _stream(_iBuffer)
{
  // Nothing to be done
}

AbstractIStream::AbstractIStream(ArCW & arCW,
                                 ArConstRef<MemoryBlock> memory,
                                 unsigned int offset) // Protected
: ArObject(arCW),
  _iBuffer(new AbstractIStream::IBuffer(this,memory,offset)),
  _stream(_iBuffer)
{
  // Nothing to be done
}

AbstractIStream::~AbstractIStream(void) // Protected
{
delete _iBuffer;
}

//-------- Stream control ----------------------------------------------------

unsigned int // number of bytes currently available in the input buffer
AbstractIStream::bufferContent(void) const
{
return(_iBuffer->in_avail());
}

istream & // embedded C++ stream
AbstractIStream::stream(void)
{
return(_stream);
}

//-------- Text input --------------------------------------------------------

bool // success
AbstractIStream::readTxtChar(char & cOut)
{
return(_stream.get(cOut));
}

bool // success
AbstractIStream::readTxtUChar(unsigned char & ucOut)
{
return(_stream.get((char &)ucOut));
}

bool // success
AbstractIStream::readTxtShort(short & sOut)
{
return(_stream >> sOut);
}

bool // success
AbstractIStream::readTxtUShort(unsigned short & usOut)
{
return(_stream >> usOut);
}

bool // success
AbstractIStream::readTxtInt(int & iOut)
{
return(_stream >> iOut);
}

bool // success
AbstractIStream::readTxtUInt(unsigned int & uiOut)
{
return(_stream >> uiOut);
}

bool // success
AbstractIStream::readTxtLLong(long long & llOut)
{
return(_stream >> llOut);
}

bool // success
AbstractIStream::readTxtULLong(unsigned long long & ullOut)
{
return(_stream >> ullOut);
}

bool // success
AbstractIStream::readTxtFloat(float & fOut)
{
return(_stream >> fOut);
}

bool // success
AbstractIStream::readTxtDouble(double & dOut)
{
return(_stream >> dOut);
}

bool // success
AbstractIStream::readTxtWord(StlString & wOut)
{
return(_stream >> wOut);
}

bool // success
AbstractIStream::readTxtLine(StlString & lOut)
{
return(getline(_stream,lOut));
}

//-------- Binary input ------------------------------------------------------

unsigned int // nb read
AbstractIStream::readData(void * buffer,
                          unsigned int size)
{
_stream.read(reinterpret_cast<char *>(buffer),size);
return(_stream.gcount());
}

bool // success
AbstractIStream::copyDataToOutput(ArRef<AbstractOStream> output)
{
char buffer[0x400];
for(;;)
  {
  unsigned int nbIn=readData(buffer,0x400);
  if(nbIn>0)
    {
    output->writeData(buffer,nbIn);
    }
  else
    {
    break;
    }
  }
return(!(fail()||output->fail()));
}

//-------- Network format binary input ---------------------------------------

bool // success
AbstractIStream::readChar(char & cOut)
{
return(_stream.read(&cOut,sizeof(char)));
}

bool // success
AbstractIStream::readUChar(unsigned char & ucOut)
{
return(_stream.read(reinterpret_cast<char *>(&ucOut),sizeof(unsigned char)));
}

bool // success
AbstractIStream::readShort(short & sOut)
{
if(_stream.read(reinterpret_cast<char *>(&sOut),sizeof(short)))
  {
  ArSystem::inOut16(&sOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readUShort(unsigned short & usOut)
{
if(_stream.read(reinterpret_cast<char *>(&usOut),sizeof(unsigned short)))
  {
  ArSystem::inOut16(&usOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readInt(int & iOut)
{
if(_stream.read(reinterpret_cast<char *>(&iOut),sizeof(int)))
  {
  ArSystem::inOut32(&iOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readUInt(unsigned int & uiOut)
{
if(_stream.read(reinterpret_cast<char *>(&uiOut),sizeof(unsigned int)))
  {
  ArSystem::inOut32(&uiOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readLLong(long long & llOut)
{
if(_stream.read(reinterpret_cast<char *>(&llOut),sizeof(long long)))
  {
  ArSystem::inOut64(&llOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readULLong(unsigned long long & ullOut)
{
if(_stream.read(reinterpret_cast<char *>(&ullOut),sizeof(unsigned long long)))
  {
  ArSystem::inOut64(&ullOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readFloat(float & fOut)
{
if(_stream.read(reinterpret_cast<char *>(&fOut),sizeof(float)))
  {
  ArSystem::inOut32(&fOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readDouble(double & dOut)
{
if(_stream.read(reinterpret_cast<char *>(&dOut),sizeof(double)))
  {
  ArSystem::inOut64(&dOut); // N to H
  return(true);
  }
else
  {
  return(false);
  }
}

bool // success
AbstractIStream::readString(StlString & sOut)
{
unsigned int size;
if(_stream.read(reinterpret_cast<char *>(&size),sizeof(unsigned int)))
  {
  ArSystem::inOut32(&size); // N to H
  if(size)
    {
    StlString tmp(size,'x');
    if(_stream.read(const_cast<char *>(tmp.c_str()),size))
      {
      sOut.swap(tmp);
      return(true);
      }
    else
      {
      return(false);
      }
    }
  else
    {
    return(true);
    }
  }
else
  {
  return(false);
  }
}

//-------- Error management --------------------------------------------------

bool // problem occured
AbstractIStream::fail(void) const
{
if(ArObject::fail())
  {
  return(true);
  }
else
  {
  if(_stream.fail())
    {
    ar_const_cast(thisRef())->setErrorMessage("istream failure");
    return(true);
    }
  else
    {
    return(false);
    }
  }
}

//-------- Forbidden operations ----------------------------------------------

int // >0: nb read   <=0: error
AbstractIStream::_readBuffer(char * /* buffer */,
                             unsigned int /* size */) // Protected
{
return(-1);
}

AbstractIStream::IBuffer::IBuffer(AbstractIStream * stream)
: _buffer(ArSystem::allocate<char>(ISTREAM_BUFFERSIZE+0)),
  _memory(),
  _offset(0),
  _stream(stream)
{
setg(_buffer,_buffer,_buffer); // empty
}

AbstractIStream::IBuffer::IBuffer(AbstractIStream * stream,
                                  ArConstRef<MemoryBlock> memory,
                                  unsigned int offset)
: _buffer((char *)0),
  _memory(memory),
  _offset(offset),
  _stream(stream)
{
char * buffer=const_cast<char *>(_memory->getMemory());
unsigned int size=_memory->getSize();
if(_offset>size)
  {
  _offset=size;
  }
setg(buffer+_offset,buffer+_offset,buffer+size); // full
}

AbstractIStream::IBuffer::~IBuffer(void)
{
if(_buffer)
  {
  ArSystem::deallocate(_buffer);
  }
}

unsigned int
AbstractIStream::IBuffer::getMemoryUsage(void) const
{
return(gptr()-eback());
}

unsigned int
AbstractIStream::IBuffer::getOffset(void) const
{
return(_offset);
}

int
AbstractIStream::IBuffer::underflow(void)
{
if(_buffer)
  {
  int r=_stream->_readBuffer(_buffer,ISTREAM_BUFFERSIZE);
  if(r<=0)
    {
    return(EOF);
    }
  else
    {
    setg(_buffer,_buffer,_buffer+r);
    return(static_cast<unsigned char>(_buffer[0]));
    }
  }
else
  {
  return(EOF);
  }
}


//----------------------------------------------------------------------------
// Class AbstractOStream (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(AbstractOStream,ArObject)

//-------- Construction / Destruction ----------------------------------------

AbstractOStream::AbstractOStream(ArCW & arCW) // Protected
: ArObject(arCW),
  _oBuffer(new AbstractOStream::OBuffer(this)),
  _stream(_oBuffer)
{
  // Nothing to be done
}

AbstractOStream::AbstractOStream(ArCW & arCW,
                                 ArRef<MemoryBlock> memory,
                                 unsigned int offset) // Protected
: ArObject(arCW),
  _oBuffer(new AbstractOStream::OBuffer(this,memory,offset)),
  _stream(_oBuffer)
{
  // Nothing to be done
}

AbstractOStream::~AbstractOStream(void) // Protected
{
delete _oBuffer;
}

//-------- Stream control ----------------------------------------------------

bool // success
AbstractOStream::flush(void)
{
return(_stream.flush());
}

ostream & // embedded C++ stream
AbstractOStream::stream(void)
{
return(_stream);
}

//-------- Text output -------------------------------------------------------

bool // success
AbstractOStream::writeTxtChar(char c)
{
return(_stream.put(c));
}

bool // success
AbstractOStream::writeTxtUChar(unsigned char uc)
{
return(_stream.put(uc));
}

bool // success
AbstractOStream::writeTxtShort(short s)
{
return(_stream << s);
}

bool // success
AbstractOStream::writeTxtUShort(unsigned short us)
{
return(_stream << us);
}

bool // success
AbstractOStream::writeTxtInt(int i)
{
return(_stream << i);
}

bool // success
AbstractOStream::writeTxtUInt(unsigned int ui)
{
return(_stream << ui);
}

bool // success
AbstractOStream::writeTxtLLong(long long ll)
{
return(_stream << ll);
}

bool // success
AbstractOStream::writeTxtULLong(unsigned long long ull)
{
return(_stream << ull);
}

bool // success
AbstractOStream::writeTxtFloat(float f)
{
return(_stream << f);
}

bool // success
AbstractOStream::writeTxtDouble(double d)
{
return(_stream << d);
}

bool // success
AbstractOStream::writeTxtString(const StlString & s)
{
return(_stream << s);
}

//-------- Binary output -----------------------------------------------------

bool // success
AbstractOStream::writeData(const void * buffer,
                           unsigned int size)
{
return(_stream.write(reinterpret_cast<const char *>(buffer),size));
}

bool // success
AbstractOStream::copyDataFromInput(ArRef<AbstractIStream> input)
{
char buffer[0x400];
for(;;)
  {
  unsigned int nbIn=input->readData(buffer,0x400);
  if(nbIn>0)
    {
    writeData(buffer,nbIn);
    }
  else
    {
    break;
    }
  }
return(!(fail()||input->fail()));
}

//-------- Network format binary output --------------------------------------

bool // success
AbstractOStream::writeChar(char c)
{
return(_stream.write(&c,sizeof(char)));
}

bool // success
AbstractOStream::writeUChar(unsigned char uc)
{
return(_stream.write(reinterpret_cast<const char *>(&uc),
                     sizeof(unsigned char)));
}

bool // success
AbstractOStream::writeShort(short s)
{
ArSystem::inOut16(&s); // H to N
return(_stream.write(reinterpret_cast<const char *>(&s),
                     sizeof(short)));
}

bool // success
AbstractOStream::writeUShort(unsigned short us)
{
ArSystem::inOut16(&us); // H to N
return(_stream.write(reinterpret_cast<const char *>(&us),
                     sizeof(unsigned short)));
}

bool // success
AbstractOStream::writeInt(int i)
{
ArSystem::inOut32(&i); // H to N
return(_stream.write(reinterpret_cast<const char *>(&i),
                     sizeof(int)));
}

bool // success
AbstractOStream::writeUInt(unsigned int ui)
{
ArSystem::inOut32(&ui); // H to N
return(_stream.write(reinterpret_cast<const char *>(&ui),
                     sizeof(unsigned int)));
}

bool // success
AbstractOStream::writeLLong(long long ll)
{
ArSystem::inOut64(&ll); // H to N
return(_stream.write(reinterpret_cast<const char *>(&ll),
                     sizeof(long long)));
}

bool // success
AbstractOStream::writeULLong(unsigned long long ull)
{
ArSystem::inOut64(&ull); // H to N
return(_stream.write(reinterpret_cast<const char *>(&ull),
                     sizeof(unsigned long long)));
}

bool // success
AbstractOStream::writeFloat(float f)
{
ArSystem::inOut32(&f); // H to N
return(_stream.write(reinterpret_cast<const char *>(&f),
                     sizeof(float)));
}

bool // success
AbstractOStream::writeDouble(double d)
{
ArSystem::inOut64(&d); // H to N
return(_stream.write(reinterpret_cast<const char *>(&d),
                     sizeof(double)));
}

bool // success
AbstractOStream::writeString(const StlString & s)
{
unsigned int sz=s.size();
ArSystem::inOut32(&sz); // H to N
if(!_stream.write(reinterpret_cast<const char *>(&sz),
                  sizeof(unsigned int)))
  {
  return(false);
  }
else
  {
  if(s.size())
    {
    return(_stream.write(s.c_str(),s.size()));
    }
  else
    {
    return(true);
    }
  }
}

//-------- Error management --------------------------------------------------

bool // problem occured
AbstractOStream::fail(void) const
{
if(ArObject::fail())
  {
  return(true);
  }
else
  {
  if(_stream.fail())
    {
    ar_const_cast(thisRef())->setErrorMessage("ostream failure");
    return(true);
    }
  else
    {
    return(false);
    }
  }
}

//-------- Forbidden operations ----------------------------------------------

int // >=0: nb written   -1: error
AbstractOStream::_writeBuffer(const char * /* buffer */,
                              unsigned int /* size */) // Protected
{
return(-1);
}

AbstractOStream::OBuffer::OBuffer(AbstractOStream * stream)
: _buffer(ArSystem::allocate<char>(OSTREAM_BUFFERSIZE+1)),
  _memory(),
  _offset(0),
  _stream(stream)
{
setp(_buffer,_buffer+OSTREAM_BUFFERSIZE);
}

AbstractOStream::OBuffer::OBuffer(AbstractOStream * stream,
                                  ArRef<MemoryBlock> memory,
                                  unsigned int offset)
: _buffer((char *)0),
  _memory(memory),
  _offset(offset),
  _stream(stream)
{
if(!_memory->getSize())
  {
  _memory->setSize(OSTREAM_BUFFERSIZE);
  }
char * buffer=_memory->accessMemory();
unsigned int size=_memory->getSize();
if(_offset>size)
  {
  _offset=size;
  }
setp(buffer+_offset,buffer+size);
}

AbstractOStream::OBuffer::~OBuffer(void)
{
if(_buffer)
  {
  ArSystem::deallocate(_buffer);
  }
}

unsigned int
AbstractOStream::OBuffer::getMemoryUsage(void) const
{
return(_buffer ? pptr()-pbase() : pptr()-(_memory->getMemory()+_offset));
}

unsigned int
AbstractOStream::OBuffer::getOffset(void) const
{
return(_offset);
}

int
AbstractOStream::OBuffer::sync(void)
{
if(_buffer)
  {
  unsigned int size=pptr()-pbase();
  int r=_stream->_writeBuffer(_buffer,size);
  if(r==-1)
    {
    return(EOF);
    }
  else
    {
    setp(_buffer,_buffer+OSTREAM_BUFFERSIZE);
    return(0);
    }
  }
else return(0);
}

int
AbstractOStream::OBuffer::overflow(int c)
{
if(_buffer)
  {
  unsigned int size=pptr()-pbase();
  if(c!=EOF) _buffer[size++]=c; // buffer size is OSTREAM_BUFFERSIZE+1
  int r=_stream->_writeBuffer(_buffer,size);
  if(r==-1)
    {
    return(EOF);
    }
  else
    {
    setp(_buffer,_buffer+OSTREAM_BUFFERSIZE);
    return(0);
    }
  }
else
  {
  if(c!=EOF)
    {
    unsigned int oldSize=_memory->getSize();
    if(_memory->setSize(2*oldSize-_offset))
      {
      char * buffer=_memory->accessMemory();
      buffer[oldSize++]=c;
      setp(buffer+oldSize,buffer+_memory->getSize());
      return(0);
      }
    else
      {
      return(EOF);
      }
    }
  else
    {
    return(0);
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
