// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/process.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/process.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>

namespace AReVi {

//----------------------------------------------------------------------------
// Class Process
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Process,ArObject)

//-------- Construction / Destruction ----------------------------------------

Process::Process(ArCW & arCW,
                 const StlVector<StlString> & commandLine,
                 bool redirectInput,
                 bool redirectOutput,
                 bool redirectError) // Protected
: ArObject(arCW),
  _id(-1),
  _ended(false),
  _result(-1)
  // _inputPipe is not initialized !
  // _outputPipe is not initialized !
  // _errorPipe is not initialized !
{
_inputPipe[0]=_inputPipe[1]=-1;
_outputPipe[0]=_outputPipe[1]=-1;
_errorPipe[0]=_errorPipe[1]=-1;
_start(commandLine,redirectInput,redirectOutput,redirectError);
}

Process::Process(ArCW & arCW,
                 const StlString & shellCommand,
                 bool redirectInput,
                 bool redirectOutput,
                 bool redirectError) // Protected
: ArObject(arCW),
  _id(-1),
  _ended(false),
  _result(-1)
  // _inputPipe is not initialized !
  // _outputPipe is not initialized !
  // _errorPipe is not initialized !
{
_inputPipe[0]=_inputPipe[1]=-1;
_outputPipe[0]=_outputPipe[1]=-1;
_errorPipe[0]=_errorPipe[1]=-1;
StlVector<StlString> commandLine;
commandLine.push_back("/bin/sh");
commandLine.push_back("-c");
commandLine.push_back(shellCommand);
_start(commandLine,redirectInput,redirectOutput,redirectError);
}

Process::~Process(void) // Protected
{
if(_id&&!isEnded())
  {
  ArSystem::sendSignal(_id,ArSystem::SIGNUM_KILL);
  }
closeInputFD();
closeOutputFD();
closeErrorFD();
}

//-------- Inspection --------------------------------------------------------

ArSystem::ProcessID // process identifier
Process::getID(void) const
{
return(_id);
}

//-------- I/O redirection ---------------------------------------------------

int // >=0: input file descriptor  -1: not redirected
Process::getInputFD(void) const
{
return(_inputPipe[1]);
}

void
Process::closeInputFD(void)
{
if(_inputPipe[0]!=-1)
  {
  ArSystem::closeFD(_inputPipe[0]);
  _inputPipe[0]=-1;
  }
if(_inputPipe[1]!=-1)
  {
  ArSystem::closeFD(_inputPipe[1]);
  _inputPipe[1]=-1;
  }
}

int // >=0: output file descriptor  -1: not redirected
Process::getOutputFD(void) const
{
return(_outputPipe[0]);
}

void
Process::closeOutputFD(void)
{
if(_outputPipe[0]!=-1)
  {
  ArSystem::closeFD(_outputPipe[0]);
  _outputPipe[0]=-1;
  }
if(_outputPipe[1]!=-1)
  {
  ArSystem::closeFD(_outputPipe[1]);
  _outputPipe[1]=-1;
  }
}

int // >=0: error file descriptor  -1: not redirected
Process::getErrorFD(void) const
{
return(_errorPipe[0]);
}

void
Process::closeErrorFD(void)
{
if(_errorPipe[0]!=-1)
  {
  ArSystem::closeFD(_errorPipe[0]);
  _errorPipe[0]=-1;
  }
if(_errorPipe[1]!=-1)
  {
  ArSystem::closeFD(_errorPipe[1]);
  _errorPipe[1]=-1;
  }
}

//-------- Ending ------------------------------------------------------------

bool // process end reached
Process::isEnded(void) const
{
if(!_ended)
  {
  Process * self=const_cast<Process *>(this);
  int status;
  switch(waitpid(self->_id,&status,WNOHANG))
    {
    case -1:
      {
      if(errno==ECHILD) // should not happen
        {
        self->_ended=true;
        }
      break;
      }
    case 0:
      {
      break;
      }
    default:
      {
      self->_ended=true;
      if(WIFEXITED(status))
        {
        self->_result=WEXITSTATUS(status);
        }
      }
    }
  }
return(_ended);
}

int // status after process end
Process::waitStatus(void)
{
if(!_ended&&(_id!=-1))
  {
  int r,status;
  do
    {
    r=waitpid(_id,&status,0);
    } while((r==-1)&&(errno==EINTR));
  if(r!=-1)
    {
    _ended=true;
    if(WIFEXITED(status))
      {
      _result=WEXITSTATUS(status);
      }
    }
  }
return(_result);
}

//-------- External commands embedding ---------------------------------------

int // status
Process::execute(const StlString & shellCommand) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,false,false,false);
if(process->fail())
  {
  return(-1);
  }
return(process->waitStatus());
}

int // status
Process::executeI(const StlString & shellCommand,
                  const StlString & input) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,true,false,false);
if(process->fail())
  {
  return(-1);
  }
if(ArSystem::writeFD(process->getInputFD(),input.c_str(),input.size())==-1)
  {
  return(-1);
  }
process->closeInputFD();
return(process->waitStatus());
}

int // status
Process::executeO(const StlString & shellCommand,
                  StlString & outputOut) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,false,true,false);
if(process->fail())
  {
  return(-1);
  }
outputOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getOutputFD(),&c,sizeof(char))>0)
    {
    outputOut+=c;
    }
  else
    {
    break;
    }
  }
if(outputOut.size()&&(outputOut[outputOut.size()-1]=='\n'))
  {
  outputOut=outputOut.substr(0,outputOut.size()-1);
  }
return(process->waitStatus());
}

int // status
Process::executeE(const StlString & shellCommand,
                  StlString & errorOut) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,false,false,true);
if(process->fail())
  {
  return(-1);
  }
errorOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getErrorFD(),&c,sizeof(char))>0)
    {
    errorOut+=c;
    }
  else
    {
    break;
    }
  }
if(errorOut.size()&&(errorOut[errorOut.size()-1]=='\n'))
  {
  errorOut=errorOut.substr(0,errorOut.size()-1);
  }
return(process->waitStatus());
}

int // status
Process::executeIO(const StlString & shellCommand,
                   const StlString & input,
                   StlString & outputOut) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,true,true,false);
if(process->fail())
  {
  return(-1);
  }
if(ArSystem::writeFD(process->getInputFD(),input.c_str(),input.size())==-1)
  {
  return(-1);
  }
process->closeInputFD();
outputOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getOutputFD(),&c,sizeof(char))>0)
    {
    outputOut+=c;
    }
  else
    {
    break;
    }
  }
if(outputOut.size()&&(outputOut[outputOut.size()-1]=='\n'))
  {
  outputOut=outputOut.substr(0,outputOut.size()-1);
  }
return(process->waitStatus());
}

int // status
Process::executeIE(const StlString & shellCommand,
                   const StlString & input,
                   StlString & errorOut) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,true,false,true);
if(process->fail())
  {
  return(-1);
  }
if(ArSystem::writeFD(process->getInputFD(),input.c_str(),input.size())==-1)
  {
  return(-1);
  }
process->closeInputFD();
errorOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getErrorFD(),&c,sizeof(char))>0)
    {
    errorOut+=c;
    }
  else
    {
    break;
    }
  }
if(errorOut.size()&&(errorOut[errorOut.size()-1]=='\n'))
  {
  errorOut=errorOut.substr(0,errorOut.size()-1);
  }
return(process->waitStatus());
}

int // status
Process::executeOE(const StlString & shellCommand,
                   StlString & outputOut,
                   StlString & errorOut) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,false,true,true);
if(process->fail())
  {
  return(-1);
  }
outputOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getOutputFD(),&c,sizeof(char))>0)
    {
    outputOut+=c;
    }
  else
    {
    break;
    }
  }
if(outputOut.size()&&(outputOut[outputOut.size()-1]=='\n'))
  {
  outputOut=outputOut.substr(0,outputOut.size()-1);
  }
errorOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getErrorFD(),&c,sizeof(char))>0)
    {
    errorOut+=c;
    }
  else
    {
    break;
    }
  }
if(errorOut.size()&&(errorOut[errorOut.size()-1]=='\n'))
  {
  errorOut=errorOut.substr(0,errorOut.size()-1);
  }
return(process->waitStatus());
}

int // status
Process::executeIOE(const StlString & shellCommand,
                    const StlString & input,
                    StlString & outputOut,
                    StlString & errorOut) // Static
{
ArRef<Process> process=Process::NEW(shellCommand,true,true,true);
if(process->fail())
  {
  return(-1);
  }
if(ArSystem::writeFD(process->getInputFD(),input.c_str(),input.size())==-1)
  {
  return(-1);
  }
process->closeInputFD();
outputOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getOutputFD(),&c,sizeof(char))>0)
    {
    outputOut+=c;
    }
  else
    {
    break;
    }
  }
if(outputOut.size()&&(outputOut[outputOut.size()-1]=='\n'))
  {
  outputOut=outputOut.substr(0,outputOut.size()-1);
  }
errorOut.erase();
for(;;)
  {
  char c;
  if(ArSystem::partialReadFD(process->getErrorFD(),&c,sizeof(char))>0)
    {
    errorOut+=c;
    }
  else
    {
    break;
    }
  }
if(errorOut.size()&&(errorOut[errorOut.size()-1]=='\n'))
  {
  errorOut=errorOut.substr(0,errorOut.size()-1);
  }
return(process->waitStatus());
}

//-------- Forbidden operations ----------------------------------------------

void
Process::_start(const StlVector<StlString> & commandLine,
                bool redirectInput,
                bool redirectOutput,
                bool redirectError) // Protected
{
if(redirectInput&&(pipe(_inputPipe)==-1))
  {
  setErrorMessage("Cannot create input pipe --- "+
                  ArSystem::printError(errno));
  return;
  }
if(redirectOutput&&(pipe(_outputPipe)==-1))
  {
  setErrorMessage("Cannot create output pipe --- "+
                  ArSystem::printError(errno));
  closeInputFD();
  return;
  }
if(redirectError&&(pipe(_errorPipe)==-1))
  {
  setErrorMessage("Cannot create error pipe --- "+
                  ArSystem::printError(errno));
  closeInputFD();
  closeOutputFD();
  return;
  }
_id=fork();
switch(_id)
  {
  case -1:
    {
    setErrorMessage("Cannot create child process --- "+
                    ArSystem::printError(errno));
    closeInputFD();
    closeOutputFD();
    closeErrorFD();
    return;
    }
  case 0: // child
    {
    int r;
    if(redirectInput)
      {
      do
        {
        r=dup2(_inputPipe[0],STDIN_FILENO);
        } while((r==-1)&&(errno==EINTR));
      closeInputFD();
      }
    if(redirectOutput)
      {
      do
        {
        r=dup2(_outputPipe[1],STDOUT_FILENO);
        } while((r==-1)&&(errno==EINTR));
      closeOutputFD();
      }
    if(redirectError)
      {
      do
        {
        r=dup2(_errorPipe[1],STDERR_FILENO);
        } while((r==-1)&&(errno==EINTR));
      closeErrorFD();
      }
    StlVector<char *> argv;
    for(unsigned int i=0;i<commandLine.size();++i)
      {
      argv.push_back(const_cast<char *>(commandLine[i].c_str()));
      }
    argv.push_back((char *)0);
    execvp(argv[0],&argv[0]);
    exit(-1); // should never be reached
    }
  default: // parent
    {
    if(redirectInput)
      {
      ArSystem::closeFD(_inputPipe[0]);
      _inputPipe[0]=-1;
      }
    if(redirectOutput)
      {
      ArSystem::closeFD(_outputPipe[1]);
      _outputPipe[1]=-1;
      }
    if(redirectError)
      {
      ArSystem::closeFD(_errorPipe[1]);
      _errorPipe[1]=-1;
      }
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
