// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/regex.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/regex.h"

#include "AReVi/arSystem.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class RegularExpression
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(RegularExpression,ArObject)

//-------- Construction / Destruction ----------------------------------------

RegularExpression::RegularExpression(ArCW & arCW,
                                     const StlString & pattern) // Protceted
: ArObject(arCW),
  _regex(),
  _match((regmatch_t *)0),
  _pattern(pattern)
{
int r=regcomp(&_regex,_pattern.c_str(),REG_EXTENDED|REG_NEWLINE);
if(r)
  {
  int n=regerror(r,&_regex,(char *)0,0);
  char * buffer=ArSystem::allocate<char>(n);
  regerror(r,&_regex,buffer,n);
  setErrorMessage(buffer);
  ArSystem::deallocate(buffer);
  }
else
  {
  _match=ArSystem::allocate<regmatch_t>(_regex.re_nsub+1);
  }
}

RegularExpression::~RegularExpression(void) // Protected
{
regfree(&_regex);
if(_match)
  {
  ArSystem::deallocate(_match);
  }
}

//-------- Pattern matching --------------------------------------------------

const StlString & // pattern to match
RegularExpression::getPattern(void) const
{
return(_pattern);
}

bool // success
RegularExpression::match(const StlString & input,
                         unsigned int & startOut,
                         unsigned int & endOut) const
{
if(_match)
  {
  if(regexec(&_regex,input.c_str(),1,_match,0))
    {
    return(false);
    }
  else
    {
    startOut=_match[0].rm_so;
    endOut=_match[0].rm_eo;
    return(true);
    }
  }
else
  {
  return(false);
  }
}

bool // success
RegularExpression::detailledMatch(const StlString & input,
                                  StlVector<int> & startOut,
                                  StlVector<int> & endOut) const
{
startOut.clear();
endOut.clear();
if(_match)
  {
  unsigned int nb=_regex.re_nsub+1;
  if(regexec(&_regex,input.c_str(),nb,_match,0))
    {
    return(false);
    }
  else
    {
    for(unsigned int i=0;i<nb;i++)
      {
      startOut.push_back(_match[i].rm_so);
      endOut.push_back(_match[i].rm_eo);
      }
    return(true);
    }
  }
else
  {
  return(false);
  }
}

//-------- String management -------------------------------------------------

unsigned int // number of replacements done
RegularExpression::replace(StlString & strInOut,
                           const StlString & replacement,
                           unsigned int maxReplacement) const
{
StlVector<StlString> subStrings;
split(strInOut,subStrings,(maxReplacement ? maxReplacement+1 : 0));
strInOut.erase();
for(unsigned int i=0;i<subStrings.size();i++)
  {
  if(i)
    {
    strInOut+=replacement;
    }
  strInOut+=subStrings[i];
  }
return(subStrings.size()-1);
}

void
RegularExpression::split(const StlString & input,
                         StlVector<StlString> & subStringsOut,
                         unsigned int maxSplit) const
{
subStringsOut.clear();
unsigned int start=0;
if(_match)
  {
  bool done=false;
  do
    {
    if(regexec(&_regex,input.c_str()+start,1,_match,0))
      {
      done=true;
      }
    else
      {
      if(_match[0].rm_so==_match[0].rm_eo)
        {
        done=true; // avoid zero length match
        }
      else
        {
        subStringsOut.push_back(input.substr(start,_match[0].rm_so));
        start+=_match[0].rm_eo;
        if(maxSplit&&(subStringsOut.size()==maxSplit-1))
          {
          done=true;
          }
        }
      }
    } while(!done);
  }
subStringsOut.push_back(input.substr(start,input.size()-start));
}

} // namespace AReVi

//----------------------------------------------------------------------------
