// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/spaceMouse.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/spaceMouse.h"

#include "AReVi/arSystem.h"

#include <errno.h>

inline static
double SpaceMouse_roundAxis(double axis)
{
return(rint(axis*10000.0)*0.0001);
}

namespace AReVi {

//----------------------------------------------------------------------------
// Class SpaceMouse
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(SpaceMouse,InputDevice)

//-------- Construction / Destruction ----------------------------------------

SpaceMouse::SpaceMouse(ArCW & arCW,
                       const StlString & device) // Protected
: InputDevice(arCW,device),
  _fd(-1),
  _savedTermios(),
  _line(),
  _lineSize(0),
  _lineLimit(0)
{
reset(_device);
}

SpaceMouse::~SpaceMouse(void) // Protected
{
if(_fd>=0)
  {
  tcsetattr(_fd,TCSANOW,&_savedTermios);
  ArSystem::closeFD(_fd);
  }
}

//-------- Device management -------------------------------------------------

bool // success
SpaceMouse::reset(const StlString & device)
{
if(_fd>=0)
  {
  tcsetattr(_fd,TCSANOW,&_savedTermios);
  ArSystem::closeFD(_fd);
  _fd=-1;
  }
_device=device;
_line[0]='\0';
_lineSize=0;
_lineLimit=0;
_buttons.clear();
_axes.clear();
unsigned int i;
for(i=9;i--;)
  {
  _buttons.push_back(false);
  }
for(i=6;i--;)
  {
  _axes.push_back(0.0);
  }
setErrorMessage("");
_fd=ArSystem::openInOutFile(_device);
if(_fd<0)
  {
  setErrorMessage("Cannot open device '"+_device+"' --- "+
                  ArSystem::printError(errno));
  return(false);
  }
if(tcgetattr(_fd,&_savedTermios)<0)
  {
  setErrorMessage("Cannot read properties of device '"+_device+"' --- "+
                  ArSystem::printError(errno));
  ArSystem::closeFD(_fd);
  _fd=-1;
  return(false);
  }
struct termios newTermios;
memset(&newTermios,0,sizeof(struct termios));
newTermios.c_cflag=CS8 | CREAD | CSTOPB | CLOCAL;
newTermios.c_cc[VMIN]=1;
cfsetispeed(&newTermios,B9600);
cfsetospeed(&newTermios,B9600);
if(tcflush(_fd,TCIFLUSH)<0)
  {
  setErrorMessage("Cannot flush device '"+_device+"' --- "+
                  ArSystem::printError(errno));
  ArSystem::closeFD(_fd);
  _fd=-1;
  return(false);
  }
if(tcsetattr(_fd,TCSANOW,&newTermios)<0)
  {
  setErrorMessage("Cannot change properties of device '"+_device+"' --- "+
                  ArSystem::printError(errno));
  ArSystem::closeFD(_fd);
  _fd=-1;
  return(false);
  }
if(!ArSystem::setBlockingFD(_fd,false))
  {
  setErrorMessage("Cannot use non-blocking mode for device '"+_device+
                  "' --- "+ArSystem::printError(errno));
  tcsetattr(_fd,TCSANOW,&_savedTermios);
  ArSystem::closeFD(_fd);
  _fd=-1;
  return(false);
  }
return(// _send("vQ\r",2000)&&  // version
       _send("p00\r",2000)&& // data rate
       _send("m3\r",2000)&&  // mode
       _send("z\r",2000)&&   // reset location
       _send("q00\r",2000)&& // gain
       _send("nM\r",2000));  // radius
}

inline static
void
decodeChar(char & c)
{
switch(c)
  {
  case '0': c='\x00'; break;
  case 'A': c='\x01'; break;
  case 'B': c='\x02'; break;
  case '3': c='\x03'; break;
  case 'D': c='\x04'; break;
  case '5': c='\x05'; break;
  case '6': c='\x06'; break;
  case 'G': c='\x07'; break;
  case 'H': c='\x08'; break;
  case '9': c='\x09'; break;
  case ':': c='\x0A'; break;
  case 'K': c='\x0B'; break;
  case '<': c='\x0C'; break;
  case 'M': c='\x0D'; break;
  case 'N': c='\x0E'; break;
  case '?': c='\x0F'; break;
  }
}

void
SpaceMouse::update(void)
{
while(!fail()&&_readLine())
  {
  switch(_line[0])
    {
    case 'k':
      {
      if(_lineSize==5) // kBBB\r
        {
        decodeChar(_line[1]);
        decodeChar(_line[2]);
        decodeChar(_line[3]);
        int value=((int)_line[1])|
                  (((int)_line[2])<<4)|
                  (((int)_line[3])<<8);
        for(unsigned int i=_buttons.size();i--;)
          {
          int index=(i==8 ? 0 : i+1);
          if(value&(1<<i))
            {
            _buttons[index]=true;
            }
          else
            {
            _buttons[index]=false;
            }
          }
        }
      break;
      }
    case 'd':
      {
      if(_lineSize==26) // dXXXXYYYYZZZZXXXXYYYYZZZZ\r
        {
        for(unsigned int i=1;i<25;i++)
          {
          decodeChar(_line[i]);
          }
        int tx=(((int)_line[1])<<12)+
               (((int)_line[2])<<8)+
               (((int)_line[3])<<4)+
                ((int)_line[4]);
        int ty=(((int)_line[5])<<12)+
               (((int)_line[6])<<8)+
               (((int)_line[7])<<4)+
                ((int)_line[8]);
        int tz=(((int)_line[9])<<12)+
               (((int)_line[10])<<8)+
               (((int)_line[11])<<4)+
                ((int)_line[12]);
        int rx=(((int)_line[13])<<12)+
               (((int)_line[14])<<8)+
               (((int)_line[15])<<4)+
                ((int)_line[16]);
        int ry=(((int)_line[17])<<12)+
               (((int)_line[18])<<8)+
               (((int)_line[19])<<4)+
                ((int)_line[20]);
        int rz=(((int)_line[21])<<12)+
               (((int)_line[22])<<8)+
               (((int)_line[23])<<4)+
                ((int)_line[24]);
        _axes[0]=SpaceMouse_roundAxis((double)(32768-tz)/32768.0);
        _axes[1]=SpaceMouse_roundAxis((double)(32768-tx)/32768.0);
        _axes[2]=SpaceMouse_roundAxis((double)(ty-32768)/32768.0);
        _axes[3]=SpaceMouse_roundAxis((double)(32768-rz)/32768.0);
        _axes[4]=SpaceMouse_roundAxis((double)(32768-rx)/32768.0);
        _axes[5]=SpaceMouse_roundAxis((double)(ry-32768)/32768.0);
        }
      break;
      }
    }
  }
}

//-------- Translation / Rotation  -------------------------------------------

void
SpaceMouse::getTranslation(double & txOut,
                           double & tyOut,
                           double & tzOut) const
{
txOut=_axes[0];
tyOut=_axes[1];
tzOut=_axes[2];
}

void
SpaceMouse::getRotation(double & rollOut,
                        double & pitchOut,
                        double & yawOut) const
{
rollOut=_axes[3];
pitchOut=_axes[4];
yawOut=_axes[5];
}

//-------- Forbidden operations ----------------------------------------------

bool // success
SpaceMouse::_send(const char * msg,
                  unsigned int timeOut) // Protected
{
unsigned int limit1=ArSystem::getTime()+timeOut;
do
  {
  if(ArSystem::writeFD(_fd,msg,strlen(msg))<0)
    {
    setErrorMessage("Error while writing to device '"+_device+"' --- "+
                    ArSystem::printError(errno));
    tcsetattr(_fd,TCSANOW,&_savedTermios);
    ArSystem::closeFD(_fd);
    _fd=-1;
    return(false);
    }
  unsigned int limit2=ArSystem::getTime()+timeOut/4;
  do
    {
    if(_readLine())
      {
      return(true);
      }
    if(fail())
      {
      return(false);
      }
    } while(ArSystem::getTime()<limit2);
  } while(ArSystem::getTime()<limit1);
setErrorMessage("Device '"+_device+"' does not respond");
tcsetattr(_fd,TCSANOW,&_savedTermios);
ArSystem::closeFD(_fd);
_fd=-1;
return(false);
}

bool // line available
SpaceMouse::_readLine(void) // Protected
{
if(_lineSize)
  {
  _lineLimit-=_lineSize;
  memmove(_line,_line+_lineSize,_lineLimit+1);
  _lineSize=0;
  }
char * cr=strchr(_line,'\r');
if(cr)
  {
  _lineSize=cr+1-_line;
  return(true);
  }
int r=ArSystem::partialReadFD(_fd,_line+_lineLimit,0x400-_lineLimit-1);
if(r>0)
  {
  _lineLimit+=r;
  _line[_lineLimit]='\0';
  cr=strchr(_line,'\r');
  if(cr)
    {
    _lineSize=cr+1-_line;
    return(true);
    }
  }
else
  {
  switch(r)
    {
    case 0:
      {
      setErrorMessage("EOF reached while reading from device '"+
                      _device+"'");
      tcsetattr(_fd,TCSANOW,&_savedTermios);
      ArSystem::closeFD(_fd);
      _fd=-1;
      return(false);
      }
    case -1:
      {
      setErrorMessage("Error while reading from device '"+
                      _device+"' --- "+
                      ArSystem::printError(errno));
      tcsetattr(_fd,TCSANOW,&_savedTermios);
      ArSystem::closeFD(_fd);
      _fd=-1;
      return(false);
      }
    }
  }
return(false);
}

} // namespace AReVi

//----------------------------------------------------------------------------
