// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/tcpUdp.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/tcpUdp.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"

#include <netinet/in.h>
#include <string.h>
#include <errno.h>

namespace AReVi {

#define SOCKET_RECEPTION_SIZE 65535

//----------------------------------------------------------------------------
// Class IPSocket (abstract)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(IPSocket,ArObject)

//-------- Construction / Destruction ----------------------------------------

IPSocket::IPSocket(ArCW & arCW) // Protected
: ArObject(arCW),
  _socket(-1),
  _blocking(true),
  _ipAddress(0),
  _portNumber(-1)
{
  // Nothing to be done
}

IPSocket::~IPSocket(void)
{
_cancel();
}

//-------- Inspection --------------------------------------------------------

unsigned int // local ip address
IPSocket::getIPAddress(void) const
{
return(_ipAddress);
}

int // local port number
IPSocket::getPortNumber(void) const
{
return(_portNumber);
}

//-------- Blocking / Non-blocking -------------------------------------------

void
IPSocket::setBlocking(bool blocking)
{
_blocking=blocking;
if(!fail()&&!ArSystem::setBlockingFD(_socket,blocking))
  {
  setErrorMessage("Cannot change blocking status --- "+
                  ArSystem::printError(errno));
  _cancel();
  }
}

bool // blocks until data arrives
IPSocket::isBlocking(void) const
{
return(_blocking);
}

int // file descriptor
IPSocket::getSocketFD(void) const
{
return(_socket);
}

//-------- Forbidden operations ----------------------------------------------

void
IPSocket::_cancel(void) // Protected
{
if(_socket!=-1)
  {
  ArSystem::closeFD(_socket);
  _socket=-1;
  }
}


//----------------------------------------------------------------------------
// Class TCPConnection
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TCPConnection,IPSocket)

//-------- Construction / Destruction ----------------------------------------

TCPConnection::TCPConnection(ArCW & arCW,
                             unsigned int distantIPAddress,
                             int distantPortNumber,
                             bool packetOriented) // Protected
: IPSocket(arCW),
  _distantIPAddress(distantIPAddress),
  _distantPortNumber(distantPortNumber),
  _packetOriented(packetOriented),
  _packetSize(0),
  _nbExtracted(0),
  _memory(MemoryBlock::NEW()),
  _pendingData()
{
_socket=ArSystem::connectTCP(_distantIPAddress,_distantPortNumber);
if(_socket==-1)
  {
  setErrorMessage("Cannot connect to `"+ipAddrToStr(_distantIPAddress)+":"+
                  intToStr(_distantPortNumber)+"' --- "+
                  ArSystem::printError(errno));
  return;
  }
if(!ArSystem::getSocketAddress(_socket,_ipAddress,_portNumber))
  {
  setErrorMessage("Cannot retrieve socket parameters --- "+
                  ArSystem::printError(errno));
  _cancel();
  return;
  }
setBlocking(false);
if(fail())
  {
  return;
  }
if(!_packetOriented)
  {
  _memory->setSize(SOCKET_RECEPTION_SIZE);
  }
}

TCPConnection::TCPConnection(ArCW & arCW,
                             int socket,
                             unsigned int distantIPAddress,
                             int distantPortNumber,
                             bool packetOriented) // Protected
: IPSocket(arCW),
  _distantIPAddress(distantIPAddress),
  _distantPortNumber(distantPortNumber),
  _packetOriented(packetOriented),
  _packetSize(0),
  _nbExtracted(0),
  _memory(MemoryBlock::NEW()),
  _pendingData()
{
_socket=socket;
if(!ArSystem::getSocketAddress(_socket,_ipAddress,_portNumber))
  {
  setErrorMessage("Cannot retrieve socket parameters --- "+
                  ArSystem::printError(errno));
  _cancel();
  return;
  }
setBlocking(false);
if(fail())
  {
  return;
  }
if(!_packetOriented)
  {
  _memory->setSize(SOCKET_RECEPTION_SIZE);
  }
}

TCPConnection::~TCPConnection(void) // Protected
{
  // Nothing to be done
}

//-------- Inspection --------------------------------------------------------

unsigned int // distant ip address
TCPConnection::getDistantIPAddress(void) const
{
return(_distantIPAddress);
}

int // distant port number
TCPConnection::getDistantPortNumber(void) const
{
return(_distantPortNumber);
}

bool // packets size transmitted
TCPConnection::isPacketOriented(void) const
{
return(_packetOriented);
}

//-------- Binary communication ----------------------------------------------

bool // bytes available
TCPConnection::readBytes(ArRef<MemoryBlock> memory)
{
if(_socket==-1) return(false);
if(_packetOriented)
  {
  switch(_extractPacket())
    {
    case 0: // nothing
      {
      return(false);
      }
    case 1: // ok
      {
      bool success=memory->swapBlock(_memory);
      _nbExtracted=0;
      _memory->setSize(0);
      return(success);
      }
    case -1: // error
      {
      return(false);
      }
    }
  return(false); // should never be reached
  }
else
  {
  for(;;)
    {
    int r=ArSystem::partialReadFD(_socket,
                                   _memory->accessMemory(),
                                   _memory->getSize());
    switch(r)
      {
      case 0: // EOF
        {
        setErrorMessage("EOF reached");
        _cancel();
        return(false);
        break;
        }
      case -1: // error
        {
        setErrorMessage("PB while reading --- "+ArSystem::printError(errno));
        _cancel();
        return(false);
        break;
        }
      case -2: // non-blocking
        {
        if(!_blocking)
          {
          return(false);
          }
        break;
        }
      default: // something
        {
        _memory->setSize(r); // shrink --> no realloc --> no copy
        bool success=memory->swapBlock(_memory);
        _memory->setSize(0); // free to avoid copy in realloc
        _memory->setSize(SOCKET_RECEPTION_SIZE);
        return(success);
        }
      }
    }
  }
}

void
TCPConnection::writeBytes(ArConstRef<MemoryBlock> memory,
                          unsigned int offset,
                          unsigned int length)
{
while(pendingPost())
  {
  pushPost();
  }
if((_socket!=-1)&&_packetOriented)
  {
  unsigned int nb=length;
  ArSystem::inOut32(&nb); // H to N
  if(ArSystem::writeFD(_socket,&nb,sizeof(unsigned int))==-1)
    {
    setErrorMessage("PB while writing --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
if(_socket!=-1)
  {
  if(ArSystem::writeFD(_socket,memory->getMemory()+offset,length)==-1)
    {
    setErrorMessage("PB while writing --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
}


void
TCPConnection::postBytes(ArConstRef<MemoryBlock> memory,
                         unsigned int offset,
                         unsigned int length,
                         bool copyMemoryIfDelayed)
{
pushPost();
if(_socket!=-1)
  {
  PendingData data(memory,offset,length,
                   (_packetOriented ? sizeof(unsigned int) : 0));
  if(_pendingData.empty())
    {
    switch(_partialWrite(data))
      {
      case -1:
        {
        return;
        }
      case 0: // non-blocking
        {
        break;
        }
      case 1: // fully written
        {
        return;
        }
      }
    }
  if(copyMemoryIfDelayed)
    {
    ArRef<MemoryBlock> copy=MemoryBlock::NEW();
    copy->setSize(data.length);
    memcpy(copy->accessMemory(),
           data.memory->getMemory()+data.offset,data.length);
    data.memory=copy;
    data.offset=0;
    }
  _pendingData.push_back(data);
  }
}

//-------- Textual communication ---------------------------------------------

bool // string available
TCPConnection::readString(StlString & resultOut)
{
if(_socket==-1) return(false);
if(_packetOriented)
  {
  switch(_extractPacket())
    {
    case 0: // nothing
      {
      return(false);
      }
    case 1: // ok
      {
      unsigned int size=_memory->getSize();
      if(size)
        {
        char * bytes=_memory->accessMemory();
        bytes[size-1]='\0'; // ensure string is ended
        resultOut=bytes;
        }
      else
        {
        resultOut.erase();
        }
      _nbExtracted=0;
      _memory->setSize(0);
      return(true);
      }
    case -1: // error
      {
      return(false);
      }
    }
  return(false); // should never be reached
  }
else
  {
  for(;;)
    {
    int r=ArSystem::partialReadFD(_socket,
                                  _memory->accessMemory(),
                                  _memory->getSize()-1);
    switch(r)
      {
      case 0: // EOF
        {
        setErrorMessage("EOF reached");
        _cancel();
        return(false);
        break;
        }
      case -1: // error
        {
        setErrorMessage("PB while reading --- "+ArSystem::printError(errno));
        _cancel();
        return(false);
        break;
        }
      case -2: // non-blocking
        {
        if(!_blocking)
          {
          return(false);
          }
        break;
        }
      default: // something
        {
        char * bytes=_memory->accessMemory();
        bytes[r]='\0'; // ensure string is ended
        resultOut=bytes;
        return(true);
        }
      }
    }
  }
}

void
TCPConnection::writeString(const StlString & str)
{
while(pendingPost())
  {
  pushPost();
  }
if((_socket!=-1)&&_packetOriented)
  {
  unsigned int length=str.size()+1;
  ArSystem::inOut32(&length); // H to N
  if(ArSystem::writeFD(_socket,&length,sizeof(unsigned int))==-1)
    {
    setErrorMessage("PB while writing --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
if(_socket!=-1)
  {
  unsigned int nb=str.size();
  if(_packetOriented) ++nb;
  if(ArSystem::writeFD(_socket,str.c_str(),nb)==-1)
    {
    setErrorMessage("PB while writing --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
}


void
TCPConnection::postString(const StlString & str)
{
ArRef<RigidMemoryBlock> block=RigidMemoryBlock::NEW((char *)str.c_str(),
                                                   str.size()+1,false);
postBytes(block,0,block->getSize(),true);
}

//-------- Writing queue -----------------------------------------------------

void
TCPConnection::pushPost(void)
{
while(pendingPost()&&(_partialWrite(_pendingData.front())==1))
  {
  _pendingData.pop_front();
  }
}

bool // posted data should be pushed
TCPConnection::pendingPost(void) const
{
return((_socket!=-1)&&!_pendingData.empty());
}

//-------- Forbidden operations ----------------------------------------------

int // -1: error  0: nothing  1: ok
TCPConnection::_extractPacket(void) // Protected
{
bool again=(_nbExtracted<sizeof(unsigned int));
while(again)
  {
  char * ptr=reinterpret_cast<char *>(&_packetSize);
  ptr+=_nbExtracted;
  unsigned int remaining=sizeof(unsigned int)-_nbExtracted;
  int r=ArSystem::partialReadFD(_socket,ptr,remaining);
  switch(r)
    {
    case 0: // EOF
      {
      setErrorMessage("EOF reached");
      _cancel();
      return(-1);
      }
    case -1:
      {
      setErrorMessage("PB while reading --- "+ArSystem::printError(errno));
      _cancel();
      return(-1);
      }
    case -2: // non-blocking
      {
      if(!_blocking)
        {
        return(0);
        }
      break;
      }
    default:
      {
      _nbExtracted+=r;
      if(_nbExtracted==sizeof(unsigned int))
        {
        ArSystem::inOut32(&_packetSize); // N to H
        _memory->setSize(_packetSize);
        again=false;
        }
      else
        {
        if(!_blocking)
          {
          return(0);
          }
        }
      }
    }
  }
again=(_nbExtracted<sizeof(unsigned int)+_packetSize);
while(again)
  {
  char * ptr=_memory->accessMemory();
  ptr+=_nbExtracted-sizeof(unsigned int);
  unsigned int remaining=sizeof(unsigned int)+_packetSize-_nbExtracted;
  int r=ArSystem::partialReadFD(_socket,ptr,remaining);
  switch(r)
    {
    case 0: // EOF
      {
      setErrorMessage("EOF reached");
      _cancel();
      return(-1);
      }
    case -1:
      {
      setErrorMessage("PB while reading --- "+ArSystem::printError(errno));
      _cancel();
      return(-1);
      }
    case -2: // non-blocking
      {
      if(!_blocking)
        {
        return(0);
        }
      break;
      }
    default:
      {
      _nbExtracted+=r;
      if(_nbExtracted==sizeof(unsigned int)+_packetSize) return(1);
      else
        {
        if(!_blocking)
          {
          return(0);
          }
        }
      }
    }
  }
return(1); // reached only when _nbBytes=0
}

int // -1: error  0: non-blocking  1: fully written
TCPConnection::_partialWrite(PendingData & data)
{
if(data.lengthRemaining) // write packet length
  {
  unsigned int length=data.length;
  ArSystem::inOut32(&length); // H to N
  char *ptr=(char *)&length;
  int r=ArSystem::partialWriteFD(_socket,
                                 ptr+(sizeof(unsigned int)-
                                      data.lengthRemaining),
                                 data.lengthRemaining);
  switch(r)
    {
    case -1:
      {
      setErrorMessage("PB while writing --- "+ArSystem::printError(errno));
      _cancel();
      return(-1);
      }
    case -2: // non-blocking
      {
      return(0);
      }
    default:
      {
      data.lengthRemaining-=r;
      }
    }
  }
if((!data.lengthRemaining)&&data.length)
  {
  int r=ArSystem::partialWriteFD(_socket,
                                 data.memory->getMemory()+data.offset,
                                 data.length);
  switch(r)
    {
    case -1:
      {
      setErrorMessage("PB while writing --- "+ArSystem::printError(errno));
      _cancel();
      return(-1);
      }
    case -2: // non-blocking
      {
      return(0);
      }
    default:
      {
      data.offset+=r;
      data.length-=r;
      }
    }
  }
return((data.lengthRemaining||data.length) ? 0 : 1);
}


//----------------------------------------------------------------------------
// Class TCPListener
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(TCPListener,IPSocket)

//-------- Construction / Destruction ----------------------------------------

TCPListener::TCPListener(ArCW & arCW,
                         int portNumber,
                         bool packetOriented) // Protected
: IPSocket(arCW),
  _packetOriented(packetOriented)
{
_socket=ArSystem::listenTCP(portNumber);
if(_socket==-1)
  {
  setErrorMessage("Cannot listen TCP port `"+intToStr(portNumber)+
                  "' --- "+ArSystem::printError(errno));
  return;
  }
if(!ArSystem::getSocketAddress(_socket,_ipAddress,_portNumber))
  {
  setErrorMessage("Cannot retrieve socket parameters --- "+
                  ArSystem::printError(errno));
  _cancel();
  return;
  }
setBlocking(false);
if(fail()) return;
}

TCPListener::~TCPListener(void) // Protected
{
  // Nothing to be done
}

//-------- Inspection --------------------------------------------------------

bool // creates packet oriented connections
TCPListener::isPacketOriented(void) const
{
return(_packetOriented);
}

//-------- Connection control ------------------------------------------------

ArRef<TCPConnection> // distant connection
TCPListener::nextConnection(void)
{
if(_socket==-1)
  {
  return(TCPConnection::nullRef());
  }
for(;;)
  {
  unsigned int ipAddr;
  int portNumber;
  int fd=ArSystem::acceptTCP(_socket,ipAddr,portNumber);
  switch(fd)
    {
    case -1: // error
      {
      setErrorMessage("Cannot accept connection --- "+
                      ArSystem::printError(errno));
      _cancel();
      return(TCPConnection::nullRef());
      }
    case -2: // non-blocking
      {
      if(!_blocking)
        {
        return(TCPConnection::nullRef());
        }
      break;
      }
    default:
      {
      ArRef<TCPConnection> ref(new TCPConnection(TCPConnection::_CLS,
                                                 fd,ipAddr,portNumber,
                                                 _packetOriented));
      ref->setTransient(true);
      return(ref);
      }
    }
  }
}


//----------------------------------------------------------------------------
// Class UDPTransmitter
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(UDPTransmitter,IPSocket)

//-------- Construction / Destruction ----------------------------------------

UDPTransmitter::UDPTransmitter(ArCW & arCW,
                               int portNumber) // Protected
: IPSocket(arCW),
  _memory(MemoryBlock::NEW())
{
_socket=ArSystem::openUDP(portNumber);
if(_socket==-1)
  {
  setErrorMessage("Cannot open UDP port `"+intToStr(portNumber)+
                  "' --- "+ArSystem::printError(errno));
  return;
  }
if(!ArSystem::getSocketAddress(_socket,_ipAddress,_portNumber))
  {
  setErrorMessage("Cannot retrieve socket parameters --- "+
                  ArSystem::printError(errno));
  _cancel();
  return;
  }
setBlocking(false);
if(fail())
  {
  return;
  }
_memory->setSize(SOCKET_RECEPTION_SIZE);
}

UDPTransmitter::~UDPTransmitter(void) // Protected
{
  // Nothing to be done
}

//-------- Binary communication ----------------------------------------------

bool // bytes available
UDPTransmitter::receiveBytes(ArRef<MemoryBlock> memory,
                             unsigned int & distantIPAddressOut,
                             int & distantPortNumberOut)
{
if(_socket==-1) return(false);
for(;;)
  {
  int r=ArSystem::receiveUDP(_socket,
                             _memory->accessMemory(),_memory->getSize(),
                             distantIPAddressOut,distantPortNumberOut);
  switch(r)
    {
    case 0: // EOF
      {
      setErrorMessage("EOF reached");
      _cancel();
      return(false);
      }
    case -1: // error
      {
      setErrorMessage("PB while receiving --- "+ArSystem::printError(errno));
      _cancel();
      return(false);
      }
    case -2: // non-blocking
      {
      if(!_blocking)
        {
        return(false);
        }
      break;
      }
    default: // something
      {
      _memory->setSize(r); // shrink --> no realloc --> no copy
      bool success=memory->swapBlock(_memory);
      _memory->setSize(0); // free to avoid copy in realloc
      _memory->setSize(SOCKET_RECEPTION_SIZE);
      return(success);
      }
    }
  }
}

void
UDPTransmitter::sendBytes(ArConstRef<MemoryBlock> memory,
                          unsigned int offset,
                          unsigned int length,
                          unsigned int distantIPAddress,
                          int distantPortNumber)
{
if(_socket!=-1)
  {
  if(ArSystem::sendUDP(_socket,memory->getMemory()+offset,length,
                       distantIPAddress,distantPortNumber)==-1)
    {
    setErrorMessage("PB while sending --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
}

void
UDPTransmitter::broadcastBytes(ArConstRef<MemoryBlock> memory,
                               unsigned int offset,
                               unsigned int length,
                               int distantPortNumber)
{
if(_socket!=-1)
  {
  const StlVector<unsigned int> & addr=ArSystem::getUDPBroadcastAddresses();
  unsigned int nbErr=0;
  for(unsigned int i=0;i<addr.size();i++)
    {
    if(ArSystem::sendUDP(_socket,memory->getMemory()+offset,length,
                         addr[i],distantPortNumber)==-1)
      {
      ++nbErr;
      }
    }
  if(nbErr==addr.size())
    {
    setErrorMessage("PB while broadcasting --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
}

//-------- Textual communication ---------------------------------------------

bool // string available
UDPTransmitter::receiveString(StlString & resultOut,
                              unsigned int & distantIPAddressOut,
                              int & distantPortNumberOut)
{
if(_socket==-1) return(false);
for(;;)
  {
  int r=ArSystem::receiveUDP(_socket,
                             _memory->accessMemory(),_memory->getSize()-1,
                             distantIPAddressOut,distantPortNumberOut);
  switch(r)
    {
    case 0: // EOF
      {
      setErrorMessage("EOF reached");
      _cancel();
      return(false);
      }
    case -1: // error
      {
      setErrorMessage("PB while receiving --- "+ArSystem::printError(errno));
      _cancel();
      return(false);
      }
    case -2: // non-blocking
      {
      if(!_blocking)
        {
        return(false);
        }
      break;
      }
    default: // something
      {
      char * bytes=_memory->accessMemory();
      bytes[r]='\0'; // ensure string is ended
      resultOut=bytes;
      return(true);
      }
    }
  }
}

void
UDPTransmitter::sendString(const StlString & str,
                           unsigned int distantIPAddress,
                           int distantPortNumber)
{
if(_socket!=-1)
  {
  if(ArSystem::sendUDP(_socket,str.c_str(),str.size()+1,
                       distantIPAddress,distantPortNumber)==-1)
    {
    setErrorMessage("PB while sending --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
}

void
UDPTransmitter::broadcastString(const StlString & str,
                                int distantPortNumber)
{
if(_socket!=-1)
  {
  const StlVector<unsigned int> & addr=ArSystem::getUDPBroadcastAddresses();
  unsigned int nbErr=0;
  for(unsigned int i=0;i<addr.size();i++)
    {
    if(ArSystem::sendUDP(_socket,str.c_str(),str.size()+1,
                         addr[i],distantPortNumber)==-1)
      {
      ++nbErr;
      }
    }
  if(nbErr==addr.size())
    {
    setErrorMessage("PB while broadcasting --- "+ArSystem::printError(errno));
    _cancel();
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
