// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/thread.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/thread.h"

#include "AReVi/arSystem.h"

extern "C" {
#include <sched.h>
}

namespace AReVi {

//----------------------------------------------------------------------------
// Class Thread
//----------------------------------------------------------------------------

AR_CLASS_DEF(Thread,ArObject)

//-------- Construction / Destruction ----------------------------------------

Thread::Thread(ArCW & arCW) // Protceted
: ArObject(arCW),
  _started(false),
  _ended(false),
  _obj(),
  _method((void (ArObject::*)(void))0),
  _th()
{
  // Nothing to be done
}

Thread::~Thread(void) // Protected
{
  // Nothing to be done
}

//-------- Ending ------------------------------------------------------------

bool // thread is running
Thread::isRunning(void) const
{
return(_started&&!_ended);
}

bool // success
Thread::waitEnd(void) const
{
if(_started&&!pthread_equal(_th,pthread_self()))
  {
  while(!_ended) // caller waits termination
    {
    sched_yield(); // not join --> many waiting threads allowed
    }
  return(true);
  }
else
  {
  return(false);
  }
}

//-------- Forbidden operations ----------------------------------------------

bool // success
Thread::_start(void) // Protected
{
_started=true;
_ended=false;
int r=pthread_create(&_th,(pthread_attr_t *)0,                                                       Thread::_threadTask,(void *)this);
if(r)
  {
  setErrorMessage("Cannot create thread --- "+ArSystem::printError(r));
  return(false);
  }
else
  {
  setErrorMessage(""); // clear previous errors
  return(true);
  }
}

void *
Thread::_threadTask(void * arg) // Static, Protected
{
ArRef<Thread> thread(static_cast<Thread *>(arg));
thread->_th=pthread_self();
pthread_detach(thread->_th);
(thread->_obj.c_ptr()->*thread->_method)();
if(thread)
  {
  thread->_ended=true;
  thread->_obj.clear();
  thread->_method=((void (ArObject::*)(void))0);
  }
return((void *)0);
}


//----------------------------------------------------------------------------
// Class ThreadMutex
//----------------------------------------------------------------------------

AR_CLASS_DEF(ThreadMutex,ArObject)

//-------- Construction / Destruction ----------------------------------------

ThreadMutex::ThreadMutex(ArCW & arCW) // Protected
: ArObject(arCW)
  // _mtx is not initialized !
{
int r=pthread_mutex_init(&_mtx,(pthread_mutexattr_t *)0);
if(r)
  {
  setErrorMessage("Cannot create mutex --- "+ArSystem::printError(r));
  }
}

ThreadMutex::~ThreadMutex(void) // Protected
{
if(!fail())
  {
  pthread_mutex_destroy(&_mtx);
  }
}

//-------- Execution control -------------------------------------------------

void
ThreadMutex::lock(void)
{
pthread_mutex_lock(&_mtx);
}

void
ThreadMutex::unlock(void)
{
pthread_mutex_unlock(&_mtx);
}

bool // lock success
ThreadMutex::tryLock(void)
{
return(!pthread_mutex_trylock(&_mtx));
}


//----------------------------------------------------------------------------
// Class RecursiveThreadMutex
//----------------------------------------------------------------------------

AR_CLASS_DEF(RecursiveThreadMutex,ThreadMutex)

//-------- Construction / Destruction ----------------------------------------

RecursiveThreadMutex::RecursiveThreadMutex(ArCW & arCW) // Protected
: ThreadMutex(arCW),
  _count(0)
  // _th is not initialized !
  // _cond is not initialized !
{
if(!fail())
  {
  int r=pthread_cond_init(&_cond,(pthread_condattr_t *)0);
  if(r)
    {
    setErrorMessage("Cannot create condition --- "+ArSystem::printError(r));
    pthread_mutex_destroy(&_mtx);
    }
  }
}

RecursiveThreadMutex::~RecursiveThreadMutex(void) // Protected
{
if(!fail())
  {
  pthread_cond_destroy(&_cond);
  }
}

//-------- Execution control -------------------------------------------------

void
RecursiveThreadMutex::lock(void)
{
pthread_mutex_lock(&_mtx);
while(_count&&!pthread_equal(_th,pthread_self()))
  {
  pthread_cond_wait(&_cond,&_mtx);
  }
if(!_count++)
  {
  _th=pthread_self();
  }
pthread_mutex_unlock(&_mtx);
}

void
RecursiveThreadMutex::unlock(void)
{
pthread_mutex_lock(&_mtx);
--_count;
pthread_cond_broadcast(&_cond);
pthread_mutex_unlock(&_mtx);
}

bool // lock success
RecursiveThreadMutex::tryLock(void)
{
pthread_mutex_lock(&_mtx);
if(_count)
  {
  if(pthread_equal(_th,pthread_self()))
    {
    ++_count;
    pthread_mutex_unlock(&_mtx);
    return(true); // recursive lock
    }
  else
    {
    pthread_mutex_unlock(&_mtx);
    return(false); // locked by another thread
    }
  }
else
  {
  ++_count;
  _th=pthread_self();
  pthread_mutex_unlock(&_mtx);
  return(true); // first lock
  }
}


//----------------------------------------------------------------------------
// Class ThreadCondition
//----------------------------------------------------------------------------

AR_CLASS_DEF(ThreadCondition,ArObject)

//-------- Construction / Destruction ----------------------------------------

ThreadCondition::ThreadCondition(ArCW & arCW) // Protected
: ArObject(arCW)
  // _cond is not initialized !
  // _mtx is not initialized !
{
int r=pthread_cond_init(&_cond,(pthread_condattr_t *)0);
if(r)
  {
  setErrorMessage("Cannot create condition --- "+ArSystem::printError(r));
  }
else
  {
  r=pthread_mutex_init(&_mtx,(pthread_mutexattr_t *)0);
  if(r)
    {
    setErrorMessage("Cannot create mutex --- "+ArSystem::printError(r));
    pthread_cond_destroy(&_cond);
    }
  }
}

ThreadCondition::~ThreadCondition(void) // Protected
{
if(!fail())
  {
  pthread_cond_destroy(&_cond);
  pthread_mutex_destroy(&_mtx);
  }
}

//-------- Execution control -------------------------------------------------

void
ThreadCondition::enter(void)
{
pthread_mutex_lock(&_mtx);
}

void
ThreadCondition::leave(void)
{
pthread_mutex_unlock(&_mtx);
}

void
ThreadCondition::wait(void)
{
pthread_cond_wait(&_cond,&_mtx);
}

void
ThreadCondition::broadcast(void)
{
pthread_cond_broadcast(&_cond);
}

} // namespace AReVi

//----------------------------------------------------------------------------
