// LI2/ENIB - AReVi - Utility services
//----------------------------------------------------------------------------
// file        : AReVi/Utils/utilSound.cpp
// author(s)   : E. Cazeaux, B. Muguet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/arSystem.h"
#include "AReVi/Utils/utilSound.h"

#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/fileStream.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class UtilSound (abstract)
//----------------------------------------------------------------------------

bool
UtilSound::createWavFileFromSoundBuffer(StlString filename,
                                        ArRef<MemoryBlock> soundBuffer,
                                        int sampleRate,
                                        short nbChannels,
                                        short bitsPerSample) // Static
{

  if(filename.empty())
    filename = "soundFile.wav";

  ArRef<FileOStream> wavStream = FileOStream::NEW(filename,false);
  if (!wavStream) {
    fprintf(stderr, "could not open %s\n", filename.c_str());
    return false;
  }

  //Check the sound parameters
  if((bitsPerSample != 8) && (bitsPerSample != 16))
    {
      cerr << "Error while creating wav file : bitsPerSample invalid" << endl;
      return false;
    }

  if((nbChannels != 1) && (nbChannels != 2))
    {
      cerr << "Error while creating wav file : nbChannels invalid" << endl;
      return false;
    }

  //----------------------------
  //--------Write header--------
  //----------------------------

  int headerSize = 44;
  int nbBytesData = soundBuffer->getSize();

  int length = (headerSize-8) + nbBytesData;

  wavStream->writeTxtString("RIFF");

  ArSystem::inOut32(&length);
  wavStream->writeInt(length);

  wavStream->writeTxtString("WAVE");

  // ------- paramètres du format 'WAV' -------------

  wavStream->writeTxtString("fmt ");

  //nombre d'octets utilisés après pour définir le format
  int nbOctets = 16;
  ArSystem::inOut32(&nbOctets);
  wavStream->writeInt(nbOctets);

  //numéro de format du fichier
  short format = 1; //pcm
  ArSystem::inOut16(&format);
  wavStream->writeShort(format);

  //nombre de canaux
  ArSystem::inOut16(&nbChannels);
  wavStream->writeShort(nbChannels);

  //fréquence d'échantillonnage (en Hz)
  ArSystem::inOut32(&sampleRate);
  wavStream->writeInt(sampleRate);

  //nombres d'octets par seconde
  int byteRate = sampleRate * nbChannels * (bitsPerSample/8);
  ArSystem::inOut32(&byteRate);
  wavStream->writeInt(byteRate);

  //Produit du nombre de canaux par le
  //nombre d'octets par échantillon (ici 1x1=1)
  short produit = nbChannels * (bitsPerSample/8);
  ArSystem::inOut16(&produit);
  wavStream->writeShort(produit);

  //bits par échantillon (valeurs
  //possibles : 8, 12 ou 16)
  ArSystem::inOut16(&bitsPerSample);
  wavStream->writeShort(bitsPerSample);

  //'data' : annonce l'arrivée des données
  wavStream->writeTxtString("data");

  //taille des données
  ArSystem::inOut32(&nbBytesData);
  wavStream->writeInt(nbBytesData);

  wavStream->flush();

  //------------------------------
  //--------Write sound data------
  //------------------------------

  wavStream->writeData((char*)(soundBuffer->accessMemory()),soundBuffer->getSize());

  wavStream->flush();

  return true;
}

bool
UtilSound::createWavFileFromSoundBuffers(StlString filename,
                                         StlList<ArRef<MemoryBlock> >
                                                                 soundBuffers,
                                         int sampleRate,
                                         short nbChannels,
                                         short bitsPerSample) // Static
{

  if(filename.empty())
    filename = "soundFile.wav";

  //Get the size of sound data
  int nbBytesData = 0;

  StlList<ArRef<MemoryBlock> >::iterator it;
  for(it = soundBuffers.begin();it != soundBuffers.end();it++)
    nbBytesData += (*it)->getSize();

  ArRef<FileOStream> wavStream = FileOStream::NEW(filename,false);
  if (!wavStream) {
    fprintf(stderr, "could not open %s\n", filename.c_str());
    return false;
  }

    //Check the sound parameters
  if((bitsPerSample != 8) && (bitsPerSample != 16))
    {
      cerr << "Error while creating wav file : bitsPerSample invalid" << endl;
      return false;
    }

  if((nbChannels != 1) && (nbChannels != 2))
    {
      cerr << "Error while creating wav file : nbChannels invalid" << endl;
      return false;
    }

  //----------------------------
  //--------Write header--------
  //----------------------------

  int headerSize = 44;

  int length = (headerSize-8) + nbBytesData;

  wavStream->writeTxtString("RIFF");

  ArSystem::inOut32(&length);
  wavStream->writeInt(length);

  wavStream->writeTxtString("WAVE");

  // ------- paramètres du format 'WAV' -------------

  wavStream->writeTxtString("fmt ");

  //nombre d'octets utilisés après pour définir le format
  int nbOctets = 16;
  ArSystem::inOut32(&nbOctets);
  wavStream->writeInt(nbOctets);

  //numéro de format du fichier
  short format = 1; //pcm
  ArSystem::inOut16(&format);
  wavStream->writeShort(format);

  //nombre de canaux
  ArSystem::inOut16(&nbChannels);
  wavStream->writeShort(nbChannels);

  //fréquence d'échantillonnage (en Hz)
  ArSystem::inOut32(&sampleRate);
  wavStream->writeInt(sampleRate);

  //nombres d'octets par seconde
  int byteRate = sampleRate * nbChannels * (bitsPerSample/8);
  ArSystem::inOut32(&byteRate);
  wavStream->writeInt(byteRate);

  //Produit du nombre de canaux par le
  //nombre d'octets par échantillon (ici 1x1=1)
  short produit = nbChannels * (bitsPerSample/8);
  ArSystem::inOut16(&produit);
  wavStream->writeShort(produit);

  //bits par échantillon (valeurs
  //possibles : 8, 12 ou 16)
  ArSystem::inOut16(&bitsPerSample);
  wavStream->writeShort(bitsPerSample);

  //'data' : annonce l'arrivée des données
  wavStream->writeTxtString("data");

  //taille des données
  ArSystem::inOut32(&nbBytesData);
  wavStream->writeInt(nbBytesData);

  wavStream->flush();

  //------------------------------
  //--------Write sound data------
  //------------------------------

  for(it = soundBuffers.begin();it != soundBuffers.end();it++)
    wavStream->writeData((char*)((*it)->accessMemory()),(*it)->getSize());

  wavStream->flush();

  return true;
}

} // namespace AReVi

//----------------------------------------------------------------------------
