// LI2/ENIB - AReVi - VRML services
//----------------------------------------------------------------------------
// file        : AReVi/VRML/extrusionNode_priv.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/VRML/extrusionNode_priv.h"

#include "AReVi/VRML/vrmlValue_priv.h"
#include "AReVi/VRML/vrmlShape3D.h"

#include "AReVi/Shapes/mesh3D.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class ExtrusionNode
//----------------------------------------------------------------------------

//-------- Construction / Destruction ----------------------------------------

ExtrusionNode::~ExtrusionNode(void)
{
  // Nothing to be done
}

//-------- Field management --------------------------------------------------

void
ExtrusionNode::addField(const StlString & fieldName,
                        VrmlValue * fieldValue)
{
bool ok=false;
if(fieldName=="beginCap")
  {
  if(fieldValue->checkSFBool())
    {
    _beginCapField=fieldValue->getBool(0);
    ok=true;
    }
  }
else if(fieldName=="ccw")
  {
  if(fieldValue->checkSFBool())
    {
    _ccwField=fieldValue->getBool(0);
    ok=true;
    }
  }
else if(fieldName=="convex")
  {
  if(fieldValue->checkSFBool())
    {
    _convexField=fieldValue->getBool(0);
    ok=true;
    }
  }
else if(fieldName=="creaseAngle")
  {
  if(fieldValue->checkSFFloat()&&
     (fieldValue->getReal(0)>=0.0))
    {
    _creaseAngleField=fieldValue->getReal(0);
    ok=true;
    }
  }
else if(fieldName=="crossSection")
  {
  if(fieldValue->checkMFVec2f())
    {
    _crossSectionField.clear();
    unsigned int nb=fieldValue->getSize();
    for(unsigned int i=0;i<nb;i+=2)
      {
      _crossSectionField.push_back(Util3D::Dbl2(fieldValue->getReal(i),
                                                fieldValue->getReal(i+1)));
      }
    ok=true;
    }
  }
else if(fieldName=="endCap")
  {
  if(fieldValue->checkSFBool())
    {
    _endCapField=fieldValue->getBool(0);
    ok=true;
    }
  }
else if(fieldName=="orientation")
  {
  if(fieldValue->checkMFRotation())
    {
    _orientationField.clear();
    unsigned int nb=fieldValue->getSize();
    for(unsigned int i=0;i<nb;i+=4)
      {
      _orientationField.push_back(Util3D::Dbl4(fieldValue->getReal(i),
                                               fieldValue->getReal(i+1),
                                               fieldValue->getReal(i+2),
                                               fieldValue->getReal(i+3)));
      }
    ok=true;
    }
  }
else if(fieldName=="scale")
  {
  if(fieldValue->checkMFVec2f())
    {
    unsigned int nb=fieldValue->getSize();
    ok=true;
    for(unsigned int i=nb;i--;)
      {
      if(fieldValue->getReal(i)<=0.0)
        {
        ok=false;
        break;
        }
      }
    if(ok)
      {
      _scaleField.clear();
      for(unsigned int i=0;i<nb;i+=2)
        {
        _scaleField.push_back(Util3D::Dbl2(fieldValue->getReal(i),
                                           fieldValue->getReal(i+1)));
        }
      }
    }
  }
else if(fieldName=="solid")
  {
  if(fieldValue->checkSFBool())
    {
    _solidField=fieldValue->getBool(0);
    ok=true;
    }
  }
else if(fieldName=="spine")
  {
  if(fieldValue->checkMFVec3f())
    {
    _spineField.clear();
    unsigned int nb=fieldValue->getSize();
    for(unsigned int i=0;i<nb;i+=3)
      {
      _spineField.push_back(Util3D::Dbl3(fieldValue->getReal(i),
                                         fieldValue->getReal(i+1),
                                         fieldValue->getReal(i+2)));
      }
    ok=true;
    }
  }
else
  {
  GeometryNode::addField(fieldName,fieldValue);
  ok=true;
  }
if(!ok)
  {
  _shape->_badFieldMsg(fieldName,_type);
  }
}

void
ExtrusionNode::check(void)
{
_fail=false;
// FIXME: do something ...
_shape->_ignoreNodeMsg(_type);
_fail=true;
if(!_fail)
  {
  rebuild();
  }
}

//-------- Rendering ---------------------------------------------------------

ArRef<ShapePart3D> // built part
ExtrusionNode::buildPart(Util3D::Transform & initTransformOut) const
{
initTransformOut.identity();
return(Mesh3D::NEW());
}

void
ExtrusionNode::updatePart(ArRef<ShapePart3D> part,
                          const MaterialNode * /* material */) const
{
ArRef<Mesh3D> mesh=ar_down_cast<Mesh3D>(part);
// FIXME: update geometry
mesh->setDoubleSided(!_solidField);
mesh->applyChanges(false);
}


//-------- Forbidden operations ----------------------------------------------

void
ExtrusionNode::_generateFields(ostream & output,
                               unsigned int indent,
                               bool showDefaultValues,
                            StlVector<const VrmlNode *> & /* defNodesInOut */)
                                                            const // Protected
{
if(showDefaultValues||
   !_beginCapField)
  {
  _generateSFBool(output,indent,"beginCap",_beginCapField);
  }
if(showDefaultValues||
   !_ccwField)
  {
  _generateSFBool(output,indent,"ccw",_ccwField);
  }
if(showDefaultValues||
   !_convexField)
  {
  _generateSFBool(output,indent,"convex",_convexField);
  }
if(showDefaultValues||
   _creaseAngleField)
  {
  _generateSFFloat(output,indent,"creaseAngle",_creaseAngleField);
  }
if(showDefaultValues||
   (_crossSectionField.size()!=5)||
   (_crossSectionField[0].x!=1.0f)||
   (_crossSectionField[0].y!=1.0f)||
   (_crossSectionField[1].x!=1.0f)||
   (_crossSectionField[1].y!=-1.0f)||
   (_crossSectionField[2].x!=-1.0f)||
   (_crossSectionField[2].y!=-1.0f)||
   (_crossSectionField[3].x!=-1.0f)||
   (_crossSectionField[3].y!=1.0f)||
   (_crossSectionField[4].x!=1.0f)||
   (_crossSectionField[4].y!=1.0f))
  {
  _generateMFVec2f(output,indent,"crossSection",_crossSectionField);
  }
if(showDefaultValues||
   !_endCapField)
  {
  _generateSFBool(output,indent,"endCap",_endCapField);
  }
if(showDefaultValues||
   (_orientationField.size()!=1)||
   _orientationField[0].x||
   _orientationField[0].y||
   (_orientationField[0].z!=1.0f)||
   _orientationField[0].a)
  {
  _generateMFRotation(output,indent,"orientation",_orientationField);
  }
if(showDefaultValues||
   (_scaleField.size()!=1)||
   (_scaleField[0].x!=1.0f)||
   (_scaleField[0].y!=1.0f))
  {
  _generateMFVec2f(output,indent,"scale",_scaleField);
  }
if(showDefaultValues||
   !_solidField)
  {
  _generateSFBool(output,indent,"solid",_solidField);
  }
if(showDefaultValues||
   (_spineField.size()!=2)||
   _spineField[0].x||
   _spineField[0].y||
   _spineField[0].z||
   _spineField[1].x||
   (_spineField[1].y!=1.0f)||
   _spineField[1].z)
  {
  _generateMFVec3f(output,indent,"spine",_spineField);
  }
}

ExtrusionNode::ExtrusionNode(const StlString & nodeType,
                             VrmlShape3D * shape) // Protected
: GeometryNode(nodeType,shape),
  _beginCapField(true),
  _ccwField(true),
  _convexField(true),
  _creaseAngleField(0.0),
  _crossSectionField(),
  _endCapField(true),
  _orientationField(),
  _scaleField(),
  _solidField(true),
  _spineField()
{
_crossSectionField.push_back(Util3D::Dbl2(1.0,1.0));
_crossSectionField.push_back(Util3D::Dbl2(1.0,-1.0));
_crossSectionField.push_back(Util3D::Dbl2(-1.0,-1.0));
_crossSectionField.push_back(Util3D::Dbl2(-1.0,1.0));
_crossSectionField.push_back(Util3D::Dbl2(1.0,1.0));
_orientationField.push_back(Util3D::Dbl4(0.0,0.0,1.0,0.0));
_scaleField.push_back(Util3D::Dbl2(1.0,1.0));
_spineField.push_back(Util3D::Dbl3(0.0,0.0,0.0));
_spineField.push_back(Util3D::Dbl3(0.0,1.0,0.0));
}

} // namespace AReVi

//----------------------------------------------------------------------------
