// LI2/ENIB - AReVi - VRML services
//----------------------------------------------------------------------------
// file        : AReVi/VRML/vrmlShape3D.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/VRML/vrmlShape3D.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/urlStream.h"
#include "AReVi/Utils/memoryStream.h"
#include "AReVi/Utils/memoryBlock.h"

#include "AReVi/VRML/vrmlParser_priv.h"
#include "AReVi/VRML/vrmlInteractor.h"
#include "AReVi/VRML/shapeNode_priv.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class VrmlShape3D
//----------------------------------------------------------------------------

AR_CLASS_DEF(VrmlShape3D,Shape3D)

//-------- Construction / Destruction ----------------------------------------

VrmlShape3D::VrmlShape3D(ArCW & arCW) // Protected
: Shape3D(arCW),
  _msg(),
  _rootNodes(),
  _namedNodes(),
  _transformShapes(),
  _url(),
  _urlDir(),
  _parser((VrmlParser *)0),
  _interactors()
{
  // Nothing to be done
}

VrmlShape3D::~VrmlShape3D(void) // Protected
{
forgetAll();
}

//-------- Parsing -----------------------------------------------------------

bool // success
VrmlShape3D::parseURL(const StlString & url)
{
forgetAll();
_url=url;
_urlDir=strDirname(_url);
ArRef<URLStream> input=URLStream::NEW(_url);
if(input->fail())
  {
  _errorMsg("Cannot read URL '"+_url+"'");
  }
else
  {
  _parser=new VrmlParser(&input->stream(),this);
  _parser->parse();
  delete _parser;
  _parser=(VrmlParser *)0;
  _buildParts();
  }
return(!_msg.size());
}

bool // success
VrmlShape3D::parseString(const StlString & vrmlString)
{
char * ptr=const_cast<char *>(vrmlString.c_str());
ArRef<RigidMemoryBlock> memory=RigidMemoryBlock::NEW(ptr,vrmlString.size(),
                                                    false);
return(parseStream(MemoryIStream::NEW(memory,0)));
}

bool // success
VrmlShape3D::parseStream(ArRef<AbstractIStream> input)
{
forgetAll();
if(input->fail())
  {
  _errorMsg("Cannot read input stream");
  }
else
  {
  _parser=new VrmlParser(&input->stream(),this);
  _parser->parse();
  delete _parser;
  _parser=(VrmlParser *)0;
  _buildParts();
  }
return(!_msg.size());
}

const StlString & // messages when parsing
VrmlShape3D::readParserMessages(void) const
{
return(_msg);
}

void
VrmlShape3D::forgetAll(void)
{
while(!_rootNodes.empty())
  {
  _rootNodes.back()->removeParent((VrmlNode *)0);
  _rootNodes.pop_back();
  }
_msg.clear();
_url.clear();
_urlDir.clear();
while(!_interactors.empty())
  {
  _removeInteractor(_interactors.back());
  }
_needsRebuild=true;
}

//-------- URL management ----------------------------------------------------

const StlString & // URL
VrmlShape3D::getURL(void) const
{
return(_url);
}

const StlString & // directory of URL
VrmlShape3D::getURLDirectory(void) const
{
return(_urlDir);
}

void
VrmlShape3D::setURLDirectory(const StlString & urlDir)
{
_urlDir=urlDir;
if(_urlDir.size()&&(_urlDir[_urlDir.size()-1]!='/'))
  {
  _urlDir+='/';
  }
}

//-------- Code generation ---------------------------------------------------

void
VrmlShape3D::generateString(StlString & vrmlStringOut,
                            unsigned int indent,
                            bool showDefaultValues) const
{
ArRef<MemoryBlock> memory=MemoryBlock::NEW();
ArRef<MemoryOStream> stream=MemoryOStream::NEW(memory,0);
generateStream(stream,indent,showDefaultValues);
vrmlStringOut=StlString(memory->getMemory(),stream->getMemoryUsage());
}

void
VrmlShape3D::generateStream(ArRef<AbstractOStream> output,
                            unsigned int indent,
                            bool showDefaultValues) const
{
StlVector<const VrmlNode *> defNodes;
for(unsigned int i=0;i<_rootNodes.size();i++)
  {
  _rootNodes[i]->generateCode(output->stream(),indent,
                              showDefaultValues,defNodes);
  }
}

//-------- Forbidden operations ----------------------------------------------

void
VrmlShape3D::_ignoreNodeMsg(const StlString & nodeType) // Protected
{
StlString msg="Ignoring node type '";
msg+=nodeType;
msg+="'";
_errorMsg(msg);
}

void
VrmlShape3D::_ignoreFieldMsg(const StlString & fieldName,
                             const StlString & nodeType) // Protected
{
StlString msg="Ignoring field '";
msg+=fieldName;
msg+="' in node type '";
msg+=nodeType;
msg+="'";
_errorMsg(msg);
}

void
VrmlShape3D::_badFieldMsg(const StlString & fieldName,
                          const StlString & nodeType) // Protected
{
StlString msg="Bad value for field '";
msg+=fieldName;
msg+="' in node type '";
msg+=nodeType;
msg+="'";
_errorMsg(msg);
}

void
VrmlShape3D::_errorMsg(const StlString & msg) // Protected
{
StlString str="line ";
if(_parser)
  {
  _parser->_nbErrors++;
  str+=intToStr(_parser->_nbLines);
  }
else
  {
  str+="???";
  }
str+=": ";
str+=msg;
str+='\n';
_msg+=str;
}

void
VrmlShape3D::_addRootNode(GroupChildNode * node) // Protected
{
node->addParent((VrmlNode *)0);
_rootNodes.push_back(node);
}

unsigned int
VrmlShape3D::_getNbRootNodes(void) const // Protected
{
return(_rootNodes.size());
}

const GroupChildNode *
VrmlShape3D::_getRootNode(unsigned int index) const // Protected
{
return(_rootNodes[index]);
}

GroupChildNode *
VrmlShape3D::_accessRootNode(unsigned int index) // Protected
{
return(_rootNodes[index]);
}

void
VrmlShape3D::_addNamedNode(VrmlNode * node) // Protected
{
_namedNodes.push_back(node);
}

void
VrmlShape3D::_removeNamedNode(VrmlNode * node) // Protected
{
for(unsigned int i=_namedNodes.size();i--;)
  {
  if(_namedNodes[i]==node)
    {
    _namedNodes.erase(_namedNodes.begin()+i);
    break;
    }
  }
}

const VrmlNode * // found node
VrmlShape3D::_getNamedNode(const StlString & nodeName,
                           int occurence) const // Protected
{
if(occurence<0)
  {
  for(unsigned int i=_namedNodes.size();i--;)
    {
    if(_namedNodes[i]->getName()==nodeName)
      {
      if(!++occurence)
        {
        return(_namedNodes[i]);
        }
      }
    }
  }
else
  {
  unsigned int nb=_namedNodes.size();
  for(unsigned int i=0;i<nb;i++)
    {
    if(_namedNodes[i]->getName()==nodeName)
      {
      if(!--occurence)
        {
        return(_namedNodes[i]);
        }
      }
    }
  }
return((const VrmlNode *)0);
}

VrmlNode * // found node
VrmlShape3D::_accessNamedNode(const StlString & nodeName,
                              int occurence) // Protected
{
if(occurence<0)
  {
  for(unsigned int i=_namedNodes.size();i--;)
    {
    if(_namedNodes[i]->getName()==nodeName)
      {
      if(!++occurence)
        {
        return(_namedNodes[i]);
        }
      }
    }
  }
else
  {
  unsigned int nb=_namedNodes.size();
  for(unsigned int i=0;i<nb;i++)
    {
    if(_namedNodes[i]->getName()==nodeName)
      {
      if(!--occurence)
        {
        return(_namedNodes[i]);
        }
      }
    }
  }
return((VrmlNode *)0);
}

void
VrmlShape3D::_buildParts(void) // Protected
{
StlVector<TransformNode *> transformPath;
for(unsigned int i=_rootNodes.size();i--;)
  {
  _rootNodes[i]->buildParts(transformPath);
  }
_rebuildIfNeeded();
}

void
VrmlShape3D::_rebuildIfNeeded(void) // Protected
{
if(_needsRebuild)
  {
  for(unsigned int i=_rootNodes.size();i--;)
    {
    _rootNodes[i]->rebuild();
    }
  while(!_transformShapes.empty())
    {
    _transformShapes.back()->applyTransformations();
    _transformShapes.pop_back();
    }
  Shape3D::_rebuildIfNeeded();
  }
}

void
VrmlShape3D::_postShapeTransformation(ShapeNode * shape) // Protected
{
_transformShapes.push_back(shape);
}

void
VrmlShape3D::_addInteractor(VrmlInteractor * interactor) // Protected
{
_interactors.push_back(interactor);
}

void
VrmlShape3D::_removeInteractor(VrmlInteractor * interactor) // Protected
{
for(unsigned int i=_interactors.size();i--;)
  {
  if(_interactors[i]==interactor)
    {
    StlVectorFastErase(_interactors,i);
    interactor->_shapeRemoved();
    break;
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
