// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/arMethod.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/arMethod.h"

#include "AReVi/arValue.h"

#include <ctype.h>

namespace AReVi {

//----------------------------------------------------------------------------
// Class ArMethod
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(ArMethod,ArObject)

//-------- Construction / Destruction ----------------------------------------

ArMethod::ArMethod(ArCW & arCW,
                   ArConstRef<ArClass> declCls,
                   const StlString & declFile,
                   unsigned int declLine,
                   const StlString & mthName,
                   bool isStatic,
                   bool isConst,
                   ArMethod::Stub stub) // Private
: ArObject(arCW),
  _declCls(declCls),
  _declFile(declFile),
  _declLine(declLine),
  _mthName(mthName),
  _constructor(mthName=="AR_CONSTRUCTOR"),
  _isStatic(isStatic),
  _isConst(isConst),
  _stub(stub),
  _result(),
  _argTypes(),
  _argNames(),
  _argSignature("void")
{
ar_const_cast(_declCls)->_methods.push_back(thisRef());
setTransient(true);
}

ArMethod::~ArMethod(void) // Protected
{
for(unsigned int i=_declCls->_methods.size();i--;)
  {
  if(_declCls->_methods[i].c_ptr()==this)
    {
    StlVectorFastErase(ar_const_cast(_declCls)->_methods,i);
    break;
    }
  }
}

//-------- Inspection --------------------------------------------------------

ArConstRef<ArClass> // class declaring this method
ArMethod::getDeclarationClass(void) const
{
return(_declCls);
}

const StlString & // declaration header file of this method
ArMethod::getDeclarationFile(void) const
{
return(_declFile);
}

unsigned int // declaration line number of this method
ArMethod::getDeclarationLine(void) const
{
return(_declLine);
}

const StlString & // name of this method
ArMethod::getName(void) const
{
return(_mthName);
}

bool // this method is a constructor
ArMethod::isConstructor(void) const
{
return(_constructor);
}

bool // this method is static
ArMethod::isStaticMethod(void) const
{
return(_isStatic);
}

bool // this method is const
ArMethod::isConstMethod(void) const
{
return(_isConst);
}

ArConstRef<ArType> // returned type if any
ArMethod::getResultType(void) const
{
return(_result);
}

unsigned int // number of arguments
ArMethod::getNbArgs(void) const
{
return(_argTypes.size());
}

ArConstRef<ArType> // type of index'th argument
ArMethod::getArgType(unsigned int index) const
{
return(_argTypes[index]);
}

const StlString & // name of index'th argument
ArMethod::getArgName(unsigned int index) const
{
return(_argNames[index]);
}

const StlString & // C++ arguments
ArMethod::getArgSignature(void) const
{
return(_argSignature);
}

bool // C++ arguments match (ignoring spaces)
ArMethod::matchesArgSignature(const StlString & argSignature) const
{
const char * p1=_argSignature.c_str();
const char * p2=argSignature.c_str();
for(;;)
  {
  while(isspace(*p1))
    {
    ++p1;
    }
  while(isspace(*p2))
    {
    ++p2;
    }
  if(*p1==*p2)
    {
    if(*p1)
      {
      ++p1;
      ++p2;
      }
    else
      {
      break;
      }
    }
  else
    {
    return(false);
    }
  }
return(true);
}

//-------- Invocation --------------------------------------------------------

bool // success
ArMethod::invoke(ArRef<ArObject> obj,
                 ArRef<ArValue> resultOut,
                 const StlVector<ArRef<ArValue> > & argsInOut,
                 bool check) const
{
if(!resultOut&&_result)
  {
  resultOut=ArValue::NEW(_result);
  }
if(check)
  {
  if(_isStatic||!(obj&&obj->getClass()->isA(_declCls))||
     (_result&&!_result->isSameType(resultOut->getType()))||
     (argsInOut.size()!=_argTypes.size()))
    {
    return(false);
    }
  for(unsigned int i=_argTypes.size();i--;)
    {
    if(!_argTypes[i]->isSameType(argsInOut[i]->getType()))
      {
      return(false);
      }
    }
  }
(*_stub)(obj,resultOut,argsInOut);
return(true);
}

bool // success
ArMethod::invokeConst(ArConstRef<ArObject> obj,
                      ArRef<ArValue> resultOut,
                      const StlVector<ArRef<ArValue> > & argsInOut,
                      bool check) const
{
if(!resultOut&&_result)
  {
  resultOut=ArValue::NEW(_result);
  }
if(check)
  {
  if(_isStatic||!_isConst||!(obj&&obj->getClass()->isA(_declCls))||
     (_result&&!_result->isSameType(resultOut->getType()))||
     (argsInOut.size()!=_argTypes.size()))
    {
    return(false);
    }
  for(unsigned int i=_argTypes.size();i--;)
    {
    if(!_argTypes[i]->isSameType(argsInOut[i]->getType()))
      {
      return(false);
      }
    }
  }
(*_stub)(ar_const_cast(obj),resultOut,argsInOut);
return(true);
}

bool // success
ArMethod::invokeStatic(ArRef<ArValue> resultOut,
                       const StlVector<ArRef<ArValue> > & argsInOut,
                       bool check) const
{
if(!resultOut&&_result)
  {
  resultOut=ArValue::NEW(_result);
  }
if(check)
  {
  if(!_isStatic||
     (_result&&!_result->isSameType(resultOut->getType()))||
     (argsInOut.size()!=_argTypes.size()))
    {
    return(false);
    }
  for(unsigned int i=_argTypes.size();i--;)
    {
    if(!_argTypes[i]->isSameType(argsInOut[i]->getType()))
      {
      return(false);
      }
    }
  }
(*_stub)(ArObject::nullRef(),resultOut,argsInOut);
return(true);
}

//-------- Forbidden operations ----------------------------------------------

void
ArMethod::_setResult(ArType::DirectType type,
                     unsigned int dimension,
                     ArType::PassingMode passing) // Private
{
ArConstRef<ArType> t(new ArType(ArType::_CLS,type,dimension,passing));
ar_const_cast(t)->setTransient(true);
_result=t;
}

void
ArMethod::_setResult(ArType::IndirectType type,
                     ArConstRef<ArClass> cls,
                     unsigned int dimension,
                     ArType::PassingMode passing) // Private
{
ArConstRef<ArType> t(new ArType(ArType::_CLS,type,cls,dimension,passing));
ar_const_cast(t)->setTransient(true);
_result=t;
}

void
ArMethod::_addArg(const StlString & name,
                  ArType::DirectType type,
                  unsigned int dimension,
                  ArType::PassingMode passing) // Private
{
ArConstRef<ArType> t(new ArType(ArType::_CLS,type,dimension,passing));
ar_const_cast(t)->setTransient(true);
_argTypes.push_back(t);
_argNames.push_back(name);
if(_argSignature=="void")
  {
  _argSignature=t->getCppType();
  }
else
  {
  _argSignature+=",";
  _argSignature+=t->getCppType();
  }
}

void
ArMethod::_addArg(const StlString & name,
                  ArType::IndirectType type,
                  ArConstRef<ArClass> cls,
                  unsigned int dimension,
                  ArType::PassingMode passing) // Private
{
ArConstRef<ArType> t(new ArType(ArType::_CLS,type,cls,dimension,passing));
ar_const_cast(t)->setTransient(true);
_argTypes.push_back(t);
_argNames.push_back(name);
if(_argSignature=="void")
  {
  _argSignature=t->getCppType();
  }
else
  {
  _argSignature+=",";
  _argSignature+=t->getCppType();
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
