// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/arObject.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/arObject.h"

#include "AReVi/Utils/abstractStream.h"
#include "AReVi/message.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class SerializationDependencies
//----------------------------------------------------------------------------

//-------- Stream operations -------------------------------------------------

void
SerializationDependencies::writeToStream(ArRef<AbstractOStream> output) const
{
unsigned int i;
output->writeUInt(_references.size());
for(i=0;i<_references.size();i++)
  {
  output->writeString(_references[i]->getClass()->getName());
  }
output->writeUInt(_urls.size());
for(i=0;i<_urls.size();i++)
  {
  output->writeString(_urls[i]);
  }
}

bool
SerializationDependencies::readFromStream(ArRef<AbstractIStream> input)
{
unsigned int i,nbArObject,nbURL;
StlString className,url;
_references.clear();
_urls.clear();
if(!input->readUInt(nbArObject))
  {
  return(false);
  }
for(i=0;i<nbArObject;i++)
  {
  if(!input->readString(className))
    {
    return(false);
    }
  ArConstRef<ArClass> cls=ArClass::find(className);
  ArRef<ArObject> inst;
  cls->newInstance(inst);
  if(inst)
    {
    if(inst->fail())
      {
      cerr << inst->getErrorMessage() << endl;
      }
    else
      {
      inst->setTransient(false);
      }
    }
  addReference(inst);
  }
if(!input->readUInt(nbURL))
  {
  return(false);
  }
for(i=0;i<nbURL;i++)
  {
  if(!input->readString(url))
    {
    return(false);
    }
  addURL(url);
  }
return(true);
}


//----------------------------------------------------------------------------
// Class ArObject
//----------------------------------------------------------------------------

//-------- Type information --------------------------------------------------

ArCW ArObject::_CLS=(ArCW)0;

void
ArObject::REGISTER_CLASS(void)
{
if(!ArObject::_CLS)
  {
  ArClass::registerClass(ArObject::_CLS,"ArObject","",
                         &ArObject::_makeInstance);
  }
}

ArConstRef<ArClass> // ArClass objet for ArObject
ArObject::CLASS(void) // Static
{
assert(ArObject::_CLS!=(ArCW)0);
return(ArConstRef<ArClass>(ArObject::_CLS));
}

ArConstRef<ArClass> // class of this instance
ArObject::getClass(void) const
{
return(ArConstRef<ArClass>(_arCW));
}

unsigned int // instance ID in its class
ArObject::getClassID(void) const
{
return(_id);
}

//-------- Construction / Destruction ----------------------------------------

ArObject::ArObject(void) // Private, only used by ArClass
: _arCW(ArClass::_CLS),
  _id(0),
  _refs(),
  _ptrs(),
  _valid(true),
  _transient(false),
  _destroyCB(thisRef()),
  _alias(),
  _messages(),
  _messageCB(thisRef()),
  _errorMessage(),
  _userData((void *)0)
{
  // this instance will be managed later (in ArClass constructor)
}

ArObject::ArObject(ArCW & arCW) // Protected
: _arCW(arCW),
  _id(0),
  _refs(),
  _ptrs(),
  _valid(true),
  _transient(false),
  _destroyCB(thisRef()),
  _alias(),
  _messages(),
  _messageCB(thisRef()),
  _errorMessage(),
  _userData((void *)0)
{
_arCW->_manageObject(this);
}

ArObject::~ArObject(void) // Protected
{
if(_arCW)
  {
  _arCW->_unmanageObject(this);
  }
unsigned int i;
for(i=_ptrs.size();i--;)
  {
  *_ptrs[i]=(ArObject *)0;
  }
for(i=_refs.size();i--;)
  {
  *_refs[i]=(ArObject *)0;
  }
}

//-------- Garbage collection control ----------------------------------------

void
ArObject::setTransient(bool transient) // NOT virtual !
{
_transient=transient;
if(_transient&&_refs.empty())
  {
  _performDestruction();
  }
}

bool // transient
ArObject::isTransient(void) const // NOT virtual !
{
return(_transient);
}

unsigned int // nb references
ArObject::getNbReferences(void) const // NOT virtual !
{
return(_refs.size());
}

//-------- Alias management --------------------------------------------------

bool // success
ArObject::setAlias(const StlString & alias)
{
if(alias.size())
  {
  ArRef<ArObject> found;
  ArObject::CLASS()->findAlias(alias,found);
  if(found)
    {
    return(false);
    }
  else
    {
    const char * s=alias.c_str();
    for(unsigned int i=0;s[i];i++) // identifier format ?
      {
      if(i)
        {
        if(!isalnum(s[i])&&(s[i]!='_'))
          {
          return(false);
          }
        }
      else
        {
        if(!isalpha(s[i])&&(s[i]!='_'))
          {
          return(false);
          }
        }
      }
    _alias=alias;
    return(true);
    }
  }
else // default alias
  {
  _alias=getClass()->getName();
  _alias+=".";
  _alias+=uintToStr(_id);
  return(true);
  }
}

const StlString & // alias or className.ID
ArObject::getAlias(void) const
{
return(_alias);
}

//-------- Message management ------------------------------------------------

unsigned int // number of pending messages
ArObject::getNbMessages(void) const
{
return(_messages.size());
}

ArConstRef<Message> // next message or 0
ArObject::getNextMessage(void)
{
while(_messages.size())
  {
  ArConstRef<Message> msg=_messages.front();
  _messages.pop_front();
  if(msg)
    {
    return(msg);
    }
  }
return(Message::nullRef());
}

//-------- Stream operations -------------------------------------------------

void
ArObject::writeToStream(ArRef<AbstractOStream> /* output */) const
{
  // Nothing to be done
}

bool // success
ArObject::readFromStream(ArRef<AbstractIStream> input)
{
return(!input->fail());
}

//-------- Serialization operations ------------------------------------------

void
ArObject::extractDependencies(SerializationDependencies & /* dependenciesOut */,
                              bool /* recursive */) const
{
  // Nothing to be done
}

void
ArObject::serialize(SerializationDependencies & /* dependenciesOut */,
                    ArRef<AbstractOStream> /* output */) const
{
  // Nothing to be done
}

bool // success
ArObject::unserialize(const SerializationDependencies & /* dependencies */,
                      ArRef<AbstractIStream> input)
{
return(!input->fail());
}

//-------- Error management --------------------------------------------------

void
ArObject::setErrorMessage(const StlString & errorMessage)
{
if(errorMessage.size())
  {
  _errorMessage=getAlias();
  _errorMessage+=": ";
  _errorMessage+=errorMessage;
  }
else
  {
  _errorMessage.erase();
  }
}

bool // problem occured
ArObject::fail(void) const
{
return(_errorMessage.size() ? true : false);
}

const StlString & // problem description
ArObject::getErrorMessage(void) const
{
return(_errorMessage);
}

//-------- User data ---------------------------------------------------------

void
ArObject::setUserData(void * userData)
{
_userData=userData;
}

void * // user data
ArObject::getUserData(void) const
{
return(_userData);
}

//-------- Forbidden operations ----------------------------------------------

ArObject * // created instance
ArObject::_makeInstance(void) // Static, Private
{
return(new ArObject(ArObject::_CLS));
}

} // namespace AReVi

//----------------------------------------------------------------------------
