// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/arSystem.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/arSystem.h"

#include "AReVi/Utils/thread.h"
#include "AReVi/Utils/javaObject.h"

#include "AReVi/Lib3D/llw_priv.h"
#include "AReVi/Lib3D/window3D.h"
#include "AReVi/Lib3D/urlTexture.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/VRML/vrmlShape3D.h"

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/select.h>
#include <dirent.h>
#include <signal.h>
#include <sched.h>
#include <netdb.h>
#include <net/if.h>
#include <netinet/in.h>

#ifdef __sgi
  typedef int socklen_t;
  typedef int __itimer_which;
#endif // not __sgi

#ifdef __CYGWIN__
# include <windows.h>
  typedef int __itimer_which;
# define fdatasync(fd) fsync(fd)
#endif // __CYGWIN__

#ifdef __APPLE__
  // typedef int socklen_t;
  typedef int __itimer_which;
# define fdatasync(fd) fsync(fd)
#endif // __APPLE__

#ifdef __FreeBSD__
  typedef int __itimer_which;
# define fdatasync(fd) fsync(fd)
#endif // __FreeBSD__

namespace AReVi {

void registerAReViClasses(void);

//----------------------------------------------------------------------------
// Class ArSystem (singleton)
//----------------------------------------------------------------------------

ArSystem::InitState
ArSystem::_initState=ArSystem::AR_DOWN;

StlVector<StlString> *
ArSystem::_commandLine=(StlVector<StlString> *)0;

bool
ArSystem::_use3D=false;

ArSystem::ProcessID
ArSystem::_pid=(ArSystem::ProcessID)0;

ArSystem::ProcessID
ArSystem::_ppid=(ArSystem::ProcessID)0;

ArRef<ThreadMutex>
ArSystem::_mtx;

StlString
ArSystem::_httpProxyHost;

int
ArSystem::_httpProxyPort=-1;

StlVector<unsigned int> *
ArSystem::_broadcast=(StlVector<unsigned int> *)0;

StlVector<int> *
ArSystem::_readingPoll=(StlVector<int> *)0;

StlVector<int> *
ArSystem::_writingPoll=(StlVector<int> *)0;

ArRef<Scheduler>
ArSystem::_sched;

bool
ArSystem::_goOn=false;

StlVector<ArSystem::PluginInfo> *
ArSystem::_plugins=(StlVector<ArSystem::PluginInfo> *)0;

unsigned int
ArSystem::_randomSeed=0;

unsigned long long
ArSystem::_randomCount=0;

//-------- Construction / Destruction ----------------------------------------

ArSystem::ArSystem(int argc,
                   char ** argv,
                   bool use3D)
{
int dummy;
char * ptr1=reinterpret_cast<char *>(&dummy);
char * ptr2=reinterpret_cast<char *>(this);
if(ptr2<ptr1)
  {
  cerr << "ArSystem should be a local variable !" << endl;
  exit(-1);
  }
ArSystem::init(argc,argv,use3D);
}

ArSystem::~ArSystem(void)
{
ArSystem::destroy();
}

//-------- Initialization ----------------------------------------------------

void
ArSystem::init(int argc,
               char ** argv,
               bool use3D) // Static
{
if(ArSystem::_initState!=ArSystem::AR_DOWN)
  {
  cerr << "ArSystem is already initialized !" << endl;
  exit(-1);
  }
else
  {
  ArSystem::_initState=ArSystem::AR_GOING_UP;
  registerAReViClasses();
  ArSystem::_commandLine=new StlVector<StlString>;
  for(int i=0;i<argc;i++)
    {
    ArSystem::_commandLine->push_back(argv[i]);
    }
  ArSystem::_use3D=use3D;
  ArSystem::_pid=getpid();
  ArSystem::_ppid=getppid();
  ArSystem::_mtx=ThreadMutex::NEW();
  if(ArSystem::_mtx->fail())
    {
    cerr << "ArSystem: PB during mutex initialisation !" << endl;
    cerr << "    " << ArSystem::_mtx->getErrorMessage() << endl;
    exit(-1);
    }
  ArSystem::_httpProxyHost.erase();
  ArSystem::_httpProxyPort=-1;
  ArSystem::_broadcast=new StlVector<unsigned int>;
  ArSystem::_readingPoll=new StlVector<int>;
  ArSystem::_writingPoll=new StlVector<int>;
  ArSystem::_plugins=new StlVector<ArSystem::PluginInfo>;
  //---- Check size of integers ----
  {
  if(sizeof(char)!=1)
    {
    cerr << "ArSystem: a `char' should be 8 bits long ("
         << sizeof(char)*8 << " detected) !" << endl;
    exit(-1);
    }
  if(sizeof(short)!=2)
    {
    cerr << "ArSystem: a `short' should be 16 bits long ("
         << sizeof(short)*8 << " detected) !" << endl;
    exit(-1);
    }
  if(sizeof(int)!=4)
    {
    cerr << "ArSystem: an `int' should be 32 bits long ("
         << sizeof(int)*8 << " detected) !" << endl;
    exit(-1);
    }
  if(sizeof(long long)!=8)
    {
    cerr << "ArSystem: a `long long' should be 64 bits long ("
         << sizeof(long long)*8 << " detected) !" << endl;
    exit(-1);
    }
  }
  //---- Check size of pointers ----
  {
  if(sizeof(void *)!=sizeof(void (*)(void)))
    {
    cerr << "ArSystem: pointers to objets and to functions "
         << "don't have the same size !" << endl;
    exit(-1);
    }
  }
  //---- Check little/big endian ----
  {
  bool fail=false;
  short test=0x1234;
  switch(*((char *)&test))
    {
    case 0x12:
      if(ArSystem::isLittleEndian()) fail=true;
      break;
    case 0x34:
      if(ArSystem::isBigEndian()) fail=true;
      break;
    default:
      fail=true;
    }
  if(fail)
    {
    cerr << "ArSystem: PB in little/big endian detection !" << endl;
    exit(-1);
    }
  }
  //---- Check float/double representation ----
  {
  unsigned int test[2]={ 0xdeadbeef , 0xdeadbeef };
  float testFloat=-6.25985339871e+18F;
  double testDouble=-1.1885959257070704e+148;
  if(memcmp(test,&testFloat,sizeof(float))||
     memcmp(test,&testDouble,sizeof(double)))
    {
    cerr << "ArSystem: PB in float/double format detection !" << endl;
    exit(-1);
    }
  }
  //---- Check double/int structure alignment ----
  {
  Util3D::Dbl3 d3[2];
  double d[6];
  Util3D::Face n3[2];
  unsigned int n[6];
  for(uint32_t i=2;i--;)
    {
    if(((&d3[i].x-&d3[0].x)!=(&d[3*i+0]-&d[0]))||
       ((&d3[i].y-&d3[0].x)!=(&d[3*i+1]-&d[0]))||
       ((&d3[i].z-&d3[0].x)!=(&d[3*i+2]-&d[0]))||
       ((&n3[i].p1-&n3[0].p1)!=(&n[3*i+0]-&n[0]))||
       ((&n3[i].p2-&n3[0].p1)!=(&n[3*i+1]-&n[0]))||
       ((&n3[i].p3-&n3[0].p1)!=(&n[3*i+2]-&n[0])))
      {
      cerr << "ArSystem: PB in double/int structure alignment !" << endl;
      exit(-1);
      }
    }
  }
  //---- Initialize various services ----
#ifdef __CYGWIN__
  char * path=getenv("PATH");
  if(path)
    {
    char * ldPath=getenv("LD_LIBRARY_PATH");
    if(ldPath&&ldPath[0]&&strcmp(ldPath,path))
      {
      cerr << "Warning, overriding LD_LIBRARY_PATH with PATH..." << endl;
      }
    setenv("LD_LIBRARY_PATH",path,1);
    }
#endif // __CYGWIN__
  ArSystem::setRandomSeed(ArSystem::getTime(),0);
  char * var=getenv("http_proxy");
  if(!var)
    {
    var=getenv("HTTP_PROXY");
    }
  if(var)
    {
    StlString proto,host,file;
    int port;
    if(ArSystem::decodeURL(var,proto,host,port,file)&&
       (proto=="http")&&(file=="/"))
      {
      ArSystem::setHTTPProxy(host,port);
      }
    }
  if(ArSystem::_use3D)
    {
    LLW::init();
    if(LLW::initFail())
      {
      cerr << "ArSystem: Cannot initialize 3D !" << endl;
      exit(-1);
      }
    else
      {
      Renderer3D::_init();
      Texture::_init();
      URLTexture::_init();
      Scene3D::_init();
      }
    }
  ArSystem::_forceSettings();
  ArSystem::_initState=ArSystem::AR_UP;
  }
}

void
ArSystem::destroy(void) // Static
{
if(ArSystem::_initState==ArSystem::AR_UP)
  {
  ArSystem::_initState=ArSystem::AR_GOING_DOWN;
  if(ArSystem::_sched)
    {
    ArSystem::_sched.destroy();
    }
  if(ArSystem::_use3D)
    {
    Scene3D::_destroy();
    URLTexture::_destroy();
    Texture::_destroy();
    Renderer3D::_destroy();
    }
  while(!ArSystem::_plugins->empty())
    {
    ArSystem::PluginInfo & info=ArSystem::_plugins->back();
    if(info.destroyFnct)
      {
      (*info.destroyFnct)();
      }
    ArSystem::closeLibrary(info.library);
    ArSystem::_plugins->pop_back();
    }
  delete ArSystem::_plugins;
  ArSystem::_plugins=(StlVector<ArSystem::PluginInfo> *)0;
  if(ArSystem::_use3D)
    {
    LLW::destroy();
    }
  ArSystem::restoreAllDefaultHandlers();
  delete ArSystem::_readingPoll;
  ArSystem::_readingPoll=(StlVector<int> *)0;
  delete ArSystem::_writingPoll;
  ArSystem::_writingPoll=(StlVector<int> *)0;
  delete ArSystem::_broadcast;
  ArSystem::_broadcast=(StlVector<unsigned int> *)0;
  ArSystem::_httpProxyPort=-1;
  ArSystem::_httpProxyHost.erase();
  ArSystem::_mtx.destroy(); // UGLY !
  ArSystem::_use3D=false;
  delete ArSystem::_commandLine;
  ArSystem::_commandLine=(StlVector<StlString> *)0;
  ArSystem::_initState=ArSystem::AR_DOWN;
  }
}

const StlVector<StlString> & // command line
ArSystem::getCommandLine(void) // Static
{
return(*ArSystem::_commandLine);
}

bool // 3D is used
ArSystem::uses3D(void) // Static
{
return(ArSystem::_use3D);
}

bool // success
ArSystem::loadPlugin(const StlString & pluginName,
                     const StlVector<StlString> * params) // Static
{
if(ArSystem::usesPlugin(pluginName))
  {
  cerr << "ArSystem: Plugin `" << pluginName << "' is already loaded" << endl;
  return(false);
  }
ArSystem::PluginInfo info;
info.library=ArSystem::openLibrary("AReVi"+pluginName,true);
if(info.library)
  {
  info.name=pluginName;
  void * symbol=ArSystem::getLibrarySymbol(info.library,pluginName+"_init");
  memcpy(&info.initFnct,&symbol,sizeof(void *));
  if(info.initFnct)
    {
    symbol=ArSystem::getLibrarySymbol(info.library,pluginName+"_update");
    memcpy(&info.updateFnct,&symbol,sizeof(void *));
    symbol=ArSystem::getLibrarySymbol(info.library,pluginName+"_destroy");
    memcpy(&info.destroyFnct,&symbol,sizeof(void *));
    if((*info.initFnct)(params))
      {
      ArSystem::_plugins->push_back(info);
      ArSystem::_forceSettings();
      return(true);
      }
    else
      {
      cerr << "ArSystem: Cannot initialize plugin `"
           << pluginName << "'" << endl;
      }
    }
  else
    {
    cerr << "ArSystem: Cannot find `" << pluginName
         << "_init()' in plugin `" << pluginName << "'" << endl;
    cerr << "    " << ArSystem::printLibraryError() << endl;
    }
  ArSystem::closeLibrary(info.library);
  }
else
  {
  cerr << "ArSystem: Cannot load plugin `" << pluginName << "'" << endl;
  cerr << "    " << ArSystem::printLibraryError() << endl;
  }
return(false);
}

bool // plugin is loaded
ArSystem::usesPlugin(const StlString & pluginName) // Static
{
for(unsigned int i=ArSystem::_plugins->size();i--;)
  {
  if((*ArSystem::_plugins)[i].name==pluginName)
    {
    return(true);
    }
  }
return(false);
}

ArSystem::InitState // init state
ArSystem::getInitState(void) // Static
{
return(ArSystem::_initState);
}

unsigned int // current AReVi version
ArSystem::getVersion(void) // Static
{
#ifdef AR_VERSION
return(AR_VERSION);
#else // not AR_VERSION
return(0);
#endif // not AR_VERSION
}

//-------- Simulation --------------------------------------------------------

bool // go on
ArSystem::enterSimulation(ArRef<Scheduler> (*simulationInit)(void)) // Static
{
if(ArSystem::_sched)
  {
  ArSystem::_sched.destroy(); // UGLY !
  }
ArSystem::_goOn=true;
ArSystem::_sched=(*simulationInit)();
if(!ArSystem::_sched||ArSystem::_sched->fail())
  {
  ArSystem::_goOn=false;
  }
else
  {
  ArSystem::_sched->_initialize();
  }
return(ArSystem::_goOn);
}

bool // go on
ArSystem::simulationStep(void) // Static
{
if(ArSystem::_use3D)
  {
  if(Renderer3D::_redisplay())
    {
    LLW::flush();
    }
  Renderer3D::_checkEvents(false);
  }
for(unsigned int i=ArSystem::_plugins->size();i--;)
  {
  void (*updateFnct)()=(*ArSystem::_plugins)[i].updateFnct;
  if(updateFnct)
    {
    (*updateFnct)();
    }
  }
if(!ArSystem::_sched->_activate()&&!ArSystem::_use3D)
  {
  ArSystem::_goOn=false;
  }
return(ArSystem::_goOn);
}

void
ArSystem::leaveSimulation(void) // Static
{
ArSystem::_goOn=false;
}

void
ArSystem::simulationLoop(ArRef<Scheduler> (*simulationInit)(void)) // Static
{
if(ArSystem::enterSimulation(simulationInit))
  {
  while(ArSystem::simulationStep());
  }
}

//-------- Error Messages ----------------------------------------------------

StlString // error description message
ArSystem::printError(int errorNumber) // Static
{
StlString message;
switch(errorNumber)
  {                                                         // Common errors
  case E2BIG:            message="E2BIG";            break;
  case EACCES:           message="EACCES";           break;
  case EADDRINUSE:       message="EADDRINUSE";       break;
  case EADDRNOTAVAIL:    message="EADDRNOTAVAIL";    break;
  case EAFNOSUPPORT:     message="EAFNOSUPPORT";     break;
  case EAGAIN:           message="EAGAIN";           break;
  case EALREADY:         message="EALREADY";         break;
  case EBADF:            message="EBADF";            break;
  case EBUSY:            message="EBUSY";            break;
  case ECHILD:           message="ECHILD";           break;
  case ECONNABORTED:     message="ECONNABORTED";     break;
  case ECONNREFUSED:     message="ECONNREFUSED";     break;
  case ECONNRESET:       message="ECONNRESET";       break;
  case EDEADLK:          message="EDEADLK";          break;
  case EDESTADDRREQ:     message="EDESTADDRREQ";     break;
  case EDOM:             message="EDOM";             break;
  case EDQUOT:           message="EDQUOT";           break;
  case EEXIST:           message="EEXIST";           break;
  case EFAULT:           message="EFAULT";           break;
  case EFBIG:            message="EFBIG";            break;
  case EHOSTDOWN:        message="EHOSTDOWN";        break;
  case EHOSTUNREACH:     message="EHOSTUNREACH";     break;
  case EINPROGRESS:      message="EINPROGRESS";      break;
  case EINTR:            message="EINTR";            break;
  case EINVAL:           message="EINVAL";           break;
  case EIO:              message="EIO";              break;
  case EISCONN:          message="EISCONN";          break;
  case EISDIR:           message="EISDIR";           break;
  case ELOOP:            message="ELOOP";            break;
  case EMFILE:           message="EMFILE";           break;
  case EMLINK:           message="EMLINK";           break;
  case EMSGSIZE:         message="EMSGSIZE";         break;
  case ENAMETOOLONG:     message="ENAMETOOLONG";     break;
  case ENETDOWN:         message="ENETDOWN";         break;
  case ENETRESET:        message="ENETRESET";        break;
  case ENETUNREACH:      message="ENETUNREACH";      break;
  case ENFILE:           message="ENFILE";           break;
  case ENOBUFS:          message="ENOBUFS";          break;
  case ENODEV:           message="ENODEV";           break;
  case ENOENT:           message="ENOENT";           break;
  case ENOEXEC:          message="ENOEXEC";          break;
  case ENOLCK:           message="ENOLCK";           break;
  case ENOMEM:           message="ENOMEM";           break;
  case ENOPROTOOPT:      message="ENOPROTOOPT";      break;
  case ENOSPC:           message="ENOSPC";           break;
  case ENOSYS:           message="ENOSYS";           break;
  case ENOTBLK:          message="ENOTBLK";          break;
  case ENOTCONN:         message="ENOTCONN";         break;
  case ENOTDIR:          message="ENOTDIR";          break;
  case ENOTEMPTY:        message="ENOTEMPTY";        break;
  case ENOTSOCK:         message="ENOTSOCK";         break;
  case ENOTTY:           message="ENOTTY";           break;
  case ENXIO:            message="ENXIO";            break;
  case EOPNOTSUPP:       message="EOPNOTSUPP";       break;
  case EPERM:            message="EPERM";            break;
  case EPFNOSUPPORT:     message="EPFNOSUPPORT";     break;
  case EPIPE:            message="EPIPE";            break;
  case EPROTONOSUPPORT:  message="EPROTONOSUPPORT";  break;
  case EPROTOTYPE:       message="EPROTOTYPE";       break;
  case ERANGE:           message="ERANGE";           break;
  case EREMOTE:          message="EREMOTE";          break;
  case EROFS:            message="EROFS";            break;
  case ESHUTDOWN:        message="ESHUTDOWN";        break;
  case ESOCKTNOSUPPORT:  message="ESOCKTNOSUPPORT";  break;
  case ESPIPE:           message="ESPIPE";           break;
  case ESRCH:            message="ESRCH";            break;
  case ESTALE:           message="ESTALE";           break;
  case ETIMEDOUT:        message="ETIMEDOUT";        break;
  case ETOOMANYREFS:     message="ETOOMANYREFS";     break;
  case ETXTBSY:          message="ETXTBSY";          break;
  case EUSERS:           message="EUSERS";           break;
  case EXDEV:            message="EXDEV";            break;
#if !defined(__APPLE__) && !defined(__FreeBSD__)   // Non Apple/FreeBSD errors
  case EADV:             message="EADV";             break;
  case EBADE:            message="EBADE";            break;
  case EBADFD:           message="EBADFD";           break;
  case EBADMSG:          message="EBADMSG";          break;
  case EBADR:            message="EBADR";            break;
  case EBADRQC:          message="EBADRQC";          break;
  case EBADSLT:          message="EBADSLT";          break;
  case EBFONT:           message="EBFONT";           break;
  case ECHRNG:           message="ECHRNG";           break;
  case ECOMM:            message="ECOMM";            break;
  case EIDRM:            message="EIDRM";            break;
  case EL2HLT:           message="EL2HLT";           break;
  case EL2NSYNC:         message="EL2NSYNC";         break;
  case EL3HLT:           message="EL3HLT";           break;
  case EL3RST:           message="EL3RST";           break;
  case ELIBACC:          message="ELIBACC";          break;
  case ELIBBAD:          message="ELIBBAD";          break;
  case ELIBEXEC:         message="ELIBEXEC";         break;
  case ELIBMAX:          message="ELIBMAX";          break;
  case ELIBSCN:          message="ELIBSCN";          break;
  case ELNRNG:           message="ELNRNG";           break;
  case EMULTIHOP:        message="EMULTIHOP";        break;
  case ENOANO:           message="ENOANO";           break;
  case ENOCSI:           message="ENOCSI";           break;
  case ENODATA:          message="ENODATA";          break;
  case ENOLINK:          message="ENOLINK";          break;
  case ENOMSG:           message="ENOMSG";           break;
  case ENONET:           message="ENONET";           break;
  case ENOPKG:           message="ENOPKG";           break;
  case ENOSR:            message="ENOSR";            break;
  case ENOSTR:           message="ENOSTR";           break;
  case ENOTUNIQ:         message="ENOTUNIQ";         break;
  case EPROTO:           message="EPROTO";           break;
  case EREMCHG:          message="EREMCHG";          break;
  case ESRMNT:           message="ESRMNT";           break;
  case ETIME:            message="ETIME";            break;
  case EUNATCH:          message="EUNATCH";          break;
  case EXFULL:           message="EXFULL";           break;
#endif // not __APPLE__ and not __FreeBSD__
#ifdef __linux                                              // Linux errors
  case ECANCELED:        message="ECANCELED";        break;
  case EDOTDOT:          message="EDOTDOT";          break;
  case EILSEQ:           message="EILSEQ";           break;
  case EISNAM:           message="EISNAM";           break;
  case ENAVAIL:          message="ENAVAIL";          break;
  case ENOMEDIUM:        message="ENOMEDIUM";        break;
  case ENOTNAM:          message="ENOTNAM";          break;
  case EOVERFLOW:        message="EOVERFLOW";        break;
  case EREMOTEIO:        message="EREMOTEIO";        break;
  case ERESTART:         message="ERESTART";         break;
  case ESTRPIPE:         message="ESTRPIPE";         break;
  case EUCLEAN:          message="EUCLEAN";          break;
#endif // __linux
#ifdef __CYGWIN__                                           // Cygwin errors
  case EDEADLOCK:        message="EDEADLOCK";        break;
  case EDOTDOT:          message="EDOTDOT";          break;
  case ELBIN:            message="ELBIN";            break;
  case ENMFILE:          message="ENMFILE";          break;
  case ENOMEDIUM:        message="ENOMEDIUM";        break;
  case ENOSHARE:         message="ENOSHARE";         break;
  case ENOTSUP:          message="ENOTSUP";          break;
  case EPROCLIM:         message="EPROCLIM";         break;
#endif // __CYGWIN__
#ifdef __sgi                                                // SGI errors
  case EBUFSIZE:         message="EBUFSIZE";         break;
  case ECANTEXTENT:      message="ECANTEXTENT ";     break;
  case ECLOCKCPU:        message="ECLOCKCPU";        break;
  case ECONTROLLER:      message="ECONTROLLER";      break;
  case EDESTROYED:       message="EDESTROYED";       break;
  case EDIRCORRUPTED:    message="EDIRCORRUPTED";    break;
  case EDISJOINT:        message="EDISJOINT";        break;
  case EEMPTY:           message="EEMPTY";           break;
  case EENDOFMINOR:      message="EENDOFMINOR";      break;
  case EENQUEUED:        message="EENQUEUED";        break;
  case EGROUPLOOP:       message="EGROUPLOOP";       break;
  case EINIT:            message="EINIT";            break;
  case EINVALMODE:       message="EINVALMODE";       break;
  case EINVALSTATE:      message="EINVALSTATE";      break;
  case EINVALTIME:       message="EINVALTIME";       break;
  case EIORESID:         message="EIORESID";         break;
  case EJOINED:          message="EJOINED";          break;
  case ELOGINLIM:        message="ELOGINLIM";        break;
  case EMUSTRUN:         message="EMUSTRUN";         break;
  case ENFSREMOTE:       message="ENFSREMOTE";       break;
  case ENOATTACH:        message="ENOATTACH";        break;
  case ENOATTR:          message="ENOATTR";          break;
  case ENOEXIST:         message="ENOEXIST";         break;
  case ENOINTRGROUP:     message="ENOINTRGROUP";     break;
  case ENOLIMFILE:       message="ENOLIMFILE";       break;
  case ENOLOGIN:         message="ENOLOGIN";         break;
  case ENOPROC:          message="ENOPROC";          break;
  case ENOTENQUEUED:     message="ENOTENQUEUED";     break;
  case ENOTJOINED:       message="ENOTJOINED";       break;
  case ENOTSTOPPED:      message="ENOTSTOPPED";      break;
  case EPROCLIM:         message="EPROCLIM";         break;
  case EREMDEV:          message="EREMDEV";          break;
#endif // __sgi
  default:               message=intToStr(errorNumber);  break;
  }
message+=" - ";
message+=strerror(errorNumber);
return(message);
}

//-------- Random Generator --------------------------------------------------

void
ArSystem::setRandomSeed(unsigned int seed,
                        unsigned long long count) // Static
{
ArSystem::_randomSeed=seed;
ArSystem::_randomCount=count;
srand(seed);
while(count--)
  {
  rand();
  }
}

void
ArSystem::getRandomSeed(unsigned int & seedOut,
                        unsigned long long & countOut) // Static
{
seedOut=ArSystem::_randomSeed;
countOut=ArSystem::_randomCount;
}

unsigned int // random in [0;limit-1]
ArSystem::integerRand(unsigned int limit) // Static
{
++ArSystem::_randomCount;
return(rand()%limit);
}

double // random in [0.0;1.0[
ArSystem::realRand(void) // Static
{
++ArSystem::_randomCount;
return((double)(rand())/((double)(RAND_MAX)+1.0));
}

//-------- Signal Management -------------------------------------------------

const ArSystem::SignalHandler
ArSystem::SIGHANDLER_ERROR=
  reinterpret_cast<ArSystem::SignalHandler>(SIG_ERR),
ArSystem::SIGHANDLER_DEFAULT
  =reinterpret_cast<ArSystem::SignalHandler>(SIG_DFL),
ArSystem::SIGHANDLER_IGNORE
  =reinterpret_cast<ArSystem::SignalHandler>(SIG_IGN);

const ArSystem::SignalNumber
ArSystem::SIGNUM_NONE   = 0,
ArSystem::SIGNUM_HUP    = SIGHUP,
ArSystem::SIGNUM_INT    = SIGINT,
ArSystem::SIGNUM_QUIT   = SIGQUIT,
ArSystem::SIGNUM_ILL    = SIGILL,
ArSystem::SIGNUM_TRAP   = SIGTRAP,
ArSystem::SIGNUM_ABRT   = SIGABRT,
ArSystem::SIGNUM_BUS    = SIGBUS,
ArSystem::SIGNUM_FPE    = SIGFPE,
ArSystem::SIGNUM_KILL   = SIGKILL,
ArSystem::SIGNUM_USR1   = SIGUSR1,
ArSystem::SIGNUM_SEGV   = SIGSEGV,
ArSystem::SIGNUM_USR2   = SIGUSR2,
ArSystem::SIGNUM_PIPE   = SIGPIPE,
ArSystem::SIGNUM_ALRM   = SIGALRM,
ArSystem::SIGNUM_TERM   = SIGTERM,
ArSystem::SIGNUM_CHLD   = SIGCHLD,
#if !defined(__APPLE__) && !defined(__FreeBSD__)
ArSystem::SIGNUM_CLD    = SIGCLD,
#else // __APPLE__ or __FreeBSD__
ArSystem::SIGNUM_CLD    = SIGCHLD, // synonym of SIGCLD
#endif // __APPLE__ or __FreeBSD__
ArSystem::SIGNUM_CONT   = SIGCONT,
ArSystem::SIGNUM_STOP   = SIGSTOP,
ArSystem::SIGNUM_TSTP   = SIGTSTP,
ArSystem::SIGNUM_TTIN   = SIGTTIN,
ArSystem::SIGNUM_TTOU   = SIGTTOU,
ArSystem::SIGNUM_URG    = SIGURG,
ArSystem::SIGNUM_XCPU   = SIGXCPU,
ArSystem::SIGNUM_XFSZ   = SIGXFSZ,
ArSystem::SIGNUM_VTALRM = SIGVTALRM,
ArSystem::SIGNUM_PROF   = SIGPROF,
ArSystem::SIGNUM_WINCH  = SIGWINCH,
#if !defined(__APPLE__) && !defined(__FreeBSD__)
ArSystem::SIGNUM_POLL   = SIGPOLL,
#else // __APPLE__ or __FreeBSD__
ArSystem::SIGNUM_POLL   = SIGIO, // synonym of SIGPOLL
#endif // __APPLE__ or __FreeBSD__
ArSystem::SIGNUM_IO     = SIGIO,
ArSystem::SIGNUM_SYS    = SIGSYS;

StlString // signal description message
ArSystem::printSignal(ArSystem::SignalNumber signum) // Static
{
if(signum==ArSystem::SIGNUM_HUP)
  {
  return("HUP - Hangup");
  }
if(signum==ArSystem::SIGNUM_INT)
  {
  return("INT - Interrupt");
  }
if(signum==ArSystem::SIGNUM_QUIT)
  {
  return("QUIT - Quit");
  }
if(signum==ArSystem::SIGNUM_ILL)
  {
  return("ILL - Illegal instruction");
  }
if(signum==ArSystem::SIGNUM_TRAP)
  {
  return("TRAP - Trace/breakpoint trap");
  }
if(signum==ArSystem::SIGNUM_ABRT)
  {
  return("ABRT - Abort");
  }
if(signum==ArSystem::SIGNUM_BUS)
  {
  return("BUS - Bus error");
  }
if(signum==ArSystem::SIGNUM_FPE)
  {
  return("FPE - Arithmetic exception");
  }
if(signum==ArSystem::SIGNUM_KILL)
  {
  return("KILL - Killed");
  }
if(signum==ArSystem::SIGNUM_USR1)
  {
  return("USR1 - User defined signal 1");
  }
if(signum==ArSystem::SIGNUM_SEGV)
  {
  return("SEGV - Segmentation violation");
  }
if(signum==ArSystem::SIGNUM_USR2)
  {
  return("USR2 - User defined signal 2");
  }
if(signum==ArSystem::SIGNUM_PIPE)
  {
  return("PIPE - Broken pipe");
  }
if(signum==ArSystem::SIGNUM_ALRM)
  {
  return("ALRM - Alarm clock");
  }
if(signum==ArSystem::SIGNUM_TERM)
  {
  return("TERM - Terminated");
  }
if(signum==ArSystem::SIGNUM_CHLD)
  {
  return("CHLD - Child status changed");
  }
if(signum==ArSystem::SIGNUM_CLD)
  {
  return("CLD - Child status changed");
  }
if(signum==ArSystem::SIGNUM_CONT)
  {
  return("CONT - Continue");
  }
if(signum==ArSystem::SIGNUM_STOP)
  {
  return("STOP - Stopped");
  }
if(signum==ArSystem::SIGNUM_TSTP)
  {
  return("TSTP - Stopped from tty");
  }
if(signum==ArSystem::SIGNUM_TTIN)
  {
  return("TTIN - Stopped, background read from tty");
  }
if(signum==ArSystem::SIGNUM_TTOU)
  {
  return("TTOU - Stopped, background write to tty");
  }
if(signum==ArSystem::SIGNUM_URG)
  {
  return("URG - Urgent I/O condition");
  }
if(signum==ArSystem::SIGNUM_XCPU)
  {
  return("XCPU - CPU time limit exceeded");
  }
if(signum==ArSystem::SIGNUM_XFSZ)
  {
  return("XFSZ - File size limit exceeded");
  }
if(signum==ArSystem::SIGNUM_VTALRM)
  {
  return("VTALRM - Virtual alarm clock");
  }
if(signum==ArSystem::SIGNUM_PROF)
  {
  return("PROF - Profiling alarm clock");
  }
if(signum==ArSystem::SIGNUM_WINCH)
  {
  return("WINCH - Window size change");
  }
if(signum==ArSystem::SIGNUM_POLL)
  {
  return("POLL - Pollable event");
  }
if(signum==ArSystem::SIGNUM_IO)
  {
  return("IO - I/O now possible");
  }
if(signum==ArSystem::SIGNUM_SYS)
  {
  return("SYS - Bad system call");
  }
return("Unknown signal "+intToStr(signum));
}

ArSystem::ProcessID // process identifier
ArSystem::processID(void) // Static
{
return(ArSystem::_pid);
}

ArSystem::ProcessID // parent identifier
ArSystem::parentID(void) // Static
{
return(ArSystem::_ppid);
}

bool // process reached
ArSystem::sendSignal(ArSystem::ProcessID processID,
                     ArSystem::SignalNumber signum) // Static
{
return(!kill(processID,signum));
}

ArSystem::SignalHandler // old handler
ArSystem::installSignalHandler(ArSystem::SignalNumber signum,
                               ArSystem::SignalHandler handler) // Static
{
struct sigaction act,old;
sigemptyset(&act.sa_mask);
act.sa_handler=reinterpret_cast<void (*)(int)>(handler);
act.sa_flags=0;
int r;
do
  {
  r=sigaction(signum,&act,&old);
  } while((r==-1)&&(errno==EINTR));
if(r!=-1) return(reinterpret_cast<ArSystem::SignalHandler>(old.sa_handler));
else return(ArSystem::SIGHANDLER_ERROR);
}

void
ArSystem::ignoreAllSignals(void) // Static
{
for(int signum=1;signum<NSIG;signum++)
  {
  ArSystem::installSignalHandler(signum,ArSystem::SIGHANDLER_IGNORE);
  }
}

void
ArSystem::restoreAllDefaultHandlers(void) // Static
{
for(int signum=1;signum<NSIG;signum++)
  {
  if(signum!=ArSystem::SIGNUM_PROF)
    {
    ArSystem::installSignalHandler(signum,ArSystem::SIGHANDLER_DEFAULT);
    }
  }
}

//-------- Time Management ---------------------------------------------------

void
ArSystem::yield(void) // Static
{
sched_yield();
}

void
ArSystem::sleep(unsigned int millisecond) // Static
{
#ifdef __CYGWIN__
usleep(millisecond*1000);
#else // not __GYGWIN__
struct timespec ts;
ts.tv_sec=static_cast<time_t>(millisecond/1000);
ts.tv_nsec=static_cast<int>(1000000*(millisecond%1000));
int r;
do
  {
  r=nanosleep(&ts,&ts);
  } while((r==-1)&&(errno==EINTR));
#endif // not __GYGWIN__
}

unsigned int // time in millisecond
ArSystem::getTime(void) // Static
{
struct timeval tv;
gettimeofday(&tv,(struct timezone *)0);
return(tv.tv_sec*1000+tv.tv_usec/1000);
}

void
ArSystem::setRealTimer(unsigned int millisecond,
                       bool repeat) // Static
{
ArSystem::_setTimer(static_cast<int>(ITIMER_REAL),millisecond,repeat);
}

#ifdef __CYGWIN__

void
ArSystem::setVirtualTimer(unsigned int /* millisecond */,
                          bool /* repeat */) // Static
{
cerr << "ArSystem: No virtual timer available on this platform !" << endl;
cerr << "          (try to use a real timer instead)" << endl;
}

void
ArSystem::setProfilingTimer(unsigned int /* millisecond */,
                            bool /* repeat */) // Static
{
cerr << "ArSystem: No profiling timer available on this platform !" << endl;
cerr << "          (try to use a real timer instead)" << endl;
}

#else // not __CYGWIN__

void
ArSystem::setVirtualTimer(unsigned int millisecond,
                          bool repeat) // Static
{
ArSystem::_setTimer(static_cast<int>(ITIMER_VIRTUAL),millisecond,repeat);
}

void
ArSystem::setProfilingTimer(unsigned int millisecond,
                            bool repeat) // Static
{
ArSystem::_setTimer(static_cast<int>(ITIMER_PROF),millisecond,repeat);
}

#endif // not __CYGWIN__

//-------- Library Loading ---------------------------------------------------

void * // !=0: library handle  0: error
ArSystem::openLibrary(const StlString & libName,
                      bool convertLibName) // Static
{
int openMode=RTLD_LAZY;
if(libName.size())
  {
  if(convertLibName)
    {
#if !defined(__CYGWIN__) && !defined(__APPLE__)
    StlString name="lib"+libName+".so";
#endif // not __CYGWIN__ and not __APPLE__
#ifdef __CYGWIN__
    StlString name=libName+".dll";
#endif // __CYGWIN__
#ifdef __APPLE__
    StlString name="lib"+libName+".dylib";
#endif // __APPLE__
    return(dlopen(name.c_str(),openMode));
    }
  else
    {
    return(dlopen(libName.c_str(),openMode));
    }
  }
else
  {
  return(dlopen((const char *)0,openMode));
  }
}

void * // !=0: symbol  0: error
ArSystem::getLibrarySymbol(void * library,
                           const StlString & symbolName) // Static
{
void * symbol=dlsym(library,symbolName.c_str());
if(!symbol)
  {
  StlString newName="_"; // under some platforms, symbol names
  newName+=symbolName;   // are added a leading underscore
  symbol=dlsym(library,newName.c_str());
  }
return(symbol);
}

StlString // library error description message
ArSystem::printLibraryError(void) // Static
{
StlString msg;
char * err=dlerror();
if(err)
  {
  msg=err;
#ifdef __CYGWIN__
  if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                   (LPCVOID)0,GetLastError(),
                   MAKELANGID(LANG_NEUTRAL,SUBLANG_NEUTRAL),
                   (LPTSTR)&err,0,(va_list *)0))
    {
    msg+=": ";
    msg+=err;
    LocalFree(err);
    }
#endif // __CYGWIN__
  }
return(msg);
}

void
ArSystem::closeLibrary(void * library) // Static
{
dlclose(library);
}

//-------- File Management ---------------------------------------------------

int // >=0: file descriptor  -1: error
ArSystem::openInFile(const StlString & fileName) // Static
{
int fd;
do
  {
  fd=open(fileName.c_str(),O_RDONLY);
  } while((fd==-1)&&(errno==EINTR));
return(fd);
}

int // >=0: file descriptor  -1: error
ArSystem::openOutFile(const StlString & fileName,
                      bool append) // Static
{
int flags=O_CREAT|O_WRONLY;
if(append) flags|=O_APPEND;
else flags|=O_TRUNC;
int fd;
do
  {
  fd=open(fileName.c_str(),flags,0666);
  } while((fd==-1)&&(errno==EINTR));
return(fd);
}

int // >=0: file descriptor  -1: error
ArSystem::openInOutFile(const StlString & fileName) // Static
{
int fd;
do
  {
  fd=open(fileName.c_str(),O_RDWR|O_CREAT,0666);
  } while((fd==-1)&&(errno==EINTR));
return(fd);
}

int // >=0: file descriptor  -1: error
ArSystem::createTmpFile(const StlString & prefix,
                        StlString & fileNameOut) // Static
{
fileNameOut=prefix+"XXXXXX";
int fd;
do
  {
  fd=mkstemp(const_cast<char *>(fileNameOut.c_str()));
  } while((fd==-1)&&(errno==EINTR));
return(fd);
}

bool // success
ArSystem::getFileKind(const StlString & fileName,
                      StlString & kindOut) // Static
{
struct stat infos;
int r;
do
  {
  r=stat(fileName.c_str(),&infos);
  } while((r==-1)&&(errno==EINTR));
if(r!=-1)
  {
  if(S_ISREG(infos.st_mode))
    {
    kindOut="REG";
    }
  else if(S_ISDIR(infos.st_mode))
    {
    kindOut="DIR";
    }
  else if(S_ISCHR(infos.st_mode))
    {
    kindOut="CHR";
    }
  else if(S_ISBLK(infos.st_mode))
    {
    kindOut="BLK";
    }
  else if(S_ISFIFO(infos.st_mode))
    {
    kindOut="FIFO";
    }
  else if(S_ISSOCK(infos.st_mode))
    {
    kindOut="SOCK";
    }
  else
    {
    kindOut="?";
    }
  return(true);
  }
else
  {
  kindOut="?";
  return(false);
  }
}

bool // success
ArSystem::removeFile(const StlString & fileName) // Static
{
int r;
do
  {
  r=unlink(fileName.c_str());
  } while((r==-1)&&(errno==EINTR));
return(!r);
}

//-------- Directory Management ----------------------------------------------

bool // success
ArSystem::getDirContent(const StlString & dirName,
                        StlVector<StlString> & contentOut) // Static
{
contentOut.clear();
DIR * dir;
do
  {
  dir=opendir(dirName.c_str());
  } while(!dir&&(errno==EINTR));
if(dir)
  {
  struct dirent * entry;
  do
    {
    entry=readdir(dir);
    if(entry&&strcmp(entry->d_name,".")&&strcmp(entry->d_name,".."))
      {
      contentOut.push_back(entry->d_name);
      }
    } while(entry);
  closedir(dir);
  return(true);
  }
else
  {
  return(false);
  }
}

int // >=0: file descriptor  -1: error
ArSystem::createDir(const StlString & dirName) // Static
{
int fd;
do
  {
  fd=mkdir(dirName.c_str(),0755);
  } while((fd==-1)&&(errno==EINTR));
return(fd);
}

int // >=0: file descriptor  -1: error
ArSystem::createTmpDir(const StlString & prefix,
                       StlString & dirNameOut) // Static
{
int fd=ArSystem::createTmpFile(prefix,dirNameOut);
if(fd>=0)
  {
  ArSystem::removeFile(dirNameOut);
  ArSystem::closeFD(fd);
  fd=ArSystem::createDir(dirNameOut);
  }
return(fd);
}

bool // success
ArSystem::removeDir(const StlString & dirName) // Static
{
int r;
do
  {
  r=rmdir(dirName.c_str());
  } while((r==-1)&&(errno==EINTR));
return(!r);
}

//-------- IP Services -------------------------------------------------------

StlString // local host name
ArSystem::getHostName(void) // Static
{
char buffer[0x100];
int r;
do
  {
  r=gethostname(buffer,0x100);
  } while((r==-1)&&(errno==EINTR));
if(r!=-1) return(buffer);
else
  {
  StlString result;
  return(result);
  }
}

unsigned int // >0: IP address  0: error
ArSystem::getIPAddress(const StlString & hostName) // Static
{
ArSystem::_mtx->lock();
struct hostent * host=gethostbyname(hostName.c_str()); // static alloc !
unsigned int ipAddr=(host ? *((unsigned int *)(host->h_addr)) : 0);
ArSystem::_mtx->unlock();
ArSystem::inOut32(&ipAddr); // N to H
return(ipAddr);
}

void
ArSystem::setHTTPProxy(const StlString & proxyHost,
                       int proxyPort) // Static
{
ArSystem::_httpProxyHost=proxyHost;
ArSystem::_httpProxyPort=proxyPort;
}

void
ArSystem::getHTTPProxy(StlString & proxyHostOut,
                       int & proxyPortOut) // Static
{
proxyHostOut=ArSystem::_httpProxyHost;
proxyPortOut=ArSystem::_httpProxyPort;
}

bool // success
ArSystem::decodeURL(const StlString & url,
                    StlString & protocolOut,
                    StlString & hostOut,
                    int & portOut,
                    StlString & fileOut) // Static
{
bool fail=false;
const char * urlStr=url.c_str();
if(!strncmp(urlStr,"http:",5))
  {
  protocolOut="http";
  const char * ptr1=urlStr+5;
  while(*ptr1=='/')
    {
    ptr1++;
    }
  const char * ptr2=ptr1;
  while(*ptr2&&(*ptr2!=':')&&(*ptr2!='/'))
    {
    ptr2++;
    }
  hostOut=StlString(ptr1,ptr2-ptr1);
  portOut=80;
  const char * remaining=ptr2;
  if(*ptr2==':')
    {
    ptr2++;
    const char * ptr3=ptr2;
    while(*ptr3&&(*ptr3>='0')&&(*ptr3<='9'))
      {
      ptr3++;
      }
    const char * ptr4=ptr3;
    while(*ptr4&&(*ptr4!='/'))
      {
      ptr4++;
      }
    if((ptr3!=ptr2)&&(ptr3==ptr4))
      {
      sscanf(ptr2,"%d",&portOut);
      }
    else
      {
      fail=true;
      }
    remaining=ptr4;
    }
  if(*remaining)
    {
    fileOut=remaining;
    }
  else
    {
    fileOut="/";
    }
  }
else if(!strncmp(urlStr,"file:",5))
  {
  protocolOut="file";
  hostOut.clear();
  portOut=-1;
  const char * ptr=urlStr+5;
  if(*ptr=='/')
    {
    while(ptr[1]=='/')
      {
      ptr++;
      }
    }
  if(*ptr)
    {
    fileOut=ptr;
    }
  else
    {
    fileOut.clear();
    fail=true;
    }
  }
else // assume url is a local file
  {
  protocolOut="file";
  hostOut.clear();
  portOut=-1;
  fileOut=url;
  }
return(!fail);
}

//-------- Socket Management -------------------------------------------------

int // >=0: file descriptor  -1: error
ArSystem::connectTCP(unsigned int ipAddress,
                     int portNumber) // Static
{
int fd=socket(PF_INET,SOCK_STREAM,0);
if(fd==-1)
  {
  return(-1);
  }
struct sockaddr_in addr;
socklen_t len=sizeof(struct sockaddr_in);
memset(&addr,0,len);
addr.sin_family=AF_INET;
ArSystem::inOut16(&portNumber); // H to N
addr.sin_port=portNumber;
ArSystem::inOut32(&ipAddress); // H to N
addr.sin_addr.s_addr=ipAddress;
if(connect(fd,reinterpret_cast<const struct sockaddr *>(&addr),len)==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
return(fd);
}

int // >=0: file descriptor  -1: error
ArSystem::listenTCP(int portNumber) // Static
{
if(portNumber<0) portNumber=0;
int fd=socket(PF_INET,SOCK_STREAM,0);
if(fd==-1)
  {
  return(-1);
  }
int on=1;
if(setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(int))==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
struct sockaddr_in addr;
socklen_t addrLength=sizeof(struct sockaddr_in);
memset(&addr,0,addrLength);
ArSystem::inOut16(&portNumber); // H to N
addr.sin_port=portNumber;
unsigned int ipAddress=INADDR_ANY;
ArSystem::inOut32(&ipAddress); // H to N
addr.sin_addr.s_addr=ipAddress;
addr.sin_family=AF_INET;
if(bind(fd,(const struct sockaddr *)&addr,addrLength)==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
if(listen(fd,10)==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
return(fd);
}

int // >=0: file descriptor  -1: error  -2: non-blocking
ArSystem::acceptTCP(int listenSocket,
                    unsigned int & ipAddressOut,
                    int & portNumberOut) // Static
{
struct sockaddr_in addr;
socklen_t addrLength=sizeof(struct sockaddr_in);
int fd;
do
  {
  fd=accept(listenSocket,
            reinterpret_cast<struct sockaddr *>(&addr),&addrLength);
  } while((fd==-1)&&(errno==EINTR));
if(fd>=0)
  {
  ipAddressOut=addr.sin_addr.s_addr;
  ArSystem::inOut32(&ipAddressOut); // N to H
  portNumberOut=addr.sin_port;
  ArSystem::inOut16(&portNumberOut); // N to H
  return(fd);
  }
else return(errno==EAGAIN ? -2 : -1);
}

bool // success
ArSystem::getSocketAddress(int sock,
                           unsigned int & localIPAddressOut,
                           int & localPortNumberOut) // Static
{
struct sockaddr_in addr;
socklen_t addrLength=sizeof(struct sockaddr_in);
if(getsockname(sock,
               reinterpret_cast<struct sockaddr *>(&addr),&addrLength)==-1)
  {
  return(false);
  }
localIPAddressOut=addr.sin_addr.s_addr;
ArSystem::inOut32(&localIPAddressOut); // N to H
localPortNumberOut=addr.sin_port;
ArSystem::inOut16(&localPortNumberOut); // N to H
if(!localIPAddressOut)
  {
  localIPAddressOut=ArSystem::getIPAddress(ArSystem::getHostName());
  }
return(true);
}

int // >=0: file descriptor  -1: error
ArSystem::openUDP(int portNumber) // Static
{
int fd=socket(PF_INET,SOCK_DGRAM,0);
if(fd==-1)
  {
  return(-1);
  }
int on=1;
if(setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(int))==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
struct sockaddr_in addr;
socklen_t addrLength=sizeof(struct sockaddr_in);
memset(&addr,0,addrLength);
addr.sin_family=AF_INET;
ArSystem::inOut16(&portNumber); // H to N
addr.sin_port=portNumber;
unsigned int ipAddress=INADDR_ANY;
ArSystem::inOut32(&ipAddress); // H to N
addr.sin_addr.s_addr=ipAddress;
if(bind(fd,(const struct sockaddr *)&addr,addrLength)==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
on=1;
if(setsockopt(fd,SOL_SOCKET,SO_BROADCAST,&on,sizeof(int))==-1)
  {
  ArSystem::closeFD(fd);
  return(-1);
  }
return(fd);
}


const StlVector<unsigned int> & // UDP broadcast addresses
ArSystem::getUDPBroadcastAddresses(void) // Static
{
if(!ArSystem::_broadcast->size()) // first call ?
  {
#ifndef __CYGWIN__
  int sock=socket(PF_INET,SOCK_DGRAM,0);
  if(sock!=-1)
    {
    char buf[BUFSIZ];
    struct ifconf ifc;
    ifc.ifc_len=BUFSIZ;
    ifc.ifc_buf=buf;
    if(ioctl(sock,SIOCGIFCONF,&ifc)!=-1)
      {
      struct ifreq * ifr=ifc.ifc_req;
      for(int n=ifc.ifc_len/sizeof(struct ifreq);n--;ifr++)
        {
        if((ifr->ifr_addr.sa_family==AF_INET)&&
           (ioctl(sock,SIOCGIFFLAGS,ifr)!=-1)&&
           (ifr->ifr_flags&IFF_UP))
          {
          const struct sockaddr_in * addr;
          if(ifr->ifr_flags&IFF_POINTOPOINT)
            {
            if(ioctl(sock,SIOCGIFDSTADDR,ifr)!=-1)
              {
              addr=reinterpret_cast<const struct sockaddr_in *>
                                   (&ifr->ifr_dstaddr);
              ArSystem::_broadcast->push_back(addr->sin_addr.s_addr);
              ArSystem::inOut32(&ArSystem::_broadcast->back()); // N to H
              }
            }
          else if(ifr->ifr_flags&IFF_BROADCAST)
            {
            if(ioctl(sock,SIOCGIFBRDADDR,ifr)!=-1)
              {
              addr=reinterpret_cast<const struct sockaddr_in *>
                                   (&ifr->ifr_broadaddr);
              ArSystem::_broadcast->push_back(addr->sin_addr.s_addr);
              ArSystem::inOut32(&ArSystem::_broadcast->back()); // N to H
              }
            }
          }
        }
      }
    ArSystem::closeFD(sock);
    }
#endif // not __CYGWIN__
  if(!ArSystem::_broadcast->size()) // no broadcast addresse found ?
    {
    ArSystem::_broadcast->push_back(0xFFFFFFFF); // try the default one ...
    }
  }
return(*ArSystem::_broadcast);
}

int // >0: nb received  0: EOF  -1: error  -2: non-blocking
ArSystem::receiveUDP(int sock,
                     void * buffer,
                     unsigned int size,
                     unsigned int & ipAddressOut,
                     int & portNumberOut) // Static
{
struct sockaddr_in addr;
socklen_t addrLength=sizeof(struct sockaddr_in);
int r;
do
  {
  r=recvfrom(sock,buffer,size,0,
             reinterpret_cast<struct sockaddr *>(&addr),&addrLength);
  } while((r==-1)&&(errno==EINTR));
switch(r)
  {
  case -1:
    {
    return(errno==EAGAIN ? -2 : -1);
    }
  case 0:
    {
    return(0);
    }
  default:
    {
    ipAddressOut=addr.sin_addr.s_addr;
    ArSystem::inOut32(&ipAddressOut); // N to H
    portNumberOut=addr.sin_port;
    ArSystem::inOut16(&portNumberOut); // N to H
    return(r);
    }
  }
}

int // >=0: nb written  -1: error
ArSystem::sendUDP(int sock,
                  const void * buffer,
                  unsigned int size,
                  unsigned int ipAddress,
                  int portNumber) // Static
{
struct sockaddr_in addr;
socklen_t addrLength=sizeof(struct sockaddr_in);
memset(&addr,0,addrLength);
addr.sin_family=AF_INET;
ArSystem::inOut32(&ipAddress); // H to N
addr.sin_addr.s_addr=ipAddress;
ArSystem::inOut16(&portNumber); // H to N
addr.sin_port=portNumber;
int r;
const char * ptr=static_cast<const char *>(buffer);
unsigned int remaining=size;
while(remaining)
  {
  r=sendto(sock,ptr,remaining,0,
           reinterpret_cast<const struct sockaddr *>(&addr),addrLength);
  if(r==-1)
    {
    if((errno!=EINTR)&&(errno!=EAGAIN)) return(-1);
    }
  else
    {
    remaining-=r;
    ptr+=r;
    }
  }
return(static_cast<int>(size-remaining));
}

//-------- File Descriptor Management ----------------------------------------

bool // success
ArSystem::closeFD(int fileDescriptor) // Static
{
int r;
do
  {
  r=close(fileDescriptor);
  } while((r==-1)&&(errno==EINTR));
unsigned int i;
for(i=ArSystem::_readingPoll->size();i--;)
  {
  if((*ArSystem::_readingPoll)[i]==fileDescriptor)
    {
    StlVectorFastErase(*ArSystem::_readingPoll,i);
    }
  }
for(i=ArSystem::_writingPoll->size();i--;)
  {
  if((*ArSystem::_writingPoll)[i]==fileDescriptor)
    {
    StlVectorFastErase(*ArSystem::_writingPoll,i);
    }
  }
return(!r);
}

bool // success
ArSystem::syncFD(int fileDescriptor) // Static
{
int r;
do
  {
  r=fdatasync(fileDescriptor);
  } while((r==-1)&&((errno==EINTR)||(errno==EAGAIN)));
return(!r);
}

bool // success
ArSystem::setBlockingFD(int fileDescriptor,
                        bool blocking) // Static
{
int r;
do
  {
  r=fcntl(fileDescriptor,F_GETFL);
  } while((r==-1)&&(r==EINTR));
if(r==-1)
  {
  return(false);
  }
int flags=(blocking ? r & ~O_NONBLOCK : r | O_NONBLOCK);
do
  {
  r=fcntl(fileDescriptor,F_SETFL,flags);
  } while((r==-1)&&(r==EINTR));
if(r==-1)
  {
  return(false);
  }
return(true);
}

bool // blocking
ArSystem::isBlockingFD(int fileDescriptor) // Static
{
int r;
do
  {
  r=fcntl(fileDescriptor,F_GETFL);
  } while((r==-1)&&(r==EINTR));
if(r==-1)
  {
  return(false);
  }
return(!(r&O_NONBLOCK));
}

int // >0: nb read  0: EOF  -1: error  -2: non-blocking
ArSystem::partialReadFD(int fileDescriptor,
                        void * buffer,
                        unsigned int size) // Static
{
int r;
do
  {
  r=read(fileDescriptor,buffer,size);
  } while((r==-1)&&(errno==EINTR));
if(r>=0)
  {
  return(r);
  }
else
  {
  return(errno==EAGAIN ? -2 : -1);
  }
}

int // >0: nb read  0: EOF  -1: error
ArSystem::readFD(int fileDescriptor,
                 void * buffer,
                 unsigned int size) // Static
{
int r;
char * ptr=static_cast<char *>(buffer);
unsigned int remaining=size;
while(remaining)
  {
  r=read(fileDescriptor,ptr,remaining);
  switch(r)
    {
    case 0:
      {
      return(0);
      }
    case -1:
      {
      if((errno!=EINTR)&&(errno!=EAGAIN)) return(-1);
      break;
      }
    default:
      {
      remaining-=r;
      ptr+=r;
      }
    }
  }
return(static_cast<int>(size-remaining));
}

int // >=0: nb written  -1: error  -2: non-blocking
ArSystem::partialWriteFD(int fileDescriptor,
                         const void * buffer,
                         unsigned int size) // Static
{
int r;
do
  {
  r=write(fileDescriptor,buffer,size);
  } while((r==-1)&&(errno==EINTR));
if(r>=0)
  {
  return(r);
  }
else
  {
  return(errno==EAGAIN ? -2 : -1);
  }
}

int // >=0: nb written  -1: error
ArSystem::writeFD(int fileDescriptor,
                  const void * buffer,
                  unsigned int size) // Static
{
const char * ptr=static_cast<const char *>(buffer);
unsigned int remaining=size;
while(remaining)
  {
  int r=write(fileDescriptor,ptr,remaining);
  if(r==-1)
    {
    if((errno!=EINTR)&&(errno!=EAGAIN)) return(-1);
    }
  else
    {
    remaining-=r;
    ptr+=r;
    }
  }
return(static_cast<int>(size-remaining));
}

//-------- Polling Management ------------------------------------------------

void
ArSystem::addReadFDPolling(int fileDescriptor) // static
{
if(fileDescriptor!=-1)
  {
  ArSystem::_readingPoll->push_back(fileDescriptor);
  }
}

void
ArSystem::removeReadFDPolling(int fileDescriptor)
{
if(fileDescriptor!=-1)
  {
  for(unsigned int i=ArSystem::_readingPoll->size();i--;)
    {
    if((*ArSystem::_readingPoll)[i]==fileDescriptor)
      {
      StlVectorFastErase(*ArSystem::_readingPoll,i);
      break;
      }
    }
  }
}

bool // fileDescriptor is polled for reading
ArSystem::hasReadFDPolling(int fileDescriptor)
{
if(fileDescriptor!=-1)
  {
  for(unsigned int i=ArSystem::_readingPoll->size();i--;)
    {
    if((*ArSystem::_readingPoll)[i]==fileDescriptor)
      {
      return(true);
      }
    }
  }
return(false);
}

void
ArSystem::addWriteFDPolling(int fileDescriptor) // static
{
if(fileDescriptor!=-1)
  {
  ArSystem::_writingPoll->push_back(fileDescriptor);
  }
}

void
ArSystem::removeWriteFDPolling(int fileDescriptor)
{
if(fileDescriptor!=-1)
  {
  for(unsigned int i=ArSystem::_writingPoll->size();i--;)
    {
    if((*ArSystem::_writingPoll)[i]==fileDescriptor)
      {
      StlVectorFastErase(*ArSystem::_writingPoll,i);
      break;
      }
    }
  }
}

bool // fileDescriptor is polled for writing
ArSystem::hasWriteFDPolling(int fileDescriptor)
{
if(fileDescriptor!=-1)
  {
  for(unsigned int i=ArSystem::_writingPoll->size();i--;)
    {
    if((*ArSystem::_writingPoll)[i]==fileDescriptor)
      {
      return(true);
      }
    }
  }
return(false);
}

int // >=0: nb ready  -1: error
ArSystem::waitForPolledFD(StlVector<int> & readingFDOut,
                          StlVector<int> & writingFDOut,
                          unsigned int millisecond) // Static
{
readingFDOut.clear();
writingFDOut.clear();
fd_set readSet,writeSet;
int maxfd=-1;
FD_ZERO(&readSet);
FD_ZERO(&writeSet);
unsigned int i;
for(i=ArSystem::_readingPoll->size();i--;)
  {
  int fd=(*ArSystem::_readingPoll)[i];
  FD_SET(fd,&readSet);
  if(fd>maxfd)
    {
    maxfd=fd;
    }
  }
for(i=ArSystem::_writingPoll->size();i--;)
  {
  int fd=(*ArSystem::_writingPoll)[i];
  FD_SET(fd,&writeSet);
  if(fd>maxfd)
    {
    maxfd=fd;
    }
  }
struct timeval tv;
tv.tv_sec=static_cast<time_t>(millisecond/1000);
tv.tv_usec=static_cast<int>(1000*(millisecond%1000));
int r=select(maxfd+1,&readSet,&writeSet,(fd_set *)0,&tv);
if((r==-1)&&(errno==EINTR))
  {
  r=0;
  }
if(r>0)
  {
  for(i=ArSystem::_readingPoll->size();i--;)
    {
    int fd=(*ArSystem::_readingPoll)[i];
    if(FD_ISSET(fd,&readSet))
      {
      readingFDOut.push_back(fd);
      }
    }
  for(i=ArSystem::_writingPoll->size();i--;)
    {
    int fd=(*ArSystem::_writingPoll)[i];
    if(FD_ISSET(fd,&writeSet))
      {
      writingFDOut.push_back(fd);
      }
    }
  }
return(r);
}

//-------- Forbidden operations ----------------------------------------------

void
ArSystem::_forceSettings(void) // Static, Private
{
//---- localisation ----
setlocale(LC_ALL,""); // allow user defined localization
setlocale(LC_NUMERIC,"C"); // ensure standard numeric printf/scanf
//---- Install exit handlers ----
ArSystem::installSignalHandler(ArSystem::SIGNUM_HUP,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_INT,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_QUIT,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_ABRT,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_BUS,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_FPE,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_PIPE,
                               ArSystem::_exitHandler);
ArSystem::installSignalHandler(ArSystem::SIGNUM_TERM,
                               ArSystem::_exitHandler);
if(!JavaObject::inJavaApp()) // FIXME: UGLY Java hack !
  {
  ArSystem::installSignalHandler(ArSystem::SIGNUM_ILL,
                                 ArSystem::_exitHandler);
  ArSystem::installSignalHandler(ArSystem::SIGNUM_SEGV,
                                 ArSystem::_exitHandler);
  }
}

void
ArSystem::_exitHandler(ArSystem::SignalNumber signum) // Static, Private
{
StlString description=ArSystem::printSignal(signum);
cerr << "ArSystem: Process " << getpid()
     << " exiting on signal " << signum
     << " (" << description << ")" << endl;
ArSystem::sendSignal(ArSystem::processID(),ArSystem::SIGNUM_KILL);
}

void
ArSystem::_setTimer(int which,
                    unsigned int millisec,
                    bool repeat) // Static, Private
{
struct itimerval itv;
itv.it_value.tv_sec=(int)(millisec/1000);
itv.it_value.tv_usec=(int)(1000*(millisec%1000));
if(repeat)
  {
  itv.it_interval.tv_sec=itv.it_value.tv_sec;
  itv.it_interval.tv_usec=itv.it_value.tv_usec;
  }
else
  {
  itv.it_interval.tv_sec=itv.it_interval.tv_usec=0;
  }
setitimer((__itimer_which)which,&itv,(struct itimerval *)0);
}

} // namespace AReVi

//----------------------------------------------------------------------------
