// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/message.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/message.h"

#include "AReVi/Utils/tcpUdp.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/memoryStream.h"
#include "AReVi/arSystem.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class Message
//----------------------------------------------------------------------------

AR_CLASS_DEF(Message,ArObject)

//-------- Construction / Destruction ----------------------------------------

Message::Message(ArCW & arCW,
                 ArRef<ArObject> emitter) // Protected
: ArObject(arCW),
  _emitter(emitter->thisPtr()),
  _emitterAlias(_emitter ? _emitter->getAlias() : ""),
  _service(),
  _text()
{
  // Nothing to be done
}

Message::Message(ArCW & arCW) // Protected
: ArObject(arCW),
  _emitter(),
  _emitterAlias(),
  _service(),
  _text()
{
  // Nothing to be done
}

Message::~Message(void) // Protected
{
  // Nothing to be done
}

//-------- Identification ----------------------------------------------------

void
Message::setEmitter(ArRef<ArObject> emitter)
{
_emitter=emitter->thisPtr();
_emitterAlias=(_emitter ? _emitter->getAlias() : "");
}

ArRef<ArObject> // reference on emitter if known
Message::getEmitter(void) const
{
return(_emitter->thisRef());
}

void
Message::setEmitterAlias(const StlString & alias)
{
_emitter.clear();
_emitterAlias=alias;
}

const StlString & // emitter's alias
Message::getEmitterAlias(void) const
{
return(_emitterAlias);
}

ArRef<MessageService> // message service that delivered this message
Message::getService(void) const
{
return(_service->thisRef());
}

//-------- Content -----------------------------------------------------------

void
Message::setText(const StlString & text)
{
_text=text;
}

const StlString & // message content
Message::getText(void) const
{
return(_text);
}

//-------- Stream operations -------------------------------------------------

void
Message::writeToStream(ArRef<AbstractOStream> output) const
{
ArObject::writeToStream(output);
output->writeString(_text);
}

bool // success
Message::readFromStream(ArRef<AbstractIStream> input)
{
return(ArObject::readFromStream(input)&&
       input->readString(_text));
}


//----------------------------------------------------------------------------
// Class MessageService
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(MessageService,ArObject)

//-------- Construction / Destruction ----------------------------------------

MessageService::MessageService(ArCW & arCW) // Protected
: ArObject(arCW),
  _bdcast()
{
  // Nothing to be done
}

MessageService::~MessageService(void) // Protected
{
  // Nothing to be done
}

//-------- Peer to peer ------------------------------------------------------

void
MessageService::sendTo(ArRef<ArObject> receiver,
                       ArRef<Message> message)
{
message->_service=thisPtr();
receiver->_messages.push_back(message);
if(!receiver->_messageCB.empty())
  {
  ArObject::MessageEvent evt;
  receiver->_messageCB.fire(evt);
  }
}

void
MessageService::sendToAlias(const StlString & receiver,
                            ArRef<Message> message)
{
ArRef<ArObject> found;
ArObject::CLASS()->findAlias(receiver,found);
if(found)
  {
  sendTo(found,message);
  }
}

//-------- Broadcast ---------------------------------------------------------

void
MessageService::subscribeBroadcast(ArConstRef<ArClass> messageType,
                                   ArRef<ArObject> receiver)
{
for(unsigned int i=_bdcast.size();i--;)
  {
  if(!_bdcast[i].receiver) // remove deleted objects
    {
    StlVectorFastErase(_bdcast,i);
    }
  else if(_bdcast[i].receiver==receiver)
    {
    for(unsigned int j=_bdcast[i].messageTypes.size();j--;)
      {
      if(_bdcast[i].messageTypes[j]==messageType)
        {
        return; // already subscribed
        }
      }
    _bdcast[i].messageTypes.push_back(messageType);
    return;
    }
  }
_bdcast.push_back(MessageService::BroadcastSubscription());
_bdcast.back().receiver=receiver->thisPtr();
_bdcast.back().messageTypes.push_back(messageType);
}

void
MessageService::unsubscribeBroadcast(ArConstRef<ArClass> messageType,
                                     ArRef<ArObject> receiver)
{
for(unsigned int i=_bdcast.size();i--;)
  {
  if(!_bdcast[i].receiver) // remove deleted objects
    {
    StlVectorFastErase(_bdcast,i);
    }
  else if(_bdcast[i].receiver==receiver)
    {
    for(unsigned int j=_bdcast[i].messageTypes.size();j--;)
      {
      if(_bdcast[i].messageTypes[j]==messageType)
        {
        StlVectorFastErase(_bdcast[i].messageTypes,j);
        return;
        }
      }
    return;
    }
  }
}

void
MessageService::broadcast(ArRef<Message> message)
{
for(unsigned int i=_bdcast.size();i--;)
  {
  if(!_bdcast[i].receiver) // remove deleted objects
    {
    StlVectorFastErase(_bdcast,i);
    }
  else
    {
    for(unsigned int j=_bdcast[i].messageTypes.size();j--;)
      {
      if(message->getClass()->isA(_bdcast[i].messageTypes[j]))
        {
        sendTo(_bdcast[i].receiver->thisRef(),message);
        break;
        }
      }
    }
  }
}


//----------------------------------------------------------------------------
// Class NetworkMessageService
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(NetworkMessageService,MessageService)

//-------- Construction / Destruction ----------------------------------------

NetworkMessageService::NetworkMessageService(ArCW & arCW,
                                             int udpPortNumber) // Protected
: MessageService(arCW),
  _transmitter(UDPTransmitter::NEW(udpPortNumber)),
  _listener(TCPListener::NEW(0,true)),
  _sessionID(),
  _services()
{
if(_transmitter->fail())
  {
  setErrorMessage(_transmitter->getErrorMessage());
  return;
  }
if(_listener->fail())
  {
  setErrorMessage(_listener->getErrorMessage());
  return;
  }
_sessionID=ArSystem::getHostName();
_sessionID+=":";
_sessionID+=intToStr(_transmitter->getPortNumber());
_sessionID+="/";
_sessionID+=intToStr(_listener->getPortNumber());
tryToReachLocalSessions();
}


NetworkMessageService::NetworkMessageService(ArCW & arCW,
                                             int udpPortNumber,
                                             int tcpPortNumber) // Protected
: MessageService(arCW),
  _transmitter(UDPTransmitter::NEW(udpPortNumber)),
  _listener(TCPListener::NEW(tcpPortNumber,true)),
  _sessionID(),
  _services()
{
if(_transmitter->fail())
  {
  setErrorMessage(_transmitter->getErrorMessage());
  return;
  }
if(_listener->fail())
  {
  setErrorMessage(_listener->getErrorMessage());
  return;
  }
_sessionID=ArSystem::getHostName();
_sessionID+=":";
_sessionID+=intToStr(_transmitter->getPortNumber());
_sessionID+="/";
_sessionID+=intToStr(_listener->getPortNumber());
tryToReachLocalSessions();
}


NetworkMessageService::~NetworkMessageService(void) // Protected
{
  // Nothing to be done
}

//-------- Session management ------------------------------------------------

const StlString & // session ID
NetworkMessageService::getSessionID(void) const
{
return(_sessionID);
}

void
NetworkMessageService::tryToReachLocalSessions(void)
{
_transmitter->broadcastString(_sessionID,_transmitter->getPortNumber());
}

void
NetworkMessageService::tryToReachSession(unsigned int distantIPAddress,
                                         int distantUdpPortNumber)
{
_transmitter->sendString(_sessionID,distantIPAddress,distantUdpPortNumber);
}

void
NetworkMessageService::getDistantSessions(StlVector<StlString> & sessionsOut)
                                                                         const
{
sessionsOut.clear();
for(unsigned int i=_services.size();i--;)
  {
  sessionsOut.push_back(_services[i].session);
  }
}

//-------- Identification ----------------------------------------------------

void
NetworkMessageService::extractLocalAlias(const StlString & alias,
                                         StlString & localAliasOut) const
{
const char * at=strchr(alias.c_str(),'@');
if(at)
  {
  localAliasOut=StlString(alias,0,(at-alias.c_str()));
  }
else
  {
  localAliasOut=alias;
  }
}

void
NetworkMessageService::extractNetworkSession(const StlString & alias,
                                             StlString & sessionOut) const
{
const char * at=strchr(alias.c_str(),'@');
if(at)
  {
  sessionOut=at+1;
  }
else
  {
  sessionOut.erase();
  }
}

void
NetworkMessageService::makeNetworkAlias(StlString & aliasInOut) const
{
extractLocalAlias(aliasInOut,aliasInOut);
aliasInOut+="@";
aliasInOut+=_sessionID;
}

//-------- Peer to peer ------------------------------------------------------

void
NetworkMessageService::sendToAlias(const StlString & receiver,
                                   ArRef<Message> message)
{
const char * at=strchr(receiver.c_str(),'@');
if(at)
  {
  StlString alias(receiver,0,(at-receiver.c_str()));
  StlString session(at+1);
  if(_sessionID==session)
    {
    MessageService::sendToAlias(alias,message);
    }
  else
    {
    ArRef<MemoryBlock> memory=MemoryBlock::NEW();
    ArRef<MemoryOStream> stream=MemoryOStream::NEW(memory,0);
    stream->writeString(message->getClass()->getName());
    stream->writeString(message->getEmitterAlias());
    stream->writeString(alias);
    message->writeToStream(stream);
    stream->flush();
    unsigned int length=stream->getMemoryUsage();
    for(unsigned int i=_services.size();i--;)
      {
      if(_services[i].session==session)
        {
        _services[i].connection->writeBytes(memory,0,length);
        break;
        }
      }
    }
  }
else
  {
  MessageService::sendToAlias(receiver,message);
  }
}

void
NetworkMessageService::sendToSameAlias(const StlString & receiver,
                                       ArRef<Message> message)
{
StlString alias;
const char * at=strchr(receiver.c_str(),'@');
if(at)
  {
  alias=StlString(receiver,0,(at-receiver.c_str()));
  }
else
  {
  alias=receiver;
  }
ArRef<MemoryBlock> memory=MemoryBlock::NEW();
ArRef<MemoryOStream> stream=MemoryOStream::NEW(memory,0);
stream->writeString(message->getClass()->getName());
stream->writeString(message->getEmitterAlias());
stream->writeString(alias);
message->writeToStream(stream);
stream->flush();
unsigned int length=stream->getMemoryUsage();
for(unsigned int i=_services.size();i--;)
  {
  _services[i].connection->writeBytes(memory,0,length);
  }
MessageService::sendToAlias(alias,message);
}

//-------- Broadcast ---------------------------------------------------------

void
NetworkMessageService::broadcast(ArRef<Message> message)
{
ArRef<MemoryBlock> memory=MemoryBlock::NEW();
ArRef<MemoryOStream> stream=MemoryOStream::NEW(memory,0);
stream->writeString(message->getClass()->getName());
stream->writeString(message->getEmitterAlias());
StlString receiver; // no receiver --> broadcast
stream->writeString(receiver);
message->writeToStream(stream);
stream->flush();
unsigned int length=stream->getMemoryUsage();
for(unsigned int i=_services.size();i--;)
  {
  _services[i].connection->writeBytes(memory,0,length);
  }
MessageService::broadcast(message);
}

//-------- Network watching --------------------------------------------------

void
NetworkMessageService::watchNetwork(void)
{
StlString content;
unsigned int ipAddress;
int portNumber;
while(_transmitter->receiveString(content,ipAddress,portNumber))
  {
  bool found=false;
  if(content==_sessionID)
    {
    found=true;
    }
  else
    {
    for(unsigned int i=_services.size();i--;)
      {
      if(_services[i].session==content)
        {
        found=true;
        break;
        }
      }
    }
  if(!found)
    {
    int udpPort,tcpPort;
    const char * colon=strchr(content.c_str(),':');
    if(colon&&(sscanf(colon+1,"%d/%d",&udpPort,&tcpPort)==2))
      {
      NetworkMessageService::ConnectedService service;
      service.session=content;
      service.connection=TCPConnection::NEW(ipAddress,tcpPort,true);
      if(!service.connection->fail())
        {
        service.connection->writeString(_sessionID);
        _services.push_back(service);
        }
      }
    }
  }
ArRef<TCPConnection> connection=_listener->nextConnection();
while(connection)
  {
  while(!connection->fail()&&!connection->readString(content))
    {
    ArSystem::yield();
    }
  if(!connection->fail())
    {
    NetworkMessageService::ConnectedService service;
    service.session=content;
    service.connection=connection;
    _services.push_back(service);
    }
  connection=_listener->nextConnection();
  }
for(unsigned int i=_services.size();i--;)
  {
  ArRef<MemoryBlock> memory=MemoryBlock::NEW();
  while(_services[i].connection->readBytes(memory))
    {
    ArRef<MemoryIStream> stream=MemoryIStream::NEW(memory,0);
    StlString messageType;
    StlString emitter;
    StlString receiver;
    if(stream->readString(messageType)&&
       stream->readString(emitter)&&
       stream->readString(receiver))
      {
      if(!strchr(emitter.c_str(),'@'))
        {
        emitter+="@";
        emitter+=_services[i].session;
        }
      ArRef<Message> msg;
      ArConstRef<ArClass> msgClass=ArClass::find(messageType);
      if(msgClass)
        {
        msgClass->newInstance(msg);
        }
      if(msg)
        {
        msg->setEmitterAlias(emitter);
        msg->readFromStream(stream);
        if(receiver.size())
          {
          MessageService::sendToAlias(receiver,msg);
          }
        else
          {
          MessageService::broadcast(msg);
          }
        }
      }
    }
  if(_services[i].connection->fail())
    {
    StlVectorFastErase(_services,i);
    }
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
