// LI2/ENIB - AReVi - Core services
//----------------------------------------------------------------------------
// file        : AReVi/scheduler.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/scheduler.h"

#include "AReVi/arSystem.h"

#include <errno.h>
#include <sys/time.h>

namespace AReVi {

//----------------------------------------------------------------------------
// Class Scheduler (abstract, singleton)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(Scheduler,ArObject)

ArPtr<Scheduler> Scheduler::_sched;

//-------- Construction / Destruction ----------------------------------------

Scheduler::Scheduler(ArCW & arCW,
                     double precision) // Protected
: ArObject(arCW),
  _time(0.0),
  _precision(precision),
  _suspended(false),
  _pollingCB(thisRef()),
  _activities(),
  _stock(),
  _current(),
  _currentActivity()
{
if(Scheduler::_sched)
  {
  Scheduler::_sched.destroy(); // UGLY !
  }
Scheduler::_sched=thisPtr();
if(precision<=0.0)
  {
  setErrorMessage("Precision should be greater than zero");
  }
}

Scheduler::~Scheduler(void) // Protected
{
while(!_current.empty())
  {
  _current.back()->abort();
  }
for(unsigned int priority=3;priority--;)
  {
  StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
  while(!infos.empty())
    {
    infos.front()->activities.back()->abort();
    }
  }
while(!_stock.empty())
  {
  delete _stock.back();
  _stock.pop_back();
  }
if(Scheduler::_sched.c_ptr()==this)
  {
  Scheduler::_sched.clear();
  }
}

//-------- Scheduling properties ---------------------------------------------

double // current time
Scheduler::getTime(void) const
{
return(_time);
}

double // time rounding precision
Scheduler::getPrecision(void) const
{
return(_precision);
}

double // time rounded to precision
Scheduler::roundTime(double time) const
{
return(Scheduler::_roundTime(this,time));
}

//-------- Activation --------------------------------------------------------

void
Scheduler::setSuspended(bool suspended)
{
_suspended=suspended;
}

bool // suspended
Scheduler::isSuspended(void) const
{
return(_suspended);
}

//-------- Forbidden operations ----------------------------------------------

void
Scheduler::_initialize(void) // Protected
{
_time=0.0;
}

bool // remaining activities
Scheduler::_activate(void) // Protected
{
if(_suspended)
  {
  Scheduler::PollingEvent evt;
  switch(ArSystem::waitForPolledFD(evt.readingFD,evt.writingFD,20))
    {
    case -1:
      {
      cerr << "Cannot poll file descriptors --- "
           << ArSystem::printError(errno) << endl;
      break;
      }
    case 0:
      {
      break; // nothing to be done
      }
    default:
      {
      _pollingCB.fire(evt);
      break;
      }
    }
  return(!(_activities[0].empty()&&
           _activities[1].empty()&&
           _activities[2].empty()));
  }
else
  {
  if(!_jumpToNextTime())
    {
    return(false);
    }
  for(unsigned int priority=3;priority--;)
    {
    StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
    while(!infos.empty())
      {
      Scheduler::ActivityInfo * info=infos.front();
      if(info->nextTime<=_time)
        {
        _current.insert(_current.end(),
                        info->activities.begin(),info->activities.end());
        info->activities.clear();
        Scheduler::_releaseInfo(this,info);
        infos.pop_front();
        }
      else
        {
        break;
        }
      }
    while(!_current.empty())
      {
      unsigned int index=ArSystem::integerRand(_current.size());
      _currentActivity=_current[index];
      StlVectorFastErase(_current,index);
      double interval=Scheduler::_roundTime(this,
                                            _time-_currentActivity->_lastTime);
      _currentActivity->_lastTime=_time;
      bool goOn=_currentActivity->_activate(interval);
      if(_currentActivity) // not destroyed during activation
        {
        if(!goOn)
          {
          _currentActivity->_ended=true;
          }
        if(!(_currentActivity->_ended||_currentActivity->_suspended))
          {
          _insertActivity(_currentActivity);
          }
        _currentActivity.clear();
        }
      }
    }
  }
return(true);
}

// bool // remaining activities
// Scheduler::_jumpToNextTime(void)=0; // Protected

void
Scheduler::_insertActivity(ArRef<Activity> activity) // Protected
{
double nextTime=Scheduler::_computeNextTime(this,activity);
int priority=activity->_priority;
StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
StlList<Scheduler::ActivityInfo *>::iterator it=infos.begin();
for(;it!=infos.end();it++)
  {
  if(nextTime<(*it)->nextTime)
    {
    infos.insert(it,Scheduler::_useInfo(this,nextTime,activity));
    return;
    }
  if(nextTime==(*it)->nextTime)
    {
    (*it)->activities.push_back(activity);
    return;
    }
  }
infos.push_back(Scheduler::_useInfo(this,nextTime,activity));
}

void
Scheduler::_removeActivity(ArRef<Activity> activity) // Protected
{
unsigned int i;
for(i=_current.size();i--;)
  {
  if(_current[i]==activity)
    {
    StlVectorFastErase(_current,i);
    return;
    }
  }
int priority=activity->_priority;
StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
StlList<Scheduler::ActivityInfo *>::iterator it=infos.begin();
for(;it!=infos.end();it++)
  {
  if((*it)->nextTime==activity->_nextTime)
    {
    StlVector<ArRef<Activity> > & act=(*it)->activities;
    for(i=act.size();i--;)
      {
      if(act[i]==activity)
        {
        StlVectorFastErase(act,i);
        if(act.empty())
          {
          Scheduler::_releaseInfo(this,*it);
          infos.erase(it);
          }
        return;
        }
      }
    }
  }
}

void
Scheduler::_changeInterval(ArRef<Activity> activity,
                           double interval) // Protected
{
activity->_interval=interval;
if(_currentActivity==activity)
  {
  return;
  }
unsigned int i;
for(i=_current.size();i--;)
  {
  if(_current[i]==activity)
    {
    return;
    }
  }
int priority=activity->_priority;
StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
StlList<Scheduler::ActivityInfo *>::iterator it=infos.begin();
for(;it!=infos.end();it++)
  {
  if((*it)->nextTime==activity->_nextTime)
    {
    StlVector<ArRef<Activity> > & act=(*it)->activities;
    for(i=act.size();i--;)
      {
      if(act[i]==activity)
        {
        StlVectorFastErase(act,i);
        if(act.empty())
          {
          Scheduler::_releaseInfo(this,*it);
          infos.erase(it);
          }
        _insertActivity(activity);
        return;
        }
      }
    }
  }
}

void
Scheduler::_changePriority(ArRef<Activity> activity,
                           int priority) // Protected
{
int oldPriority=activity->_priority;
activity->_priority=priority;
if(_currentActivity==activity)
  {
  return;
  }
unsigned int i;
for(i=_current.size();i--;)
  {
  if(_current[i]==activity)
    {
    return;
    }
  }
StlList<Scheduler::ActivityInfo *> & infos=_activities[oldPriority];
StlList<Scheduler::ActivityInfo *>::iterator it=infos.begin();
for(;it!=infos.end();it++)
  {
  if((*it)->nextTime==activity->_nextTime)
    {
    StlVector<ArRef<Activity> > & act=(*it)->activities;
    for(i=act.size();i--;)
      {
      if(act[i]==activity)
        {
        StlVectorFastErase(act,i);
        if(act.empty())
          {
          Scheduler::_releaseInfo(this,*it);
          infos.erase(it);
          }
        _insertActivity(activity);
        return;
        }
      }
    }
  }
}


//----------------------------------------------------------------------------
// Class VirtualTimeScheduler (singleton)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(VirtualTimeScheduler,Scheduler)

//-------- Construction / Destruction ----------------------------------------

VirtualTimeScheduler::VirtualTimeScheduler(ArCW & arCW,
                                           double precision) // Protected
: Scheduler(arCW,precision)
{
  // Nothing to be done
}

VirtualTimeScheduler::~VirtualTimeScheduler(void) // Protected
{
  // Nothing to be done
}

//-------- Forbidden operations ----------------------------------------------

bool // remaining activities
VirtualTimeScheduler::_jumpToNextTime(void) // Protected
{
bool remaining=false;
double nextTime=DBL_MAX;
for(unsigned int priority=3;priority--;)
  {
  StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
  StlList<Scheduler::ActivityInfo *>::const_iterator it=infos.begin();
  for(;it!=infos.end();it++)
    {
    remaining=true;
    double t=(*it)->nextTime;
    if((t!=_time)&&(t<nextTime))
      {
      nextTime=t;
      break;
      }
    }
  }
if(nextTime==DBL_MAX)
  {
  nextTime=Scheduler::_roundTime(this,_time+_precision); // as fast as possible
  }
if(remaining)
  {
  _time=nextTime;
  Scheduler::PollingEvent evt;
  switch(ArSystem::waitForPolledFD(evt.readingFD,evt.writingFD,0))
    {
    case -1:
      {
      cerr << "Cannot poll file descriptors --- "
           << ArSystem::printError(errno) << endl;
      break;
      }
    case 0:
      {
      break; // nothing to be done
      }
    default:
      {
      _pollingCB.fire(evt);
      break;
      }
    }
  return(true);
  }
else
  {
  if(_pollingCB.empty())
    {
    return(false);
    }
  else
    {
    Scheduler::PollingEvent evt;
    switch(ArSystem::waitForPolledFD(evt.readingFD,evt.writingFD,20))
      {
      case -1:
        {
        cerr << "Cannot poll file descriptors --- "
             << ArSystem::printError(errno) << endl;
        break;
        }
      case 0:
        {
        break; // nothing to be done
        }
      default:
        {
        _pollingCB.fire(evt);
        break;
        }
      }
    return(true);
    }
  }
}


//----------------------------------------------------------------------------
// Class RealTimeScheduler (singleton)
//----------------------------------------------------------------------------

AR_CLASS_NOVOID_DEF(RealTimeScheduler,Scheduler)

//-------- Construction / Destruction ----------------------------------------

RealTimeScheduler::RealTimeScheduler(ArCW & arCW,
                                     double precision) // Protected
: Scheduler(arCW,precision),
  _realTime(0.0),
  _timeFactor(1.0)
{
  // Nothing to be done
}

RealTimeScheduler::~RealTimeScheduler(void) // Protected
{
  // Nothing to be done
}

//-------- Scheduling properties ---------------------------------------------

void
RealTimeScheduler::setTimeFactor(double timeFactor)
{
if(timeFactor>0.0)
  {
  _timeFactor=timeFactor;
  }
}

double // time factor
RealTimeScheduler::getTimeFactor(void) const
{
return(_timeFactor);
}

//-------- Activation --------------------------------------------------------

void
RealTimeScheduler::setSuspended(bool suspended)
{
if(_suspended!=suspended)
  {
  _suspended=suspended;
  if(!_suspended)
    {
    struct timeval tv;
    gettimeofday(&tv,(struct timezone *)0);
    _realTime=Scheduler::_roundTime(this,
                                    ((double)tv.tv_sec)+
                                    (((double)tv.tv_usec)*1e-6));
    }
  }
}

//-------- Forbidden operations ----------------------------------------------

void
RealTimeScheduler::_initialize(void) // Protected
{
struct timeval tv;
gettimeofday(&tv,(struct timezone *)0);
_realTime=Scheduler::_roundTime(this,
                                ((double)tv.tv_sec)+
                                (((double)tv.tv_usec)*1e-6));
_time=0.0;
}

bool // remaining activities
RealTimeScheduler::_jumpToNextTime(void) // Protected
{
bool remaining=false;
double nextTime=DBL_MAX;
for(unsigned int priority=3;priority--;)
  {
  StlList<Scheduler::ActivityInfo *> & infos=_activities[priority];
  StlList<Scheduler::ActivityInfo *>::const_iterator it=infos.begin();
  for(;it!=infos.end();it++)
    {
    remaining=true;
    double t=(*it)->nextTime;
    if(t<nextTime)
      {
      nextTime=t;
      break;
      }
    }
  }
if(nextTime==DBL_MAX)
  {
  nextTime=Scheduler::_roundTime(this,_time+_precision); // as fast as possible
  }
if(remaining)
  {
  struct timeval tv;
  gettimeofday(&tv,(struct timezone *)0);
  double newReal=((double)tv.tv_sec)+(((double)tv.tv_usec)*1e-6);
  double delta=(newReal-_realTime)*_timeFactor;
  if((delta>=_precision)&&(_time+delta>=nextTime))
    {
    _realTime=newReal;
    _time=Scheduler::_roundTime(this,_time+delta);
    if(!_pollingCB.empty())
      {
      Scheduler::PollingEvent evt;
      switch(ArSystem::waitForPolledFD(evt.readingFD,evt.writingFD,0))
        {
        case -1:
          {
          cerr << "Cannot poll file descriptors --- "
               << ArSystem::printError(errno) << endl;
          break;
          }
        case 0:
          {
          break; // nothing to be done
          }
        default:
          {
          _pollingCB.fire(evt);
          break;
          }
        }
      }
    }
  else // nothing to be done until nextTime
    {
    double sleepTime=(nextTime-_time)/_timeFactor;
    if(sleepTime>0.02)
      {
      sleepTime=0.02; // keep the application interactive
      }
    Scheduler::PollingEvent evt;
    switch(ArSystem::waitForPolledFD(evt.readingFD,evt.writingFD,
                                     (unsigned int)(1000.0*sleepTime)))
      {
      case -1:
        {
        cerr << "Cannot poll file descriptors --- "
             << ArSystem::printError(errno) << endl;
        break;
        }
      case 0:
        {
        break; // nothing to be done
        }
      default:
        {
        _pollingCB.fire(evt);
        break;
        }
      }
    }
  }
else
  {
  if(!_pollingCB.empty())
    {
    remaining=true;
    }
  Scheduler::PollingEvent evt;
  switch(ArSystem::waitForPolledFD(evt.readingFD,evt.writingFD,20))
    {
    case -1:
      {
      cerr << "Cannot poll file descriptors --- "
           << ArSystem::printError(errno) << endl;
      break;
      }
    case 0:
      {
      break; // nothing to be done
      }
    default:
      {
      _pollingCB.fire(evt);
      break;
      }
    }
  }
return(remaining);
}

} // namespace AReVi

//----------------------------------------------------------------------------
