// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/Events/nsEventDataTexture.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/Events/nsEventDataTexture.h"
#include "AReVi/Utils/abstractStream.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Lib3D/dataTexture.h"
#include "AReVi/Lib3D/llw_priv.h"

namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : NSEventDataTextureReload
//----------------------------------------------------------------------------
AR_CLASS_DEF(NSEventDataTextureReload,NSEvent)
//---------------------- Constructors / Destructor ---------------------------
NSEventDataTextureReload::NSEventDataTextureReload(ArCW & arCW) // Protected
  : NSEvent(arCW),
    _width(0),
    _height(0),
    _components(0),
    _needsBlending(false),
    _needsAlphaTest(false),
    _useMipmap(false),
    _compressData(false),
    _data(MemoryBlock::NEW())
{
}

NSEventDataTextureReload::~NSEventDataTextureReload(void)
{}

//---------------------------- Public methods --------------------------------
// Read/Write data parameters
void
NSEventDataTextureReload::writeDataTexture(ArConstRef<DataTexture> dataTexture)
{
  _useMipmap=dataTexture->usesMipmap();
  _needsBlending=dataTexture->getBlending();
  _needsAlphaTest=dataTexture->getAlphaTest();
  _useMipmap=dataTexture->usesMipmap();

// Compress dataTexture To JPEG
  if (!LLW::saveRGBImageToJPEG(dataTexture->getData(),
                          dataTexture->getWidth(),
                          dataTexture->getHeight(),
                          DataTexture::getQuality(),
                          _data))
  {
  // Uncompressed version
    _data->setSize(dataTexture->getData()->getSize());
    memcpy(_data->accessMemory(),dataTexture->getData()->getMemory(),
                                 _data->getSize());
    _compressData=false;
    setCompress(true); // compress message!!!
  }
  else
  {
    _compressData=true;
    setCompress(false); // don't recompress!!!
  }
}

void
NSEventDataTextureReload::readDataTexture(ArRef<DataTexture> dataTexture) const
{
  if (_compressData)
  {
    // compressed data
    unsigned int w,h;
    ArRef<MemoryBlock> buffer=MemoryBlock::NEW();

    if (!LLW::loadRGBImageFromJPEG(_data,w,h,buffer))
    {
      cerr << "NSEventDataTextureReload::can't uncompress dataTexture " <<endl
          << "     => " << dataTexture->accessData()->getErrorMessage() << endl;
      exit(-1);
    }

    if (w!=dataTexture->getWidth() || h!=dataTexture->getHeight())
    {
      cerr << "NSEventDataTextureReload::uncompress dataTexture ERROR,"
              " bad data size (w,h)" <<endl;
      exit(-1);
    }
    if (_data->getSize()>dataTexture->getData()->getSize())
    {
    cerr << "NSEventDataTextureReload::readDataTexture()=>"
            " ERROR, BUFFER SIZE" <<endl;
    }

    buffer->swapBlock(dataTexture->accessData());
  }
  else
  {
    // Uncompressed data
    if (_data->getSize()>dataTexture->getData()->getSize())
    {
    cerr << "NSEventDataTextureReload::readDataTexture()=>"
            " ERROR, BUFFER SIZE" <<endl;
    }
    else
    {
    memcpy(dataTexture->accessData()->accessMemory(),_data->getMemory(),
                                                     _data->getSize());
    }
  }
  dataTexture->reload(false,_useMipmap);
  dataTexture->setBlending(_needsBlending);
  dataTexture->setAlphaTest(_needsAlphaTest);
}

//-------------------------- Stream operations -------------------------------
void
NSEventDataTextureReload::writeToStream(ArRef<AbstractOStream> output) const
{
  NSEvent::writeToStream(output);

  unsigned int flags=(_needsBlending  ? 1<<0 : 0)+
                     (_needsAlphaTest ? 1<<1 : 0)+
                     (_useMipmap      ? 1<<2 : 0)+
                     (_compressData   ? 1<<3 : 0);

  output->writeUInt(flags);

  output->writeUInt(_data->getSize());
  output->writeData(_data->getMemory(),_data->getSize());
}

bool // success
NSEventDataTextureReload::readFromStream(ArRef<AbstractIStream> input)
{
  unsigned int flags,size;

  if (!NSEvent::readFromStream(input)||
      !input->readUInt(flags)        ||
      !input->readUInt(size))
  {
  return(false);
  }
  _data->setSize(size);
  if (!input->readData(_data->accessMemory(),size))
  {
  return(false);
  }

  _needsBlending=flags&(1<<0);
  _needsAlphaTest=flags&(1<<1);
  _useMipmap=flags&(1<<2);
  _compressData=flags&(1<<3);
  return true;
}

//----------------------------------------------------------------------------
// CLASS       : NSEventDataTextureSetup
//----------------------------------------------------------------------------
AR_CLASS_DEF(NSEventDataTextureSetup,NSEvent)
//---------------------- Constructors / Destructor ---------------------------
NSEventDataTextureSetup::NSEventDataTextureSetup(ArCW & arCW) // Protected
  : NSEvent(arCW),
    _needsBlending(false),
    _needsAlphaTest(false)
{}

NSEventDataTextureSetup::~NSEventDataTextureSetup(void)
{}

//---------------------------- Public methods --------------------------------
// Read/Write data parameters
void
NSEventDataTextureSetup::writeDataTexture(ArConstRef<DataTexture> dataTexture)
{
  _needsBlending=dataTexture->getBlending();
  _needsAlphaTest=dataTexture->getAlphaTest();
}

void
NSEventDataTextureSetup::readDataTexture(ArRef<DataTexture> dataTexture) const
{
  dataTexture->setBlending(_needsBlending);
  dataTexture->setAlphaTest(_needsAlphaTest);
}

//-------------------------- Stream operations -------------------------------
void
NSEventDataTextureSetup::writeToStream(ArRef<AbstractOStream> output) const
{
  NSEvent::writeToStream(output);

  unsigned int flags=(_needsBlending  ? 1<<0 : 0)+
                     (_needsAlphaTest ? 1<<1 : 0);
  output->writeUInt(flags);
}

bool // success
NSEventDataTextureSetup::readFromStream(ArRef<AbstractIStream> input)
{
  unsigned int flags;

  if (!NSEvent::readFromStream(input)||
      !input->readUInt(flags))
  {
  return(false);
  }

  _needsBlending=flags&(1<<0);
  _needsAlphaTest=flags&(1<<1);
  return true;
}


} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
