// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/Events/nsEventObject3D.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/Events/nsEventObject3D.h"
#include "AReVi/Utils/abstractStream.h"
#include "AReVi/Shapes/shape3D.h"
#include "AReVi/Lib3D/object3D.h"
#include "AReVi/Lib3D/material3D.h"

namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : NSEventObject3DShapeChange
//----------------------------------------------------------------------------
AR_CLASS_DEF(NSEventObject3DShapeChange,NSEvent)
//---------------------- Constructors / Destructor ---------------------------
NSEventObject3DShapeChange::NSEventObject3DShapeChange(ArCW & arCW) // Protected
  : NSEvent(arCW),_shapes(),_distances()
{
}

NSEventObject3DShapeChange::~NSEventObject3DShapeChange(void)
{
}
//---------------------------- Public methods --------------------------------
void
NSEventObject3DShapeChange::writeObject3D(ArConstRef<Object3D> object)
{
  _shapes.clear();
  _distances.clear();

  StlVector<ArRef<Shape3D> > shapes;

  object->queryShapes(_distances,shapes);
  for (unsigned i=shapes.size();i--;)
  {
    _shapes.push_back(RemoteRef(shapes[i]));
  }
}

bool // If fail => write error detail in object (getErrorMessage())
NSEventObject3DShapeChange::readObject3D(ArRef<Object3D> object,
                                         ArRef<RemoteDepService> depService) const
{
  StlVector<ArRef<Shape3D> > shapes;

  for (unsigned i=_shapes.size();i--;)
  {
    ArRef<Shape3D> shape;

    if (!_shapes[i].request(depService,shape))
    {
      object->setErrorMessage("Can not resolve dependencies => NSEventObject3DShapeChange fail");
      return false;
    }
    shapes.push_back(shape);
  }

  if (_distances.size()) // LOD
    object->setLODShapes(_distances, shapes);
  else if (shapes.size()) // OK
    object->setShape(shapes[0]);
  else
    object->setShape(Shape3D::nullRef());

  return true;
}

//-------------------------- Stream operations -------------------------------
bool // success
NSEventObject3DShapeChange::readFromStream(ArRef<AbstractIStream> input)
{
//cerr << "NSEventObject3DShapeChange::readFromStream" << endl;
  if (!NSEvent::readFromStream(input))
    return false;

  unsigned nbElts;
  input->readUInt(nbElts);
  _shapes.resize(nbElts);

  for (unsigned i=_shapes.size();i--;)
    _shapes[i].readFromStream(input);

  input->readUInt(nbElts);
  _distances.resize(nbElts);
  for (unsigned i=_distances.size();i--;)
    input->readDouble(_distances[i]);

  return true;
}

void
NSEventObject3DShapeChange::writeToStream(ArRef<AbstractOStream> output) const
{
  NSEvent::writeToStream(output);

  output->writeUInt(_shapes.size());
  for (unsigned i=_shapes.size();i--;)
    _shapes[i].writeToStream(output);


  output->writeUInt(_distances.size());
  for (unsigned i=_distances.size();i--;)
    output->writeDouble(_distances[i]);

}

//----------------------------------------------------------------------------
// CLASS       : NSEventObject3DAppearance
//----------------------------------------------------------------------------
AR_CLASS_DEF(NSEventObject3DAppearance,NSEvent)
//---------------------- Constructors / Destructor ---------------------------
NSEventObject3DAppearance::NSEventObject3DAppearance(ArCW & arCW) // Protected
  : NSEvent(arCW),
    _wireFrame(false),
    _drawShapeBBox(false),
    _drawShapePartBBox(false),
    _material(Material3D::NEW())
{}

NSEventObject3DAppearance::~NSEventObject3DAppearance(void)
{}

//---------------------------- Public methods --------------------------------
void
NSEventObject3DAppearance::writeObject3D(ArConstRef<Object3D> object)
{
  _wireFrame=object->getWireFrame();
  _drawShapeBBox=object->getShapeBoundingBox();
  _drawShapePartBBox=object->getShapePartBoundingBox();
  object->readSelectedMaterial(_material);
}

void
NSEventObject3DAppearance::readObject3D(ArRef<Object3D> object) const
{
  object->setWireFrame(_wireFrame);
  object->setShapeBoundingBox(_drawShapeBBox);
  object->setShapePartBoundingBox(_drawShapePartBBox);
  object->writeSelectedMaterial(_material);
}

//-------------------------- Stream operations -------------------------------
bool // success
NSEventObject3DAppearance::readFromStream(ArRef<AbstractIStream> input)
{
  if (!NSEvent::readFromStream(input))
    return false;
  unsigned int flags;
  if(!input->readUInt(flags)||
     !_material->readFromStream(input))
    {
    return(false);
    }
  _wireFrame=flags&(1<<0);
  _drawShapeBBox=flags&(1<<1);
  _drawShapePartBBox=flags&(1<<2);
  return(true);
}

void
NSEventObject3DAppearance::writeToStream(ArRef<AbstractOStream> output) const
{
  NSEvent::writeToStream(output);
  unsigned int flags=(_wireFrame         ? (1<<0) : 0)+
                     (_drawShapeBBox     ? (1<<1) : 0)+
                     (_drawShapePartBBox ? (1<<2) : 0);
  output->writeUInt(flags);
  _material->writeToStream(output);
}

} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
