// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/localDepService.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/localDepService.h"
#include "AReViNS/Events/nsEventObjectRef.h"
#include "AReViNS/Events/nsEventFile.h"
#include "AReViNS/Events/nsEventControler.h"
#include "AReViNS/Controlers/nsControler.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/memoryStream.h"

namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : LocalDepService
//----------------------------------------------------------------------------
AR_CLASS_DEF(LocalDepService,LocalService)
//---------------------- Constructors / Destructor ---------------------------
LocalDepService::LocalDepService(ArCW & arCW) // Protected
  : LocalService(arCW),
    _listenersMap(),
    _downCastMap(),
    _listeners(),
    _waitingClients()
{
  _initLocalDepService();
}

LocalDepService::LocalDepService(ArCW & arCW, const int &port) // Protected
  : LocalService(arCW,port),
    _listenersMap(),
    _downCastMap(),
    _listeners(),
    _waitingClients()
{
  _initLocalDepService();
}

LocalDepService::~LocalDepService(void)
{
  // Nothing to be done
}

//---------------------------- Public methods --------------------------------
// send object to specified or all connected service(s)
bool
LocalDepService::shareObject(ArRef<ArObject> sharedObject, int serviceID)
{
  ArRef<NSEventObjectRef> eventRef=NSEventObjectRef::NEW();
  eventRef->setRef(sharedObject);

 // cerr << "Share " << sharedObject->getAlias() << endl;

  if (!sendEvent(eventRef,serviceID))
      return false;

  sharedObject->removeDestroyCB(thisRef(),&LocalDepService::_onSharedObjectDestroyed);
  sharedObject->addDestroyCB(thisRef(),&LocalDepService::_onSharedObjectDestroyed);


  return _attachListener(sharedObject);
}


// Register down casted shared object
//   => client class is a parent of the server class
bool // success if clientClass is parent of serverClass
LocalDepService::registerDownCast(const StlString &serverClass,
                                  const StlString &clientClass,
                                  bool subclass)
{
  ArConstRef<ArClass> server(ArClass::find(serverClass)),
                      client(ArClass::find(clientClass));

  if (!server.valid() || !client.valid()) // server and client must exist!!
  {
    if (!server.valid())
      setErrorMessage("serverClass=" + serverClass + " must exist!");
    if (!client.valid())
      setErrorMessage("clientClass=" + clientClass + " must exist!");
    return false;
  }

  if (!server->isA(client)) // server must derive client
  {
    setErrorMessage(clientClass + " must derive from " + serverClass);
    return false;
  }

  _downCastMap[serverClass]=clientClass;

  if (subclass) // All subclass of 'serverClass' are also mapped into 'clientClass'
  {
    StlVector<ArConstRef<ArClass> > descendantsOut;
    server->getDescendants(descendantsOut);
    for (unsigned int i=descendantsOut.size();i--;)
      _downCastMap[descendantsOut[i]->getName()]=clientClass;
  }

// Send info to clients
// NOTE => not important if register is down before client connection!
  ArRef<NSEventDownCastMap> event=NSEventDownCastMap::NEW();
  event->setMap(_downCastMap);
  sendEvent(event);
  return true;
}

// send file pointed by url to specified serviceID or all if serviceID=-1
bool
LocalDepService::shareURL(const StlString &url, int serviceID)
{
  ArRef<NSEventFile> fileEvent=NSEventFile::NEW();
  fileEvent->setURL(url);

  sendEvent(fileEvent,serviceID);

  setErrorMessage(fileEvent->getErrorMessage());

  //cerr << "Send URL=" <<url << "   to=" << serviceID << "    ERROR=" << getErrorMessage() << endl;
  return !fail();
}

//------------------------  Controlers management ----------------------------
// Register controlers
bool
LocalDepService::registerListener(const StlString &objectClass,
                                  const StlString &ctrlClass,
                                  bool subclass)
{
  ArConstRef<ArClass> object(ArClass::find(objectClass)),
                      ctrl(ArClass::find(ctrlClass));

  if (!object.valid() || !ctrl.valid()) // server and client must exist!!
  {
    if (!object.valid())
      setErrorMessage("objectClass=" + objectClass + " must exist!");
    if (!ctrl.valid())
      setErrorMessage("ctrlClass=" + ctrlClass + " must exist!");
    return false;
  }

  _listenersMap[objectClass]=ctrlClass;

  if (subclass) // All subclass of 'serverClass' are also registered
                // as using the same listener
  {
    StlVector<ArConstRef<ArClass> > descendantsOut;
    object->getDescendants(descendantsOut);
    for (unsigned int i=descendantsOut.size();i--;)
      _listenersMap[descendantsOut[i]->getName()]=ctrlClass;
  }

  return true;
}

//------------------------- Forbiden operations ------------------------------
// CB : inform that a client need an object
NSEvent::EventProcess
LocalDepService::_onRequestRemoteReference(ArConstRef<NSEvent> eventObj)
{
//cerr << "LocalDepService::_onRequestRemoteObject" << endl;
  ArConstRef<NSEventObjectRefRequest> event=ar_down_cast<NSEventObjectRefRequest>(eventObj);
//cerr << "Request of " << event->getID() << endl;
  ArRef<ArObject> object;
  ArObject::CLASS()->findAlias(event->getID(),object);

  if (object)
  {
    shareObject(object,eventObj->getServiceId());
  }
  else // request alias not found ( no more exist?) => return empty ref
  {
    //cerr << "WARNING : Can't find requested local alias : '" << event->getID() << "'" << endl;
    ArRef<NSEventObjectRef> eventRef=NSEventObjectRef::NEW();
    eventRef->setRef(ArObject::nullRef());
    eventRef->setID(event->getID());
    sendEvent(eventRef,eventObj->getServiceId());
  }
  return NSEvent::EP_OK;
}

// CB : inform that a client need a listener for gived reference
NSEvent::EventProcess
LocalDepService::_onRequestRemoteReferenceListener(ArConstRef<NSEvent> eventObj)
{
  ArConstRef<NSEventObjectListenerRequest> event=ar_down_cast<NSEventObjectListenerRequest>(eventObj);
  ArRef<ArObject> object;
  ArObject::CLASS()->findAlias(event->getID(),object);

  if (object)
  {
  _attachListener(object);
  }
  else // request alias not found ( no more exist?) => return empty ref
  {
    ArRef<NSEventObjectRef> eventRef=NSEventObjectRef::NEW();
    eventRef->setRef(ArObject::nullRef());
    eventRef->setID(event->getID());
    sendEvent(eventRef,eventObj->getServiceId());
  }
  return NSEvent::EP_OK;
}


// CB : inform that a shared object is destroyed
void
LocalDepService::_onSharedObjectDestroyed(const ArObject::DestroyEvent &event)
{
  StlMap<StlString, ArRef<NSListener> >::iterator it=_listeners.find(event.source->getAlias());

  if (it!=_listeners.end()) // pending message to this listener will be ingnored.
  {
    it->second.destroy(); // UGLY !
    _listeners.erase(it);
  }

  ArRef<NSEventObjectRef> eventRef=NSEventObjectRef::NEW();
  eventRef->setRef(ArObject::nullRef());
  eventRef->setID(event.source->getAlias());
  sendEvent(eventRef);
}

// CB : inform that a client need an object
NSEvent::EventProcess
LocalDepService::_onRequestRemoteFile(ArConstRef<NSEvent> eventObj)
{
  //cerr << "LocalDepService::_onRequestRemoteFile" << endl;
  ArConstRef<NSEventFileRequest> event=ar_down_cast<NSEventFileRequest>(eventObj);

  shareURL(event->getURL(),eventObj->getServiceId());
  return NSEvent::EP_OK;
}

// CB : inform that a new service is connected => send informations.
void
LocalDepService::_onNewService(const LocalService::InformNewRemoteServiceEvent &evt)
{
//  cerr << "Send Map To NEW SERVICE size=" << _downCastMap.size() << endl;
  // Send info to clients => Map for down cast!!!
  ArRef<NSEventDownCastMap> event=NSEventDownCastMap::NEW();
  event->setMap(_downCastMap);
  sendEvent(event,evt.serviceID);

  if (_waitingClients.empty())
    _flushPendingEvents();
}

// CB When a remoteDepService deconnect!
void
LocalDepService::_onDestroyService(const NetworkService::ServiceDeconnectEvent &evt)
{
StlVector<int>::iterator it=find(_waitingClients.begin(),_waitingClients.end(),evt.serviceID);

if (it!=_waitingClients.end())
  {
  _waitingClients.erase(it);

  if (_waitingClients.empty())
    _flushPendingEvents();
  }
}

bool
LocalDepService::_attachListener(ArRef<ArObject> sharedObject)
{
StlString className=sharedObject->getClass()->getName();

// Look if a listener already exist for this object.
if (_listeners.find(sharedObject->getAlias())==_listeners.end())
  {
  // Add Listener to shared object if any
  StlMap<StlString, StlString>::const_iterator it=_listenersMap.find(className);

  if (it==_listenersMap.end()) // No listener
    {
    if (_downCastMap.find(className)!=_downCastMap.end())
      {
      className=_downCastMap[className];
      it=_listenersMap.find(className);
      }
    }

  if (it!=_listenersMap.end()) // If specified, use it!
    {
    ArConstRef<ArClass> ctrlClass=ArClass::find(it->second);
    ArRef<NSListener> ctrlObject;

    if(ctrlClass)
      {
      ctrlClass->newInstance(ctrlObject);

      if(!ctrlObject)
        {
        setErrorMessage("Can't create nsControler of type =>'" + it->second + "' => FAIL");
        _onError(NS_ERROR_LISTENER_FAIL);
        }
      else
        {
        ctrlObject->init(thisRef(),sharedObject);
        _listeners[sharedObject->getAlias()]=ctrlObject;
        }
      }
    else
      {
      setErrorMessage("Can't find NSListener of type =>'" + it->second + "' for object class '" + it->first + "' => FAIL");
      _onError(NS_ERROR_LISTENER_FAIL);
      }
    }
    else // try find one automatically
    {
      StlString ltClassName="NSListener"+className;

      ArConstRef<ArClass> ctrlClass=ArClass::find(ltClassName);
      ArRef<NSListener> ctrlObject;

      if(ctrlClass)
      {
        ctrlClass->newInstance(ctrlObject);

        if(ctrlObject)
        {
          ctrlObject->init(thisRef(),sharedObject);
          _listeners[sharedObject->getAlias()]=ctrlObject;
        }
      }
    }
  }
return true;
}

bool
LocalDepService::_updateNetwork(ArRef<Activity> , double )
{
//   static ArRef<NSEventControler> eventCtrl(NSEventControler::NEW());
//
//   if(!_networkAct.valid())
//     return false;
//
//   StlMap<StlString, ArRef<NSListener> >::iterator itCtrl=_listeners.begin();
//   StlMap<StlString, ArRef<NSListener> >::iterator itCtrl_end=_listeners.end();
//
//   while(itCtrl!=itCtrl_end)
//   {
//     StlList<ArRef<NSEvent> > events;
//
//     itCtrl->second->getPendingEvents(events);
//     StlList<ArRef<NSEvent> >::iterator it=events.begin();
//     StlList<ArRef<NSEvent> >::iterator it_end=events.end();
//
//     eventCtrl->setID(itCtrl->first);
//
//     while(it!=it_end)
//     {
//       eventCtrl->setEvent(*it);
//       sendEvent(eventCtrl,-1);
//       it++;
//     }
//     itCtrl++;
//   }
//
//   return _networkAct.valid();

if (_waitingClients.empty()) // EveryBody is ready
  return _flushPendingEvents();
else
  return _networkAct.valid();
}

// CB : inform that a client have processed all his messages
NSEvent::EventProcess
LocalDepService::_onRemoteDepServiceReady(ArConstRef<NSEvent> eventObj)
{
StlVector<int>::iterator it=find(_waitingClients.begin(),_waitingClients.end(),eventObj->getServiceId());

if (it!=_waitingClients.end())
  _waitingClients.erase(it);
//   cerr << "Client Ready Empty=" << _waitingClients.empty() << endl;
if (_waitingClients.empty()) // EveryBody is ready
  _flushPendingEvents();

return NSEvent::EP_OK;
}


bool
LocalDepService::_flushPendingEvents(void)
{
 // NEW VERSION

  static ArRef<NSEventControlerPack> eventCtrlPack(NSEventControlerPack::NEW());
  ArRef<NSEventControler> eventCtrl;
  if(!_networkAct.valid())
    return false;

// Look for all pending messages
  StlMap<StlString, ArRef<NSListener> >::iterator itCtrl=_listeners.begin();
  StlMap<StlString, ArRef<NSListener> >::iterator itCtrl_end=_listeners.end();
  StlList<ArRef<NSEvent> > eventsPack;

  while(itCtrl!=itCtrl_end)
  {
    StlList<ArRef<NSEvent> > events;

    itCtrl->second->getPendingEvents(events);
    StlList<ArRef<NSEvent> >::iterator it=events.begin();
    StlList<ArRef<NSEvent> >::iterator it_end=events.end();

    while(it!=it_end)
    {
      eventCtrl=NSEventControler::NEW();
      eventCtrl->setID(itCtrl->first);
      eventCtrl->setEvent(*it);
      //sendEvent(eventCtrl,-1);
      eventsPack.push_back(eventCtrl);
      it++;
    }
    itCtrl++;
  }
  if (eventsPack.size())
    {
    eventCtrlPack->setEvents(eventsPack);
    sendEvent(eventCtrlPack,-1);
    _waitingClients.clear();
    getServices(_waitingClients);
    }

  return _networkAct.valid();
}


// Init localDepService instance (only called by constructor!)
void
LocalDepService::_initLocalDepService(void) // private
{
  _setConnectionMessage("SharedObjects");

  if (fail())
    return;

  addInformNewRemoteServiceCB(thisRef(),&LocalDepService::_onNewService);
  addServiceDeconnectCB(thisRef(),&LocalDepService::_onDestroyService);

  registerEvent(NSEventObjectRefRequest::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalDepService::_onRequestRemoteReference);
  registerEvent(NSEventObjectListenerRequest::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalDepService::_onRequestRemoteReferenceListener);
  registerEvent(NSEventFileRequest::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalDepService::_onRequestRemoteFile);
  registerEvent(NSEventRemoteDepReady::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalDepService::_onRemoteDepServiceReady);

// register non automatics listeners (default are NSListener+ClassName).
// eg: Cow derive from Object3D (and exist on client side!) => use
//     LocalDepService::registerListener("Cow","NSListenerObject3D");
//     If cow non exist on client side use directly
//     LocalDepService::registerDownCast("Cow","Object3D");
  registerListener("LineSet3D","NSListenerShapePart3D",true);
  registerListener("PointSet3D","NSListenerShapePart3D",true);
  registerListener("URLSound3D","NSListenerSound3D",true);

  // NOTICE : Window3D and all subClass are shared as Renderer3D

  registerDownCast("ArTweaker","ArConsole",true);
  registerDownCast("Window3D","Renderer3D",true);
  registerDownCast("Text2Speech","DataSound3D",true);

  //registerListener("TextureRenderer3D","NSListenerRenderer3D");
}


} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
