// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/localService.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/localService.h"
#include "AReViNS/Events/nsEventObjectRef.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/memoryStream.h"
#include "AReVi/Utils/tcpUdp.h"
#include "AReVi/Utils/thread.h"

namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : LocalService
//----------------------------------------------------------------------------
AR_CLASS_DEF(LocalService,NetworkService)
//---------------------- Constructors / Destructor ---------------------------
LocalService::LocalService(ArCW & arCW) // Protected
  : NetworkService(arCW),
    _listener(),
    _informNewRemoteServiceCB(thisRef())
{
  openSlot(0);

  if(_listener->fail())
  {
  setErrorMessage(_listener->getErrorMessage());
  return;
  }
}

LocalService::LocalService(ArCW & arCW, const int &port) // Protected
  : NetworkService(arCW),
    _listener(),
    _informNewRemoteServiceCB(thisRef())
{
  openSlot(port);

  if(_listener->fail())
  {
    setErrorMessage(_listener->getErrorMessage());
    return;
  }
}

LocalService::~LocalService(void)
{
openSlot(-1);//close listener
}
//---------------------------- Public methods --------------------------------
// Service allow connection on specified port number
bool
LocalService::openSlot(int portNumber)
{
  if (_listener.valid()) // Erase waiting connections!
    {
    ArSystem::removeReadFDPolling(_listener->getSocketFD());
    _listener.destroy(); // UGLY !
    }
  _listener.clear();

  if (portNumber>=0)
  {
    _listener=TCPListener::NEW(portNumber,true);
    if (!_listener->fail())
      {
      ArSystem::addReadFDPolling(_listener->getSocketFD());
      return true;
      }
    else
      return false;
  }
  else
    return true;
}

// Port number of listenner for new connections.
int
LocalService::getListenningPortNumber(void) const
{
  return _listener->getPortNumber();
}

// Port number of service.
void
LocalService::getServicePortNumber(int serviceId, int &readPort, int &writePort) const
{
  NetworkService::_getPortNumber(serviceId,readPort,writePort);
}

// IP of service.
unsigned int
LocalService::getServiceIPAdress(int serviceId) const
{
  return NetworkService::_getDistantIPAdress(serviceId);
}

// Distant port number of service.
void
LocalService::getServiceDistantPortNumber(int serviceId, int &readPort, int &writePort) const
{
NetworkService::_getDistantPortNumber(serviceId, readPort, writePort);
}


// Read data from this service
unsigned int
LocalService::getDataReadCount(int serviceId) const
{
  return NetworkService::_getDataReadCount(serviceId);
}

// Write data to this service
unsigned int
LocalService::getDataWriteCount(int serviceId) const
{
  return NetworkService::_getDataWriteCount(serviceId);
}

// Service session name
StlString  // "" on error
LocalService::getServiceName(int serviceId) const
{
  return NetworkService::_getServiceSessionName(serviceId);
}

// Service host dname
StlString  // "" on error
LocalService::getServiceHostName(int serviceId) const
{
  return NetworkService::_getServiceHostName(serviceId);
}

// Connection method Remote Side
// 1 - Remote connect to local service (write)
// 2 - Remote open a listener waiting to local service connection (read)
// 3 - Remote send to local service listening port number.
// 4 - Remote accept (or not) local connection
// 5 - Send session name
// 6 - Wait for session name
// 7 - Finalize connection
// Connection method Local Side
// 1 - Local open a listener waiting to remote service connection
// 2 - Local accept (or not) remote connection (read connection)
// 3 - Local get remote port number listener
// 4 - Local connect to remote service (write)
// 5 - Wait session name
// 6 - Send session name
// 7 - Finalize connection


bool
LocalService::_getNetworkConnections(void)
{
  if (!_listener.valid())
    return true; // no TCPListenner

  StlString message,sessionName,hostName;
  ArRef<TCPConnection> connectionRead;

// Watch for new connections
while(1)
  {
  connectionRead=NetworkService::_acceptConnection(_listener, NS_REQUEST_CONNECTION_MESSAGE+_connectionMessage);
  if (!connectionRead.valid() || connectionRead->fail()) // ignore
    break;

// get Writing connection port
  unsigned int t0=ArSystem::getTime();
  int port(-1);

  while(!connectionRead->fail()&&!connectionRead->readString(message)&&(ArSystem::getTime()-t0)<NetworkService::getTimeOut())
    {
    ArSystem::yield();
    }
  if (connectionRead->fail() || (ArSystem::getTime()-t0)>=NetworkService::getTimeOut() || !strToInt(port,message))
    continue; // ignore->go next

// Now connect to writing connection
  ArRef<TCPConnection> connectionWrite=_connect(connectionRead->getDistantIPAddress(),port);
  if (!connectionWrite.valid() || connectionWrite->fail())
    continue; // ignore->go next

// Wait session@host Name
  t0=ArSystem::getTime();
  while(!connectionRead->fail()&&!connectionRead->readString(sessionName)&&(ArSystem::getTime()-t0)<NetworkService::getTimeOut())
    {
    ArSystem::yield();
    }
  if (connectionRead->fail() || (ArSystem::getTime()-t0)>=NetworkService::getTimeOut() || port<=0)
    continue; // ignore->go next

  const char * at=strchr(sessionName.c_str(),'@');
  if(at)
  {
    hostName=StlString(at+1);
    sessionName=StlString(sessionName,0,(at-sessionName.c_str()));
  }
  else
    hostName="";

// Send session Name
  connectionWrite->writeString(_sessionName+"@"+ArSystem::getHostName());

// Finalize
  int id=_open(connectionRead,connectionWrite,sessionName,hostName);

  if(!_informNewRemoteServiceCB.empty())
    {
    LocalService::InformNewRemoteServiceEvent event;
    event.serviceID=id;
    _informNewRemoteServiceCB.fire(event);
    }
  }
  return true;
}

void
LocalService::_onPolling(const Scheduler::PollingEvent &evt)
{
if (evt.writingFD.size() && !_setNetworkEvents())
  _networkAct.clear();

if (evt.readingFD.size() && !_getNetworkEvents())
  _networkAct.clear();

if (!processEvents())
  _networkAct.clear();

if (_listener.valid())
  if (find(evt.readingFD.begin(),evt.readingFD.end(),_listener->getSocketFD())!=evt.readingFD.end())
    _getNetworkConnections();
}


} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
