// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/localViewer3D.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/localViewer3D.h"
#include "AReViNS/Controlers/nsCtrlViewer3D.h"
#include "AReViNS/Events/nsEventViewer3D.h"
#include "AReViNS/Events/nsEventWindow3D.h"
#include "AReViNS/Events/nsEventControler.h"

#include "AReVi/activity.h"
#include "AReVi/Lib3D/scene3D.h"
#include "AReVi/Lib3D/sound3D.h"



//------------- PERF TESTS
#include "AReVi/arSystem.h"

namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : LocalViewer3D
//----------------------------------------------------------------------------
AR_CLASS_NOVOID_DEF(LocalViewer3D,LocalService)

//---------------------- Constructors / Destructor ---------------------------
LocalViewer3D::LocalViewer3D(ArCW & arCW,
                               int             port,
                               ArRef<LocalDepService> depService) // Protected
  : LocalService(arCW,port),
    _infos(),
    _depService(depService),
    _viewerSetup()
{
  _initLocalViewer3D();
}

LocalViewer3D::~LocalViewer3D(void)
{
  close(); // call close here while _infos are valid.

  // Clear all infos
  StlMap<int, ClientInfos *>::iterator it=_infos.begin();
  StlMap<int, ClientInfos *>::iterator it_end=_infos.end();
  while(it!=it_end)
  {
    if (it->second)
      delete it->second;
    it++;
  }
  _infos.clear();
}

//---------------------------- Public methods --------------------------------
ArRef<LocalDepService>
LocalViewer3D::accessDepService(void)
{
  if (_depService.valid())
    return _depService;
  else
    return LocalDepService::nullRef();
}

ArRef<Viewer3D>
LocalViewer3D::accessViewer(int serviceID)
{
  if (_infos.find(serviceID)!=_infos.end())
    return _infos[serviceID]->_viewer;
  else
    return Viewer3D::nullRef();
}

ArRef<Listener3D>
LocalViewer3D::accessListener(int serviceID)
{
  if (_infos.find(serviceID)!=_infos.end())
    return _infos[serviceID]->_soundListener;
  else
    return Listener3D::nullRef();
}

bool
LocalViewer3D::setListener(int serviceID, ArRef<Listener3D> listener)
{
  if (_infos.find(serviceID)!=_infos.end())
  {
    _infos[serviceID]->_soundListener=listener;
    ArRef<NSEventViewer3DListener> event=NSEventViewer3DListener::NEW();
    event->setRef(listener);
    return sendEvent(event,serviceID);
  }
  else
    return false;
}

// Register and unregister events
// Method must return a new valid viewer else connection will be closed
bool
LocalViewer3D::setViewerSetup(ArRef<ArObject> performer,
                               ViewerSetupMethod  method)
{
  if (performer.valid())
  {
    _viewerSetup.performer=performer;
    _viewerSetup.method=method;
    if (_viewerSetup.performer.c_ptr()==thisRef().c_ptr())
      _viewerSetup.autoref=true;
    else
      _viewerSetup.autoref=false;
  }


  return true;
}

//------------------------- Forbiden operations ------------------------------
void
LocalViewer3D::_initLocalViewer3D(void)
{
  _setConnectionMessage("SharedViewer3D");
  if (fail())
    return;
  if (_depService->fail())
  {
    setErrorMessage("Can't init viewerServer cause of DepService Error : \n     =>" + _depService->getErrorMessage());
    return;
  }
  setViewerSetup(thisRef(),(LocalViewer3D::ViewerSetupMethod)&LocalViewer3D::_onViewerSetup);

  addInformNewRemoteServiceCB(thisRef(),&LocalViewer3D::_onNewRemoteViewer3D);
  addServiceDeconnectCB(thisRef(),&LocalViewer3D::_onViewer3DDeconnect);

  registerEvent(NSEventWindow3DGeometry::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalViewer3D::_onGeometryEvent);
  registerEvent(NSEventWindow3DMouseMotion::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalViewer3D::_onMouseMotionEvent);
  registerEvent(NSEventWindow3DMouseButton::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalViewer3D::_onMouseButtonEvent);
  registerEvent(NSEventWindow3DKeyboard::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalViewer3D::_onKeyboardEvent);
  registerEvent(NSEventWindow3DClose::CLASS()->getName(),
                thisRef(),
                (NetworkService::EventMethod)&LocalViewer3D::_onCloseEvent);
}

ArRef<Viewer3D>
LocalViewer3D::_onViewerSetup(unsigned int)
{
  return Viewer3D::NEW();
}

//
void
LocalViewer3D::_onNewRemoteViewer3D(const LocalService::InformNewRemoteServiceEvent &event)
{
  if (_infos.find(event.serviceID)!=_infos.end())
  {
    setErrorMessage("Service ID is already known ("+intToStr(event.serviceID)+")=> FAIL");
    _onError(NS_ERROR_SERVICE_ALREADYCONNECTED);
    return;
  }

  ArRef<ArObject> perf;
  if (_viewerSetup.autoref)
    perf=thisRef();
  else
    perf=_viewerSetup.performer;

  if (!perf.valid())
  {
    setErrorMessage("Can't perfom viewer setup cause of an invalid ref to the specified performer");
    _onError(NS_ERROR_SETUPREMOTEVIEWER3D);
    return;
  }

  _infos[event.serviceID]=new ClientInfos();
  _infos[event.serviceID]->_viewer=Viewer3D::nullRef();
  _infos[event.serviceID]->_soundListener=Listener3D::nullRef();
  _infos[event.serviceID]->_viewerListener=NSListenerViewer3D::nullRef();

  ArRef<Viewer3D> viewer=((perf.c_ptr()->*(_viewerSetup.method))(event.serviceID));
  if (viewer.valid())
  {
    _infos[event.serviceID]->_viewer=viewer;
    _infos[event.serviceID]->_viewerListener=NSListenerViewer3D::NEW();
    _infos[event.serviceID]->_viewerListener->init(_depService,viewer);

  // Register to receive Viewer destroy event.
    _infos[event.serviceID]->_viewer->addDestroyCB(thisRef(),&LocalViewer3D::_onDestroyCB);
  }
  else
  {
    close(event.serviceID);
    if (_infos.find(event.serviceID)!=_infos.end())
    {
      delete _infos[event.serviceID];
      _infos.erase(_infos.find(event.serviceID));
    }
  }
}

void
LocalViewer3D::_onViewer3DDeconnect(const NetworkService::ServiceDeconnectEvent &event)
{
  if (_infos.find(event.serviceID)!=_infos.end())
  {
    if (_infos[event.serviceID]->_viewer.valid())
      _infos[event.serviceID]->_viewer.destroy(); // UGLY !

    if (_infos[event.serviceID]->_soundListener.valid())
      _infos[event.serviceID]->_soundListener.destroy(); // UGLY !

    if (_infos[event.serviceID]->_viewerListener.valid())
      _infos[event.serviceID]->_viewerListener.destroy(); // UGLY !

    delete _infos[event.serviceID];

    //exit (1);
    _infos.erase(_infos.find(event.serviceID));
  }
}

// Viewer destroyed => transmit to client
void
LocalViewer3D::_onDestroyCB(const ArObject::DestroyEvent &evt)
{
  int serviceId;
  if (_findServiceByViewer(evt.source->getAlias(), serviceId))
  {
    ArRef<NSEvent> event=NSEventWindow3DClose::NEW();
    sendEvent(event,serviceId);
    close(serviceId); // close connection
  }
}


bool
LocalViewer3D::_findServiceByViewer(const StlString &viewerAlias, int &serviceIdOut)
{
  StlMap<int, ClientInfos *>::const_iterator it=_infos.begin();
  StlMap<int, ClientInfos *>::const_iterator it_end=_infos.end();
  serviceIdOut=-1;

  while(it!=it_end)
  {
    if(it->second->_viewer->getAlias()==viewerAlias)
    {
      serviceIdOut=it->first;
      return true;
    }
    it++;
  }
  return false;
}

//----------------- Messages from client -------------------------------------
NSEvent::EventProcess
LocalViewer3D::_onGeometryEvent(ArConstRef<NSEvent> eventObj)
{
  ArRef<Viewer3D> viewer=accessViewer(eventObj->getServiceId());
  ArConstRef<NSEventWindow3DGeometry> event=ar_down_cast<NSEventWindow3DGeometry>(eventObj);


  unsigned int w,h;
  int x,y;
  event->getWindowGeometry(x,y,w,h);
  if (viewer.valid())
    viewer->setWindowGeometry(x,y,w,h);

  return NSEvent::EP_OK;
}

// On Keyboards events
NSEvent::EventProcess
LocalViewer3D::_onKeyboardEvent(ArConstRef<NSEvent> eventObj)
{
  ArConstRef<NSEventWindow3DKeyboard> event=ar_down_cast<NSEventWindow3DKeyboard>(eventObj);
  ArRef<Viewer3D> viewer=accessViewer(eventObj->getServiceId());

  if (viewer.valid())
    viewer->simulateKeyboard((Renderer3D::Modifier)event->modifiers(),
                             event->key(),event->pressed(),
                             event->x(),event->y());

  return NSEvent::EP_OK;
}

// On MOUSE events
NSEvent::EventProcess
LocalViewer3D::_onMouseMotionEvent(ArConstRef<NSEvent> eventObj)
{
   //cerr<<"_onMouseMotionEvent"<<endl;
  ArConstRef<NSEventWindow3DMouseMotion> event=ar_down_cast<NSEventWindow3DMouseMotion>(eventObj);
  ArRef<Viewer3D> viewer=accessViewer(eventObj->getServiceId());

  if (viewer.valid()){
    viewer->simulateMouseMotion((Renderer3D::Modifier)event->modifiers(),
                                event->x(),event->y());
    //cerr<<"_onMouseMotionEvent simulateMouseMotion"<<endl;
  }

  return NSEvent::EP_OK;
}

// On MOUSE Buttons events
NSEvent::EventProcess
LocalViewer3D::_onMouseButtonEvent(ArConstRef<NSEvent> eventObj)
{
  ArConstRef<NSEventWindow3DMouseButton> event=ar_down_cast<NSEventWindow3DMouseButton>(eventObj);
  ArRef<Viewer3D> viewer=accessViewer(eventObj->getServiceId());

  if (viewer.valid()){
    viewer->simulateMouseButton((Renderer3D::Modifier)event->modifiers(),
                                event->button(),event->pressed(),
                                event->x(),event->y());
  }

  return NSEvent::EP_OK;
}


// VIEWER CLOSE
NSEvent::EventProcess
LocalViewer3D::_onCloseEvent(ArConstRef<NSEvent> eventObj)
{
  ArRef<Viewer3D> viewer=accessViewer(eventObj->getServiceId());

  if (viewer.valid())
    viewer->simulateClose();

  return NSEvent::EP_OK;
}


bool //virtual
LocalViewer3D::_updateNetwork(ArRef<Activity> , double )
{
  static ArRef<NSEventControler> eventCtrl(NSEventControler::NEW());

  if(!_networkAct.valid())
    return false;

  StlMap<int, ClientInfos *>::const_iterator itCtrl=_infos.begin();
  StlMap<int, ClientInfos * >::const_iterator itCtrl_end=_infos.end();

  while(itCtrl!=itCtrl_end) // for each viewer, send events to connected client.
  {
    StlList<ArRef<NSEvent> > events;

    itCtrl->second->_viewerListener->getPendingEvents(events);
    if (events.size())
      {
      StlList<ArRef<NSEvent> >::iterator it=events.begin();
      StlList<ArRef<NSEvent> >::iterator it_end=events.end();

      eventCtrl->setID(itCtrl->second->_viewerListener->getAlias());

      while(it!=it_end)
        {
        eventCtrl->setEvent(*it);
        sendEvent(eventCtrl, itCtrl->first);
        it++;
        }
      }
    itCtrl++;
  }

  return _networkAct.valid();
}

} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
