// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/networkService.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/networkService.h"
#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/memoryStream.h"
#include "AReVi/Utils/tcpUdp.h"

//#include <zlib.h>


namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : NetworkService
//----------------------------------------------------------------------------
bool NetworkService::_compressData=false; // Usefull only for nbClient2
unsigned int NetworkService:: _timeOut=2000; // on connection, waiting for reply (ms)

AR_CLASS_NOVOID_DEF(NetworkService,ArObject)
//---------------------- Constructors / Destructor ---------------------------
NetworkService::NetworkService(ArCW & arCW) // Protected
  : ArObject(arCW),
    _eventsTable(),
    _connectionMessage(""),
    _sessionName("NoName"),
    _eventsReceived(),
    _eventsPosted(),
    _serviceDeconnectCB(thisPtr()),
    _networkServiceErrorCB(thisPtr()),
    _networkAct(),
    _services(),
    _serviceId(0)
{
  ArSystem::installSignalHandler(ArSystem::SIGNUM_PIPE,ArSystem::SIGHANDLER_IGNORE);

  /*if(!ArSystem::usesPlugin("ZLib"))
    cerr << "WARNING : ZLib is not loaded; it is recommended to use ZLib plugin with NetworkService" << endl;
  if(!ArSystem::usesPlugin("JpegLib"))
    cerr << "WARNING : JpegLib is not loaded; it is recommended to use JpegLib plugin with NetworkService" << endl;
*/
  registerEvent(NSEventPing::CLASS()->getName(),
                  thisRef(),
                  (NetworkService::EventMethod)&NetworkService::_onPingEvent);
  registerEvent(NSEventPingRequest::CLASS()->getName(),
                  thisRef(),
                  (NetworkService::EventMethod)&NetworkService::_onRequestPingEvent);
  registerEvent(NSEventSessionName::CLASS()->getName(),
                  thisRef(),
                  (NetworkService::EventMethod)&NetworkService::_onSessionNameEvent);

  Scheduler::access()->addPollingCB(thisRef(),&NetworkService::_onPolling);
  setUpdateInterval(0.01);
}

NetworkService::~NetworkService(void)
{
  _networkAct=Activity::nullRef();
  close();// close all services
}

//---------------------------- Public methods --------------------------------
// Close connection with specified service or all
bool //success
NetworkService::close(int serviceId)
{
  if (serviceId<0) // All services
  {
    StlMap<int,ServiceInfo>::iterator it=_services.begin();
    StlVector<int> closedServices;
    while(it!=_services.end())
    {
      if (it->second._connectionRead.valid())
        it->second._connectionRead.destroy(); // UGLY !

      if (it->second._connectionWrite.valid())
        it->second._connectionWrite.destroy(); // UGLY !

      closedServices.push_back(it->first);
      it++;
    }
    _services.clear();

    for (unsigned i=0;i<closedServices.size();i++)
      if(!_serviceDeconnectCB.empty())
        {
          NetworkService::ServiceDeconnectEvent event;
          event.serviceID=closedServices[i];
          _serviceDeconnectCB.fire(event);
        }
  }
  else
  {
    StlMap<int,ServiceInfo >::iterator it=_services.find(serviceId);
    if (it!=_services.end())
    {
      int serviceID=it->first;

      if (it->second._connectionRead.valid())
        it->second._connectionRead.destroy(); // UGLY !

      if (it->second._connectionWrite.valid())
        it->second._connectionWrite.destroy(); // UGLY !

      _services.erase(it);

      if(!_serviceDeconnectCB.empty())
      {
        NetworkService::ServiceDeconnectEvent event;
        event.serviceID=serviceID;
        _serviceDeconnectCB.fire(event);
      }
    }
    else
      return false;
  }

  return true;
}

unsigned int
NetworkService::getNbServices(void) const
{
  return _services.size();
}

void
NetworkService::getServices(StlVector<int> &servicesIdOut) const
{
  StlMap<int,ServiceInfo>::const_iterator it=_services.begin();
  while(it!=_services.end())
  {
    servicesIdOut.push_back(it->first);
    it++;
  }
}


void
NetworkService::setUpdateInterval(double dt)
{
  if (dt>=0.0)
  {
    if (_networkAct.valid())
      _networkAct->setInterval(dt);
    else
    {
      _networkAct=Activity::NEW(dt);
      _networkAct->setBehavior(thisRef(),&NetworkService::_updateNetwork);
    }
  }
  else
    _networkAct=Activity::nullRef();

}

// Set session name
void
NetworkService::setSessionName(const StlString &sessionName)
{
  _sessionName=sessionName;

  ArRef<NSEventSessionName> event=NSEventSessionName::NEW();
  event->setID(_sessionName);
  sendEvent(event);
}

// Get session name
const StlString &
NetworkService::getSessionName(void) const
{
  return _sessionName;
}

// Register and unregister events
bool // virtual
NetworkService::registerEvent(const StlString &eventClass,
              ArRef<ArObject> performer,
              EventMethod     method)
{
  if (_eventsTable.find(eventClass)!=_eventsTable.end())
    return false;

  if (performer.c_ptr()==thisRef().c_ptr())
  {
    _eventsTable[eventClass].autoref=true;
    _eventsTable[eventClass].performer=ArObject::nullRef();
  }
  else
  {
    _eventsTable[eventClass].autoref=false;
    _eventsTable[eventClass].performer=performer;
  }
  _eventsTable[eventClass].method=method;

  return true;
}

bool  // virtual
NetworkService::unregisterEvent(const StlString &eventClass)
{
  if (_eventsTable.find(eventClass)==_eventsTable.end())
    return false;

  _eventsTable.erase(_eventsTable.find(eventClass));
  return true;
}

bool
NetworkService::processEvents(void)
{
// First process more recent message received
  while (_eventsReceived.size())
  {
    ArRef<NSEvent> event=_eventsReceived.front();
    _eventsReceived.pop_front();

    if(event.valid())
    {
      StlMap<StlString, EventPerformer>::iterator it=_eventsTable.find(event->getClass()->getName());

      if (it==_eventsTable.end()) // no CB mapped => ignore
        continue;

      if (!it->second.performer && !it->second.autoref) // performer destroyed=>remove from table
        _eventsTable.erase(it);
      else
      {
        ArRef<ArObject> perf;
        if (it->second.autoref)
          perf=thisRef();
        else
          perf=it->second.performer;

        EventMethod method=it->second.method;
        NSEvent::EventProcess ret=((perf.c_ptr()->*method)(event));
        if (ret!=NSEvent::EP_OK)
        {
          if(ret==NSEvent::EP_POST)
          {
            _eventsPosted.push_back(event);
          }
          else // NSEvent::EP_FAIL => FIRE CB TO INFORM
          {
            _onError(NS_ERROR_PROCESSEVENT);
          }
        }
      }
    }
  }

// Process posted events (if all passed without any new processed ... exit)
  unsigned int nbTry=_eventsPosted.size();
  while(nbTry)
  {
    ArRef<NSEvent> event=_eventsPosted.front();
    _eventsPosted.pop_front();
    nbTry--;

    if(event.valid())
    {
      StlMap<StlString, EventPerformer>::iterator it=_eventsTable.find(event->getClass()->getName());

      if (it==_eventsTable.end()) // no CB mapped => ignore
        continue;

      if (!it->second.performer && !it->second.autoref) // performer destroyed=>remove from table
        _eventsTable.erase(it);
      else
      {
        ArRef<ArObject> perf;
        if (it->second.autoref)
          perf=thisRef();
        else
          perf=it->second.performer;

        EventMethod method=it->second.method;
        NSEvent::EventProcess ret=((perf.c_ptr()->*method)(event));
        if (ret!=NSEvent::EP_OK)
        {
          if(ret==NSEvent::EP_POST)
          {
          // push_back to posted.
            _eventsPosted.push_back(event);
          }
          else // NSEvent::EP_FAIL  || NSEvent::EP_LATER => FIRE CB TO INFORM
          {
            _onError(NS_ERROR_PROCESSEVENT);
          }
        }
        else
          nbTry=_eventsPosted.size();
      }
    }
  }

  return true;
}


// Add an event to send to services
bool
NetworkService::sendEvent(ArRef<NSEvent> event,int serviceId/*=-1*/)
{
  if (!_services.size()) // nothing to do
    return true;

 // Write event to memoryblock
  static ArRef<MemoryBlock> memory=MemoryBlock::NEW();

  if (serviceId<0) // All services
  {
  // If message must be sent more than once => compress data
    _makeMemoryBlockFromEvent(event,memory,_services.size()>2 && event->getCompress());

    StlMap<int,ServiceInfo>::iterator it=_services.begin();
    StlList<StlMap<int,ServiceInfo >::iterator> _servicesToErase;

    while(it!=_services.end())
    {
      it->second._dataWriteCount+=memory->getSize();
      it->second._connectionWrite->postBytes(memory,0,memory->getSize(),true);

      if (it->second._connectionWrite->pendingPost()&&
          !ArSystem::hasWriteFDPolling(it->second._connectionWrite
                                                 ->getSocketFD()))
        {
        ArSystem::addWriteFDPolling(it->second._connectionWrite
                                               ->getSocketFD());
        }

      if (it->second._connectionWrite->fail())
        _servicesToErase.push_back(it);

      it++;
    }

  // Erase failed connections
    for (StlList<StlMap<int,ServiceInfo>::iterator>::iterator itl=_servicesToErase.begin();
        itl!=_servicesToErase.end();
        itl++)
     close((*itl)->first);
  }
  else
  {
    _makeMemoryBlockFromEvent(event,memory,false); // don't compress point2point events.

    StlMap<int,ServiceInfo>::iterator it=_services.find(serviceId);
    if (it!=_services.end())
    {
      it->second._dataWriteCount+=memory->getSize();
      it->second._connectionWrite->postBytes(memory,0,memory->getSize(),true);
      if (it->second._connectionWrite->pendingPost()&&
          !ArSystem::hasWriteFDPolling(it->second._connectionWrite
                                                 ->getSocketFD()))
        {
        ArSystem::addWriteFDPolling(it->second._connectionWrite
                                               ->getSocketFD());
        }


      if (it->second._connectionWrite->fail())
        close(it->first);
    }
  }
  return true;
}

// Open a new connection into service
int //serviceID
NetworkService::_open(ArRef<TCPConnection> connectionIn,
                      ArRef<TCPConnection> connectionOut,
                      const StlString &sessionName,
                      const StlString &hostName)
{
  StlMap<int,ServiceInfo>::const_iterator it=_services.begin();
  while(it!=_services.end())
  {
    if(it->second._connectionRead==connectionIn ||
       it->second._connectionWrite==connectionOut)
      return -1;
    it++;
  }

  if (connectionIn->getSocketFD()!=-1)
    ArSystem::addReadFDPolling(connectionIn->getSocketFD());

  _services[_serviceId]._dataWriteCount=0;
  _services[_serviceId]._dataReadCount=0;
  _services[_serviceId]._connectionRead=connectionIn;
  _services[_serviceId]._connectionWrite=connectionOut;
  _services[_serviceId]._sessionName=sessionName;
  _services[_serviceId]._hostName=hostName;
  return _serviceId++;
}


// Port number of service.
void
NetworkService::_getPortNumber(int serviceId, int &inputPort, int &outputPort) const
{
StlMap<int,ServiceInfo>::const_iterator it=_services.find(serviceId);
if (it!=_services.end())
  {
  inputPort=it->second._connectionRead->getPortNumber();
  outputPort=it->second._connectionWrite->getPortNumber();
  }
else
  {
  inputPort=-1;
  outputPort=-1;
  }
}

// Port number of remote service.
void
NetworkService::_getDistantPortNumber(int serviceId, int &inputPort, int &outputPort) const
{
StlMap<int,ServiceInfo>::const_iterator it=_services.find(serviceId);
if (it!=_services.end())
  {
  inputPort=it->second._connectionRead->getDistantPortNumber();
  outputPort=it->second._connectionWrite->getDistantPortNumber();
  }
else
  {
  inputPort=-1;
  outputPort=-1;
  }
}

// IP address of remote service.
unsigned int
NetworkService::_getDistantIPAdress(int serviceId) const
{
  StlMap<int,ServiceInfo>::const_iterator it=_services.find(serviceId);
  if (it!=_services.end())
    return it->second._connectionRead->getDistantIPAddress();
  else
    return 0;
}

// Set Connection message
void
NetworkService::_setConnectionMessage(const StlString &connectionMessage)
{
  _connectionMessage=connectionMessage;
}

// Service session name
StlString // "" on error
NetworkService::_getServiceSessionName(int serviceId) const
{
  StlMap<int,ServiceInfo>::const_iterator it=_services.find(serviceId);
  if (it!=_services.end())
    return it->second._sessionName;
  else
    return "";
}

// Get service host name
StlString
NetworkService::_getServiceHostName(int serviceId) const
{
  StlMap<int,ServiceInfo>::const_iterator it=_services.find(serviceId);
  if (it!=_services.end())
    return it->second._hostName;
  else
    return "";
}


// Dispatch events for this service
ArRef<NSEvent>
NetworkService::_makeEventFromMemoryBlock(ArConstRef<MemoryBlock> memory,
                                          int serviceId)
{
// Look if is a compressed block
  ArRef<MemoryIStream> stream=MemoryIStream::NEW(memory,1);
  ArRef<MemoryBlock> buffer;
  char compressed=memory->getMemory()[0];

  if (compressed)
  {
    unsigned int uncompressedSize,compressedSize;
    ArRef<MemoryBlock> cBuffer=MemoryBlock::NEW();
    ArRef<MemoryBlock> buffer=MemoryBlock::NEW();

    stream->readUInt(uncompressedSize); // uncompressed size
    stream->readUInt(compressedSize); // compressed size

    //cerr << "COMPRESSED BLOCK USIZE=" << uncompressedSize << "    CSIZE=" << compressedSize << endl;
    cBuffer->setSize(compressedSize);
    buffer->setSize(uncompressedSize);
    stream->readData(cBuffer->accessMemory(),compressedSize);

    if (!buffer->uncompressFrom(cBuffer))
    {
      cerr << getAlias() << "FATAL ERROR : ON DECOMPRESS EVENT (ZLIB ERROR)=>EXIT" << endl << buffer->getErrorMessage() << endl;
      exit(1);
    }
    stream=MemoryIStream::NEW(buffer,0);
  }


// Read Event
  StlString eventType;
  ArRef<NSEvent> event;
  if(stream->readString(eventType))
  {
    ArConstRef<ArClass> eventClass=ArClass::find(eventType);

    if(eventClass)
      eventClass->newInstance(event);
    else // Event type unknown
    {
      setErrorMessage("Unable to find eventClass of type '" + eventType + "'");
      _onError(NS_ERROR_UNKNOWEVENT);

      return NSEvent::nullRef();
    }

    if(event)
    {
      event->_setNetworkService(thisRef(),serviceId);
      if (!event->readFromStream(stream))
      {
        setErrorMessage("Error extracting event infos for " + event->getAlias()+"\n    =>"+event->getErrorMessage());
        _onError(NS_ERROR_INVALIDEVENT);
        return NSEvent::nullRef();
      }
    }
    else
    {
      setErrorMessage("Unable to instanciate event of type '" + eventType + "' ... need default constructor");
      _onError(NS_ERROR_INSTANCIATEEVENT);

      return NSEvent::nullRef();
    }
  }
  else
  {
    cerr << "NETWORK SERVICE FALTAL ERROR : INVALID MESSAGE!! => EXIT" <<  endl;
    exit(1);
  }

  return event;
}

bool
NetworkService::_makeMemoryBlockFromEvent(ArRef<NSEvent> event,
                                           ArRef<MemoryBlock> memory,
                                           bool compressFlag/*=true*/)
{
  // Get event memory
  ArRef<MemoryOStream> stream=MemoryOStream::NEW(memory,1);
  stream->writeString(event->getClass()->getName());
  event->writeToStream(stream);
  stream->flush();
  memory->setSize(stream->getMemoryUsage()+1);

  //compressFlag=false;
  if (NetworkService::_compressData && compressFlag && memory->getSize()>1024) // block size must worth a compression!
  {
  // compress memory
    unsigned int uncompressedSize(memory->getSize());
    unsigned int compressedSize(uncompressedSize);

    ArRef<MemoryBlock> cBuffer=MemoryBlock::NEW();
    cBuffer->setSize((unsigned)(uncompressedSize*1.2));

    ArRef<RigidMemoryBlock> src=RigidMemoryBlock::NEW(memory->accessMemory()+1,memory->getSize()-1,false);
    if (cBuffer->compressFrom(src,1))
    {
      compressedSize=cBuffer->getSize();

      ArRef<MemoryOStream> compressedStream=MemoryOStream::NEW(memory,1);
      compressedStream->writeUInt(uncompressedSize); // uncompressed size
      compressedStream->writeUInt(compressedSize); // compressed size
      compressedStream->writeData(cBuffer->getMemory(),compressedSize);
      compressedStream->flush();
      memory->setSize(compressedStream->getMemoryUsage()+1);
      memory->accessMemory()[0]=1;

      //cerr << "ratio=" << compressedSize / (double) uncompressedSize << endl;
    }
    else // not compression available => send uncompressed block.
      memory->accessMemory()[0]=0;
  }
  else
    memory->accessMemory()[0]=0;

  return true;
}

// write to network (write messages).
bool
NetworkService::_setNetworkEvents(void)
{
  StlMap<int,ServiceInfo>::iterator it=_services.begin();
  StlList<StlMap<int,ServiceInfo >::iterator> _servicesToErase;

  while(it!=_services.end())
  {
    it->second._connectionWrite->pushPost();

    if (!it->second._connectionWrite->pendingPost())
      ArSystem::removeWriteFDPolling(it->second._connectionWrite
                                                ->getSocketFD());

    if (it->second._connectionWrite->fail())
      _servicesToErase.push_back(it);
    it++;
  }

// Erase failed connections
  for (StlList<StlMap<int,ServiceInfo>::iterator>::iterator itl=_servicesToErase.begin();
      itl!=_servicesToErase.end();
      itl++)
    close((*itl)->first);


  return true;
}

// watch network activity
// call _watchNetworkCB while return true.
bool
NetworkService::_getNetworkEvents(void)
{
  static ArRef<MemoryBlock> memory=MemoryBlock::NEW();

// Read events from services
  StlMap<int,ServiceInfo >::iterator it=_services.begin();
  StlMap<int,ServiceInfo >::iterator last_it=_services.end();
  StlList<StlMap<int,ServiceInfo >::iterator> _servicesToErase;

  while(it!=last_it)
  {
    while(it->second._connectionRead->readBytes(memory))
    {
      it->second._dataReadCount+=memory->getSize();
      ArRef<NSEvent> event(_makeEventFromMemoryBlock(memory,it->first));
      if (event)
        _eventsReceived.push_back(event);
    }

    if (it->second._connectionRead->fail())
      _servicesToErase.push_back(it);

    it++;
  }

  for (StlList<StlMap<int,ServiceInfo >::iterator>::iterator itl=_servicesToErase.begin();
        itl!=_servicesToErase.end();
        itl++)
     close((*itl)->first);

  return true;
}

bool
NetworkService::_updateNetwork(ArRef<Activity>, double)
{
  return _networkAct.valid();
}

void
NetworkService::_onPolling(const Scheduler::PollingEvent &evt)
{
// Write pending events
if (evt.writingFD.size() && !_setNetworkEvents())
  _networkAct.clear();

// Read pending events
if (evt.readingFD.size() && !_getNetworkEvents())
    _networkAct.clear();

if (!processEvents())
  _networkAct.clear();
}

// Connection method
ArRef<TCPConnection>
NetworkService::_connect(unsigned int distantIPAddress,
                         int           distantPort)
{
  ArRef<TCPConnection> connection=(TCPConnection::NEW(distantIPAddress,distantPort,true));
  if(connection->fail())
  {
    setErrorMessage("connect failed : " + connection->getErrorMessage());
    connection.destroy(); // UGLY !
    return TCPConnection::nullRef();
  }
  else
  {
    connection->writeString(NS_REQUEST_CONNECTION_MESSAGE+_connectionMessage);
    if (connection->fail())
    {
      setErrorMessage("connect request failed : " + connection->getErrorMessage());
      return TCPConnection::nullRef();
    }

    // Waiting connection confirmation
    StlString message;
    unsigned int t0=ArSystem::getTime();
    while(!connection->fail()&&!connection->readString(message))
    {
      if ((ArSystem::getTime()-t0)>NetworkService::getTimeOut())
      {
        setErrorMessage("connect request failed : TIMEOUT");
        return TCPConnection::nullRef();
      }
      ArSystem::yield();
    }

    if (connection->fail())
    {
      setErrorMessage("connect request failed : " + connection->getErrorMessage());
      return TCPConnection::nullRef();
    }
    if (message!=_connectionMessage)
    {
      setErrorMessage("connect request failed cause of bad response. Received='"+
                      message+"' Expected='"+_connectionMessage+"'");
      return TCPConnection::nullRef();
    }
    return connection;
  }
}


// Get next valid connection request
ArRef<TCPConnection>
NetworkService::_acceptConnection(ArRef<TCPListener> listener, const StlString &requestedMessage)
{
if (!listener.valid())
  return TCPConnection::nullRef(); // no TCPListenner


StlString message;

// Watch for new connections
ArRef<TCPConnection> connection=listener->nextConnection();
while(connection)
  {
// Waiting for message
  unsigned int t0=ArSystem::getTime();
  while(!connection->fail()&&!connection->readString(message)&&(ArSystem::getTime()-t0)<NetworkService::getTimeOut())
    {
    ArSystem::yield();
    }

  if(!connection->fail() && (message==requestedMessage) && (ArSystem::getTime()-t0<NetworkService::getTimeOut()))
    {
    connection->writeString(_connectionMessage); // Accept connection.
    if(!connection->fail())
      break;
    }
  connection=listener->nextConnection();
  }
return connection;
}

//------------------------- Network Statistics ------------------------------

// Read data from this service
unsigned int
NetworkService::_getDataReadCount(int serviceId) const
{
  StlMap<int,ServiceInfo >::const_iterator it=_services.find(serviceId);
  if (it!=_services.end())
    return it->second._dataReadCount;
  else
    return 0;
}

// Write data to this service
unsigned int
NetworkService::_getDataWriteCount(int serviceId) const
{
  StlMap<int,ServiceInfo>::const_iterator it=_services.find(serviceId);
  if (it!=_services.end())
    return it->second._dataWriteCount;
  else
    return 0;
}

// Ping specified service (or all with -1)
bool
NetworkService::_ping(int serviceId)
{
  if (serviceId<0)
  {
    StlMap<int,ServiceInfo>::iterator it=_services.begin();
    StlList<int> _servicesToPing;

    while(it!=_services.end())
    {
      if (_pingTimes.find(it->first)==_pingTimes.end()) // not yet requested
        _servicesToPing.push_back(it->first);
      it++;
    }
    for (StlList<int>::iterator itl=_servicesToPing.begin();
        itl!=_servicesToPing.end();
        itl++)
     _ping(*itl);
  }
  else if (_pingTimes.find(serviceId)==_pingTimes.end()) // not yet requested
  {
    ArRef<NSEventPingRequest> pingEvent=NSEventPingRequest::NEW();
    if (sendEvent(pingEvent,serviceId))
      _pingTimes[serviceId]=ArSystem::getTime();
    else
      return false;
  }
  return true;
}

// CB : reception of a pingRequest Event
NSEvent::EventProcess
NetworkService::_onRequestPingEvent(ArConstRef<NSEvent> eventObj)
{
  ArRef<NSEventPing> pingEvent=NSEventPing::NEW();
  sendEvent(pingEvent,eventObj->getServiceId());
  return NSEvent::EP_OK;
}

// CB : reception of a ping Event
NSEvent::EventProcess
NetworkService::_onPingEvent(ArConstRef<NSEvent> eventObj)
{
  // Just stop "waiting" reply
  _pingTimes.erase(_pingTimes.find(eventObj->getServiceId()));
  return NSEvent::EP_OK;
}

// CB : notify distant session name changed
NSEvent::EventProcess
NetworkService::_onSessionNameEvent(ArConstRef<NSEvent> eventObj)
{
  ArConstRef<NSEventSessionName> event=ar_down_cast<NSEventSessionName>(eventObj);

  StlMap<int,ServiceInfo>::iterator it=_services.find(event->getServiceId());
  if (it!=_services.end())
    it->second._sessionName=event->getID();

  return NSEvent::EP_OK;
}


//! handle error
void
NetworkService::_onError(unsigned int errorId)
{
  if(!_networkServiceErrorCB.empty())
  {
    NetworkService::NetworkServiceErrorEvent notifyEvent;
    notifyEvent.errorID=errorId;
    _networkServiceErrorCB.fire(notifyEvent);
  }
  else // else display
    cerr << getErrorMessage() << endl;
}

//-------- Static attributes -------------------------------------------------

bool //static
NetworkService::getCompressData(void)
{
  return _compressData;
}

void //static
NetworkService::setCompressData(bool compressData)
{
  _compressData=compressData;
}

unsigned int //static
NetworkService::getTimeOut(void)
{
  return _timeOut;
}

void //static
NetworkService::setTimeOut(unsigned int timeOut)
{
  _timeOut=timeOut;
}

} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
