// LI2/ENIB - AReVi - Network services
//----------------------------------------------------------------------------
// file        : AReViNS/remoteService.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReViNS/remoteService.h"
#include "AReViNS/Events/nsEventObjectRef.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"
#include "AReVi/Utils/memoryStream.h"
#include "AReVi/Utils/tcpUdp.h"

namespace AReVi {
//----------------------------------------------------------------------------
// CLASS       : RemoteService
//----------------------------------------------------------------------------
AR_CLASS_DEF(RemoteService,NetworkService)
//---------------------- Constructors / Destructor ---------------------------
RemoteService::RemoteService(ArCW & arCW) // Protected
  : NetworkService(arCW),
  _pingCB(thisPtr())
{}

RemoteService::~RemoteService(void)
{
}
//---------------------------- Public methods --------------------------------
// Connection method Remote Side
// 1 - Remote connect to local service (write)
// 2 - Remote open a listener waiting to local service connection (read)
// 3 - Remote send to local service listening port number.
// 4 - Remote accept (or not) local connection
// 5 - Send session name
// 6 - Wait for session name
// 7 - Finalize connection
// Connection method Local Side
// 1 - Local open a listener waiting to remote service connection
// 2 - Local accept (or not) remote connection (read connection)
// 3 - Local get remote port number listener
// 4 - Local connect to remote service (write)
// 5 - Wait session name
// 6 - Send session name
// 7 - Finalize connection


bool
RemoteService::connect(unsigned int distantIPAddress,
                       int           distantPort)
{
  if (getNbServices()) // already connected!
    return false;

// Request connection
  ArRef<TCPConnection> connectionWrite=_connect(distantIPAddress,distantPort);
  if (!connectionWrite.valid() || connectionWrite->fail())
    return false;

// Now create a listener to accept second connection
  ArRef<TCPListener> listener=TCPListener::NEW(0,true);

  if (connectionWrite->fail())
    return false;
  connectionWrite->writeString(intToStr(listener->getPortNumber()));

// Now wait for second connection request
  unsigned int t0=ArSystem::getTime();
  ArRef<TCPConnection> connectionRead;
  do
  {
  connectionRead=_acceptConnection(listener,NS_REQUEST_CONNECTION_MESSAGE+_connectionMessage);
  }
  while((ArSystem::getTime()-t0)<NetworkService::getTimeOut() && !connectionRead.valid());

  if (!connectionRead.valid() && (ArSystem::getTime()-t0)>NetworkService::getTimeOut())
  {
    setErrorMessage("connect request failed (write connection): TIMEOUT");
    return false;
  }
  if (connectionRead->fail())
  {
    setErrorMessage("connect request failed (write connection): " + connectionRead->getErrorMessage());
    return false;
  }

// Send session name
  if (connectionWrite->fail())
    return false;
  connectionWrite->writeString(_sessionName+"@"+ArSystem::getHostName());

// Waiting for session Name
  StlString sessionName,hostName;
  t0=ArSystem::getTime();
  while(!connectionRead->fail()&&!connectionRead->readString(sessionName))
  {
    if ((ArSystem::getTime()-t0)>NetworkService::getTimeOut())
    {
      setErrorMessage("connect request failed : TIMEOUT");
      return false;
    }
    ArSystem::yield();
  }

  const char * at=strchr(sessionName.c_str(),'@');
  if(at)
  {
    hostName=StlString(at+1);
    sessionName=StlString(sessionName,0,(at-sessionName.c_str()));
  }
  else
    hostName="";


// Finalize
  _open(connectionRead,connectionWrite,sessionName,hostName);
  return true;
}

// Port number of service.
void
RemoteService::getPortNumber(int &readPort, int &writePort) const
{
StlVector<int> services;
getServices(services);
if (services.size())
  NetworkService::_getPortNumber(services[0],readPort,writePort);
else
  {
  readPort=-1;
  writePort=-1;
  }
}

// Server session name
StlString  // "" on error
RemoteService::getServerName(void) const
{
  StlVector<int> services;
  getServices(services);
  if (services.size())
    return NetworkService::_getServiceSessionName(services[0]);
  else
    return "";
}

// Server host name
StlString  // "" on error
RemoteService::getServerHostName(void) const
{
StlVector<int> services;
getServices(services);
if (services.size())
  return NetworkService::_getServiceHostName(services[0]);
else
  return "";
}

// Read data from this service
unsigned int
RemoteService::getDataReadCount(void) const
{
  StlVector<int> services;
  getServices(services);
  if (services.size())
    return NetworkService::_getDataReadCount(services[0]);
  else
    return 0;
}

// Write data to this service
unsigned int
RemoteService::getDataWriteCount(void) const
{
    StlVector<int> services;
  getServices(services);
  if (services.size())
    return NetworkService::_getDataWriteCount(services[0]);
  else
    return 0;
}

// Ping server
bool
RemoteService::ping(void)
{
    StlVector<int> services;
  getServices(services);
  if (services.size())
    return NetworkService::_ping(services[0]);
  else
    return false;
}


// CB : reception of a ping Event
NSEvent::EventProcess
RemoteService::_onPingEvent(ArConstRef<NSEvent> eventObj)
{
  if(!_pingCB.empty())
  {
    RemoteService::PingEvent event;
    event.dt=ArSystem::getTime()-_pingTimes[eventObj->getServiceId()];
    _pingCB.fire(event);
  }

  return NetworkService::_onPingEvent(eventObj);
}


} // namespace NetworkServices
//---End-Of-File---------------------------------------------------------------
