// LI2/ENIB - AReVi - Image loader plugin
//----------------------------------------------------------------------------
// file        : Ext/MagickImageLoader/magickImageLoader.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This plugin uses the ImageMagick API (http://www.imagemagick.org).
//
//  The resulting dynamic library can be loaded at run-time
//  (fails if a dependency problem occurs).
//
//  These functions uses a `width x height x components' byte array
//  with rows ordered from bottom to top.
//
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/llw_priv.h"

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <magick/api.h>

namespace AReVi {

//----------------------------------------------------------------------------

static
bool // success (dataOut filled from bottom to top)
Magick_imageLoader(const char * fileName,
                   unsigned int & widthOut,
                   unsigned int & heightOut,
                   unsigned int & componentsOut,
                   ArRef<MemoryBlock> dataOut)
{
bool result=false;
ExceptionInfo exception;
GetExceptionInfo(&exception);
ImageInfo * imageInfo=CloneImageInfo((ImageInfo *)0);
strcpy(imageInfo->filename,fileName);
Image * image=ReadImage(imageInfo,&exception);
if(image)
  {
  widthOut=image->columns;
  heightOut=image->rows;
  if(IsGrayImage(image,&exception))
    {
    if(IsOpaqueImage(image,&exception))
      {
      componentsOut=1; // GRAY
      }
    else
      {
      componentsOut=2; // GRAY - ALPHA
      }
    }
  else
    {
    if(IsOpaqueImage(image,&exception))
      {
      componentsOut=3; // RED - GREEN - BLUE
      }
    else
      {
      componentsOut=4; // RED - GREEN - BLUE - ALPHA
      }
    }
  if(dataOut->ensureMinSize(componentsOut*widthOut*heightOut))
    {
    char * ptr=dataOut->accessMemory();
    switch(componentsOut)
      {
      case 1: // GRAY
        {
        for(unsigned int y=heightOut;y--;)
          {
          int h=(int)(heightOut-1-y);
          const PixelPacket * pixels=AcquireImagePixels(image,0,h,widthOut,1,
                                                        &exception);
          if(!pixels)
            {
            goto pixelProblem;
            }
          for(unsigned int x=widthOut;x--;)
            {
            unsigned int index=x+(widthOut*y);
            ptr[index]=Downscale(pixels[x].red);
            }
          }
        break;
        }
      case 2: // GRAY - ALPHA
        {
        for(unsigned int y=heightOut;y--;)
          {
          int h=(int)(heightOut-1-y);
          const PixelPacket * pixels=AcquireImagePixels(image,0,h,widthOut,1,
                                                        &exception);
          if(!pixels)
            {
            goto pixelProblem;
            }
          for(unsigned int x=widthOut;x--;)
            {
            unsigned int index=2*(x+(widthOut*y));
            ptr[index]=Downscale(pixels[x].red);
            ptr[index+1]=255-Downscale(pixels[x].opacity);
            }
          }
        break;
        }
      case 3: // RED - GREEN - BLUE
        {
        for(unsigned int y=heightOut;y--;)
          {
          int h=(int)(heightOut-1-y);
          const PixelPacket * pixels=AcquireImagePixels(image,0,h,widthOut,1,
                                                        &exception);
          if(!pixels)
            {
            goto pixelProblem;
            }
          for(unsigned int x=widthOut;x--;)
            {
            unsigned int index=3*(x+(widthOut*y));
            ptr[index]=Downscale(pixels[x].red);
            ptr[index+1]=Downscale(pixels[x].green);
            ptr[index+2]=Downscale(pixels[x].blue);
            }
          }
        break;
        }
      case 4: // RED - GREEN - BLUE - ALPHA
        {
        for(unsigned int y=heightOut;y--;)
          {
          int h=(int)(heightOut-1-y);
          const PixelPacket * pixels=AcquireImagePixels(image,0,h,widthOut,1,
                                                        &exception);
          if(!pixels)
            {
            goto pixelProblem;
            }
          for(unsigned int x=widthOut;x--;)
            {
            unsigned int index=4*(x+(widthOut*y));
            ptr[index]=Downscale(pixels[x].red);
            ptr[index+1]=Downscale(pixels[x].green);
            ptr[index+2]=Downscale(pixels[x].blue);
            ptr[index+3]=255-Downscale(pixels[x].opacity);
            }
          }
        break;
        }
      }
    result=true;
    pixelProblem:; // problem while reading pixels
    }
  DestroyImage(image);
  }
DestroyImageInfo(imageInfo);
DestroyExceptionInfo(&exception);
return(result);
}

static
bool // success
Magick_imageSaver(const char * fileName,
                  unsigned int width,
                  unsigned int height,
                  unsigned int components,
                  const char * data)
{
const char * comp;
switch(components)
  {
  case 1:
    comp="I";
    break;
  case 2:
    comp="IA";
    break;
  case 3:
    comp="RGB";
    break;
  case 4:
    comp="RGBA";
    break;
  default:
    return(false);
  }
ExceptionInfo exception;
GetExceptionInfo(&exception);
Image * image=ConstituteImage(width,height,comp,CharPixel,data,&exception);
bool result=false;
if(image)
  {
  Image * flip=FlipImage(image,&exception);
  if(flip)
    {
    ImageInfo * imageInfo=CloneImageInfo((ImageInfo *)0);
    strcpy(flip->filename,fileName);
    result=WriteImage(imageInfo,flip);
    DestroyImageInfo(imageInfo);
    DestroyImage(flip);
    }
  DestroyImage(image);
  }
return(result);
}

//-------- Initialization ----------------------------------------------------

bool // success
ImageLoader_init(const StlVector<StlString> * /* params */)
{
// FIXME: initialisation causes a SEGV on exit !
// InitializeMagick("AReViMagickImageLoader");
LLW::_imageLoaderFncts->push_back(&Magick_imageLoader);
LLW::_imageSaverFncts->push_back(&Magick_imageSaver);
return(true);
}

extern "C"
bool // success
MagickImageLoader_init(const StlVector<StlString> * params)
{
return(ImageLoader_init(params)); // friend of LLW
}

void
ImageLoader_destroy(void)
{
unsigned int i;
for(i=LLW::_imageLoaderFncts->size();i--;)
  {
  LLW::ImageLoaderFnct fnct=(*LLW::_imageLoaderFncts)[i];
  if(fnct==&Magick_imageLoader)
    {
    StlVectorFastErase(*LLW::_imageLoaderFncts,i);
    break;
    }
  }
for(i=LLW::_imageSaverFncts->size();i--;)
  {
  LLW::ImageSaverFnct fnct=(*LLW::_imageSaverFncts)[i];
  if(fnct==&Magick_imageSaver)
    {
    StlVectorFastErase(*LLW::_imageSaverFncts,i);
    break;
    }
  }
DestroyMagick();
}

extern "C"
void
MagickImageLoader_destroy(void)
{
ImageLoader_destroy(); // friend of LLW
}

} // namespace AReVi

//----------------------------------------------------------------------------
