// LI2/ENIB - AReVi - 3D sound plugin
//----------------------------------------------------------------------------
// file        : Ext/Sound3DPlayer/sound3DPlayer.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This plugin uses the OpenAL API (http://www.openal.org).
//
//  The resulting dynamic library can be loaded at run-time
//  (fails if a dependency problem occurs).
//
//----------------------------------------------------------------------------

#include "AReVi/Lib3D/urlSound3D.h"

#include "AReVi/arSystem.h"

#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alut.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>


namespace AReVi {

//-------- Static data -------------------------------------------------------

static ALCdevice * Sound3D_device=(ALCdevice *)0;

static ALCcontext * Sound3D_globalContext=(ALCcontext *)0;

static ALboolean (*vorbisLoader_ext)(ALuint,ALvoid *,ALint);
static void (*reverbScale_ext)(ALuint,ALfloat);
static void (*reverbDelay_ext)(ALuint,ALfloat);

//-------- Buffer management -------------------------------------------------

static
int // buffer ID or -1
Sound3D_loadBuffer(const char * fileName)
{
unsigned int bufferID=AL_NONE;
bufferID=alutCreateBufferFromFile(fileName);
if((bufferID==AL_NONE)&&vorbisLoader_ext)
  {
  struct stat infos;
  if(stat(fileName,&infos)!=-1)
    {
    FILE * input=fopen(fileName,"rb");
    if(input)
      {
      ALvoid * data=ArSystem::allocate<char>(infos.st_size);
      if(fread(data,infos.st_size,1,input)==1)
        {
        alGenBuffers(1,&bufferID);
        if((*vorbisLoader_ext)(bufferID,data,infos.st_size)!=AL_TRUE)
          {
          alDeleteBuffers(1,&bufferID);
          bufferID=(unsigned int)-1;
          }
        }
      ArSystem::deallocate(data);
      fclose(input);
      }
    }
  }
alGetError(); // clear error state
alutGetError(); // clear error state
return(bufferID==AL_NONE ? -1 : (int)bufferID);
}

static
int // buffer ID or -1
Sound3D_createBufferFromMemory(const char * data,
                               unsigned int size,
                               unsigned int freq,
                               UtilSound::SoundFormat format)
{
unsigned int bufferID=(unsigned int)-1;
alGenBuffers(1,&bufferID);
int alFormat;
switch(format)
  {
  case UtilSound::MONO8:
    alFormat=AL_FORMAT_MONO8;
    break;
  case UtilSound::MONO16:
    alFormat=AL_FORMAT_MONO16;
    break;
  case UtilSound::STEREO8:
    alFormat=AL_FORMAT_STEREO8;
    break;
  case UtilSound::STEREO16:
    alFormat=AL_FORMAT_STEREO16;
    break;
  default:
    return(-1);
  }
alBufferData(bufferID,alFormat,(ALvoid *)data,size,freq);
return(bufferID);
}

static
void
Sound3D_freeBuffer(int bufferID)
{
alDeleteBuffers(1,reinterpret_cast<unsigned int *>(&bufferID));
}

//-------- Source management -------------------------------------------------

static
int // source ID
Sound3D_createSource(void)
{
unsigned int sourceID;
alGenSources(1,&sourceID);
alSourcef(sourceID,AL_MIN_GAIN,0.0);
return(sourceID);
}

static
void
Sound3D_destroySource(int sourceID)
{
alDeleteSources(1,reinterpret_cast<unsigned int *>(&sourceID));
}

static
void
Sound3D_setSourceBuffer(int sourceID,
                        int bufferID)
{
alSourcei(sourceID,AL_BUFFER,bufferID);
/***
if(reverbScale_ext&&reverbDelay_ext)
  {
  (*reverbScale_ext)(sourceID,0.35);
  (*reverbDelay_ext)(sourceID,1.0);
  }
***/
}

static
void
Sound3D_addSourceBuffer(int sourceID,
                        int bufferID)
{
unsigned int buffers=bufferID;
alSourceQueueBuffers(sourceID,1,&buffers);
}


static
void
Sound3D_getSourceBuffersState(int sourceID,
                              int & nbBuffersOut,
                              int & nbProcessedOut)
{
alGetSourcei(sourceID,AL_BUFFERS_QUEUED,&nbBuffersOut);
alGetSourcei(sourceID,AL_BUFFERS_PROCESSED,&nbProcessedOut);
}

static
void
Sound3D_removeSourceBuffer(int sourceID,
                           int nbBuffers,
                           StlVector<unsigned int> & buffers)
{
buffers.resize(nbBuffers);
alSourceUnqueueBuffers(sourceID , nbBuffers, &buffers[0]);
// user must free buffers!
}

static
void
Sound3D_setSourceGain(int sourceID,
                      double gain)
{
if(gain)
  {
  alSourcef(sourceID,AL_GAIN,gain);
  alSourcef(sourceID,AL_MAX_GAIN,1.0);
  }
else
  {
  alSourcef(sourceID,AL_GAIN,0.0);
  alSourcef(sourceID,AL_MAX_GAIN,0.0);
  }
}

static
void
Sound3D_setSourcePitch(int sourceID,
                       double pitch)
{
alSourcef(sourceID,AL_PITCH,pitch);
}

static
void
Sound3D_setSourceRepeat(int sourceID,
                        bool repeat)
{
alSourcei(sourceID,AL_LOOPING,repeat ? AL_TRUE : AL_FALSE);
}

static
void
Sound3D_setSourceRelative(int sourceID,
                          bool relative)
{
alSourcei(sourceID,AL_SOURCE_RELATIVE,relative ? AL_TRUE : AL_FALSE);
}

static
void
Sound3D_setSourcePosition(int sourceID,
                          float pos[3])
{
alSourcefv(sourceID,AL_POSITION,pos);
}

static
void
Sound3D_setSourceVelocity(int sourceID,
                          float vel[3])
{
alSourcefv(sourceID,AL_VELOCITY,vel);
}

static
void
Sound3D_playSources(int * sourceIDs,
                    unsigned int nbSources)
{
alSourcePlayv(nbSources,reinterpret_cast<unsigned int *>(sourceIDs));
}

static
void
Sound3D_pauseSources(int * sourceIDs,
                     unsigned int nbSources)
{
alSourcePausev(nbSources,reinterpret_cast<unsigned int *>(sourceIDs));
}

static
void
Sound3D_stopSources(int * sourceIDs,
                    unsigned int nbSources)
{
alSourceStopv(nbSources,reinterpret_cast<unsigned int *>(sourceIDs));
}

static
int // -1:stop  0: pause  1: play
Sound3D_getSourceState(int sourceID)
{
int state;
alGetSourcei(sourceID,AL_SOURCE_STATE,&state);
switch(state)
  {
  case AL_PAUSED:
    return(0);
  case AL_PLAYING:
    return(1);
  default:
    return(-1);
  }
}

//-------- Listener management -----------------------------------------------

static
void
Sound3D_setListenerPosition(float pos[3])
{
alListenerfv(AL_POSITION,pos);
}

static
void
Sound3D_setListenerVelocity(float vel[3])
{
alListenerfv(AL_VELOCITY,vel);
}

static
void
Sound3D_setListenerOrientation(float localXandZ[6])
{
alListenerfv(AL_ORIENTATION,localXandZ);
}

static
void
Sound3D_setListenerGain(double gain)
{
alListenerf(AL_GAIN,gain);
}

//-------- Initialization ----------------------------------------------------

extern "C"
bool // success
Sound3DPlayer_init(const StlVector<StlString> * params)
{
Sound3D::_initSuccess=false;

StlString openArgs;
if(params&&!params->empty())
  {
  for(unsigned int i=0;i<params->size();i++)
    {
    if(i) openArgs+=" ";
    openArgs+=(*params)[i];
    }
  }
else
  {
  openArgs="esd native";
  }
openArgs="'( ( devices '( "+openArgs+" ) ) )";

// Probe for esd availability
Sound3D_device=alcOpenDevice(openArgs.c_str());
if(!Sound3D_device)
  {
  return(false);
  }

Sound3D_globalContext=alcCreateContext(Sound3D_device,(ALint *)0);
if(!Sound3D_globalContext)
  {
  alcCloseDevice(Sound3D_device);
  Sound3D_globalContext=(ALCcontext *)0;
  return(false);
  }

alcMakeContextCurrent(Sound3D_globalContext);
alDistanceModel(AL_NONE);
void * symbol=alGetProcAddress("alutLoadVorbis_LOKI");
memcpy(&vorbisLoader_ext,&symbol,sizeof(void *));
symbol=alGetProcAddress("alReverbScale_LOKI");
memcpy(&reverbScale_ext,&symbol,sizeof(void *));
symbol=alGetProcAddress("alReverbDelay_LOKI");
memcpy(&reverbDelay_ext,&symbol,sizeof(void *));
Sound3D::_currentListener.clear();
//Sound3D::_buffers=new StlVector<Sound3D::BufferInfo>;
Sound3D::_loadBufferFnct=&Sound3D_loadBuffer;
Sound3D::_createBufferFromMemoryFnct=&Sound3D_createBufferFromMemory;
Sound3D::_freeBufferFnct=&Sound3D_freeBuffer;
Sound3D::_createSourceFnct=&Sound3D_createSource;
Sound3D::_destroySourceFnct=&Sound3D_destroySource;
Sound3D::_setSourceBufferFnct=&Sound3D_setSourceBuffer;
Sound3D::_addSourceBufferFnct=&Sound3D_addSourceBuffer;
Sound3D::_removeSourceBufferFnct=&Sound3D_removeSourceBuffer;
Sound3D::_getSourceBuffersStateFnct=&Sound3D_getSourceBuffersState;
Sound3D::_setSourceGainFnct=&Sound3D_setSourceGain;
Sound3D::_setSourcePitchFnct=&Sound3D_setSourcePitch;
Sound3D::_setSourceRepeatFnct=&Sound3D_setSourceRepeat;
Sound3D::_setSourceRelativeFnct=&Sound3D_setSourceRelative;
Sound3D::_setSourcePositionFnct=&Sound3D_setSourcePosition;
Sound3D::_setSourceVelocityFnct=&Sound3D_setSourceVelocity;
Sound3D::_playSourcesFnct=&Sound3D_playSources;
Sound3D::_pauseSourcesFnct=&Sound3D_pauseSources;
Sound3D::_stopSourcesFnct=&Sound3D_stopSources;
Sound3D::_getSourceStateFnct=&Sound3D_getSourceState;
Sound3D::_setListenerPositionFnct=&Sound3D_setListenerPosition;
Sound3D::_setListenerVelocityFnct=&Sound3D_setListenerVelocity;
Sound3D::_setListenerOrientationFnct=&Sound3D_setListenerOrientation;
Sound3D::_setListenerGainFnct=&Sound3D_setListenerGain;

alGetError();
alutInitWithoutContext((int *)0,(char **)0);
alutGetError();

URLSound3D::_init();

Sound3D::_initSuccess=true;
return(true);
}

extern "C"
void
Sound3DPlayer_destroy(void)
{
Sound3D::_currentListener.clear();
/*for(unsigned int i=Sound3D::_buffers->size();i--;)
  {
  (*Sound3D::_freeBuffer)((*Sound3D::_buffers)[i].bufferID);
  }
delete Sound3D::_buffers;
Sound3D::_buffers=(StlVector<Sound3D::BufferInfo> *)0;
*/
URLSound3D::_destroy();
Sound3D::_initSuccess=false;
if(Sound3D_globalContext)
  {
  alcDestroyContext(Sound3D_globalContext);
  Sound3D_globalContext=(ALCcontext *)0;
  }
if(Sound3D_device)
  {
  alcCloseDevice(Sound3D_device);
  Sound3D_device=(ALCdevice *)0;
  }
}

} // namespace AReVi

//----------------------------------------------------------------------------
