// LI2/ENIB - AReVi - Tcl wrapper library
//----------------------------------------------------------------------------
// file        : Ext/TclWrapper/tclWrapper.cpp
// author(s)   : F. Harrouet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This library uses the Tcl API (http://tcl.sourceforge.net).
//
//  The resulting dynamic library can be loaded at run-time from TCL.
//
//----------------------------------------------------------------------------

#include "AReVi/Utils/tclCommand.h"

#include "AReVi/arSystem.h"
#include "AReVi/arValue.h"
#include "AReVi/arMethod.h"
#include "AReVi/arConstant.h"
#include "AReVi/Utils/urlStream.h"

#include <tcl.h>

namespace AReVi {

static Tcl_Interp * TCL_interp=(Tcl_Interp *)0;

static bool (*TCL_AReViAppInitFnct)(int,char **)=(bool (*)(int,char **))0;
static bool (*TCL_AReViAppUpdateFnct)(void)=(bool (*)(void))0;
static void (*TCL_AReViAppDestroyFnct)(void)=(void (*)(void))0;

static Tcl_ObjType * TCL_ArRefType=(Tcl_ObjType *)0;
static StlVector<ArRef<ArObject> *> * TCL_refs=(StlVector<ArRef<ArObject> *> *)0;

//-------- ArRef TCL type ----------------------------------------------------

inline static
ArRef<ArObject> *
TCL_accessArRef(Tcl_Obj * obj)
{
return(reinterpret_cast<ArRef<ArObject> *>(obj->internalRep.otherValuePtr));
}

static
void
TCL_FreeArRefInternalRep(Tcl_Obj * obj)
{
ArRef<ArObject> * ref=TCL_accessArRef(obj);
if(TCL_refs)
  {
  for(unsigned int i=TCL_refs->size();i--;)
    {
    if((*TCL_refs)[i]==ref)
      {
      StlVectorFastErase(*TCL_refs,i);
      break;
      }
    }
  }
delete ref;
}

static
void
TCL_DupArRefInternalRep(Tcl_Obj * src,
                        Tcl_Obj * dup)
{
ArRef<ArObject> * r=TCL_accessArRef(src);
Tcl_InvalidateStringRep(dup);
dup->typePtr=src->typePtr;
if(TCL_refs)
  {
  TCL_refs->push_back(new ArRef<ArObject>(*r));
  dup->internalRep.otherValuePtr=TCL_refs->back();
  }
else
  {
  dup->internalRep.otherValuePtr=new ArRef<ArObject>;
  }
}

static
void
TCL_UpdateStringOfArRef(Tcl_Obj * obj)
{
ArRef<ArObject> * r=TCL_accessArRef(obj);
if(*r)
  {
  obj->length=(*r)->getAlias().size();
  obj->bytes=Tcl_Alloc(obj->length+1);
  strcpy(obj->bytes,(*r)->getAlias().c_str());
  }
else
  {
  obj->length=0;
  obj->bytes=Tcl_Alloc(obj->length+1);
  obj->bytes[0]='\0';
  }
}

static
int
TCL_SetArRefFromAny(Tcl_Interp * interp,
                    Tcl_Obj * obj)
{
int len;
char * s=Tcl_GetStringFromObj(obj,&len);
ArRef<ArObject> r;
ArObject::CLASS()->findAlias(StlString(s,len),r);
if(!r)
  {
  if(interp)
    {
    StlString msg="Cannot find ArObject '";
    msg+=s;
    msg+="'";
    Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
    }
  return(TCL_ERROR);
  }
if(obj->typePtr&&obj->typePtr->freeIntRepProc)
  {
  (*(obj->typePtr->freeIntRepProc))(obj);
  }
obj->typePtr=TCL_ArRefType;
if(TCL_refs)
  {
  TCL_refs->push_back(new ArRef<ArObject>(r));
  obj->internalRep.otherValuePtr=TCL_refs->back();
  }
else
  {
  obj->internalRep.otherValuePtr=new ArRef<ArObject>;
  }
return(TCL_OK);
}

static
Tcl_Obj *
Tcl_NewArRefObj(ArRef<ArObject> ref)
{
Tcl_Obj * obj=Tcl_NewObj();
Tcl_InvalidateStringRep(obj);
obj->typePtr=TCL_ArRefType;
if(TCL_refs)
  {
  TCL_refs->push_back(new ArRef<ArObject>(ref));
  obj->internalRep.otherValuePtr=TCL_refs->back();
  }
else
  {
  obj->internalRep.otherValuePtr=new ArRef<ArObject>;
  }
return(obj);
}

static
int
Tcl_GetArRefFromObj(Tcl_Interp * interp,
                    Tcl_Obj * obj,
                    ArRef<ArObject> * refOut)
{
if(obj->typePtr!=TCL_ArRefType)
  {
  int result=TCL_SetArRefFromAny(interp,obj);
  if(result!=TCL_OK)
    {
    return(result);
    }
  }
*refOut=*TCL_accessArRef(obj);
return(TCL_OK);
}

static
void
Tcl_SetArRefObj(Tcl_Obj * obj,
                ArRef<ArObject> ref)
{
if(Tcl_IsShared(obj))
  {
  Tcl_Panic("Tcl_SetArRefObj called with shared object");
  }
if(obj->typePtr!=TCL_ArRefType)
  {
  if(obj->typePtr&&obj->typePtr->freeIntRepProc)
    {
    (*(obj->typePtr->freeIntRepProc))(obj);
    }
  obj->typePtr=TCL_ArRefType;
  if(TCL_refs)
    {
    TCL_refs->push_back(new ArRef<ArObject>);
    obj->internalRep.otherValuePtr=TCL_refs->back();
    }
  else
    {
    obj->internalRep.otherValuePtr=new ArRef<ArObject>;
    }
  }
Tcl_InvalidateStringRep(obj);
*TCL_accessArRef(obj)=(TCL_refs ? ref : ArObject::nullRef());
}

//-------- Conversions between Tcl_Obj and ArValue ---------------------------

#define TCL_READ_VALUE(TCL_TYPE,TCL_FNCT,CPP_TYPE,CPP_FNCT)                   \
  switch(type->getDimension())                                                \
    {                                                                         \
    case 0:                                                                   \
      {                                                                       \
      TCL_TYPE r;                                                             \
      if(Tcl_Get##TCL_FNCT##FromObj(interp,obj,&r)!=TCL_OK) return(false);    \
      value->access##CPP_FNCT()=(CPP_TYPE)r;                                  \
      return(true);                                                           \
      }                                                                       \
    case 1:                                                                   \
      {                                                                       \
      StlVector<CPP_TYPE > & v=value->access##CPP_FNCT##V();                  \
      int sz;                                                                 \
      if(Tcl_ListObjLength(interp,obj,&sz)!=TCL_OK) return(false);            \
      for(int i=0;i<sz;i++)                                                   \
        {                                                                     \
        Tcl_Obj * l;                                                          \
        if(Tcl_ListObjIndex(interp,obj,i,&l)!=TCL_OK) return(false);          \
        TCL_TYPE r;                                                           \
        if(Tcl_Get##TCL_FNCT##FromObj(interp,l,&r)!=TCL_OK) return(false);    \
        v.push_back((CPP_TYPE)r);                                             \
        }                                                                     \
      return(true);                                                           \
      }                                                                       \
    case 2:                                                                   \
      {                                                                       \
      StlVector<StlVector<CPP_TYPE > > & vv=value->access##CPP_FNCT##VV();    \
      int sz1;                                                                \
      if(Tcl_ListObjLength(interp,obj,&sz1)!=TCL_OK) return(false);           \
      for(int i=0;i<sz1;i++)                                                  \
        {                                                                     \
        Tcl_Obj * l1;                                                         \
        if(Tcl_ListObjIndex(interp,obj,i,&l1)!=TCL_OK) return(false);         \
        vv.push_back(StlVector<CPP_TYPE >());                                 \
        int sz2;                                                              \
        if(Tcl_ListObjLength(interp,l1,&sz2)!=TCL_OK) return(false);          \
        for(int j=0;j<sz2;j++)                                                \
          {                                                                   \
          Tcl_Obj * l2;                                                       \
          if(Tcl_ListObjIndex(interp,l1,j,&l2)!=TCL_OK) return(false);        \
          TCL_TYPE r;                                                         \
          if(Tcl_Get##TCL_FNCT##FromObj(interp,l2,&r)!=TCL_OK) return(false); \
          vv.back().push_back((CPP_TYPE)r);                                   \
          }                                                                   \
        }                                                                     \
      return(true);                                                           \
      }                                                                       \
    }                                                                         \
  break;

static
bool // success
TCL_readValue(Tcl_Interp * interp,
              Tcl_Obj * obj,
              ArRef<ArValue> value)
{
ArConstRef<ArType> type=value->getType();
if(type->isDirect())
  {
  switch(type->getDirectType())
    {
    case ArType::DT_BOOL:
      {
      TCL_READ_VALUE(int,Boolean,bool,Bool)
      }
    case ArType::DT_CHAR:
      {
      TCL_READ_VALUE(int,Int,char,Char)
      }
    case ArType::DT_UCHAR:
      {
      TCL_READ_VALUE(int,Int,unsigned char,UChar)
      }
    case ArType::DT_SHORT:
      {
      TCL_READ_VALUE(int,Int,short,Short)
      }
    case ArType::DT_USHORT:
      {
      TCL_READ_VALUE(int,Int,unsigned short,UShort)
      }
    case ArType::DT_INT:
      {
      TCL_READ_VALUE(int,Int,int,Int)
      }
    case ArType::DT_UINT:
      {
      TCL_READ_VALUE(int,Int,unsigned int,UInt)
      }
    case ArType::DT_LLONG:
      {
      TCL_READ_VALUE(Tcl_WideInt,WideInt,long long,LLong)
      }
    case ArType::DT_ULLONG:
      {
      TCL_READ_VALUE(Tcl_WideInt,WideInt,unsigned long long,ULLong)
      }
    case ArType::DT_FLOAT:
      {
      TCL_READ_VALUE(double,Double,float,Float)
      }
    case ArType::DT_DOUBLE:
      {
      TCL_READ_VALUE(double,Double,double,Double)
      }
    case ArType::DT_STRING:
      {
      switch(type->getDimension())
        {
        case 0:
          {
          int len;
          char * r=Tcl_GetStringFromObj(obj,&len);
          value->accessString().assign(r,len);
          return(true);
          }
        case 1:
          {
          StlVector<StlString> & v=value->accessStringV();
          int sz;
          if(Tcl_ListObjLength(interp,obj,&sz)!=TCL_OK) return(false);
          for(int i=0;i<sz;i++)
            {
            Tcl_Obj * l;
            if(Tcl_ListObjIndex(interp,obj,i,&l)!=TCL_OK) return(false);
            int len;
            char * r=Tcl_GetStringFromObj(l,&len);
            v.push_back(StlString(r,len));
            }
          return(true);
          }
        case 2:
          {
          StlVector<StlVector<StlString> > & vv=value->accessStringVV();
          int sz1;
          if(Tcl_ListObjLength(interp,obj,&sz1)!=TCL_OK) return(false);
          for(int i=0;i<sz1;i++)
            {
            Tcl_Obj * l1;
            if(Tcl_ListObjIndex(interp,obj,i,&l1)!=TCL_OK) return(false);
            vv.push_back(StlVector<StlString>());
            int sz2;
            if(Tcl_ListObjLength(interp,l1,&sz2)!=TCL_OK) return(false);
            for(int j=0;j<sz2;j++)
              {
              Tcl_Obj * l2;
              if(Tcl_ListObjIndex(interp,l1,j,&l2)!=TCL_OK) return(false);
              int len;
              char * r=Tcl_GetStringFromObj(l2,&len);
              vv.back().push_back(StlString(r,len));
              }
            }
          return(true);
          }
        }
      break;
      }
    default:
      {
      break; // should not happen
      }
    }
  }
else
  {
  switch(type->getIndirectType())
    {
    case ArType::IT_REF:
      {
      TCL_READ_VALUE(ArRef<ArObject>,ArRef,ArRef<ArObject>,Ref)
      }
    case ArType::IT_CONST_REF:
      {
      TCL_READ_VALUE(ArRef<ArObject>,ArRef,ArConstRef<ArObject>,ConstRef)
      }
    case ArType::IT_PTR:
      {
      TCL_READ_VALUE(ArRef<ArObject>,ArRef,ArPtr<ArObject>,Ptr)
      }
    case ArType::IT_CONST_PTR:
      {
      TCL_READ_VALUE(ArRef<ArObject>,ArRef,ArConstPtr<ArObject>,ConstPtr)
      }
    default:
      {
      break; // should not happen
      }
    }
  }
return(false);
}

#define TCL_WRITE_VALUE(TCL_TYPE,TCL_FNCT,CPP_TYPE,CPP_FNCT)                  \
  switch(type->getDimension())                                                \
    {                                                                         \
    case 0:                                                                   \
      {                                                                       \
      Tcl_Set##TCL_FNCT##Obj(obj,(TCL_TYPE)value->get##CPP_FNCT());           \
      return(true);                                                           \
      }                                                                       \
    case 1:                                                                   \
      {                                                                       \
      const StlVector<CPP_TYPE > & v=value->get##CPP_FNCT##V();               \
      StlVector<Tcl_Obj *> l(v.size(),(Tcl_Obj *)0);                          \
      for(unsigned int i=v.size();i--;)                                       \
        {                                                                     \
        l[i]=Tcl_New##TCL_FNCT##Obj((TCL_TYPE)v[i]);                          \
        }                                                                     \
      Tcl_SetListObj(obj,l.size(),&l.front());                                \
      return(true);                                                           \
      }                                                                       \
    case 2:                                                                   \
      {                                                                       \
      const StlVector<StlVector<CPP_TYPE > > & vv=value->get##CPP_FNCT##VV(); \
      StlVector<Tcl_Obj *> l1(vv.size());                                     \
      for(unsigned int i=vv.size();i--;)                                      \
        {                                                                     \
        const StlVector<CPP_TYPE > & v=vv[i];                                 \
        StlVector<Tcl_Obj *> l2(v.size());                                    \
        for(unsigned int j=v.size();j--;)                                     \
          {                                                                   \
          l2[j]=Tcl_New##TCL_FNCT##Obj((TCL_TYPE)v[j]);                       \
          }                                                                   \
        l1[i]=Tcl_NewListObj(l2.size(),&l2.front());                          \
        }                                                                     \
      Tcl_SetListObj(obj,l1.size(),&l1.front());                              \
      return(true);                                                           \
      }                                                                       \
    }                                                                         \
  break;

#define TCL_WRITE_CONST_ARREF(CPP_TYPE,CPP_FNCT)                              \
  switch(type->getDimension())                                                \
    {                                                                         \
    case 0:                                                                   \
      {                                                                       \
      Tcl_SetArRefObj(obj,ar_const_cast(value->get##CPP_FNCT()));             \
      return(true);                                                           \
      }                                                                       \
    case 1:                                                                   \
      {                                                                       \
      const StlVector<CPP_TYPE > & v=value->get##CPP_FNCT##V();               \
      StlVector<Tcl_Obj *> l(v.size(),(Tcl_Obj *)0);                          \
      for(unsigned int i=v.size();i--;)                                       \
        {                                                                     \
        l[i]=Tcl_NewArRefObj(ar_const_cast(v[i]));                            \
        }                                                                     \
      Tcl_SetListObj(obj,l.size(),&l.front());                                \
      return(true);                                                           \
      }                                                                       \
    case 2:                                                                   \
      {                                                                       \
      const StlVector<StlVector<CPP_TYPE > > & vv=value->get##CPP_FNCT##VV(); \
      StlVector<Tcl_Obj *> l1(vv.size());                                     \
      for(unsigned int i=vv.size();i--;)                                      \
        {                                                                     \
        const StlVector<CPP_TYPE > & v=vv[i];                                 \
        StlVector<Tcl_Obj *> l2(v.size());                                    \
        for(unsigned int j=v.size();j--;)                                     \
          {                                                                   \
          l2[j]=Tcl_NewArRefObj(ar_const_cast(v[j]));                         \
          }                                                                   \
        l1[i]=Tcl_NewListObj(l2.size(),&l2.front());                          \
        }                                                                     \
      Tcl_SetListObj(obj,l1.size(),&l1.front());                              \
      return(true);                                                           \
      }                                                                       \
    }                                                                         \
  break;

static
bool // success
TCL_writeValue(Tcl_Obj * obj,
               ArConstRef<ArValue> value)
{
ArConstRef<ArType> type=value->getType();
if(type->isDirect())
  {
  switch(type->getDirectType())
    {
    case ArType::DT_BOOL:
      {
      TCL_WRITE_VALUE(int,Boolean,bool,Bool)
      }
    case ArType::DT_CHAR:
      {
      TCL_WRITE_VALUE(int,Int,char,Char)
      }
    case ArType::DT_UCHAR:
      {
      TCL_WRITE_VALUE(int,Int,unsigned char,UChar)
      }
    case ArType::DT_SHORT:
      {
      TCL_WRITE_VALUE(int,Int,short,Short)
      }
    case ArType::DT_USHORT:
      {
      TCL_WRITE_VALUE(int,Int,unsigned short,UShort)
      }
    case ArType::DT_INT:
      {
      TCL_WRITE_VALUE(int,Int,int,Int)
      }
    case ArType::DT_UINT:
      {
      TCL_WRITE_VALUE(int,Int,unsigned int,UInt)
      }
    case ArType::DT_LLONG:
      {
      TCL_WRITE_VALUE(long long,WideInt,long long,LLong)
      }
    case ArType::DT_ULLONG:
      {
      TCL_WRITE_VALUE(long long,WideInt,unsigned long long,ULLong)
      }
    case ArType::DT_FLOAT:
      {
      TCL_WRITE_VALUE(double,Double,float,Float)
      }
    case ArType::DT_DOUBLE:
      {
      TCL_WRITE_VALUE(double,Double,double,Double)
      }
    case ArType::DT_STRING:
      {
      switch(type->getDimension())
        {
        case 0:
          {
          const StlString & str=value->getString();
          Tcl_SetStringObj(obj,str.c_str(),str.size());
          return(true);
          }
        case 1:
          {
          const StlVector<StlString> & v=value->getStringV();
          StlVector<Tcl_Obj *> l(v.size(),(Tcl_Obj *)0);
          for(unsigned int i=v.size();i--;)
            {
            l[i]=Tcl_NewStringObj(v[i].c_str(),v[i].size());
            }
          Tcl_SetListObj(obj,l.size(),&l.front());
          return(true);
          }
        case 2:
          {
          const StlVector<StlVector<StlString> > & vv=value->getStringVV();
          StlVector<Tcl_Obj *> l1(vv.size());
          for(unsigned int i=vv.size();i--;)
            {
            const StlVector<StlString> & v=vv[i];
            StlVector<Tcl_Obj *> l2(v.size());
            for(unsigned int j=v.size();j--;)
              {
              l2[j]=Tcl_NewStringObj(v[j].c_str(),v[j].size());
              }
            l1[i]=Tcl_NewListObj(l2.size(),&l2.front());
            }
          Tcl_SetListObj(obj,l1.size(),&l1.front());
          return(true);
          }
        }
      break;
      }
    default:
      {
      break; // should not happen
      }
    }
  }
else
  {
  switch(type->getIndirectType())
    {
    case ArType::IT_REF:
      {
      TCL_WRITE_VALUE(ArRef<ArObject>,ArRef,ArRef<ArObject>,Ref)
      }
    case ArType::IT_CONST_REF: // FIXME: const/non-const distinction ?
      {
      TCL_WRITE_CONST_ARREF(ArConstRef<ArObject>,ConstRef)
      }
    case ArType::IT_PTR:
      {
      TCL_WRITE_VALUE(ArRef<ArObject>,ArRef,ArPtr<ArObject>,Ptr)
      }
    case ArType::IT_CONST_PTR: // FIXME: const/non-const distinction ?
      {
      TCL_WRITE_CONST_ARREF(ArConstPtr<ArObject>,ConstPtr)
      }
    default:
      {
      break; // should not happen
      }
    }
  }
return(false);
}

//-------- New TCL commands to access AReVi ----------------------------------

static
int
TCL_initApplicationCmd(ClientData /* data */,
                       Tcl_Interp * interp,
                       int objc,
                       Tcl_Obj * CONST objv[])
{
if(!TCL_init())
  {
  Tcl_SetResult(interp,"false",TCL_VOLATILE);
  return(TCL_OK);
  }
int len;
StlVector<char *> args;
args.reserve(objc+1);
for(int i=0;i<objc;i++)
  {
  args.push_back(Tcl_GetStringFromObj(objv[i],&len));
  }
args.push_back((char *)0);
bool result=(*TCL_AReViAppInitFnct)(objc,&args.front());
Tcl_SetResult(interp,
              const_cast<char *>(result ? "true" : "false"),
              TCL_VOLATILE);
return(TCL_OK);
}

static
int
TCL_updateApplicationCmd(ClientData /* data */,
                         Tcl_Interp * interp,
                         int objc,
                         Tcl_Obj * CONST objv[])
{
if(objc>1)
  {
  int len;
  StlString msg="Command '";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+="' does not require any argument";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
bool result=(*TCL_AReViAppUpdateFnct)();
Tcl_SetResult(interp,
              const_cast<char *>(result ? "true" : "false"),
              TCL_VOLATILE);
return(TCL_OK);
}

static
int
TCL_destroyApplicationCmd(ClientData /* data */,
                          Tcl_Interp * interp,
                          int objc,
                          Tcl_Obj * CONST objv[])
{
if(objc>1)
  {
  int len;
  StlString msg="Command '";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+="' does not require any argument";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
(*TCL_AReViAppDestroyFnct)();
TCL_destroy();
if(TCL_ArRefType)
  {
  delete TCL_ArRefType;
  TCL_ArRefType=(Tcl_ObjType *)0;
  }
return(TCL_OK);
}

static
int
TCL_invokeCmd(ClientData /* data */,
              Tcl_Interp * interp,
              int objc,
              Tcl_Obj * CONST objv[])
{
int len;
if(objc<3)
  {
  StlString msg="bad args: should be `";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+=" ?-static? ?ArRef? method args'";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
ArRef<ArObject> r;
char * s=Tcl_GetStringFromObj(objv[1],&len);
if(strncmp(s,"-static",len))
  {
  if(Tcl_GetArRefFromObj(interp,objv[1],&r)!=TCL_OK)
    {
    StlString msg="Cannot use `";
    msg+=Tcl_GetStringFromObj(objv[0],&len);
    msg+="' on unknown object `";
    msg+=Tcl_GetStringFromObj(objv[1],&len);
    msg+="'";
    Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
    return(TCL_ERROR);
    }
  }
StlString mthName1;
StlString mthName2;
StlString argSignature;
char * mthStr=Tcl_GetStringFromObj(objv[2],&len);
char * pos=strchr(mthStr,'(');
if(pos&&(mthStr[len-1]==')'))
  {
  mthName1.assign(mthStr,pos-mthStr);
  argSignature.assign(pos+1,len-2-mthName1.size());
  if(!argSignature.size())
    {
    argSignature="void";
    }
  }
else
  {
  mthName1.assign(mthStr,len);
  }
pos=strstr(mthName1.c_str(),"::");
if(pos)
  {
  mthName2=pos+2;
  mthName1=mthName1.substr(0,mthName1.size()-mthName2.size()-2);
  }
ArConstRef<ArMethod> mth;
if(argSignature.size())
  {
  if(r)
    {
    if(mthName2.size())
      {
      ArConstRef<ArClass> cls=ArClass::find(mthName1);
      if(cls&&r->getClass()->isA(cls))
        {
        mth=r->getClass()->findMethodWithSignature(mthName2,argSignature);
        }
      }
    else
      {
      mth=r->getClass()->findMethodWithSignature(mthName1,argSignature);
      }
    }
  else
    {
    ArConstRef<ArClass> cls=ArClass::find(mthName1);
    if(cls)
      {
      if(mthName2.size())
        {
        mth=cls->findStaticMethodWithSignature(mthName2,argSignature);
        }
      else
        {
        mth=cls->findConstructorWithSignature(argSignature);
        }
      }
    }
  if(mth&&(mth->getNbArgs()!=(unsigned int)objc-3))
    {
    StlString msg="Bad number of method arguments in `";
    msg+=Tcl_GetStringFromObj(objv[0],&len);
    msg+=" ";
    msg+=Tcl_GetStringFromObj(objv[1],&len);
    msg+=" ";
    msg+=Tcl_GetStringFromObj(objv[2],&len);
    msg+="' (expected ";
    msg+=uintToStr(mth->getNbArgs());
    msg+=", but got ";
    msg+=intToStr(objc-3);
    msg+=")";
    Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
    return(TCL_ERROR);
    }
  }
else
  {
  StlVector<ArConstRef<ArMethod> > methods;
  if(r)
    {
    if(mthName2.size())
      {
      ArConstRef<ArClass> cls=ArClass::find(mthName1);
      if(cls&&r->getClass()->isA(cls))
        {
        r->getClass()->findMethods(methods,mthName2);
        }
      }
    else
      {
      r->getClass()->findMethods(methods,mthName1);
      }
    }
  else
    {
    ArConstRef<ArClass> cls=ArClass::find(mthName1);
    if(cls)
      {
      if(mthName2.size())
        {
        cls->findStaticMethods(methods,mthName2);
        }
      else
        {
        cls->findConstructors(methods);
        }
      }
    }
  unsigned int nbArgs=objc-3;
  for(unsigned int i=methods.size();i--;)
    {
    if(methods[i]->getNbArgs()!=nbArgs)
      {
      StlVectorFastErase(methods,i);
      }
    }
  switch(methods.size())
    {
    case 0:
      {
      break;
      }
    case 1:
      {
      mth=methods.front();
      break;
      }
    default:
      {
      StlString msg="Possible ";
      msg+=uintToStr(nbArgs);
      msg+="-argument signatures for `";
      msg+=Tcl_GetStringFromObj(objv[0],&len);
      msg+=" ";
      msg+=Tcl_GetStringFromObj(objv[1],&len);
      msg+=" ";
      msg+=Tcl_GetStringFromObj(objv[2],&len);
      msg+="' are:";
      for(unsigned int i=methods.size();i--;)
        {
        msg+=" (";
        msg+=methods[i]->getArgSignature();
        msg+=")";
        }
      Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
      return(TCL_ERROR);
      }
    }
  }
if(!mth)
  {
  StlString msg="Cannot find any suitable ";
  if(!argSignature.size())
    {
    msg+=uintToStr(objc-3);
    msg+="-argument ";
    }
  msg+="method `";
  msg+=Tcl_GetStringFromObj(objv[2],&len);
  msg+="' in `";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+=" ";
  msg+=Tcl_GetStringFromObj(objv[1],&len);
  msg+="'";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
StlVector<ArRef<ArValue> > args;
unsigned int nbArgs=mth->getNbArgs();
unsigned int i;
for(i=0;i<nbArgs;i++)
  {
  ArConstRef<ArType> type=mth->getArgType(i);
  args.push_back(ArValue::NEW(type));
  Tcl_Obj * arg;
  if(type->getPassingMode()==ArType::PM_REF)
    {
    arg=Tcl_ObjGetVar2(interp,objv[i+3],(Tcl_Obj *)0,0);
    }
  else
    {
    arg=objv[i+3];
    }
  if(arg)
    {
    TCL_readValue(interp,arg,args.back());
    }
  }
ArRef<ArValue> result;
if(mth->getResultType())
  {
  result=ArValue::NEW(mth->getResultType());
  }
if(mth->isStaticMethod())
  {
  mth->invokeStatic(result,args,false);
  }
else
  {
  mth->invoke(r,result,args,false);
  }
for(i=nbArgs;i--;)
  {
  if(mth->getArgType(i)->getPassingMode()==ArType::PM_REF)
    {
    Tcl_Obj * arg=Tcl_ObjGetVar2(interp,objv[i+3],(Tcl_Obj *)0,0);
    if(arg&&!Tcl_IsShared(arg))
      {
      TCL_writeValue(arg,args[i]);
      }
    else
      {
      arg=Tcl_NewObj();
      TCL_writeValue(arg,args[i]);
      if(!Tcl_ObjSetVar2(interp,objv[i+3],(Tcl_Obj *)0,arg,0))
        {
        Tcl_IncrRefCount(arg);
        Tcl_DecrRefCount(arg); // free arg here
        }
      }
    }
  }
if(result)
  {
  Tcl_Obj * obj=Tcl_NewObj();
  TCL_writeValue(obj,result);
  Tcl_SetObjResult(interp,obj);
  }
else
  {
  Tcl_SetResult(interp,(char *)0,TCL_VOLATILE);
  }
return(TCL_OK);
}

static
int
TCL_constantCmd(ClientData /* data */,
                Tcl_Interp * interp,
                int objc,
                Tcl_Obj * CONST objv[])
{
int len;
if(objc!=2)
  {
  StlString msg="bad args: should be `";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+=" ClassName::ConstantName'";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
StlString className;
StlString constantName;
char * s=Tcl_GetStringFromObj(objv[1],&len);
char * pos=strstr(s,"::");
if(pos)
  {
  className.assign(s,pos-s);
  constantName.assign(pos+2);
  }
if(!className.size()||!constantName.size())
  {
  StlString msg="Cannot understand constant `";
  msg+=Tcl_GetStringFromObj(objv[1],&len);
  msg+="' in `";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+="'";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
ArConstRef<ArConstant> cst;
ArConstRef<ArClass> cls=ArClass::find(className);
if(cls)
  {
  cst=cls->findConstant(constantName);
  }
if(cst)
  {
  Tcl_Obj * obj=Tcl_NewObj();
  TCL_writeValue(obj,cst->getValue());
  Tcl_SetObjResult(interp,obj);
  return(TCL_OK);
  }
else
  {
  StlString msg="Cannot find any suitable constant `";
  msg+=Tcl_GetStringFromObj(objv[1],&len);
  msg+="' in `";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+="'";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
}

//-------- Command management ------------------------------------------------

static
int
TCL_AReViCommand(ClientData data,
                 Tcl_Interp * interp,
                 int objc,
                 Tcl_Obj * CONST objv[])
{
_TclCommandInfo * info=reinterpret_cast<_TclCommandInfo *>(data);
int len;
if(info->obj)
  {
  StlVector<StlString> args;
  args.reserve(objc);
  for(int i=0;i<objc;i++)
    {
    args.push_back(Tcl_GetStringFromObj(objv[i],&len));
    }
  StlString result;
  bool success=(info->obj.c_ptr()->*info->method)(args,result);
  Tcl_SetResult(interp,const_cast<char *>(result.c_str()),TCL_VOLATILE);
  return(success ? TCL_OK : TCL_ERROR);
  }
else
  {
  StlString msg="AReVi side of command `";
  msg+=Tcl_GetStringFromObj(objv[0],&len);
  msg+="' does not exist anymore";
  Tcl_SetResult(interp,const_cast<char *>(msg.c_str()),TCL_VOLATILE);
  return(TCL_ERROR);
  }
}

static
void
TCL_add(_TclCommandInfo * cmd)
{
Tcl_CmdInfo info;
if(Tcl_GetCommandInfo(TCL_interp,cmd->name.c_str(),&info))
  {
  cmd->token=(Tcl_Command)0;
  }
else
  {
  cmd->token=Tcl_CreateObjCommand(TCL_interp,cmd->name.c_str(),
                                  &TCL_AReViCommand,cmd,
                                  (Tcl_CmdDeleteProc *)0);
  }
}

static
void
TCL_remove(_TclCommandInfo * cmd)
{
if(cmd->token)
  {
  Tcl_DeleteCommandFromToken(TCL_interp,(Tcl_Command)(cmd->token));
  }
}

//-------- Tcl invocation ----------------------------------------------------

static
bool // command is complete
TCL_isComplete(const StlString & command)
{
return(Tcl_CommandComplete(command.c_str()) ? true : false);
}

static
bool // success
TCL_eval(const StlString & command,
         StlString & resultOut)
{
int err=Tcl_Eval(TCL_interp,command.c_str());
if(TCL_interp->result)
  {
  resultOut=TCL_interp->result;
  }
else
  {
  resultOut.erase();
  }
return(err==TCL_ERROR ? false : true);
}

static
bool // success
TCL_evalURL(const StlString & url,
            StlString & resultOut)
{
ArRef<URLStream> input=URLStream::NEW(url);
if(input->fail())
  {
  resultOut="Cannot retreive URL `"+url+"'";
  return(false);
  }
if(input->getProtocol()=="file")
  {
  StlString fileName=input->getFileName();
  input.clear();
  int err=Tcl_EvalFile(TCL_interp,fileName.c_str());
  if(TCL_interp->result)
    {
    resultOut=TCL_interp->result;
    }
  else
    {
    resultOut.erase();
    }
  return(err==TCL_ERROR ? false : true);
  }
else
  {
  StlString tmpName;
  int tmpFD=ArSystem::createTmpFile("./TCL_url_",tmpName);
  if(tmpFD==-1)
    {
    resultOut="Cannot retreive URL `"+url+"'";
    return(false);
    }
  ArRef<FDOStream> output=FDOStream::NEW(tmpFD,true);
  if(output->fail())
    {
    resultOut="Cannot retreive URL `"+url+"'";
    ArSystem::removeFile(tmpName);
    return(false);
    }
  else
    {
    output->copyDataFromInput(input);
    input.clear();
    output.clear();
    int err=Tcl_EvalFile(TCL_interp,tmpName.c_str());
    if(TCL_interp->result)
      {
      resultOut=TCL_interp->result;
      }
    else
      {
      resultOut.erase();
      }
    ArSystem::removeFile(tmpName);
    return(err==TCL_ERROR ? false : true);
    }
  }
}

//-------- Initialization ----------------------------------------------------

bool // success
TCL_init(void)
{
TclCommand::_initSuccess=true;
TclCommand::_commands=new StlVector<_TclCommandInfo *>;
TclCommand::_isCompleteFnct=&TCL_isComplete;
TclCommand::_evalFnct=&TCL_eval;
TclCommand::_evalURLFnct=&TCL_evalURL;
TclCommand::_addFnct=&TCL_add;
TclCommand::_removeFnct=&TCL_remove;
return(true);
}

void
TCL_destroy(void)
{
if(TclCommand::_initSuccess)
  {
  while(!TclCommand::_commands->empty())
    {
    TclCommand::remove(TclCommand::_commands->back()->name);
    }
  delete TclCommand::_commands;
  }
TCL_interp=(Tcl_Interp *)0;
TCL_AReViAppInitFnct=(bool (*)(int,char **))0;
TCL_AReViAppUpdateFnct=(bool (*)(void))0;
TCL_AReViAppDestroyFnct=(void (*)(void))0;
if(TCL_refs)
  {
  while(!TCL_refs->empty())
    {
    ArRef<ArObject> * ref=TCL_refs->back();
    TCL_refs->pop_back();
    ref->clear();
    }
  delete TCL_refs;
  TCL_refs=(StlVector<ArRef<ArObject> *> *)0;
  }
TclCommand::_initSuccess=false;
TclCommand::_commands=(StlVector<_TclCommandInfo *> *)0;
TclCommand::_isCompleteFnct=(bool (*)(const StlString &))0;
TclCommand::_evalFnct=(bool (*)(const StlString &,StlString &))0;
TclCommand::_evalURLFnct=(bool (*)(const StlString &,StlString &))0;
TclCommand::_addFnct=(void (*)(_TclCommandInfo *))0;
TclCommand::_removeFnct=(void (*)(_TclCommandInfo *))0;
}

int
TCL_InitAReViApp(void * tclInterp,
                 bool (*initAppFnct)(int,char **),
                 bool (*updateAppFnct)(void),
                 void (*destroyAppFnct)(void))
{
TCL_interp=reinterpret_cast<Tcl_Interp *>(tclInterp);
TCL_AReViAppInitFnct=initAppFnct;
TCL_AReViAppUpdateFnct=updateAppFnct;
TCL_AReViAppDestroyFnct=destroyAppFnct;
TCL_ArRefType=new Tcl_ObjType;
TCL_ArRefType->name="ArRef";
TCL_ArRefType->freeIntRepProc=&TCL_FreeArRefInternalRep;
TCL_ArRefType->dupIntRepProc=&TCL_DupArRefInternalRep;
TCL_ArRefType->updateStringProc=&TCL_UpdateStringOfArRef;
TCL_ArRefType->setFromAnyProc=&TCL_SetArRefFromAny;
Tcl_RegisterObjType(TCL_ArRefType);
TCL_refs=new StlVector<ArRef<ArObject> *>;
Tcl_CreateObjCommand(TCL_interp,"::AReViApp::init",
                     &TCL_initApplicationCmd,
                     (ClientData)0,(Tcl_CmdDeleteProc *)0);
Tcl_CreateObjCommand(TCL_interp,"::AReViApp::update",
                     &TCL_updateApplicationCmd,
                     (ClientData)0,(Tcl_CmdDeleteProc *)0);
Tcl_CreateObjCommand(TCL_interp,"::AReViApp::destroy",
                     &TCL_destroyApplicationCmd,
                     (ClientData)0,(Tcl_CmdDeleteProc *)0);
Tcl_CreateObjCommand(TCL_interp,"::AReViApp::invoke",
                     &TCL_invokeCmd,
                     (ClientData)0,(Tcl_CmdDeleteProc *)0);
Tcl_CreateObjCommand(TCL_interp,"::AReViApp::constant",
                     &TCL_constantCmd,
                     (ClientData)0,(Tcl_CmdDeleteProc *)0);
Tcl_PkgProvide(TCL_interp,"AReViApp","1.0");
return(TCL_OK);
}

} // namespace AReVi

//----------------------------------------------------------------------------
