// LI2/ENIB - AReVi-Contrib - Audio & Video Encoder plugin
//----------------------------------------------------------------------------
// file        : ExtContrib/AudioVideoEncoder/audioVideoEncoder.cpp
// author(s)   : T. Jourdan, B. Muguet
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This plugin uses the FFMpeg API (http://ffmpeg.sourceforge.net).
//
//  The resulting dynamic library can be loaded at run-time
//  (fails if a dependency problem occurs).
//
//----------------------------------------------------------------------------

#include "AReVi/Contrib/aviRecorder.h"
#include "AReVi/Contrib/soundEncoder.h"

#include "AReVi/arSystem.h"
#include "AReVi/Utils/memoryBlock.h"

#include <ffmpeg/avformat.h>
#include <ffmpeg/avcodec.h>

#if (LIBAVCODEC_BUILD >= 4754)
# define STREAM_CODEC_CONTEXT(st) ((st)->codec)
#else
# define STREAM_CODEC_CONTEXT(st) (&(st)->codec)
#endif

#include <errno.h>

#define IO_BUFFER_SIZE 32768

namespace AReVi {

struct AviRecorderData
{
AVFormatContext * oc;

AVStream * video_st;
AVStream * audio_st;

unsigned char * videoOutBuffer;
int videoOutBufferSize;

uint8_t* audioOutBuffer;
int audioOutBufferSize;
};

struct SoundEncoderData
{
  AVOutputFormat * fmt;

  AVCodec *encoder;
  AVCodec *decoder;

  AVCodecContext *encoderContext;
  AVCodecContext *decoderContext;
};

//-------- Recording management ----------------------------------------------

/* add a video output stream */
static
AVStream *
addVideoStream(AVFormatContext * oc,
               CodecID codec_id,
               double frameRate,
               unsigned int bitRate,
               unsigned int width,
               unsigned int height)
{
AVCodecContext * c;
AVStream * st;
st=av_new_stream(oc,0);
if(!st)
  {
  return((AVStream *)0);
  }
c=&st->codec;
c->codec_id=codec_id;
c->codec_type=CODEC_TYPE_VIDEO;
c->pix_fmt=PIX_FMT_YUV420P;

/* put sample parameters */
c->bit_rate=bitRate;
/* resolution must be a multiple of two */
c->width=width;
c->height=height;
/* frames per second */
#if (LIBAVCODEC_BUILD >= 4754)
c->time_base.den=(int)frameRate;
c->time_base.num=1;
#else
c->frame_rate=(int)frameRate;
c->frame_rate_base=1;
#endif
c->gop_size=12; /* emit one intra frame every twelve frames at most */
if(c->codec_id==CODEC_ID_MPEG2VIDEO)
  {
  /* just for testing, we also add B frames */
  c->max_b_frames=2;
  }
if(c->codec_id==CODEC_ID_MPEG1VIDEO)
  {
  /* needed to avoid using macroblocks in which some coeffs overflow
     this doesnt happen with normal video, it just happens here as the
     motion of the chroma plane doesnt match the luma plane */
  c->mb_decision=2;
  }
// some formats want stream headers to be seperate
if(!strcmp(oc->oformat->name,"mp4")||
   !strcmp(oc->oformat->name,"mov")||
   !strcmp(oc->oformat->name,"3gp"))
  {
  c->flags|=CODEC_FLAG_GLOBAL_HEADER;
  }
return(st);
}


// Add a audio streams
static
AVStream *addAudioStream(AVFormatContext *oc,
			 CodecID codec_id,
			 int bitRate,
			 int sampleRate,
			 int channels)
{
  AVCodecContext *c;
  AVStream *st;

  st = av_new_stream(oc, 1);
  if (!st) {
    fprintf(stderr, "Could not alloc stream\n");
    exit(1);
  }

  c = STREAM_CODEC_CONTEXT(st);
  c->codec_id = codec_id;
  c->codec_type = CODEC_TYPE_AUDIO;

  /* put sample parameters */
  c->bit_rate = bitRate;
  c->sample_rate = sampleRate;
  c->channels = channels;

  return st;
}

static
AVFrame *
allocPicture(int pixFmt,
             int width,
             int height,
             bool allocBuffer)
{
AVFrame * picture;
unsigned char * pictureBuf;
int size;
picture=avcodec_alloc_frame();
if(!picture)
  {
  return((AVFrame*)0);
  }
if(!allocBuffer)
  {
  return(picture);
  }
size=avpicture_get_size(pixFmt,width,height);
pictureBuf=(unsigned char*)malloc(size);
if(!pictureBuf)
  {
  av_free(picture);
  return((AVFrame*)0);
  }
avpicture_fill((AVPicture*)picture,pictureBuf,pixFmt,width,height);
return(picture);
}

static
bool
openVideo(AVFormatContext * oc,
          AVStream * st,
          unsigned char * & videoOutBuffer,
          int & videoOutBufferSize)
{

  AVCodec * codec;
  AVCodecContext * c = STREAM_CODEC_CONTEXT(st);

  codec=avcodec_find_encoder(c->codec_id);

  if(!codec)
    {
      cout << "error in openVideo : avcodec_find_encoder" << endl;
      return(false);
    }
  /* open the codec */
  if(avcodec_open(c,codec)<0)
    {
      cout << "error in openVideo : avcodec_open ( " << codec->name << ")" << endl;
      return(false);
    }
  videoOutBuffer=NULL;
  if(!(oc->oformat->flags&AVFMT_RAWPICTURE))
  {
    /* allocate output buffer */
    /* XXX: API change will be done */
    videoOutBufferSize=200000;
    videoOutBuffer=(unsigned char*)malloc(videoOutBufferSize);
  }

return(true);
}


static
bool
openAudio(AVFormatContext */*oc*/,
	   AVStream *st,
	   unsigned char * & audioOutBuffer,
	   int & audioOutBufferSize)
{
   AVCodecContext * c = STREAM_CODEC_CONTEXT(st);
   AVCodec *codec;

   /* find the audio encoder */
   codec = avcodec_find_encoder(c->codec_id);
   if (!codec) {
     fprintf(stderr, "codec not found\n");
     exit(1);
   }

   /* open it */
   if (avcodec_open(c, codec) < 0) {
     fprintf(stderr, "could not open audio codec\n");
     exit(1);
   }

   /* allocate output buffer */
   /* XXX: API change will be done */
   audioOutBufferSize = c->frame_size * 2;
   audioOutBuffer = (unsigned char*)malloc(audioOutBufferSize);

   return(true);
}


static
bool
writeVideoFrame(AVFormatContext * oc,
                AVStream * st,
                unsigned int & frameCount,
                unsigned int frameWidth,
                unsigned int frameHeight,
                ArConstRef<MemoryBlock> videoFrameData,
                unsigned char * videoOutBuffer,
                int videoOutBufferSize)
{
int outSize, ret;
AVCodecContext * c=STREAM_CODEC_CONTEXT(st);
AVFrame * picturePtr;
ArRef<MemoryBlock> tmpVideoData=MemoryBlock::NEW();
tmpVideoData->setSize(videoFrameData->getSize());
char * tmpVDPtr=tmpVideoData->accessMemory();
const char * VDPtr=videoFrameData->getMemory();
for(unsigned int line=0;line<frameHeight;line++)
  {
  memcpy(&tmpVDPtr[(frameHeight-1-line)*frameWidth*3],
         &VDPtr[line*frameWidth*3],frameWidth*3);
  }
/* Create the picture */
AVFrame * tmpPicture=allocPicture(PIX_FMT_RGB24,frameWidth,frameHeight,false);
avpicture_fill((AVPicture *)tmpPicture,(unsigned char *)tmpVideoData->getMemory(),
               PIX_FMT_RGB24,frameWidth,frameHeight);
// If the pixel format of the image is different than the encoder, convert it
if(c->pix_fmt!=PIX_FMT_RGB24)
  {
  AVFrame * picture=allocPicture(c->pix_fmt,frameWidth,frameHeight,true);
  img_convert((AVPicture *)picture,c->pix_fmt,
              (AVPicture *)tmpPicture,PIX_FMT_RGB24,
              frameWidth,frameHeight);
  picturePtr=picture;
  av_free(tmpPicture);
  }
else
  {
  picturePtr=tmpPicture;
  }
/* If the frame size is different than the encoder, resample */
if(((int)frameWidth!=c->width)||
   ((int)frameHeight!=c->height))
  {
  ImgReSampleContext * sc=img_resample_init(c->width,c->height,
                                            frameWidth,frameHeight);
  AVFrame * outPut=allocPicture(c->pix_fmt,c->width,c->height,true);
  img_resample(sc,(AVPicture *)outPut,(AVPicture *)picturePtr);
  img_resample_close(sc);
  if(picturePtr->data[0]!=(unsigned char *)tmpVDPtr)
    {
    av_free(picturePtr->data[0]);
    }
  av_free(picturePtr);
  picturePtr=outPut;
  }
if(oc->oformat->flags&AVFMT_RAWPICTURE)
  {
    /* raw video case */
    AVPacket* packet = new AVPacket();
    av_init_packet(packet);
    packet->stream_index = st->index;
    packet->data = (unsigned char*)picturePtr;
    packet->size = sizeof(AVPicture);
    ret=av_write_frame(oc,packet);
  }
else
  {
  /* encode the image */
    outSize=avcodec_encode_video(c,videoOutBuffer,videoOutBufferSize,picturePtr);
    /* if zero size, it means the image was buffered */
    if(outSize!=0)
      {
	/* write the compressed frame in the media file */
	/* XXX: in case of B frames, the pts is not yet valid */
	AVPacket* packet = new AVPacket();
	av_init_packet(packet);
	packet->stream_index = st->index;
	packet->data = videoOutBuffer;
	packet->size = outSize;
	ret=av_write_frame(oc,packet);
      }
    else
    {
      ret=0;
    }
  }
if(picturePtr->data[0]!=(unsigned char *)tmpVDPtr)
  {
  av_free(picturePtr->data[0]);
  }
av_free(picturePtr);
if(ret!=0)
  {
  return(false);
  }
frameCount++;
return(true);
}


static
bool
writeAudioFrame(AVFormatContext *oc,
		AVStream *st,
		ArConstRef<MemoryBlock> audioFrameData,
                uint8_t * audioOutBuffer,
                int audioOutBufferSize)
 {
     int out_size;
     AVCodecContext * c=STREAM_CODEC_CONTEXT(st);

     short* sample = (short *)malloc(c->frame_size * sizeof(short) * c->channels);
     memcpy((char *)sample,(char *)(audioFrameData->getMemory()),c->frame_size*2);

     out_size = avcodec_encode_audio(c,audioOutBuffer,audioOutBufferSize,sample);

     /* write the compressed frame in the media file */
     AVPacket* packet = new AVPacket();
     av_init_packet(packet);
     packet->stream_index = st->index;
     packet->data = audioOutBuffer;
     packet->size = out_size;
     if (av_write_frame(oc,packet) != 0) {
       cout << "Error while writing audio frame" << endl;
       return false;
     }
     return(true);
}

static
void
closeVideo(AVStream * st,
           unsigned char * videoOutBuffer)
{
avcodec_close(STREAM_CODEC_CONTEXT(st));
av_free(videoOutBuffer);
}

static
void
closeAudio(AVStream * st)
{
avcodec_close(STREAM_CODEC_CONTEXT(st));
}

// Streaming fonctions
static
#if (LIBAVCODEC_BUILD >= 4754)
int
#else
void
#endif
stream_write_packet(void * opaque,
                    unsigned char * buf,
                    int buf_size)
{
AbstractOStream * stream=(AbstractOStream *)opaque;
stream->writeData(buf,buf_size);
#if (LIBAVCODEC_BUILD >= 4754)
if(stream->fail())
  {
  return(-EIO);
  }
else
  {
  return(buf_size);
  }
#endif
}

static
int
stream_read_packet(void * opaque,
                   unsigned char * buf,
                   int buf_size)
{
AbstractIStream * stream=(AbstractIStream *)opaque;
buf_size=stream->readData(buf,buf_size);
if(stream->fail())
  {
  return(-EIO);
  }
else
  {
  return(buf_size);
  }
}

static
int
ostream_open(ByteIOContext * s,
             ArRef<AbstractOStream> stream,
             unsigned long buffer_size)
{
unsigned char * buffer;
buffer=(unsigned char *)av_malloc(buffer_size);
if(!buffer)
  {
  return(-ENOMEM);
  }
 if(init_put_byte(s,buffer,buffer_size,true,stream.c_ptr(),
		  stream_read_packet,stream_write_packet,NULL)<0) // no seek!
  {
  av_free(buffer);
  return(-EIO);
  }
s->is_streamed=true;
s->max_packet_size=buffer_size;
return(0);
}

static
int
ostream_close(ByteIOContext * s)
{
if(s->buffer)
  {
  av_free(s->buffer);
  }
memset(s,0,sizeof(ByteIOContext));
return(0);
}

void
AviRecorder_start(AviRecorder * avi)
{

  avi->_aviData=new AviRecorderData;

  // Create output media context
  avi->_aviData->oc=av_alloc_format_context();
  //avi->_aviData->oc=(AVFormatContext *)av_mallocz(sizeof(AVFormatContext));

  if(avi->_stream.valid()) // Detect output format
    {
      avi->_aviData->oc->oformat=guess_stream_format(avi->_format.c_str(),NULL,NULL);
    }
  else
    {
      avi->_aviData->oc->oformat=guess_format(NULL,avi->_fileName.c_str(),NULL);
    }
  if(!avi->_aviData->oc->oformat) // Failed --> will use mpeg
    {
      cout << "failed !!!!!!!!!!!!!!!!!!" << endl;
      avi->_aviData->oc->oformat=guess_format("mpeg",NULL,NULL);
    }

  av_set_parameters(avi->_aviData->oc,NULL);

  // Add audio & video streams
  avi->_aviData->video_st=NULL;
  avi->_aviData->audio_st=NULL;

  if(avi->_aviData->oc->oformat->video_codec!=CODEC_ID_NONE)
    {
      avi->_aviData->video_st=addVideoStream(avi->_aviData->oc,
					     avi->_aviData->oc->oformat->video_codec,
					     avi->_frameRate,avi->_videoBitRate,
					     avi->_frameWidth,avi->_frameHeight);
    }



  if(avi->_useAudio)
    {
      // Add audio stream
      if(avi->_aviData->oc->oformat->audio_codec!=CODEC_ID_NONE)
	{
	  avi->_aviData->audio_st=addAudioStream(avi->_aviData->oc,
						 avi->_aviData->oc->oformat->audio_codec,
						 avi->_audioBitRate,avi->_sampleRate,
						 avi->_channels);
	}
    }

  // Open video & audio codecs
  if(avi->_aviData->video_st)
    {
      openVideo(avi->_aviData->oc,avi->_aviData->video_st,
		avi->_aviData->videoOutBuffer,avi->_aviData->videoOutBufferSize);
    }

  if(avi->_useAudio)
    {
      // Open audio codecs
      if(avi->_aviData->audio_st)
	{
	  if(!openAudio(avi->_aviData->oc,avi->_aviData->audio_st,
			avi->_aviData->audioOutBuffer,avi->_aviData->audioOutBufferSize))
	    cout << "error in openAudio" << endl;
	}
    }

  if(avi->_stream.valid()) // Open ostream
    {
      ostream_open(&avi->_aviData->oc->pb, avi->_stream, IO_BUFFER_SIZE);
    }
  else // Open the file
    {
      url_fopen(&avi->_aviData->oc->pb,avi->_fileName.c_str(),URL_WRONLY);
    }


// Write stream header
 av_write_header(avi->_aviData->oc);

}

void
AviRecorder_stop(AviRecorder * avi)
{
  if(avi->_aviData->video_st) // Close each codec
  {
    closeVideo(avi->_aviData->video_st,avi->_aviData->videoOutBuffer);
  }

  if(avi->_useAudio)
    {
      if(avi->_aviData->audio_st)
	{
	  closeAudio(avi->_aviData->audio_st);
	}
    }

av_write_trailer(avi->_aviData->oc); // Write trailer if any
for(int i=0;i<avi->_aviData->oc->nb_streams;i++) // Free the streams
  {
  av_freep(&avi->_aviData->oc->streams[i]);
  }
if(avi->_stream.valid()) // Close ostream
  {
  ostream_close(&avi->_aviData->oc->pb);
  }
else if(!(avi->_aviData->oc->oformat->flags&AVFMT_NOFILE)) // Close output file
  {
  url_fclose(&avi->_aviData->oc->pb);
  }
// Free the stream
av_free(avi->_aviData->oc);
delete avi->_aviData;
avi->_aviData=(AviRecorderData *)0;
}

bool // success
AviRecorder_addVideoFrame(AviRecorder * avi,
			    unsigned int frameWidth,
			    unsigned int frameHeight,
			    ArConstRef<MemoryBlock> videoFrameData)
{
  // Write interleaved audio and video frames
  return(writeVideoFrame(avi->_aviData->oc,
			 avi->_aviData->video_st,
			 avi->_nbFrames,frameWidth,frameHeight,
			 videoFrameData,
			 avi->_aviData->videoOutBuffer,
			 avi->_aviData->videoOutBufferSize));
}


bool // success
AviRecorder_addAudioFrame(AviRecorder * avi,
			  ArConstRef<MemoryBlock> audioFrameData)
{
  if(audioFrameData->getSize() == 0)
    return false;

  // Write interleaved audio and audio frames
  return(writeAudioFrame(avi->_aviData->oc,
			 avi->_aviData->audio_st,
			 audioFrameData,
			 avi->_aviData->audioOutBuffer,
			 avi->_aviData->audioOutBufferSize));
}

int
AviRecorder_getAudioFrameSize(AviRecorder * avi)
  {
    AVCodecContext * c=STREAM_CODEC_CONTEXT(avi->_aviData->audio_st);
    int frame_size = c->frame_size;
    return frame_size;
  }

StlList<StlString>
AviRecorder_getListOfAvailableCodecs()
{
  AVOutputFormat* of;
  StlList<StlString> codecsList;

  of = first_oformat;

  StlString infoCodecs;
  infoCodecs.append("!!OUTPUT FORMATS!!");
  while(of != NULL)
    {
      infoCodecs.append(of->name);
      infoCodecs.append("  ");
      if (of->extensions)
	infoCodecs.append(of->extensions);
      infoCodecs.append("  ");

      codecsList.push_back(infoCodecs);
      infoCodecs.clear();
      of = of->next;
    }

  AVCodec* codec;

  //codecsList.clear();

  codec = first_avcodec;

  infoCodecs.append("!!CODECS!!");

  while(codec != NULL)
    {
      infoCodecs.append(codec->name);

      infoCodecs.append("   ");

      StlString type;
      switch(codec->type)
	{
	case 0:
	  type = "CODEC_TYPE_VIDEO";
	  break;
	case 1:
	  type = "CODEC_TYPE_AUDIO";
	  break;
	case 2:
	  type = "CODEC_TYPE_DATA";
	  break;
	default:
	  break;
	}

      infoCodecs.append(type);

      infoCodecs.append("   ");

      infoCodecs.append(intToStr(avcodec_get_pix_fmt(codec->name)));

      codecsList.push_back(infoCodecs);
      infoCodecs.clear();
      codec = codec->next;
    }
  return codecsList;
}

//-------- Sound Encoding and Decoding management ----------------------------------------------



void
SoundEncoder_start(SoundEncoder * soundEncoder,const char* codecName,
		   int bitRate, int sampleRate, int nbChannels)
{
  soundEncoder->_soundEncoderData=new SoundEncoderData;

  // init encoder

  /* find the MP2 encoder */
  if(!strcmp(codecName,"MP2"))
    {
      soundEncoder->_soundEncoderData->encoder = avcodec_find_encoder(CODEC_ID_MP2);
    }
  else if(!strcmp(codecName,"VORBIS"))
    soundEncoder->_soundEncoderData->encoder = avcodec_find_encoder(CODEC_ID_VORBIS);

  if (!soundEncoder->_soundEncoderData->encoder) {
    fprintf(stderr, "codec not found\n");
    exit(1);
  }
  else
    cout << "using " << soundEncoder->_soundEncoderData->encoder->name << " encoder" << endl;

  soundEncoder->_soundEncoderData->encoderContext= avcodec_alloc_context();

  /* put sample parameters */
  soundEncoder->_soundEncoderData->encoderContext->bit_rate = bitRate;
  soundEncoder->_soundEncoderData->encoderContext->sample_rate = sampleRate;
  soundEncoder->_soundEncoderData->encoderContext->channels = nbChannels;

  /* open it */
  if (avcodec_open(soundEncoder->_soundEncoderData->encoderContext, soundEncoder->_soundEncoderData->encoder) < 0) {
        fprintf(stderr, "could not open codec\n");
        exit(1);
  }

  // init decoder

    /* find the mpeg audio decoder */
  if(!strcmp(codecName,"MP2"))
    soundEncoder->_soundEncoderData->decoder = avcodec_find_decoder(CODEC_ID_MP2);

  if(!strcmp(codecName,"VORBIS"))
    soundEncoder->_soundEncoderData->decoder = avcodec_find_decoder(CODEC_ID_VORBIS);


    if (!soundEncoder->_soundEncoderData->decoder) {
      fprintf(stderr, "codec not found\n");
      exit(1);
    }
    else
      cout << "using " << soundEncoder->_soundEncoderData->decoder->name << " decoder" << endl;


    soundEncoder->_soundEncoderData->decoderContext = avcodec_alloc_context();

    /* open it */
    if (avcodec_open(soundEncoder->_soundEncoderData->decoderContext, soundEncoder->_soundEncoderData->decoder) < 0) {
      fprintf(stderr, "could not open codec\n");
      exit(1);
    }



}

void
SoundEncoder_stop(SoundEncoder * soundEncoder)
{
  if (soundEncoder->_soundEncoderData->encoderContext)
          avcodec_close(soundEncoder->_soundEncoderData->encoderContext);

  if (soundEncoder->_soundEncoderData->decoderContext)
        avcodec_close(soundEncoder->_soundEncoderData->decoderContext);
}

int
SoundEncoder_encodeSoundSample(SoundEncoder * soundEncoder,
			       unsigned char * encodedSample,
			       short* sample)
{
  return avcodec_encode_audio(soundEncoder->_soundEncoderData->encoderContext,
				      encodedSample,
				      soundEncoder->_soundEncoderData->encoderContext->frame_size*2,
				      sample);
}

int
SoundEncoder_decodeSoundSample(SoundEncoder * soundEncoder,
			       unsigned int *outbuf,
			       int* out_size,
			       unsigned char *inbuf,
			       int in_size)
{
  return avcodec_decode_audio(soundEncoder->_soundEncoderData->decoderContext,
			      (short*)outbuf, out_size,
			      inbuf, in_size);
}

int
SoundEncoder_getFrameSize(SoundEncoder * soundEncoder)
  {
    int frame_size = soundEncoder->_soundEncoderData->encoderContext->frame_size;
    return frame_size;
  }

int
SoundEncoder_getMaxAudioFrameSize()
{
  int maxAudioFrameSize = AVCODEC_MAX_AUDIO_FRAME_SIZE;
  return maxAudioFrameSize;
}

StlList<StlString>
SoundEncoder_getListOfAvailableCodecs()
{
  AVCodec* codec;
  StlList<StlString> codecsList;

  codec = first_avcodec;

  StlString infoCodecs;

  while(codec != NULL)
    {
      infoCodecs.append(codec->name);
      codecsList.push_back(infoCodecs);
      infoCodecs.clear();
      codec = codec->next;
    }

  return codecsList;
}

//-------- Initialization ----------------------------------------------------

extern "C"
bool // success
AudioVideoEncoder_init(const StlVector<StlString> * /* params */)
{
av_register_all();

AviRecorder::_startFnct=&AviRecorder_start;
AviRecorder::_stopFnct=&AviRecorder_stop;
AviRecorder::_addVideoFrameFnct=&AviRecorder_addVideoFrame;
AviRecorder::_addAudioFrameFnct=&AviRecorder_addAudioFrame;
AviRecorder::_getAudioFrameSizeFnct = &AviRecorder_getAudioFrameSize;
AviRecorder::_getListOfAvailableCodecsFnct = &AviRecorder_getListOfAvailableCodecs;
AviRecorder::_initSuccess=true;

SoundEncoder::_startFnct=&SoundEncoder_start;
SoundEncoder::_stopFnct=&SoundEncoder_stop;
SoundEncoder::_encodeSoundSampleFnct = &SoundEncoder_encodeSoundSample;
SoundEncoder::_decodeSoundSampleFnct = &SoundEncoder_decodeSoundSample;
SoundEncoder::_getFrameSizeFnct = &SoundEncoder_getFrameSize;
SoundEncoder::_getMaxAudioFrameSizeFnct = &SoundEncoder_getMaxAudioFrameSize;
SoundEncoder::_getListOfAvailableCodecsFnct = &SoundEncoder_getListOfAvailableCodecs;
SoundEncoder::_initSuccess=true;

return(true);
}

extern "C"
void
AudioVideoEncoder_destroy(void)
{
AviRecorder::_initSuccess=false;
SoundEncoder::_initSuccess=false;
}

} // namespace AReVi

//----------------------------------------------------------------------------
