// LI2/ENIB - AReVi-Contrib - Cg plugin
//----------------------------------------------------------------------------
// file        : ExtContrib/CgWrapper/cgWrapper.cpp
// author(s)   : T. Jourdan
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This plugin uses the Cg API (http://developer.nvidia.com).
//
//  The resulting dynamic library can be loaded at run-time
//  (fails if a dependency problem occurs).
//
//----------------------------------------------------------------------------

#include "AReVi/Contrib/cgContext.h"

#include <Cg/cg.h>
#include <Cg/cgGL.h>

namespace AReVi {

static CGprofile CgWrapper_selectProfile(CgContext::Profile profile) {
  CGprofile cgProf = (CGprofile)-1;
  switch(profile) {
  case CgContext::PROFILE_ARBVP1 :
    cgProf = CG_PROFILE_ARBVP1;
    break;
  case CgContext::PROFILE_ARBFP1 :
    cgProf = CG_PROFILE_ARBFP1;
    break;
  case CgContext::PROFILE_VP20 :
    cgProf = CG_PROFILE_VP20;
    break;
  case CgContext::PROFILE_FP20 :
    cgProf = CG_PROFILE_FP20;
    break;
  case CgContext::PROFILE_VP30 :
    cgProf = CG_PROFILE_VP30;
    break;
  case CgContext::PROFILE_FP30 :
    cgProf = CG_PROFILE_FP30;
    break;
  case CgContext::PROFILE_VP40 :
    cgProf = CG_PROFILE_VP40;
    break;
  case CgContext::PROFILE_FP40 :
    cgProf = CG_PROFILE_FP40;
    break;
  case CgContext::LATEST_VERTEX :
    cgProf = cgGLGetLatestProfile(CG_GL_VERTEX);
    break;
  case CgContext::LATEST_FRAGMENT :
    cgProf = cgGLGetLatestProfile(CG_GL_FRAGMENT);
    break;
  }

  return cgProf;
}

static CGGLenum CgWrapper_selectStateMatrixType(CgContext::StateMatrixType state) {
  CGGLenum stateMatrixType = (CGGLenum)-1;

  switch(state) {
  case CgContext::MODELVIEW_MATRIX :
    stateMatrixType = CG_GL_MODELVIEW_MATRIX;
    break;
  case CgContext::PROJECTION_MATRIX :
    stateMatrixType = CG_GL_PROJECTION_MATRIX;
    break;
  case CgContext::TEXTURE_MATRIX :
    stateMatrixType = CG_GL_TEXTURE_MATRIX;
    break;
  case CgContext::MODELVIEW_PROJECTION_MATRIX :
    stateMatrixType = CG_GL_MODELVIEW_PROJECTION_MATRIX;
    break;
  };

  return stateMatrixType;
}

static CGGLenum CgWrapper_selectTransformMatrixType(CgContext::TransformMatrixType transform) {
  CGGLenum transformMatrixType = (CGGLenum)-1;

  switch(transform) {
  case CgContext::MATRIX_IDENTITY :
    transformMatrixType = CG_GL_MATRIX_IDENTITY;
    break;
  case CgContext::MATRIX_TRANSPOSE :
    transformMatrixType = CG_GL_MATRIX_TRANSPOSE;
    break;
  case CgContext::MATRIX_INVERSE :
    transformMatrixType = CG_GL_MATRIX_INVERSE;
    break;
  case CgContext::MATRIX_INVERSE_TRANSPOSE :
    transformMatrixType = CG_GL_MATRIX_INVERSE_TRANSPOSE;
    break;
  }

  return transformMatrixType;
}

static const char** CgWrapper_makeArgs(const StlVector<StlString> * args) {
  if(args == NULL) {
    return NULL;
  }

  const char** argsArray = new const char*[args->size()];
  for(size_t i = 0; i < args->size(); i++) {
    argsArray[i] = (*args)[i].c_str();
  }
  return argsArray;
}

static
CgContext::CgCTX
CgWrapper_createContext(void) {
  return (CgContext::CgCTX)cgCreateContext();
}

static
void
CgWrapper_destroyContext(CgContext::CgCTX context) {
  cgDestroyContext((CGcontext)context);
}

static
CgContext::Program
CgWrapper_createProgram(CgContext::CgCTX context,
			const StlString& code,
			CgContext::Profile profile,
			const StlString& entry,
			const StlVector<StlString> * args) {
  const char** argsArray = CgWrapper_makeArgs(args);
  CgContext::Program program = cgCreateProgram((CGcontext) context,
				      CG_SOURCE,
				      code.c_str(),
				      CgWrapper_selectProfile(profile),
				      entry.c_str(),
				      argsArray);
  if(argsArray) {
    delete [] argsArray;
  }

  return program;
}



static
CgContext::Program
CgWrapper_createProgramFromFile(CgContext::CgCTX context,
				const StlString& filename,
				CgContext::Profile profile,
				const StlString& entry,
				const StlVector<StlString> * args) {
  const char** argsArray = CgWrapper_makeArgs(args);
  CgContext::Program program = cgCreateProgramFromFile((CGcontext)context,
					      CG_SOURCE,
					      filename.c_str(),
					      CgWrapper_selectProfile(profile),
					      entry.c_str(),
					      argsArray);
  if(argsArray) {
    delete [] argsArray;
  }

  return program;
}

static
void
CgWrapper_destroyProgram(CgContext::Program program) {
  cgDestroyProgram((CGprogram)program);
}

static
void
CgWrapper_enableProfile(CgContext::Profile profile) {
  cgGLEnableProfile(CgWrapper_selectProfile(profile));
}

static
void
CgWrapper_disableProfile(CgContext::Profile profile) {
  cgGLDisableProfile(CgWrapper_selectProfile(profile));
}

static
bool
CgWrapper_isProfileSupported(CgContext::Profile profile) {
  return cgGLIsProfileSupported(CgWrapper_selectProfile(profile));
}

static
void
CgWrapper_loadProgram(CgContext::Program program) {
  cgGLLoadProgram((CGprogram)program);
}

static
void
CgWrapper_bindProgram(CgContext::Program program) {
  cgGLBindProgram((CGprogram)program);
}

static
const char*
CgWrapper_getError(void) {
  return cgGetErrorString(cgGetError());
}

static
void
CgWrapper_setErrorCallback(void (*callback)(void)) {
  cgSetErrorCallback(callback);
}

static
CgContext::Param
CgWrapper_getNamedParameter(CgContext::Program program,
			    const StlString& name) {
  return (CgContext::Param)cgGetNamedParameter((CGprogram)program,
					  name.c_str());
}

static
void
CgWrapper_setParameter1f(CgContext::Param parameter, float x) {
  cgGLSetParameter1f((CGparameter)parameter, x);
}

static
void
CgWrapper_setParameter2f(CgContext::Param parameter, float x, float y) {
  cgGLSetParameter2f((CGparameter)parameter, x, y);
}

static
void
CgWrapper_setParameter3f(CgContext::Param parameter, float x, float y, float z) {
  cgGLSetParameter3f((CGparameter)parameter, x, y, z);
}

static
void
CgWrapper_setParameter4f(CgContext::Param parameter, float x, float y, float z, float w) {
  cgGLSetParameter4f((CGparameter)parameter, x, y, z, w);
}

static
void
CgWrapper_setParameter1fv(CgContext::Param parameter, const float* array) {
  cgGLSetParameter1fv((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter2fv(CgContext::Param parameter, const float* array) {
  cgGLSetParameter2fv((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter3fv(CgContext::Param parameter, const float* array) {
  cgGLSetParameter3fv((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter4fv(CgContext::Param parameter, const float* array) {
  cgGLSetParameter4fv((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter1f(CgContext::Param parameter, float* array) {
  cgGLGetParameter1f((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter2f(CgContext::Param parameter, float* array) {
  cgGLGetParameter2f((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter3f(CgContext::Param parameter, float* array) {
  cgGLGetParameter3f((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter4f(CgContext::Param parameter, float* array) {
  cgGLGetParameter4f((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter1d(CgContext::Param parameter, double x) {
  cgGLSetParameter1d((CGparameter)parameter, x);
}

static
void
CgWrapper_setParameter2d(CgContext::Param parameter, double x, double y) {
  cgGLSetParameter2d((CGparameter)parameter, x, y);
}

static
void
CgWrapper_setParameter3d(CgContext::Param parameter, double x, double y, double z) {
  cgGLSetParameter3d((CGparameter)parameter, x, y, z);
}

static
void
CgWrapper_setParameter4d(CgContext::Param parameter, double x, double y, double z, double w) {
  cgGLSetParameter4d((CGparameter)parameter, x, y, z, w);
}

static
void
CgWrapper_setParameter1dv(CgContext::Param parameter, const double* array) {
  cgGLSetParameter1dv((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter2dv(CgContext::Param parameter, const double* array) {
  cgGLSetParameter2dv((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter3dv(CgContext::Param parameter, const double* array) {
  cgGLSetParameter3dv((CGparameter)parameter, array);
}

static
void
CgWrapper_setParameter4dv(CgContext::Param parameter, const double* array) {
  cgGLSetParameter4dv((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter1d(CgContext::Param parameter, double* array) {
  cgGLGetParameter1d((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter2d(CgContext::Param parameter, double* array) {
  cgGLGetParameter2d((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter3d(CgContext::Param parameter, double* array) {
  cgGLGetParameter3d((CGparameter)parameter, array);
}

static
void
CgWrapper_getParameter4d(CgContext::Param parameter, double* array) {
  cgGLGetParameter4d((CGparameter)parameter, array);
}

static
void
CgWrapper_setMatrixParameterfr(CgContext::Param parameter, const float* matrix) {
  cgGLSetMatrixParameterfr((CGparameter)parameter, matrix);
}

static
void
CgWrapper_setMatrixParameterfc(CgContext::Param parameter, const float* matrix) {
  cgGLSetMatrixParameterfc((CGparameter)parameter, matrix);
}

static
void
CgWrapper_setMatrixParameterdr(CgContext::Param parameter, const double* matrix) {
  cgGLSetMatrixParameterdr((CGparameter)parameter, matrix);
}

static
void
CgWrapper_setMatrixParameterdc(CgContext::Param parameter, const double* matrix) {
  cgGLSetMatrixParameterdc((CGparameter)parameter, matrix);
}

static
void
CgWrapper_getMatrixParameterfr(CgContext::Param parameter, float* matrix) {
  cgGLGetMatrixParameterfr((CGparameter)parameter, matrix);
}

static
void
CgWrapper_getMatrixParameterfc(CgContext::Param parameter, float* matrix) {
  cgGLGetMatrixParameterfc((CGparameter)parameter, matrix);
}

static
void
CgWrapper_getMatrixParameterdr(CgContext::Param parameter, double* matrix) {
  cgGLGetMatrixParameterdr((CGparameter)parameter, matrix);
}

static
void
CgWrapper_getMatrixParameterdc(CgContext::Param parameter, double* matrix) {
  cgGLGetMatrixParameterdc((CGparameter)parameter, matrix);
}

static
void
CgWrapper_setStateMatrixParameter(CgContext::Param parameter,
				  CgContext::StateMatrixType state,
				  CgContext::TransformMatrixType transform) {
  cgGLSetStateMatrixParameter((CGparameter)parameter,
			      CgWrapper_selectStateMatrixType(state),
			      CgWrapper_selectTransformMatrixType(transform));
}

static
void
CgWrapper_setParameterArray1f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const float* array) {
  cgGLSetParameterArray1f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray2f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const float* array) {
  cgGLSetParameterArray2f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray3f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const float* array) {
  cgGLSetParameterArray3f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray4f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const float* array) {
  cgGLSetParameterArray4f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray1f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, float* array) {
  cgGLGetParameterArray1f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray2f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, float* array) {
  cgGLGetParameterArray2f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray3f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, float* array) {
  cgGLGetParameterArray3f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray4f(CgContext::Param parameter, int startIndex,
			      int numberOfElements, float* array) {
  cgGLGetParameterArray4f((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray1d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const double* array) {
  cgGLSetParameterArray1d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray2d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const double* array) {
  cgGLSetParameterArray2d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray3d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const double* array) {
  cgGLSetParameterArray3d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setParameterArray4d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, const double* array) {
  cgGLSetParameterArray4d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray1d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, double* array) {
  cgGLGetParameterArray1d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray2d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, double* array) {
  cgGLGetParameterArray2d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray3d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, double* array) {
  cgGLGetParameterArray3d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_getParameterArray4d(CgContext::Param parameter, int startIndex,
			      int numberOfElements, double* array) {
  cgGLGetParameterArray4d((CGparameter)parameter, startIndex,
			  numberOfElements, array);
}

static
void
CgWrapper_setMatrixParameterArrayfr(CgContext::Param parameter, int startIndex,
				    int numberOfElements, const float* array) {
  cgGLSetMatrixParameterArrayfr((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_setMatrixParameterArrayfc(CgContext::Param parameter, int startIndex,
				    int numberOfElements, const float* array) {
  cgGLSetMatrixParameterArrayfc((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_getMatrixParameterArrayfr(CgContext::Param parameter, int startIndex,
				    int numberOfElements, float* array) {
  cgGLGetMatrixParameterArrayfr((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_getMatrixParameterArrayfc(CgContext::Param parameter, int startIndex,
				    int numberOfElements, float* array) {
  cgGLGetMatrixParameterArrayfc((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_setMatrixParameterArraydr(CgContext::Param parameter, int startIndex,
				    int numberOfElements, const double* array) {
  cgGLSetMatrixParameterArraydr((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_setMatrixParameterArraydc(CgContext::Param parameter, int startIndex,
				    int numberOfElements, const double* array) {
  cgGLSetMatrixParameterArraydc((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_getMatrixParameterArraydr(CgContext::Param parameter, int startIndex,
				    int numberOfElements, double* array) {
  cgGLGetMatrixParameterArraydr((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_getMatrixParameterArraydc(CgContext::Param parameter, int startIndex,
				    int numberOfElements, double* array) {
  cgGLGetMatrixParameterArraydc((CGparameter)parameter, startIndex,
				numberOfElements, array);
}

static
void
CgWrapper_enableClientState(CgContext::Param parameter) {
  cgGLEnableClientState((CGparameter)parameter);
}

static
void
CgWrapper_disableClientState(CgContext::Param parameter) {
  cgGLDisableClientState((CGparameter)parameter);
}

static
void
CgWrapper_setParameterPointer(CgContext::Param parameter, GLint size, GLenum type,
			      GLsizei stride, GLvoid* array) {
  cgGLSetParameterPointer((CGparameter)parameter, size, type,
			  stride, array);
}

static
void
CgWrapper_enableTextureParameter(CgContext::Param parameter) {
  cgGLEnableTextureParameter((CGparameter)parameter);
}

static
void
CgWrapper_disableTextureParameter(CgContext::Param parameter) {
  cgGLDisableTextureParameter((CGparameter)parameter);
}

static
GLuint
CgWrapper_getTextureParameter(CgContext::Param parameter) {
  return cgGLGetTextureParameter((CGparameter)parameter);
}

static
void
CgWrapper_setTextureParameter(CgContext::Param parameter, GLuint textureId) {
  cgGLSetTextureParameter((CGparameter)parameter, textureId);
}

extern "C"
bool
CgWrapper_init(const StlVector<StlString> *) {
  CgContext::_createContext = &CgWrapper_createContext;
  CgContext::_destroyContext = &CgWrapper_destroyContext;
  CgContext::_createProgram = &CgWrapper_createProgram;
  CgContext::_createProgramFromFile = &CgWrapper_createProgramFromFile;
  CgContext::_destroyProgram = &CgWrapper_destroyProgram;
  CgContext::_enableProfile = &CgWrapper_enableProfile;
  CgContext::_disableProfile = &CgWrapper_disableProfile;
  CgContext::_isProfileSupported = &CgWrapper_isProfileSupported;
  CgContext::_loadProgram = &CgWrapper_loadProgram;
  CgContext::_bindProgram = &CgWrapper_bindProgram;
  CgContext::_getError = &CgWrapper_getError;
  CgContext::_setErrorCallback = &CgWrapper_setErrorCallback;
  CgContext::_getNamedParameter = &CgWrapper_getNamedParameter;
  CgContext::_setParameter1f = &CgWrapper_setParameter1f;
  CgContext::_setParameter2f = &CgWrapper_setParameter2f;
  CgContext::_setParameter3f = &CgWrapper_setParameter3f;
  CgContext::_setParameter4f = &CgWrapper_setParameter4f;
  CgContext::_setParameter1fv = &CgWrapper_setParameter1fv;
  CgContext::_setParameter2fv = &CgWrapper_setParameter2fv;
  CgContext::_setParameter3fv = &CgWrapper_setParameter3fv;
  CgContext::_setParameter4fv = &CgWrapper_setParameter4fv;
  CgContext::_getParameter1f = &CgWrapper_getParameter1f;
  CgContext::_getParameter2f = &CgWrapper_getParameter2f;
  CgContext::_getParameter3f = &CgWrapper_getParameter3f;
  CgContext::_getParameter4f = &CgWrapper_getParameter4f;
  CgContext::_setParameter1d = &CgWrapper_setParameter1d;
  CgContext::_setParameter2d = &CgWrapper_setParameter2d;
  CgContext::_setParameter3d = &CgWrapper_setParameter3d;
  CgContext::_setParameter4d = &CgWrapper_setParameter4d;
  CgContext::_setParameter1dv = &CgWrapper_setParameter1dv;
  CgContext::_setParameter2dv = &CgWrapper_setParameter2dv;
  CgContext::_setParameter3dv = &CgWrapper_setParameter3dv;
  CgContext::_setParameter4dv = &CgWrapper_setParameter4dv;
  CgContext::_getParameter1d = &CgWrapper_getParameter1d;
  CgContext::_getParameter2d = &CgWrapper_getParameter2d;
  CgContext::_getParameter3d = &CgWrapper_getParameter3d;
  CgContext::_getParameter4d = &CgWrapper_getParameter4d;
  CgContext::_setMatrixParameterfr = &CgWrapper_setMatrixParameterfr;
  CgContext::_setMatrixParameterfc = &CgWrapper_setMatrixParameterfc;
  CgContext::_setMatrixParameterdr = &CgWrapper_setMatrixParameterdr;
  CgContext::_setMatrixParameterdc = &CgWrapper_setMatrixParameterdc;
  CgContext::_getMatrixParameterfr = &CgWrapper_getMatrixParameterfr;
  CgContext::_getMatrixParameterfc = &CgWrapper_getMatrixParameterfc;
  CgContext::_getMatrixParameterdr = &CgWrapper_getMatrixParameterdr;
  CgContext::_getMatrixParameterdc = &CgWrapper_getMatrixParameterdc;
  CgContext::_setStateMatrixParameter = &CgWrapper_setStateMatrixParameter;
  CgContext::_setParameterArray1f = &CgWrapper_setParameterArray1f;
  CgContext::_setParameterArray2f = &CgWrapper_setParameterArray2f;
  CgContext::_setParameterArray3f = &CgWrapper_setParameterArray3f;
  CgContext::_setParameterArray4f = &CgWrapper_setParameterArray4f;
  CgContext::_getParameterArray1f = &CgWrapper_getParameterArray1f;
  CgContext::_getParameterArray2f = &CgWrapper_getParameterArray2f;
  CgContext::_getParameterArray3f = &CgWrapper_getParameterArray3f;
  CgContext::_getParameterArray4f = &CgWrapper_getParameterArray4f;
  CgContext::_setParameterArray1d = &CgWrapper_setParameterArray1d;
  CgContext::_setParameterArray2d = &CgWrapper_setParameterArray2d;
  CgContext::_setParameterArray3d = &CgWrapper_setParameterArray3d;
  CgContext::_setParameterArray4d = &CgWrapper_setParameterArray4d;
  CgContext::_getParameterArray1d = &CgWrapper_getParameterArray1d;
  CgContext::_getParameterArray2d = &CgWrapper_getParameterArray2d;
  CgContext::_getParameterArray3d = &CgWrapper_getParameterArray3d;
  CgContext::_getParameterArray4d = &CgWrapper_getParameterArray4d;
  CgContext::_setMatrixParameterArrayfr = &CgWrapper_setMatrixParameterArrayfr;
  CgContext::_setMatrixParameterArrayfc = &CgWrapper_setMatrixParameterArrayfc;
  CgContext::_getMatrixParameterArrayfr = &CgWrapper_getMatrixParameterArrayfr;
  CgContext::_getMatrixParameterArrayfc = &CgWrapper_getMatrixParameterArrayfc;
  CgContext::_setMatrixParameterArraydr = &CgWrapper_setMatrixParameterArraydr;
  CgContext::_setMatrixParameterArraydc = &CgWrapper_setMatrixParameterArraydc;
  CgContext::_getMatrixParameterArraydr = &CgWrapper_getMatrixParameterArraydr;
  CgContext::_getMatrixParameterArraydc = &CgWrapper_getMatrixParameterArraydc;
  CgContext::_enableClientState = &CgWrapper_enableClientState;
  CgContext::_disableClientState = &CgWrapper_disableClientState;
  CgContext::_setParameterPointer = &CgWrapper_setParameterPointer;
  CgContext::_enableTextureParameter = &CgWrapper_enableTextureParameter;
  CgContext::_disableTextureParameter = &CgWrapper_disableTextureParameter;
  CgContext::_getTextureParameter = &CgWrapper_getTextureParameter;
  CgContext::_setTextureParameter = &CgWrapper_setTextureParameter;
  CgContext::_initSuccess = true;

  return true;
}

extern "C"
void CgWrapper_destroy(void) {
  CgContext::_initSuccess = false;
}

} // namespace AReVi

//----------------------------------------------------------------------------
