// LI2/ENIB - AReVi-Contrib - Jpeg compression plugin
//----------------------------------------------------------------------------
// file        : ExtContrib/JpegLib/zlibTools.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

#include "AReVi/Utils/abstractStream.h"

#ifdef __cplusplus // for bad jpeglib.h files
extern "C" {
#endif

#include <jpeglib.h>

#ifdef __cplusplus
}
#endif

#include <jerror.h>

namespace AReVi {

//----------------------------------------------------------------------------
// JPEG COMPRESSION TOOLS (Priv)
//----------------------------------------------------------------------------
// This file contains compression data destination routines for the case of
// emitting JPEG data to any AbstractStream.
// IMPORTANT: we assume that char is 8 bit encoded
// this is not a core library module, so it doesn't define JPEG_INTERNALS


// Expanded data destination object for stream output
typedef struct
{
  struct jpeg_destination_mgr pub; // public fields

  //ArPtr<AbstractOStream> stream;
  void *pStream;
  JOCTET * buffer;    //start of buffer
} stream_destination_mgr;

typedef stream_destination_mgr * stream_dst_ptr;

#define OUTPUT_BUF_SIZE  4096 // Arbitrary size to write into stream

/*
 * Initialize destination --- called by jpeg_start_compress
 * before any data is actually written.
 */
METHODDEF(void)
init_destination(j_compress_ptr cinfo)
{
  stream_dst_ptr dest = (stream_dst_ptr) cinfo->dest;

  // Allocate the output buffer --- it will be released when done with image
  dest->buffer = (JOCTET *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
          OUTPUT_BUF_SIZE * sizeof(JOCTET));

  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;
}


/*
 * Empty the output buffer --- called whenever buffer fills up.
 *
 * In typical applications, this should write the entire output buffer
 * (ignoring the current state of next_output_byte & free_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been dumped.
 *
 * In applications that need to be able to suspend compression due to output
 * overrun, a FALSE return indicates that the buffer cannot be emptied now.
 * In this situation, the compressor will return to its caller (possibly with
 * an indication that it has not accepted all the supplied scanlines).  The
 * application should resume compression after it has made more room in the
 * output buffer.  Note that there are substantial restrictions on the use of
 * suspension --- see the documentation.
 *
 * When suspending, the compressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_output_byte & free_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point will be regenerated after resumption, so do not
 * write it out when emptying the buffer externally.
 */

METHODDEF(boolean)
empty_output_buffer (j_compress_ptr cinfo)
{
  stream_dst_ptr dest = (stream_dst_ptr) cinfo->dest;


//  if (!dest->stream->writeData(dest->buffer,OUTPUT_BUF_SIZE))
  if (!(reinterpret_cast<AbstractOStream *>(dest->pStream))->writeData(dest->buffer,OUTPUT_BUF_SIZE))
    ERREXIT(cinfo, JERR_FILE_WRITE);

  // empty buffer
  dest->pub.next_output_byte = dest->buffer;
  dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;

  return TRUE;
}


/*
 * Terminate destination --- called by jpeg_finish_compress
 * after all data has been written.  Usually needs to flush buffer.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF(void)
term_destination(j_compress_ptr cinfo)
{
  stream_dst_ptr dest = (stream_dst_ptr) cinfo->dest;
  size_t datacount = OUTPUT_BUF_SIZE - dest->pub.free_in_buffer;

  // Write any data remaining in the buffer
  if (datacount > 0) {
    //if (!dest->stream->writeData(dest->buffer,datacount))
    if (!(reinterpret_cast<AbstractOStream *>(dest->pStream))->writeData(dest->buffer,datacount))
    ERREXIT(cinfo, JERR_FILE_WRITE);
  }

  //cerr << "flush stream=" << (reinterpret_cast<AbstractOStream *>(dest->pStream)) << endl;
  //dest->stream->flush();
  reinterpret_cast<AbstractOStream *>(dest->pStream)->flush();

  // Make sure we wrote the output file OK
  //if (dest->stream->fail())
  if ((reinterpret_cast<AbstractOStream *>(dest->pStream))->fail())
    ERREXIT(cinfo, JERR_FILE_WRITE);
}


/*
 * Prepare for output to a stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing compression.
 */

GLOBAL(void)
jpeg_stream_dest(j_compress_ptr cinfo, ArRef<AbstractOStream> outStream)
{
  stream_dst_ptr dest;

  /* The destination object is made permanent so that multiple JPEG images
   * can be written to the same file without re-executing jpeg_stream_dest.
   * This makes it dangerous to use this manager and a different destination
   * manager serially with the same JPEG object, because their private object
   * sizes may be different.  Caveat programmer.
   */
  if (cinfo->dest == NULL)
  {
    // first time for this JPEG object?
    cinfo->dest = (struct jpeg_destination_mgr *)
                  (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                  sizeof(stream_destination_mgr));
  }

  dest = (stream_dst_ptr) cinfo->dest;
  dest->pub.init_destination = init_destination;
  dest->pub.empty_output_buffer = empty_output_buffer;
  dest->pub.term_destination = term_destination;
//  dest->stream = outStream;
  dest->pStream = outStream.c_ptr();
}


//----------------------------------------------------------------------------
// JPEG DECOMPRESSION TOOLS (Priv)
//----------------------------------------------------------------------------
// Expanded data source object for stream input
typedef struct
{
  struct jpeg_source_mgr pub; // public fields

  //ArRef<AbstractIStream> stream; // source stream
  void *pStream; // source stream
  JOCTET * buffer; // start of buffer
  boolean start_of_file;	// have we gotten any data yet?
} stream_source_mgr;

typedef stream_source_mgr * stream_src_ptr;

#define INPUT_BUF_SIZE  OUTPUT_BUF_SIZE // can be different


// Initialize source --- called by jpeg_read_header
// before any data is actually read.
METHODDEF(void)
init_source(j_decompress_ptr cinfo)
{
  stream_src_ptr src = (stream_src_ptr) cinfo->src;

  // We reset the empty-input-file flag for each image,
  // but we don't clear the input buffer.
  // This is correct behavior for reading a series of images from one source.
  //
  src->start_of_file = TRUE;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In typical applications, this should read fresh data into the buffer
 * (ignoring the current state of next_input_byte & bytes_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been reloaded.  It is not necessary to
 * fill the buffer entirely, only to obtain at least one more byte.
 *
 * There is no such thing as an EOF return.  If the end of the file has been
 * reached, the routine has a choice of ERREXIT() or inserting fake data into
 * the buffer.  In most cases, generating a warning message and inserting a
 * fake EOI marker is the best course of action --- this will allow the
 * decompressor to output however much of the image is there.  However,
 * the resulting error message is misleading if the real problem is an empty
 * input file, so we handle that case specially.
 *
 * In applications that need to be able to suspend compression due to input
 * not being available yet, a FALSE return indicates that no more data can be
 * obtained right now, but more may be forthcoming later.  In this situation,
 * the decompressor will return to its caller (with an indication of the
 * number of scanlines it has read, if any).  The application should resume
 * decompression after it has loaded more data into the input buffer.  Note
 * that there are substantial restrictions on the use of suspension --- see
 * the documentation.
 *
 * When suspending, the decompressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_input_byte & bytes_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point must be rescanned after resumption, so move it to
 * the front of the buffer rather than discarding it.
 */

METHODDEF(boolean)
fill_input_buffer (j_decompress_ptr cinfo)
{
  stream_src_ptr src = (stream_src_ptr) cinfo->src;
  size_t nbytes;

  //nbytes = src->stream->readData(src->buffer,INPUT_BUF_SIZE);
  nbytes = (reinterpret_cast<AbstractIStream *>(src->pStream))->readData(src->buffer,INPUT_BUF_SIZE);

  if (nbytes <= 0)
  {
    if (src->start_of_file) // Treat empty input file as fatal error
      //ERREXIT(cinfo, JERR_INPUT_EMPTY);
      ERREXIT(cinfo, JWRN_HIT_MARKER);

    WARNMS(cinfo, JWRN_JPEG_EOF);

    // Insert a fake EOI marker
    src->buffer[0] = (JOCTET) 0xFF;
    src->buffer[1] = (JOCTET) JPEG_EOI;
    nbytes = 2;
  }
  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->start_of_file = FALSE;

  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

METHODDEF(void)
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  stream_src_ptr src = (stream_src_ptr) cinfo->src;
  /* Just a dumb implementation for now.  Could use fseek() except
   * it doesn't work on pipes.  Not clear that being smart is worth
   * any trouble anyway --- large skips are infrequent.
   */
  if (num_bytes > 0)
  {
    while (num_bytes > (long) src->pub.bytes_in_buffer)
    {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      (void) fill_input_buffer(cinfo);
      /* note we assume that fill_input_buffer will never return FALSE,
       * so suspension need not be handled.
       */
    }
    src->pub.next_input_byte += (size_t) num_bytes;
    src->pub.bytes_in_buffer -= (size_t) num_bytes;
  }
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */

/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF(void)
term_source (j_decompress_ptr /*cinfo*/)
{
  // no work necessary here
}


/*
 * Prepare for input from a stream stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

GLOBAL(void)
jpeg_stream_src(j_decompress_ptr cinfo, ArRef<AbstractIStream> inStream)
{
  stream_src_ptr src;

  /* The source object and input buffer are made permanent so that a series
   * of JPEG images can be read from the same file by calling jpeg_stream_src
   * only before the first one.  (If we discarded the buffer at the end of
   * one image, we'd likely lose the start of the next one.)
   * This makes it unsafe to use this manager and a different source
   * manager serially with the same JPEG object.  Caveat programmer.
   */
  if (cinfo->src == NULL)
  {
    // first time for this JPEG object?
    cinfo->src = (struct jpeg_source_mgr *)
                 (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                 sizeof(stream_source_mgr));

    src = (stream_src_ptr) cinfo->src;
    src->buffer = (JOCTET *)
        (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
        INPUT_BUF_SIZE * sizeof(JOCTET));
  }

  src = (stream_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
  src->pub.term_source = term_source;
  //src->stream = inStream;
  src->pStream = inStream.c_ptr();
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
  src->pub.next_input_byte = NULL; /* until buffer loaded */
}


} // namespace AReVi

//---End-Of-File--------------------------------------------------------------
