// LI2/ENIB - AReVi-Contrib - sound plugin using ESD
//----------------------------------------------------------------------------
// file        : ExtContrib/SoundESD/soundPlayerESD.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This plugin uses the ESD API (http://www.tux.org/~ricdude/EsounD.html).
//
//  The resulting dynamic library can be loaded at run-time
//  (fails if a dependency problem occurs).
//
//----------------------------------------------------------------------------

#include "AReVi/Contrib/soundPlayer.h"
#include "AReVi/Utils/memoryBlock.h"

#include "AReVi/arSystem.h"
#include "soundPlayerESD.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class SoundPlayerESD
//----------------------------------------------------------------------------
//-------- Construction / Destruction ----------------------------------------

SoundPlayerESD::SoundPlayerESD(SoundPlayer *player, const StlString &/*deviceName*/)
  : _player(player),
    _mixerID(-1),
    _bufferIDRef(0),
    _deviceID(-1),
    _buffers(),
    _currentBuffer(_buffers.end()),
    _currentBufferPosition(0),
    _th(),
    _threadEnded(false),
    _playing(false),
    _paused(false),
    _autoErase(true),
    _pendingBufferData(MemoryBlock::NEW()),
    _pendingBufferSize(0)
{
_pipeID[0]=-1;
_pipeID[1]=-1;

esd_format_t format;
int sampleSize;

switch(_player->getFormat())
  {
  case UtilSound::MONO8:
    format=ESD_BITS8 | ESD_MONO;
    sampleSize=1;
    break;
  case UtilSound::MONO16:
    format=ESD_BITS16 | ESD_MONO;
    sampleSize=2;
    break;
  case UtilSound::STEREO8:
    format=ESD_BITS8 | ESD_STEREO;
    sampleSize=2;
    break;
  case UtilSound::STEREO16:
    format=ESD_BITS16 | ESD_STEREO;
    sampleSize=4;
    break;
  default:
    player->setErrorMessage("Invalid audio format!");
    return;
  }

// openDevice with ESD
format|=ESD_STREAM | ESD_RECORD;

StlString deviceNameId=intToStr(ArSystem::processID())+"::"+_player->getAlias();

_deviceID = esd_play_stream_fallback( format, player->getRate(), NULL, deviceNameId.c_str());

if (_deviceID <0)
  {

  player->setErrorMessage("Fail to open audio device");
  return;
  }
// Retrieve MixerID
int esdID=esd_open_sound(NULL); // get the server

esd_info_t *infos=esd_get_all_info(esdID);
esd_player_info_t *esd_player_info=infos->player_list;

while(esd_player_info)
  {
  if (!strcmp(esd_player_info->name,deviceNameId.c_str()))
    {
    _mixerID=esd_player_info->source_id;
    break;
    }
  esd_player_info=esd_player_info->next;
  }

esd_free_all_info(infos);

// Setup Sound volume
double left,right;
player->getVolume(left,right);
esd_set_stream_pan(esdID,_mixerID,(int)(left*ESD_VOLUME_BASE),(int)(right*ESD_VOLUME_BASE));

esd_close(esdID);

// Setup MinBufferData Size for ESD
_bufferMinSize=(int)(player->getRate()*BUFFER_MIN_SIZE_SCALE*sampleSize);
_bufferMinSize+=_bufferMinSize%sampleSize;
_pendingBufferData->setSize(_bufferMinSize);

// Create pipe

if (pipe(_pipeID))
  {
  player->setErrorMessage("Fail to create pipe");
  cerr << "Pipe creation failed." << endl;
  return;
  }
ArSystem::setBlockingFD(_deviceID,false);
ArSystem::setBlockingFD(_pipeID[0],false);
ArSystem::setBlockingFD(_pipeID[1],false);

// Start Thread
_startThread();

}

SoundPlayerESD::~SoundPlayerESD(void) // not virtual
{
popBuffers(true);

_lock();
int deviceID=_deviceID;
_deviceID=-1; // so thread exit!!
_unlock();

pthread_cond_broadcast(&_cond);
_endThread();

if (deviceID>=0) // remove player from list
  {
  esd_close(deviceID);
  }

_mixerID=-1;
_player=NULL;

if (_pipeID[0]>=0)
  close(_pipeID[0]);

if (_pipeID[1]>=0)
  close(_pipeID[1]);
}


//-------- SoundPlayerESD inspectors -----------------------------------------
int
SoundPlayerESD::getDeviceID(void) const
{
// Device ID is pipeReadMode ID ... Polling!!!
// 0 is Read 1 is Write
return(_pipeID[0]);
}

int
SoundPlayerESD::getMixerID(void) const
{
return(_mixerID);
}


bool
SoundPlayerESD::setVolume(void)
{
if (_mixerID<0)
  return(false);

int id=esd_open_sound(NULL); // get the server

double left,right;
_player->getVolume(left,right);

esd_set_stream_pan(id,_mixerID,(int)(left*ESD_VOLUME_BASE),(int)(right*ESD_VOLUME_BASE));
esd_close(id);

return(true);
}

bool
SoundPlayerESD::setPlay(bool play)
{
_lock();
_playing=play;
_unlock();

if (play) // Broadcast info! (If thread was waiting, wake up it)
  pthread_cond_broadcast(&_cond);

return(true);
}

bool
SoundPlayerESD::isPlaying(void) const
{
bool result;
_lock();
result=_playing&&(!_paused);
_unlock();

return(result);
}


//-------- Buffers control --------------------------------------------------

// Copy buffer into list of fixed size buffers.
int
SoundPlayerESD::pushBuffer(ArConstRef<MemoryBlock> buffer)
{
ArRef<MemoryBlock> copy;

unsigned bufferSize=buffer->getSize();
unsigned bufferPos=0;
int bufferID=_bufferIDRef;

while(bufferSize)
  {
  if (_pendingBufferSize) // have pendingData ... only first loop!
    {
    if (bufferSize<_bufferMinSize-_pendingBufferSize) // still in pendingBuffer
      {
      memcpy(_pendingBufferData->accessMemory()+_pendingBufferSize,buffer->getMemory(),bufferSize);
      _pendingBufferSize+=bufferSize;
      bufferSize=0; // EXIT LOOP
      }
    else
      {
      bufferPos=_bufferMinSize-_pendingBufferSize;
      bufferSize-=bufferPos;

      memcpy(_pendingBufferData->accessMemory()+_pendingBufferSize,buffer->getMemory(),bufferPos);

      _pushBuffer(_pendingBufferData->accessMemory(),_pendingBufferData->getSize());
      _pendingBufferSize=0;
      }
    }
  else
    {
    if (bufferSize<_bufferMinSize) // go in pendingBuffer
      {
      memcpy(_pendingBufferData->accessMemory(),buffer->getMemory()+bufferPos,bufferSize);
      _pendingBufferSize=bufferSize;
      bufferSize=0; // exit loop
      }
    else
      {
      _pushBuffer(buffer->getMemory()+bufferPos,_bufferMinSize);
      bufferPos+=_bufferMinSize;
      bufferSize-=_bufferMinSize;
      }
    }
  ArSystem::yield();
  }
return(bufferID);
}

unsigned
SoundPlayerESD::popBuffers(bool all)
{
if (all)
  {

  _lock(); // Get Access to buffers

  unsigned nb(_buffers.size());
  while(!_buffers.empty())
    {
    free(_buffers.begin()->bufferData);
    _buffers.pop_front();
    }

  _currentBuffer=_buffers.end();
  _currentBufferPosition=0;
  _unlock(); // Release Access to buffers


  _pendingBufferSize=0;
  return(nb);
  }
else
  {
  _lock(); // Get Access to buffers
  unsigned nb(0);

  while(_buffers.begin()!=_currentBuffer && (!_buffers.empty()))
    {
    free(_buffers.begin()->bufferData);
    _buffers.erase(_buffers.begin());
    nb++;
    }

  _currentBuffer=_buffers.begin();
  _currentBufferPosition=0;
  _unlock(); // Release Access to buffers

  return(nb);
  }
}

int
SoundPlayerESD::getCurrentBuffer(void) const
{
int bufferID;

_lock();
if (_currentBuffer!=_buffers.end())
  {
  bufferID=_currentBuffer->bufferID;
  }
else
  bufferID=-1;
_unlock();

return(bufferID);
}


bool
SoundPlayerESD::setCurrentBuffer(int bufferID)
{
_lock();

list<SoundData>::iterator it=_buffers.begin();

while(it!=_buffers.end())
  {
  if (it->bufferID==bufferID)
    break;

  it++;
  }

if (it!=_buffers.end())
  {
  _currentBuffer=it;
  _unlock();
  pthread_cond_broadcast(&_cond);
  return(true);
  }
else // fail
  {
  _unlock();
  return(false);
  }
}


unsigned
SoundPlayerESD::getNbBuffers(void) const
{
_lock(); // Get Access to buffers
unsigned nb(_buffers.size());
_unlock(); // Release Access to buffers

return(nb);
}

//-------- Time control --------------------------------------------------

unsigned int // time in millisecond
SoundPlayerESD::getTime(void) const
{
unsigned int bufferAbsolutePosition=0;
unsigned int soundUnitSize=0;

_lock();
list<SoundData>::const_iterator it;

for (it=_buffers.begin();it!=_currentBuffer;it++)
  {
  bufferAbsolutePosition+=it->bufferSize;
  }
if (it!=_buffers.end())
  bufferAbsolutePosition+=_currentBufferPosition;

_unlock();

switch(_player->getFormat())
  {
  case UtilSound::MONO8:
    soundUnitSize=_player->getRate();
    break;
  case UtilSound::MONO16:
  case UtilSound::STEREO8:
    soundUnitSize=(_player->getRate()<<1);
    break;
  case UtilSound::STEREO16:
    soundUnitSize=(_player->getRate()<<2); // Each sample is 16 (2bytes) * 2 for stereo=> 4
    break;
  default:
    return(0);
  }

return((bufferAbsolutePosition*1000)/soundUnitSize);

}

bool
SoundPlayerESD::setTime(unsigned int newTimePosition)
{
unsigned int bufferAbsolutePosition=0,newAbsolutePosition=0;
unsigned int soundUnitSize=0;

switch(_player->getFormat())
  {
  case UtilSound::MONO8:
    soundUnitSize=(1);
    break;
  case UtilSound::MONO16:
  case UtilSound::STEREO8:
    soundUnitSize=(2);
    break;
  case UtilSound::STEREO16:
    soundUnitSize=(4); // Each sample is 16 (2bytes) * 2 for stereo=> 4
    break;
  default:
    return(false);
  }

newAbsolutePosition=(newTimePosition*soundUnitSize*_player->getRate())/1000;


_lock();
_currentBuffer=_buffers.begin();
while((bufferAbsolutePosition<newAbsolutePosition) && (_currentBuffer!=_buffers.end()))
  {
  bufferAbsolutePosition+=_currentBuffer->bufferSize;
  _currentBuffer++;
  }
_unlock();
pthread_cond_broadcast(&_cond);
return(true);
}

unsigned int // total buffers duration
SoundPlayerESD::getDuration(void) const
{
unsigned int bufferAbsolutePosition=0;
unsigned int soundUnitSize=0;

_lock();
list<SoundData>::const_iterator  it;

for (it=_buffers.begin();it!=_buffers.end();it++)
  {
  bufferAbsolutePosition+=it->bufferSize;
  }
_unlock();

switch(_player->getFormat())
  {
  case UtilSound::MONO8:
    soundUnitSize=_player->getRate();
    break;
  case UtilSound::MONO16:
  case UtilSound::STEREO8:
    soundUnitSize=(_player->getRate()<<1);
    break;
  case UtilSound::STEREO16:
    soundUnitSize=(_player->getRate()<<2); // Each sample is 16 (2bytes) * 2 for stereo=> 4
    break;
  default:
    return(0);
  }

return(bufferAbsolutePosition*1000/soundUnitSize);
}


//-------- Thread -----------------------------------------------------------
void
SoundPlayerESD::_pushBuffer(const char *data,unsigned dataSize)
{
SoundData sd;

sd.bufferData=(char *)malloc(dataSize);
memcpy(sd.bufferData,data,dataSize);

sd.bufferID=_bufferIDRef++;
sd.bufferSize=dataSize;

_lock(); // Get Access to buffers
_buffers.push_back(sd);

if (_currentBuffer==_buffers.end())
  {
  _currentBuffer--;
  }

_unlock(); // Release Access to buffers

if (_playing)// Broadcast info! WakeUp Thread if it was sleeping!
  pthread_cond_broadcast(&_cond);
}

void
SoundPlayerESD::_playingThread(void)
{
// Entering Thread
while(1) // play
  {
  _lock();

  if (_deviceID<0)
      {
      _unlock();
      break;
      }

  if ((_currentBuffer==_buffers.end()) || (!_playing))
    {
    _paused=true;
    _wait(); // Wait until have data to write or another event
    _paused=false;
    if (_deviceID<0)
      {
      _unlock();
      break;
      }
    else
      {
      _unlock();
      continue;
      }
    }

  // else buffer to play!

   int lengthToWrite=(_currentBuffer->bufferSize-_currentBufferPosition);
   int length=ArSystem::partialWriteFD(_deviceID,_currentBuffer->bufferData+_currentBufferPosition,lengthToWrite);

   if (length>0)
    {
     _currentBufferPosition+=length;

     if (_currentBufferPosition>=_currentBuffer->bufferSize) // Next buffer
      {
      _informPosition();
      _currentBufferPosition=0;
      _currentBuffer++;
      }
    }

  _unlock();
  ArSystem::yield();
  }

// Exit Thread
}

void
SoundPlayerESD::_informPosition(void) const
{
if (_pipeID[1]>=0)
  {
  char data=0;
  ArSystem::partialWriteFD(_pipeID[1],&data,sizeof(char));
  }
}


// ----- Thread management
void
SoundPlayerESD::_startThread(void)
{
_threadEnded=false;
int r=pthread_cond_init(&_cond,(pthread_condattr_t *)0);
if(r)
  {
  cerr << ("Cannot create condition --- "+ArSystem::printError(r)) << endl;
  return;
  }

r=pthread_mutex_init(&_mtx,(pthread_mutexattr_t *)0);
if(r)
  {
  cerr << ("Cannot create mutex --- "+ArSystem::printError(r)) << endl;
  pthread_cond_destroy(&_cond);
  return;
  }

r=pthread_create(&_th,(pthread_attr_t *)0, SoundPlayerESD::_threadTask,(void *)this);
if(r)
  {
  cerr << ("Cannot create thread --- "+ArSystem::printError(r))  << endl;
  pthread_cond_destroy(&_cond);
  pthread_mutex_destroy(&_mtx);
  return;
  }

}

void
SoundPlayerESD::_endThread(void)
{
if(!pthread_equal(_th,pthread_self()))
  {
  while(!_threadEnded) // caller waits termination
    {
    sched_yield(); // not join --> many waiting threads allowed
    }

  pthread_cond_destroy(&_cond);
  pthread_mutex_destroy(&_mtx);

  }
}


void *
SoundPlayerESD::_threadTask(void * arg) // Static, Protected
{
SoundPlayerESD *thread(static_cast<SoundPlayerESD *>(arg));
thread->_th=pthread_self();
pthread_detach(thread->_th);
thread->_playingThread();

thread->_threadEnded=true;
return((void *)0);
}

void
SoundPlayerESD::_lock(void) const
{
pthread_mutex_lock((pthread_mutex_t *)&_mtx);
}

void
SoundPlayerESD::_unlock(void) const
{
pthread_mutex_unlock((pthread_mutex_t *)&_mtx);
}

void
SoundPlayerESD::_wait(void) const
{
pthread_cond_wait((pthread_cond_t *)&_cond,(pthread_mutex_t *)&_mtx);
}

} // namespace AReVi

//----------------------------------------------------------------------------
