// LI2/ENIB - AReVi-Contrib - sound plugin using ESD
//----------------------------------------------------------------------------
// file        : ExtContrib/SoundESD/soundRecorderESD.cpp
// author(s)   : E. Cazeaux
//----------------------------------------------------------------------------
// Copyright (C) 2002-2006  LI2/ENIB
//   Laboratoire d'Ingenierie Informatique (LI2)
//   Ecole Nationale d'Ingenieurs de Brest (ENIB)
//   CS 73862 - 29238 BREST Cedex 3 - France
//   Tel: +33(0)298 05 89 89, Fax: +33(0)298 05 89 79, e-mail: li2@enib.fr
//
// This software package is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
// This software package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
// General Public License (file LICENSE-LGPL) for more details.
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//
//  This plugin uses the ESD API (http://www.tux.org/~ricdude/EsounD.html).
//
//  The resulting dynamic library can be loaded at run-time
//  (fails if a dependency problem occurs).
//
//----------------------------------------------------------------------------

#include "AReVi/Contrib/soundRecorder.h"
#include "AReVi/Utils/memoryBlock.h"

#include <unistd.h>

#include "AReVi/arSystem.h"
#include "soundRecorderESD.h"

namespace AReVi {

//----------------------------------------------------------------------------
// Class SoundRecorderESD
//----------------------------------------------------------------------------
//-------- Construction / Destruction ----------------------------------------

SoundRecorderESD::SoundRecorderESD(SoundRecorder *recorder, const StlString &deviceName)
  : _recorder(recorder),
    _esdRecord(false),
    _bufferIDRef(0),
    _deviceID(-1), // Not valid ID
    _buffers(),
    _th(),
    _threadEnded(false),
    _recording(false)
{
_pipeID[0]=-1;
_pipeID[1]=-1;

_esdRecord=(deviceName=="DeviceOutput");

// Create pipe

if (pipe(_pipeID))
  {
  cerr << "Pipe creation failed." << endl;
  return;
  }

// ArSystem::setBlockingFD(_deviceID,false);
ArSystem::setBlockingFD(_pipeID[0],false);
ArSystem::setBlockingFD(_pipeID[1],false);


// Setup temp buffer
_tempBuffer.bufferSize=10000;
_tempBuffer.bufferData=(char *)malloc(sizeof(char)*_tempBuffer.bufferSize);
_tempBuffer.bufferID=-1;

// Start Thread
_startThread();
}

SoundRecorderESD::~SoundRecorderESD(void) // not virtual
{

_lock();
int deviceID=_deviceID;
_deviceID=-2; // so thread exit!!
_unlock();

pthread_cond_broadcast(&_cond);
_endThread();

// Erase all non popped buffers!
while(!_buffers.empty())
    {
    free(_buffers.begin()->bufferData);
    _buffers.erase(_buffers.begin());
    }

if (deviceID>=0) // remove recorder from list
  {
  esd_close(deviceID);
  }

_recorder=NULL;

if (_pipeID[0]>=0)
  close(_pipeID[0]);

if (_pipeID[1]>=0)
  close(_pipeID[1]);
}


//-------- SoundRecorderESD inspectors -----------------------------------------
int
SoundRecorderESD::getDeviceID(void) const
{
// Device ID is pipeReadMode ID ... Polling!!!
// 0 is Read 1 is Write
return(_pipeID[0]);
}

bool
SoundRecorderESD::setRecord(bool record)
{
_lock();
_recording=record;

if (_recording && _deviceID<0)
  {
  esd_format_t format;
  int sampleSize;

  switch(_recorder->getFormat())
    {
    case UtilSound::MONO8:
      format=ESD_BITS8 | ESD_MONO;
      sampleSize=1;
      break;
    case UtilSound::MONO16:
      format=ESD_BITS16 | ESD_MONO;
      sampleSize=2;
      break;
    case UtilSound::STEREO8:
      format=ESD_BITS8 | ESD_STEREO;
      sampleSize=2;
      break;
    case UtilSound::STEREO16:
      format=ESD_BITS16 | ESD_STEREO;
      sampleSize=4;
      break;
    default:
      _recorder->setErrorMessage("Invalid audio format!");
      return(false);
    }

  // openDevice with ESD
  format|=ESD_STREAM | ESD_RECORD;

  StlString deviceNameId=intToStr(ArSystem::processID())+"::"+_recorder->getAlias();

  if (_esdRecord)
    _deviceID = esd_filter_stream( format, _recorder->getRate(), NULL, deviceNameId.c_str());
  else
    _deviceID = esd_record_stream_fallback( format, _recorder->getRate(), NULL, deviceNameId.c_str());

  if (_deviceID <0)
    {
    _recorder->setErrorMessage("Fail to open audio device to record");
    return(false);
    }
  else
    ArSystem::setBlockingFD(_deviceID,false);
  }
else if (!_recording && _deviceID>=0)
  {
  esd_close(_deviceID);
  _deviceID=-1;
  }

_unlock();

if (record) // Broadcast info! (If thread was waiting, wake up it)
  pthread_cond_broadcast(&_cond);

return(true);
}

bool
SoundRecorderESD::isRecording(void) const
{
bool result;
_lock();
result=_recording;
_unlock();

return(result);
}


//-------- Buffers control --------------------------------------------------

// Copy buffer into list of fixed size buffers.
bool
SoundRecorderESD::popPendingBuffer(ArRef<MemoryBlock> bufferOut)
{
bool success(false);
_lock();

if (_buffers.size())
  {
  bufferOut->setSize(_buffers.begin()->bufferSize);
  memcpy(bufferOut->accessMemory(),_buffers.begin()->bufferData,_buffers.begin()->bufferSize);
  free(_buffers.begin()->bufferData);
  _buffers.erase(_buffers.begin());
  success=true;
  }
else
  success=false;

_unlock();

return(success);
}

unsigned
SoundRecorderESD::getNbPendingBuffers(void) const
{
_lock(); // Get Access to buffers
unsigned nb(_buffers.size());
_unlock(); // Release Access to buffers

return(nb);
}

//-------- Thread -----------------------------------------------------------

void
SoundRecorderESD::_recordingThread(void)
{
// Entering Thread
while(1) // record
  {
  _lock();

  if (_deviceID==-2)
      {
      _unlock();
      break;
      }

  if (!_recording)
    {
    _wait(); // Wait until have data te record or another event
    if (_deviceID==-2) // exit cond
      {
      _unlock();
      break;
      }
   else
      {
      _unlock();
      continue;
      }
    }

  // else record buffers!

   int length=ArSystem::partialReadFD(_deviceID,_tempBuffer.bufferData,_tempBuffer.bufferSize);

   if (length>0)
    {
    SoundData buffer;
    buffer.bufferID=_bufferIDRef++;
    buffer.bufferSize=length;
    buffer.bufferData=(char *)malloc(sizeof(char)*buffer.bufferSize);
    memcpy(buffer.bufferData,_tempBuffer.bufferData,buffer.bufferSize);

    _buffers.push_back(buffer);
    _informPendingBuffer();
    }

  _unlock();
   ArSystem::yield();
  }

// Exit Thread
}

void
SoundRecorderESD::_informPendingBuffer(void) const
{
if (_pipeID[1]>=0)
  {
  char data=0;
  ArSystem::partialWriteFD(_pipeID[1],&data,sizeof(char));
  }
}


// ----- Thread management
void
SoundRecorderESD::_startThread(void)
{
_threadEnded=false;
int r=pthread_cond_init(&_cond,(pthread_condattr_t *)0);
if(r)
  {
  cerr << ("Cannot create condition --- "+ArSystem::printError(r)) << endl;
  return;
  }

r=pthread_mutex_init(&_mtx,(pthread_mutexattr_t *)0);
if(r)
  {
  cerr << ("Cannot create mutex --- "+ArSystem::printError(r)) << endl;
  pthread_cond_destroy(&_cond);
  return;
  }

r=pthread_create(&_th,(pthread_attr_t *)0, SoundRecorderESD::_threadTask,(void *)this);
if(r)
  {
  cerr << ("Cannot create thread --- "+ArSystem::printError(r))  << endl;
  pthread_cond_destroy(&_cond);
  pthread_mutex_destroy(&_mtx);
  return;
  }

}

void
SoundRecorderESD::_endThread(void)
{
if(!pthread_equal(_th,pthread_self()))
  {
  while(!_threadEnded) // caller waits termination
    {
    sched_yield(); // not join --> many waiting threads allowed
    }

  pthread_cond_destroy(&_cond);
  pthread_mutex_destroy(&_mtx);

  }
}


void *
SoundRecorderESD::_threadTask(void * arg) // Static, Protected
{
SoundRecorderESD *thread(static_cast<SoundRecorderESD *>(arg));
thread->_th=pthread_self();
pthread_detach(thread->_th);
thread->_recordingThread();

thread->_threadEnded=true;
return((void *)0);
}

void
SoundRecorderESD::_lock(void) const
{
pthread_mutex_lock((pthread_mutex_t *)&_mtx);
}

void
SoundRecorderESD::_unlock(void) const
{
pthread_mutex_unlock((pthread_mutex_t *)&_mtx);
}

void
SoundRecorderESD::_wait(void) const
{
pthread_cond_wait((pthread_cond_t *)&_cond,(pthread_mutex_t *)&_mtx);
}

} // namespace AReVi

//----------------------------------------------------------------------------
